# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
#
#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import webob
from webob.exc import HTTPError
import six
from oslo_log import log as logging
import oslo_messaging as messaging
from cinder import exception
from cinder import volume as vol_api
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.compute import nova as nova_interface
from powervc_cinder import _
from powervc_cinder.volume import rpcapi_product as vol_rpcapi

LOG = logging.getLogger(__name__)
authorize = extensions.extension_authorizer('volume', 'extend-volume')

EXTEND_VOLUME_FAILURE_KEY = "Extending volume Failure description"
EXTEND_VOLUME_NOIMPLEMENTED = "not implemented"


class ExtendVolumeControllerExtension(wsgi.Controller):

    def __init__(self, *args, **kwargs):
        super(ExtendVolumeControllerExtension, self).__init__(*args,
                                                              **kwargs)
        self.volume_rpcapi = vol_rpcapi.VolumeAPIProduct()
        self.volume_api = vol_api.API()

    @wsgi.action('ibm-extend')
    def ibm_extend_volume(self, req, id, body):
        context = req.environ['cinder.context']
        authorize(context, action='ibm-extend')

        try:
            volume = self.volume_api.get(context, id)
        except exception.VolumeNotFound as ex:
            msg = ex.message % {'volume_id': id}
            raise webob.exc.HTTPNotFound(explanation=six.text_type(msg))

        if req.path.strip('/').startswith('v1'):
            try:
                val = int(body['ibm-extend']['new_size'])
            except ValueError:
                msg = (_("New volume size must be specified as an "
                         "integer."))
                raise webob.exc.HTTPBadRequest(
                    explanation=six.text_type(msg))
        else:
            val = body['ibm-extend']['new_size']
            if not isinstance(val, int):
                msg = (_("New volume size must be specified as an "
                         "integer."))
                raise webob.exc.HTTPBadRequest(
                    explanation=six.text_type(msg))

        size_increase = val - volume['size']
        if size_increase < 0:
            msg = (_("New size for volume extend must be greater "
                     "than current size. (current: %(size)s, "
                     "extended: %(new_size)s)") %
                   dict(new_size=val, size=volume['size']))
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(msg))
        if size_increase == 0:
            if req.path.strip('/').startswith('v1'):
                return webob.Response(status_int=200)
            else:
                return webob.Response(status_int=204)
        try:
            nova_api = nova_interface.API()
            server_id = volume.get('volume_attachment')[0].get('instance_uuid')
            server = nova_api.get_server(context, server_id)
            os_distro = server.os_distro
            if os_distro == 'ibmi':
                msg = (_("New volume size is not allowed for VM type 'ibm-i'"))
                raise webob.exc.HTTPBadRequest(
                    explanation=six.text_type(msg))
        except HTTPError as http_ex:
            raise http_ex
        except Exception as e:
            # volume is not attached, continue
            LOG.debug(e)

        try:
            self.volume_rpcapi.ibm_extend_volume(context, volume, val)
        except Exception as ex:
            if isinstance(ex, messaging.RemoteError):
                exc_type = ex.exc_type
                exc_value = ex.value
                if exc_type == 'PVCExpendvdiskCapacityException':
                    # request is too big to be satisfied by storage.
                    msg = (_("PVCExpendvdiskCapacityException: "
                             "%(error)s") % dict(error=exc_value))
                    raise webob.exc.HTTPInsufficientStorage(
                        explanation=six.text_type(msg))
                elif exc_type == 'PVCHMCVersionIncompatible':
                    # HMC is not of valid version
                    msg = (_("PVCHMCVersionInvalid: "
                             "%(error)s") % dict(error=exc_value))
                    raise webob.exc.HTTPConflict(
                        explanation=six.text_type(msg))
                elif exc_type == 'SVCVdiskNotFoundException':
                    # volume has been deleted. Need to translate
                    # the SVCVdiskNotFoundException's vdisk id
                    # to some one more meaningful message for caller.
                    msg = (_("SVCVdiskNotFoundException: The volume resize"
                             " request failed since the backing disk does "
                             "not exist on storage %(host)s. Error "
                             "details: %(error)s") %
                           dict(host=volume.get('host'), error=exc_value))
                    raise webob.exc.HTTPNotFound(
                        explanation=six.text_type(msg))
                elif exc_type == 'VolumeSizeExceedsAvailableQuota':
                    msg = (_("VolumeSizeExceedsAvailableQuota: %(error)s") %
                           dict(error=exc_value))
                    raise webob.exc.HTTPConflict(
                        explanation=six.text_type(msg))
                elif exc_type == 'PVCExpendvdiskFCMapException':
                    msg = (_("PVCExpendvdiskFCMapException: %(error)s") %
                           dict(error=exc_value))
                    raise webob.exc.HTTPConflict(
                        explanation=six.text_type(msg))
                elif exc_type == 'NotImplementedError':
                    msg = (_("NotImplementedError: Extend volume is not "
                             "supported for boot volume on "
                             "storage host %(host)s") %
                           dict(host=volume.get('host')))
                    raise webob.exc.HTTPConflict(
                        explanation=six.text_type(msg))
                else:
                    # other volume errors
                    msg = (_("CinderException: %(exc)s: %(error)s") %
                           dict(exc=exc_type, error=ex.msg))
                    raise webob.exc.HTTPInternalServerError(
                        explanation=six.text_type(msg))
            elif ex.__class__.__name__.find('Timeout') >= 0:
                # getting RPC timeout error.
                msg = (_("RPCTimeout: Volume Service on host %(host)s is "
                         "down %(error)s") %
                       dict(host=volume.get('host'), error=ex))
                raise webob.exc.HTTPInternalServerError(
                    explanation=six.text_type(msg))
            else:
                msg = (_("RPCException: Cinder API error: "
                         "%s") % ex)
                raise webob.exc.HTTPInternalServerError(
                    explanation=six.text_type(msg))

        if req.path.strip('/').startswith('v1'):
            return webob.Response(status_int=200)
        else:
            return webob.Response(status_int=204)


class Extend_volume(extensions.ExtensionDescriptor):
    """Capability to Extend/ resize a volume"""
    name = "extend volume"
    alias = "ibm-extend-volume"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_extend-volume/api/v1.1"
    updated = "2013-08-14T21:00:00-06:00"

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hosts Resource"""
        extension_list = []
        extension = extensions.ControllerExtension(
            self, 'volumes', ExtendVolumeControllerExtension())
        extension_list.append(extension)
        return extension_list
