# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import webob
import six
from oslo_log import log as logging
from cinder import exception
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.volume import utils as volume_utils
from cinder import db

LOG = logging.getLogger(__name__)


class VolumeMetadataControllerExtension(wsgi.Controller):
    """Main Controller Class to extend the volumes API for Registration
    Below methods are overridden to add the notification messages
    for GUI to handle the refresh when metadata properties are changed
    """

    @wsgi.extends
    def update(self, req, volume_id, id, body):
        """Extend Rest PUT Method to Add notifications"""
        # Note: Since we are just adding notifications here, the policy
        #       check that the community code has in place applies and we
        #       do not need to do additional authorization. See the
        #       @wrap_check_policy decorator on update_volume_metadata().
        context = req.environ['cinder.context']
        volume = self._check_volume(context, id)
        LOG.info("Updating the Volume metadata properties..")
        volume_utils.notify_about_volume_usage(context, volume,
                                               'metadata.update.start')
        resp_obj = (yield)
        self._yield_our_response(req, resp_obj)
        volume_utils.notify_about_volume_usage(context, volume,
                                               'metadata.update.end')

    @wsgi.extends
    def update_all(self, req, volume_id, body):
        """Extend Rest PUT Method to Add notifications"""
        # Note: Since we are just adding notifications here, the policy
        #       check that the community code has in place applies and we
        #       do not need to do additional authorization. See the
        #       @wrap_check_policy decorator on update_volume_metadata().
        context = req.environ['cinder.context']
        volume = self._check_volume(context, volume_id)
        LOG.info("Updating the Volume metadata properties..")
        volume_utils.notify_about_volume_usage(context, volume,
                                               'metadata.update.start')
        resp_obj = (yield)
        self._yield_our_response(req, resp_obj)
        volume_utils.notify_about_volume_usage(context, volume,
                                               'metadata.update.end')

    def _check_volume(self, context, volume_id):
        """
        Checks for the existence of the specified by volume by attempting to
        retrieve it's details and raising an exception if the volume was not
        found.
        """
        try:
            return db.volume_get(context, volume_id)
        except exception.NotFound as ex:
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))

    def _yield_our_response(self, req, resp_obj):
        """Helper method to yield the response until the parent is done"""
        pass


class Volume_Metadata(extensions.ExtensionDescriptor):
    """Capability to add Notification while Metadata
    properties of a Volume are updated"""
    name = "Update the Volume Metadata Properties"
    alias = "volume-metadata"
    namespace = "http://docs.openstack.org/volume/ext/" + \
                "cinder/api/v2.0"
    updated = "2015-02-17T21:00:00-06:00"

    def get_controller_extensions(self):
        """Provide a Controller Extension to the volumes Resource"""
        extension1 = extensions.ControllerExtension(
            self, 'volume_metadata', VolumeMetadataControllerExtension())
        return [extension1]
