# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Update the MultiAttach Flag"""

import webob.exc
from oslo_log import log as logging
from oslo_utils import strutils

from cinder import db
from cinder.api import extensions
from cinder.api.openstack import wsgi

from powervc_cinder import _

LOG = logging.getLogger(__name__)


class VolumeMultiattachControllerExtension(wsgi.Controller):
    """Main Controller Class to extend the volumes API for Registration"""

    @wsgi.extends
    def update(self, req, id, body):
        """Extend Rest PUT Method to Update a Volume Multiattach property"""
        context = req.environ['cinder.context']
        multiattach = None
        update = False
        if 'multiattach' in body['volume']:
            multiattach = body.get('volume', {}).get('multiattach')
        # Only something to do if they provided the MultiAttach flag
        if multiattach is not None:
            multiattach = strutils.bool_from_string(multiattach, default=None)
            # Make sure it is a boolean or a string form of a boolean
            if multiattach is None or not isinstance(multiattach, bool):
                msg = _("The multiattach attribute must be a boolean")
                raise webob.exc.HTTPBadRequest(explanation=msg)
            update = True
            volume = db.volume_get(context, id)
            if volume['status'] == 'error':
                msg = _("Multiattach Property cannot be set to false "
                        "if Volume is in error state")
                raise webob.exc.HTTPBadRequest(explanation=msg)
            if update:
                # Update the MultiAttach Database Property for the Volume
                db.volume_update(context, id,
                                 dict(multiattach=multiattach))
        # Yield the adding of the response until the parent was called
        resp_obj = (yield)
        self._add_our_response(req, resp_obj, update, multiattach)

    def _add_our_response(self, req, resp_obj, update, multiattach):
        """Internal Helper Method to add our response for the Update"""
        if 'volume' in resp_obj.obj and update:
            resp_obj.obj['volume']['multiattach'] = multiattach


class Volume_Multiattach(extensions.ExtensionDescriptor):
    """Capability to Update MultiAttach on a Volume"""
    name = "Set the Volume Multi-attach Property"
    alias = "volume-multiattach"
    namespace = "http://docs.openstack.org/volume/ext/" + \
                "volume_multiattach/api/v1.1"
    updated = "2015-01-27T21:00:00-06:00"

    def get_controller_extensions(self):
        """Provide a Controller Extension to the volumes Resource"""
        extension1 = extensions.ControllerExtension(
            self, 'volumes', VolumeMultiattachControllerExtension())
        return [extension1]
