# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from webob import exc
from oslo_log import log as logging

from cinder.api import extensions
from cinder.api.openstack import wsgi
from powervc_cinder.volume import rpcapi_product as vol_rpcapi
from cinder.db import api as db
from powervc_cinder.db import api as powervc_db


LOG = logging.getLogger(__name__)
authorizeVol = extensions.extension_authorizer('volume', 'volume_pools')
authorizeHostVol = extensions.extension_authorizer('volume', 'host_vol_pools')


class VolumePoolsController(wsgi.Controller):
    """Main Controller Extension Class to provide the
       volume-pools API"""
    def __init__(self, *args, **kwargs):
        super(VolumePoolsController, self).__init__(*args, **kwargs)
        self.volume_rpcapi = vol_rpcapi.VolumeAPIProduct()

    @wsgi.extends
    def show(self, req, id):
        context = req.environ['cinder.context']
        authorizeVol(context, action='show')
        metadata = {}
        try:
            if not id or len(id) == 0:
                raise exc.HTTPNotFound()
            hosts = powervc_db.service_get_all_by_host(context, id)
            # Essentially not host matches that id, throw 404
            if len(hosts) == 0:
                raise exc.HTTPNotFound()
            host_id = hosts[0].get('host')
            metadata = self.volume_rpcapi.get_storage_metadata(
                context, host_id)
            # Remove the metadata attribute of compression_supported for v2
            if not req.path.strip('/').startswith('v1'):
                metadata.pop('compression_supported', None)
        except Exception as e:
            LOG.debug(e)
        return {host_id: metadata}

    @wsgi.extends
    def index(self, req):
        metadata_list = {}
        metadata = {}
        context = req.environ['cinder.context']
        authorizeVol(context, action='index')
        hosts = db.service_get_all(context)
        # If no hosts have been added, the only host is
        # the scheduler
        for host in hosts:
            if host.topic == 'cinder-volume':
                host_id = host.get('host')
                try:
                    metadata = (self.volume_rpcapi.get_storage_metadata(
                        context, host_id))
                    # Remove the metadata attribute
                    # of compression_supported for v2
                    if not req.path.strip('/').startswith('v1'):
                        metadata.pop('compression_supported', None)
                except Exception as e:
                    LOG.debug(e)
                metadata_list[host_id] = metadata
        return metadata_list


class HostVolumePoolsController(wsgi.Controller):
    """Main Controller Class to extend the v2 os-hosts API for volume-pools"""
    def __init__(self, *args, **kwargs):
        super(HostVolumePoolsController, self).__init__(*args, **kwargs)
        self.volume_rpcapi = vol_rpcapi.VolumeAPIProduct()

    @wsgi.extends
    def index(self, req, host_name_id):
        context = req.environ['cinder.context']
        authorizeHostVol(context, action='index')
        metadata = {}
        if not host_name_id or len(host_name_id) == 0:
            raise exc.HTTPNotFound()
        try:
            hosts = powervc_db.service_get_all_by_host(context, host_name_id)
            # Essentially not host matches that host_name_id, throw 404
            if len(hosts) == 0:
                raise exc.HTTPNotFound()
            host_id = hosts[0].get('host')
            metadata = self.volume_rpcapi.get_storage_metadata(
                context, host_id)
            # Remove the metadata attribute of compression_supported for v2
            if not req.path.strip('/').startswith('v1'):
                metadata.pop('compression_supported', None)
        except Exception as e:
            LOG.debug(e)
        return metadata


class Volume_pools(extensions.ExtensionDescriptor):
    """
    API Extension to provide metadata for a storage.
    """
    name = "Volume Pools"
    alias = "volume_pools"
    namespace = "http://www.ibm.com/openstack/cinder/contrib/powervc/v1.0"
    updated = ""

    def get_resources(self):
        resources = []

        # New API resource. This resource supports
        # v2/{tenant_id}/os-hosts/{id}/volume
        resource1 = extensions.ResourceExtension(
            'volume-pools', controller=HostVolumePoolsController(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resources.append(resource1)

        # New API resource. This resource supports
        # v2/{tenant_id}/volume-pools
        resource2 = extensions.ResourceExtension(
            'volume-pools', controller=VolumePoolsController())
        resources.append(resource2)

        return resources
