# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import webob
import six
from oslo_log import log as logging

from cinder import exception
from cinder.api import extensions
from cinder.api.openstack import wsgi
from powervc_cinder import _
from cinder import rpc
from powervc_cinder.db import api as powervc_db_api
from cinder.db import api as db_api

LOG = logging.getLogger(__name__)
authorize = extensions.extension_authorizer('volume', 'restricted_metadata')


class VolumeRestrictedMetadataController(wsgi.Controller):
    """Provides access to key/value metadata associated with a Volume, but
    accessed through an alternative API and thus can be access controlled.
    """

    def _get_metadata(self, context, volume_id):
        """
        Returns a dictionary of key/value pairs containing the restricted
        metadata for the specified volume, e.g.:
        { key1: value1, key2: value2 }

        Will return an empty dictionary for nonexistent volumes - no check for
        the validity of the passed-in volume is performed here.
        """
        metadata = powervc_db_api.volume_restricted_metadata_get(context,
                                                                 volume_id)
        return metadata

    def _check_volume(self, context, volume_id):
        """
        Checks for the existence of the specified by volume by attempting to
        retrieve it's details and raising an exception if the volume was not
        found.
        """
        try:
            db_api.volume_get(context, volume_id)
        except exception.NotFound as ex:
            raise webob.exc.HTTPNotFound(explanation=six.text_type(ex))

    def index(self, req, volume_id):
        """Returns the list of restricted metadata for a given volume.
        e.g. GET to /volumes/{volume_id}/restricted_metadata
             returns
              {"metadata": {"key2": "value2", "key1": "value1"}}
        """
        context = req.environ['cinder.context']
        authorize(context, action='index')
        self._check_volume(context, volume_id)
        metadata = self._get_metadata(context, volume_id)
        if not metadata:
            raise webob.exc.HTTPNotFound()
        else:
            return {'metadata': metadata}

    def create(self, req, volume_id, body=None):
        """Creates new metadata items, echoing the passed-in items back in the
        response body.

        e.g. POST to /volumes/{volume_id}/restricted_metadata
             with body
              {"metadata": {"key2": "value2", "key1": "value1"}}
            returns
              {"metadata": {"key2": "value2", "key1": "value1"}}
        """
        context = req.environ['cinder.context']
        authorize(context, action='create')

        if not self.is_valid_body(body, 'metadata'):
            raise webob.exc.HTTPBadRequest()

        self._check_volume(context, volume_id)

        metadata = body['metadata']
        powervc_db_api.volume_restricted_metadata_update_or_create(context,
                                                                   volume_id,
                                                                   metadata)

        notifier_info = dict(volume_id=volume_id, metadata=metadata)
        anotifier = rpc.get_notifier(publisher_id='volumeRestrictedMetadata')
        anotifier.info(
            context, 'volume_restricted_metadata.create', notifier_info)
        return body

    def update(self, req, volume_id, id, body=None):
        """Modifies or creates a specific metadata item.

        e.g. PUT to /volumes/{volume_id}/restricted_metadata/key2
             with body
              {"key2" : "new_value"}
             returns
              {"key2" : "new_value"}

        If the specified key (key2 in the example above) exists, its value is
        updated.  If it does not exist, it is created.

        Only a single value can be updated with this function.  To set multiple
        values, perform a POST on restricted_metadata
        """
        context = req.environ['cinder.context']
        authorize(context, action='update')
        if not body:
            expl = _('Request body empty')
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(expl))
        self._check_volume(context, volume_id)
        if id not in body:
            expl = _('Request body and URI mismatch')
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(expl))
        if len(body) > 1:
            expl = _('Request body contains too many items')
            raise webob.exc.HTTPBadRequest(explanation=six.text_type(expl))
        powervc_db_api.volume_restricted_metadata_update_or_create(context,
                                                                   volume_id,
                                                                   body)

        notifier_info = dict(volume_id=volume_id, metadata=body)
        anotifier = rpc.get_notifier(publisher_id='volumeRestrictedMetadata')
        anotifier.info(
            context, 'volume_restricted_metadata.update', notifier_info)
        return body

    def show(self, req, volume_id, id):
        """Displays a single metadata item.

        e.g. GET to /volumes/{volume_id}/restricted_metadata/existing_key
             returns
             {"existing_key" : "value"}

             DELETE to /volumes/{volume_id}/restricted_metadata/nonexistent_key
             returns
             404 Not Found
        """
        context = req.environ['cinder.context']
        authorize(context, action='show')
        self._check_volume(context, volume_id)
        metadata = self._get_metadata(context, volume_id)
        if id in metadata:
            return {id: metadata[id]}
        else:
            raise webob.exc.HTTPNotFound()

    def delete(self, req, volume_id, id):
        """Deletes an existing restricted volume metadata item.

        e.g. DELETE to /volumes/{volume_id}/restricted_metadata/existing_key
        returns 202 Accepted

        e.g. DELETE to /volumes/{volume_id}/restricted_metadata/nonexistent_key
        returns 404 Not Found
        """
        context = req.environ['cinder.context']
        self._check_volume(context, volume_id)
        authorize(context, action='delete')

        try:
            powervc_db_api.volume_restricted_metadata_delete(context,
                                                             volume_id,
                                                             id)
        except exception.VolumeMetadataNotFound as error:
            raise webob.exc.HTTPNotFound(explanation=six.text_type(error))

        notifier_info = dict(volume_id=volume_id, id=id)
        anotifier = rpc.get_notifier(publisher_id='volumeRestrictedMetadata')
        anotifier.info(
            context, 'volume_restricted_metadata.delete', notifier_info)
        return webob.Response(status_int=202)


class VolRestrictedMetaCtrlOld(VolumeRestrictedMetadataController):
    """Provides v1 support for restricted_metadata form"""

    def index(self, req, volume_id):
        if not req.path.strip('/').startswith('v1'):
            raise webob.exc.HTTPNotFound()
        return super(VolRestrictedMetaCtrlOld, self).index(req, volume_id)

    def create(self, req, volume_id, body=None):
        if not req.path.strip('/').startswith('v1'):
            raise webob.exc.HTTPNotFound()
        return super(VolRestrictedMetaCtrlOld, self).create(req, volume_id,
                                                            body=body)

    def update(self, req, volume_id, id, body=None):
        if not req.path.strip('/').startswith('v1'):
            raise webob.exc.HTTPNotFound()
        return super(VolRestrictedMetaCtrlOld, self).update(req, volume_id,
                                                            id, body=body)

    def show(self, req, volume_id, id):
        if not req.path.strip('/').startswith('v1'):
            raise webob.exc.HTTPNotFound()
        return super(VolRestrictedMetaCtrlOld, self).show(req, volume_id, id)

    def delete(self, req, volume_id, id):
        if not req.path.strip('/').startswith('v1'):
            raise webob.exc.HTTPNotFound()
        return super(VolRestrictedMetaCtrlOld, self).delete(req, volume_id, id)


class Volume_restricted_metadata(extensions.ExtensionDescriptor):
    """
    API Extension to provide restricted metadata for Cinder Volumes.
    """
    name = "Volume restricted metadata"
    alias = "volume_restricted_metadata"
    namespace = "http://www.ibm.com/openstack/cinder/contrib/powervc/v1.0"
    updated = ""

    def get_resources(self):
        VolumeRestrictedMetadataController()
        resources = []

        res = extensions.\
            ResourceExtension('restricted-metadata',
                              VolumeRestrictedMetadataController(),
                              parent=dict(member_name='volume',
                                          collection_name='volumes')
                              )
        resources.append(res)
        # NOTE (edmondsw): restricted_metadata left for backward compatibility
        # with v1, but no longer supported with v2.
        res2 = extensions.\
            ResourceExtension('restricted_metadata',
                              VolRestrictedMetaCtrlOld(),
                              parent=dict(member_name='volume',
                                          collection_name='volumes')
                              )
        resources.append(res2)
        return resources
