# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from datetime import datetime
import os
import pwd
from sqlalchemy import MetaData, Table
from sqlalchemy.sql import bindparam, expression, and_

from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config.data_utils import V120, V121, V122, V123, V130, V999,\
    V1302
from powervc_cinder.config import storage_config

CINDER_CONF = '/etc/cinder/cinder.conf'
FABRIC_CONF = '/etc/cinder/fabrics.conf'
PASTE_CONF = '/etc/cinder/api-paste.ini'
POWERVC_MANAGE = '/opt/ibm/powervc/bin/powervc-manage'


######################################################
#         Cinder Data Manager Definition             #
######################################################
class CinderDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'cinder'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db2', 'rabbitmq', 'httpd', 'keystone', 'glance']

    ######################################################
    ####  Data Manager - powervc-config cmd parser  ######
    ######################################################
    def invoke_config_parser(self, subparser):
        """Invokes the parser for each component for powervc-config cmd"""
        cinder_parser = super(
            CinderDataManager,
            self).invoke_config_parser(subparser)
        config = self._get_added_conf_properties()
        storage_config.setup_parser(
            cinder_parser,
            config,
            self.get_component_name())

    ###############################################
    #      Data Manager - Upgrade Interface       #
    ###############################################
    def upgrade_conf_files(self, version):
        """Performs the upgrade of the Component's CONF Files"""
        super(CinderDataManager, self).upgrade_conf_files(version)
        # If this was ever 1.2.1 we need to upgrade the fabrics.conf file
        self.__upgrade_fabrics_conf()

    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        # If coming from before 1.3, need to adjust rsakeys permissions
        if utils.compare_version(version, V130) < 0:
            if os.path.exists('/etc/cinder/rsakeys'):
                pw_info = pwd.getpwnam('cinder')
                os.chown('/etc/cinder/rsakeys', pw_info.pw_uid, pw_info.pw_gid)
                os.chmod('/etc/cinder/rsakeys', 0o750)
                for file_name in os.listdir('/etc/cinder/rsakeys'):
                    file_path = '/etc/cinder/rsakeys/' + file_name
                    os.chown(file_path, pw_info.pw_uid, pw_info.pw_gid)
                    os.chmod(file_path, 0o640)

    ###############################################
    #       Data Manager - Backup Interface       #
    ###############################################
    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config = utils.get_config_data()
        # Generate a unique time-stamp to put in the existing cinder.conf files
        timestamp = datetime.now().strftime('%Y%m%d%H%M%S%f')
        config['cinder_restore_timestamp'] = timestamp
        # Loop through each CONF file adding a time-stamp to determine changes
        for host in self.__get_storage_providers():
            conf_file = '/etc/cinder/cinder-%s.conf' % host
            utils.set_openstack_config_value(
                conf_file, 'restore_timestamp', timestamp)

    def restore_component(self, upgrading):
        """Performs any Component-specific restoration that is needed"""
        config = utils.get_config_data()
        # Loop through each CONF file seeing if the host is new or old
        for host in self.__get_storage_providers():
            conf_file = '/etc/cinder/cinder-%s.conf' % host
            val = utils.get_openstack_config_value(
                conf_file, 'restore_timestamp')
            # If file came from the restore, we need to verify service exists
            if val is None or val != config['cinder_restore_timestamp']:
                self.__update_new_provider(host)
            # If file is left from before the restore, need to delete it
            else:
                self.__remove_old_provider(host)

    ###############################################
    #      Data Manager - Synchronize UUID's      #
    ###############################################
    def synchronize_uuids(self, host, uuid_map):
        """Synchronizes the UUID's that are needed for the given Host"""
        self.__synchronize_instances(host, uuid_map.get('instances', {}))

    ###############################################
    #    Data Manager - Purge Database Records    #
    ###############################################
    def purge_deleted_database_records(self, timestamp, chunksize):
        """Purge deleted records from the Component's Database"""
        self.__purge_deleted_volumes(timestamp, chunksize)

    ###############################################
    #   Data Manager - Subclass Implementation    #
    ###############################################
    def _get_service_user(self):
        """Provides the name of the Service User for the Component"""
        return 'cinder'

    def _get_enabled_services(self):
        """Provides a list of the Services for the Component"""
        services = ['api', 'scheduler']
        services.extend(self.__get_volume_services())
        return ['openstack-cinder-%s' % svc for svc in services]

    def _get_disabled_services(self):
        """Provides a list of the Disabled Services for the Component"""
        services = ['backup', 'volume']
        return ['openstack-cinder-%s' % svc for svc in services]

    def _get_service_log_files(self):
        """Provides a list of Service Log Files for the Component"""
        logs = ['api', 'scheduler', 'cinder-manage',
                'httpd-error', 'powervc-dbsync']
        return ['/var/log/cinder/%s.log' % log for log in logs]

    def _get_database_info(self):
        """Provides the Database Name/User Tuple for the Component"""
        return ('CINDER', 'cinder')

    def _get_dbsync_commands(self):
        """Provides a command list for the Component's DB Synchronization"""
        return ['/usr/bin/cinder-manage db sync',
                '/opt/ibm/powervc/bin/powervc-dbsync cinder']

    def _get_added_conf_properties(self):
        """Provides a dictionary of configuration properties to add"""
        config = {CINDER_CONF: dict()}
        # If this is the Mgmt System, also need to update the Cinder Conf Files
        for provider in self.__get_storage_providers():
            config['/etc/cinder/cinder-%s.conf' % provider] = dict()
        # Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._added_auth_conf_properties(config[key])
            self._added_messaging_conf_properties(config[key],
                                                  extra_thread_pool_size=128)
            self._added_ras_conf_properties(config[key])
            self.__added_cinder_mgmtsystem_conf_properties(config[key])
        return config

    def _get_deleted_conf_properties(self):
        """Provides a dictionary of configuration properties to remove"""
        config = {CINDER_CONF: dict()}
        # If this is the Mgmt System, also need to update the Cinder Conf Files
        for provider in self.__get_storage_providers():
            config['/etc/cinder/cinder-%s.conf' % provider] = dict()
        # Loop through each of the Config Files, adding the Properties
        for key in config.keys():
            self._deleted_auth_conf_properties(config[key])
            self._deleted_messaging_conf_properties(config[key])
            config[key]['DEFAULT']['enabled_apis'] = V122
            config[key]['DEFAULT']['sql_connection'] = V123
            config[key]['DEFAULT']['lock_path'] = V123
            config[key]['DEFAULT']['osapi_max_request_body_size'] = V130
        return config

    def _get_audit_conf_properties(self, version=None):
        """Provides a tuple of configuration properties for auditing"""
        sects = ['composite:openstack_volume_api_v1',
                 'composite:openstack_volume_api_v2']
        return (PASTE_CONF, sects, ['keystone', 'keystone_nolimit'])

    def _get_included_backup_files(self):
        """Provides a list of Component files to include in a backup"""
        return ['/etc/cinder']

    def _get_excluded_backup_files(self):
        """Provides a list of Component files to exclude from a backup"""
        return ['/etc/cinder/volumes']

    def _get_service_password_properties(self):
        """Provides a list of the configuration properties for passwords"""
        return [('cinder', CINDER_CONF,
                 'keystone_authtoken', 'password')]

    def _temporary_openstack_patches(self):
        """Perform any temporary openstack patching for the Component"""
        pass

    ###############################################
    #    Internal Data Manager Helper Methods     #
    ###############################################
    def __get_storage_providers(self):
        """Helper method to get a list of registered storage providers"""
        files = utils.get_matching_files('/etc/cinder', 'cinder-', '.conf')
        return [f[7:-5] for f in files]

    def __get_volume_services(self):
        """Helper method to get a list of cinder-volume-xxx services"""
        files1 = utils.get_matching_files(
            '/etc/init.d', 'openstack-cinder-volume-', '')
        files2 = utils.get_matching_files(
            '/usr/lib/systemd/system', 'openstack-cinder-volume-', '')
        if files1:
            return [f[17:] for f in files1]
        return [f[17:-8] for f in files2]

    def __added_cinder_mgmtsystem_conf_properties(self, config):
        """Add in the Management System cinder.conf Configuration Properties"""
        config.setdefault('database', dict())
        # Add in the properties for the Database Section
        config['database'].update(dict(
            pool_timeout=('60', V123),
            connection=('aes:ibm_db_sa://cinder:%(cinder_'
                        'password)s@%(local_ip)s:50110/cinder', V123)))
        # Add in the properties for the Oslo Currency Section
        config['oslo_concurrency'] = \
            dict(lock_path=('/var/lib/cinder/tmp', V123))
        config['oslo_middleware'] = dict(
            osapi_max_request_body_size=('1099511627776', V130))
        # Add in the properties for the Default Section
        config['DEFAULT'].update(dict(
            quota_volumes=('100000', V121),
            quota_gigabytes=('1000000', V121),
            log_dir=('/var/log/cinder', V130),
            state_path=('/var/lib/cinder', V130),
            development_ssh_san_conn=('False', V999),
            osapi_volume_listen_port=('9000', V120),
            default_log_levels=('urllib3=WARN', V120),
            strict_ssh_host_key_policy=('True', V122),
            ibmpowervc_policy_granular=('True', V120),
            osapi_volume_listen=('%(local_ip)s', V121),
            enable_v1_api=('False', V130),
            enable_v2_api=('True', V130),
            wsgi_keep_alive=('False', V130),
            api_paste_config=('/etc/cinder/api-paste.ini', V120),
            ssh_hosts_key_file=('/opt/ibm/powervc/data/known_hosts', V122),
            service_down_time=('120', V122), osapi_max_limit=('10000', V1302),
            glance_host=('%(local_ip)s', V122), report_interval=('30', V122),
            osapi_volume_workers=('4', V122), quota_snapshots=('100000', V121),
            volume_group=('cinder-volumes', V120), monkey_patch=('True', V121),
            volume_name_template=('<volume-%s>', V120), verbose=('True', V120),
            scheduler_manager=('powervc_cinder.scheduler.manager.'
                               'PowerVCSchedulerManager', V120),
            scheduler_driver=('powervc_cinder.scheduler.sched_drv.'
                              'PowerVCSchedulerDriver', V120),
            scheduler_host_manager=('powervc_cinder.scheduler.host_'
                                    'manager.PowerVCHostManager', V120),
            osapi_volume_extension=(
                'powervc_cinder.api.plugins.api_extensions_v2', V121),
            volume_manager=(
                'powervc_cinder.volume.manager.PowerVCVolumeManager', V120),
            monkey_patch_modules=('cinder.volume.volume_types:powervc_cinder.'
                                  'volume.volume_type_decorator,cinder.volume.'
                                  'api:powervc_cinder.volume.volume_create_'
                                  'decorator', V121),
            nova_catalog_info=('compute:nova:publicURL', V130),
            nova_ca_certificates_file=(
                '/etc/pki/tls/certs/powervc.crt', V130)))
        # Add the Properties specific to whether it is PowerVM or PowerKVM
        if utils.get_offering() == 'powervm':
            self.__added_cinder_powervm_conf_properties(config)
        else:
            self.__added_cinder_powerkvm_conf_properties(config)

    def __added_cinder_powervm_conf_properties(self, config):
        """Add in the PowerVM-specific cinder.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            supported_host_types=(
                'svc,ssp,xiv,emc_vmax,emc_vnx,ds8k', V130),
            storwize_svc_connection_protocol=('fc', V122),
            storwize_svc_multihostmap_enabled=('True', V123)))

    def __added_cinder_powerkvm_conf_properties(self, config):
        """Add in the PowerKVM-specific cinder.conf Configuration Properties"""
        config['DEFAULT'].update(dict(
            supported_host_types=('svc', V122),
            storwize_svc_connection_protocol=('iscsi', V122),
            storwize_svc_multihostmap_enabled=('True', V120)))

    def __update_new_provider(self, provider):
        """Helper method to make sure the service exists for the provider"""
        service = 'openstack-cinder-volume-%s' % provider
        conf_file = '/etc/cinder/cinder-%s.conf' % provider
        # We only want to create the service if this is a PowerVM Mgmt System
        if not self.__service_file_exists(provider):
            utils.move_file(conf_file, conf_file + '_backup')
            utils.run_command(POWERVC_MANAGE + ' -o addhost -s ' +
                              service + ' -f ' + conf_file + '_backup')
            utils.remove_file(conf_file + '_backup')

    def __remove_old_provider(self, provider):
        """Helper method to make sure service is removed for old provider"""
        service = 'openstack-cinder-volume-%s' % provider
        conf_file = '/etc/cinder/cinder-%s.conf' % provider
        # We only want to delete the service if this is a PowerVM Mgmt System
        if self.__service_file_exists(provider):
            utils.run_command(POWERVC_MANAGE + ' -o removehost -s ' + service)
        # Make sure the CONF file is gone even it the svc-manage didn't
        utils.remove_file(conf_file)

    def __service_file_exists(self, provider):
        """Helper method to see if the service file exists for the provider"""
        svc = 'openstack-cinder-volume-%s' % provider
        svc1 = '/etc/rc.d/init.d/%s' % svc
        svc2 = '/usr/lib/systemd/system/%s.service' % svc
        return utils.file_exists(svc1) or utils.file_exists(svc2)

    def __upgrade_fabrics_conf(self):
        """Helper method to upgrade the fabrics.conf file"""
        props = ['address', 'user', 'password', 'uniqueid', 'display_name']
        fabs = utils.get_openstack_config_value(FABRIC_CONF, 'fc_fabric_names')
        # If there aren't any Fabrics set, then nothing to do
        if not fabs:
            return
        # Loop through each of the Fabrics, adding in the new properties
        for fab in fabs.split(','):
            for prop in props:
                newprop = 'fc_fabric_%s' % prop
                oldprop = 'fc_fabric_%s_%s' % (prop, fab)
                oldval = utils.get_openstack_config_value(FABRIC_CONF, oldprop)
                # If the value is set for the property, than add the new one
                if oldval:
                    # If this is the password property we need to encrypt it
                    if prop == 'password':
                        oldval = utils.encrypt_data(oldval)
                    # Now we can set the new property in the fabric's section
                    utils.set_openstack_config_value(
                        FABRIC_CONF, newprop, oldval, fab)
            # Since 1.2.1 only supported Brocade, we will hard-code to brocade
            if utils.get_openstack_config_value(
                    FABRIC_CONF, 'fc_fabric_type') is None:
                utils.set_openstack_config_value(
                    FABRIC_CONF, 'fc_fabric_type', 'brocade')
            if utils.get_openstack_config_value(
                    FABRIC_CONF, 'fabric_type', fab) is None:
                utils.set_openstack_config_value(
                    FABRIC_CONF, 'fabric_type', 'brocade', fab)
        # Also need to duplicate the fabric names in fc-zone-manager section
        utils.set_openstack_config_value(
            FABRIC_CONF, 'fc_fabric_names', fabs, 'fc-zone-manager')

    def __load_association_tables(self, metadata):
        """Helper method to load the Association Tables for the Volume"""
        tables = dict()
        assoc_names = [
            'volume_attachment', 'volume_metadata', 'volume_health_status',
            'volume_admin_metadata', 'volume_restricted_metadata']
        for name in assoc_names:
            tables[name] = Table(name, metadata, autoload=True)
        return tables

    def __get_table_uuid_column(self, table, table_name):
        """Helper method to get the UUID column from the given table"""
        if table_name == 'volume_health_status':
            return table.c.id
        return table.c.volume_id

    def __purge_deleted_volumes(self, timestamp, chunksize):
        """Helper method to purge deleted volumes from the database"""
        engine = utils.get_database_engine(
            CINDER_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='cinder')
        vol_table = Table('volumes', metadata, autoload=True)
        assoc_tables = self.__load_association_tables(metadata)
        # Query the Volumes table to get the UUID's of the deleted Volumes
        vol_query = expression.select([vol_table.c.id]).where(and_(
            vol_table.c.deleted > 0, vol_table.c.deleted_at < timestamp))
        conn = engine.connect()
        # We want to close the ResultSet for the query after construct it
        vol_rs = conn.execute(vol_query)
        vol_uuids = [r[0] for r in vol_rs.fetchall()]
        vol_rs.close()
        # Break the UUID's into Chunks so that DB2 can handle
        for index in xrange(0, len(vol_uuids), chunksize):
            uuid_chunk = vol_uuids[index: index + chunksize]
            with conn.begin():
                # Loop through the Association Tables updating those UUID's
                for name, table in assoc_tables.iteritems():
                    uuidcol = self.__get_table_uuid_column(table, name)
                    # Once we know the UUID column, then can delete the entries
                    conn.execute(table.delete().where(uuidcol.in_(uuid_chunk)))
                # Now we can deleted the actual rows from the volumes table
                conn.execute(vol_table.delete().where(
                    vol_table.c.id.in_(uuid_chunk)))
        conn.close()

    def __synchronize_instances(self, host, uuid_map):
        """Helper method to synchronize the VM UUID's that are given"""
        # Get a connection to the Database and load the needed Tables
        engine = utils.get_database_engine(
            CINDER_CONF, 'connection', 'database')
        metadata = MetaData(bind=engine, schema='cinder')
        # Setup the Update Statement to update the UUID in the Ports Table
        meta_table = Table('volume_metadata', metadata, autoload=True)
        attach_table = Table('volume_attachment', metadata, autoload=True)
        meta_update = meta_table.update().where(and_(
            meta_table.c.key == 'instance_uuid',
            meta_table.c.value == bindparam('olduuid'))).values(
                value=bindparam('newuuid'))
        attach_update = attach_table.update().where(
            attach_table.c.instance_uuid == bindparam('olduuid')).values(
                instance_uuid=bindparam('newuuid'))
        # Loop through each of the UUID's updating the Volume Tables
        for old_uuid, new_uuid in uuid_map.iteritems():
            engine.execute(attach_update, olduuid=old_uuid, newuuid=new_uuid)
            engine.execute(meta_update, olduuid=old_uuid, newuuid=new_uuid)
