# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import os
import sys
import xml.etree.ElementTree as ET
from powervc_oslo.config import data_utils as utils
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import config_utils as cutils
from powervc_cinder import _

LOCATION = "/etc/cinder/cinder_emc_config_%s.xml"
service = None


def config_port_group(args):
    #First check to see that the EMC VMAX xml exists
    location = LOCATION % args.arrayid
    if not os.path.exists(location):
        #If the file does not exist, tell the user and add to log the msg
        msg = _("The Storage Array ID specified does not correspond to any "
                "storage providers currently managed by PowerVC.")
        cutils.safe_print(msg)
        sys.exit(1)
    #Get a copy of the xml
    tree = ET.parse(location)
    root = tree.getroot()

    #If the list action or no action at all was specified, list the PGs
    if (args.list or len(sys.argv) < 6):
        cutils.safe_print("PortGroups:")
        #Loop through the PortGroups printing out all PortGroup listings
        for group in root.findall("PortGroups"):
            for pg in group.iter():
                if pg.text is not None:
                    cutils.safe_print("%s" % pg.text)
        sys.exit(0)

    #Adds the specified PortGroup to the xml
    if args.add:
        if not args.portgroup:
            msg = _("A PortGroup was not specified. Please specify a "
                    "PortGroup")
            cutils.safe_print(msg)
            sys.exit(2)
        #Loop through PortGroups to first check if the PortGroup exists
        for group in root.findall("PortGroups"):
            for pg in group.iter():
                if pg.text == args.portgroup:
                    msg = _("PortGroup: %s already exists in the EMC VMAX "
                            "configuration.") % args.portgroup
                    cutils.safe_print(msg)
                    sys.exit(3)
        #Add the new PortGroup entry to PortGroups
        for group in root.findall("PortGroups"):
            new = ET.SubElement(group, 'PortGroup')
            new.text = args.portgroup

        #Write the new xml to the conf file
        tree.write(location)

        #Add a message displayed to the user.
        msg = _("PortGroup %(port)s was successfully added to the pool of "
                "candidate PortGroups for EMC VMAX storage array "
                "configuration %(conf)s\n") % dict(port=args.portgroup,
                                                   conf=args.arrayid)
        cutils.safe_print(msg)

    #Removes the specified PortGroup from the xml
    if args.remove:
        if not args.portgroup:
            msg = _("A PortGroup was not specified. Please specify a "
                    "PortGroup")
            cutils.safe_print(msg)
            sys.exit(4)
        pgs = []
        #Loop through the PortGroups making a list composed of each PortGroup
        for group in root.findall("PortGroups"):
            for portgroup in group.iter():
                if portgroup.text is not None:
                    pgs.append(portgroup.text)
        #Check to see that the PortGroup given is in the PortGroup list
        if args.portgroup not in pgs:
            msg = _("The PortGroup specified for removal does not exist in "
                    "the list of candidate PortGroups.")
            cutils.safe_print(msg)
            sys.exit(5)
        #If this is the only PortGroup in the XML, prompt user
        if len(pgs) == 1:
            warning = _("PortGroup: %s is the only candidate PortGroup in the "
                        "EMC VMAX configuration. Removing this PortGroup may "
                        "cause unwanted results. Do you want to continue? "
                        "with this operation? (y/N):") % args.portgroup
            if not utils.prompt_for_confirmation(warning):
                sys.exit(0)
        #Now we can remove the PortGroup from the xml
        for group in root.findall("PortGroups"):
            for portgroup in group.iter():
                if portgroup.text == args.portgroup:
                    group.remove(portgroup)

        #Write the new xml to the conf file
        tree.write(location)
        #Add a message to be logged and displayed to the user
        msg = _("PortGroup %(port)s was successfully removed from the pool "
                "of candidate PortGroups for EMC VMAX storage array "
                "configuration %(conf)s\n") % dict(port=args.portgroup,
                                                   conf=args.arrayid)
        cutils.safe_print(msg)

    if not args.restart:
        msg = _("The EMC VMAX Cinder driver must be restarted to pick up the "
                "configuration changes.")
        cutils.safe_print(msg)
    else:
        msg = _("Restarting the %s service") % service
        cutils.safe_print(msg)
        mgr.restart_services(service)


def setup_parser(cinder_subp, config, service_name):
    global service
    service = service_name
    pg_parser = cinder_subp.add_parser('portgroup',
                                       help=_("Configure the EMC VMAX "
                                              "PortGroup list"),
                                       description=_("View or restrict the "
                                                     "EMC VMAX PortGroups "
                                                     "that PowerVC can use"
                                                     ))
    pg_parser.add_argument('--arrayid',
                           help=_("Storage Array ID to perform the port "
                                  "removal on"),
                           metavar='ArrayID',
                           required=True,
                           dest='arrayid',
                           action='store')
    pg_parser.add_argument('--portgroup',
                           help=_("Target PortGroup to be added or removed "
                                  "from the VMAX PortGroup candidates"),
                           metavar='PortGroup',
                           dest="portgroup",
                           action='store')
    pg_parser.add_argument('--add',
                           help=_("Adds the given PortGroup to the specified "
                                  "array"),
                           dest='add',
                           action='store_true')
    pg_parser.add_argument('--remove',
                           help=_("Removes the given PortGroup from the "
                                  "specified array"),
                           dest='remove',
                           action='store_true')
    pg_parser.add_argument('--list',
                           help=_("Displays a list of each EMC VMAX candidate "
                                  "PortGroup"),
                           dest='list',
                           action='store_true')
    pg_parser.add_argument('--restart',
                           help=_("Restart the cinder EMC VMAX driver"),
                           dest='restart',
                           action='store_true')
    pg_parser.set_defaults(func=config_port_group)
