#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
IBM PowerVC, 2013, Cinder extensions:
This is the object-relational mapping (ORM) for PVC's Cinder sub-domains.
It defines the mappings between PVC resource types, such as StorageNode
and each type's corresponding database table.
Using SqlAlchemy, each resource type is mapped as Data Transfer Object,
such as StorageNodeDTO, because it has attributes but no behavior.
"""
from cinder.i18n import _
from oslo_config import cfg

from sqlalchemy import Column, ForeignKey
from sqlalchemy import Integer, String, Float, DateTime, Text
from sqlalchemy.orm import relationship

from oslo_log import log as logging
from oslo_utils import timeutils
import cinder.db.sqlalchemy.models as cinder_models

NAME_LEN = 255
DESC_LEN = 1023
UUID_LEN = 36
STATUS_LEN = 31
USER_LEN = 63
CRED_LEN = 255
IP_LEN = 39
MODE_LEN = 31

LOG = logging.getLogger(__name__)

name_opts = [
    cfg.StrOpt('pvc_volume_name_template',
               default='',
               help=_("PowerVC specific customization for"
                      " volume name")),
]

cfg.CONF.register_opts(name_opts)


######################################################
#                HMC Model Definition                #
######################################################
class HmcDTO(cinder_models.BASE, cinder_models.CinderBase):
    """The HmcDTO class provides the ORM for the ibm_hmcs table"""
    __tablename__ = 'ibm_hmcs'
    __table_args__ = ()
    # HMC DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    hmc_uuid = Column('uuid', String(UUID_LEN), nullable=False)
    hmc_display_name = Column('display_name', String(NAME_LEN), nullable=False)
    access_ip = Column('access_ip', String(NAME_LEN), nullable=False)
    user_id = Column('user_id', String(USER_LEN), nullable=False)
    password = Column('user_credentials', String(CRED_LEN), nullable=False)
    registered_at = Column('registered_at', DateTime, default=timeutils.utcnow)


class HmcClustersDTO(cinder_models.BASE, cinder_models.CinderBase):
    """The HmcClustersDTO class provides the ORM for the hmc_clusters table"""
    __tablename__ = 'ibm_hmc_clusters'
    __table_args__ = ()
    # HMC Clusters DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    hmc_uuid = Column('hmc_uuid', String(UUID_LEN), nullable=False)
    host_name = Column('host_name', String(NAME_LEN), nullable=False)


######################################################
#           Storage Node Model Definition            #
######################################################
class StorageNodeDTO(cinder_models.BASE, cinder_models.CinderBase):
    """The StorageNodeDTO class provides the ORM for storage_nodes table"""
    __tablename__ = 'storage_nodes'
    __table_args__ = ()
    # Storage Node DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    service_id = Column('service_id', Integer,
                        ForeignKey('services.id'), nullable=False)
    storage_hostname = Column('storage_hostname',
                              String(NAME_LEN), nullable=False)
    backend_state = Column('backend_state', String(STATUS_LEN))
    backend_type = Column('backend_type', String(MODE_LEN))
    backend_id = Column('backend_id', String(NAME_LEN))
    total_capacity_gb = Column('total_capacity_gb', Float)
    free_capacity_gb = Column('free_capacity_gb', Float)
    volume_count = Column('volume_count', Integer)
    compression_supported = Column('compression_supported', String(8))


######################################################
#             Meta-data Model Definition             #
######################################################
class VolumeRestrictedMetadataDTO(cinder_models.BASE,
                                  cinder_models.CinderBase):
    """The VolumeMetaDTO class provides the ORM for restricted_meta table"""
    __tablename__ = 'volume_restricted_metadata'
    # Volume Restricted Meta-data DTO Model Attributes
    id = Column(Integer, primary_key=True)
    key = Column(String(NAME_LEN))
    value = Column(String(NAME_LEN))
    volume_id = Column(String(UUID_LEN),
                       ForeignKey('volumes.id'), nullable=False)
    # Volume Restricted Meta-data DTO Model Relationships
    volume = relationship(cinder_models.Volume,
                          backref="volume_restricted_metadata",
                          foreign_keys=volume_id,
                          primaryjoin='and_('
                          'VolumeRestrictedMetadataDTO.volume_id == Volume.id,'
                          'VolumeRestrictedMetadataDTO.deleted == 0)')


######################################################
#           Health Status Model Definition           #
######################################################
class VolumeHealthStatusDTO(cinder_models.BASE,
                            cinder_models.CinderBase):
    """The Volume Health DTO class provides the ORM for health_status table"""
    __tablename__ = 'volume_health_status'
    __table_args__ = ()
    # Volume Health Status DTO Model Attributes
    id = Column(String(UUID_LEN), primary_key=True)
    health_state = Column(String(STATUS_LEN))
    reason = Column(Text)
    unknown_reason_details = Column(Text)


class StorageNodeHealthStatusDTO(cinder_models.BASE,
                                 cinder_models.CinderBase):
    """The StorageNode Health DTO class provides the ORM for
       health_status table"""
    __tablename__ = 'storage_node_health_status'
    __table_args__ = ()
    # StorageNode Health Status DTO Model Attributes
    id = Column(String(UUID_LEN), primary_key=True)
    health_state = Column(String(STATUS_LEN))
    reason = Column(Text)
    unknown_reason_details = Column(Text)


######################################################
#           On-board Task Model Definition           #
######################################################
class OnboardTaskDTO(cinder_models.BASE, cinder_models.CinderBase):
    """The OnboardTaskDTO class provides the ORM for the onboard_tasks table"""
    __tablename__ = 'onboard_tasks'
    __table_args__ = ()
    # On-board Task DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    host = Column('host', String(NAME_LEN), nullable=False)
    status = Column('status', String(STATUS_LEN), default='waiting')
    started = Column('started_at', DateTime, default=timeutils.utcnow)
    ended = Column('ended_at', DateTime)
    progress = Column('progress', Integer, default=0)


class OnboardTaskVolumeDTO(cinder_models.BASE, cinder_models.CinderBase):
    """The OnboardTaskVolumeDTO class provides the ORM for the tasks table"""
    __tablename__ = 'onboard_task_volumes'
    __table_args__ = ()
    # On-board Task Volume DTO Model Attributes
    id = Column('id', Integer, primary_key=True)
    task_id = Column('task_id', Integer, nullable=False)
    volume_uuid = Column('volume_uuid', String(UUID_LEN), nullable=False)
    volume_name = Column('volume_name', String(NAME_LEN), nullable=False)
    status = Column('status', String(STATUS_LEN), nullable=False)
    fault_message = Column('fault_message', String(NAME_LEN))


###################################################
#            List of PowerVC 1.2 DTO's            #
###################################################
POWERVC_V1R2_DTOS = [
    HmcDTO, HmcClustersDTO, StorageNodeDTO, VolumeRestrictedMetadataDTO,
    VolumeHealthStatusDTO, OnboardTaskDTO, OnboardTaskVolumeDTO,
    StorageNodeHealthStatusDTO
]


###################################################
#          Override Name Method on Volume         #
###################################################
# The way OpenStack is setup each Volume basically has 2 names, the
# "display name" that the user specifies and the "name" that is internally
# generated based on a template and what is used to actually create the
# volume on the Storage Provider.  This template name isn't actually persisted
# and is generated on the fly when someone asks for it.  This works well for
# volumes created through OpenStack, but for any volumes that are changed
# outside of PowerVC or are on-boarded, the template name won't match the
# actual name on the Storage Provider.
#
# This code below is to inject method overrides for the name property on the
# OpenStack SQLAlchemy Volume class so that that the name property returns the
# display name rather than the generated template name if it was on-boarded.
def _volume_name(self):
    """Override the SQLAlchemy Volume Name method"""
    # If on-boarded (i.e not scheduled for create), use the display name
    if self.get('scheduled_at') is None:
        return self.get('display_name')
    # If this wasn't an on-boarded Volume, fall back to the original
    name = _VOLUME_NAME_TEMPL % self.__dict__

    # if pvc_volume_template_name is specified in cinder configuration, use it
    # Optionally, if a custom name should be assigned to volume name,
    # add pvc_volume_name_template to storage provider specific cinder
    # configuration file in /etc/cinder.
    # example : pvc_volume_name_template = custom-%(display_name)s
    # or pvc_volume_template = %(display_name)s

    if _PVC_VOLUME_NAME_TEMPL:
        name = _PVC_VOLUME_NAME_TEMPL % self.__dict__
    return _sanitize_volume_name(name)

_VOLUME_NAME_TEMPL = 'volume-%(display_name).37s-%(id).13s'

_PVC_VOLUME_NAME_TEMPL = cfg.CONF.pvc_volume_name_template
_INVALID_NAME_CHARS = '()\<>*$&?|[]\'"`,/#=:;^!@%{}+'
_INVALID_NAME_CHAR_MAP = dict([(c, '') for c in list(_INVALID_NAME_CHARS)])


def _sanitize_volume_name(vol_name):
    """Replace invalid characters in the template name for the Volume"""
    replaced_char = False
    # If the Volume Name is None or an empty string, just return it
    if vol_name:
        new_name = list(vol_name)
        # Loop through each character in the string, substituting if invalid
        for index in range(len(new_name)):
            # Try to use ord for the character since it may fail in some cases
            try:
                ordinal = ord(new_name[index])
            except Exception:
                # It failed due to being a special character so we replace it
                ordinal = 127
            out_of_range = ordinal < 33 or ordinal > 126
            # Replace reserved or non-ASCII characters with the caret symbol
            # 39655 - change tilde character to underscore
            if out_of_range or new_name[index] in _INVALID_NAME_CHAR_MAP:
                new_name[index] = '_'
                replaced_char = True
        # For performance only convert new name to a string if it changed
        if replaced_char:
            return "".join(new_name)
    return vol_name

_orig_dbvol_name = getattr(cinder_models.Volume, 'name')
setattr(cinder_models.Volume, '_orig_name', _orig_dbvol_name.fget)
setattr(cinder_models.Volume, 'name', property(fget=_volume_name))
