# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import ConfigParser
from oslo_log import log as logging
from oslo_concurrency.lockutils import synchronized
from powervc_cinder import _
from cinder.volume.configuration import Configuration
from webob import exc
import six

from powervc_cinder.zonemanager import powervc_fc_zone_manager as fabmanager


LOG = logging.getLogger(__name__)


class MapInitiators():
    """
    Class for doing interfacing with the PowerVCZoneManager to do correlation
    with WWPN initiators that are plugged into the managed fabric(s).

    The use case for this class is the storage topology collection on the
    nova side. When FC ports are inventoried, the function in this module
    is used set the switch fabric designation on the port entry in the
    nova DB. Rather than extend the volume service with a new method that
    does not really belong there, this class will do the extra work needed
    of constructing the zone/fabric manager in the REST API flow. This
    allows the function to work independent of any storage providers being
    registered.
    """

    def __init__(self):
        """ Do configuration work. Construct the fabric manager """
        self.opts = None
        self.fabricmanager = None
        self.conf = self._load_fabric_opts()

    @synchronized('fabric_map', 'pvc-', external=True)
    def get_wwpn_fabric_map(self, wwpn_list):
        """
        Given a list of WWPN initiators, return a dictionary mapping to
        the fabrics where the WWPN is logged into.

        :param wwpn_list: The list of wwpns. These strings should not be
                tokenized with colons (:).
                    Example:
                        [
                            "10000090FA2A5866",
                            "10000090FA2A8923",
                            "c0507606d56e03af"
                        ]
        :returns: A dictionary mapping of input wwpns to discovered fabric
                names. Example:
                        {
                            "10000090FA2A8923": "B",
                            "10000090FA2A5866": "A",
                            "c0507606d56e03af": None
                        }
        """
        if self.conf and not self.fabricmanager:
            self.fabricmanager = fabmanager.PowerVCZoneManager(
                self.conf, overrides=self.opts)
            LOG.debug("PowerVCZoneManager initialized.")
        if not self.fabricmanager:
            msg = _("Could not obtain a fabric manager client. Fibre "
                    "Channel switch fabrics may not be registered.")
            raise exc.HTTPInternalServerError(explanation=six.text_type(msg))

        # Add colons into input wwpns and put in a mapping dictionary.
        with_colons = dict([[fabmanager.get_formatted_wwn(w.lower()), w]
                            for w in wwpn_list])
        LOG.debug("ENTER get_wwpn_fabric_map: colon_map = %s" % with_colons)
        input_dict = {'initiators': with_colons.keys()}
        try:
            data = self.fabricmanager.get_san_context(input_dict)
        except Exception as ex:
            # The exception here may not be the root cause. If we cannot
            # get a normal response, then we raise our own error here.
            LOG.exception(ex)
            msg = (_("Unable to retrieve WWPN initiator information from "
                     "one or more fabric switches. Error: %(ex)s") %
                   dict(ex=ex))
            raise exc.HTTPInternalServerError(explanation=six.text_type(msg))
        LOG.debug("Zonemanager.get_san_context() returns: %s" % data)

        # Build a dictionary mapping of the original wwpn passed to the
        # result of the fabric lookup.
        map_dict = dict([[w, self._match_fabric(c, data)]
                         for c, w in with_colons.iteritems()])

        LOG.debug("RETURNS wwpn_mapping_dict: %s" % map_dict)
        return map_dict

    def _match_fabric(self, wwpn, fabricmanager_data):
        """ Helper to map the wwpn to a fabric returned in the passed data """
        for fabric, items in fabricmanager_data.iteritems():
            if wwpn in items['initiators']:
                # We upper case the fabric designator for API consumption.
                # e.g. 'a' --> 'A'
                return fabric.upper()
        return None

    def get_zones_fabric_map(self, initiators=[], targets=[]):
        """
        Given a filter list of initiators (and or) targets wwpns,
        returns zones with matching members. If a single list is provided,
        then filter against only that list. If both list provided,
        then filter against both lists. If no filter is supplied
        then a list of all active zones will be returned
        Example:
        active_zones-
            { "zone1": ["init_wwpn1", "target_wwpn1"],
              "zone2": ["init_wwpn2", "target_wwpn2"]
            }
        Example1-
            initiators: ["init_wwpn1", "init_wwpn2"]
            targets   : ["target_wwpn1"]
            Returns   : ["zone1"]
        Example2-
            initiators: ["init_wwpn1", "init_wwpn2"]
            targets   : []
            Returns   : ["zone1", "zone2"]
        """
        if self.conf and not self.fabricmanager:
            self.fabricmanager = fabmanager.PowerVCZoneManager(
                self.conf, overrides=self.opts)
            LOG.debug("PowerVCZoneManager initialized.")
        if not self.fabricmanager:
            msg = _("Could not obtain a fabric manager client. Fibre "
                    "Channel switch fabrics may not be registered.")
            raise exc.HTTPInternalServerError(explanation=six.text_type(msg))
        try:
            zonenames = self.fabricmanager.get_zones(initiators, targets)
        except Exception as ex:
            # The exception here may not be the root cause. If we cannot
            # get a normal response, then we raise our own error here.
            LOG.exception(ex)
            msg = (_("Unable to retrieve zones from "
                     "one or more fabric switches. Error: %(ex)s"
                     "Zones: %(zones)") %
                   dict(ex=ex, zones=zonenames))
            raise exc.HTTPInternalServerError(explanation=six.text_type(msg))
        LOG.debug("Zonemanager.get_zones() returns: %s" % zonenames)
        return zonenames

    def _load_fabric_opts(self):
        """
        We are not running in the volume service, so programmatically
        parse the fabrics.conf file and retrieve options that we care
        about so that fabric manager (a.k.a zone manager) can be
        initialized. Return a Configuration object with the fc_fabric_names
        property set. Return None if no fabrics are registered.
        """
        FC_ZONE_MANAGER = "fc-zone-manager"
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/cinder/fabrics.conf')
        if not parser.has_option(FC_ZONE_MANAGER, 'fc_fabric_names'):
            LOG.info(_("No fabrics registered."))
            return None
        fabric_names = parser.get(FC_ZONE_MANAGER, 'fc_fabric_names')
        fabric_names_list = fabric_names.split(',')
        if not fabric_names_list:
            LOG.info(_("No fabric names registered."))

        key_prefixes = ['fc_fabric_address', 'fc_fabric_user',
                        'fc_fabric_password']
        key_prefixes_debug = ['fc_fabric_address', 'fc_fabric_user']
        options = {}  # dict of fabs containing list of option tuples
        options_debug = {}  # dict of fabs with list of option tuples for debug
        # Loop over the fabric names (e.g. 'a', 'b')
        for fab in fabric_names_list:
            # First get required config props
            fab_opts = [(k, parser.get(fab, k))
                        for k in key_prefixes]
            fab_opts_debug = [(k, parser.get(fab, k))
                              for k in key_prefixes_debug]

            # Now get non-required config props that we can default.
            if parser.has_option(fab, 'fabric_type'):
                fabric_type = parser.get(fab, 'fabric_type')
            else:
                fabric_type = 'brocade'  # default
            fab_opts.append(('fabric_type', fabric_type))
            fab_opts_debug.append(('fabric_type', fabric_type))

            if parser.has_option(fab, 'fc_fabric_display_name'):
                dn = parser.get(fab, 'fc_fabric_display_name')
            else:
                dn = 'fabric_' + fab  # default
            fab_opts.append(('fc_fabric_display_name', dn))
            fab_opts_debug.append(('fc_fabric_display_name', dn))
            if parser.has_option(fab, 'fc_fabric_vsan'):
                vsan = parser.get(fab, 'fc_fabric_vsan')
                fab_opts.append(('fc_fabric_vsan', vsan))
                fab_opts_debug.append(('fc_fabric_vsan', vsan))
            options[fab] = fab_opts
            options_debug[fab] = fab_opts_debug
        LOG.debug("Fabric options: %s" % options_debug)

        # Construct the empty configuration, then set the fabric names
        # to it, for use by the zone manager.
        conf = Configuration([])
        """conf.local_conf.import_opt(
            'fc_fabric_names',
            'powervc_cinder.zonemanager.powervc_fc_zone_manager')"""
        conf.local_conf.set_override('fc_fabric_names',
                                     fabric_names,
                                     'fc-zone-manager')
        self.opts = options
        return conf
