# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from oslo_log import log as logging
from powervc_cinder import _
LOG = logging.getLogger(__name__)


class Fabric():
    """
    We use this Fabric object to build up a list of all the zones that
    we want to create.  It knows about the initiators, storage controller
    nodes and the chosen WWPN on each node.
    """
    def __init__(self, name, phy_initiator_wwpns, phy_to_virt_map):
        """
        Pass in a list of all physical initiator WWPNs, and a
        phy_to_virt_map that looks like:

        {
          phy_wwpn_X: [ virt_wwpn_a, virt_wwpn_b ],
          phy_wwpn_Y: [ virt_wwpn_c, virt_wwpn_d ]
        }

        No target node information is passed in here, we call
        add_node_wwpn() for every node that we want to use.
        """
        self.name = name
        self._phy_initiator_wwpns = phy_initiator_wwpns
        self._phy_to_virt_map = phy_to_virt_map
        self._nodes = []
        self._zones = {}

    def add_node_wwpn(self, target_wwpn):
        """
        For each target node, we have chosen a WWPN to which we
        want to connect.  We should call this once for each target node
        on the fabric. If storwize_svc_multipath_enabled is True, than
        we will add every target_wwpn for each node
        """
        if isinstance(target_wwpn, list):
            self._nodes.extend(target_wwpn)
        else:
            self._nodes.append(target_wwpn)

    def zone_initiator_to_node(self, phy_initiator_index, node_index):
        """
        Adds a zone from each virtual wwpn associated with the
        specified physical wwpn to the chosen target WWPN on the
        specified node.
        """
        LOG.info(_("Mapping initiator %(phy)s to node %(node)s")
                 % {'phy': phy_initiator_index,
                    'node': node_index})

        phy_initiator_wwpn = \
            self._phy_initiator_wwpns[phy_initiator_index]
        for virt_wwpn in self._phy_to_virt_map[phy_initiator_wwpn]:
            if virt_wwpn not in self._zones:
                self._zones[virt_wwpn] = []
            self._zones[virt_wwpn].append(self._nodes[node_index])

    def initiator_count(self):
        return len(self._phy_initiator_wwpns)

    def node_count(self):
        return len(self._nodes)

    def zones(self):
        return self._zones
