# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
PowerVC Extension to the Cinder filtered Scheduler Driver
"""

from cinder import db
from oslo_log import log as logging
from oslo_utils import timeutils
from cinder.scheduler import driver as base_driver
from cinder.scheduler import filter_scheduler as driver

from powervc_cinder import _

LOG = logging.getLogger(__name__)


class PowerVCSchedulerDriver(driver.FilterScheduler):
    """PowerVC extension to the Cinder Filter Scheduler."""

    def __init__(self, *args, **kwargs):
        super(PowerVCSchedulerDriver, self).__init__(*args, **kwargs)

    def _schedule(self, context, request_spec, filter_properties=None):
        """
        Returns a list of hosts that meet the required specs,
        ordered by their fitness.
        """
        s = super(PowerVCSchedulerDriver, self)
        hosts = s._schedule(context, request_spec,
                            filter_properties=filter_properties)

        if not hosts:
            # no hosts fitted. At least we cannot find the hosts
            # that matches capacity requirement. Log an error to
            # to volume meta data.

            # collect request related information
            volume_id = request_spec['volume_id']
            vol_properties = request_spec['volume_properties']
            req_size = vol_properties['size']

            # collect host_state information
            elevated = context.elevated()
            all_hosts = self.host_manager.get_all_host_states(elevated)

            # For now we are only focusing on the capacity.
            req_info = (_('volume request: '
                          'requested size: %(size)s. ') % {'size': req_size})

            info_list = []
            for hstate_info in all_hosts:
                ts = timeutils.isotime(at=hstate_info.updated)
                info = (_("{host: %(hostname)s, free_capacity: %(free_cap)s, "
                          "total_capacity: %(total)s, reserved_percentage:"
                          " %(reserved)s, last update: %(time_updated)s}. ") %
                        {'hostname': hstate_info.host,
                         'free_cap': hstate_info.free_capacity_gb,
                         'total': hstate_info.total_capacity_gb,
                         'reserved': hstate_info.reserved_percentage,
                         'time_updated': ts})
                info_list.append(info)
            if len(info_list) > 0:
                info_list = ''.join(info_list)
                msg = (_('request exceeds capacity: %(req_info)'
                         ('available capacity: %(info_list)s') %
                         {'info_list': info_list,
                          'req_info': req_info}))
                meta_data_msg = (_('Volume request for '
                                   '%(req_size)d exceeds capacity.' %
                                   {'req_size': req_size}))
                meta_data = {'schedule Failure description': meta_data_msg}
            else:
                msg = (_("No storage has been registered. %s" % req_info))
                meta_data_msg = (_("No storage has been registered."))
                meta_data = {'schedule Failure description': meta_data_msg}

            LOG.error(("Schedule Failure: volume_id: %(volume_id)s - %(msg)s "
                       % {'volume_id': volume_id,
                          'msg': msg}))

            db.volume_update(context, volume_id, {'metadata': meta_data})

            return None
        else:
            return hosts


###################################################
#       Override DB Volume Update on Driver       #
###################################################
# Since the Cinder Scheduler now includes the Pool together with the Host to
# be unique, the value that is ending up in the DB for the Host is host#pool
# This is causing problems for the UI/etc assuming the Host is really the Host,
# so we are going to strip off the Pool before putting in the DB so is the Host
def _volume_update_db(context, volume_id, host):
    """Set the host and set the scheduled_at field of a volume."""
    # Strip off the Pool from the Host if it is provided
    if host and host.find('#') > 0:
        host = host[:host.find('#')]
    return _old_volume_update_db(context, volume_id, host)

_old_volume_update_db = base_driver.volume_update_db
base_driver.volume_update_db = _volume_update_db
