#
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Extended Volume Driver interface to discover/query resource information"""
from cinder import exception
from oslo_log import log as logging

from powervc_cinder import _

LOG = logging.getLogger(__name__)
CONNECTION_TYPE_NPIV = 'npiv'
CONNECTION_TYPE_VSCSI = 'pv_vscsi'


class VolumeDiscoveryDriver(object):

    """
    Extended Volume Driver interface for drivers to implementation to
    discover/query additional resource information from the managed system.

    The base VolumeDriver interface provides methods to take actions on the
    Storage Provider and its Volumes but given the premise that OpenStack is
    the authoritative management source for the Hosts being managed, it assumes
    the resources created by OpenStack are an accurate representation of the
    current state of the resources, so it provides very limited information
    through the driver interface about those resources.

    This interface extends those driver capabilities by asking the driver to
    provide 4 levels of information about the resources being managed:
        1) discover - provide a list of all Volumes that exist on the Provider
        2) query - provide enough info about the Volumes to import into OS
        3) inventory - provide additional details about the Volumes specified
        4) metrics - provide additional metric information about the Volumes
    """

    def discover_volumes(self, context, filters=None):
        """
        Returns a list of all of the Volumes that exist on the given Provider.
        For each Volumes the driver needs to return a dictionary containing
        the following attributes:
            name:      The Name of the Volume defined on the Storage Provider
            status:    The Status of the Volume, matching the status definition
            uuid:      The UUID of the VM when created thru OS (Optional)
            status:    The Status of the Volume, matching the definition
            size:      The Size of the Volume in GB
            restricted_metadata: The Additional Meta-data from the Driver
              vdisk_id:   The Identifier for the Volume on the Back-end
              vdisk_name: The Name of the Volume on the Back-end
            support:   Dictionary stating whether the Volume can be managed
              status:  Whether or not it is "supported" or "not_supported"
              reasons: List of Text Strings as to why it isn't supported

        :param context:   The security context for the query
        :param filters:   The filters to apply, such as {'wwpns': ['wwpn1',..]
        """
        raise NotImplementedError()

    def query_volumes(self, context, volumes, server_info={}, mark_boot=True):
        """
        Returns a list of Volumes (matching those specified on input) with
        enough additional details about each Volume to be able to import the
        Volume into the Cinder Database such as OpenStack can start managing.
        For each Volume the driver needs to return a dictionary containing
        the following attributes:
            uuid:      The UUID of the Volume when created thru OS
            name:      The Name of the Volume defined on the Storage Provider
            status:    The Status of the Volume, matching the definition
            size:      The Size of the Volume in GB
            restricted_metadata: The Additional Meta-data from the Driver
              vdisk_id:   The Identifier for the Volume on the Back-end
              vdisk_name: The Name of the Volume on the Back-end

        :param context:   The security context for the query
        :param volumes: A list of dictionary objects for each Volume with:
            uuid:      The UUID of the Volume when created thru OS
            name:      The Name of the Volume defined on the Storage Provider
            restricted_metadata: The Additional Meta-data from the Driver
              vdisk_id:   The Identifier for the Volume on the Back-end
              vdisk_name: The Name of the Volume on the Back-end
        :param server_info: The host info for the attached servers
        """
        # Currently the discover/query methods return the same data, so we can
        # return the values passed in since discover was called to get the info
        return volumes

    def inventory_volumes(self, context, volumes):
        """
        Provides a mechanism for the Driver to gather Inventory-related
        information for the Volumes provided off of the Back-end at
        periodic intervals.  The Driver is free from there to populate
        the information directly in the Database rather than return it.

        :param context: The security context for the query
        :param volumes: A list of dictionary objects for each Volume with:
            uuid:      The UUID of the Volume when created thru OS
            name:      The Name of the Volume defined on the Storage Provider
            restricted_metadata: The Additional Meta-data from the Driver
              vdisk_id:   The Identifier for the Volume on the Back-end
              vdisk_name: The Name of the Volume on the Back-end
        """
        pass

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        return []

    def _get_vscsi_itl_info(self, server_infos, volume):
        """
        Return the vsci connection info for a given volume
        and given instance
        """
        valid_itls = []
        if server_infos is None:
            return valid_itls
        itl_list = volume.get('itl_list', [])
        if itl_list:
            for cinfo in itl_list:
                for server_wwpns in server_infos.values():
                    # make the wwpn to upper
                    vol_wwpns = cinfo.get('vios_wwpn', [])
                    vol_wwpns = [wwpn.upper() for wwpn in vol_wwpns]
                    server_wwpns = [wwpn.upper() for wwpn in server_wwpns]
                    common_wwpn = list(set(vol_wwpns) & set(server_wwpns))
                    if common_wwpn and len(common_wwpn) > 0:
                        valid_itls.append(cinfo)
            # END FOR-Loop
        return valid_itls

    def _add_attach_info(self, volume, boot_map, connection_info,
                         server_id, mountpoint=None, mark_boot=False):
        """
        The attachment information of volume to given server_id is set
        in the volume passed in. the attach info looks like
        volume['attach_info'] = {
            'server_id' :
                {'connection_info':
                    {'driver_volume_type': 'fibre_channel',
                     'data': {'target_lun': u'1', 'qos_spec': None,
                              'target_wwn': [u'5005076802432ADE',
                                             u'5005076802232ADF'],
                              'target_discovered': False, 'access_mode': 'rw'},
                     'volume_connector': {'connector': None,
                             'connection-type': 'npiv'}},
                'mountpoint': '/dev/sdb'}
            }
        """
        attach_info = dict(connection_info=connection_info)
        if mountpoint:
            attach_info['mountpoint'] = mountpoint
        volume['attach_info'][server_id] = attach_info
        # See if this should be flagged as the Server's Boot Volume
        if mark_boot is True and server_id not in boot_map:
            volume['bootable'] = True
            boot_map[server_id] = volume['uuid']

    def _get_connection_info(self, pg83_naa, connect_info,
                             conn_type=None, udid=None):
        """Helper method to get Connection Info for the Discovered Volume
        This method creates the connector for the fibre-channel volume type
        and updates the connection info as per the connection type(npiv/vscsi)
        """
        target_lun = connect_info.get('target_lun')
        target_wwn = connect_info.get('target_wwn')
        data = dict(access_mode='rw', qos_spec=None, target_discovered=False,
                    target_lun=target_lun, target_wwn=target_wwn)
        # Add the Data to the overall Connection Info that is returned
        connect_info = dict(driver_volume_type='fibre_channel', data=data)
        # If this is NPIV, we want to set the connection type as such
        if conn_type == CONNECTION_TYPE_NPIV:
            connect_info['volume_connector'] = {
                'connection-type': CONNECTION_TYPE_NPIV, 'connector': None}
        # If this is vscsi, ...
        elif conn_type == CONNECTION_TYPE_VSCSI:
            connect_info['volume_connector'] = {
                'connection-type': CONNECTION_TYPE_VSCSI, 'connector': None}
            connect_info['data']['target_UDID'] = udid
            if pg83_naa:
                connect_info['data']['pg83NAA'] = pg83_naa
        else:
            msg = (_('Can not establish the connection type '
                     'for the %(type)s ' %
                     {'type': conn_type})
                   )
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return connect_info
