# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

# standard libs
from collections import OrderedDict
import grp
import os.path
import re
import requests
from requests.exceptions import SSLError
import six
import ssl
import urlparse

# related third party libs
from cinder import context
from cinder import exception
from cinder.db import api as db_api
from cinder.volume import volume_types
from cinder.volume.drivers.ibm import xiv_ds8k
from cinder.zonemanager import utils as fczm_utils
from oslo_log import log as logging
from oslo_utils import excutils
# ds8k closed driver imports
from xiv_ds8k_openstack.utils.powervc_ds8k import DS8KPowerVCHelpers
from xiv_ds8k_openstack import ds8k_nova_proxy

# local/powervc libs
from powervc_cinder.volume.drivers.xiv_ds8k import common_driver
from powervc_cinder.zonemanager import powervc_utils as pvc_fczm_utils
__import__('powervc_oslo.storage.ds8k')


# don't do any zoning requests
fczm_utils.AddFCZone = pvc_fczm_utils.ignore_zoning_request
fczm_utils.RemoveFCZone = pvc_fczm_utils.ignore_zoning_request

LOG = logging.getLogger(__name__)
DATA_TYPE_KEY = 'data_type'

from powervc_cinder.volume.drivers.xiv_ds8k.common_driver\
    import RESTRICTED_METADATA_VDISK_ID_KEY,\
    RESTRICTED_METADATA_VDISK_UID_KEY,\
    RESTRICTED_METADATA_VDISK_NAME_KEY,\
    METADATA_VOLUME_WWN


class DS8KProductDriver(common_driver.XIVDS8KDriver):
    def __init__(self, *args, **kwargs):
        super(DS8KProductDriver, self).__init__(*args, **kwargs)
        self._setup_exception = None
        self.endpoint_desc = "DS8KProductDriver provider"
        self.helpers = DS8KPowerVCHelpers()
        self._verified_conn = False
        self.xiv_ds8k_proxy._connector_obj = DS8KRESTConnector()

    def get_connection(self, pool=None, metadata={}):
        """Get a connection context for DS8K """
        LOG.debug("Create resource ds8k_context object ..")

        try:
            ds8k_userid = self.configuration.san_login
            ds8k_password = self.configuration.san_password
            ds8k_password = self._decode_password(ds8k_password)
            ds8k_ip = self.configuration.san_ip
            return_context = {'username': ds8k_userid,
                              'password': ds8k_password,
                              'address': ds8k_ip,
                              'http_connector':
                              self.xiv_ds8k_proxy._connector_obj}
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Fetching DS8K connection "
                            "details failed."))
                LOG.exception(e)

        if not self._verified_conn:
            # Here we create a connection to verify the sign on information.

            # This call connects and updates the context with the helper client
            # connection cache
            if not self.helpers.verify_admin_privileges(return_context):
                msg = _("Error connecting to the DS8K.")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            self._verified_conn = True

        try:
            if pool is None:
                return_context['pool'] = self.configuration.san_clustername
            else:
                return_context['pool'] = pool

            return_context['volume_metadata'] = metadata or {}
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Fetching DS8K connection "
                            "details failed."))
                LOG.exception(e)

        return return_context

    def _build_default_opts(self):
        conn = self.get_connection()
        pool_dict = self.helpers.get_pools_list(conn)
        configured_pool = pool_dict.get(self.configuration.san_clustername)
        if not configured_pool:
            names = [p['name'] for p in pool_dict.values()]
            msg = _('Unable to find the pool name for the configured pool '
                    '%(config_value)s in the pool list name %(pool_list)s.' %
                    {'config_value': self.configuration.san_clustername,
                     'pool_list': names})
            raise exception.VolumeBackendAPIException(data=msg)
        # Get pool name from ID
        """ Get default opts to allow manager to build the default vol type """
        default_opts = {'storage_pool':
                        configured_pool['name'],
                        'thin_provision': "True"
                        }
        return default_opts

    def get_storage_metadata(self):
        """Returns the additional information that PowerVC uses to populate
        various UI elements.
        """
        compression = False
        try:
            LOG.debug("Get helper class handle")
            ds8k_context = self.get_connection()
            pools = self.helpers.get_pools_list(ds8k_context)

            # convert dict to list (UI expects list)
            poollist = []
            temp = []
            for value in pools.iteritems():
                temp = value
                poollist.append(temp)

            # add free_capacity, capacity and name
            volume_pools = []
            total, free, ptotal, pavail = (0, 0, 0, 0)
            for key, pool in enumerate(poollist):
                # sum up capacity stats for storage
                ptotal = int(pool[1]['soft_size'])
                pavail = int(pool[1]['free_size'])
                # assumption 1: available size = free_size of pool
                # assumption 2: total size is soft_size of pool
                total, free = (total + ptotal, free + pavail)
                # add volume pool meta data
                volume_pool = {'name': pool[1]['name'], 'free_capacity':
                               (pool[1]['free_size'] + " GB"),
                               'capacity': (pool[1]['soft_size'] + " GB")}
                volume_pools.append(volume_pool)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Fetching pool details failed."))
                LOG.exception(e)

        return {'volume_pools': volume_pools,
                'compression_supported': compression}

    def create_volume(self, volume):
        """ Create a DS8K volume and add PowerVC restricted metadata """
        self._add_default_volume_type(volume)
        # create a volume
        vol = super(DS8KProductDriver, self).create_volume(volume)
        # update volume metadata
        LOG.info(_("Updating volume meta data.."))
        meta = vol.get('provider_location')
        # Note we do not have to convert provider_location to a dict as
        # the helper methods handle that.
        ds8k_context = self.get_connection(metadata=meta)
        be_vol = self.helpers.get_volume(ds8k_context)
        return self._update_metadata(vol, volume['id'], be_vol['id'],
                                     be_vol['name'],
                                     be_vol['id'], be_vol['wwn'])

    def create_cloned_volume(self, tgt_volume, src_volume):
        """
        Overrides the superclass create_cloned_volume. Adds
        restricted metadata information to the cloned volume
        and sets default volume type if one is not specified
        """
        # create a volume
        self._add_default_volume_type(tgt_volume)
        s = super(DS8KProductDriver, self)
        vol = s.create_cloned_volume(tgt_volume, src_volume)
        # update volume metadata
        LOG.info(_("Updating volume meta data.."))
        meta = vol.get('provider_location')
        # Note we do not have to convert provider_location to a dict as
        # the helper methods handle that.
        ds8k_context = self.get_connection(metadata=meta)
        be_vol = self.helpers.get_volume(ds8k_context)
        return self._update_metadata(vol, tgt_volume['id'], be_vol['id'],
                                     be_vol['name'], be_vol['id'],
                                     be_vol['wwn'])

    def _sanitize_host(self, connector):
        # The backend driver allows 32 chars but 7 of those
        # are consumed by OShost: construct host name from
        # the first 6 chars of the host plus the first wwpn
        if len(connector['host']) < 26:
            return
        hostname = connector['host'][:5]
        connector['host'] = hostname + "_" + sorted(connector['wwpns'])[0]

    @pvc_fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        info = None
        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns
        self._sanitize_host(connector)

        LOG.info(_("Initializing connection with "
                   "parameters %(connector)s" %
                   {'connector': connector}))
        if (connector['connection_type'] and
                'pv_vscsi' in connector['connection_type']):
            # for a VIOS use the AIX host type which does has a volume
            # limit above 256.
            connector['os_type'] = 'AIX'
        elif 'os_type' in connector:
            if connector['os_type'] == 'ibmi':
                connector['os_type'] = 'OS400'
            elif connector['os_type'] == 'aix':
                connector['os_type'] = 'AIX'
            else:
                # Use default host type on the DS8K
                connector.pop('os_type', None)
        else:
            msg = _("The Virtual machine associated with connector "
                    "%(connector) does not have an operating system "
                    "identified. The process for bringing the virtual "
                    "machine under management by PowerVC must be completed "
                    "before attaching an NPIV volume to the VM for"
                    "DS8000 controllers")
            raise exception.VolumeBackendAPIException(data=msg)

        s = super(DS8KProductDriver, self)
        info = s.initialize_connection(volume, connector)
        LOG.info(info)
        if info is not None:
            info['data']['vendor'] = "IBM"
            info['data']['volume_id'] = volume['id']
            try:
                if (connector['connection_type'] and
                        'pv_vscsi' in connector['connection_type']):
                    ctxt = context.get_admin_context()
                    metadata = db_api.volume_metadata_get(ctxt, volume['id'])
                    info['data']['pg83NAA'] = metadata['volume_wwn']
            except Exception as e:
                # don't worry if we can't add code page 83
                pass
        LOG.info(info)
        return info

    @pvc_fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns
        self._sanitize_host(connector)

        LOG.info(_("Terminating connection with parameters %(connector)s" %
                   {'connector': connector}))

        # Perform the masking on the storage controller
        s = super(DS8KProductDriver, self)
        info = s.terminate_connection(volume, connector, **kwargs)
        LOG.info(info)
        return info

    def discover_volumes(self, context, filters=None):
        LOG.debug("discover_volumes filters %s" % filters)
        ds8k_context = self.get_connection()
        # Set pool to empty to get all volumes in all pools
        ds8k_context['pool'] = ''
        volumes = {}
        connection_info = {}
        # If we need to filter volumes we need to call the more expensive
        # get_volumes_and_hosts method to get host information.  If we don't
        # need to filter volumes we can call the faster get_volume_list
        # method.
        if filters:
            LOG.debug('Getting all ds8k information')
            volumes, connection_info = self.helpers.\
                get_volumes_and_hosts(ds8k_context)
            LOG.debug(('Finished getting all ds8k information. Volumes: %s '
                      'Hosts: %s') % (volumes, connection_info))
        else:
            LOG.debug('Getting ds8k volume information')
            volumes = self.helpers.get_volume_list(ds8k_context)
            LOG.debug('Finished getting ds8k volume information %s' % volumes)
        # The helpers return both pool ID and pool name on separate keys of the
        # volume.  We put pool_name on the 'pool' key so the common code
        # puts the DS8K pool display name in the field used by the GUI.
        id_to_vol = {}
        for v in volumes.values():
            v['pool'] = v['pool_name']
            id_to_vol[v['id']] = v
        return_volumes = self._discover_volumes(context, volumes,
                                                connection_info,
                                                filters=filters)

        for v in return_volumes:
            # We need to copy the volume hex ID from restricted metadata to
            # the provider_location so the community driver has access to
            # the ID.
            hex_id = v['restricted_metadata'][RESTRICTED_METADATA_VDISK_ID_KEY]
            pl = six.text_type({'vol_hex_id': hex_id,
                                'replication_vol_hex_id': None})
            v['provider_location'] = pl
            be_vol = id_to_vol[hex_id]
            # We copy the data type key from the backend volume to the
            # Cinder volume metadata for OS400 volume types.
            # For 512 byte sector volumes we set the metadata the same as
            # the driver does for volume creation.
            os400_type = self._is_os400_type(be_vol)
            if os400_type:
                v['metadata'][DATA_TYPE_KEY] = be_vol[DATA_TYPE_KEY]
            elif be_vol[DATA_TYPE_KEY] == '900 FB 512':
                v['metadata'][DATA_TYPE_KEY] = 'FB 512'

        LOG.debug("discover_volumes return %s" % return_volumes)
        return return_volumes

    def check_for_deleted_volumes(self, context, volumes):
        ds8k_context = self.get_connection()
        # Set pool to empty to get all volumes in all pools
        ds8k_context['pool'] = ''
        backend_vols = self.helpers.get_volume_id_list(ds8k_context)
        return self._check_for_deleted_volumes(context, volumes, backend_vols)

    def do_setup(self, ctxt):
        """The community volume manager logs and swallows exceptions
        raised by do_setup when it calls it in init_host.
        Other volume drivers do not contact their backend during
        do_setup and their certificate validation exceptions come out
        later in init_host when volume stats are retrieved.

        DS8K contacts its backend in do_setup.
        Here we stash any exception that comes out so we can
        re-raise it when the PowerVC extended volume manager's init_host
        method re-calls check_for_setup_error
        """
        try:
            super(DS8KProductDriver, self).do_setup(ctxt)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                self._setup_exception = e

    def check_for_setup_error(self):
        super(DS8KProductDriver, self).check_for_setup_error()
        if self._setup_exception:
            raise self._setup_exception

    @staticmethod
    def _get_os400_A_type(backend_volume):
        # Returns os400 A type volume type from the backend volume
        # Returns None if the volume isn't an A type
        type = DS8KProductDriver._get_data_type_model(backend_volume)
        return type if type.startswith('A') else None

    @staticmethod
    def _get_data_type_model(backend_volume):
        if DATA_TYPE_KEY in backend_volume:
            dt = backend_volume[DATA_TYPE_KEY].upper()
            parts = dt.split()
            if len(parts) > 1 and len(parts[0]) == 3:
                return parts[0]
        return ''

    @staticmethod
    def _is_os400_type(backend_volume):
        type = DS8KProductDriver._get_data_type_model(backend_volume)
        if type in ds8k_nova_proxy.VALID_OS400_VOLUME_TYPES:
            return True
        return False

    def _check_supported_size(self, volume, volume_from_backend):
        a_type = self._get_os400_A_type(volume_from_backend)
        if a_type:
            # We support onboard of parital GiB sized volumes of OS400 A**
            # types because the size is ignored on the backend creates and
            # existing customers have A type volumes.
            # We need to round up to the next GiB so Cinder size.
            size = ds8k_nova_proxy.VALID_OS400_VOLUME_TYPES.get(a_type, None)
            if size:
                volume['size'] = size
            return
        # Check to see if the backend size is an exact multiple of GiB.
        remainder = volume_from_backend['size_in_bytes'] % (2 ** 30)
        if remainder:
            volume['support']['status'] = 'not_supported'
            if 'reasons' not in volume['support']:
                volume['support']['reasons'] = []
            reason = _('This volume is not a candidate for management '
                       'because its size is not a multiple of one gibibyte.')
            volume['support']['reasons'].append(reason)
            volume['block_vm_onboard'] = {'reasons': [reason]}


class DS8KProxy(ds8k_nova_proxy.DS8KNovaProxy):

    def _create_lun(self, lun, pool_id=None):
        # Override _create_lun to get the storage pool from the storage
        # template.

        # If a specific pool ID is passed in, use it.
        if pool_id:
            return super(DS8KProxy, self)._create_lun(lun, pool_id=pool_id)

        # Try to get the pool ID from the storage pool specified in the
        # extra spec.
        storage_pool = None
        pool_id = None
        if lun.volume_type_id:
            specs = volume_types.\
                get_volume_type_extra_specs(lun.volume_type_id)
            if specs:
                storage_pool = specs.get('drivers:storage_pool')
        if not storage_pool:
            # No storage pool found in the template.  Call parent method
            # and let it decide the pool.
            return super(DS8KProxy, self)._create_lun(lun, pool_id=pool_id)

        # Reverse look up pool ID for the name in the storage pool list.
        for pid, pool in self._storage_pools.iteritems():
            if pool['name'] == storage_pool:
                pool_id = pid
                break
        if not pool_id:
            msg = _('Unable to find the pool ID for the pool '
                    'name %(pool_name)s '
                    'in the list of pools %(pools)s' %
                    {'pool_name': storage_pool,
                     'pools': self._storage_pools.values()})
            raise exception.VolumeBackendAPIException(data=msg)
        return super(DS8KProxy, self)._create_lun(lun, pool_id=pool_id)

    def _get_pools(self, pool_str):
        # We override _get_pools to return all pools instead of just the
        # pools in the configuration.

        # Call super if the conf value isn't passed, this is currently
        # only in the migrate_volume case which PowerVC doesn't support.
        if pool_str != self.configuration.san_clustername:
            return super(DS8KProxy, self)._get_pools(pool_str)

        pools = self._client.fetchall('GET', '/pools')
        return OrderedDict(sorted(
            ((p['id'], {'node': int(p['node']),
                        'cap': int(p['cap']),
                        'capavail': int(p['capavail']),
                        'name': p['name'] + '_' + p['id'][1:]}) for p in pools
             ), key=lambda i: i[1]['capavail'], reverse=True))


class DS8KRESTConnector(ds8k_nova_proxy.AbstractRESTConnector):
    def __init__(self):
        super(DS8KRESTConnector, self).__init__()
        self.session = requests.Session()
        # Setup the adapter to use our added URLLIB3 Connection Pool
        self.session.mount('httpsds8k://', requests.adapters.HTTPAdapter())

    def send(self, method='', url='', headers={}, payload='', ):
        """Override the send to be able to do certificate validation"""
        self.session.verify = self.__get_rest_certificate(url)
        LOG.debug("url: %s" % url)
        url = url.replace('https://', 'httpsds8k://')
        try:
            # make a REST GET to the system
            resp = self.session.request(
                method, url, headers=headers, data=payload)
            return resp.status_code, resp.text
        except SSLError as exc:
            LOG.exception(six.text_type(exc))
            raise ssl.SSLError(six.text_type(exc))

    def __get_rest_certificate(self, url):
        """Helper method to get the certificate file to use for rest"""
        cert_file = re.sub('[:.]', '_', urlparse.urlsplit(
            url.strip()).hostname)
        cert_file = '/opt/ibm/powervc/data/certs/' + cert_file + '.crt'
        LOG.debug("cert_file: %s" % cert_file)
        # Use the certificate if it exists, otherwise use the System CA Bundle
        if os.path.exists(cert_file):
            return cert_file
        else:
            LOG.debug("cert_file not found, returning True")
            return True
