# =================================================================
#
# Copyright 2014, 2016 IBM Corp.
# All Rights Reserved.
# =================================================================
# Copyright (c) 2012 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import hashlib
from cinder import context, exception
from cinder.i18n import _, _LE, _LI, _LW
from cinder.volume.drivers.emc import emc_vmax_common, emc_vmax_masking, \
    emc_vmax_fast, emc_vmax_provision, emc_vmax_provision_v3
from cinder.volume.drivers.emc import emc_vmax_utils
from cinder.volume.drivers.emc.emc_vmax_common import CINDER_EMC_CONFIG_FILE, \
    CINDER_EMC_CONFIG_FILE_PREFIX, CINDER_EMC_CONFIG_FILE_POSTFIX, \
    FASTPOLICY
import copy
import os.path
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
import random
import time

import six

from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import restricted_metadata_keys as rmk
from powervc_oslo.wbem import pywbem


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
QUIESCE_TIME = 10

vmax_cfg_options = [
    # Set default wait period to be longer than the 60 retries by
    # default for the community driver.
    # Note that this default is also set separately by the registration code
    # in the cinder_emc_config_<id>.xml file. So the first two CONF options
    # below only apply when the community driver code is not using the
    # extra_specs from the XML config file.
    cfg.IntOpt('vmax_job_retries',
               default=720,
               help="Maximum number of retries to check for Job "
               "operation to complete. SMI-Provider polling option. This "
               "is only used if the <Retries> entry does not exist in the "
               "cinder_emc_config XML file."
               ),
    cfg.IntOpt('vmax_job_interval',
               default=5,
               help="Delay interval in seconds to wait before the next "
               "attempt. SMI-Provider polling option. This "
               "is only used if the <Retries> entry does not exist in the "
               "cinder_emc_config XML file."
               ),
    cfg.BoolOpt('vmax_allow_manage_inuse',
                default=False,
                help="Allow volumes to be managed that are marked as in-use "
                "by the VMAX provider. VMAX volumes may be marked as in-use "
                "because they participate in a masking group, but not be "
                "mapped to a VM."),
    cfg.BoolOpt('vmax_allow_cached_cim_inst',
                default=True,
                help="Allow volume CIM instances to be cached only during "
                "a particular request. This avoids volume lookup multiple "
                "times during a single request."),
    cfg.BoolOpt('vmax_wait_for_sync',
                default=True,
                help="Controls whether the VMAX cinder driver waits for "
                "replication synchronization to complete before returning "
                "from a clone operation API."),
    cfg.BoolOpt('vmax_restrict_pool_to_xml_conf',
                default=False,
                help="If True, then the VMAX cinder driver will only report "
                "a pool or pools matching what is specified in the XML "
                "configuration file. Storage template creation will be "
                "restricted as well."),
    cfg.BoolOpt('vmax_use_thin_capacity',
                default=False,
                help="If True, then storage pool free capacity reports as"
                "the remaining actual/thin capacity rather than the "
                "unsubscribed capacity."),
    cfg.BoolOpt('vmax_force_skip_verify',
                default=False,
                help="Override internal skip_verify flag and force it to be "
                "True if this option is set to True. Else honor the internal "
                "flag. This governs verification that a DeviceID for a LUN "
                "has not been reclaimed and reused by another management "
                "system (out-of-band for this PowerVC instance).")
]

CONF.register_opts(vmax_cfg_options)


# *****************************************************
# **            VMAX Common Extension                **
# *****************************************************
class EMCVMAXCommonExtension(emc_vmax_common.EMCVMAXCommon):
    """ PowerVC cinder volume extension to the EMC VMAX Common utility class
    """
    def __init__(self, prtcl, version, configuration=None):
        super(EMCVMAXCommonExtension, self).__init__(prtcl,
                                                     version,
                                                     configuration)
        # set the following overrides that hang off the 'common' class.
        self.masking = EMCVMAXMaskingExtension(prtcl)
        self.utils = EMCVMAXUtilsExtension(prtcl)
        self.provision = EMCVMAXProvisionExtension(prtcl)
        self.provisionv3 = EMCVMAXProvision3Extension(prtcl)
        # Just reference the one created instead of constructing new utils.
        self.provision.utils = self.utils
        self.provisionv3.utils = self.utils
        self.masking.utils = self.utils
        self.masking.provision.utils = self.utils
        self.fast.utils = self.utils
        self.fast.provision = self.provision
        self.fast.provision.utils = self.utils
        self.sync_jobs = {}
        self.first_time = True
        LOG.debug("EMCVMAXCommonExtension initialized. "
                  "CONF.vmax_allow_cached_cim_inst=%s, "
                  "CONF.vmax_wait_for_sync=%s, "
                  "CONF.vmax_restrict_pool_to_xml_conf=%s, "
                  "CONF.vmax_use_thin_capacity=%s, "
                  "CONF.vmax_force_skip_verify=%s." %
                  (CONF.vmax_allow_cached_cim_inst, CONF.vmax_wait_for_sync,
                   CONF.vmax_restrict_pool_to_xml_conf,
                   CONF.vmax_use_thin_capacity, CONF.vmax_force_skip_verify))

    def _get_array_id(self):
        # TODO: Expose EMC support for multi-array and multi-pool.
        return self.pool_info['arrays_info'][0]['SerialNumber']

    def _register_config_file_from_config_group(self, configGroupName):
        """Override this method on EMCVMAXCommon.
           Community has rejected bug 1364235. Their direction is to use
           cinder-multi-backend support instead of fixing this case.

        :param configGroupName: the config group name
        :returns: string configurationFile
        """
        LOG.debug("PowerVC Override.")
        if hasattr(self.configuration, 'cinder_emc_config_file'):
            configurationFile = self.configuration.cinder_emc_config_file
        elif configGroupName is None:
            # self._set_ecom_credentials(CINDER_EMC_CONFIG_FILE)
            return CINDER_EMC_CONFIG_FILE
        else:
            configurationFile = (
                ("%(prefix)s%(configGroupName)s%(postfix)s"
                 % {'prefix': CINDER_EMC_CONFIG_FILE_PREFIX,
                    'configGroupName': configGroupName,
                    'postfix': CINDER_EMC_CONFIG_FILE_POSTFIX}))

        # The file saved in self.configuration may not be the correct one,
        # double check.
        if configGroupName and configGroupName not in configurationFile:
            configurationFile = (
                ("%(prefix)s%(configGroupName)s%(postfix)s"
                 % {'prefix': CINDER_EMC_CONFIG_FILE_PREFIX,
                    'configGroupName': configGroupName,
                    'postfix': CINDER_EMC_CONFIG_FILE_POSTFIX}))

        if os.path.isfile(configurationFile):
            LOG.debug("Configuration file : %(configurationFile)s exists.",
                      {'configurationFile': configurationFile})
        else:
            exceptionMessage = (_(
                "Configuration file %(configurationFile)s does not exist.")
                % {'configurationFile': configurationFile})
            LOG.error(exceptionMessage)
            raise exception.VolumeBackendAPIException(data=exceptionMessage)

        # self._set_ecom_credentials(configurationFile)
        return configurationFile

    def _initial_setup(self, volume, volumeTypeId=None):
        """Necessary setup to accumulate the relevant information.

        This method overrides the method in:
            cinder.volume.drivers.emc.emc_vmax_common

        Currently we override to allow the volume's "type" to specify
        the pool name and the fast policy name if present, otherwise
        obtain the same way the VMAX driver does.
        Community will not likely accept a similar fix since their
        direction is to use cinder-multi-backend support instead.

        :param volume: the volume Object
        :returns: The extra specs dictionary
        """
        LOG.debug("PowerVC Override.")
        extraSpecs = super(EMCVMAXCommonExtension, self)._initial_setup(
            volume, volumeTypeId)
        # Volume type fields that override xml config file field are
        # handled here. If the following two things happen, then this
        # override method can go away:
        # 1. PowerVC switches to allow the 'storagetype' scope (not just
        #    'drivers') in build_default_opts() flow.
        # 2. EMC changes their driver to have the volume type extra_specs
        #    override the extra specs coming from the XML config file,
        #    rather than the other way around.
        driver_scope = 'storagetype:'
        pvc_scope = 'drivers:'
        for key in self._get_default_opts().keys():
            pvc_key = pvc_scope + key
            if pvc_key in extraSpecs:
                # The pool property is a special case for PVC.
                if key == 'storage_pool':
                    emc_key = driver_scope + 'pool'
                else:
                    emc_key = driver_scope + key
                # Map PowerVC key scope to cinder VMAX scope
                extraSpecs[emc_key] = extraSpecs[pvc_key]
        if 'storagetype:stripecount' in extraSpecs and\
                int(extraSpecs['storagetype:stripecount']) == 1 and\
                'storagetype:compositetype' in extraSpecs and\
                extraSpecs['storagetype:compositetype'] == 'striped':
            LOG.debug("Converting compositetype to 'concatenated' since "
                      "stripecount is 1.")
            extraSpecs['storagetype:compositetype'] = 'concatenated'

        LOG.debug("Aggregate extra_specs: %s " % extraSpecs)
        if self.sync_jobs:
            # Query for and log some info about synchronization jobs tracked.
            # Note that a special config option is required to get into this
            # code block and if we end up supporting this, then we would
            # need a periodic task to handle this work rather than doing
            # it inline here.
            self._update_sync_job_data()
        return extraSpecs

    def _validate_pool(self, volume):
        """
        PowerVC Override. Always say pool is valid. We don't tie into thier
        multi-pool stuff yet.
        TODO: Remove after EMC fixes this.
        """
        return None
#         if 'provider_location' in volume and not volume['provider_location']:
#             LOG.debug("PowerVC override returns None.")
#             return None
#         if 'metadata' in volume and 'storage_pool' in volume['metadata']:
#             LOG.debug("PowerVC override returns %s" %
#                       volume['metadata']['storage_pool'])
#             return volume['metadata']['storage_pool']
#         ctxt = context.get_admin_context()
#         meta = db_api.volume_metadata_get(ctxt, volume['id'])
#         if meta and 'storage_pool' in meta:
#             LOG.debug("PowerVC override returns db queried: %s" %
#                       meta['storage_pool'])
#             return meta['storage_pool']
#         # fallback
#         return super(EMCVMAXCommonExtension, self)._validate_pool(volume)

    @lockutils.synchronized('syncJobs', 'vmax-')
    def _update_sync_job_data(self):
        """ Run through sync jobs, update percent complete info, and
            break clone relationships if jobs are done. This is similar
            to flash copy management.
        """
        to_remove = []
        for vol_id, sync_data in self.sync_jobs.iteritems():
            percentSynced = 0
            try:
                syncInstance = self.conn.GetInstance(
                    sync_data['syncName'], LocalOnly=False)
                percentSynced = syncInstance['PercentSynced']
            except Exception as ex:
                LOG.exception(ex)
                LOG.info("Exception taken, so presume the sync job is "
                         "gone and the operation is complete for %s" %
                         vol_id)
                percentSynced = 100
            LOG.debug("Clone name '%s' with ID %s goes from %d to %d "
                      "percent synced." % (sync_data['cloneName'],
                                           vol_id, sync_data['percentSynced'],
                                           percentSynced))
            sync_data['percentSynced'] = percentSynced
            if percentSynced >= 100:
                to_remove.append(vol_id)
                # Remove the Clone relationship so it can be used as a
                # regular lun.
                try:
                    LOG.debug("Break clone relationship for vol %s." % vol_id)
                    rc, job = self.provision.delete_clone_relationship(
                        self.conn, sync_data['repServiceInstanceName'],
                        sync_data['syncName'], sync_data['extraSpecs'],
                        force=True)
                    LOG.debug("delete_clone_relationship for sync data %s "
                              "returns rc=%d, job=%s." % (sync_data, rc, job))
                except Exception as ex:
                    LOG.exception(ex)
                    LOG.debug("Continue after exception with: %s" % sync_data)
        # Any item that reports as complete will be removed from dict even
        # if sync rel did not get successfully removed.
        for vol_id in to_remove:
            LOG.debug("Removing completed sync job: %s" %
                      self.sync_jobs[vol_id])
            del self.sync_jobs[vol_id]

    def _get_default_opts(self):
        """
        Define the default options for the default volume type extra_specs
        in the "drivers:" scope. The drivers scope is a PowerVC convention
        for extra_specs options that the cinder driver honors. So we
        provide this information to integrate the EMC driver into PowerVC.
        Note that the EMC VMAX driver only supports 'thin' pools.
        """
        default_opts = {'storage_pool':
                        self.pool_info['arrays_info'][0]['PoolName'],
                        'fastpolicy': None,
                        'stripecount': '1',
                        'slo': None,
                        'workload': None,
                        'provision_type': 'thin'
                        }
        return default_opts

    def update_volume_stats(self):
        """Override the Volume Stats to return all of the Pools"""
        total, free = (0, 0)
        # Call the Parent Class to retrieve most of the Statistics
        stats = super(EMCVMAXCommonExtension, self).update_volume_stats()
        LOG.debug("volume_stats from EMC: %s" % stats)
        # add multiattach support for powervc
        for pool in stats['pools']:
            pool['multiattach'] = True

        if 'stg_sys_inst_name' in self.pool_info and 'stg_sys_name'\
                in self.pool_info:
            stg_sys_inst_name = self.pool_info['stg_sys_inst_name']
        else:
            # TODO: Refactor to expose multi-array/pool support (would
            #       require PowerVC enhancement to handle that type of
            #       driver support)
            stg_sys_inst_name = self.utils.find_storageSystem(
                self.conn, self._get_array_id())
            self.pool_info['stg_sys_inst_name'] = stg_sys_inst_name
            self.pool_info['stg_sys_name'] = stg_sys_inst_name['Name']
            self.pool_info['is_v3'] = self.utils.isArrayV3(
                self.conn, self._get_array_id())
            if not self.masking.pool_info:
                self.masking.pool_info = self.pool_info

        if 'provider' not in self.pool_info:
            smis = self.conn.EnumerateInstances(
                'SE_ManagementServerSoftwareIdentity')
            if smis:
                self.pool_info['provider'] = smis[0]['ElementName']
            else:
                self.pool_info['provider'] = 'unknown'

        # Retrieve the Capacity from each of the Pools and use the total
        # if V2 and special config opt is not set to use single pool.
        pools = []
        if not (self.pool_info['is_v3'] or
                CONF.vmax_restrict_pool_to_xml_conf):
            try:
                pools = self.conn.Associators(
                    stg_sys_inst_name,
                    resultClass='EMC_VirtualProvisioningPool')
            except Exception as ex:
                LOG.debug("Could not get EMC_VirtualProvisioningPool: %s" %
                          six.text_type(ex))
        if len(pools) > 1:
            # Note that in 1.2.3, for the FAST policy support, it was decided
            # that we would not accumulate the total for the pools associated
            # with a fast policy on the volume-pools API - because it would
            # be to confusing for customers using our UI. However, this method
            # is different from volume-pools API flow and the pool capacities
            # coming from the EMC driver code here may not match the capacities
            # reported for the volume-pools API. This is expected.
            for p in pools:
                ptotal = self.utils.convert_bits_to_gbs(p['TotalManagedSpace'])
                palloc = self.utils.convert_bits_to_gbs(
                    p['EMCSubscribedCapacity'])
                pfree = self.utils.convert_bits_to_gbs(
                    p['RemainingManagedSpace'])
                ptotal = max(ptotal, 0)
                unsubscribed = ptotal - palloc
                if unsubscribed < 0:
                    LOG.debug("Capping negative unsubscribed of %d to 0." %
                              unsubscribed)
                    unsubscribed = 0
                total += ptotal
                if CONF.vmax_use_thin_capacity:
                    pfree = max(pfree, 0)
                    free += pfree
                else:
                    # Normal case consistent with EMC driver.
                    free += unsubscribed

        if (len(pools) == 1 or CONF.vmax_restrict_pool_to_xml_conf or
                self.pool_info['is_v3']):
            # Get total values from the pool info that driver gave us if
            # they are not already setting this info in stats. This is
            # important for vmax3 as we don't use PowerVC's multiple pool
            # support and so fall into this branch for V3.
            if stats['free_capacity_gb'] == 0:
                stats['total_capacity_gb'] =\
                    stats['pools'][0]['total_capacity_gb']
                stats['free_capacity_gb'] =\
                    stats['pools'][0]['free_capacity_gb']
        else:
            # Get values from vars accumulating totals in earlier pool loop
            stats['total_capacity_gb'] = total
            stats['free_capacity_gb'] = free

        stats['storage_protocol'] = 'FC'  # override 'unknown'
        self.first_time = False  # Have connected and collected at least once
        LOG.debug("VMAX stats refreshed for PowerVC: %s." % stats)
        return stats

    def find_device_number(self, volume, host):
        """
        PowerVC version until community driver is fixed to honor finding
        the device number with respect to a set of initiator wwpns.

        TODO: The LP# 1382641 community fix is suspected to NOT be a
              complete solution. Leave this here until we can collect proper
              data to open new bug and have community resolve if needed.

              LP# 1498964 was not a complete fix either. The new one to
              follow is 1532283.
        """
        if 'initiator_wwpns' not in volume:
            return super(EMCVMAXCommonExtension, self).find_device_number(
                volume, host)
        # otherwise, use our override
        wwpns = [wwpn.lower() for wwpn in volume['initiator_wwpns']]
        foundNumDeviceNumber = None
        lunMaskingView = None
        volumeInstance = self._find_lun(volume)
        LOG.debug("PowerVC override. volume_devid=%s, wwpns=%s" %
                  (volumeInstance['DeviceID'], wwpns))
        storageSystemName = volumeInstance['SystemName']
        unitnames = self.conn.ReferenceNames(
            volumeInstance.path,
            ResultClass='CIM_ProtocolControllerForUnit')

        for unitname in unitnames:
            controller = unitname['Antecedent']
            classname = controller['CreationClassName']
            index = classname.find('Symm_LunMaskingView')
            if index > -1:
                LOG.debug("PowerVC check initiator ports for %s." % controller)
                hwid_inst_names = self.conn.AssociatorNames(
                    controller, ResultClass='EMC_StorageHardwareID')
                hwid_wwpns = self.utils.hwid_names_to_wwpns(hwid_inst_names)
                found_match = None
                for hwid in hwid_wwpns:
                    if hwid in wwpns:
                        found_match = hwid
                        break
                # do we need to special case if connector or wwpns is not
                # passed in?
                if found_match:
                    LOG.debug("First matching wwpn: %s" % found_match)
                    unitinstance = self.conn.GetInstance(unitname,
                                                         LocalOnly=False)
                    foundNumDeviceNumber = int(unitinstance['DeviceNumber'],
                                               16)
                    lunMaskingView = controller
                    break
                else:
                    LOG.debug("Did not find matching initiator wwpn in: "
                              "%s" % hwid_wwpns)

        if foundNumDeviceNumber is None:
            LOG.debug(
                "Device number not found for volume "
                "%(volumeName)s %(volumeInstance)s."
                % {'volumeName': volume['name'],
                   'volumeInstance': volumeInstance.path})

        data = {'hostlunid': foundNumDeviceNumber,
                'storagesystem': storageSystemName,
                'maskingview': lunMaskingView
                }

        LOG.debug("Device info: %(data)s." % {'data': data})
        return data

    def _set_ecom_credentials(self, arrayInfo):
        """
        Override the Set Credentials method to handle backwards
        compatibility with 1.2.2.x registrations.
        """
        super(EMCVMAXCommonExtension, self)._set_ecom_credentials(arrayInfo)
        if not self.ecomCACert:
            # 'Cert' is the old xml key for 1.2.2.x
            self.ecomCACert = self._parse_from_file(
                self.configuration.cinder_emc_config_file, 'Cert')
            if self.ecomCACert:
                LOG.debug('Picked up Cert instead of EcomCACert.')
            else:
                LOG.warn(_LW("Could not find certificate reference in "
                             "configuration file for secure https."))

    def _parse_from_file(self, fileName, stringToParse):
        """Parse the string from XML.
        OLD method removed from emc_vmax_utils.
        TODO: Remove this after PowerVC upgrade processing is enhanced
              to convert old registration format to new Liberty format.
        """
        from xml.dom import minidom
        retString = None
        myFile = open(fileName, 'r')
        data = myFile.read()
        myFile.close()
        dom = minidom.parseString(data)
        tag = dom.getElementsByTagName(stringToParse)
        if tag is not None and len(tag) > 0:
            strXml = tag[0].toxml()
            strXml = strXml.replace('<%s>' % stringToParse, '')
            strXml = strXml.replace('\n', '')
            strXml = strXml.replace('\r', '')
            strXml = strXml.replace('\t', '')
            retString = strXml.replace('</%s>' % stringToParse, '')
            retString = retString.strip()
        return retString

    def _get_ecom_connection(self):
        """
        Override the Get Connection method to not pass x509.
        The community flow still does not connect successfully.
        """
        # Continue to replace 'http:' to force a failure in case secure
        # options were manually removed from the XML config file.
        self.url = self.url.replace('http://', 'https://')
        conn = pywbem.WBEMConnection(self.url, (self.user, self.passwd),
                                     default_namespace='root/emc', x509=None,
                                     verify_callback=None,
                                     ca_certs=self.ecomCACert,
                                     no_verification=self.ecomNoVerification)
        # If we can't connect to the SMI-S Proxy, throw an exception
        if conn is None:
            exception_message = (_LE("Cannot connect to ECOM server"))
            raise exception.VolumeBackendAPIException(data=exception_message)
        conn.debug = True
        return conn

    def _find_lun(self, volume, skip_verify=False, skip_cache=False):
        """
        PowerVC override to avoid multiple lookups per request and to do
        validation that the original volume was not deleted out-of-band
        and the deviceID is re-used. The community handles this in a way
        not compatible with our current implementation.
        :param skip_verify: Optional param. If True, then will not check and
                    raise exception if backend instance does not match
                    correlation restricted_metadata we track. Otherwise
                    the verification check is done.
        :param skip_cache: Optional param. If True, then skip the cache and
                    always perform the lookup.
        """
        if CONF.vmax_allow_cached_cim_inst and '_retrieved_inst' in volume\
                and not skip_cache:
            LOG.debug("PowerVC override to returned already retrieved "
                      "instance: %s" % (volume['_retrieved_inst']))
            return volume['_retrieved_inst']
        else:
            lu = super(EMCVMAXCommonExtension, self)._find_lun(volume)
            if lu:
                # Handle the case where the original volume was deleted and
                # the device ID was re-used for a different volume.
                emc_name = lu['ElementName']
                if (emc_name and not skip_verify and
                        not CONF.vmax_force_skip_verify):
                    metad = None
                    if 'restricted_metadata' in volume:
                        metad = volume['restricted_metadata']
                    else:
                        try:
                            grm = powervc_db_api.volume_restricted_metadata_get
                            metad = grm(context.get_admin_context(),
                                        volume['id'])
                        except Exception as ex:
                            LOG.exception(ex)
                    # LOG.debug("restricted_metadata=%s" % metad)
                    if not metad or rmk.VDISK_NAME_KEY not in metad:
                        # This can happen if volume just created and metadata
                        # is not persisted yet.
                        LOG.info(_LI("PowerVC metadata looked up is "
                                     "incomplete for VMAX volume: %(data)s ") %
                                 {'data': metad})
                    elif metad[rmk.VDISK_NAME_KEY] != emc_name:
                        LOG.info(
                            _LI("Volume %(vid)s EMC name '%(mname)s' has "
                                "changed on the backend to %(emc_name)s. "
                                "Lookup fails.") %
                            dict(vid=volume['id'],
                                 mname=metad[rmk.VDISK_NAME_KEY],
                                 emc_name=emc_name))
                        # The lun_status here is temporary - not persisted.
                        # The health monitor will periodically call
                        # check_for_deleted_volumes() and persist the actual
                        # state/status properties accordingly.
                        volume['lun_status'] = _LI("Deleted out-of-band "
                                                   "and DeviceID reused")
                        return None
                volume['_retrieved_inst'] = lu
            else:
                # The status here is temporary - not persisted.
                volume['lun_status'] = _LI("Not found")
            return lu

    def _create_v2_replica_and_delete_clone_relationship(
            self, repServiceInstanceName, cloneVolume, sourceVolume,
            sourceInstance, targetInstance, extraSpecs, isSnapshot=False):
        """
        Create a replica and delete the clone relationship.

        PowerVC Override:
            If targetInstance is None, we need to create it before
            calling create_element_replica() such that the target has
            the extraSpecs from the desired storage template plus the
            disk geometry of the source volume.

        :param repServiceInstanceName: the replication service
        :param cloneVolume: the clone volume object
        :param sourceVolume: the source volume object
        :param sourceInstance: the source volume instance
        :param targetInstance: the target volume instance
        :param extraSpecs: extra specifications
        :param isSnapshot: check to see if it is a snapshot
        :returns: int -- return code
        :returns: dict -- cloneDict
        """
        sourceName = sourceVolume['name']
        cloneName = cloneVolume['name']
        poolInstName, storageSystemName = (
            self._get_pool_and_storage_system(extraSpecs))
        srcPoolInstName = ""  # init
        # Should be safe going just for V2 pool types since this is a V2
        # method. The idea here is to check if source and target pools
        # are the same. If so, then we don't need to pre-create target.
        src_pools = self.conn.AssociatorNames(
            sourceInstance.path,
            ResultClass='Symm_VirtualProvisioningPool')
        if src_pools:
            srcPoolInstName = src_pools[0]
        LOG.debug("Source poolInst is %s, Target poolInst is %s." %
                  (srcPoolInstName, poolInstName))

        # Check for targetInstance and create if needed.
        # If the source is a striped volume however, we cannot pre-create
        # the target. Also, skip pre-create if target pool is same as source
        # since that will match the community driver logic better.
        if not targetInstance and\
                extraSpecs['storagetype:compositetype'] != 'striped' and\
                srcPoolInstName != poolInstName:
            # Set-up for calling the lower-level create_composite_volume()
            # method in the same fashion that create_volume() does,
            # except that we hard code the member count to be 1 and the
            # composite type to be 'concatenated'. The reason we do this is
            # because in our override flow (1) we know that targetInstance
            # does not get passed only when the source is a simple volume
            # of this nature and (2) we need to override those two extra
            # specs since they must get ignored from the target volume type.
            # If they are not ignored, then volume copy that happens later
            # can fail.
            # First, determine the target volume size from the source volume
            # size in bytes, not converted from the cinder 'size' in GiB,
            # which may have a rounding error for on-boarded volumes.
            volumeSize = int(
                self.utils.get_volume_size(self.conn, sourceInstance))
            if (volumeSize != (int(sourceVolume['size']) * (2 ** 30))):
                LOG.info(_("Source volume '%(name)s' size of %(ssb)d bytes "
                           "does not match cinder volume size of %(gbs)s GiB."
                           " Bytes difference=%(diff)d.")
                         % {'name': sourceName, 'ssb': volumeSize,
                            'gbs': sourceVolume['size'], 'diff':
                            volumeSize - (int(sourceVolume['size']) *
                                          (2 ** 30))})
            volumeName = cloneVolume['id']
            compositeType = self.utils.get_composite_type('concatenated')
            elementCompositionService = (
                self.utils.find_element_composition_service(self.conn,
                                                            storageSystemName))
            LOG.debug("PowerVC: Create concatenated clone volume target. "
                      "Volume: %(cloneName)s, Source Volume: %(sourceName)s, "
                      "SourceElement: %(sourceelement)s, "
                      "extra_specs: %(extraspecs)s, "
                      "poolInstName=%(poolInstName)s, "
                      "Stripecount=1."
                      % {'cloneName': cloneName, 'sourceName': sourceName,
                         'sourceelement': six.text_type(sourceInstance.path),
                         'extraspecs': extraSpecs,
                         'poolInstName': poolInstName})

            # Create the target volume shell thin device.
            target_voldict, rc = self.provision.create_composite_volume(
                self.conn, elementCompositionService, volumeSize, volumeName,
                poolInstName, compositeType, 1, extraSpecs)
            # rc is guaranteed to be 0 at this point.
            LOG.debug("Created thin concatenated volume for cinder id %s, "
                      "RC=%d." % (volumeName, rc))

            # Add provider location so that we can look up the CIM inst.
            cloneVolume['provider_location'] = six.text_type(target_voldict)
            targetInstance = self._find_lun(cloneVolume, skip_verify=True)
            if not targetInstance:
                message = (_("Could not locate the clone volume target "
                             "on the array for provider_location: %(loc)s.")
                           % {'loc': cloneVolume['provider_location']})
                raise exception.VolumeDriverException(message=message)
            LOG.debug("Target thin volume created per extra specs. Now make "
                      "an ElementReplica into: %s" %
                      six.text_type(targetInstance.path))
        else:
            LOG.debug("Target instance already created or source is striped "
                      "or target pool matches source. "
                      "Meta-config=%s. Pool=%s. Continue with ElementReplica "
                      "into: %s" %
                      (extraSpecs['storagetype:compositetype'], poolInstName,
                       six.text_type(targetInstance.path) if
                       targetInstance else 'none.'))

        # TODO: Call the super method. For now we keep the super processing
        #       inline to investigate not waiting for synchronization
        #       (aka flash copy) to complete.
        try:
            rc, job = self.provision.create_element_replica(
                self.conn, repServiceInstanceName, cloneName, sourceName,
                sourceInstance, targetInstance, extraSpecs)
        except Exception:
            LOG.warning(_LW(
                "Clone failed. Cleaning up the target volume. "
                "Clone name: %(cloneName)s "
                "Source name: %(sourceName)s "
                "Extra specs: %(extraSpecs)s "),
                {'cloneName': cloneName,
                 'sourceName': sourceName,
                 'extraSpecs': extraSpecs})
            if targetInstance is not None:
                # Check if the copy session exists.
                storageSystem = targetInstance['SystemName']
                syncInstanceName = self.utils.find_sync_sv_by_target(
                    self.conn, storageSystem, targetInstance, False)
                if syncInstanceName is not None:
                    # Remove the Clone relationship.
                    rc, job = self.provision.delete_clone_relationship(
                        self.conn, repServiceInstanceName, syncInstanceName,
                        extraSpecs, force=True)
                storageConfigService = (
                    self.utils.find_storage_configuration_service(
                        self.conn, storageSystem))
                self._delete_target_volume_v2(
                    storageConfigService, targetInstance, extraSpecs)
                if 'provider_location' in cloneVolume:
                    # clear it since rollback already performed.
                    cloneVolume['provider_location'] = ""
            # Re-throw the exception.
            raise
        cloneDict = self.provision.get_volume_dict_from_job(
            self.conn, job['Job'])

        fastPolicyName = extraSpecs[FASTPOLICY]
        if isSnapshot:
            if fastPolicyName is not None:
                storageSystemName = sourceInstance['SystemName']
                self._add_clone_to_default_storage_group(
                    fastPolicyName, storageSystemName, cloneDict, cloneName)
            LOG.info(_LI("Snapshot creation %(cloneName)s completed. "
                         "Source Volume: %(sourceName)s."),
                     {'cloneName': cloneName,
                      'sourceName': sourceName})

            return rc, cloneDict

        cloneVolume['provider_location'] = six.text_type(cloneDict)
        # Note - this point, the cloneVolume will be cleaned up at the
        # top emc_vmax_product driver level if an exception is raised later
        # in this method. By Liberty, there will be community fix that
        # correctly handles the cleanup within this method in a parallel
        # fashion to vmax V3 cleanup. TODO: We will re-evaluate then.
        syncInstanceName, storageSystemName = (
            self._find_storage_sync_sv_sv(
                cloneVolume, sourceVolume, extraSpecs,
                waitforsync=CONF.vmax_wait_for_sync))

        if CONF.vmax_wait_for_sync:
            # Remove the Clone relationship so it can be used as a regular lun.
            rc, job = self.provision.delete_clone_relationship(
                self.conn, repServiceInstanceName, syncInstanceName,
                extraSpecs, force=True)
        elif syncInstanceName:
            # Don't break the clone relationship, but keep track of it
            # so that it can be managed and removed later.
            sync_data = {'syncName': syncInstanceName,
                         'cloneLoc': cloneVolume['provider_location'],
                         'repServiceInstanceName': repServiceInstanceName,
                         'extraSpecs': copy.deepcopy(extraSpecs),
                         'cloneName': cloneName,
                         'percentSynced': 0}
            self.sync_jobs[cloneVolume['id']] = sync_data
            LOG.debug("Added synchronization entry for cloned ID %s: %s" %
                      (cloneVolume['id'], sync_data))
        else:
            LOG.warn(_("The find synchronization call did not return a "
                       "Sync instance of SE_StorageSynchronized_SV_SV"))

        # I wonder if the vmax driver is incorrect in adding the volume
        # to the default FAST policy group after the synchronization is
        # complete?
        LOG.debug("FAST policy is: %s." % fastPolicyName)
        if fastPolicyName is not None:
            self._add_clone_to_default_storage_group(
                fastPolicyName, storageSystemName, cloneDict, cloneName,
                extraSpecs)
        return rc, cloneDict

    def _create_clone_v2(self, repServiceInstanceName, cloneVolume,
                         sourceVolume, sourceInstance, isSnapshot, extraSpecs):
        """Create a clone (v2).
           PowerVC override. TODO: This is a large method, but given the
           discussion with EMC during 1.2.3 development, it is likely to
           remain technical debt for some time since they treat the
           extra specs differently and have multi-backend support.

        :param repServiceInstanceName: the replication service
        :param cloneVolume: the clone volume object
        :param sourceVolume: the source volume object
        :param sourceInstance: the device ID of the volume
        :param extraSpecs: extra specifications
        :param isSnapshot: check to see if it is a snapshot
        :returns: int -- return code
        :raises: VolumeBackendAPIException
        """
        # Check if the source volume contains any meta devices.
        metaHeadInstanceName = self.utils.get_volume_meta_head(
            self.conn, sourceInstance.path)

        if metaHeadInstanceName is None:  # Simple volume.
            LOG.debug("No meta-head. Call PowerVC override for simple "
                      "volume cloning.")
            return self._create_v2_replica_and_delete_clone_relationship(
                repServiceInstanceName, cloneVolume, sourceVolume,
                sourceInstance, None, extraSpecs, isSnapshot=isSnapshot)
        else:
            # Composite volume with meta device members.
            # Check the meta members capacity info.
            elementCompositionService = (
                self.utils.find_element_composition_service(
                    self.conn, self.pool_info['stg_sys_name']))
            metaMemberInstanceNames = self.utils.get_volume_members(
                self.conn, sourceInstance, metaHeadInstanceName,
                elementCompositionService)
            volumeCapacities = self.utils.get_meta_members_capacity_in_byte(
                self.conn, metaMemberInstanceNames)
            # Override extra specs such that target vol matches source for
            # the meta configuration (geometry)
            cloneExtraSpecs = copy.deepcopy(extraSpecs)
            metaHeadInst = self.conn.GetInstance(metaHeadInstanceName)
            propList = metaHeadInst.properties.items()
            LOG.debug("metaHead propList: %s " % propList)
            isConcatenated = 'True'
            for properties in propList:
                if properties[0] == 'IsConcatenated':
                    isConcatenated = six.text_type(properties[1].value)
            sourceType = 'concatenated' if isConcatenated == 'True'\
                else 'striped'
            cloneExtraSpecs['storagetype:compositetype'] = sourceType
            cloneExtraSpecs['storagetype:membercount'] =\
                cloneExtraSpecs['storagetype:stripecount'] =\
                str(len(volumeCapacities))
            # If striped, the fastpolicy needs to inherit from the source
            # volume as well.
            if sourceType == 'striped':
                cloneExtraSpecs[FASTPOLICY] =\
                    self._get_fast_policy_for_vol(sourceInstance)
            LOG.debug("SourceType=%s, %d source Meta volumes, Capacities=%s. "
                      "CloneExtraSpecs: %s" %
                      (sourceType, len(volumeCapacities), volumeCapacities,
                       cloneExtraSpecs))

            # If only one meta or striped, do the same as above. But for
            # multiple metas, even if all the same size, fall through to loop
            # processing below. (PowerVC change). If source volume is
            # striped, then we can assume all the metas are the same size.
            if len(volumeCapacities) == 1 or sourceType == 'striped':
                LOG.debug("Only one meta or striped: %s" %
                          sourceVolume['name'])
                return self._create_v2_replica_and_delete_clone_relationship(
                    repServiceInstanceName, cloneVolume, sourceVolume,
                    sourceInstance, None, cloneExtraSpecs, isSnapshot)

            baseTargetVolumeInstance = None
            # Force member count to be 1 for a base concatenated vol.
            cloneExtraSpecs['storagetype:membercount'] = '1'
            LOG.debug("Volume capacities:  %s. With membercount=1, "
                      "continue on to creating the target meta volume for %s."
                      % (volumeCapacities, sourceVolume['name']))
            for volumeSizeInbits in volumeCapacities:
                if baseTargetVolumeInstance is None:  # Create base volume.
                    baseVolumeName = "TargetBaseVol"
                    volume = {'size': int(self.utils.convert_bits_to_gbs(
                        volumeSizeInbits))}
                    __, baseVolumeDict, storageSystemName = (
                        self._create_composite_volume(volume, baseVolumeName,
                                                      volumeSizeInbits,
                                                      cloneExtraSpecs))
                    baseTargetVolumeInstance = self.utils.find_volume_instance(
                        self.conn, baseVolumeDict, baseVolumeName)
                    LOG.debug("Base target volume %(targetVol)s created. "
                              "capacity requested: %(capInBits)lu.",
                              {'capInBits': volumeSizeInbits,
                               'targetVol': baseTargetVolumeInstance.path})
                    ret_base_size = int(self.utils.get_volume_size(
                        self.conn, baseTargetVolumeInstance))
                    if ret_base_size != volumeSizeInbits:
                        LOG.warn("Source=%s, Target=base. The created target "
                                 "size of %d does not match requested size of "
                                 "%d." % (sourceVolume['name'], ret_base_size,
                                          volumeSizeInbits))
                else:  # Create append volume
                    targetVolumeName = "MetaVol"
                    volume = {'size': int(self.utils.convert_bits_to_gbs(
                        volumeSizeInbits))}
                    storageConfigService = (
                        self.utils.find_storage_configuration_service(
                            self.conn, storageSystemName))
                    unboundVolumeInstance = (
                        self._create_and_get_unbound_volume(
                            self.conn, storageConfigService,
                            baseTargetVolumeInstance.path, volumeSizeInbits,
                            cloneExtraSpecs))
                    if unboundVolumeInstance is None:
                        exceptionMessage = (_LE(
                            "Error Creating unbound volume."))
                        LOG.error(exceptionMessage)
                        # Remove target volume (cleanup)
                        self._delete_target_volume_v2(storageConfigService,
                                                      baseTargetVolumeInstance,
                                                      cloneExtraSpecs)
                        raise exception.VolumeBackendAPIException(
                            data=exceptionMessage)
                    LOG.debug("New unbound volume member %s to append: %s " %
                              (targetVolumeName,
                               six.text_type(unboundVolumeInstance.path)))
                    ret_meta_size = int(self.utils.get_volume_size(
                        self.conn, unboundVolumeInstance))
                    if ret_meta_size != volumeSizeInbits:
                        LOG.warn("Source=%s, Target=meta. The created target "
                                 "size of %d does not match requested size of "
                                 "%d." % (sourceVolume['name'], ret_meta_size,
                                          volumeSizeInbits))

                    # Append the new unbound volume to the
                    # base target composite volume.
                    baseTargetVolumeInstance = self.utils.find_volume_instance(
                        self.conn, baseVolumeDict, baseVolumeName)
                    try:
                        elementCompositionService = (
                            self.utils.find_element_composition_service(
                                self.conn, storageSystemName))
                        compositeType = self.utils.get_composite_type(
                            cloneExtraSpecs['storagetype:compositetype'])
                        modifiedVolumeDict = None
                        __, modifiedVolumeDict = (
                            self._modify_and_get_composite_volume_instance(
                                self.conn,
                                elementCompositionService,
                                baseTargetVolumeInstance,
                                unboundVolumeInstance.path,
                                targetVolumeName, compositeType,
                                cloneExtraSpecs))
                        if modifiedVolumeDict is None:
                            exceptionMessage = (_LE(
                                "Error appending volume %(volumename)s to "
                                "target base volume.")
                                % {'volumename': targetVolumeName})
                            LOG.error(exceptionMessage)
                            raise exception.VolumeBackendAPIException(
                                data=exceptionMessage)
                    except Exception:
                        exceptionMessage = (_LE(
                            "Exception appending meta volume to target volume "
                            "%(volumename)s.")
                            % {'volumename': baseVolumeName})
                        LOG.error(exceptionMessage)
                        # Remove append volume and target base volume
                        self._delete_target_volume_v2(
                            storageConfigService, unboundVolumeInstance,
                            cloneExtraSpecs)
                        self._delete_target_volume_v2(
                            storageConfigService, baseTargetVolumeInstance,
                            cloneExtraSpecs)
                        LOG.info(_LI("Rollback happened for base and append "
                                     "volume."))
                        raise

            LOG.debug("Create V2 replica for multiple meta members.")
            return self._create_v2_replica_and_delete_clone_relationship(
                repServiceInstanceName, cloneVolume, sourceVolume,
                sourceInstance, baseTargetVolumeInstance, cloneExtraSpecs,
                isSnapshot)

    def _get_fast_policy_for_vol(self, volumeInstance):
        """ Return fast policy controlling volume if one, else None """
        sg_instNames = self.conn.AssociatorNames(
            volumeInstance.path, ResultClass='CIM_DeviceMaskingGroup')
        if sg_instNames:
            prex = emc_vmax_fast.DEFAULT_SG_PREFIX
            postx = emc_vmax_fast.DEFAULT_SG_POSTFIX
            for sg_instName in sg_instNames:
                if prex in sg_instName['InstanceID']:
                    # found a fast policy
                    tail = sg_instName['InstanceID'].split(prex)[1]
                    fastpolicy = tail.split(postx)[0]
                    LOG.debug("Volume %s is under fast control of: %s " %
                              (volumeInstance.path['DeviceID'], fastpolicy))
                    return fastpolicy
        return None

    def _get_volume_dict_from_path(self, volpath):
        """Given the volume path, determine the volume dictionary.
        This is also known as the provider_location.
        Compare to emc_vmax_provision->get_volume_dict_from_job().
        """
        volumeDict = {}
        volumeDict['classname'] = volpath.classname
        keys = {}
        keys['CreationClassName'] = volpath['CreationClassName']
        keys['SystemName'] = volpath['SystemName']
        keys['DeviceID'] = volpath['DeviceID']
        keys['SystemCreationClassName'] = volpath['SystemCreationClassName']
        volumeDict['keybindings'] = keys
        return volumeDict

    def _check_and_remove_sync_rels(self, volume):
        """ Helper to do work of checking for and removing any
        SE_StorageSynchronized_SV_SV clone relationships.
        TODO: I don't know if this fix makes sense for the community(?)
              This can probably be removed with a more robust driver in place
              for Liberty.
        """
        vol_dict = ast.literal_eval(volume['provider_location'])
        vol_path = self.utils.get_instance_name(
            vol_dict['classname'], vol_dict['keybindings'])
        storageSystem = vol_dict['keybindings']['SystemName']
        sv_sync_ans = self.conn.AssociatorNames(
            vol_path, AssocClass='SE_StorageSynchronized_SV_SV',
            ResultClass='Symm_StorageVolume')
        LOG.debug("SE_StorageSynchronized_SV_SV association targets of "
                  "%s: %s." % (vol_dict['keybindings']['DeviceID'],
                               sv_sync_ans))
        if not sv_sync_ans:
            return
        repSvc = self.utils.find_replication_service(
            self.conn, storageSystem)
        num_removed = 0
        for targ in sv_sync_ans:
            # Try finding the actual association by using one direction...
            sync_path = self.utils.get_instance_name(
                'SE_StorageSynchronized_SV_SV',
                {'SyncedElement': vol_path, 'SystemElement': targ})
            try:
                sync_inst = self.conn.GetInstance(sync_path)
            except Exception:
                sync_inst = None
            LOG.debug("sync_path=%s.  GetInsance=%s" % (sync_path, sync_inst))
            if not sync_inst:
                # ...and now try the other direction...
                sync_path = self.utils.get_instance_name(
                    'SE_StorageSynchronized_SV_SV',
                    {'SyncedElement': targ, 'SystemElement': vol_path})
                sync_inst = self.conn.GetInstance(sync_path)
                LOG.debug("#2 sync_path=%s.  GetInsance=%s" % (sync_path,
                                                               sync_inst))
            if not sync_inst:
                LOG.info(_LI("Cannot get SE_StorageSynchronized_SV_SV for: %s")
                         % six.text_type(targ))
                continue
            LOG.info(_LI("Break clone relationship for "
                         "synchronization instance: %s") % sync_path)
            extraSpecs = self._initial_setup(volume)
            rc, job = self.provision.delete_clone_relationship(
                self.conn, repSvc, sync_path, extraSpecs, force=True)
            LOG.debug("Remove clone relationship returns: RC=%d, job=%s" %
                      (rc, six.text_type(job)))
            if rc == 0:
                num_removed += 1
        LOG.debug("Return. Num_removed=%d." % num_removed)

    def _populate_masking_dict(self, volume, connector, extraSpecs):
        """ PowerVC override for:
            https://bugs.launchpad.net/cinder/+bug/1501925
        """
        m_dict = super(EMCVMAXCommonExtension, self)._populate_masking_dict(
            volume, connector, extraSpecs)
        if len(m_dict['sgGroupName']) <= 64:
            return m_dict

        LOG.debug("PowerVC resource name truncation fallback for: %s" %
                  m_dict['sgGroupName'])
        uniqueName = self.utils.generate_unique_trunc_pool(
            extraSpecs['storagetype:pool'])
        shortHostName = self.utils.get_host_short_name(connector['host'])
        shortHostName = self.get_unique_trunc_host(shortHostName)
        protocol = self.utils.get_short_protocol_type(self.protocol)
        if m_dict['isV3']:
            m_dict['pool'] = uniqueName
            prefix = (
                ("OS-%(shortHostName)s-%(poolName)s-%(slo)s-%(workload)s"
                 % {'shortHostName': shortHostName,
                    'poolName': uniqueName,
                    'slo': m_dict['slo'],
                    'workload': m_dict['workload']}))
        else:
            prefix = (
                ("OS-%(shortHostName)s-%(poolName)s-%(protocol)s"
                 % {'shortHostName': shortHostName,
                    'poolName': uniqueName,
                    'protocol': protocol}))

        m_dict['sgGroupName'] = ("%(prefix)s-SG" % {'prefix': prefix})
        m_dict['maskingViewName'] = ("%(prefix)s-MV" % {'prefix': prefix})
        m_dict['igGroupName'] = (
            ("OS-%(shortHostName)s-%(protocol)s-IG"
             % {'shortHostName': shortHostName,
                'protocol': protocol}))
        LOG.debug("Populated masking dict: %s" % m_dict)
        return m_dict

    def get_unique_trunc_host(self, hostName):
        if hostName and len(hostName) > 38:
            hostName = hostName.lower()
            m = hashlib.md5()
            m.update(hostName.encode('utf-8'))
            uuid = m.hexdigest()
            return(
                ("%(host)s%(uuid)s"
                 % {'host': hostName[-6:],
                    'uuid': uuid}))
        else:
            return hostName

    def extend_volume(self, volume, newSize):
        """
        PowerVC override to pick up Mitaka change. TODO: Remove when
        backported to Liberty: https://bugs.launchpad.net/cinder/+bug/1515190

        Extends an existing volume.

        Prequisites:
        1. The volume must be composite e.g StorageVolume.EMCIsComposite=True
        2. The volume can only be concatenated
           e.g StorageExtent.IsConcatenated=True

        :params volume: the volume Object
        :params newSize: the new size to increase the volume to
        :returns: dict -- modifiedVolumeDict - the extended volume Object
        :raises: VolumeBackendAPIException
        """
        originalVolumeSize = volume['size']
        volumeName = volume['name']
        extraSpecs = self._initial_setup(volume)
        self.conn = self._get_ecom_connection()
        volumeInstance = self._find_lun(volume)
        if volumeInstance is None:
            exceptionMessage = (_("Cannot find Volume: %(volumename)s. "
                                  "Extend operation.  Exiting....")
                                % {'volumename': volumeName})
            LOG.error(exceptionMessage)
            raise exception.VolumeBackendAPIException(data=exceptionMessage)

        if int(originalVolumeSize) > int(newSize):
            exceptionMessage = (_(
                "Your original size: %(originalVolumeSize)s GB is greater "
                "than: %(newSize)s GB. Only Extend is supported. Exiting...")
                % {'originalVolumeSize': originalVolumeSize,
                   'newSize': newSize})
            LOG.error(exceptionMessage)
            raise exception.VolumeBackendAPIException(data=exceptionMessage)

        additionalVolumeSize = six.text_type(
            int(newSize) - int(originalVolumeSize))
        additionalVolumeSize = self.utils.convert_gb_to_bits(
            additionalVolumeSize)
        LOG.debug("PowerVC extension: newSize=%s, additionalVolSize=%s, "
                  "extra_specs=%s" % (str(newSize), additionalVolumeSize,
                                      extraSpecs))
        if extraSpecs['isV3']:
            rc, modifiedVolumeDict = self._extend_v3_volume(
                volumeInstance, volumeName, newSize, extraSpecs)
        else:
            # This is V2.
            rc, modifiedVolumeDict = self._extend_composite_volume(
                volumeInstance, volumeName, newSize, additionalVolumeSize,
                extraSpecs)

        # Check the occupied space of the new extended volume.
        extendedVolumeInstance = self.utils.find_volume_instance(
            self.conn, modifiedVolumeDict, volumeName)
        extendedVolumeSize = self.utils.get_volume_size(
            self.conn, extendedVolumeInstance)
        LOG.debug(
            "The actual volume size of the extended volume: %(volumeName)s "
            "is %(volumeSize)s.",
            {'volumeName': volumeName,
             'volumeSize': extendedVolumeSize})

        # If the requested size and the actual size don't
        # tally throw an exception.
        newSizeBits = self.utils.convert_gb_to_bits(newSize)
        diffVolumeSize = self.utils.compare_size(
            newSizeBits, extendedVolumeSize)
        if diffVolumeSize != 0:
            exceptionMessage = (_(
                "The requested size : %(requestedSize)s is not the same as "
                "resulting size: %(resultSize)s.")
                % {'requestedSize': newSizeBits,
                   'resultSize': extendedVolumeSize})
            LOG.error(exceptionMessage)
            raise exception.VolumeBackendAPIException(data=exceptionMessage)

        LOG.debug(
            "Leaving extend_volume: %(volumeName)s. "
            "Return code: %(rc)lu, "
            "volume dict: %(name)s.",
            {'volumeName': volumeName,
             'rc': rc,
             'name': modifiedVolumeDict})

        return modifiedVolumeDict

    def _extend_v3_volume(self, volumeInstance, volumeName, newSize,
                          extraSpecs):
        """
        PowerVC override to pick up Mitaka change. TODO: Remove when
        backported to Liberty: https://bugs.launchpad.net/cinder/+bug/1515190

        Extends a VMAX3 volume.

        :param volumeInstance: volume instance
        :param volumeName: volume name
        :param newSize: new size the volume will be increased to
        :param extraSpecs: extra specifications
        :returns: int -- return code
        :returns: volumeDict
        """
        new_size_in_bits = int(self.utils.convert_gb_to_bits(newSize))
        storageConfigService = self.utils.find_storage_configuration_service(
            self.conn, volumeInstance['SystemName'])
        volumeDict, rc = self.provisionv3.extend_volume_in_SG(
            self.conn, storageConfigService, volumeInstance.path,
            volumeName, new_size_in_bits, extraSpecs)

        return rc, volumeDict


# *****************************************************
# **            VMAX Masking Extension               **
# *****************************************************
class EMCVMAXMaskingExtension(emc_vmax_masking.EMCVMAXMasking):
    """ PowerVC Override to EMC VMAX masking class """
    def __init__(self, prtcl):
        super(EMCVMAXMaskingExtension, self).__init__(prtcl)
        self.hwidService = None
        self.pool_info = None

    def _get_storage_hardware_id_instance_names(
            self, conn, initiatorNames, storageSystemName):
        """Given a list of initiator names find CIM_StorageHardwareID instance.
           This is one of the worst offenders in the VMAX driver so
           overriding here.
        :param conn: the connection to the ecom server
        :param initiatorName: the list of initiator names
        :param storageSystemName: the storage system name
        :returns: foundHardwardIDsInstanceNames
        """
        LOG.debug("Find SE_StorageHardwareID CIM paths for wwpns: %s" %
                  initiatorNames)
        foundHWIDNames = []
        if not self.hwidService:
            self.hwidService = self.utils.find_storage_hardwareid_service(
                conn, storageSystemName)
        i_wwpns = set([i.lower() for i in initiatorNames])
        hwInstNames = conn.AssociatorNames(
            self.hwidService, ResultClass='SE_StorageHardwareID')
        LOG.debug("Total number of hwInstNames=%d" % len(hwInstNames))
        for hwInstName in hwInstNames:
            # Note: thread-safety fix not needed because our override
            #       does not look up the instance for performance.
            storageId = six.text_type(hwInstName).strip('\"')[-16:].lower()
            if storageId in i_wwpns:
                foundHWIDNames.append(hwInstName)
                i_wwpns.remove(storageId)
            if not i_wwpns:
                break
        LOG.debug("The found hardware IDs are: %s. WWPNs not found: %s" %
                  (foundHWIDNames, i_wwpns))
        return foundHWIDNames

    def _get_sg_or_mv_associated_with_initiator(
            self, conn, controllerConfigService, volumeInstanceName,
            volumeName, connector, getSG=True):
        """Get storage group or masking view associated with connector.
        TODO: Remove after: https://bugs.launchpad.net/cinder/+bug/1511060
        """
        foundInstanceName = None
        if 'wwpns' not in connector:
            LOG.info(_("No 'wwpns' in connector."))
            return foundInstanceName
        initiatorNames = self._find_initiator_names(conn, connector)
        igInstanceNameFromConnector = self._find_initiator_masking_group(
            conn, controllerConfigService, initiatorNames)
        # Device can be shared by multi-SGs in a multi-host attach case.
        storageGroupInstanceNames = (
            self.get_associated_masking_groups_from_device(
                conn, volumeInstanceName))
        LOG.debug("Found storage groups volume "
                  "%(volumeName)s is in: %(storageGroups)s",
                  {'volumeName': volumeName,
                   'storageGroups': storageGroupInstanceNames})
        if storageGroupInstanceNames:  # not empty
            # Get the SG by IGs.
            for sgInstanceName in storageGroupInstanceNames:
                # Get maskingview from storage group.
                mvInstanceName = self.get_masking_view_from_storage_group(
                    conn, sgInstanceName)
                # PowerVC patched location
                if not mvInstanceName:
                    LOG.info(_LI("No masking view assocated with SG "
                                 "%(storageGroup)s."),
                             {'storageGroup': sgInstanceName})
                    if getSG is True:
                        return sgInstanceName
                    else:
                        return None
                LOG.debug("Found masking view associated with SG "
                          "%(storageGroup)s: %(maskingview)s",
                          {'maskingview': mvInstanceName,
                           'storageGroup': sgInstanceName})
                # Get initiator group from masking view.
                igInstanceName = (
                    self.get_initiator_group_from_masking_view(
                        conn, mvInstanceName))
                LOG.debug("Initiator Group in masking view %(ig)s: "
                          "IG associated with connector%(igFromConnector)s",
                          {'ig': igInstanceName,
                           'igFromConnector': igInstanceNameFromConnector})
                if igInstanceName == igInstanceNameFromConnector:
                    if getSG is True:
                        foundInstanceName = sgInstanceName
                        LOG.debug("Found the storage group associated with "
                                  "initiator %(initiator)s: %(storageGroup)s",
                                  {'initiator': initiatorNames,
                                   'storageGroup': foundInstanceName})
                    else:
                        foundInstanceName = mvInstanceName
                        LOG.debug("Found the masking view associated with "
                                  "initiator %(initiator)s: %(maskingview)s.",
                                  {'initiator': initiatorNames,
                                   'maskingview': foundInstanceName})

                    break
        return foundInstanceName

    def get_or_create_masking_view_and_map_lun(self, conn, maskingViewDict,
                                               extraSpecs):
        """Get or Create a masking view and add a volume to the storage group.
        TODO: Remove when fixed: https://bugs.launchpad.net/cinder/+bug/1515176
        """
        rollbackDict = {}
        LOG.debug("PowerVC override - interim fix.")
        controllerConfigService = maskingViewDict['controllerConfigService']
        volumeInstance = maskingViewDict['volumeInstance']
        maskingViewName = maskingViewDict['maskingViewName']
        volumeName = maskingViewDict['volumeName']
        isV3 = maskingViewDict['isV3']
        isLiveMigration = maskingViewDict['isLiveMigration']
        maskingViewDict['extraSpecs'] = extraSpecs
        defaultStorageGroupInstanceName = None
        fastPolicyName = None
        assocStorageGroupName = None
        if isLiveMigration is False:
            if isV3:
                defaultStorageGroupInstanceName = (
                    self._get_v3_default_storagegroup_instancename(
                        conn, volumeInstance, maskingViewDict,
                        controllerConfigService, volumeName))

            else:
                fastPolicyName = maskingViewDict['fastPolicy']
                # If FAST is enabled remove the volume from the default SG.
                if fastPolicyName is not None:
                    defaultStorageGroupInstanceName = (
                        self._get_and_remove_from_storage_group_v2(
                            conn, controllerConfigService,
                            volumeInstance.path,
                            volumeName, fastPolicyName,
                            extraSpecs))

        # If anything has gone wrong with the masking view we rollback
        try:
            maskingViewInstanceName, storageGroupInstanceName, errorMessage = (
                self._validate_masking_view(conn, maskingViewDict,
                                            defaultStorageGroupInstanceName,
                                            extraSpecs))
            LOG.debug(
                "The masking view in the attach operation is "
                "%(maskingViewInstanceName)s. The storage group "
                "in the masking view is %(storageGroupInstanceName)s.",
                {'maskingViewInstanceName': maskingViewInstanceName,
                 'storageGroupInstanceName': storageGroupInstanceName})
        except Exception as e:
            LOG.exception(_LE(
                "Masking View creation or retrieval was not successful "
                "for masking view %(maskingViewName)s. "
                "Attempting rollback."),
                {'maskingViewName': maskingViewDict['maskingViewName']})
            errorMessage = e

        if not errorMessage:
            # Only after the masking view has been validated, add the
            # volume to the storage group and recheck that it has been
            # successfully added.
            errorMessage = self._check_adding_volume_to_storage_group(
                conn, maskingViewDict, storageGroupInstanceName)

        rollbackDict['controllerConfigService'] = controllerConfigService
        rollbackDict['defaultStorageGroupInstanceName'] = (
            defaultStorageGroupInstanceName)
        rollbackDict['volumeInstance'] = volumeInstance
        rollbackDict['volumeName'] = volumeName
        rollbackDict['fastPolicyName'] = fastPolicyName
        rollbackDict['isV3'] = isV3
        rollbackDict['extraSpecs'] = extraSpecs
        rollbackDict['sgName'] = maskingViewDict['sgGroupName']

        if errorMessage:
            # Rollback code if we cannot complete any of the steps above
            # successfully then we must roll back by adding the volume back to
            # the default storage group for that fast policy.
            if (fastPolicyName is not None):
                # If the errorMessage was returned before the volume
                # was removed from the default storage group no action.
                self._check_if_rollback_action_for_masking_required(
                    conn, rollbackDict)
            if isV3:
                rollbackDict['sgGroupName'] = assocStorageGroupName
                rollbackDict['storageSystemName'] = (
                    maskingViewDict['storageSystemName'])
                self._check_if_rollback_action_for_masking_required(
                    conn, rollbackDict)

            exceptionMessage = (_(
                "Failed to get, create or add volume %(volumeName)s "
                "to masking view %(maskingViewName)s. "
                "The error message received was %(errorMessage)s.")
                % {'maskingViewName': maskingViewName,
                   'volumeName': volumeName,
                   'errorMessage': errorMessage})
            LOG.error(exceptionMessage)
            raise exception.VolumeBackendAPIException(data=exceptionMessage)

        return rollbackDict

    def _get_v3_default_storagegroup_instancename(self, conn, volumeinstance,
                                                  maskingviewdict,
                                                  controllerConfigService,
                                                  volumeName):
        """
        TODO: Remove when fixed: https://bugs.launchpad.net/cinder/+bug/1515176
        """
        LOG.debug("PowerVC override - interim fix.")
        defaultStorageGroupInstanceName = None
        defaultSgGroupName = self.utils.get_v3_storage_group_name(
            maskingviewdict['pool'],
            maskingviewdict['slo'],
            maskingviewdict['workload'])
        assocStorageGroupInstanceNames = (
            self.utils.get_storage_groups_from_volume(
                conn, volumeinstance.path))
        for assocStorageGroupInstanceName in (
                assocStorageGroupInstanceNames):
            instance = conn.GetInstance(
                assocStorageGroupInstanceName, LocalOnly=False)
            assocStorageGroupName = instance['ElementName']

            if assocStorageGroupName == defaultSgGroupName:
                defaultStorageGroupInstanceName = (
                    assocStorageGroupInstanceName)
                break
        if defaultStorageGroupInstanceName:
            self._get_and_remove_from_storage_group_v3(
                conn, controllerConfigService, volumeinstance.path,
                volumeName, maskingviewdict,
                defaultStorageGroupInstanceName)
        else:
            LOG.warning(_LW(
                "Volume: %(volumeName)s does not belong "
                "to storage group %(defaultSgGroupName)s."),
                {'volumeName': volumeName,
                 'defaultSgGroupName': defaultSgGroupName})
        return defaultStorageGroupInstanceName

    def remove_and_reset_members(
            self, conn, controllerConfigService, volumeInstance,
            volumeName, extraSpecs, connector=None, noReset=None):
        """
        PowerVC extension. TODO: Remove when fixed in branch.
            https://bugs.launchpad.net/cinder/+bug/1512796
        Reason: Accommodate VMAX3 branches and new methods added below this.

        Part of unmap device or rollback.

        Removes volume from the Device Masking Group that belongs to a
        Masking View. Check if fast policy is in the extra specs, if it isn't
        we do not need to do any thing for FAST. Assume that
        isTieringPolicySupported is False unless the FAST policy is in
        the extra specs and tiering is enabled on the array.

        :param conn: connection the the ecom server
        :param controllerConfigService: the controller configuration service
        :param volumeInstance: the volume Instance
        :param volumeName: the volume name
        :param extraSpecs: additional info
        :param connector: optional
        :param noReset: optional, if none, then reset
        :returns: storageGroupInstanceName
        """
        fastPolicyName = extraSpecs.get(FASTPOLICY, None)
        isV3 = extraSpecs['isV3']
        LOG.debug("PowerVC override. volume=%s, isV3=%s" %
                  (volumeName, isV3))
        storageGroupInstanceName = None
        if connector is not None:
            storageGroupInstanceName = self._get_sg_associated_with_connector(
                conn, controllerConfigService, volumeInstance.path,
                volumeName, connector)
            if storageGroupInstanceName is None:
                return None
        else:  # Connector is None in V3 volume deletion case.
            storageGroupInstanceNames = (
                self.get_associated_masking_groups_from_device(
                    conn, volumeInstance.path))
            if storageGroupInstanceNames:
                storageGroupInstanceName = storageGroupInstanceNames[0]
            else:
                return None
        instance = conn.GetInstance(storageGroupInstanceName, LocalOnly=False)
        storageGroupName = instance['ElementName']

        volumeInstanceNames = self.get_devices_from_storage_group(
            conn, storageGroupInstanceName)

        numVolInStorageGroup = len(volumeInstanceNames)
        LOG.debug(
            "There are %(numVol)d volumes in the storage group "
            "%(maskingGroup)s.",
            {'numVol': numVolInStorageGroup,
             'maskingGroup': storageGroupInstanceName})

        if numVolInStorageGroup == 1:
            # Last volume in the storage group.
            self._last_vol_in_SG(
                conn, controllerConfigService, storageGroupInstanceName,
                storageGroupName, volumeInstance, volumeName, extraSpecs)
        else:
            # Not the last volume so remove it from storage group in
            # the masking view.
            self._multiple_vols_in_SG(
                conn, controllerConfigService, storageGroupInstanceName,
                storageGroupName, volumeInstance, volumeName,
                numVolInStorageGroup, fastPolicyName, extraSpecs)

        return storageGroupInstanceName

    def _last_vol_in_SG(
            self, conn, controllerConfigService, storageGroupInstanceName,
            storageGroupName, volumeInstance, volumeName, extraSpecs):
        """
        PowerVC extension. TODO: Remove when fixed in branch.
            https://bugs.launchpad.net/cinder/+bug/1512796

        Steps if the volume is the last in a storage group.

        1. Check if the volume is in a masking view.
        2. If it is in a masking view, delete the masking view,
           remove the volume from the storage group, and delete
           the storage group.
        3. If it is not in a masking view, remove the volume from the
           storage group and delete the storage group.

        :param conn: the ecom connection
        :param controllerConfigService: storage system instance name
        :param storageGroupInstanceName: the SG instance name
        :param storageGroupName: the Storage group name (String)
        :param volumeInstance: the volume instance
        :param volumeName: the volume name
        :param extraSpecs: the extra specifications
        """
        status = False
        LOG.debug("Only one volume remains in storage group "
                  "%(sgname)s. Driver will attempt cleanup.",
                  {'sgname': storageGroupName})
        mvInstanceName = self.get_masking_view_from_storage_group(
            conn, storageGroupInstanceName)
        if mvInstanceName is None:
            LOG.debug("Unable to get masking view %(maskingView)s "
                      "from storage group.",
                      {'maskingView': mvInstanceName})
        else:
            maskingViewInstance = conn.GetInstance(
                mvInstanceName, LocalOnly=False)
            maskingViewName = maskingViewInstance['ElementName']

        if mvInstanceName:
            maskingViewInstance = conn.GetInstance(
                mvInstanceName, LocalOnly=False)
            maskingViewName = maskingViewInstance['ElementName']

            @lockutils.synchronized(maskingViewName,
                                    "emc-mv-", True)
            def do_delete_mv_and_sg():
                return self._delete_mv_and_sg(
                    conn, controllerConfigService, mvInstanceName,
                    maskingViewName, storageGroupInstanceName,
                    storageGroupName, volumeInstance, volumeName,
                    extraSpecs)
            do_delete_mv_and_sg()
            status = True
        else:
            # Remove the volume from the storage group and delete the SG.
            self._remove_last_vol_and_delete_sg(
                conn, controllerConfigService,
                storageGroupInstanceName,
                storageGroupName, volumeInstance.path,
                volumeName, extraSpecs)
            status = True
        return status

    def _multiple_vols_in_SG(
            self, conn, controllerConfigService, storageGroupInstanceName,
            storageGroupName, volumeInstance, volumeName, numVolsInSG,
            fastPolicyName, extraSpecs):
        """
        PowerVC extension. TODO: Remove when fixed in branch.
            https://bugs.launchpad.net/cinder/+bug/1512796

        Necessary steps if the volume is not the last in the SG.

        1. Remove the volume from the SG.
        2. Return the volume to default SG if necessary.

        :param conn: the ecom connection
        :param controllerConfigService: storage system instance name
        :param storageGroupInstanceName: the SG instance name
        :param storageGroupName: the Storage group name (String)
        :param volumeInstance: the volume instance
        :param volumeName: the volume name
        :param numVolsInSG: the number of volumes in the SG
        :param fastPolicyName: the FAST policy name
        :param extraSpecs: the extra specifications
        """
        storageSystemInstanceName = self.utils.find_storage_system(
            conn, controllerConfigService)
        if not extraSpecs['isV3']:
            isTieringPolicySupported, __ = (
                self._get_tiering_info(conn, storageSystemInstanceName,
                                       fastPolicyName))
        LOG.debug("Start: number of volumes in masking storage group: "
                  "%(numVol)d", {'numVol': numVolsInSG})
        self.provision.remove_device_from_storage_group(
            conn, controllerConfigService, storageGroupInstanceName,
            volumeInstance.path, volumeName, extraSpecs)

        LOG.debug(
            "RemoveMembers for volume %(volumeName)s completed "
            "successfully.", {'volumeName': volumeName})

        # Add it back to the default storage group.
        if extraSpecs['isV3']:
            self._return_volume_to_default_storage_group_v3(
                conn, controllerConfigService, storageGroupName,
                volumeInstance, volumeName, storageSystemInstanceName,
                extraSpecs)
        else:
            # V2 if FAST POLICY enabled, move the volume to the default SG.
            if fastPolicyName is not None and isTieringPolicySupported:
                self._cleanup_tiering(
                    conn, controllerConfigService, fastPolicyName,
                    volumeInstance, volumeName, extraSpecs)

        volumeInstanceNames = self.get_devices_from_storage_group(
            conn, storageGroupInstanceName)
        LOG.debug(
            "End: number of volumes in masking storage group: %(numVol)d.",
            {'numVol': len(volumeInstanceNames)})

    def _delete_mv_and_sg(self, conn, controllerConfigService, mvInstanceName,
                          maskingViewName, storageGroupInstanceName,
                          storageGroupName, volumeInstance, volumeName,
                          extraSpecs):
        """Delete the Masking view and the Storage Group.

        Also does necessary cleanup like removing the policy from the
        storage group for V2 and returning the volume to the default
        storage group.

        TODO: PowerVC override, remove after integration of:
            https://bugs.launchpad.net/cinder/+bug/1520549
        Reason: Distinquish between VMAX2 and VMAX3
        """
        isV3 = extraSpecs['isV3']
        fastPolicyName = extraSpecs.get(FASTPOLICY, None)

        storageSystemInstanceName = self.utils.find_storage_system(
            conn, controllerConfigService)

        self._last_volume_delete_masking_view(
            conn, controllerConfigService, mvInstanceName,
            maskingViewName, extraSpecs)
        if not isV3:
            isTieringPolicySupported, tierPolicyServiceInstanceName = (
                self._get_tiering_info(conn, storageSystemInstanceName,
                                       fastPolicyName))
            self._get_and_remove_rule_association(
                conn, fastPolicyName,
                isTieringPolicySupported,
                tierPolicyServiceInstanceName,
                storageSystemInstanceName['Name'],
                storageGroupInstanceName, extraSpecs)

        self._cleanup_last_vol(
            conn, controllerConfigService, storageGroupInstanceName,
            storageGroupName, volumeInstance, volumeName,
            storageSystemInstanceName, isTieringPolicySupported, extraSpecs)

    def _cleanup_last_vol(
            self, conn, controllerConfigService, storageGroupInstanceName,
            storageGroupName, volumeInstance, volumeName,
            storageSystemInstanceName, isTieringPolicySupported, extraSpecs):
        """Do necessary cleanup when the volume is the last in the SG.

        This includes removing the last volume from the SG and deleting the
        SG.  It also means moving the volume to the default SG for VMAX3 and
        FAST for VMAX2.

        TODO: PowerVC override, remove after integration of:
            https://bugs.launchpad.net/cinder/+bug/1520549
        Reason: Distinquish between VMAX2 and VMAX3
        """
        LOG.debug("PowerVC _cleanup_last_vol() for %s with extra_specs: %s" %
                  (volumeInstance.path, extraSpecs))
        # Remove the last volume and delete the storage group.
        self._remove_last_vol_and_delete_sg(
            conn, controllerConfigService, storageGroupInstanceName,
            storageGroupName, volumeInstance.path, volumeName,
            extraSpecs)
        # Add it back to the default storage group.
        if extraSpecs['isV3']:
            self._return_volume_to_default_storage_group_v3(
                conn, controllerConfigService, storageGroupName,
                volumeInstance, volumeName, storageSystemInstanceName,
                extraSpecs)
        else:
            # V2 if FAST POLICY enabled, move the volume to the default
            # SG.
            fastPolicyName = extraSpecs.get(FASTPOLICY, None)
            if fastPolicyName is not None and isTieringPolicySupported:
                self._cleanup_tiering(
                    conn, controllerConfigService, fastPolicyName,
                    volumeInstance, volumeName, extraSpecs)
        LOG.debug(
            "Volume %(volumeName)s successfully removed from SG and "
            "returned to default storage group where applicable. "
            "Storage Group %(storageGroupName)s successfully deleted. ",
            {'volumeName': volumeName,
             'storageGroupName': storageGroupName})

    def _create_or_get_initiator_group(
            self, conn, controllerConfigService, igGroupName,
            connector, storageSystemName, extraSpecs):
        """Attempt to create an initiatorGroup.

        PowerVC override. Part of same fix as _find_initiator_masking_group().
        What we are doing is avoiding multiple calls to
        _get_storage_hardware_id_instance_names() in a row (expensive).
        TODO: Remove when _find_initiator_masking_group() override goes away.
        """
        initiatorNames = self._find_initiator_names(conn, connector)
        LOG.debug("The initiator name(s) are: %(initiatorNames)s.",
                  {'initiatorNames': initiatorNames})
        storageHardwareIDInstanceNames = (
            self._get_storage_hardware_id_instance_names(
                conn, initiatorNames, storageSystemName))
        foundInitiatorGroupInstanceName = self._find_initiator_masking_group(
            conn, controllerConfigService, initiatorNames,
            hwNames=storageHardwareIDInstanceNames)

        # If you cannot find an initiatorGroup that matches the connector
        # info create a new initiatorGroup.
        if foundInitiatorGroupInstanceName is None:
            # Check that our connector information matches the
            # hardwareId(s) on the vmax.
            if not storageHardwareIDInstanceNames:
                LOG.info(_LI(
                    "Initiator Name(s) %(initiatorNames)s are not on array "
                    "%(storageSystemName)s."),
                    {'initiatorNames': initiatorNames,
                     'storageSystemName': storageSystemName})
                storageHardwareIDInstanceNames = (
                    self._create_hardware_ids(conn, initiatorNames,
                                              storageSystemName))
                if not storageHardwareIDInstanceNames:
                    msg = (_("Failed to create hardware id(s) on "
                             "%(storageSystemName)s.")
                           % {'storageSystemName': storageSystemName})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

            foundInitiatorGroupInstanceName = self._create_initiator_Group(
                conn, controllerConfigService, igGroupName,
                storageHardwareIDInstanceNames, extraSpecs)

            LOG.info(_LI(
                "Created new initiator group name: %(igGroupName)s."),
                {'igGroupName': igGroupName})
        else:
            LOG.info(_LI(
                "Using existing initiator group name: %(igGroupName)s."),
                {'igGroupName': igGroupName})

        return foundInitiatorGroupInstanceName

    def _find_initiator_masking_group(
            self, conn, controllerConfigService, initiatorNames, hwNames=None):
        """Check to see if an initiatorGroup already exists.
        PowerVC override. Possible performance issue with getting all IGs
        and looping over them to get HWID associators.
        PowerVC will instead use _get_storage_hardware_id_instance_names()
        to lookup all HWIDs and from there, map to initiatorNames, then
        CIM query to the IG. That should be immensely faster.
        TODO: not a functional problem but need to open a LP bug and
              propose fix in community.

        NOTE:  An initiator/wwn can only belong to one initiatorGroup.
        If we were to attempt to create one with an initiator/wwn that is
        already belong to another initiatorGroup, it would fail.

        :param conn: the connection to the ecom server
        :param controllerConfigService: the controller configuration service
        :param initiatorNames: the list of initiator names
        :returns: foundInitiatorMaskingGroup
        """
        foundInitiatorMaskingGroupInstanceName = None
        systemName = self.pool_info['stg_sys_name']
        LOG.debug("PowerVC: initiatorNames=%s, stgSysName=%s, hwNames=%s" %
                  (initiatorNames, systemName, hwNames))
        if hwNames is None:
            hwNames = self._get_storage_hardware_id_instance_names(
                conn, initiatorNames, systemName)
        # Mimic EMC impl that checks whether any connector initiator is
        # in an initiator group and return the first found.
        for hwName in hwNames:
            igNames = conn.AssociatorNames(
                hwName, ResultClass='CIM_InitiatorMaskingGroup')
            if igNames:
                foundInitiatorMaskingGroupInstanceName = igNames[0]
                break

        LOG.debug("PowerVC: returning looked up IG InstanceName %s" %
                  str(foundInitiatorMaskingGroupInstanceName))

        return foundInitiatorMaskingGroupInstanceName

    def _create_initiator_Group(
            self, conn, controllerConfigService, igGroupName,
            hardwareIdinstanceNames, extraSpecs):
        """
        Create a new initiator group.
        Given a list of hardwareId Instance name create a new
        initiator group.
        PowerVC override to work around SMI-S failure for two-part
        IG creation. Here, we create in one call.

        TODO: Remove when community fixes: LP bug is pending.
        """
        rc, job = conn.InvokeMethod(
            'CreateGroup', controllerConfigService, GroupName=igGroupName,
            Type=self.utils.get_num(2, '16'),
            Members=hardwareIdinstanceNames)

        if rc != 0L:
            rc, errordesc = self.utils.wait_for_job_complete(conn, job,
                                                             extraSpecs)
            if rc != 0L:
                exceptionMessage = (_(
                    "PVC: Error Create Group: %(groupName)s with "
                    "hardwareIdinstanceNames %(hwInstNames)s. "
                    "Return code: %(rc)lu.  Error: %(error)s.")
                    % {'groupName': igGroupName,
                       'rc': rc, 'hwInstNames': hardwareIdinstanceNames,
                       'error': errordesc})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(
                    data=exceptionMessage)
        foundInitiatorGroupInstanceName = self._find_new_initiator_group(
            conn, job)
        LOG.debug("PowerVC override: Created new initiator group %s with "
                  "%d HardwareID members." %
                  (six.text_type(foundInitiatorGroupInstanceName),
                   len(hardwareIdinstanceNames) if hardwareIdinstanceNames
                   else 0))

        return foundInitiatorGroupInstanceName

    def _check_adding_volume_to_storage_group(
            self, conn, maskingViewDict, storageGroupInstanceName):
        """
        Add the volume to the storage group and double check it is there.
        PowreVC override - don't double-check that vol was added to SG.
        Issue found in 1.2.3.3, not fixed on liberty so pulled forward.
        TODO: Remove if work-around is not valid or if fixed in community.
        """
        controllerConfigService = maskingViewDict['controllerConfigService']
        sgGroupName = maskingViewDict['sgGroupName']
        volumeInstance = maskingViewDict['volumeInstance']
        volumeName = maskingViewDict['volumeName']
        msg = None
        if self._is_volume_in_storage_group(
                conn, storageGroupInstanceName,
                volumeInstance, sgGroupName):
            LOG.debug(
                "Volume: %(volumeName)s is already part "
                "of storage group %(sgGroupName)s.",
                {'volumeName': volumeName,
                 'sgGroupName': sgGroupName})
        else:
            self.add_volume_to_storage_group(
                conn, controllerConfigService,
                storageGroupInstanceName, volumeInstance, volumeName,
                sgGroupName, maskingViewDict['extraSpecs'])
            LOG.debug("PVC: Successfully added %(volumeName)s to "
                      "%(sgGroupName)s.",
                      {'volumeName': volumeName,
                       'sgGroupName': sgGroupName})

        return msg


# *****************************************************
# **            VMAX Utils Extension                 **
# *****************************************************
class EMCVMAXUtilsExtension(emc_vmax_utils.EMCVMAXUtils):
    """Utility class for SMI-S based EMC volume drivers.

    This Utility class is for EMC volume drivers based on SMI-S.
    It supports VMAX arrays.
    """

    def __init__(self, prtcl):
        super(EMCVMAXUtilsExtension, self).__init__(prtcl)
        self.smi_version_int = 0

    def hwid_names_to_wwpns(self, hwid_name_list, toupper=False):
        """ Utility method to map HWID inst names to WWPNs """
        if toupper:
            return [six.text_type(hn).strip('\"')[-16:].upper()
                    for hn in hwid_name_list]
        else:
            return [six.text_type(hn).strip('\"')[-16:].lower()
                    for hn in hwid_name_list]

    def _get_max_job_retries(self, extraSpecs):
        """Get max job retries either default or user defined
        """
        if extraSpecs and 'storagetype:retries' in extraSpecs:
            jobRetries = super(
                EMCVMAXUtilsExtension, self)._get_max_job_retries(extraSpecs)
        else:
            jobRetries = CONF.vmax_job_retries  # fall-back to old option
        return int(jobRetries)

    def _get_interval_in_secs(self, extraSpecs):
        """Get interval in secs, either default or user defined
        """
        if extraSpecs and 'storagetype:interval' in extraSpecs:
            intervalInSecs = super(
                EMCVMAXUtilsExtension, self)._get_interval_in_secs(extraSpecs)
        else:
            intervalInSecs = CONF.vmax_job_interval  # fall-back to old option
        return int(intervalInSecs)

    def get_volume_members(self, conn, volInstance, metaHeadInstName,
                           elementCompositionService):
        """
        Get composite volume member instances for a volume.
        First try to use GetCompositeElements method before falling back
        to cinder Kilo code to support newer SMI-S providers.
        In my testing, I was not able to remove this. TODO: revist later
        after discussion with EMC.
        :param conn:
        :param volInstnace:
        :param metaHeadInstName:
        :param elementCompositionService:
        :returns: List of CIM element instance names representing the
                  composition elements. The CIM class may be different
                  depending on whether GetCompositeElements returns as
                  expected. Returns None if not a composite volume.
        """
        members = None
        rc, output = conn.InvokeMethod(
            'GetCompositeElements', elementCompositionService,
            TheElement=volInstance.path)
        msg = _("Method GetCompositeElements: %(volume)s. "
                "Return code: %(rc)lu.  Output: %(output)s.") %\
            {'volume': volInstance.path,
             'rc': rc, 'output': output}
        if rc != 0L:
            LOG.info(msg)
        else:
            LOG.debug(msg)
        if output and 'OutElements' in output:
            members = output['OutElements']
        elif metaHeadInstName:
            LOG.debug("Fallback processing...")
            members = self.utils.get_meta_members_of_composite_volume(
                conn, metaHeadInstName)
        LOG.debug("Returning member instances: %s " % members)
        return members

    def _get_random_portgroup(self, element):
        """PowerVC override.
        TODO: Remove after fix: https://bugs.launchpad.net/cinder/+bug/1501919
        """
        portGroupElements = element.getElementsByTagName('PortGroup')
        if portGroupElements and len(portGroupElements) > 0:
            portGroupNames_l = []
            for pgElem in portGroupElements:
                portGroupName = (pgElem.childNodes[0].nodeValue).strip()
                if portGroupName:
                    portGroupNames_l.append(portGroupName)
            portGroupNames_s = set(portGroupNames_l)
            LOG.debug("portGroupNames: %s. List len was %d."
                      % (portGroupNames_s, len(portGroupNames_l)))
            numPortGroups = len(portGroupNames_s)
            if numPortGroups > 0:
                portGroupNames = list(portGroupNames_s)
                selectedPortGroupName = (
                    portGroupNames[random.randint(0, numPortGroups - 1)])
                LOG.debug("Returning selected PortGroup: "
                          "%(selectedPortGroupName)s.",
                          {'selectedPortGroupName': selectedPortGroupName})
                return selectedPortGroupName

        exception_message = (_("No PortGroup elements found in config file."))
        LOG.error(exception_message)
        raise exception.VolumeBackendAPIException(data=exception_message)

    def get_pool_capacities(self, conn, poolName, storageSystemName):
        """PowerVC override to support config option of thin free capacity
        """
        if CONF.vmax_use_thin_capacity:
            poolInstanceName = self.get_pool_by_name(
                conn, poolName, storageSystemName)
            if poolInstanceName is None:
                LOG.error(_LE(
                    "Unable to retrieve pool instance of %(poolName)s on "
                    "array %(array)s."),
                    {'poolName': poolName, 'array': storageSystemName})
                return (0, 0)
            storagePoolInstance = conn.GetInstance(
                poolInstanceName, LocalOnly=False)
            total_capacity_gb = self.convert_bits_to_gbs(
                storagePoolInstance['TotalManagedSpace'])
            free_capacity_gb = self.convert_bits_to_gbs(
                storagePoolInstance['RemainingManagedSpace'])
            free_capacity_gb = max(free_capacity_gb, 0)
            LOG.debug("PowerVC thin config option. Free cap for pool %s "
                      "is %d" % (poolName, free_capacity_gb))
            return (total_capacity_gb, free_capacity_gb)
        else:
            return super(EMCVMAXUtilsExtension, self).get_pool_capacities(
                conn, poolName, storageSystemName)

    def generate_unique_trunc_pool(self, poolName):
        """PowerVC supplied: Create a unique pool name under 16 chars.
        https://bugs.launchpad.net/cinder/+bug/1501925

        :param poolName: long pool name
        :returns: truncated pool name
        """
        if poolName and len(poolName) > 16:
            return (
                ("%(first)s_%(last)s"
                 % {'first': poolName[:8],
                    'last': poolName[-7:]}))
        else:
            return poolName

    def _is_sync_complete(self, conn, syncName, logSyncName=None, retry=0):
        """Check if the job is finished.
        PowerVC override. TODO: Remove when fixed in community for
        particular OpenStack release:
        See LP bug: https://bugs.launchpad.net/cinder/+bug/1517103
        Reason: 'CopyState' should be used instead of 'PercentSynced'

        :param conn: connection to the ecom server
        :param syncName: the sync name
        :returns: True if fully synchronized; False if not;
        """
        syncInstance = conn.GetInstance(syncName,
                                        LocalOnly=False)
        copyState = syncInstance['CopyState']
        pct = -1L
        if 'PercentSynced' in syncInstance:
            pct = syncInstance['PercentSynced']
        LOG.debug("%s: Poll #%d. CopyState is %lu. Percent synced is %lu" %
                  (logSyncName, retry, copyState, pct))
        return copyState == 4

    def wait_for_job_complete(self, conn, job, extraSpecs=None):
        """
        Override for: https://bugs.launchpad.net/cinder/+bug/1538214
        This should not be retired unless their fix has the proposed
        changes PowerVC has asked for.
        Reason: So that RC and error description gets correctly passed to
                caller. And so that getInstance() on the job is not called
                multiple times per thread wake-up.

        Given the job wait for it to complete.

        :param conn: connection to the ecom server
        :param job: the job dict
        :param extraSpecs: the extraSpecs dict. Defaults to None
        :raises: loopingcall.LoopingCallDone
        :raises: VolumeBackendAPIException
        """
        t1 = time.time()
        kwargs = {'retries': 0,
                  'rc': 0,
                  'errordesc': None}
        jobNum = str(job['Job']).split('=')[-1]
        maxJobRetries = self._get_max_job_retries(extraSpecs)
        intervalInSecs = self._get_interval_in_secs(extraSpecs)
        job_to_log = jobNum if jobNum else str(job['Job'])

        def _wait_for_job_complete():
            # Called at an interval until the job is finished.
            kwargs['retries'] = kwargs['retries'] + 1
            is_job_finished = False
            try:
                is_job_finished = self._is_job_finished(conn, job)
            except Exception as e:
                LOG.exception(e)
                exceptionMessage = (
                    _("Issue encountered waiting for job %(job)s.") %
                    {'job': job_to_log})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)

            if is_job_finished:
                kwargs['rc'], kwargs['errordesc'] = (
                    self._verify_job_state(conn, job))
                LOG.debug("Poll attempt %d satisfied for job %s" %
                          (kwargs['retries'], job_to_log))
                raise loopingcall.LoopingCallDone()

            # Using the '>' comparison below, we actually end up checking
            # for maxJobRetries+1, but since the first check happens before
            # any wait interval, this extra check ensures we do wait the
            # full maxJobRetries*intervalInSecs. To only check once with no
            # waiting, one would set Retries=0, which is the expected semantic
            # since the first check is a 'try', not a retry :-)
            if kwargs['retries'] > maxJobRetries:
                kwargs['rc'], kwargs['errordesc'] = (
                    self._verify_job_state(conn, job))
                LOG.error(_LE("_wait_for_job_complete of %(job)s timed out "
                              "after %(retries)d tries, or %(mins).1f "
                              "minutes."),
                          {'job': job_to_log, 'retries': kwargs['retries'],
                           'mins': (maxJobRetries * intervalInSecs) / 60.0})
                raise loopingcall.LoopingCallDone()
            # LOG.debug("Job %s: Did poll retry %d. Sleeping..." %
            #           (job_to_log, kwargs['retries']))

        LOG.debug("BEGIN_JOB: Wait on completion for: %s" % job_to_log)
        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_job_complete)
        timer.start(interval=intervalInSecs).wait()
        if kwargs['errordesc']:
            errStr = "Error Description is: %s." % kwargs['errordesc']
        else:
            errStr = ""
        LOG.debug("END_JOB %(job)s: Return code = %(rc)d. Seconds=%(secs)d. "
                  "%(errStr)s",
                  {'job': job_to_log, 'rc': int(kwargs['rc']),
                   'secs': int(time.time() - t1), 'errStr': errStr})
        return kwargs['rc'], kwargs['errordesc']

    def _verify_job_state(self, conn, job):
        """
        Override for: https://bugs.launchpad.net/cinder/+bug/1538214
        Reason: Part of above fix.

        Check if the job is finished.

        :param conn: connection to the ecom server
        :param job: the job dict
        :returns: boolean -- True if finished; False if not finished;
        """
        jobstatedict = {2: 'New',
                        3: 'Starting',
                        4: 'Running',
                        5: 'Suspended',
                        6: 'Shutting Down',
                        7: 'Completed',
                        8: 'Terminated',
                        9: 'Killed',
                        10: 'Exception',
                        11: 'Service',
                        32767: 'Queue Pending',
                        32768: 'DMTF Reserved',
                        65535: 'Vendor Reserved'}
        jobInstanceName = job['Job']
        jobinstance = conn.GetInstance(jobInstanceName,
                                       LocalOnly=False)
        operationalstatus = jobinstance['OperationalStatus']
        if not operationalstatus:
            jobstate = int(jobinstance['JobState'])
            # PowerVC extra check
            if jobstate not in jobstatedict:
                jobstatedict[jobstate] = str(jobstate)
            errordescription = (_(
                "The job has not completed and is in a '%(state)s' "
                "state. If this problem continues to be seen, the Retries "
                "configuration option may need to be adjusted upward to "
                "wait longer for SMI-S Provider jobs.")
                % {'state': jobstatedict[jobstate]})
            LOG.error(_LE("%(errordescription)s ") %
                      {'errordescription': errordescription})
            errorcode = -1
        else:
            errordescription = jobinstance['ErrorDescription']
            errorcode = jobinstance['ErrorCode']
        return errorcode, errordescription

    def wait_for_sync(self, conn, syncName, extraSpecs=None):
        """
        Override for: https://bugs.launchpad.net/cinder/+bug/1538214
        This should not be retired unless their fix has the proposed
        changes PowerVC has asked for.
        Reason: Don't check sync instance multiple times for each poll.
                Pass a logable version of sync name to distinquish between
                concurrent clones.

        Given the sync name wait for it to fully synchronize.

        :param conn: connection to the ecom server
        :param syncName: the syncName
        :param extraSpecs: extra specifications
        :raises: loopingcall.LoopingCallDone
        :raises: VolumeBackendAPIException
        """
        t1 = time.time()
        # When % complete is reported for a sync operation, we cannot
        # tell which operation it belongs to for concurrent clones,
        # so adding this logic to pull out the ID info that can be
        # used for logging.
        s1 = str(syncName).split(',DeviceID=')
        if len(s1) > 2:
            p1 = s1[0].split('SYMMETRIX')[-1:][0].strip('"')
            p2 = s1[1].split(',')[0].strip('"')
            p3 = s1[1].split('SYMMETRIX')[-1:][0].strip('"')
            p4 = s1[2].split(',')[0].strip('"')
            logSyncName = ("From %s,DeviceID=%s _To_ %s,DeviceID=%s" %
                           (p3, p4, p1, p2))
        else:
            logSyncName = str(syncName)
        kwargs = {'retries': 0}
        intervalInSecs = self._get_interval_in_secs(extraSpecs)
        maxJobRetries = self._get_max_job_retries(extraSpecs)

        def _wait_for_sync():
            """Called at an interval until the synchronization is finished.

            :raises: loopingcall.LoopingCallDone
            :raises: VolumeBackendAPIException
            """
            kwargs['retries'] = kwargs['retries'] + 1
            is_complete = False
            try:
                is_complete = self._is_sync_complete(
                    conn, syncName, logSyncName=logSyncName,
                    retry=kwargs['retries'])
            except Exception as e:
                LOG.exception(e)
                exceptionMessage = (
                    _("Issue encountered waiting for synchronization: "
                      "%(sync)s ") %
                    {'sync': logSyncName})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(exceptionMessage)

            if is_complete:
                LOG.debug("Poll attempt %d satisfied for synchronization: %s" %
                          (kwargs['retries'], logSyncName))
                raise loopingcall.LoopingCallDone()

            # Using the '>' comparison below, we actually end up checking
            # for maxJobRetries+1, but since the first check happens before
            # any wait interval, this extra check ensures we do wait the
            # full maxJobRetries*intervalInSecs. To only check once with no
            # waiting, one would set Retries=0, which is the expected semantic
            # since the first check is a 'try', not a retry :-)
            if kwargs['retries'] > maxJobRetries:
                err = (_LE("Synchronization (clone) for %(sync)s timed out "
                           "after %(retries)d tries, or %(mins).1f "
                           "minutes."),
                       {'sync': syncName, 'retries': kwargs['retries'],
                        'mins': (maxJobRetries * intervalInSecs) / 60.0})
                raise exception.VolumeBackendAPIException(err)

        LOG.debug("BEGIN_SYNC: %s" % logSyncName)
        timer = loopingcall.FixedIntervalLoopingCall(_wait_for_sync)
        timer.start(interval=intervalInSecs).wait()
        LOG.debug("END_SYNC %s: Total wait was %d seconds. Polls=%d" %
                  (logSyncName, int(time.time() - t1), kwargs['retries']))

    def get_smi_version(self, conn):
        """Get the SMI_S version.
        PowerVC override since community driver impl does not fit with
        PowerVC function of checking for deleted volumes. We avoid getting
        CIM instance for every managed volume, every 5 mins.
        (starting on Liberty)
        """
        if self.smi_version_int != 0:
            return self.smi_version_int
        # Else community driver path
        intVersion = 0
        swIndentityInstances = conn.EnumerateInstances(
            'SE_ManagementServerSoftwareIdentity')
        if swIndentityInstances:
            swIndentityInstance = swIndentityInstances[0]
            majorVersion = swIndentityInstance['MajorVersion']
            minorVersion = swIndentityInstance['MinorVersion']
            revisionNumber = swIndentityInstance['RevisionNumber']

            intVersion = int(six.text_type(majorVersion) +
                             six.text_type(minorVersion) +
                             six.text_type(revisionNumber))

            LOG.debug("Major version: %(majV)lu, Minor version: %(minV)lu, "
                      "Revision number: %(revNum)lu, Version: %(intV)lu.",
                      {'majV': majorVersion,
                       'minV': minorVersion,
                       'revNum': revisionNumber,
                       'intV': intVersion})
        self.smi_version_int = intVersion
        return intVersion


# *****************************************************
# **            VMAX Provision Extension             **
# *****************************************************
class EMCVMAXProvisionExtension(emc_vmax_provision.EMCVMAXProvision):
    def __init__(self, prtcl):
        super(EMCVMAXProvisionExtension, self).__init__(prtcl)

    def delete_volume_from_pool(
            self, conn, storageConfigservice, volumeInstanceName, volumeName,
            extraSpecs):
        """OVERRIDE: Given the volume instance remove it from the pool.
        TODO: Remove after LP 1488173 is integrated in PowerVC.

        :param conn: connection the the ecom server
        :param storageConfigservice: volume created from job
        :param volumeInstanceName: the volume instance name
        :param volumeName: the volume name (String)
        :param extraSpecs: additional info
        :returns: int -- return code
        :raises: VolumeBackendAPIException
        """
        try:
            LOG.debug("PowerVC: Unbind volume in case it is still bound: %s"
                      % volumeInstanceName)
            rc, __ = self.unbind_volume_from_storage_pool(
                conn, storageConfigservice,
                volumeInstanceName, volumeName, extraSpecs)
            LOG.debug("Unbind returns: %d" % rc)
            if rc == 0:
                # volume was unbound, so wait some seconds becuase there
                # is an issue with v8.0 environment where the delete
                # op will fail below with rc=10207.
                #    Error: "Thin device is bound to a pool."
                LOG.debug("Unbind was successful. Short sleep...")
                time.sleep(QUIESCE_TIME)
        except Exception as ex:
            LOG.debug("Tolerated backend error when vol not bound: %s" %
                      six.text_type(ex))
        startTime = time.time()
        # Change to use one of two methods (based on
        # multi-delete or not), instead of the old EMCReturnToStoragePool
        # method which was deprecated and removed.
        if isinstance(volumeInstanceName, list):
            theElements = volumeInstanceName
            volumeName = 'Bulk Delete'
            LOG.debug("Calling ReturnElementsToStoragePool...")
            rc, job = conn.InvokeMethod(
                'ReturnElementsToStoragePool', storageConfigservice,
                TheElements=theElements)
        else:
            LOG.debug("PowerVC: Calling ReturnToStoragePool...")
            rc, job = conn.InvokeMethod(
                'ReturnToStoragePool', storageConfigservice,
                TheElement=volumeInstanceName)

        LOG.debug('Return to pool response: rc=%d, job=%s ' % (rc, job))
        if rc != 0:
            rc, errordesc = self.utils.wait_for_job_complete(conn, job,
                                                             extraSpecs)
            if rc != 0:
                exceptionMessage = (_(
                    "Error Delete Volume: %(volumeName)s. "
                    "Return code: %(rc)lu.  Error: %(error)s.")
                    % {'volumeName': volumeName,
                       'rc': rc,
                       'error': errordesc})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(
                    data=exceptionMessage)

        LOG.debug("InvokeMethod EMCReturnToStoragePool took: "
                  "%(delta)s H:MM:SS.",
                  {'delta': self.utils.get_time_delta(startTime,
                                                      time.time())})
        return rc


# *****************************************************
# **            VMAX Provision V3 Extension          **
# *****************************************************
class EMCVMAXProvision3Extension(emc_vmax_provision_v3.EMCVMAXProvisionV3):
    def __init__(self, prtcl):
        super(EMCVMAXProvision3Extension, self).__init__(prtcl)

    def extend_volume_in_SG(
            self, conn, storageConfigService, volumeInstanceName,
            volumeName, volumeSize, extraSpecs):
        """
        PowerVC override to pick up Mitaka change. TODO: Remove if
        backported to Liberty: https://bugs.launchpad.net/cinder/+bug/1515190
        Reason: Extend volume in the V3 provisioning module is new support
                in Mitaka.

        Extend a volume instance.

        :param conn: connection the the ecom server
        :param storageConfigservice: the storage configuration service
        :param volumeInstanceName: the volume instance name
        :param volumeName: the volume name (String)
        :param volumeSize: the volume size
        :param extraSpecs: additional info
        :returns: volumeDict
        :returns: int -- return code
        :raises: VolumeBackendAPIException
        """
        startTime = time.time()

        rc, job = conn.InvokeMethod(
            'CreateOrModifyElementFromStoragePool',
            storageConfigService, TheElement=volumeInstanceName,
            Size=self.utils.get_num(volumeSize, '64'))

        LOG.debug("Extend Volume: %(volumename)s. Return code: %(rc)lu.",
                  {'volumename': volumeName,
                   'rc': rc})

        if rc != 0:
            rc, error_desc = self.utils.wait_for_job_complete(conn, job,
                                                              extraSpecs)
            if rc != 0:
                exceptionMessage = (_(
                    "Error Extend Volume: %(volumeName)s. "
                    "Return code: %(rc)lu.  Error: %(error)s.")
                    % {'volumeName': volumeName,
                       'rc': rc,
                       'error': error_desc})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(
                    data=exceptionMessage)

        LOG.debug("InvokeMethod CreateOrModifyElementFromStoragePool "
                  "took: %(delta)s H:MM:SS.",
                  {'delta': self.utils.get_time_delta(startTime,
                                                      time.time())})

        # Find the newly created volume.
        volumeDict = self.get_volume_dict_from_job(conn, job['Job'])
        return volumeDict, rc
