# =================================================================
#
# Copyright 2014, 2016 IBM Corp.
# All Rights Reserved.
# =================================================================
# Copyright (c) 2012 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
from cinder import context, exception
from cinder.db import api as db_api
from cinder.i18n import _LW
from cinder.volume import volume_types
from cinder.volume.drivers.emc import emc_vmax_fc
from cinder.zonemanager import utils as fczm_utils
import copy
from oslo_config import cfg
from oslo_log import log as logging
import re
import ssl
from xml.dom.minidom import parseString

from requests.exceptions import SSLError
import six
import time

from powervc_cinder import _, exception as powervc_exception
from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import discovery_driver, \
    restricted_metadata_keys as rmk, metada_keys as mk
from powervc_cinder.volume.discovery_driver import \
    CONNECTION_TYPE_NPIV, CONNECTION_TYPE_VSCSI
import powervc_cinder.volume.drivers.emc.emc_vmax_common_ext as common_ext
from powervc_cinder.zonemanager import powervc_utils as pvc_fczm_utils


# Handle zoning in a common way for PowerVC
fczm_utils.AddFCZone = pvc_fczm_utils.ignore_zoning_request
fczm_utils.RemoveFCZone = pvc_fczm_utils.ignore_zoning_request

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
# If the following is true, then PowerVC will attempt to pre-allocate the
# target volume of a clone operation using the volume type extra_specs
# set for the target. Otherwise the normal driver clone path will be taken.
PRE_CREATE_TARGET = True
MAX_CHECK_DEL_PCT = 10
FULL_DEL_VOL_CHK_INTERVAL = 10800  # three hours
CONF.import_opt('vmax_allow_manage_inuse',
                'powervc_cinder.volume.drivers.emc.emc_vmax_common_ext')
CONF.import_opt('vmax_restrict_pool_to_xml_conf',
                'powervc_cinder.volume.drivers.emc.emc_vmax_common_ext')
CONF.import_opt('vmax_use_thin_capacity',
                'powervc_cinder.volume.drivers.emc.emc_vmax_common_ext')


class EMCVMAXProductDriver(
        discovery_driver.VolumeDiscoveryDriver,
        emc_vmax_fc.EMCVMAXFCDriver):
    """ PowerVC cinder volume extension to the EMC Product Driver
        for VMAX using SMI-S.
    """

    def __init__(self, *args, **kwargs):
        self.emc_super = super(EMCVMAXProductDriver, self)
        super(emc_vmax_fc.EMCVMAXFCDriver, self).__init__(*args, **kwargs)
        # self.emc_super.__init__(*args, **kwargs)
        self.endpoint_desc = "EMCVMAXProductDriver provider"
        # Extend the Common module as well to support PowerVC volume types
        self.common = common_ext.EMCVMAXCommonExtension(
            'FC', emc_vmax_fc.EMCVMAXFCDriver.VERSION,
            configuration=self.configuration)
        self.dv_count = 0
        # Define instance vars for check_for_deleted_vols optimization
        self.num_array_vols = -1    # number of volumes on vmax in pools
        self.time_last_dcheck = 0   # time last del vol check
        self.time_last_fcheck = 0   # time last del vol full check
        self.time_last_mcheck = 0   # time last del vol managed check
        self.delta_last_fcheck = 0  # seconds for last del vol managed check
        self.delta_last_mcheck = 0  # seconds for last del vol full check
        LOG.debug("Driver instance initialized. CONF.vmax_allow_manage_inuse"
                  "=%s." % CONF.vmax_allow_manage_inuse)

    def _build_default_opts(self):
        """ Get default opts to allow manager to build the default vol type """
        default_opts = self.common._get_default_opts()
        # Remove the options that have a value of None. They need not be
        # included in the volume type at all.
        o_list = [(k, v) for k, v in default_opts.iteritems() if v is not None]
        return dict(o_list)

    @pvc_fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        """ Wrapper around EMC VMAX driver method.
            We distinguish between vSCSI and NPIV connectivity here to
            skip zoning for vSCSI (mainly for performance) and add
            any extra properties expected in the returned data structure.
        """
        vol_name = volume.get('name')
        LOG.info(_("PowerVC VMAX map volume '%(vol)s' with ID %(id)s to "
                   "ports %(ports)s") %
                 dict(vol=vol_name, id=volume['id'],
                      ports=connector.get('wwpns', '[missing]')))
        if 'connection_type' not in connector:
            msg = (_("connection_type is missing from the attach volume "
                     "connector. An NPIV connectivity type will be assumed."))
            LOG.warn(msg)
            connector['connection_type'] = ['npiv']
        conn_type = 'npiv' if 'npiv' in connector['connection_type'] else\
            'pv_vscsi' if 'pv_vscsi' in connector['connection_type'] else None
        if not conn_type:
            msg = (_("The connection_type passed on an attach volume request "
                     "is not supported. The value passed is: %(val)s. "
                     "The supported values are 'npiv' or 'pv_vscsi'") %
                   dict(val=connector['connection_type']))
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(msg)
        LOG.debug("%s: Incoming volume=%s, connector=%s" %
                  (conn_type,
                   six.text_type(volume.__dict__),
                   six.text_type(connector)))
        # Temporarily shove wwpns in the volume dictionary so that
        # powervc find_device_number() override can lookup the proper LUN ID
        # in a multi-attach scenario. TODO: Remove after fix integrated for:
        # https://bugs.launchpad.net/cinder/+bug/1532283
        if 'wwpns' in connector and connector['wwpns']:
            volume['initiator_wwpns'] = connector['wwpns']
        # if conn_type == 'npiv':
            # Do some extra work to prep for the NPIV flow.
            # TODO: EMC has stated they are implementing a solution that
            #       does not require the vFC wwpns to be logged into the
            #       fabric. The idea was to remove the following call when we
            #       picked up changes for cinder bug 1379156. However, testing
            #       and analysis of that change shows it will not cause
            #       SE_StorageHardwareID resources to be created in the VMAX
            #       database for the npiv ports if they aren't logged into
            #       the fabric. So we will continue to add & clean these up
            #       in this PowerVC extension code.
            #  self._prep_for_npiv(volume, connector)

        # Bypass emc_super and go directly to common.initialize_connection
        # since we want to control zoning. For vSCSI, no zoning.
        try:
            device_info = self.common.initialize_connection(
                volume, connector)
            device_number = device_info['hostlunid']
            storage_system = device_info['storagesystem']
            lunMaskView = device_info.get('maskingview', None)
            target_wwns, init_targ_map = self._build_initiator_target_map(
                storage_system, volume, connector, maskingView=lunMaskView)
        except Exception as ex:
            # log the exception stack and re-raise the exception.
            # TODO: Is logic needed to rollback the initialize connection
            #       if failure happens during _build_initiator_target_map.
            #       Revisit later as we may be moving back to EMC's method
            #       for building the target map.
            LOG.exception(ex)
            raise ex

        attach_data = {'driver_volume_type': 'fibre_channel',
                       'data': {'target_lun': device_number,
                                'target_discovered': True,
                                'target_wwn': target_wwns,
                                'vendor': 'EMC',
                                'volume_id': volume['id']}}
        # get the emcwwn to be returned as pg83_naa
        metadata = self._get_restricted_metadata(volume)
        if metadata and rmk.VDISK_UID_KEY in metadata:
            attach_data['data']['pg83NAA'] = metadata[rmk.VDISK_UID_KEY]
            attach_data['data']['deviceID'] = metadata[rmk.VDISK_ID_KEY]
        # must return the initiator target map
        attach_data['data']['initiator_target_map'] = init_targ_map

        if 'SE_StorageHardwareIDs' in connector:
            connector.pop('SE_StorageHardwareIDs')
        LOG.info(_("PowerVC VMAX map returns attach data %(data)s with "
                   "connector %(connector)s.") %
                 dict(data=attach_data, connector=connector))
        return attach_data

    @pvc_fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        """
        Wrapper around VMAX driver terminate_connection.
        The lock synchronization is intentionally for 'attach'
        to allow only one of either attach or detach at a time.
        """
        vol_name = volume.get('name')
        if 'phy_to_virt_initiators' in connector and\
                'connection_type' not in connector:
            msg = (_("connection_type is missing from the attach volume "
                     "connector. An NPIV connectivity type will be assumed."))
            LOG.warn(msg)
            connector['connection_type'] = ['npiv']
        LOG.info(_("PowerVC VMAX unmap volume '%(vol)s' from ports %(ports)s."
                   " Volume ID = %(id)s")
                 % dict(vol=vol_name,
                        ports=connector.get('wwpns', '[missing]'),
                        id=volume['id']))
        LOG.debug("Incoming volume=%s, connector=%s" %
                  (six.text_type(volume.__dict__),
                   six.text_type(connector)))

        target_wwns = []
        init_targ_map = {}
        data = {'driver_volume_type': 'fibre_channel', 'data': {}}
        # Temporarily shove wwpns in the volume dictionary so that
        # powervc find_device_number() override can lookup the proper LUN ID
        # TODO: Remove after fix integrated for:
        # https://bugs.launchpad.net/cinder/+bug/1532283
        # Outer try block for detach processing
        try:
            # Do the superclass work in-line.
            mvInstanceName = None
            if 'wwpns' in connector and connector['wwpns']:
                volume['initiator_wwpns'] = connector['wwpns']
                mvInstanceName = self.common.get_masking_view_by_volume(
                    volume, connector)
            if mvInstanceName is None:
                if ('connection_type' in connector and
                        'npiv' in connector['connection_type'] and
                        'initiator_wwpns' in volume):
                    # There is a situation where there is no masking view
                    # but the initiator group and wwpn resources are still
                    # orphaned, so we can call our post-terminate cleanup
                    # to cover this case. However, this will not handle
                    # any zoning cleanup. Without a masking view, we do not
                    # know the port group, and so cannot construct the
                    # init_targ_map to do zoning cleanup.
                    self._post_for_npiv(volume, connector, mvInstanceName)
                LOG.warn(_LW("Volume %(volume)s is NOT in any masking view. "
                             "PowerVC VMAX unmap returns detach data "
                             "%(data)s with connector: %(connector)s."),
                         {'volume': volume['name'], 'data': data,
                          'connector': connector})
                return data

            # Pre-compute the map because the rels may be gone after the
            # terminate connection.
            target_wwns, init_targ_map = self._build_initiator_target_map(
                self._get_array_name(), volume, connector,
                maskingView=mvInstanceName)
            self.common.terminate_connection(volume, connector)
            # End superclass work in-line

            # Signal RemoveFCZones to remove NPIV zones if needed
            if self._ready_to_remove_npiv_zones(
                    volume, connector, mvInstanceName):
                # Add data from pre-computed targets before detach.
                data['data']['target_wwn'] = target_wwns
                data['data']['initiator_target_map'] = init_targ_map
                data['volume_id'] = volume['id']

                # Now we need to check whether the initiator
                # group still exists and it is not associated with a masking
                # view. In that case, delete it along with the HW resource
                # IDs for the NPIV wwpns.
                LOG.debug("Do NPIV cleanup for %s." % connector['wwpns'])
                self._post_for_npiv(volume, connector, mvInstanceName)
        except Exception as ex:
            # Log the stack and re-raise
            LOG.exception(ex)
            raise ex
        LOG.info(_("PowerVC VMAX unmap returns detach data %(data)s with "
                   "connector: %(connector)s.") %
                 dict(data=data, connector=connector))
        return data

    def _ready_to_remove_npiv_zones(self, volume, connector, mvInstName):
        """ Check to see if the previously looked up masking view
            still exists. If the EMC driver removed it, then there must be
            no more volume connections and the zones can also be removed.
        """
        if ('connection_type' in connector and
                'npiv' in connector['connection_type']):
            mvInst = self.common.utils.get_existing_instance(self.common.conn,
                                                             mvInstName)
            if not mvInst:
                LOG.debug("CIM instance not found for masking view %s, "
                          "so ready to remove zoning." % mvInstName)
                return True
            LOG.debug("CIM instance still exists for masking view: %s" %
                      mvInst['ElementName'])

        # This could be iscsi or vscsi so indicate false in this method
        return False

    def _post_for_npiv(self, volume, connector, mvInstName):
        """ Need to do extra cleanup for the NPIV case
            TODO: Remove when: https://bugs.launchpad.net/cinder/+bug/1504192
        """
        ctrlConfigService = self.common.utils.\
            find_controller_configuration_service(self.common.conn,
                                                  self._get_array_name())
        ig = self.common.masking._find_initiator_masking_group(
            self.common.conn, ctrlConfigService, connector['wwpns'])
        if ig:
            LOG.debug("Initiator group found: %s" % ig)
            mvs = self.common.conn.AssociatorNames(
                ig, ResultClass='EMC_LunMaskingSCSIProtocolController')
            if mvs:
                # Make this a warning since this should only get called
                # if asked to clean up the zones (boot volume going away)
                LOG.warn(_("Initiator group %(ig)s is still associated "
                           "with a masking view, so cannot clean up. The "
                           "masking view is %(mask)s and the view looked "
                           "up prior to terminiate_connection is %(prior)s.")
                         % dict(ig=ig, mask=mvs[0], prior=mvInstName))
            else:
                LOG.debug("Remove n_port wwpns: %s" % connector['wwpns'])
                HWIdService = (
                    self.common.utils.find_storage_hardwareid_service(
                        self.common.conn, self._get_array_name()))
                if 'initiator_hwnames' in connector:
                    hwNames = connector['initiator_hwnames']
                else:
                    hwNames = self.common.masking.\
                        _get_storage_hardware_id_instance_names(
                            self.common.conn, connector['wwpns'],
                            self._get_array_name())
                if '-+-OS' in ig['InstanceID']:
                    # Newer backend, so we need to remove the wwpns from the
                    # initiator group first.
                    self.remove_ig_members(ctrlConfigService, ig, hwNames)
                    LOG.debug("Removed HW members from IG.")
                LOG.info(_("Remove initiator group and HW_IDs for: %s") % ig)
                self.common.conn.DeleteInstance(ig)
                # cleanup HW resource IDs (wwpns)
                for hwid_name in hwNames:
                    rc, ret = self.common.conn.InvokeMethod(
                        'DeleteStorageHardwareID', HWIdService,
                        HardwareID=hwid_name)
                    LOG.debug("DeleteStorageHardwareID %s: rc=%d, ret=%s" %
                              (hwid_name, rc, ret))
        else:
            LOG.debug("No initiator group found after terminate_conn.")

    def remove_ig_members(self, controllerConfigService,
                          initiatorGroupInstanceName,
                          hardwareIdinstanceNames):
        """ Remove initiator group members - helper """
        LOG.debug("Removing members from IG: %s" % hardwareIdinstanceNames)
        rc, job = self.common.conn.InvokeMethod(
            'RemoveMembers', controllerConfigService,
            MaskingGroup=initiatorGroupInstanceName,
            Members=hardwareIdinstanceNames)

        if rc != 0:
            rc, errordesc = (
                self.common.utils.wait_for_job_complete(self.common.conn, job))
            if rc != 0:
                exceptionMessage = (_(
                    "Error removing initiators from group : %(groupName)s. "
                    "Return code: %(rc)lu.  Error: %(error)s.")
                    % {'groupName': initiatorGroupInstanceName,
                       'rc': rc,
                       'error': errordesc})
                LOG.error(exceptionMessage)
                raise exception.VolumeBackendAPIException(
                    data=exceptionMessage)

    def create_volume(self, volume):
        """ Create a VMAX volume and add PowerVC restricted metadata """
        LOG.info(_("PowerVC VMAX create volume %(vol)s.") %
                 dict(vol=volume.__dict__))
        ctxt = context.get_admin_context()
        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('PowerVC VMAX adding volume_type_id \
                    to volume=%s') % volume)
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})

        # munge the display_name if needed
        volume_name = self._generate_candidate_caption(volume)
        model_update = self.emc_super.create_volume(volume)
        LOG.debug("EMC create_volume return object: %s." % model_update)

        # Persist the restricted metadata for volume
        metadata = self._new_vol_sync_restricted_metadata(volume, volume_name)
        # Persist geometry information as VMAX function is sensitive to this.
        # Note that this is not kept in sync with backend at this time.
        # It just provides useful data for VMAX volumes.
        try:
            geometry = self._get_volume_geometry(volume, set_metadata=False)
            model_update['metadata'] = geometry['metadata']
            model_update['metadata'][mk.VOLUME_WWN] = \
                metadata[rmk.VDISK_UID_KEY]
            LOG.info(_("Volume %(vol_id)s geometry and metadata: %(geometry)s"
                       " ") % {'vol_id': volume['id'], 'geometry': geometry})
        except Exception as ex:
            LOG.exception(ex)
            LOG.warn(_("Could not get volume geometry. Continuing."))

        LOG.info(_("PowerVC VMAX create returns model data = %(model)s, "
                   "r_metadata = %(meta)s") %
                 dict(model=model_update, meta=metadata))
        return model_update

    def delete_volume(self, volume):
        """Deletes an EMC volume - call into super for work."""
        LOG.info(_("PowerVC VMAX delete volume %(vol)s.") %
                 dict(vol=volume.__dict__))
        # First check volume health
        try:
            self.check_volume_health(volume)
        except powervc_exception.EMCVdiskNotFoundException:
            LOG.warn(_("The volume '%(name)s' with ID %(id)s does not exist "
                       "on the VMAX array. Continuing with removal of "
                       "volume record from PowerVC.") %
                     dict(name=volume['name'], id=volume['id']))
            return
        # Check if there are any copy (clone) sessions associated
        # with this volume. If so, they may or may not be stale, but they
        # need to be removed prior to deletion or delete will fail.
        if 'provider_location' in volume:
            try:
                self.common._check_and_remove_sync_rels(volume)
            except Exception as ex:
                LOG.info(_("Unable to handle check-and-remove synchronized "
                           "clone relationship removal. Continuing. Error: "
                           "%s") % six.text_type(ex))
        # Call driver method to do the delete.
        self.emc_super.delete_volume(volume)
        LOG.info(_("PowerVC VMAX delete returns complete for volume "
                   "%(name)s.") % dict(name=volume['name']))

    def create_cloned_volume(self, target_vol, source_vol):
        """Creates a cloned volume."""
        LOG.info(_("PowerVC VMAX clone source '%(source)s' to target "
                   "'%(target)s'.") % dict(source=source_vol['name'],
                                           target=target_vol['name']))
        LOG.debug("create_cloned_volume(): target_vol=%s, **** source_vol=%s."
                  % (target_vol.__dict__, source_vol.__dict__))

        ctxt = context.get_admin_context()
        volume_type_id = target_vol.get('volume_type_id')
        if not volume_type_id:
            # No volume type has been set yet on the target volume object,
            # so set the default one. Note that not all properties of the
            # volume type may be applicable to the clone operation. For
            # example, stripecount will be ignored because the target will
            # receive the stripecount setting of the source volume.
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                target_vol['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % target_vol)
                self.db.volume_update(ctxt, target_vol['id'],
                                      {'volume_type_id': volume_type_id})
        try:
            # Call the super to do the clone
            model_update = self.emc_super.create_cloned_volume(target_vol,
                                                               source_vol)
        except Exception as ex:
            LOG.exception(ex)
            err = _("Could not clone a volume. The error is: %(reason)s")
            err_msg = err % dict(reason=six.text_type(ex))
            LOG.error(err_msg)
            # Check if we have cleanup to do.
            if 'provider_location' in target_vol and\
                    target_vol['provider_location']:
                # The target exists so roll it back.
                LOG.info(_("Roll-back processing due to error. Clean up "
                           "clone target..."))
                try:
                    self.emc_super.delete_volume(target_vol)
                except Exception as del_ex:
                    LOG.warn(_("An error occurred during rollback "
                               "processing: %s") % six.text_type(del_ex))
            # raise the original
            raise exception.VolumeBackendAPIException(data=err_msg)

        # Persist the restricted metadata for volume
        metadata = self._new_vol_sync_restricted_metadata(
            target_vol, target_vol['name'])

        # Persist geometry information as VMAX function is sensitive to this
        try:
            # Note that this is not kept in sync with backend at this time.
            # It just provides useful data for VMAX volumes.
            geometry = self._get_volume_geometry(target_vol,
                                                 set_metadata=False)
            model_update['metadata'] = geometry['metadata']
            model_update['metadata'][mk.VOLUME_WWN] = \
                metadata[rmk.VDISK_UID_KEY]
            LOG.info(_("Volume %(vol_id)s geometry and metadata: %(geometry)s"
                       " ") % {'vol_id': target_vol['id'],
                               'geometry': geometry})
        except Exception as ex:
            LOG.exception(ex)
            LOG.warn(_("Could not get volume geometry. Continuing."))

        LOG.info(_("PowerVC VMAX clone returns with model data %(model)s, "
                   "r_metadata=%(meta)s.") %
                 dict(model=model_update, meta=metadata))
        return model_update

    def extend_volume(self, volume, new_size):
        """ PowerVC override for extend_volume so that the volume's
            geometry metadata can be updated after the extension.
        """
        LOG.debug("PowerVC VMAX extend volume %s." % volume['name'])
        try:
            super(EMCVMAXProductDriver, self).extend_volume(volume, new_size)
        except Exception as ex:
            LOG.exception(ex)  # ensure traceback is logged.
            raise ex
        # Persist geometry information but tolerate errors
        try:
            new_geom = self._get_volume_geometry(volume)
            LOG.info(_("Volume %(vol_id)s geometry and metadata: %(geometry)s"
                       " ") % {'vol_id': volume['id'], 'geometry': new_geom})
        except Exception as ex:
            LOG.exception(ex)
            LOG.warn(_("Could not get volume geometry. Continuing."))

    def _new_vol_sync_restricted_metadata(self, volume, volume_name):
        """ Does the common work for create/clone/etc of setting the
            PowerVC restricted metadata properties. And synchronizes
            the backend ElmentName property to match the vdisk_name
            metadata such that PowerVC can tell if a volume has been
            removed and recreated (re-using the same 'DeviceId'.
            Notes: The EMCWWN property does not appear to change for a
                   re-used TDEV. This is unfortunate since it resembles a
                   page83 NAA name, which I believe Storwize generates
                   uniquely across time.
                   ElementName is the only mutable volume property for the
                   CIM instance on 5876 vmax firmware.
        """
        # Lookup up the new backend CIM instance
        instance = self.common._find_lun(volume, skip_verify=True,
                                         skip_cache=True)
        if not instance:
            err_msg = _("Cannot find the volume just created on the "
                        "backend: %(location)s") %\
                dict(location=volume['provider_location'])
            raise exception.VolumeBackendAPIException(data=err_msg)
        oldName = instance['ElementName']
        # PVC 1.2.3 - this is no different from what we did in prior releases.
        # except we truncate to match the automatic truncation that happens
        # so the ElementName value on the backend. With this same naming
        # format, if a volume is deleted and recreated out-of band with the
        # same user-specified name AND the same DeviceID is reused, then it
        # will look like the same volume to this powerVC instance. We could
        # add random digits to the syncName here to prevent this case,
        # but it seems low enough in probability to be not worth the extra
        # strangeness a user may encounter in onboarding volumes with these
        # random digits compared to onboarding in 1.2.2.
        syncName = instance['DeviceID'] + '+' + volume_name
        syncName = syncName[:64]
        instance['ElementName'] = syncName
        LOG.debug("To Set: Old ElementName=%s, New ElementName=%s." %
                  (oldName, instance['ElementName']))

        to_remove = ['EMCConsumableBlocksWritten', 'EMCSpaceConsumed',
                     'EMCFullyAllocated', 'OperationalStatus',
                     'StatusDescriptions']
        for prop in to_remove:
            # If property is not supported for ModifyInstance, then remove.
            # ECOM CIM does not support the PropertyList optional parameter
            # on a ModifyInstance, so we cannot specify only what we are
            # changing.
            if prop in instance:
                del instance[prop]
        # Now update the instance with the common name.
        # Note that after discussion with EMC on the associated bug for this,
        # they opted to save the cinder volume ID in ElementName instead.
        # This works well for openstack since the ID is unique for that
        # cinder and across time, and they don't support volume onboarding
        # natively. It does not work for us since we support onboarding and
        # can have multiple PowerVC instances managing the same volume.
        try:
            self.common.conn.ModifyInstance(instance)
        except Exception as ex:
            LOG.warn(_("Error updating the volume CIM instance %(vol)s. "
                       "The backend volume name will remain '%(cur_name)s'. "
                       "Error: %(err)s") %
                     {'vol': instance.path, 'cur_name': oldName,
                      'err': six.text_type(ex)})
            syncName = oldName
            instance['ElementName'] = oldName

        # Now set the PowerVC restricted metadata.
        loc = ast.literal_eval(volume['provider_location'])
        dev_id = loc['keybindings']['DeviceID']
        metadata = {rmk.VDISK_ID_KEY: dev_id,
                    rmk.VDISK_NAME_KEY: syncName,
                    rmk.VDISK_UID_KEY: instance['EMCWWN']
                    }
        LOG.debug("Set restricted metadata: %s" % metadata)
        ctxt = context.get_admin_context()
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, volume['id'],
                                                        metadata)
        return metadata

    def _get_restricted_metadata(self, volume):
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume['id'])
        return metadata

    def get_default_vol_type(self):
        """Returns the default volume type from the emc's .conf file"""
        vtn = self.common.configuration.default_volume_type
        vtn = vtn.decode('utf-8') if vtn else vtn
        LOG.debug("return configuration.default_volume_type= %s." % vtn)
        return vtn

    def _get_array_name(self):
        if 'stg_sys_name' not in self.common.pool_info:
            self.get_volume_stats(refresh=True)
        return self.common.pool_info['stg_sys_name']

    def get_storage_metadata(self):
        """
        Returns the additional information that PowerVC uses to populate
        various UI elements. This is a synchronous back-end call.
        """
        LOG.info(_("PowerVC VMAX metadata request."))
        volume_pools = []
        config_data = {}
        stats = self.get_volume_stats(refresh=self.common.first_time)
        # TODO: The remaining part of the method and helpers could be
        #       refactored to better reuse data from stats just obtained.
        array_id = self.common._get_array_id()
        try:
            volume_pools = self._get_pool_metadata(array_id)
            for pool in volume_pools:
                # Remove the cim_path object since it is not a simple string
                if 'cim_path' in pool:
                    del pool['cim_path']
            config_data = self._get_config_data(
                self.common.configuration.cinder_emc_config_file)
            config_data['symmetrix_array'] = array_id
            config_data['generation'] = 'V3' if\
                self.common.pool_info['is_v3'] else 'V2'
            config_data['firmware_level'] = self.common.utils.\
                get_firmware_version(self.common.conn, array_id)
            config_data['total_capacity_gb'] = stats['total_capacity_gb']
            config_data['free_capacity_gb'] = stats['free_capacity_gb']
            config_data['provider'] = self.common.pool_info.get('provider',
                                                                'unknown')
        except Exception as ex:
            LOG.exception(ex)
            err = _("Could not retrieve pools through ECOM. %(reason)s")
            err_msg = err % dict(reason=six.text_type(ex))
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        config_data['volume_pools'] = volume_pools
        # endpoint_desc describes the storage controller, handy for attaching
        # to exceptions or mentioning in log messages. We re-use the SVC
        # constructor since it is defined at a common level.
        self.endpoint_desc = powervc_exception.SVCDescriptor(
            self.common.configuration.host,
            config_data['symmetrix_array'],
            config_data['ecom_user'],
            config_data['ecom_server'],
            config_data['ecom_port'])
        LOG.debug("Endpoint_Description: %s." % self.endpoint_desc)
        LOG.info(_("PowerVC VMAX metadata returns: %(data)s.") %
                 dict(data=config_data))
        return config_data

    def _get_config_data(self, fileName):
        """ Return config data for this storage driver """
        LOG.debug("EMC filename = %s." % fileName)
        try:
            array_info = self.common.pool_info['arrays_info'][0]  # one array
            ecom_user = array_info['EcomUserName']
            ecom_ip = array_info['EcomServerIp']
            ecom_port = array_info['EcomServerPort']
            # NOTE: the following code is borrowed from the vmax cinder
            #      driver that is embedded in another method. I've added
            #      a '\n' for one of the replace substitutions to not
            #      depend on newlines for how they are splitting the port
            #      groups.
            myFile = open(fileName, 'r')
            data = myFile.read()
            myFile.close()
            dom = parseString(data)
            portGroups = dom.getElementsByTagName('PortGroups')
            portGroupNames = []
            if portGroups is not None and len(portGroups) > 0:
                portGroupsXml = portGroups[0].toxml()
                portGroupsXml = portGroupsXml.replace('<PortGroups>', '')
                portGroupsXml = portGroupsXml.replace('</PortGroups>', '')
                portGroupsXml = portGroupsXml.replace('<PortGroup>', '|')
                portGroupsXml = portGroupsXml.replace('</PortGroup>', '')
                portGroupsXml = portGroupsXml.replace('\n', '')
                portGroupsXml = portGroupsXml.replace('\t', '')
                portGroupsXml = portGroupsXml[1:]
                # convert the | separated string to a list
                portGroupNames = (
                    [s.strip() for s in portGroupsXml.split('|') if s])

        except Exception as ex:
            LOG.exception(ex)
            return {}
        # Notes: EMC team confirmed that currently, VMAX does not support
        #        compression for our pool purposes. Same goes for encryption.
        #        The FAST policy we support is an extra_spec on the volume
        #        type, not the one from the XML file.
        return {'port_groups': portGroupNames,
                'compression_supported': False,
                'ecom_server': ecom_ip,
                'ecom_port': ecom_port,
                'ecom_user': ecom_user}

    def _get_pool_metadata(self, array_id, minimal=False):
        """Return pool metadata for VMAX storage pools for PowerVC
           volume-pools API.

        :param array_id: the array Serial number (String)
        :returns: list of pool dictionaries
        """
        pool_list = []
        storageSystemInstanceName = self.common.utils.find_storageSystem(
            self.common.conn, array_id)

        if not minimal:
            # Get FAST VP information to correlate with pools later.
            fast_data = self._get_fast_policy_data(storageSystemInstanceName)

        # Now collect the pools
        if self.common.pool_info['is_v3']:
            vpools = self.common.conn.Associators(
                storageSystemInstanceName,
                resultClass='Symm_SRPStoragePool')
        else:
            vpools = self.common.conn.Associators(
                storageSystemInstanceName,
                resultClass='EMC_VirtualProvisioningPool')

        for vpool in vpools:
            if CONF.vmax_restrict_pool_to_xml_conf and vpool['PoolID'] !=\
                    self.common.pool_info['arrays_info'][0]['PoolName']:
                LOG.debug("Skip %s" % vpool['PoolID'])
                continue
            LOG.debug("CIM poolinst.path=%s" % vpool.path)
            if minimal:
                pool_dict = {'name': vpool['PoolID'],
                             'cim_path': vpool.path}
                pool_list.append(pool_dict)
                continue

            unsubscribed_num = vpool['TotalManagedSpace'] -\
                vpool['EMCSubscribedCapacity']
            # Check for negative capacity space both here and for
            # remaining space below. Rather than exclude the pool from
            # the output for such a condition, we force the value to 0
            # avoid a negative, and to avoid a pool disappearing than
            # was previously used.
            if vpool['TotalManagedSpace'] < 0:
                total = self.bytes_to_converted_str(0)
                LOG.info(_("The TotalManagedSpace reports as less than "
                           "zero, so force capacity value to zero for "
                           "pool %s.") % vpool['PoolID'])
            else:
                total = self.bytes_to_converted_str(
                    vpool['TotalManagedSpace'])
            free = self.bytes_to_converted_str(vpool['RemainingManagedSpace'])
            used = self.bytes_to_converted_str(
                vpool['TotalManagedSpace'] - vpool['RemainingManagedSpace'])
            if vpool['RemainingManagedSpace'] < 0:
                LOG.info(_("RemainingManagedSpace reports as negative "
                           " [%(free)s] so use 0.") % {'free': free})
                used = self.bytes_to_converted_str(vpool['TotalManagedSpace'])
                free = self.bytes_to_converted_str(0)
            unsubscribed = self.bytes_to_converted_str(unsubscribed_num)
            if unsubscribed_num < 0:
                unsubscribed = self.bytes_to_converted_str(0)
                unsubscribed_num = 0
                LOG.info(_("The unsubscribed space and "
                           "RemainingManagedSpace are negative so "
                           "force a zero value for pool %s.") %
                         vpool['PoolID'])
            free_capacity = unsubscribed
            if CONF.vmax_use_thin_capacity:
                LOG.debug("Using thin free capacity rather than "
                          "unsubscribed capacity.")
                free_capacity = free

            # Put values in dictionary for the pool. The attributes that
            # the console cares about are name, capacity and free_capacity.
            # In 1.2.3, we leave free_capacity calculation as being based
            # on subscribed space rather than thin volume usage.
            # TODO: In a future release, we expect the vmax driver to support
            #       overcommitment of storage so that free capacity would
            #       be based on different metrics.
            # Also, on later version of SMI-S provider (8.0+), some earlier
            # attributes we were obtaining are no longer available.
            # e.g. EMCRemainingRawCapacity, EMCTotalRawCapacity. Since this
            # API is not yet published, they have just been removed.
            pool_dict = {'name': vpool['PoolID'],
                         'capacity': total,
                         'used_capacity': used,
                         'free_capacity': free_capacity,
                         'free_space': free,
                         'unsubscribed': unsubscribed,
                         'cim_path': vpool.path,
                         'id': vpool['InstanceID'],
                         'associated_fast_policies': []}
            if 'EMCRemainingRawCapacity' in vpool:
                pool_dict['raw_free_capacity'] = self.bytes_to_converted_str(
                    vpool['EMCRemainingRawCapacity'])
            if 'EMCPercentSubscribed' in vpool:
                pool_dict['percentage_subscribed'] =\
                    vpool['EMCPercentSubscribed']
            # We check prop existence first for some things because, for
            # example, V3 does not have EMCRaidLevel prop in its pool.
            if 'EMCRaidLevel' in vpool:
                pool_dict['raid_level'] = vpool['EMCRaidLevel']
            if 'EMCPercentageUsed' in vpool:
                pool_dict['percentage_used'] = vpool['EMCPercentageUsed']

            pool_str = six.text_type(vpool.path)
            if pool_str in fast_data:
                for pol_name in fast_data[pool_str]['fast_policies']:
                    pol_info = fast_data['policies'][pol_name]
                    pool_dict['associated_fast_policies'].append(pol_info)
            pool_list.append(pool_dict)
        LOG.debug("Pools info found: %s." % pool_list)
        return pool_list

    def _get_fast_policy_data(self, storageSystemInstanceName):
        """
        Return a dictionary of pool-to-FAST_policy information.
        :param storageSystemInstanceName: The CIM instance name of the VMAX
                                          Symm_StorageSystem.
        """
        LOG.debug("Acquire FAST policy metadata...")
        utils = self.common.utils
        fastvp = self.common.fast
        conn = self.common.conn
        fast_data = {'policies': {}}
        if not fastvp._check_if_fast_supported(conn,
                                               storageSystemInstanceName):
            return fast_data

        tierPolicyServiceInstanceName = utils.get_tier_policy_service(
            conn, storageSystemInstanceName)

        tier_policies = fastvp._get_existing_tier_policies(
            conn, tierPolicyServiceInstanceName)
        LOG.debug("Tier policies: %s" % tier_policies)
        for policy in tier_policies:
            policy_name = policy['PolicyRuleName']
            policy_info = {'fast_policy_name': policy_name}
            fast_data['policies'][policy_name] = policy_info
            storage_tiers = conn.AssociatorNames(
                policy, AssocClass='CIM_AssociatedTierPolicy')
            if not storage_tiers:
                LOG.debug("No storage tiers associated with policy %s" %
                          policy_name)
                continue
            total_gb, free_gb = fastvp.get_capacities_associated_to_policy(
                conn, self.common._get_array_id(), policy_name)
            policy_info['total_gb'] = total_gb
            policy_info['free_gb'] = free_gb
            policy_info['storage_tiers'] = []
            for tier in storage_tiers:
                policy_info['storage_tiers'].append(tier['InstanceID'])
                pools = fastvp.get_associated_pools_from_tier(conn, tier)
                for pool in pools:
                    pool_str = six.text_type(pool)
                    if pool_str not in fast_data:
                        fast_data[pool_str] = {'fast_policies': []}
                    fast_data[pool_str]['fast_policies'].append(policy_name)
        LOG.debug("Returning fast_data: %s" % fast_data)
        return fast_data

    def discover_volumes(self, context, filters=None):
        """ PowerVC specific onboarding function for VMAX volumes.

        Currently supported filters:

          'host_refs': [ WWPN, ... ]
            Returns only volumes that are mapped to at least one of the WWPNS
            in the specified list.  Matching is case-insensitive.

          'vol_refs':
        """
        # Wrapping discoer_volumes in a try/except to catch any SSLErrors
        # that occur from discovery & re-raise with correct string format
        try:
            self.dv_count += 1
            LOG.info(_("PowerVC VMAX discover request [%(num)d]. Filter "
                       "list on next line:\n  %(filters)s") %
                     dict(num=self.dv_count, filters=filters))
            # If we are filtering, create a set from the WWPN list to make
            # filtering easier. Convert the WWPNs to upper-case for consistency
            filter_set = None
            if filters is not None and 'host_refs' in filters:
                upper_refs = [ref.upper() for ref in filters['host_refs']]
                filter_set = set(upper_refs)
                LOG.debug("Discover volume host_refs: %s" % filter_set)
            # If we are filtering get the volume references to get the k2udid
            vol_refs = None
            if filters is not None and 'vol_refs' in filters:
                vol_refs = copy.deepcopy(filters['vol_refs'])
                LOG.debug("Discover volume vol_refs: %s" % vol_refs)
            volume_list = []

            # Get configuration data for backend to refresh info.
            arrayid = self.common._get_array_id()
            pools = self._get_pool_metadata(arrayid, minimal=True)
            arrayEnd = arrayid[-6:]
            masking_view_cache = dict()
            deviceIDs = []
            # loop over supported pools
            for pool in pools:
                # loop over volumes in the pool
                i = 0
                for volInst in self._get_thin_bound_vols(pool['cim_path']):
                    i += 1
                    if volInst['DeviceID'] in deviceIDs:
                        LOG.debug("Volume %s with deviceID %s already "
                                  "found in a previous pool. Skip." %
                                  (volInst['ElementName'],
                                   volInst['DeviceID']))
                        continue
                    LOG.debug("Pool=%s: CIM-volume #%d: %s " %
                              (pool['name'], i, volInst['ElementName']))

                    volume, match_found = self._discover_volume(
                        arrayEnd, pool, volInst, filter_set, vol_refs,
                        masking_view_cache)
                    # If we are not filtering on WWPNs or UID's
                    # and there is a match, add this disk to the list.
                    if (filter_set is None or match_found):
                        LOG.debug("Volume added: %s" % volume)
                        volume_list.append(volume)
                        deviceIDs.append(volInst['DeviceID'])
                        # We check that vol_refs did come into method, but
                        # now have all been found. If so and no NPIV ones to
                        # find, we can break out.
                        if vol_refs is not None and not vol_refs and\
                                not filter_set:
                            LOG.debug("Filter sets are empty...escape loops.")
                            break
                if vol_refs is not None and not vol_refs and not filter_set:
                    break
            LOG.info(_("PowerVC VMAX discover returns %(num)d volumes.") %
                     dict(num=len(volume_list)))
            return volume_list
        except SSLError as exc:
            LOG.exception(six.text_type(exc))
            raise ssl.SSLError(six.text_type(exc))

    def _discover_volume(self, arrayEnd, pool, volInst, filter_set, vol_refs,
                         masking_view_cache):
        """ Helper method to process a single volume cim instance """
        match_found = False
        v = volInst
        volume = {}
        numBlocks = v['NumberOfBlocks']
        if 'ConsumableBlocks' in v:
            numBlocks = v['ConsumableBlocks']
        blockSize = v['BlockSize']
        # divide bytes by bytes-per-Gigabyte for size in GiB
        size = int((numBlocks * blockSize) / 1073741824)
        volume['size'] = size
        statuses = v['StatusDescriptions']
        status = 'available' if 'ONLINE' in statuses else 'error'
        # Volume names not already prefixed by the device ID will have the
        # DeviceID prefix added here. This is important for filtering volume
        # management lists from a storage-admin perspective.
        if len(v['ElementName']) < 6 or v['ElementName'][5] != '+':
            volume['name'] = "%s+%s" % (v['DeviceID'], v['ElementName'])
        else:
            volume['name'] = v['ElementName']
        volume['storage_pool'] = pool['name']
        volume['emcwwn'] = v['EMCWWN']
        volume['is_composite'] = v['IsComposite']
        volume['is_mappedemc'] = v['EMCIsMapped']
        volume['is_thin'] = v['ThinlyProvisioned']
        vol_loc = self._get_volume_dict_from_instance(v)
        r_meta = {rmk.VDISK_ID_KEY: v['DeviceID'],
                  rmk.VDISK_NAME_KEY: v['ElementName'],
                  rmk.VDISK_UID_KEY: v['EMCWWN'],
                  rmk.BYTE_SIZE_KEY: numBlocks * blockSize
                  }
        volume['restricted_metadata'] = r_meta
        volume['provider_location'] = six.text_type(vol_loc)
        volume['is_mapped'] = False  # default
        # If EMCIsMapped is false then volume has not been mapped in the
        # past, so can skip detail checks for mappings.
        if v['EMCIsMapped']:
            # The volume is mapped to at least one or more target ports.
            # Now we check if any initiator ports are masked-in to that
            # mapping.
            initiator_info = self._get_volume_initiator_info(
                v, masking_view_cache)
            if initiator_info['wwpn_set']:
                # Initiator ports are masked in, so we have at least one ITL.
                volume['connection_info'] = initiator_info
                volume['is_mapped'] = True
                if filter_set and (filter_set & initiator_info['wwpn_set']):
                    LOG.debug("Initiator_info has wwpn match/intersection "
                              "with given filter_set. wwpn_set=%s" %
                              (initiator_info['wwpn_set']))
                    match_found = True
            else:
                LOG.debug("Volume deviceID %s reports as EMCIsMapped, "
                          "but no associated Symm_LunMaskingView, "
                          "so onboarding is allowed." % v['DeviceID'])

        volume['support'] = {"status": "supported"}
        if volume['is_mapped'] and not CONF.vmax_allow_manage_inuse:
            status = 'in-use'
            volume['support'] = {
                "status": "not_supported",
                "reasons": [
                    _('This volume is not a candidate for management '
                      'because it is already attached to a virtual '
                      'machine.  To manage this volume with PowerVC, '
                      'you must bring the virtual machine under '
                      'management.  Select to manage the virtual '
                      'machine that has the volume attached.  The '
                      'attached volume will be automatically included '
                      'for management.')
                ]
            }
        elif 'ONLINE' not in statuses:
            status = 'error'
        else:
            status = 'available'
        volume['status'] = status
        # TODO: Need to validate how uuids look for device IDs with 4 & 5
        #      significant digits. Always keep at least 3 digits of DeviceID
        uuid = arrayEnd + v['DeviceID'][:2].lstrip('0') + v['DeviceID'][-3:]
        vol_naa = v['EMCWWN']
        if vol_refs:
            for vol_idx in vol_refs:
                if 'pg83NAA' in vol_idx and vol_naa == vol_idx['pg83NAA']:
                    volume['k2udid'] = vol_idx['k2udid']
                    volume['pg83NAA'] = vol_naa
                    match_found = True
                    # Remove match from both filter lists
                    vol_refs.remove(vol_idx)
                    if vol_idx['pg83NAA'].upper() in filter_set:
                        filter_set.remove(vol_idx['pg83NAA'].upper())
                    break
                match2 = vol_idx['uid'].split("SYMMETRIX")
                # 200970049894B09SYMMETRIX03EMCfcp
                if len(match2) > 1:
                    if uuid in match2[0]:
                        volume['k2udid'] = vol_idx['k2udid']
                        LOG.debug("vol_ref volume Found - adding.")
                        match_found = True
                        # Remove match from both filter lists
                        vol_refs.remove(vol_idx)
                        if vol_idx['uid'].upper() in filter_set:
                            filter_set.remove(vol_idx['uid'].upper())
                        break
        return volume, match_found

    def _get_vmax_full_connect_info(self, volume, connect_info):
        """
        Obtain the attachment information for the volume. This method
        now looks up the full set of initiator-target-lun (ITLs) for
        the volume during discovery rather than splitting the work and
        deferring queries for targets and LUN IDs to the query_volumes flow.
        """
        # initialize list of ITLs
        connect_info['itl_list'] = []
        connect_info['source_wwn'] = []
        num_ignored = 0
        # Get the ProtocolControllerForUnit references
        pcf_units = self.common.conn.References(
            connect_info['vol_inst_path'],
            ResultClass='CIM_ProtocolControllerForUnit')
        for pcf_unit in pcf_units:
            if 'Antecedent' in pcf_unit.path and 'Symm_LunMaskingView' in\
                    pcf_unit.path['Antecedent']['CreationClassName']:
                LOG.debug("Found masking view reference: %s " % pcf_unit.path)
                mv_inst_name = pcf_unit.path['Antecedent']
                if 'DeviceNumber' in pcf_unit:
                    lun_id = int(pcf_unit['DeviceNumber'], 16)

                    # We've already looked up the source wwpns previously,
                    # so just retrieve the mapping for them here.
                    mv_name_text = six.text_type(mv_inst_name)
                    source_wwpns = connect_info['mv_to_wwpns'][mv_name_text]
                    target_names = self.common.conn.AssociatorNames(
                        mv_inst_name,
                        resultClass='Symm_FCSCSIProtocolEndpoint')
                    target_wwpns = [x['Name'] for x in target_names]
                    cur_conn_info = {'vios_wwpn': source_wwpns,
                                     'target_wwpn': target_wwpns,
                                     'target_lun': lun_id}
                    # TODO: We will keep the list of mapped ITLs, but until
                    # 'multi-attach' is complete, the current APIs only
                    # only expect one set, so the root element ITL data will
                    # only come from the last PFC in this loop until this
                    # todo is handled. HOWEVER, 'source_wwn' needs to be
                    # an accumulated set so that dual-VIOS attachments work
                    # like before.
                    connect_info['itl_list'].append(cur_conn_info)
                    connect_info['target_wwn'] = target_wwpns
                    if 'target_lun' in connect_info:
                        LOG.info(_("Volume %(vol)s has more than one "
                                   "attachment. ITL list: %(itls)s") %
                                 dict(vol=volume['name'],
                                      itls=connect_info['itl_list']))
                    connect_info['target_lun'] = lun_id
                connect_info['source_wwn'] = list(connect_info['wwpn_set'])
            else:
                num_ignored += 1

        # Remove old initiator information.
        del connect_info['wwpn_set']
        del connect_info['mv_to_wwpns']
        del connect_info['vol_inst_path']
        LOG.debug("Volume connect_info=%s, references_ignored=%d." %
                  (connect_info, num_ignored))
        return connect_info

    def _get_volume_initiator_info(self, volume_inst, masking_view_dict):
        """
        This is a performance optimized method that uses two calls and
        is faster than a single call to get SE_StorageHardwareID instance
        names from the volume_inst.
        :returns: List of source wwpns that the volume is attached to.
                  List of SE_StorageHardwareID CIM instance names for the
                  wwpns.
        """
        initiator_info = {'wwpn_set': set(), 'mv_to_wwpns': {},
                          'vol_inst_path': volume_inst.path}
        masking_view_inst_names = self.common.conn.AssociatorNames(
            volume_inst.path,
            ResultClass='EMC_LunMaskingSCSIProtocolController')
        if len(masking_view_inst_names) > 0:
            for mv_inst_name in masking_view_inst_names:
                src_wwpns = None
                # try to get from the prepopulated dict
                if masking_view_dict.get(six.text_type(mv_inst_name)):
                    LOG.debug("masking view found in cache %s:" %
                              six.text_type(mv_inst_name))
                    src_wwpns = masking_view_dict[six.text_type(mv_inst_name)]
                # if it is not already queried we go to the backend
                if not src_wwpns:
                    hwid_names = self.common.conn.AssociatorNames(
                        mv_inst_name, ResultClass='SE_StorageHardwareID')
                    # Get the upper case list of initiator wwpns for the
                    # current masking view and add to mapping dictionary.
                    src_wwpns = set(self.common.utils.hwid_names_to_wwpns(
                        hwid_names, toupper=True))
                    masking_view_dict[six.text_type(mv_inst_name)] = src_wwpns

                initiator_info['mv_to_wwpns'][six.text_type(mv_inst_name)] =\
                    src_wwpns
                initiator_info['wwpn_set'].update(src_wwpns)
        if initiator_info['wwpn_set']:
            LOG.debug("Initiator_info=%s. Volume=%s." %
                      (initiator_info, volume_inst['Name']))
        else:
            LOG.debug("No initiator WWPNs masked-in. Volume=%s" %
                      volume_inst['Name'])
        return initiator_info

    def _get_volume_dict_from_instance(self, volInst):
        """Given the volume instance, determine the volume dictionary.
        This is also known as the provider_location.
        Compare to emc_vmax_provision->get_volume_dict_from_job().
        """
        return self.common._get_volume_dict_from_path(volInst.path)

    def _get_thin_bound_vols(self, pool_cim_path):
        """ Get thin TDEV associators bound to pool """
        LOG.debug("pool_cim_path=%s" % pool_cim_path)
        bound_vols = []
        associators = self.common.conn.Associators(
            pool_cim_path,
            ResultClass='CIM_StorageVolume',
            AssocClass='CIM_AllocatedFromStoragePool')
        j = 0
        for dev in associators:
            LOG.debug("Dev %d: %s, caption=%s " %
                      (j, dev['DeviceID'], dev['Caption']))
            j += 1
            if dev['ThinlyProvisioned']:
                bound_vols.append(dev)
        LOG.debug("Returning %d allocated & bound volumes for %s." %
                  (len(bound_vols), pool_cim_path))
        return bound_vols

    def query_volumes(self, context, volumes, server_info={}, mark_boot=True):
        """Returns import-related details for the Volumes from the Provider"""
        LOG.info(_("PowerVC VMAX query request: num_vols=%(num)d,...\n  "
                   "server_info=%(info)s, mark_boot=%(boot)s") %
                 dict(num=len(volumes), info=server_info, boot=mark_boot))
        # TODO: If there is significant overlap between SVC, EMC, XIV, etc,
        #      then logic should be restructured at the volume manager
        #      layer.
        server_copy = copy.deepcopy(server_info)

        uid_map, wwpn_map = (dict(), dict())
        returned_volumes, boot_map = (list(), dict())
        for server_id, server in server_info.iteritems():
            LOG.debug("server_id=%s, server=%s" % (server_id, server))
            mark_boot = (server.get('boot') is not True)
            for volume in server.get('volumes', []):
                uid = volume.get('unique_device_id')
                wwpns = volume.get('wwpns', [])
                udid = volume.get('k2udid')
                pg83_naa = volume.get('pg83NAA')
                mount = volume.get('mountpoint')
                vios_wwpn = volume.get('vios_wwpns', None)
                # Build maps of the UID and WWPN depending on which is set
                if pg83_naa is not None:
                    if pg83_naa.upper() not in uid_map:
                        server_dict = {server_id: (mark_boot, mount,
                                                   vios_wwpn, udid)}
                    else:
                        server_dict = uid_map[pg83_naa.upper()]
                        server_dict[server_id] = (mark_boot, mount,
                                                  vios_wwpn, udid)

                    uid_map[pg83_naa] = server_dict
                    LOG.debug("uid_map[%s] = %s" % (pg83_naa.upper(),
                                                    uid_map[pg83_naa.upper()]))
                for wwpn in wwpns:
                    wwpn_map[wwpn.upper()] = (server_id, mark_boot, udid)
                if wwpns:
                    LOG.debug("wwpn_map for upper %s = %s" %
                              (wwpns, (server_id, mark_boot, udid)))
        # Loop over passed in managed volumes
        vn = 0
        for volume in volumes:
            vn += 1
            volume_copy = volume.copy()
            server_id = None
            LOG.debug("Iter-volume #%d: %s." % (vn, volume_copy))
            # Parse the UID out of the MetaData to use for matching Servers
            metadata = volume_copy.get('restricted_metadata', {})
            uid = metadata.get(rmk.VDISK_ID_KEY)

            # TODO: Need to validate how uuids look for device IDs with 4 & 5
            #      significant digits. Always keep at least 3 digits.
            udid = volume_copy.get('k2udid', None)
            vol_pg83 = volume_copy.get('pg83NAA', None)
            volume_copy['attach_info'] = {}
            connect_info = volume_copy.pop('connection_info', None)
            if connect_info:
                # Get the missing target and LUN id information now.
                connect_info = self._get_vmax_full_connect_info(volume_copy,
                                                                connect_info)
                LOG.debug("connection_info: %s" % connect_info)
                # Default to the first one if the WWPN's weren't specified
                itl_list = connect_info.get('itl_list', [])
                if itl_list:
                    volume_copy['itl_list'] = itl_list
                    for cinfo in itl_list:
                        wwpns = cinfo.get('vios_wwpn', [])
                        LOG.debug("wwpns in query volume : %s" % wwpns)
                        # Loop through each WWPN's looking for the Server
                        for wwpn in wwpns:
                            # If we found a Server for the WWPN, use this Info
                            if wwpn_map.get(wwpn.upper()) is not None:
                                server_id = wwpn_map[wwpn.upper()][0]
                                if server_id not in volume_copy['attach_info']:
                                    mark_boot = wwpn_map[wwpn.upper()][1]
                                    mount_point = wwpn_map[wwpn.upper()][2]
                                    connection_info = None
                                    if cinfo:
                                        conn_info = dict()
                                        conn_info['target_wwn'] =\
                                            copy.deepcopy(cinfo['target_wwpn'])
                                        conn_info['source_wwn'] =\
                                            copy.deepcopy(cinfo['vios_wwpn'])
                                        conn_info['target_lun'] =\
                                            copy.deepcopy(cinfo['target_lun'])
                                        connection_info =\
                                            self._get_connection_info(
                                                uid, conn_info,
                                                CONNECTION_TYPE_NPIV)
                                        self._add_attach_info(
                                            volume_copy, boot_map,
                                            connection_info, server_id,
                                            mountpoint=mount_point,
                                            mark_boot=mark_boot)
            if vol_pg83 in uid_map:
                server_dict = uid_map[vol_pg83]
                for server_id in server_dict.keys():
                    if server_id not in volume_copy['attach_info']:
                        mark_boot = server_dict[server_id][0]
                        mount_point = server_dict[server_id][1]
                        connection_info = None
                        server_wwpn = server_dict[server_id][2]
                        udid = server_dict[server_id][3]
                        # get correct connection info and itl list
                        valid_itls = self._get_vscsi_itl_info(server_wwpn,
                                                              volume_copy)
                        if valid_itls:
                            # If given the VDisk UID from the VM, then
                            # it isn't NPIV
                            # pass the first itl to extract the target_wwn
                            cinfo = valid_itls[0]
                            connection_info = self._get_connection_info(
                                vol_pg83, cinfo, CONNECTION_TYPE_VSCSI,
                                udid=udid)
                            connection_info['vscsi_target_lun'] =\
                                valid_itls
                            connection_info['data']['vendor'] =\
                                "EMC"
                            self._add_attach_info(volume_copy, boot_map,
                                                  connection_info, server_id,
                                                  mountpoint=mount_point,
                                                  mark_boot=mark_boot)
            if server_id in server_copy:
                del server_copy[server_id]
            try:
                # Set the metadata information. Note that this is not
                # kept in sync with backend at this time. It just provides
                # useful data for VMAX volumes
                geom = self._get_volume_geometry(volume_copy,
                                                 set_metadata=False)
                volume_copy['metadata'] = geom['metadata']
            except Exception as ex:
                # Log and continue for now. This is nice-to-have only.
                LOG.exception(ex)
            # Add the Copy of the Volume to the list to be returned
            LOG.debug("#%d: New volume_copy entry: %s" % (vn, volume_copy))
            returned_volumes.append(volume_copy)
        if server_copy:
            LOG.warn(_("Server entries not matched: %s.") % server_copy)
        LOG.info(_("PowerVC VMAX query returns volume list of length %(num)d"
                   "...\n  %(vols)s") %
                 dict(num=len(returned_volumes), vols=returned_volumes))
        return returned_volumes

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        num_vols = len(volumes)
        delvols, devid_to_name = (list(), dict())
        t1 = time.time()

        def _log_time():
            betw = 0
            if self.time_last_dcheck is not None:
                betw = t1 - self.time_last_dcheck
            self.time_last_dcheck = time.time()
            delta = self.time_last_dcheck - t1
            if delta == 0:
                delta = 0.0001  # avoid any div-by-0
            percent = (delta / (delta + betw)) * 100
            LOG.debug("Existence not verified for %d of %d volumes. Check time"
                      "=%.3f, Percent of last cycle = %.2f %%, last total "
                      "volume count = %d." %
                      (len(delvols), num_vols, delta, percent,
                       self.num_array_vols))
            if betw > 0 and percent > MAX_CHECK_DEL_PCT:
                LOG.warn(_("Spent more than %(pct)d%% of the "
                           "deleted_volume_check_interval time verifying "
                           "VMAX volumes. Consider specifying an interval "
                           "option value significantly longer than "
                           "%(secs).0f seconds.") %
                         {'pct': MAX_CHECK_DEL_PCT, 'secs': betw})
            return delta

        # For VMAX, it is straightforward to verify the existence of a
        # DeviceID, but that does not tell us whether it has been deleted
        # out-of-band and then recreated here with a new ID. To verify
        # volumes, the ElementName must be validated against what PowerVC
        # thinks it should be and that can be performance instensive,
        # depending on the number of volumes.
        #
        # There is the regular path check (or full check), and there is
        # an alternate way of checking for deleted volumes.
        # Here is where we check wether to do the alternate path, which
        # loops over every volume to check and does a _find_lun() on it
        # individually. This should be faster when checking a smaller
        # % of volumes versus how many total vols are on the back end.
        #
        # Take the alternative path if:
        #   (1) The FULL_DEL_VOL_CHK_INTERVAL in seconds has not passed
        #       since the last time we took the regular path (3 hours).  AND
        #   (2) % of total backend volumes to check is less than 5%   OR
        #       The alternative path time was at least as fast as
        #       the regular path time the last we checked.
        pct_to_check = (float(num_vols) / float(self.num_array_vols)) * 100
        secs_since_last_fcheck = t1 - self.time_last_fcheck
        LOG.debug("Check for existence of %d volumes. %% to check="
                  "%.1f, last regular check=%d seconds ago. " %
                  (num_vols, pct_to_check, secs_since_last_fcheck))
        if ((secs_since_last_fcheck < FULL_DEL_VOL_CHK_INTERVAL) and
                ((pct_to_check < 5) or
                 (self.delta_last_mcheck <= self.delta_last_fcheck))):
            LOG.debug("Go through managed volumes individually.")
            for vol in volumes:
                # The following check is here to mimic regular path below
                if not vol.get('provider_location'):
                    LOG.debug("No provider_location for %s." % vol['id'])
                    continue
                if not self.common._find_lun(vol, skip_cache=True):
                    delvols.append(vol)
            self.delta_last_mcheck = _log_time()
            self.time_last_mcheck = t1
            return delvols

        # Otherwise full path...
        LOG.debug("Retrieve all VMAX volume data for comparison.")
        # We do a quick check here to see if we should reset the time it
        # took to do that alternate path (managed vol check) if a default
        # of 6 hours have passed, and if the conf inteval between checks
        # is not more than that same 6 hours.
        if (t1 - self.time_last_dcheck < (FULL_DEL_VOL_CHK_INTERVAL * 2) and
                t1 - self.time_last_mcheck >
                (FULL_DEL_VOL_CHK_INTERVAL * 2) and pct_to_check < 75):
            self.delta_last_mcheck = 0
            LOG.debug("Reset delta_last_mcheck to get new baseline for "
                      "alternate path on next check.")
        arrayid = self.common._get_array_id()
        pools = self._get_pool_metadata(arrayid, minimal=True)
        # Loop through the Support Pools trying to get the list of Volumes
        for pool in pools:
            # Get backend vols
            bvols = self.common.conn.Associators(
                pool['cim_path'], ResultClass='CIM_StorageVolume',
                AssocClass='CIM_AllocatedFromStoragePool')
            # Loop through the Volume Paths, building a mapping of deviceIDs
            # to volume names. We have changed to index by deviceID instead
            # of the whole provider_location because we cannot guarantee
            # that we will build the provider_location in the same way
            # that the backend builds it.
            for bvol in bvols:
                deviceID = bvol.path['DeviceID']
                devid_to_name[deviceID] = bvol['ElementName'] if\
                    'ElementName' in bvol else ''

        # track time we last retrieved all volumes and the number of them.
        self.num_array_vols = len(devid_to_name)
        self.time_last_fcheck = t1
        # Collect metadata for extra volume corrlelation
        metads = powervc_db_api.volume_restricted_metadata_get_all(context)

        # Now loop through the passed in volumes seeing which don't exist
        for volume in volumes:
            volloc = volume.get('provider_location')
            if not volloc:
                LOG.debug("No provider_location.")
                continue
            cur_devid = ast.literal_eval(volloc)['keybindings']['DeviceID']
            # If the Volume has a Path and isn't on the system, mark as deleted
            if cur_devid not in devid_to_name or volume['id'] not in metads:
                LOG.debug("volloc not in devid_to_name=%s, id not in "
                          "metads=%s" % (volloc not in devid_to_name,
                                         volume['id'] not in metads))
                delvols.append(volume)
                continue
            # Also handle case where Path still is on the backend VMAX, but
            # it represents a different volume with a reused DeviceID.
            metadata = metads[volume['id']]
            m_dict = dict([(m['key'], m['value']) for m in metadata])
            if rmk.VDISK_NAME_KEY in m_dict and m_dict[rmk.VDISK_NAME_KEY]:
                metadata_name = m_dict[rmk.VDISK_NAME_KEY]
            if devid_to_name[cur_devid] != metadata_name:
                LOG.info(_("Mismatched backend ElementName '%(bname)s' for "
                           "metadata name originally created with: %(mname)s. "
                           "Append to deleted volumes.") %
                         dict(bname=devid_to_name[cur_devid],
                              mname=metadata_name))
                delvols.append(volume)

        self.delta_last_fcheck = _log_time()
        return delvols

    def check_volume_health(self, volume):
        """
        Throws exceptions if something is wrong with the passed-in volume.
        Checks for existence of the underlying vdisk
        """
        LOG.debug("check_volume_health(): volume=%s" % volume.__dict__)
        vdisk_id = self._incorporate_restricted_metadata(volume)

        if vdisk_id is None or not self._check_vdisk_exists(volume, vdisk_id):
            status = volume['lun_status'] if 'lun_status' in volume else\
                volume['status']
            raise powervc_exception.EMCVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id,
                status=status)

    def _check_vdisk_exists(self, volume, vdisk_id):
        """ Check for the volume on the backend array """
        instance = None
        try:
            instance = self.common._find_lun(volume)
            if not instance:
                uid = volume['restricted_metadata'][rmk.VDISK_UID_KEY]
                LOG.info(_("Volume not found on Symmetrix array for id "
                           "%(vdisk_id)s and name '%(name)s' and uid "
                           "%(vdisk_uid)s.") %
                         dict(vdisk_id=vdisk_id, name=volume['name'],
                              vdisk_uid=uid))
        except Exception as ex:
            LOG.exception(ex)  # Can log now since this is unexpected
            LOG.debug("Volume '%s' with id=%s and vdisk_id=%s not found. "
                      "Error: %s" %
                      (volume['name'], volume['id'], vdisk_id,
                       six.text_type(ex)))
            # The status here is temporary - not persisted.
            volume['lun_status'] = _("Unable to lookup LUN")
        if not instance:
            return False
        else:
            LOG.debug("Volume found for vdisk_id=%s and display_name: %s" %
                      (vdisk_id, volume['display_name']))
            return True

    def _incorporate_restricted_metadata(self, volume_ref):
        """Query the Cinder database to retrieve any access-restricted
           metadata and attach it to the volume data.
           Returns the vdisk id, which most of the time is what we were after.
           If there is no restricted metadata, tries to find the disk and
           creates it.
        """
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume_ref['id'])

        # If the vdisk ID is not in the metadata, this could be because of an
        # upgrade from a version that didn't use it.  We try to find it by
        # looking for a volume on the SAN with a matching name
        if rmk.VDISK_ID_KEY not in metadata:
            LOG.info(_("There is no restricted metadata for volume: %s.") %
                     volume_ref)
            return None

        volume_ref['restricted_metadata'] = metadata
        return metadata[rmk.VDISK_ID_KEY]

    def bytes_to_converted_str(self, num_bytes, precision=1):
        """ Return the formatted string value of the input bytes.
            Returned string will be in GB, unless value is less than 1.
            Then MB units returned.
        """
        prec = '%' + (".%df" % precision)
        if hasattr(num_bytes, 'value'):
            num_bytes = num_bytes.value
            LOG.debug("num_bytes=%d, dir()=%s." % (num_bytes, dir(num_bytes)))
        oneMB = 1024 ** 2
        mb = float(num_bytes) / float(oneMB)
        if mb < 1:
            return (prec % mb) + ' MB'
        elif mb < 1024:
            return (prec % mb).strip('0').strip('.') + ' MB'
        else:
            return (prec % (mb / 1024)).strip('0').strip('.') + ' GB'
        # else:
        #     return (prec % (mb / oneMB)).strip('0').strip('.') + ' TB'

    def _generate_candidate_caption(self, volume):
        """ Generate candidate caption/friendly name for vol creation"""
        if volume.get("display_name"):
            LOG.debug('ENTER: display_name %s' % volume['display_name'])

            # What we are generating here is a caption for the volume.
            # It will be used as a name for onboarding, but the string
            # does not have the extensive constraints that the SVC does.
            # But we still limit the string to the following characters.
            display_name_nospace = re.sub(r'[^a-zA-Z0-9_\-#@.!,]', "_",
                                          volume['display_name'])

            # Then we substitute it into the volume_name_template specified
            # in our configuration file
            friendly_name = CONF.volume_name_template % display_name_nospace

            # We would like to use this name, but fall back to the generic
            # (UUID-based) name if that doesn't work for some reason.
            LOG.debug('EXIT: friendly_name %s' % friendly_name)
            return friendly_name
        else:
            return 'TDEV - ' + volume['name']

    def validate_connector(self, connector):
        """
        Override the FibreChannelDriver version of validate_connector since
        it requires that wwnns be provided as well, and PowerVC does not
        do that. Only validate wwpns here.
        """
        self.validate_connector_has_setting(connector, 'wwpns')

    def _build_initiator_target_map(self, storage_system, volume, connector,
                                    maskingView=None):
        """
        Override the superclass version of this method.
        TODO: Push or merge this with community fix. Waiting for community
              to remove the dependency that the NPIV wwpns are already
              visible to the array as SE_StorageHardwareID CIM instances.
        """
        if 'targ_wwpns' in connector:
            # if already looked up during this operation, then just return
            LOG.debug("Return from computed connector: %s" %
                      connector['targ_map'])
            return connector['targ_wwpns'], connector['targ_map']
        # Find targets by association rather than by the
        # EMCGetTargetEndpoints method on the hardware ID service.
        LOG.debug("PowerVC impl of obtaining target wwpns.")
        target_wwns = []
        if maskingView:
            LOG.debug("Use masking view to lookup target ports: %s" %
                      six.text_type(maskingView))
            # If masking view passed, then the correct lookup is to use that.
            endpoints = self.common.conn.Associators(
                maskingView, resultClass='Symm_FCSCSIProtocolEndpoint')
            if endpoints:
                target_wwns = [x['Name'] for x in endpoints]
            else:
                LOG.warn(_("No Symm_FCSCSIProtocolEndpoints associated "
                           "with masking view: %(view)s") %
                         dict(view=six.text_type(maskingView)))
        else:
            # Otherwise lookup by initiator association. This could be
            # sufficient for NPIV.
            target_wwns = self._get_target_wwns_by_assoc(storage_system,
                                                         connector)
        init_targ_map = {}
        for initiator in connector['wwpns']:
            init_targ_map[initiator] = target_wwns
        LOG.debug("Initiator-target-map: %s" % init_targ_map)
        return target_wwns, init_targ_map

    def _get_target_wwns_by_assoc(self, storage_system, connector):
        """
        Lookup target wwpns by Symm_FCSCSIProtocolEndpoint association.
        """
        masker = self.common.masking
        hwNames = masker._get_storage_hardware_id_instance_names(
            self.common.conn, connector['wwpns'], storage_system)
        LOG.debug("hwNames: %s" % hwNames)
        connector['initiator_hwnames'] = hwNames
        target_set = set()
        for hwName in hwNames:
            endpoints = self.common.conn.Associators(
                hwName, resultClass='Symm_FCSCSIProtocolEndpoint')
            if endpoints:
                endpoint_names = [x['Name'] for x in endpoints]
                target_set.update(set(endpoint_names))
            else:
                LOG.info(_("No endpoints found for Hardware ID: %s") % hwName)
        LOG.debug("Return target_wwn_set: %s" % target_set)
        return list(target_set)

    def _get_volume_geometry(self, volume, set_metadata=True):
        """
        Try to obtain volume geometry information from the provider.
        Return a dictionary with the information.

        :param volume: The volume to get the geometry/metadata information
                       for.
        :param set_metadata: By default this method will update the volume
                       metadata with specific pieces of information. This
                       is instead of the caller using the returned data to
                       update the 'provider_geometry' property directly in
                       the volume. Since provider_geometry does not get
                       returned by the volume cinder API, it is not as useful
                       as putting certain properties in the metadata for
                       use by PowerVC UI and test verification of storage
                       template options. Pass False to skip setting metadata.
        """
        LOG.debug("Geometry ENTRY for volume %s." % volume['name'])
        geometry = {}
        capacities = []
        ctxt = context.get_admin_context()
        instance = self.common._find_lun(volume, skip_verify=True,
                                         skip_cache=True)
        if not instance:
            return _('Volume not found yet.')
        vol_path = instance.path

        isComposite = self.common.utils.check_if_volume_is_composite(
            self.common.conn, instance)
        geometry['is_composite'] = isComposite
        geometry['blocks'] = int(instance['ConsumableBlocks'])
        if 'EMCConsumableBlocksWritten' in instance:
            wroteb = float(instance['EMCConsumableBlocksWritten'])
            wrote_prct = (wroteb / float(geometry['blocks'])) * 100.0
            geometry['pct_consumed'] = "%.1f" % wrote_prct
        if 'EMCRaidLevel' in instance:
            geometry['raid_level'] = six.text_type(instance['EMCRaidLevel'])
        # fast policy
        geometry['fast_policy'] = self.common._get_fast_policy_for_vol(
            instance)
        # check for meta-head
        metaHeadInstanceName = self.common.utils.get_volume_meta_head(
            self.common.conn, vol_path)
        LOG.debug("metaHeadInstanceName: %s" % metaHeadInstanceName)
        if metaHeadInstanceName is None:  # simple volume
            geometry['meta_config'] = 'concatenated'
            geometry['num_metas'] = 0
        else:  # composite volume with meta device members
            metaHeadInst = self.common.conn.GetInstance(metaHeadInstanceName)
            propList = metaHeadInst.properties.items()
            for properties in propList:
                if properties[0] == 'IsConcatenated':
                    isConcatenated = six.text_type(properties[1].value)
                    if 'True' in isConcatenated:
                        geometry['meta_config'] = 'concatenated'
                    else:
                        geometry['meta_config'] = 'striped'
                elif properties[0] == 'ExtentStripeLength':
                    stripeLen = int(six.text_type(properties[1].value))
                    geometry['extent_stripe_len'] = stripeLen
            # Get meta members' capacity info
            elementCompositionService = (
                self.common.utils.find_element_composition_service(
                    self.common.conn, self._get_array_name()))
            meta_extentNames = self.common.utils.get_volume_members(
                self.common.conn, instance, metaHeadInstanceName,
                elementCompositionService)
            if meta_extentNames:
                geometry['num_metas'] = len(meta_extentNames)
                # geometry['metas'] = {}
                LOG.debug("meta_extent inst names = %s" %
                          six.text_type(meta_extentNames))
                capacities = (
                    self.common.utils.get_meta_members_capacity_in_byte(
                        self.common.conn, meta_extentNames))
                # Recompute the head element capacity if needed
                totalCapacity = (instance['ConsumableBlocks'] *
                                 instance['BlockSize'])
                if capacities[0] == totalCapacity:
                    capacities[0] = totalCapacity - sum(capacities[1:])
                    LOG.debug("Recalculate meta head capacity of %d to %d." %
                              (totalCapacity, capacities[0]))
                else:
                    LOG.debug("Capacities are as expected.")
                if len(set(capacities)) == 1:
                    geometry['equal_metas'] = 'True'
                else:
                    geometry['equal_metas'] = 'False'
            else:
                geometry['num_metas'] = 0
                LOG.warn(_("%(vol)s is a composite volume, but cannot find "
                           "any meta members."))

        # Determine the storage pool for V2 arrays.
        # We add this to the metadata because the volume could be bound
        # to a different pool than the one in the storage template (type)
        # that the volume was created with - a quirk of VMAX cloning.
        meta = dict()
        if not self.common.pool_info['is_v3']:
            pools = self.common.conn.AssociatorNames(
                vol_path, ResultClass='Symm_VirtualProvisioningPool')
            if pools:
                # Handle differences between SMI-S v4.x and v8.x.
                if '+TP+' in pools[0]['InstanceID']:
                    poolid = pools[0]['InstanceID'].split('+TP+')[1]
                elif '+-TP-+-' in pools[0]['InstanceID']:
                    poolid = pools[0]['InstanceID'].split('+-TP-+-')[1]
                else:
                    LOG.debug('Not a recognized pool format for %s. '
                              'Continuing.' % pools[0]['InstanceID'])
                    poolid = pools[0]['InstanceID']
                geometry['storage_pool'] = poolid
                meta['storage_pool'] = poolid

        # Do metadata processing
        meta['geom_meta_config'] = geometry['meta_config']
        meta['fast_policy'] = geometry['fast_policy'] if\
            geometry['fast_policy'] else 'None'
        meta['geom_num_members'] = geometry['num_metas'] if\
            geometry['num_metas'] >= 1 else 1
        if capacities:
            gb_str = ""
            for cap_bytes in capacities:
                size_str = self.bytes_to_converted_str(cap_bytes)
                if gb_str:
                    gb_str += (', ' + size_str)
                else:
                    gb_str = size_str
            meta['geom_member_sizes'] = gb_str
        LOG.debug("Volume metadata: %s " % meta)
        cur_meta = {}
        if 'id' in volume:
            cur_meta = db_api.volume_metadata_get(ctxt, volume['id'])
        cur_meta.update(meta)
        LOG.debug("Existing volume metadata to replace: %s " % cur_meta)
        if set_metadata:
            # Load existing volume metadata and update it with current info.
            db_api.volume_metadata_update(ctxt, volume['id'], cur_meta, False)
        geometry['metadata'] = cur_meta

        return geometry


def add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in
                     range(0, len(wwpn), 2)]).lower()


def strip_colons(wwpn):
    if len(wwpn) > 16:
        return wwpn.lower().replace(":", "")
    else:
        return wwpn.lower()
