# =================================================================
#
# Copyright 2014, 2016 IBM Corp.
# All Rights Reserved.
# =================================================================
# Copyright (c) 2012 - 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
import copy
from oslo_log import log as logging
from oslo_utils import excutils
from cinder import context, exception
from cinder.zonemanager import utils as fczm_utils
from powervc_cinder.zonemanager import powervc_utils as pvc_fczm_utils
# don't do any zoning requests
fczm_utils.AddFCZone = pvc_fczm_utils.ignore_zoning_request
fczm_utils.RemoveFCZone = pvc_fczm_utils.ignore_zoning_request
from cinder.volume.drivers.emc import emc_cli_fc
from cinder.volume.drivers.emc import emc_vnx_cli
from cinder.exception import EMCVnxCLICmdError
from cinder.volume.drivers.emc.emc_vnx_cli import PropertyDescriptor
from cinder import db as db_api
from cinder.volume import volume_types

from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import discovery_driver
from powervc_cinder.volume.discovery_driver import \
    CONNECTION_TYPE_NPIV, CONNECTION_TYPE_VSCSI
from powervc_cinder import _
from powervc_cinder import exception as powervc_exception

LOG = logging.getLogger(__name__)

# TODO: move to common location
RESTRICTED_METADATA_VDISK_ID_KEY = "vdisk_id"
RESTRICTED_METADATA_VDISK_UID_KEY = "vdisk_uid"
RESTRICTED_METADATA_VDISK_NAME_KEY = "vdisk_name"
RESTRICTED_METADATA_BYTE_SIZE_KEY = "byte_size"
RESTRICTED_METADATA_UNIQUE_ID = "uid"
METADATA_VOLUME_WWN = "volume_wwn"


class EMCVNXProductDriver(
        discovery_driver.VolumeDiscoveryDriver,
        emc_cli_fc.EMCCLIFCDriver):

    POOL_USER_CAPACITY = PropertyDescriptor(
        '-userCap',
        'User Capacity *\(GBs\) *:\s*(.*)\s*',
        'user_capacity_gb',
        float)

    LUN_STATE = PropertyDescriptor('-state',
                                   'Current State:\s*(.*)\s*',
                                   'state')
    LUN_CAPACITY = PropertyDescriptor('-userCap',
                                      'User Capacity \(GBs\):\s*(.*)\s*',
                                      'total_capacity_gb',
                                      float)
    LUN_NAME = PropertyDescriptor('-name',
                                  'Name:\s*(.*)\s*',
                                  'lun_name')
    LUN_ID = PropertyDescriptor('-id',
                                'LOGICAL UNIT NUMBER\s*(\d+)\s*',
                                'lun_id',
                                int)
    LUN_POOL_NAME = PropertyDescriptor('-poolName',
                                       'Pool Name:\s*(.*)\s*',
                                       'pool_name')

    LUN_UID = PropertyDescriptor('-uid',
                                 'UID:\s*(.*)\s*',
                                 'uid')

    def __init__(self, *args, **kwargs):
        super(EMCVNXProductDriver, self).__init__(*args, **kwargs)

    def _build_default_opts(self):
        default_opts = {
            'storage_pool': self.configuration.storage_vnx_pool_names,
            'storagetype:pool': 'a'}
        return default_opts

    def create_cloned_volume(self, tgt_volume, src_volume):
        """
        Overrides the superclass create_cloned_volume. Adds
        restricted metadata information to the cloned volume
        and sets default volume type if one is not specified
        """
        s = super(EMCVNXProductDriver, self)

        ctxt = context.get_admin_context()
        volume_type_id = tgt_volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                tgt_volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % tgt_volume)
                self.db.volume_update(ctxt, tgt_volume['id'],
                                      {'volume_type_id': volume_type_id})
        host_name = tgt_volume['host']
        try:
            vol = s.create_cloned_volume(tgt_volume, src_volume)
        except exception.VolumeBackendAPIException as ex:
            LOG.error(_("Error when creating cloned volume %s") % ex)
            errors = {'Create Failure description': ex.message}
            powervc_db_api.\
                volume_restricted_metadata_update_or_create(ctxt,
                                                            tgt_volume['id'],
                                                            errors)
            self._rollback_snapshot(tgt_volume, src_volume)
            raise ex
        vol['host'] = host_name
        # Add restricted metadata to volume
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: tgt_volume['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: tgt_volume['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: tgt_volume['name']
                    }

        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, tgt_volume['id'],
                                                        metadata)
        if vol is None:
            vol = {}
        meta = db_api.volume_metadata_get(ctxt, tgt_volume['id'])
        meta = meta if meta else dict()
        lun_id = self.get_lun_uid(tgt_volume['name'])
        if lun_id:
            lun_id = lun_id.replace(':', '')
            meta[METADATA_VOLUME_WWN] = lun_id
        vol['metadata'] = meta
        return vol

    def _rollback_snapshot(self, tgt_volume, src_volume):
        source_volume_name = src_volume['name']
        volume_size = src_volume['size']
        consistencygroup_id = src_volume['consistencygroup_id']
        snapshot_name = 'tmp-snap-%s' % tgt_volume['id']
        tmp_cgsnapshot_name = None
        if consistencygroup_id:
            tmp_cgsnapshot_name = 'tmp-cgsnapshot-%s' % tgt_volume['id']

        snapshot = {
            'name': snapshot_name,
            'volume_name': source_volume_name,
            'volume_size': volume_size,
            'cgsnapshot_id': tmp_cgsnapshot_name,
            'consistencygroup_id': consistencygroup_id,
            'id': tmp_cgsnapshot_name
        }

        try:
            # detach snapshot
            self._detach_snap_mount_pnt(tgt_volume['name'])
            # Delete temp Snapshot
            if consistencygroup_id:
                self.cli._client.delete_cgsnapshot(snapshot)
            else:
                self.delete_snapshot(snapshot)
            self._delete_smp(tgt_volume['name'])
        except Exception as ex:
            LOG.error(_("Error during rollback_snapshot for source vol "
                        "%(src_vol)s target vol %(tgt_vol)s snapshot "
                        "%(snap)s. Exception %(exc)s") %
                      dict(src_vol=src_volume['id'],
                           tgt_vol=tgt_volume['id'],
                           snap=snapshot_name,
                           exc=ex))

    def _detach_snap_mount_pnt(self, snap_mount_pnt):
        command_detach_smp = ('lun', '-detach',
                              '-name', snap_mount_pnt)

        out, rc = self.cli._client.command_execute(*command_detach_smp)
        if rc != 0:
            raise EMCVnxCLICmdError(command_detach_smp, rc, out)
        else:
            LOG.info(_('Snapshot mount point %s was detached successfully.') %
                     snap_mount_pnt)
        return True

    def _delete_smp(self, snap_mount_pnt):
        command_delete_smp = ('lun', '-destroy', '-name', snap_mount_pnt,
                              '-destroySnapshots', '-forceDetach', '-o')
        out, rc = self.cli._client.command_execute(*command_delete_smp)
        if rc != 0:
            raise EMCVnxCLICmdError(command_delete_smp, rc, out)
        else:
            LOG.info(_('Snapshot mount point %s was deleted successfully.') %
                     snap_mount_pnt)
        return True

    def discover_volumes(self, context, filters=None):
        # If we are filtering, create a set from the WWPN list to make
        # filtering easier.  Convert the WWPNs to upper-case for consistency
        filter_set = None
        if filters is not None and 'host_refs' in filters:
            upper_refs = [ref.upper() for ref in filters['host_refs']]
            filter_set = set(upper_refs)
        # If we are filtering get the volume references to get the k2udid
        vol_refs = None
        if filters is not None and 'vol_refs' in filters:
            vol_refs = filters['vol_refs']
            LOG.debug("Discover volume vol_refs: %s" % vol_refs)

        luns = self._get_lun_list(filter_set, vol_refs)

        return luns

    def validate_connector(self, connector):
        """
        Override the FibreChannelDriver version of validate_connector since
        it requires that wwnns be provided as well, and PowerVC does not
        do that. Only validate wwpns here.
        """
        self.validate_connector_has_setting(connector, 'wwpns')

    def get_pool(self, volume):
        """
        Override the community code to return none. This method is currently
        used in cinder manager.py to compute allocated capacity for each pool.
        PowerVC does not use that property from stats now. Though it is
        important to move powervc cinder to use the host#pool terminology soon
        """
        return None

    def query_volumes(self, context, volumes, server_info={}, mark_boot=True):
        LOG.info(_("PowerVC VNX query request: num_vols=%(num)d,...\n  "
                   "server_info=%(info)s, mark_boot=%(boot)s") %
                 dict(num=len(volumes), info=server_info, boot=mark_boot))
        uid_map, wwpn_map = (dict(), dict())
        returned_volumes, boot_map = (list(), dict())
        for server_id, server in server_info.iteritems():
            mark_boot = server.get('boot') is not True
            for volume in server.get('volumes', []):
                uid = volume.get('unique_device_id', None)
                wwpns = volume.get('wwpns', [])
                mount = volume.get('mountpoint')
                vios_wwpn = volume.get('vios_wwpns', None)
                # Build maps of the UID and WWPN depending on which is set
                if uid is not None:
                    if uid.upper() not in uid_map:
                        server_dict = {server_id: (mark_boot, mount,
                                                   vios_wwpn)}
                    else:
                        server_dict = uid_map[uid.upper()]
                        server_dict[server_id] = (mark_boot, mount,
                                                  vios_wwpn)
                    uid_map[uid.upper()] = server_dict
                for wwpn in wwpns:
                    wwpn_map[wwpn.upper()] = (server_id, mark_boot, mount)

        if uid_map:
            LOG.info(_("VNX uid_map = %s") % uid_map)
        if wwpn_map:
            LOG.info(_("VNX wwpn_map = %s") % wwpn_map)
        vn = 0
        for volume in volumes:
            volume_copy = volume.copy()
            vn = vn + 1
            LOG.info("Iter-volume #%d: %s." % (vn, volume_copy))
            server_id, mark_boot = (None, False)
            # Parse the UID out of the MetaData to use for matching Servers
            metadata = volume_copy.get('restricted_metadata', {})
            uid = metadata.get(RESTRICTED_METADATA_UNIQUE_ID)
            udid = None
            vol_pg83_naa = None
            if 'k2udid' in volume_copy.keys():
                udid = volume_copy['k2udid']
            if 'pg83NAA' in volume_copy.keys():
                vol_pg83_naa = volume_copy['pg83NAA']
            volume_copy['attach_info'] = {}
            # Parse the correct Connection Info based on the WWPN's
            volume_copy.pop('connection_info', {})
            connect_infos = volume_copy.get('itl_list', [])
            if len(connect_infos) > 0:
                # Default to the first one if the WWPN's weren't specified
                for cinfo in connect_infos:
                    wwpns = cinfo.get('vios_wwpn', [])
                    # Loop through each of the WWPN's looking for the Server
                    for wwpn in wwpns:
                        # If we found a Server for the WWPN, use this Info
                        if wwpn_map.get(wwpn.upper()) is not None:
                            server_id = wwpn_map[wwpn.upper()][0]
                            if server_id not in volume_copy['attach_info']:
                                mark_boot = wwpn_map[wwpn.upper()][1]
                                mount_point = wwpn_map[wwpn.upper()][2]
                                connection_info = None
                                # sanitize connection info for npiv
                                conn_info = dict()
                                conn_info['target_wwn'] =\
                                    copy.deepcopy(cinfo['target_wwpn'])
                                conn_info['source_wwn'] =\
                                    copy.deepcopy(cinfo['vios_wwpn'])
                                conn_info['target_lun'] =\
                                    copy.deepcopy(cinfo['target_lun'])
                                connection_info =\
                                    self._get_connection_info(
                                        uid, conn_info, CONNECTION_TYPE_NPIV)
                                self._add_attach_info(
                                    volume_copy, boot_map,
                                    connection_info, server_id,
                                    mountpoint=mount_point,
                                    mark_boot=mark_boot)

            for idx in uid_map:
                if vol_pg83_naa == idx or uid in idx:
                    server_dict = uid_map[idx]
                    for server_id in server_dict.keys():
                        if server_id not in volume_copy['attach_info']:
                            mark_boot = server_dict[server_id][0]
                            mount_point = server_dict[server_id][1]
                            udid = volume.get('k2udid', None)
                            connection_info = None
                            server_wwpn = server_dict[server_id][2]
                            # get correct connection info and itl list
                            valid_itls = self._get_vscsi_itl_info(server_wwpn,
                                                                  volume_copy)
                            if valid_itls:
                                # If given the VDisk UID from the VM, then
                                # it isn't NPIV
                                # pass the first itl to extract the target_wwn
                                cinfo = valid_itls[0]
                                cinfo['target_wwn'] = cinfo.get('target_wwpn')
                                connection_info = self._get_connection_info(
                                    vol_pg83_naa, cinfo, CONNECTION_TYPE_VSCSI,
                                    udid=udid)
                                connection_info['vscsi_target_lun'] =\
                                    valid_itls
                                connection_info['data']['vendor'] =\
                                    "EMC"
                                self._add_attach_info(volume_copy, boot_map,
                                                      connection_info,
                                                      server_id,
                                                      mountpoint=mount_point,
                                                      mark_boot=mark_boot)
            # Add the Copy of the Volume to the list to be returned
            LOG.debug(_("QUERY VOLUMES RETURNED: %s") % volume_copy)
            returned_volumes.append(volume_copy)
        LOG.info(_("QUERY VOLUMES RETURNED: %s" % len(returned_volumes)))
        return returned_volumes

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        delvols, volnames = (list(), dict())
        # Run the LUN-list command to query the list of volumes on the SVC
        out, rc = self.cli._client.command_execute(*('lun', '-list', '-state'))
        if rc != 0:
            raise EMCVnxCLICmdError(('lun', '-list'), rc, out)
        # Loop through each line adding the Name to the output list
        for lun in out.split('\n\n'):
            if len(lun.strip()) > 0:
                name = self.cli._client._get_property_value(lun, self.LUN_NAME)
                volnames[name] = ''
        # Now loop through the passed in volumes seeing which don't exist
        for volume in volumes:
            # If the Volume has a Name and isn't on the system, mark as deleted
            if volume['name'] not in volnames:
                delvols.append(volume)
        return delvols

    def update_volume_stats(self):
        """Override the Volume Stats to get the Capacity for all Pools"""
        total, free = (0.0, 0.0)
        # Call the Parent Class to retrieve most of the Statistics
        super(EMCVNXProductDriver, self).update_volume_stats()
        # Retrieve the Capacity from each of the Pools and use the total
        for pool in self._get_pool_list():
            total += float(pool['user_space'])
            free += float(pool['free_space'])
        # Now we can actually override the Free/Total Capacity values
        self._stats['total_capacity_gb'] = int(total)
        self._stats['free_capacity_gb'] = int(free)
        for pool in self._stats['pools']:
            pool['multiattach'] = True

    def create_volume(self, volume):
        """ Create a VNX volume and add PowerVC restricted metadata """
        ctxt = context.get_admin_context()
        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % volume)
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})
        try:
            s = super(EMCVNXProductDriver, self)
            vol = s.create_volume(volume)
            # Add restricted metadata to volume
            metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: volume['name'],
                        RESTRICTED_METADATA_VDISK_NAME_KEY: volume['name'],
                        RESTRICTED_METADATA_VDISK_UID_KEY: volume['name']
                        }
            powervc_db_api.\
                volume_restricted_metadata_update_or_create(ctxt,
                                                            volume['id'],
                                                            metadata)
        except Exception as e:
            errors = {'Create Failure description': e.message}
            # Update the Volume Metadata in the DB with the Error Message
            powervc_db_api.\
                volume_restricted_metadata_update_or_create(ctxt,
                                                            volume['id'],
                                                            errors)
            raise e
        if vol is None:
            vol = {}
        meta = db_api.volume_metadata_get(ctxt, volume['id'])
        meta = meta if meta else dict()
        lun_id = self.get_lun_uid(volume['name'])
        if lun_id:
            lun_id = lun_id.replace(':', '')
            meta[METADATA_VOLUME_WWN] = lun_id
        vol['metadata'] = meta
        return vol

    def get_storage_metadata(self):
        compression = False

        pools = self._get_pool_list()
        volume_pools = []

        for pool in pools:
            volume_pool = {'name': pool['name'], 'free_capacity':
                           str(int(pool['free_space'])) + " GB",
                           'capacity': str(int(pool['user_space'])) + " GB"}
            volume_pools.append(volume_pool)

        return {'volume_pools': volume_pools,
                'compression_supported': compression}

    def get_default_vol_type(self):
        """Returns the default volume type from the emc's .conf file"""
        vtn = self.configuration.default_volume_type
        return vtn.decode('utf-8') if vtn else vtn

    @pvc_fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        # Ensure that the passed-in data structure is correct.
        expected_fc_keys = set(['host', 'wwpns'])
        expected_iscsi_keys = set(['host', 'initiator'])
        if not (expected_fc_keys.issubset(connector) |
                expected_iscsi_keys.issubset(connector)):
            err_msg = (_('The connector %(connector)s does not contain the '
                         'required keys, which are %(expected_fc_keys)s.'
                         'or %(expected_iscsi_keys)s') %
                       {'connector': connector,
                        'expected_fc_keys': expected_fc_keys,
                        'expected_iscsi_keys': expected_iscsi_keys})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        if 'ip' not in connector:
            # Collect local ip address
            local_ip = \
                socket.getaddrinfo(socket.gethostname(), 0, socket.AF_INET)
            connector['ip'] = local_ip[0][4][0]

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        # for EMC VNX, auto registration is enabled. Underlying
        # driver requires 'ip' value to setup initiators in VNX.
        # for vSCSI, initiators are already setup with VIOS wwpns
        # for npiv, we need to use IP address of PoewrVC instance
        # as the ip and LPAR VM
        # name as hostname/storage group name.
        # for vSCSI, auto registration will not do anything as
        # initiators are already in.

        # Add ip to connector object
        # for NPIV, this is required
        # for vSCSI, this is required to support zero VIOS support
        # - if VIOS is not registerd in VNX it has to be auto-registered
        # 40619 - ip is now filled in nova side itself
        # For vSCSI, if an initiator is already setup with storage
        # group other than CEC MTMS_LPAR_ID, below code will retrieve
        # it and use it to associate boot or attached volume to it.
        # This will help customers who already use
        # VIOS's initiators with a custom storage group
        correct_host = ""
        if 'phy_to_virt_initiators' in connector:
            LOG.info(_("NPIV case. No need to update host"))
        else:
            connection_info = self._get_hosts_list()
            for sgroup in connection_info:
                hostwwpns = connection_info[sgroup]['host_wwpns']
                for wwpn in connector['wwpns']:
                    for hostwwpn in hostwwpns:
                        if wwpn.upper() in hostwwpn:
                            correct_host = sgroup
                            # if at-least one hit, set and quit
                            connector['host'] = correct_host
                            LOG.info(_('Host in connector=%s') % correct_host)
                            break

        LOG.info(_('Host in connector=%s') % connector['host'])

        # Fill up wwnns with same values as wwpns
        # This is needed to auto generate initiators
        if 'npiv' in connector['connection_type']:
            connector['wwnns'] = connector['wwpns']
        else:
            wwnn = []
            for entry in connector['wwpns']:
                wwnn.append("2" + entry[1:])
            connector["wwnns"] = wwnn

        # Perform the masking on the storage controller
        s = super(EMCVNXProductDriver, self)

        # For both vSCSI and NPIV, do not call super initialize_connection
        # as it does zoning. No community based zoning for both technologies.
        # Instead of that, call underlying code and retrieve storage wwpns
        # and prepare initiator_target_map

        info = self.cli.initialize_connection(volume, connector)
        if 'npiv' in connector['connection_type']:
            # In NPIV, fabric is registered in PowerVC, get storage wwpns
            # using lookoup service
            target_wwns = self.cli.get_initiator_target_map(
                connector['wwpns'],
                self.cli.get_status_up_ports(connector))[0]
            if target_wwns:
                info['data']['target_wwn'] = target_wwns
            else:
                storage_wwpns = self.cli.get_status_up_ports(connector)
                info['data']['target_wwn'] = storage_wwpns

            # prepare initiator_target_map
            storage_wwpns = info['data']['target_wwn']
            initiator_wwpns = connector['wwpns']
            init_targ_map = {}
            for initiator in initiator_wwpns:
                init_targ_map[initiator] = storage_wwpns
            info['data']['initiator_target_map'] = init_targ_map
        else:
            # No fabric is registered in vSCSI, so get storage wwpns
            # using naviseccli
            storage_wwpns = self.cli.get_status_up_ports(connector)
            initiator_wwpns = connector['wwpns']
            info['data']['target_wwn'] = storage_wwpns

        LOG.info("Exit initialize_connection "
                 " - Returning FC connection info: %(info)s."
                 % {'info': info})

        info['data']['vendor'] = "EMC"

        # get the uid to be returned as pg83_naa
        lun_id = self.get_lun_uid(volume['name'])
        if lun_id:
            lun_id = lun_id.replace(':', '')
            info['data']['pg83NAA'] = lun_id

        return info

    def get_lun_uid(self, volname):
        LUN_UID = PropertyDescriptor(
            '-uid',
            'UID:\s*(.*)\s*',
            'uid')
        data = self.cli._client.get_lun_properties(('-name', volname),
                                                   [LUN_UID])
        if 'uid' in data:
            return data['uid']
        else:
            return None

    @pvc_fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        LOG.info(_("Terminating connection with parameters %(connector)s")
                 % {'connector': connector})

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        # host in connector may not be right for vSCSI under
        # dual vios environment. Same host ECNAME_lparid is called
        # twice. We need to figure right host and update based on
        # wwpns
        correct_host = ""
        if 'phy_to_virt_initiators' in connector:
            LOG.info(_("NPIV case. no need to update host"))
        else:
            connection_info = self._get_hosts_list()
            for sgroup in connection_info:
                hostwwpns = connection_info[sgroup]['host_wwpns']
                for wwpn in connector['wwpns']:
                    for hostwwpn in hostwwpns:
                        if wwpn.upper() in hostwwpn:
                            correct_host = sgroup
                            # if atleast one hit, quit
                            break
            connector['host'] = correct_host

        # Perform the masking on the storage controller
        super(EMCVNXProductDriver, self)
        conn_info = self.cli.terminate_connection(volume, connector)

        return conn_info

    def _get_pool_list(self):
        temp_cache = []
        cmd = ('storagepool', '-list', '-availableCap', '-userCap', '-state')
        out, rc = self.cli._client.command_execute(*cmd)
        if rc != 0:
            raise EMCVnxCLICmdError(cmd, rc, out)

        try:
            for pool in out.split('\n\n'):
                if len(pool.strip()) == 0:
                    continue
                obj = {}
                obj['name'] = self.cli._client._get_property_value(
                    pool, self.cli._client.POOL_NAME)
                obj['free_space'] = self.cli._client._get_property_value(
                    pool, self.cli._client.POOL_FREE_CAPACITY)
                obj['user_space'] = self.cli._client._get_property_value(
                    pool, self.POOL_USER_CAPACITY)

                temp_cache.append(obj)
        except Exception as ex:
            LOG.error("Error happens during storage pool querying, %s"
                      % ex)
            # NOTE: Do not want to continue raise the exception
            # as the pools may temporarly unavailable
            pass
        return temp_cache

    def _get_itl_list(self, storagegroup, connection_info, lun_id,
                      storage_groups_info):
        """ Gets the ITL for the mapped volume """
        # Avoid calling get_status_up_port for each volume *
        # storage group combination. This code will optimize that
        # and call get_status_up_port per storage group entry alone.
        if storagegroup in storage_groups_info:
            target_wwpns = storage_groups_info[storagegroup]
        else:
            connector = {'host': storagegroup}
            target_wwpns = self.cli.get_status_up_ports(connector)
            storage_groups_info[storagegroup] = target_wwpns

        source_wwpns = connection_info[storagegroup]['source_wwns']
        lunmap = connection_info[storagegroup]['luns']
        target_lun = lunmap[lun_id]
        cur_conn_info = {'vios_wwpn': copy.deepcopy(list(set(source_wwpns))),
                         'target_wwpn': copy.deepcopy(list(set(target_wwpns))),
                         'target_lun': target_lun}
        return cur_conn_info

    def _get_lun_list(self, filter_set=None, vol_refs=None):
        return_volumes = []
        connection_info = self._get_hosts_list()
        storage_groups_info = dict()
        cmd = ('lun', '-list', '-state', '-userCap', '-poolName', '-uid')
        out, rc = self.cli._client.command_execute(*cmd)
        if rc != 0:
            raise EMCVnxCLICmdError(cmd, rc, out)

        try:
            for lun in out.split('\n\n'):
                if len(lun.strip()) == 0:
                    continue
                obj = {}
                obj['name'] = self.cli._client._get_property_value(
                    lun, self.LUN_NAME)
                obj['storage_pool'] = self.cli._client._get_property_value(
                    lun, self.LUN_POOL_NAME)
                obj['size'] = self.cli._client._get_property_value(
                    lun, self.LUN_CAPACITY)
                obj['state'] = self.cli._client._get_property_value(
                    lun, self.LUN_STATE)
                obj['status'] = 'available'

                uid = self.cli._client._get_property_value(lun, self.LUN_UID)
                uid = uid.replace(':', '')

                vdisk_uid = obj['name']

                obj['lun_id'] = self.cli._client._get_property_value(
                    lun, self.LUN_ID)
                obj['provider_location'] = self.cli.\
                    _build_provider_location(obj['lun_id'])
                obj['restricted_metadata'] = {
                    RESTRICTED_METADATA_VDISK_ID_KEY: obj['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: obj['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: obj['name'],
                    RESTRICTED_METADATA_BYTE_SIZE_KEY:  obj['size'],
                    RESTRICTED_METADATA_UNIQUE_ID: uid
                }
                obj['support'] = {"status": "supported"}
                if vol_refs is not None:
                    for vol_idx in vol_refs:
                        if 'pg83NAA' in vol_idx and uid == vol_idx['pg83NAA']:
                            obj['k2udid'] = vol_idx['k2udid']
                            obj['pg83NAA'] = vol_idx['pg83NAA']
                            break
                        if uid == vol_idx['uid']:
                            obj['k2udid'] = vol_idx['k2udid']
                            break

                # look for entries for this lun in connnection_info
                obj['mapped_wwpns'] = []
                obj['itl_list'] = []
                for storagegroup in connection_info:
                    lunkeys = connection_info[storagegroup]['luns'].keys()
                    if obj['lun_id'] in lunkeys:
                        conninfo_wwpns =\
                            connection_info[storagegroup]['host_wwpns']
                        for cwwpn in conninfo_wwpns:
                            if cwwpn not in obj['mapped_wwpns']:
                                obj['mapped_wwpns'].append(cwwpn)
                        # Get the ITL list for connection-info
                        if len(obj['mapped_wwpns']) > 0:
                            itls = self._get_itl_list(storagegroup,
                                                      connection_info,
                                                      obj['lun_id'],
                                                      storage_groups_info)
                            obj['itl_list'].append(itls)
                if filter_set is not None:
                    rmt_wwpns = set([wwpn for wwpn in obj['mapped_wwpns']])
                if len(obj['mapped_wwpns']) > 0:
                    obj['status'] = 'in-use'
                    obj['support'] = {
                        "status": "not_supported",
                        "reasons": [
                            'This volume is not a candidate for management '
                            'because it is already attached to a virtual '
                            'machine.  To manage this volume with PowerVC, '
                            'you must bring the virtual machine under '
                            'management.  Select to manage the virtual '
                            'machine that has the volume attached.  The '
                            'attached volume will be automatically included '
                            'for management.'
                        ]
                    }

                for k, v in connection_info.items():
                    conninfo = {}
                    hwwpns = v['host_wwpns']
                    for w in obj['mapped_wwpns']:
                        if w in hwwpns:
                            conninfo[k] = v
                            obj['connection_info'] = conninfo
                            continue
                    # To allow shareable volume
                    # if conninfo:
                    #    continue

                # If we are not filtering on WWPNs or UID's,or weARE filtering
                # and there is a match, add this disk to thelist tobe returned
                if ((filter_set is None or len(filter_set & rmt_wwpns) > 0 or
                     vdisk_uid in filter_set or uid in filter_set)):
                    return_volumes.append(obj)
            return return_volumes
        except Exception as ex:
            LOG.error("Error happens during discover volumes, %s"
                      % ex)
            pass

    def _get_hosts_list(self):
        hosts_list = dict()
        in_hlu, in_hba = (False, False)
        lunmap, wwpns = (dict(), list())
        # get a list of all storage groups
        command_get_storage_group = ('storagegroup', '-list')

        out, rc = self.cli._client.command_execute(*command_get_storage_group)
        if rc != 0:
            raise EMCVnxCLICmdError(command_get_storage_group, rc, out)

        """
        Sample output of the command:
        Storage Group Name:    SG_EMC_4Q_tb2_1_EMC_4Q_tb2_2
        Storage Group UID:     7E:C0:66:16:48:39:E4:11:B8:DD:00:60:16:36:D5:06

        HBA/SP Pairs:

        HBA UID                                          SP Name     SPPort
        -------                                          -------     ------
        20:00:00:90:FA:53:0C:4E:10:00:00:90:FA:53:0C:4E   SP B         7

        HLU/ALU Pairs:

        HLU Number     ALU Number
        ----------     ----------
        0               50
        24              8125
        27              8133
        95              8132
        Shareable:             YES
        """
        for stor in out.split('\n'):
            stor = stor.strip()
            if len(stor) <= 0 or stor.startswith('----'):
                continue
            # Parse out the Storage Group Name as the Key for the Data
            if stor.startswith('Storage Group Name:'):
                name = stor[19:].strip()
                lunmap, wwpns, s_wwpns = (dict(), list(), list())
                hosts_list[name] = dict(luns=lunmap, host_wwpns=wwpns,
                                        source_wwns=s_wwpns)
                in_hlu, in_hba = (False, False)
            # The LUN Mappings will start after the HLU Number header
            elif stor.startswith('HLU Number'):
                in_hlu, in_hba = (True, False)
            # The WWPN's will start after the HBA UID header
            elif stor.startswith('HBA UID'):
                in_hba, in_hlu = (True, False)
            # See if we have hit a case where we are past our pair list
            elif ((stor.startswith('HBA/SP Pairs') or
                   stor.startswith('HLU/ALU Pairs') or
                   stor.startswith('Shareable'))):
                in_hlu, in_hba = (False, False)
            # If we are in the list of LUN Mappings, parse them out
            elif in_hlu:
                hlu = stor.split()
                lunmap[int(hlu[1].strip())] = int(hlu[0].strip())
            # If we are in the list of WWPN's, parse them out
            elif in_hba:
                wwpnstr = stor.split()[0].strip()
                wwpns.append(wwpnstr[0:23].upper().replace(':', ''))
                wwpns.append(wwpnstr[24:].upper().replace(':', ''))
                s_wwpns.append(wwpnstr[24:].upper().replace(':', ''))
        """
        Sample output from this method:
        'SG_EMC_4Q_tb2_1_EMC_4Q_tb2_2':{'host_wwpns':[20:00:00:90:FA: cont...
        53:0C:4E:10:00:00:90:FA:53:0C:4E'],
                                         'luns': {50: 0,
                                                  8125: 24,
                                                  8132: 95,
                                                  8133: 27}},
        """
        return hosts_list

    def _find_vdisk_by_name(self, vdisk_name, volumes):
        """ Look through the list of volumes for a match.
            For VNX, the only attribute needed is the name which
            will be used for the metadata update.
        """
        for key, vol in volumes.items():
            if vol['name'] == vdisk_name:
                attributes = {}
                attributes['name'] = vdisk_name
                return attributes
        return None

    def check_volume_health(self, volume):
        """
        Throws exceptions if something is wrong with the passed-in volume.

        Checks for existence of the underlying vdisk
        """
        volumes = self._get_lun_list()

        vdisk_id = self._incorporate_restricted_metadata(volume, volumes)

        if not self._check_vdisk_exists(vdisk_id):
            raise powervc_exception.XIVVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id)

    def _check_vdisk_exists(self, vdisk_id):

        luns = self._get_lun_list()

        for lun in luns:
            if lun['name'] == vdisk_id:
                return True

        return False

    def _incorporate_restricted_metadata(self, volume_ref, volumes):
        """Query the Cinder database to retrieve any access-restricted
           metadata and attach it to the volume data.
           Returns the vdisk id, which most of the time is what we were after.
           If there is no restricted metadata, tries to find the disk and
           creates it.
        """
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume_ref['id'])

        # If the vdisk ID is not in the metadata, this could be because of an
        # upgrade from a version that didn't use it.  We try to find it by
        # looking for a volume on the SAN with a matching name
        if RESTRICTED_METADATA_VDISK_ID_KEY not in metadata:
            # Previous versions would have used this as the name on the SAN
            vdisk = self._find_vdisk_by_name(volume_ref['name'], volumes)
            if vdisk is not None:
                self._update_restricted_metadata(volume_ref['id'], vdisk)
                metadata = powervc_db_api.\
                    volume_restricted_metadata_get(ctxt, volume_ref['id'])
            else:
                return None

        volume_ref['restricted_metadata'] = metadata
        return metadata[RESTRICTED_METADATA_VDISK_ID_KEY]


def add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in range(0, len(wwpn), 2)]).lower()


def strip_colons(wwpn):
    wwpn = wwpn.lower()
    if len(wwpn) > 16:
        return wwpn.replace(":", "")
    else:
        return wwpn


###################################################
#       Override Create Volume on VNX Driver      #
###################################################
# Since the VNX Driver doesn't currently provide a way to specify a
# different pool on create_volume than what is the default specified
# when registered, we need to override the CLI to allow that pool
# to be specified until the VNX team provides support in the driver.
#
# We are also overriding the Command Execute on the CommandHelper
# to specify a response of 3 if there is a missing/invalid Certificate
# so that it fails in this case rather than auto-accepts the Certificate
class VnxProductCli(emc_vnx_cli.EMCVnxCliBase):

    def get_target_storagepool(self, volume=None, source_volume_name=None):
        """Override to get the Target Storage Pool from the Extra Specs"""
        if volume is not None:
            specs = self.get_volumetype_extraspecs(volume)
            # If the Pool Extra Spec is there, return that one
            if specs and specs.get('drivers:storage_pool'):
                return specs['drivers:storage_pool']
        return self.storage_pool


def _powervc_getEMCVnxCli(prtcl, configuration=None):
    """Override the Get VNX CLI method to return our extension"""
    configuration.append_config_values(emc_vnx_cli.loc_opts)
    return VnxProductCli(prtcl, configuration=configuration)


def _powervc_command_execute(self, *command, **kwargv):
    """Override the Command Execute to fail if missing Certificate"""
    # We want to fail if there isn't a valid certificate (option 3)
    if "process_input" not in kwargv:
        kwargv["process_input"] = '3'
    return self._orig_command_execute(*command, **kwargv)


emc_vnx_cli.getEMCVnxCli = _powervc_getEMCVnxCli
emc_vnx_cli.CommandLineHelper._orig_command_execute =\
    emc_vnx_cli.CommandLineHelper.command_execute
emc_vnx_cli.CommandLineHelper.command_execute = _powervc_command_execute
