# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013, 2015 IBM Corp.
# All Rights Reserved.
#
# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Authors:
#   Ronen Kat <ronenkat@il.ibm.com>
#   Avishay Traeger <avishay@il.ibm.com>

import paramiko

from eventlet import greenthread
import random
import re
import string
import time

from cinder import ssh_utils

from cinder import exception
from cinder import context
from cinder import db as db_api
from oslo_log import log as logging
from powervc_cinder import _
from oslo_concurrency import processutils
from oslo_utils import strutils
from oslo_utils import units
from cinder.zonemanager import utils as fczm_utils
from powervc_cinder.zonemanager import powervc_utils as pvc_fczm_utils
# don't do any zoning requests
fczm_utils.AddFCZone = pvc_fczm_utils.ignore_zoning_request
fczm_utils.RemoveFCZone = pvc_fczm_utils.ignore_zoning_request
from cinder.volume.drivers.ibm import storwize_svc as svcinit
from cinder.volume.drivers.ibm.storwize_svc import replication as storwize_rep

from oslo_concurrency.lockutils import synchronized
from cinder.volume import utils
from cinder.volume import volume_types

from collections import deque
from eventlet.event import Event

from powervc_cinder.volume import discovery_driver
from powervc_cinder.fabric.powervc_fabric_utils import Fabric
from powervc_cinder.volume.drivers.svc import storwize_helpers_ext\
    as helpers_ext
from cinder.volume.drivers.ibm.storwize_svc import ssh as storwize_ssh

from powervc_cinder.db import api as powervc_db_api
from powervc_cinder import exception as powervc_exception

from oslo_utils import excutils
from oslo_config import cfg
from cinder import rpc

import math
import six

LOG = logging.getLogger(__name__)

RESTRICTED_METADATA_VDISK_ID_KEY = "vdisk_id"
RESTRICTED_METADATA_VDISK_UID_KEY = "vdisk_uid"
RESTRICTED_METADATA_VDISK_NAME_KEY = "vdisk_name"
METADATA_VOLUME_WWN = "volume_wwn"
RESTRICTED_METADATA_BYTE_SIZE_KEY = "byte_size"

# Regex to match WWPN!C0507606D56F01F2 from lshost
RE_WWPN = re.compile('WWPN!([A-Z]*[0-9]+[A-Z0-9]*)')

# Regex to match iscsi_name!iqn
RE_ISCSI_NAME = re.compile('iscsi_name!(.)')

# constant for flashcopy mapping processing
FCMAP_COPY_COMPLETE_PCT = 100

storwize_svc_opts = [
    cfg.IntOpt('storwize_fcmap_poll_interval',
               default=600,
               help='A tunable determines how frequent Storwize driver should '
                    'poll the status of in-flight flashcopy monitor queue. '
                    'The poll interval is in seconds'),
    cfg.StrOpt('storwize_fcmap_delete_copy_clean_rate',
               default='0',
               help='A tunable to decide copy/clean rate used during'
                    ' a delete operation involving a fcmap.')
]

CONF = cfg.CONF

CONF.register_opts(storwize_svc_opts)

ssh_san_opts = [
    cfg.BoolOpt('development_ssh_san_conn',
                default=False,
                help='Controls whether or not the SSH Connections are dev or '
                     'production'),
    cfg.IntOpt('san_retry_min_delay', default=2,
               help="The minimum number of seconds to wait before retrying a"
                    "command"),
    cfg.IntOpt('san_retry_max_delay', default=10,
               help="The maximum number of seconds to wait before retrying a"
                    "command"),
]

CONF.register_opts(ssh_san_opts)


class PowerVCSSHPool(ssh_utils.SSHPool):

    def __init__(self, *args, **kwargs):
        return super(PowerVCSSHPool, self).__init__(*args, **kwargs)

    """
    Override pools.Pool's put method, which is called after
    each iteration of "with self.sshpool.item()" in
    _run_ssh. If development_ssh_san_conn flag is True,
    the cinder.utils remove() method will be invoked
    to close the ssh connection after each time a
    connection is get().
    """
    def put(self, item):
        # do max_size check because SSHPool isn't calling remove in that case
        # like it should
        if CONF.development_ssh_san_conn or self.current_size > self.max_size:
            self.current_size -= 1
            self.remove(item)
        else:
            super(PowerVCSSHPool, self).put(item)


class StorwizeSVCProductDriver(
        discovery_driver.VolumeDiscoveryDriver,
        svcinit.StorwizeSVCDriver):
    def __init__(self, *args, **kwargs):
        super(StorwizeSVCProductDriver, self).__init__(*args, **kwargs)

        # have helpers extension to override community code
        self._helpers = helpers_ext.PowerVCStorwizeHelpers(self._run_ssh)
        # add in ssh extensions
        self.ssh = storwize_ssh.StorwizeSSH(self._run_ssh)

        # create deque to hold all the vdisk delete requests that has
        # pending flashcopy mapping.
        self.fcmap = deque()
        self.fcmap_timer = []

        self._extent_size = None

        # This describes the storage controller, handy for attaching to
        # exceptions or mentioning in log messages.
        self.endpoint_desc = powervc_exception.SVCDescriptor(
            self.configuration.host,
            self.configuration.host_display_name,
            self.configuration.san_login,
            self.configuration.san_ip,
            self.configuration.san_ssh_port)

        # override the polling interval with something reasonable.
        self._helpers.check_fcmapping_interval = \
            self.configuration.storwize_fcmap_poll_interval

        # Build cleanup translation tables for host names
        invalid_ch_in_host = ''
        for num in range(0, 128):
            ch = str(chr(num))
            if (not ch.isalnum() and ch != ' ' and ch != '.' and
                    ch != '-' and ch != '_'):
                invalid_ch_in_host = invalid_ch_in_host + ch
        self._string_host_name_filter = string.maketrans(
            invalid_ch_in_host, '-' * len(invalid_ch_in_host))
        self._unicode_host_name_filter = dict((ord(unicode(char)), u'-')
                                              for char in invalid_ch_in_host)

    def check_for_setup_error(self):
        if self._extent_size is None:
            exception_msg = (_('Unable to determine pool extent size'))
            raise exception.VolumeBackendAPIException(data=exception_msg)

        super(StorwizeSVCProductDriver, self).check_for_setup_error()

    """
    def check_for_setup_error(self):
        #Ensure that the flags are set properly.
        LOG.debug('enter: check_for_setup_error')

        # Check that we have the system ID information
        if self._state['system_name'] is None:
            exception_msg = (_('Unable to determine system name'))
            raise exception.VolumeBackendAPIException(data=exception_msg)
        if self._state['system_id'] is None:
            exception_msg = (_('Unable to determine system id'))
            raise exception.VolumeBackendAPIException(data=exception_msg)
        if self._extent_size is None:
            exception_msg = (_('Unable to determine pool extent size'))
            raise exception.VolumeBackendAPIException(data=exception_msg)

        required_flags = ['san_ip', 'san_ssh_port', 'san_login',
                          'storwize_svc_volpool_name']
        for flag in required_flags:
            if not self.configuration.safe_get(flag):
                raise exception.InvalidInput(reason=_('%s is not set') % flag)

        # Ensure that either password or keyfile were set
        if not (self.configuration.san_password or
                self.configuration.san_private_key):
            raise exception.InvalidInput(
                reason=_('Password or SSH private key is required for '
                         'authentication: set either san_password or '
                         'san_private_key option'))

        # Check that flashcopy_timeout is not more than 10 minutes
        flashcopy_timeout = self.configuration.storwize_svc_flashcopy_timeout
        if not (flashcopy_timeout > 0 and flashcopy_timeout <= 3600):
            raise exception.InvalidInput(
                reason=_('Illegal value %d specified for '
                         'storwize_svc_flashcopy_timeout: '
                         'valid values are between 0 and 3600')
                % flashcopy_timeout)

        opts = self._helpers.build_default_opts(self.configuration)
        self._helpers.check_vdisk_opts(self._state, opts)

        LOG.debug('leave: check_for_setup_error')
    """
    def _get_iscsi_ip_addrs(self):
        generator = self._port_conf_generator(['svcinfo', 'lsportip'])
        header = next(generator, None)
        if not header:
            return

        for port_data in generator:
            try:
                port_node_id = port_data['node_id']
                port_ipv4 = port_data['IP_address']
                port_ipv6 = port_data['IP_address_6']
                state = port_data['state']
            except KeyError:
                self._handle_keyerror('lsportip', header)

            if port_node_id in self._state['storage_nodes'] and (
                    state == 'configured' or state == 'online'):
                node = self._state['storage_nodes'][port_node_id]
                if len(port_ipv4):
                    node['ipv4'].append(port_ipv4)
                if len(port_ipv6):
                    node['ipv6'].append(port_ipv6)

    def _get_fc_wwpns(self):
        for key in self._state['storage_nodes']:
            node = self._state['storage_nodes'][key]
            ssh_cmd = ['svcinfo', 'lsnode', '-delim', '!', node['id']]
            raw = self._run_ssh(ssh_cmd)
            resp = CLIResponse(raw, delim='!')
            wwpns = set(node['WWPN'])
            for i, s in resp.select('port_id', 'port_status'):
                if 'active' == s:
                    wwpns.add(i)
            node['WWPN'] = list(wwpns)
            LOG.info(_('WWPN on node %(node)s: %(wwpn)s')
                     % {'node': node['id'], 'wwpn': node['WWPN']})

    def _incorporate_restricted_metadata(self, volume_ref):
        """Query the Cinder database to retrieve any access-restricted
           metadata and attach it to the volume data.
           Returns the vdisk id, which most of the time is what we were after.
           If there is no restricted metadata, tries to find the disk and
           creates it.
        """
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume_ref['id'])

        # If the vdisk ID is not in the metadata, this could be because of an
        # upgrade from a version that didn't use it.  We try to find it by
        # looking for a volume on the SAN with a matching name
        if RESTRICTED_METADATA_VDISK_ID_KEY not in metadata:
            # Previous versions would have used this as the name on the SAN
            vdisk = self._find_vdisk_by_name(volume_ref['name'])
            if vdisk is not None:
                self._update_restricted_metadata(volume_ref['id'], vdisk)
                metadata = powervc_db_api.\
                    volume_restricted_metadata_get(ctxt, volume_ref['id'])
            else:
                return None

        volume_ref['restricted_metadata'] = metadata
        return metadata[RESTRICTED_METADATA_VDISK_ID_KEY]

    def _check_vdisk_exists(self, vdisk_id):
        ssh_cmd = ['svcinfo', 'lsvdisk', vdisk_id]

        out, err = self._run_ssh(ssh_cmd, check_exit_code=False)
        return err == ""

    def _find_vdisk_by_name(self, vdisk_name):
        ssh_cmd = ['svcinfo', 'lsvdisk', '-delim', '!', vdisk_name]

        out, err = self._run_ssh(ssh_cmd, check_exit_code=False)
        if "CMMVC5754E" in err:
            # Can't find the volume
            return None

        attributes = {}
        for attrib_line in out.split('\n'):
            key, foo, value = attrib_line.partition('!')
            if key is not None and len(key.strip()):
                attributes[key] = value

        return attributes

    def _update_restricted_metadata(self, volume_id, vdisk):
        """Perform a database update to associate the specified information
           with the specified volume.
        """

        required_keys = set(['id', 'name', 'vdisk_UID'])
        has_required_keys = required_keys.issubset(vdisk.keys())
        msg = _("Required keys %(required)s not found in vdisk dictionary "
                "%(dictionary)r") \
            % {'required': ', '.join(required_keys),
               'dictionary': vdisk}

        self._driver_assert(has_required_keys, msg)

        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: vdisk['id'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: vdisk['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: vdisk['vdisk_UID'],
                    METADATA_VOLUME_WWN: vdisk['vdisk_UID']
                    }
        ctxt = context.get_admin_context()
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, volume_id,
                                                        metadata)

    def do_setup(self, ctxt):

        super(StorwizeSVCProductDriver, self).do_setup(ctxt)

        # Set extend size with attribute value
        pool = self.configuration.storwize_svc_volpool_name
        ssh_cmd = ['svcinfo', 'lsmdiskgrp', '-bytes', '-delim', '!',
                   '"%s"' % pool]
        attributes = self._execute_command_and_parse_attributes(ssh_cmd)
        if not attributes:
            msg = (_('do_setup: Pool %s does not exist') % pool)
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)
        self._extent_size = attributes['extent_size']

    def get_compression_enabled(self):
        """Helper Method to get compression enablement"""
        return self._state['compression_enabled']

    def _build_default_opts(self):
        # Ignore capitalization
        protocol = self.configuration.storwize_svc_connection_protocol
        if protocol.lower() == 'fc':
            protocol = 'FC'
        elif protocol.lower() == 'iscsi':
            protocol = 'iSCSI'

        opt = {'rsize': self.configuration.storwize_svc_vol_rsize,
               'warning': self.configuration.storwize_svc_vol_warning,
               'autoexpand': self.configuration.storwize_svc_vol_autoexpand,
               'grainsize': self.configuration.storwize_svc_vol_grainsize,
               'compression': self.configuration.storwize_svc_vol_compression,
               'easytier': self.configuration.storwize_svc_vol_easytier,
               'protocol': protocol,
               'multipath': self.configuration.storwize_svc_multipath_enabled,
               'iogrp': self.configuration.storwize_svc_vol_iogrp,
               'qos': None,
               'storage_pool': self.configuration.storwize_svc_volpool_name,
               'mirror_pool': None,
               'replication': 0}
        return opt

    def ensure_export(self, ctxt, volume):
        """Check that the volume exists on the storage.

        The system does not "export" volumes as a Linux iSCSI target does,
        and therefore we just check that the volume exists on the storage.
        """
        vdisk_id = self._incorporate_restricted_metadata(volume)
        if vdisk_id is None:
            LOG.error(_('ensure_export: Volume %(volume_id)s does not '
                        'reference a vdisk')
                      % {'volume_id': volume['id']})
            return

        volume_defined = self._is_vdisk_defined(vdisk_id)
        if not volume_defined:
            LOG.error(_('ensure_export: Volume %(volume_id)s references '
                        'vdisk %(vdisk)s, which was not found on storage')
                      % {'volume_id': volume['id'],
                         'vdisk': vdisk_id})

    def create_export(self, ctxt, volume, connector):
        model_update = None
        return model_update

    def remove_export(self, ctxt, volume):
        pass

    def _add_chapsecret_to_host(self, host_name):
        """Generate and store a randomly-generated CHAP secret for the host."""

        chap_secret = utils.generate_password()
        ssh_cmd = ['svctask', 'chhost', '-chapsecret', chap_secret, host_name]
        ssh_cmd_out = ['svctask', 'chhost', '-chapsecret',
                       '************', host_name]

        out, err = self._run_ssh(ssh_cmd)
        # No output should be returned from chhost
        self._assert_ssh_return(len(out.strip()) == 0,
                                '_add_chapsecret_to_host',
                                ssh_cmd_out, out, err)
        return chap_secret

    def _get_chap_secret_for_host(self, host_name):
        """Return the CHAP secret for the given host."""

        LOG.debug('enter: _get_chap_secret_for_host: host name %s'
                  % host_name)

        ssh_cmd = ['svcinfo', 'lsiscsiauth', '-delim', '!']
        out, err = self._run_ssh(ssh_cmd)

        if not len(out.strip()):
            return None

        host_lines = out.strip().split('\n')
        # strip out the chap password from stdout
        STDOUT = re.sub(r'chap!(.*)$', "chap!****************", out)
        self._assert_ssh_return(len(host_lines), '_get_chap_secret_for_host',
                                ssh_cmd, STDOUT, err)

        header = host_lines.pop(0).split('!')
        self._assert_ssh_return('name' in header, '_get_chap_secret_for_host',
                                ssh_cmd, STDOUT, err)
        self._assert_ssh_return('iscsi_auth_method' in header,
                                '_get_chap_secret_for_host', ssh_cmd,
                                STDOUT, err)
        self._assert_ssh_return('iscsi_chap_secret' in header,
                                '_get_chap_secret_for_host', ssh_cmd,
                                STDOUT, err)
        name_index = header.index('name')
        method_index = header.index('iscsi_auth_method')
        secret_index = header.index('iscsi_chap_secret')

        chap_secret = None
        host_found = False
        for line in host_lines:
            info = line.split('!')
            if info[name_index] == host_name:
                host_found = True
                if info[method_index] == 'chap':
                    chap_secret = info[secret_index]

        self._assert_ssh_return(host_found, '_get_chap_secret_for_host',
                                ssh_cmd, STDOUT, err)

        LOG.debug('leave: _get_chap_secret_for_host: host name '
                  '%(host_name)s with secret %(chap_secret)s'
                  % {'host_name': host_name, 'chap_secret': "*************"})

        return chap_secret

    def _connector_to_hostname_prefix(self, connector):
        """Translate connector info to storage system host name.

        Translate a host's name and IP to the prefix of its hostname on the
        storage subsystem.  We create a host name host name from the host and
        IP address, replacing any invalid characters (at most 55 characters),
        and adding a random 8-character suffix to avoid collisions. The total
        length should be at most 63 characters.

        """

        host_name = connector['host']
        if isinstance(host_name, unicode):
            host_name = host_name.translate(self._unicode_host_name_filter)
        elif isinstance(host_name, str):
            host_name = host_name.translate(self._string_host_name_filter)
        else:
            msg = _('_create_host: Cannot clean host name. Host name '
                    'is not unicode or string')
            LOG.error(msg)
            raise exception.NoValidHost(reason=msg)

        host_name = str(host_name)

        if not re.match('^[A-Za-z]', host_name):
            host_name = "_" + host_name

        return host_name[:55]

    def _find_host_from_attached_volume(self, connector, vdisk_id):
        """ Another fastpath to find the existing hostmap that
        hasn't been logged in to SAN. So it will not show by
        lsfabric. But the disk may have been mapped.
        """
        ssh_cmd = ['lsvdiskhostmap', '-delim', '!', vdisk_id]

        if (not connector or (not connector.get('wwpns') and
                              not connector.get('initiator'))):
            return None
        try:
            out, err = self._run_ssh(ssh_cmd, attempts=2)
        except processutils.ProcessExecutionError as e:
            if 'CMMVC5753E' in e.stderr:
                # CMMVC5753E: The specified object does not exist or is not a
                #             suitable candidate.
                return None
            else:
                # something bad happened
                raise
        if not len(out.strip()):
            return None
        if 'wwpns' in connector:
            # The connector wwpns passed in is unicode. Fix it for
            # the set compare.
            conn_wwpns = [x.encode('ascii', 'ignore').upper()
                          for x in connector.get('wwpns')]

        host_lines = out.strip().split('\n')
        header = host_lines.pop(0).split('!')
        self._assert_ssh_return('host_id' in header and
                                'host_name' in header,
                                '_find_host_from_attached_volume',
                                ssh_cmd, out, err)
        host_id_idx = header.index('host_id')
        host_name_idx = header.index('host_name')

        hostname = None
        for line in host_lines:
            host_id = line.split('!')[host_id_idx]
            host_name = line.split('!')[host_name_idx]
            ssh_cmd = ['lshost', '-delim', '!', host_id]
            try:
                out, err = self._run_ssh(ssh_cmd, attempts=2)
            except processutils.ProcessExecutionError:
                continue
            if not len(out.strip()):
                continue

            if 'wwpns' in connector:
                # find all the WWPNs from the lshost output. Expect all
                # the WWPNs in the lshost output are in upper case.
                wwpns = RE_WWPN.findall(out)

                if (set(conn_wwpns) == set(wwpns) or
                        set(conn_wwpns).issubset(set(wwpns))):
                    hostname = host_name
                    break
            else:
                if connector['initiator'] in set(RE_ISCSI_NAME.findall(out)):
                    hostname = host_name
                    break

        return hostname

    def _find_host_from_wwpn(self, connector):
        # SVC uses upper-case WWPNs
        for wwpn in [x.upper() for x in connector['wwpns']]:
            ssh_cmd = ['svcinfo', 'lsfabric', '-wwpn', wwpn, '-delim', '!']
            out, err = self._run_ssh(ssh_cmd)

            if not len(out.strip()):
                # This WWPN is not in use
                continue

            host_lines = out.strip().split('\n')
            header = host_lines.pop(0).split('!')
            self._assert_ssh_return('remote_wwpn' in header and
                                    'name' in header,
                                    '_find_host_from_wwpn',
                                    ssh_cmd, out, err)
            rmt_wwpn_idx = header.index('remote_wwpn')
            name_idx = header.index('name')

            wwpns = map(lambda x: x.split('!')[rmt_wwpn_idx], host_lines)

            if wwpn in wwpns:
                # All the wwpns will be the mapping for the same
                # host from this WWPN-based query. Just pick
                # the name from first line.
                hostname = host_lines[0].split('!')[name_idx]
                return hostname

        # Didn't find a host
        return None

    def _find_host_exhaustive(self, connector, hosts):
        for host in hosts:
            # The hosts list may go stale while the list is processed.
            # Handle it accordingly.
            stringify_host = "'" + host + "'"
            ssh_cmd = ['svcinfo', 'lshost', '-delim', '!', stringify_host]
            try:
                out, err = self._run_ssh(ssh_cmd)
            except processutils.ProcessExecutionError as e:
                if 'CMMVC5754E' in e.stderr:
                    # CMMVC5754E: The specified object does not exist
                    #             The host has been deleted while walking
                    #             the list.
                    continue
                else:
                    # something bad happened.
                    raise e

            self._assert_ssh_return(len(out.strip()),
                                    '_find_host_exhaustive',
                                    ssh_cmd, out, err)
            for attr_line in out.split('\n'):
                # If '!' not found, return the string and two empty strings
                attr_name, foo, attr_val = attr_line.partition('!')
                if (attr_name == 'iscsi_name' and
                        'initiator' in connector and
                        attr_val == connector['initiator']):
                    return host
                elif (attr_name == 'WWPN' and
                      'wwpns' in connector and
                      attr_val.lower() in
                      map(str.lower, map(str, connector['wwpns']))):
                    return host
        return None

    def _get_host_from_connector(self, connector, vdisk_id, iogrp,
                                 is_initconn=True):
        """List the hosts defined in the storage.

        Return the host name with the given connection info, or None if there
        is no host fitting that information.

        """
        # If the backend_host group is passed in the connector use the as the
        # host name.
        backend_host_group = connector.get('backend_host_group')
        LOG.debug('enter: _get_host_from_connector: backend_host_group %s' %
                  backend_host_group)
        if connector.get('connection_type') and\
           'pv_vscsi' in connector['connection_type'] and\
           backend_host_group:
            # check if the backend_host_group is valid
            self._validate_host_group(backend_host_group)
            return backend_host_group

        prefix = self._connector_to_hostname_prefix(connector)
        LOG.debug('prefix %s' % prefix)

        # GMN - this block of code fast-paths the common NPIV case where
        # a host definition does not exist, and the WWPNs are not yet logged
        # into the fabric.  Without this case, lsfabric can't help us to find
        # a host, so we end up falling back to the slow exhaustive method,
        # of doing lshost per host, which is very slow if you have a lot of
        # hosts (which you do for NPIV environments).
        #
        # What we do here is just jump in and try to create a new host
        # with all the desired WWPNs.  It will fail if the host already
        # exists or if some of the port specifiers are already present in
        # another host
        hostname = None
        if is_initconn:
            # For initialize_connection, the host may not be defined
            # yet. Try it first.
            try:
                if (connector.get('connection_type') and
                        'npiv' in connector.get('connection_type')):
                    return self._create_host(connector, iogrp, False)
            except powervc_exception.SVCHostDefsException:
                # Tried to create a host definition, but couldn't because we
                # have reached the limit.  Fail as there's nothing we can do
                # here.
                with excutils.save_and_reraise_exception():
                    LOG.info("The maximum number of hosts for "
                             "one or more IO groups is already configured. "
                             "No more host definitions can be created.")
            except Exception as ex:
                # Anything else and we carry on, looking further for an
                # existing host definition.
                LOG.info(_("Continue to fall back processing after initial "
                           "failure to make a host on the SAN controller: "
                           "%s") % ex)
                pass
        else:
            # gfm - this is only valid for terminate_connection.
            # ajiang - Another fastpath host lookup. If the _create_host failed
            # we know the wwpns have been defined in some hosts on SVC.
            # try to do fast lookup in two ways:
            # 1. check whether the volume has been mapped to the connector.
            #    If so, return the matching host
            LOG.debug("Trying to lookup SVC host from the vdisk ID.")
            hostname = self._find_host_from_attached_volume(connector,
                                                            vdisk_id)

        # 2. If #1 doesn't find the host, look for lsfabric to see
        #    whether the wwpn has been logged in. If so, find the matching
        #    host.
        # If neither #1 or #2 finds any matching host, we have to go
        # through the host definition one by one which will be the slow path.

        if not hostname:
            if 'wwpns' in connector:
                # If we have FC information, we have a faster lookup option
                hostname = self._find_host_from_wwpn(connector)

        # If we don't have a hostname yet, try the long way
        if not hostname:
            LOG.debug("Trying to lookup up the host the long way...")
            # Get list of host in the storage
            ssh_cmd = ['svcinfo', 'lshost', '-delim', '!']
            out, err = self._run_ssh(ssh_cmd)

            if not len(out.strip()):
                return None

            host_lines = out.strip().split('\n')
            self._assert_ssh_return(len(host_lines),
                                    '_get_host_from_connector',
                                    ssh_cmd, out, err)
            header = host_lines.pop(0).split('!')
            self._assert_ssh_return('name' in header,
                                    '_get_host_from_connector',
                                    ssh_cmd, out, err)
            name_index = header.index('name')
            hosts = map(lambda x: x.split('!')[name_index], host_lines)
            hostname = self._find_host_exhaustive(connector, hosts)

        LOG.debug('leave: _get_host_from_connector: host %s' % hostname)

        return hostname

    def _create_host(self, connector, iogrp, check_exit_code=True):
        """Create a new host on the storage system.

        We create a host name and associate it with the given connection
        information.

        """

        LOG.debug('enter: _create_host: host %s' % connector['host'])

        rand_id = str(random.randint(0, 99999999)).zfill(8)
        host_name = '%s-%s' % (self._connector_to_hostname_prefix(connector),
                               rand_id)

        # Get all port information from the connector
        ports = []
        if 'initiator' in connector:
            ports.append('-iscsiname %s' % connector['initiator'])
        if 'wwpns' in connector:
            for wwpn in connector['wwpns']:
                ports.append('-hbawwpn %s' % wwpn)

        # When creating a host, we need one port
        self._driver_assert(len(ports), _('_create_host: No connector ports'))
        port1 = ports.pop(0)
        arg_name, arg_val = port1.split()
        multipath = connector.get('multipath', False)
        if iogrp is None or multipath:
            ssh_cmd = ['svctask', 'mkhost', '-force', arg_name, arg_val,
                       '-name', '"%s"' % host_name]
        else:
            ssh_cmd = ['svctask', 'mkhost', '-force', arg_name, arg_val,
                       '-iogrp', iogrp, '-name', '"%s"' % host_name]
        out, err = self._run_ssh(ssh_cmd, check_exit_code=check_exit_code)
        if check_exit_code:
            self._assert_ssh_return('successfully created' in out,
                                    '_create_host', ssh_cmd, out, err)
        else:
            # We just want an exception if we didn't create the host
            if ('CMMVC6220E' or 'CMMVC6221E' or 'CMMVC6222E') in err:
                raise powervc_exception.SVCHostDefsException(
                    self.endpoint_desc)

            if 'successfully created' not in out:
                raise exception.VolumeBackendAPIException(data=err)

        # Add any additional ports to the host
        try:
            for port in ports:
                arg_name, arg_val = port.split()
                ssh_cmd = ['svctask', 'addhostport', '-force',
                           arg_name, arg_val, host_name]
                out, err = self._run_ssh(ssh_cmd,
                                         check_exit_code=check_exit_code)
                if err:
                    raise exception.VolumeBackendAPIException(data=err)
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                ssh_cmd = ['svctask', 'rmhost', host_name]
                self._run_ssh(ssh_cmd, check_exit_code=check_exit_code)

        LOG.debug('leave: _create_host: host %(host)s - %(host_name)s' %
                  {'host': connector['host'], 'host_name': host_name})
        return host_name

    def _validate_host_group(self, backend_host_grp):
        ssh_cmd = ['svcinfo', 'lshost', '-delim', '!', backend_host_grp]
        out, err = self._run_ssh(ssh_cmd)
        if err:
            raise powervc_exception.SVCInvalidHostGroupName(
                self.endpoint_desc,
                name=backend_host_grp)

    def _get_hostvdisk_mappings(self, host_name):
        """Return the defined storage mappings for a host, as vdisk IDs."""

        return_data = {}
        stringify_host = "'" + host_name + "'"
        ssh_cmd = ['svcinfo', 'lshostvdiskmap', '-delim', '!', stringify_host]
        out, err = self._run_ssh(ssh_cmd)

        mappings = out.strip().split('\n')
        if len(mappings):
            header = mappings.pop(0)
            for mapping_line in mappings:
                mapping_data = self._get_hdr_dic(header, mapping_line, '!')
                return_data[mapping_data['vdisk_id']] = mapping_data

        return return_data

    def _delete_host(self, host_name):
        """Delete a host on the storage system."""

        LOG.debug('enter: _delete_host: host %s ' % host_name)

        stringify_host = "'" + host_name + "'"
        ssh_cmd = ['svctask', 'rmhost', stringify_host]
        out, err = self._run_ssh(ssh_cmd)
        # No output should be returned from rmhost
        self._assert_ssh_return(len(out.strip()) == 0,
                                '_delete_host', ssh_cmd, out, err)

        LOG.debug('leave: _delete_host: host %s ' % host_name)

    def _create_vdisk(self, names, size, units, opts):
        """Create a new vdisk, returns vdisk ID
        names contains a list of names to use in order of preference.  We try
        them all until one works, as a simple way of handling names that are
        invalid or already in use on the storage controller.
        """

        names_string = ', '.join(names)
        LOG.debug(_('enter: _create_vdisk: candidate names %s ') %
                  names_string)

        easytier = 'on' if opts['easytier'] else 'off'

        ###############################################################
        # This section is the delta from community driver which only
        # use storwize_svc_volpool_name config option to choose
        # storage pool. PowerVC supports dynamic pool selection
        # from extra_spects. Remove this _create_vdisk if we
        # contribute this back.

        # Set the default Storage Pool from the configuration file.
        mdiskgrp = self.configuration.storwize_svc_volpool_name

        # Check for override values specified in the extra-specs
        for opt, val in opts.iteritems():
            # Override the Storage Pool if specified.
            if opt == 'storage_pool':
                mdiskgrp = val
        ###############################################################

        # Set space-efficient options
        if opts['rsize'] == -1:
            ssh_cmd_se_opt = []
        else:
            ssh_cmd_se_opt = ['-rsize', '%s%%' % str(opts['rsize']),
                              '-autoexpand', '-warning',
                              '%s%%' % str(opts['warning'])]
            if not opts['autoexpand']:
                ssh_cmd_se_opt.remove('-autoexpand')

            if opts['compression']:
                ssh_cmd_se_opt.append('-compressed')
            else:
                ssh_cmd_se_opt.extend(['-grainsize', str(opts['grainsize'])])

        # Iterate over all names.  We break out of the loop as soon as we find
        # a name that works.
        for name in names:

            # If an exception is raised here, we remember it in case we need to
            # raise it later.
            last_exception = None

            # We now pull the iogrp out of the storage template.
            # This makes that iogrp the perferred iogrp
            ssh_cmd = ['svctask', 'mkvdisk', '-name', name,
                       '-mdiskgrp', '"%s"' % mdiskgrp,
                       '-iogrp', str(opts['iogrp']),
                       '-size', size, '-unit', units,
                       '-easytier', easytier] + ssh_cmd_se_opt

            # We do our own error checking because we want to be able to
            # silently retry if we fail.
            out, err = self._run_ssh(ssh_cmd, check_exit_code=False)

            if err:
                cmd = ' '.join(ssh_cmd)

                if ("CMMVC6035E" in err or
                        "CMMVC6527E" in err or
                        "CMMVC5738E" in err):
                    # CMMVC6035E: Name already in use.
                    # CMMVC6527E: Invalid name
                    # CMMVC5738E: Name too long
                    # If we failed with one of the above errors, try again with
                    # a different name.  However, remember this exception, in
                    # case we have run out of names and really do have to raise
                    # this exception.
                    e = processutils.ProcessExecutionError(exit_code=1,
                                                           stdout=out,
                                                           stderr=err,
                                                           cmd=cmd)
                    last_exception = e
                    continue
                else:
                    # Some other error occurred - raise the exception
                    # immediately.
                    raise processutils.ProcessExecutionError(exit_code=1,
                                                             stdout=out,
                                                             stderr=err,
                                                             cmd=cmd)

            self._assert_ssh_return(len(out.strip()), '_create_vdisk',
                                    ssh_cmd, out, err)

            # If we get here then we don't need to try any more names
            break

        # If we came out of the loop with last_exception set, then this is
        # because we ran out of names that we could try, and we really should
        # raise the exception that we remembered.
        if last_exception:
            raise last_exception

        # Ensure that the output is as expected
        match_obj = re.search('Virtual Disk, id \[([0-9]+)\], '
                              'successfully created', out)
        # Make sure we got a "successfully created" message with vdisk id
        self._driver_assert(
            match_obj is not None,
            _('_create_vdisk %(name)s - did not find '
              'success message in CLI output.\n '
              'stdout: %(out)s\n stderr: %(err)s')
            % {'name': name, 'out': str(out), 'err': str(err)})

        vdisk_id = match_obj.group(1)

        LOG.debug(_('leave: _create_vdisk: volume %(name)s ID %(id)s ') %
                  {'name': name, 'id': vdisk_id})

        return vdisk_id, name

    def _make_fc_map(self, source_vdisk_id, target_vdisk_id, full_copy):
        fc_map_cli_cmd = ['svctask', 'mkfcmap', '-source', source_vdisk_id,
                          '-target', target_vdisk_id, '-autodelete']
        if not full_copy:
            fc_map_cli_cmd.extend(['-copyrate', '0'])
        try:
            out, err = self._run_ssh(fc_map_cli_cmd)
        except processutils.ProcessExecutionError as e:
            if "CMMVC6425E" in e.stderr:
                # Maximum flashcopy limit reached, can't create more.
                # Raise an SVCException so we can present it nicely to the user
                raise powervc_exception.SVCFCMapsException(self.endpoint_desc)
            else:
                raise

        self._driver_assert(
            len(out.strip()),
            _('create FC mapping from %(source)s to %(target)s - '
              'did not find success message in CLI output.\n'
              ' stdout: %(out)s\n stderr: %(err)s\n')
            % {'source': source_vdisk_id,
               'target': target_vdisk_id,
               'out': str(out),
               'err': str(err)})

        # Ensure that the output is as expected
        match_obj = re.search('FlashCopy Mapping, id \[([0-9]+)\], '
                              'successfully created', out)
        # Make sure we got a "successfully created" message with vdisk id
        self._driver_assert(
            match_obj is not None,
            _('create FC mapping from %(source)s to %(target)s - '
              'did not find success message in CLI output.\n'
              ' stdout: %(out)s\n stderr: %(err)s\n')
            % {'source': source_vdisk_id,
               'target': target_vdisk_id,
               'out': str(out),
               'err': str(err)})

        try:
            fc_map_id = match_obj.group(1)
            self._driver_assert(
                fc_map_id is not None,
                _('create FC mapping from %(source)s to %(target)s - '
                  'did not find mapping id in CLI output.\n'
                  ' stdout: %(out)s\n stderr: %(err)s\n')
                % {'source': source_vdisk_id,
                   'target': target_vdisk_id,
                   'out': str(out),
                   'err': str(err)})
        except IndexError:
            self._driver_assert(
                False,
                _('create FC mapping from %(source)s to %(target)s - '
                  'did not find mapping id in CLI output.\n'
                  ' stdout: %(out)s\n stderr: %(err)s\n')
                % {'source': source_vdisk_id,
                   'target': target_vdisk_id,
                   'out': str(out),
                   'err': str(err)})
        return fc_map_id

    def _call_prepare_fc_map(self, fc_map_id, source_vdisk_id,
                             target_vdisk_id):
        try:
            out, err = self._run_ssh(['svctask', 'prestartfcmap', fc_map_id])
        except processutils.ProcessExecutionError as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_('_prepare_fc_map: Failed to prepare FlashCopy '
                            'from %(source)s to %(target)s.\n'
                            'stdout: %(out)s\n stderr: %(err)s')
                          % {'source': source_vdisk_id,
                             'target': target_vdisk_id,
                             'out': e.stdout,
                             'err': e.stderr})

    def _prepare_fc_map(self, fc_map_id, source_vdisk_id, target_vdisk_id):
        self._call_prepare_fc_map(fc_map_id, source_vdisk_id, target_vdisk_id)
        mapping_ready = False
        wait_time = 5
        # Allow waiting of up to timeout (set as parameter)
        timeout = self.configuration.storwize_svc_flashcopy_timeout
        max_retries = (timeout / wait_time) + 1
        for try_number in range(1, max_retries):
            mapping_attrs = self._get_flashcopy_mapping_attributes(fc_map_id)
            if (mapping_attrs is None or
                    'status' not in mapping_attrs):
                break
            if mapping_attrs['status'] == 'prepared':
                mapping_ready = True
                break
            elif mapping_attrs['status'] == 'stopped':
                self._call_prepare_fc_map(fc_map_id, source_vdisk_id,
                                          target_vdisk_id)
            elif mapping_attrs['status'] != 'preparing':
                # Unexpected mapping status
                exception_msg = (_('Unexecpted mapping status %(status)s '
                                   'for mapping %(id)s. Attributes: '
                                   '%(attr)s')
                                 % {'status': mapping_attrs['status'],
                                    'id': fc_map_id,
                                    'attr': mapping_attrs})
                LOG.error(exception_msg)
                raise exception.VolumeBackendAPIException(data=exception_msg)
            # Need to wait for mapping to be prepared, wait a few seconds
            time.sleep(wait_time)

        if not mapping_ready:
            exception_msg = (_('Mapping %(id)s prepare failed to complete '
                               'within the allotted %(to)d seconds timeout. '
                               'Terminating.')
                             % {'id': fc_map_id,
                                'to': timeout})
            LOG.error(_('_prepare_fc_map: Failed to start FlashCopy '
                        'from %(source)s to %(target)s with '
                        'exception %(ex)s')
                      % {'source': source_vdisk_id,
                         'target': target_vdisk_id,
                         'ex': exception_msg})
            raise exception.InvalidSnapshot(
                reason=_('_prepare_fc_map: %s') % exception_msg)

    def _start_fc_map(self, fc_map_id, source_vdisk_id, target_vdisk_id):
        try:
            out, err = self._run_ssh(['svctask', 'startfcmap', fc_map_id])
        except processutils.ProcessExecutionError as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_('_start_fc_map: Failed to start FlashCopy '
                            'from %(source)s to %(target)s.\n'
                            'stdout: %(out)s\n stderr: %(err)s')
                          % {'source': source_vdisk_id,
                             'target': target_vdisk_id,
                             'out': e.stdout,
                             'err': e.stderr})

    def _run_flashcopy(self, source_vdisk_id, target_vdisk_id, full_copy=True):
        """Create a FlashCopy mapping from the source to the target."""

        LOG.debug('enter: _run_flashcopy: execute FlashCopy from source '
                  '%(source)s to target %(target)s' %
                  {'source': source_vdisk_id, 'target': target_vdisk_id})

        fc_map_id = self._make_fc_map(source_vdisk_id, target_vdisk_id,
                                      full_copy)
        try:
            self._prepare_fc_map(fc_map_id, source_vdisk_id, target_vdisk_id)
            self._start_fc_map(fc_map_id, source_vdisk_id, target_vdisk_id)
        except Exception:
            with excutils.save_and_reraise_exception():
                self._delete_vdisk(target_vdisk_id, True)

        LOG.debug('leave: _run_flashcopy: FlashCopy started from '
                  '%(source)s to %(target)s' %
                  {'source': source_vdisk_id, 'target': target_vdisk_id})

    def _create_copy(self, src_vdisk_id, tgt_vdisk_names, full_copy, opts,
                     src_id, from_vol):
        """Create a new snapshot using FlashCopy."""

        tgt_vdisk_name_string = ', '.join(tgt_vdisk_names)
        LOG.debug('enter: _create_copy: snapshot with possible names '
                  '%(names)s from vdisk %(src_vdisk)s' %
                  {'names': tgt_vdisk_name_string,
                   'src_vdisk': src_vdisk_id})

        src_vdisk_attributes = self._get_vdisk_attributes(src_vdisk_id)
        if src_vdisk_attributes is None:
            exception_msg = (
                _('_create_copy: Source vdisk %s does not exist')
                % src_vdisk_id)
            LOG.error(exception_msg)
            if from_vol:
                raise powervc_exception.SVCVdiskNotFoundException(
                    self.endpoint_desc, src_id, vdisk_id=src_vdisk_id)
            else:
                raise exception.SnapshotNotFound(exception_msg,
                                                 snapshot_id=src_id)

        self._driver_assert(
            'capacity' in src_vdisk_attributes,
            _('_create_copy: cannot get source vdisk '
              '%(src)s capacity from vdisk attributes '
              '%(attr)s')
            % {'src': src_vdisk_id,
               'attr': src_vdisk_attributes})

        src_vdisk_size = src_vdisk_attributes['capacity']
        tgt_vdisk_id, tgt_vdisk_name = self._create_vdisk(tgt_vdisk_names,
                                                          src_vdisk_size, 'b',
                                                          opts)

        # Run the flashcopy.  If we fail to initiate (e.g. max out the number
        # of concurrent flashcopies, clean up.
        try:
            self._run_flashcopy(src_vdisk_id, tgt_vdisk_id, full_copy)
        except Exception:
            with excutils.save_and_reraise_exception():
                self._delete_vdisk(tgt_vdisk_id, True)

        LOG.debug('leave: _create_copy: snapshot vdisk %(tgt_vdisk)s '
                  'from vdisk %(src_vdisk)s' %
                  {'tgt_vdisk': tgt_vdisk_id, 'src_vdisk': src_vdisk_id})

        return tgt_vdisk_id, tgt_vdisk_name

    def _get_flashcopy_mapping_attributes(self, fc_map_id):
        LOG.debug('enter: _get_flashcopy_mapping_attributes: mapping %s'
                  % fc_map_id)

        fc_ls_map_cmd = ['svcinfo', 'lsfcmap', '-delim', '!', fc_map_id]
        # If we hit an execption return None. Everybody now deals with
        # None being a retry.
        try:
            out, err = self._run_ssh(fc_ls_map_cmd)
        except Exception:
            return None

        # Get list of FlashCopy mappings
        # We expect zero or one line if mapping does not exist,
        # two lines if it does exist, otherwise error
        lines = out.strip().split('\n')
        self._assert_ssh_return(lines[0] == ("id!" + fc_map_id),
                                '_get_flashcopy_mapping_attributes',
                                fc_ls_map_cmd, out, err)

        # since lambda function doesn't allow multiple statements,
        # create a helper function for map().
        def _helper_func(x):
            res = x.split('!')
            return (res[0], res[1])

        fcmaps = map(_helper_func, lines)
        attributes = dict(fcmaps)

        LOG.debug('leave: _get_flashcopy_mapping_attributes: mapping '
                  '%(fc_map_id)s, attributes %(attributes)s' %
                  {'fc_map_id': fc_map_id, 'attributes': attributes})

        return attributes

    def _is_vdisk_defined(self, vdisk_id):
        """Check if vdisk is defined."""
        LOG.debug('enter: _is_vdisk_defined: vdisk %s ' % vdisk_id)
        vdisk_attributes = self._get_vdisk_attributes(vdisk_id)
        LOG.debug('leave: _is_vdisk_defined: vdisk %(vdisk)s with %(str)s '
                  % {'vdisk': vdisk_id,
                     'str': vdisk_attributes is not None})
        if vdisk_attributes is None:
            return False
        else:
            return True

    def get_default_vol_type(self):
        """Returns the default volume type from the svc's .conf file"""
        vtn = self.configuration.default_volume_type
        return vtn.decode('utf-8') if vtn else vtn

    def get_storage_metadata(self):
        """Helper Method to Retrieve meteadata for a host on the SVC"""
        volume_pools = []
        ssh_cmd = ['svcinfo', 'lsmdiskgrp', '-delim', ':']
        out, err = self._run_ssh(ssh_cmd)
        # Out should return a string with the values
        # of the keys as the first item, id:name:status...
        # and the values for the keys in the following items
        # ex: 0:mdiskgrp0:online:1:175.....
        self._assert_ssh_return(out is not None and len(out) > 0,
                                'get_volume_pools',
                                ssh_cmd, out, err)
        pool_lines = out.strip().split('\n')
        table_headers = pool_lines.pop(0).split(':')
        compression = 'yes' if self.get_compression_enabled() else 'no'
        # pool_lines.pop(0) ex: ['id:name:status:mdisk_count:more_fields']
        # need to use .split to get ['id','name','mdisk_count','more_fields']
        for line in pool_lines:
            line = line.split(':')
            volume_pool = dict(zip(table_headers, line))
            volume_pools.append(volume_pool)
        return {'volume_pools': volume_pools,
                'compression_supported': compression}

    def _get_itl_list(self, cinfo):
        """ Gets the ITL for the mapped volume """
        itl_list = list()
        for hidx in cinfo.keys():
            cur_itls = {'vios_wwpn': cinfo[hidx]['source_wwn'],
                        'target_wwpn': cinfo[hidx]['target_wwn'],
                        'target_lun': cinfo[hidx]['target_lun']}
            itl_list.append(cur_itls)
        return itl_list

    def discover_volumes(self, context, filters=None):
        """Helper Method to Retrieve a list of volumes for a host
        Currently supported filters:

         'host_refs': [ WWPN, ... ]
           Returns only volumes that are mapped to at least one of the WWPNS in
           the specified list.  Matching is case-insensitive.
        """

        # If we are filtering, create a set from the WWPN list to make
        # filtering easier.  Convert the WWPNs to upper-case for consistency
        filter_set = None
        if filters is not None and 'host_refs' in filters:
            upper_refs = [ref.upper() for ref in filters['host_refs']]
            filter_set = set(upper_refs)

        # If we are filtering get the volume references to get the k2udid
        vol_refs = None
        if filters is not None and 'vol_refs' in filters:
            vol_refs = filters['vol_refs']
        # Obain vdisk-to-wwpn mapping information, which we use to
        # augment the vdisk list, and also possibly filter.
        # We will also return the vdisk-to-scsi_id mapping information
        vdisk_to_rmt_wwpn, vdisk_connect_info = self._get_vdisk_to_wwpn_map()

        volume_list = []
        # retrieve size in bytes
        ssh_cmd = ['svcinfo', 'lsvdisk', '-bytes', '-delim', ':']
        out, err = self._run_ssh(ssh_cmd)
        # self.discover_volumes(ssh_cmd)
        # Out in string format a list containing
        # a row of headers followed by rows of values
        # example of the header status:se_copy_count:IO_group_id:
        # capacity:name and so on
        # example of the values online:0:0:10.00GB:etc

        # Successful SVC query with no out/err is valid no volumes condition,
        # so simply return the empty volume _list.
        if ((len(out) == 0) and (len(err) == 0)):
            return volume_list

        list_lines = out.strip().split('\n')
        table_headers = list_lines.pop(0).split(':')
        for line in list_lines:
            line = line.split(':')
            volume = {}
            volume_table = {}
            volume_table = dict(zip(table_headers, line))
            volume['name'] = volume_table['name']
            volume['storage_pool'] = volume_table['mdisk_grp_name']
            volume['status'] = self._map_volume_status(volume_table['status'])
            volume['size'] = self._parse_volume_size(volume_table['capacity'])
            vdisk_uid = volume_table['vdisk_UID']
            volume['restricted_metadata'] = {
                RESTRICTED_METADATA_VDISK_ID_KEY: volume_table['id'],
                RESTRICTED_METADATA_VDISK_UID_KEY: vdisk_uid,
                RESTRICTED_METADATA_VDISK_NAME_KEY: volume_table['name'],
                METADATA_VOLUME_WWN: vdisk_uid,
                RESTRICTED_METADATA_BYTE_SIZE_KEY: volume_table['capacity']
            }
            volume['metadata'] = {METADATA_VOLUME_WWN: vdisk_uid}
            if volume['status'] == 'error':
                volume['support'] = {
                    "status": "not_supported",
                    "reasons": [
                        _('This volume is not a candidate for management '
                          'because it is offline. '
                          'A volume is offline and unavailable if one '
                          'of the following takes place: '
                          'both nodes in the I/O group are missing, '
                          'none of the nodes in the I/O group that are '
                          'present can access the volume, '
                          'all synchronized copies for the volume '
                          'are in storage pools that are offline, '
                          'the volume is formatting. ')
                    ]
                }
                err_msg = (_("The volume %(name)s is in the error state "
                             "and not a candidate for management.  "
                             "vdisk status: %(status)s  "
                             "vdisk id: %(vdisk_id)s  "
                             "vdisk uid: %(vdisk_uid)s  ") %
                           {'name': volume['name'],
                            'status': volume['status'],
                            'vdisk_id': volume_table['id'],
                            'vdisk_uid': vdisk_uid})
                LOG.error(err_msg)
            else:
                volume['support'] = {"status": "supported"}

            # Add mapping data, if it exists.
            rmt_wwpns, lcl_wwpns = (set(), set())
            if volume_table['id'] in vdisk_to_rmt_wwpn:
                wwpns = vdisk_to_rmt_wwpn[volume_table['id']]
                volume['mapped_wwpns'] = wwpns
                if filter_set is not None:
                    rmt_wwpns = set([wwpn.upper() for wwpn in wwpns])
                # If the Volume is attached, then Import isn't Supported
                if len(wwpns) > 0:
                    volume['status'] = 'in-use'
                    volume['support'] = {
                        "status": "not_supported",
                        "reasons": [
                            _('This volume is not a candidate for management '
                              'because it is already attached to a virtual '
                              'machine.  To manage this volume with PowerVC, '
                              'you must bring the virtual machine under '
                              'management.  Select to manage the virtual '
                              'machine that has the volume attached.  The '
                              'attached volume will be automatically included '
                              'for management.')
                        ]
                    }
            # If there is the Target Port Info, add it to the Connection Info
            if volume_table['id'] in vdisk_connect_info:
                connect_info = vdisk_connect_info[volume_table['id']]
                volume['itl_list'] = self._get_itl_list(connect_info)
                wwpns = connect_info.get('target_wwn', [])
                if filter_set is not None:
                    lcl_wwpns = set([wwpn.upper() for wwpn in wwpns])
                if len(connect_info) > 0:
                    volume['connection_info'] = connect_info

            # Store the volume k2udid
            if vol_refs:
                for vol_idx in vol_refs:
                    if vol_idx['uid'] == vdisk_uid:
                        volume['k2udid'] = vol_idx['k2udid']
                        break
            # If we are not filtering on WWPNs or UID's, or we ARE filtering
            # and there is a match, add this disk to the list to be returned.
            if ((filter_set is None or len(filter_set & rmt_wwpns) > 0 or
                 len(filter_set & lcl_wwpns) > 0 or vdisk_uid in filter_set)):
                volume_list.append(volume)

        return volume_list

    def query_volumes(self, context, volumes, server_info={}, mark_boot=True):
        """Returns import-related details for the Volumes from the Provider"""
        LOG.debug("Enter:query_volumes %(vol)s %(info)s" %
                  dict(vol=volumes, info=server_info))
        uid_map, wwpn_map = (dict(), dict())
        returned_volumes, boot_map = (list(), dict())
        udid = None
        # Build a map of the WWPN/UID's for each Server for quicker lookup
        for server_id, server in server_info.iteritems():
            mark_boot = server.get('boot') is not True
            for volume in server.get('volumes', []):
                uid = volume.get('unique_device_id')
                wwpns = volume.get('wwpns', [])
                mount = volume.get('mountpoint')
                vios_wwpn = volume.get('vios_wwpns', None)
                # Build maps of the UID and WWPN depending on which is set
                if uid is not None:
                    if uid.upper() not in uid_map:
                        server_dict = {server_id: (mark_boot, mount,
                                                   vios_wwpn)}
                    else:
                        server_dict = uid_map[uid.upper()]
                        server_dict[server_id] = (mark_boot, mount,
                                                  vios_wwpn)
                    uid_map[uid.upper()] = server_dict
                for wwpn in wwpns:
                    wwpn_map[wwpn.upper()] = (server_id, mark_boot, mount)
        # Loop through the Volumes, setting the attachment info
        for volume in volumes:
            volume_copy = volume.copy()
            server_id = None
            mark_boot, mount_point = (False, None)
            # Parse the UID out of the MetaData to use for matching Servers
            metadata = volume_copy.get('restricted_metadata', {})
            uid = metadata.get(RESTRICTED_METADATA_VDISK_UID_KEY)
            vdisk_id = metadata.get(RESTRICTED_METADATA_VDISK_ID_KEY)
            volume_copy['attach_info'] = {}
            # Parse the correct Connection Info based on the WWPN's
            connect_infos = volume_copy.pop('connection_info', {})
            if len(connect_infos) > 0:
                for cinfo in connect_infos.values():
                    wwpns = cinfo.get('source_wwn', [])
                    # Loop through each of the WWPN's looking for the Server
                    for wwpn in wwpns:
                        # If we found a Server for the WWPN, use this Info
                        if wwpn_map.get(wwpn.upper()) is not None:
                            server_id = wwpn_map[wwpn.upper()][0]
                            if server_id not in volume_copy['attach_info']:
                                mark_boot = wwpn_map[wwpn.upper()][1]
                                mount_point = wwpn_map[wwpn.upper()][2]
                                connection_info = None

                                # sanitize connection info for npiv
                                if cinfo is not None:
                                    conn_type = \
                                        discovery_driver.CONNECTION_TYPE_NPIV
                                    connection_info =\
                                        self._get_connection_info(
                                            vdisk_id, uid, cinfo,
                                            conn_type=conn_type)
                                    self._add_attach_info(
                                        volume_copy, boot_map,
                                        connection_info, server_id,
                                        mountpoint=mount_point,
                                        mark_boot=mark_boot)

            # If the Server for this UID is in the Map, add attachment
            if uid in uid_map:
                server_dict = uid_map[uid]
                for server_id in server_dict.keys():
                    if server_id not in volume_copy['attach_info']:
                        mark_boot = server_dict[server_id][0]
                        mount_point = server_dict[server_id][1]
                        udid = volume.get('k2udid')
                        connection_info = None
                        prot = self.configuration.\
                            storwize_svc_connection_protocol.lower()
                        if prot == 'iscsi':
                            if connect_infos:
                                cinfo = connect_infos.values()[0]
                                connection_info = self._get_connection_info(
                                    vdisk_id, uid, cinfo, conn_type=None,
                                    udid=udid)
                                self._add_attach_info(volume_copy, boot_map,
                                                      connection_info,
                                                      server_id,
                                                      mountpoint=mount_point,
                                                      mark_boot=mark_boot)
                        else:
                            server_wwpn = server_dict[server_id][2]
                            # get correct connection info and itl list
                            valid_itls = self._get_vscsi_itl_info(server_wwpn,
                                                                  volume_copy)
                            if valid_itls:
                                # If given the VDisk UID from the VM, then
                                # it isn't NPIV
                                # pass the first itl to extract the target_wwn
                                cinfo = valid_itls[0]
                                conn_type = discovery_driver.\
                                    CONNECTION_TYPE_VSCSI
                                connection_info = self._get_connection_info(
                                    vdisk_id, uid, cinfo, conn_type=conn_type,
                                    udid=udid)
                                connection_info['vscsi_target_lun'] =\
                                    valid_itls
                                connection_info['data']['vendor'] =\
                                    "IBM"
                                self._add_attach_info(volume_copy, boot_map,
                                                      connection_info,
                                                      server_id,
                                                      mountpoint=mount_point,
                                                      mark_boot=mark_boot)
            # Add the Copy of the Volume to the list to be returned
            returned_volumes.append(volume_copy)
        LOG.info("Exit query_volumes %s" % returned_volumes)
        return returned_volumes

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        delvols, voluids = (list(), dict())
        # Run the lsvdisk command to query the list of volumes on the SVC
        ssh_cmd = ['svcinfo', 'lsvdisk', '-bytes', '-delim', ':']
        out, _err = self._run_ssh(ssh_cmd)
        # Parse out the lines and the table header from the output
        list_lines = out.strip().split('\n')
        table_headers = list_lines.pop(0).split(':')
        # Loop through each line adding the UID to the output list
        for line in list_lines:
            line = line.split(':')
            volume_table = dict(zip(table_headers, line))
            voluids[volume_table['vdisk_UID']] = ''
        # Now loop through the passed in volumes seeing which don't exist
        for volume in volumes:
            metadata = volume.get('volume_restricted_metadata', list())
            metadata = dict([(m['key'], m['value']) for m in metadata])
            voluid = metadata.get(RESTRICTED_METADATA_VDISK_UID_KEY)
            # If the Volume has a UID and isn't on the system, mark as deleted
            if voluid and voluid not in voluids:
                delvols.append(volume)
        return delvols

    def _get_connection_info(self, vdisk_id, vdisk_uid,
                             connect_info, conn_type=None, udid=None):
        """Helper method to get Connection Info for the Discovered Volume"""
        prot = self.configuration.storwize_svc_connection_protocol.lower()
        driver_volume_type = 'iscsi' if prot == 'iscsi' else 'fibre_channel'
        volume_attrs = self._get_vdisk_attributes(vdisk_id)
        # If this is iSCSI then we need to add a few additional attributes
        if prot == 'iscsi':
            target_lun = connect_info.get('target_lun')
            target_wwn = connect_info.get('target_wwn')
            data = dict(access_mode='rw', qos_spec=None,
                        target_discovered=False,
                        target_lun=target_lun, target_wwn=target_wwn)
            preferred_node = None
            # Loop through and determine the Node that should be used
            for node in self._state['storage_nodes'].values():
                if node['id'] == volume_attrs['preferred_node_id']:
                    preferred_node = node
            # If we found the node, that add the IQN and Portal values
            if preferred_node:
                if len(preferred_node['ipv4']):
                    ipaddr = preferred_node['ipv4'][0]
                else:
                    ipaddr = preferred_node['ipv6'][0]
                # Add the iSCSI-specific attributes to the connection info
                device_path = '/dev/disk/by-id/wwn-0x%s' % vdisk_uid.lower()
                data['target_portal'] = '%s:%s' % (ipaddr, '3260')
                data['target_iqn'] = preferred_node['iscsi_name']
                data['device_path'] = device_path
                data['target_UID'] = vdisk_uid
                data.pop('target_wwn', None)
            # Add the Data to the overall Connection Info that is returned
            connect_info = dict(driver_volume_type=driver_volume_type,
                                data=data)
            return(connect_info)

        # call get_connection_info in discovery driver for other cases
        s = super(StorwizeSVCProductDriver, self)
        connect_info = s._get_connection_info(vdisk_uid, connect_info,
                                              conn_type, udid)

        return connect_info

    @staticmethod
    def _parse_volume_size(size_bytes):
        """Helper Method to convert bytes to GB"""
        size = float(size_bytes) / 1073741824
        # round up to nearest GB
        size = int(math.ceil(size))
        return size

    @staticmethod
    def _map_volume_status(statusstr):
        """Helper Method to Map the SVC Status to Openstack Statuses"""
        # Currently anything other than online would be considered error
        return 'available' if statusstr == 'online' else 'error'

    def _svc_cmd_to_map(self, cmd, delim, key1, key2):
        ret = {}

        out, err = self._run_ssh(cmd)
        self._assert_ssh_return(out is not None,
                                '_svc_cmd_to_map',
                                cmd, out, err)
        list_lines = out.strip().split('\n')
        table_headers = list_lines.pop(0).split(delim)
        for line in list_lines:
            line = line.split(delim)
            row = dict(zip(table_headers, line))
            if not row[key1] in ret:
                ret[row[key1]] = {}
            ret[row[key1]][row[key2]] = row

        return ret

    def _get_wwpns_for_hosts(self, host_ids, host_names, host_wwpns):
        host_id = ''
        ssh_cmd = list()
        for host_id in host_ids:
            ssh_cmd.extend([';', 'svcinfo', 'lshost', '-delim', '!', host_id])

        out, err = self._run_ssh(ssh_cmd[1:])
        self._assert_ssh_return(len(out.strip()),
                                '_get_wwpns_for_host',
                                ssh_cmd, out, err)

        for attr_line in out.split('\n'):
            # If '!' not found, return the string and two empty strings
            attr_name, foo, attr_val = attr_line.partition('!')
            if attr_name == 'id':
                host_id = attr_val
                host_wwpns[host_id] = list()
            if attr_name == 'name':
                host_names[host_id] = attr_val
            if attr_name == 'WWPN':
                host_wwpns[host_id].append(attr_val)

    def _get_vdisk_to_wwpn_map(self):
        """Via a call to lshostviskmap and repeated calls to lshost, produces
           a map from vdisk ID to a list of WWPNs"""
        scsi_ids = dict()
        rmt_wwpns, connect_info = (dict(), dict())
        # Build a Map of the Output for the each Virtual Disk
        vdisk_map = self._svc_cmd_to_map(['svcinfo', 'lshostvdiskmap',
                                          '-delim', '!'],
                                         '!', 'vdisk_id', 'id')
        # Build a Map of the Local WWPN's for each of the Hosts
        lclhost_wwns = self._svc_cmd_to_map(['svcinfo', 'lsfabric', '-delim',
                                             '!'], '!', 'name', 'local_wwpn')
        # Build a Map of the Remote WWPN's for each of the Hosts
        host_ids = list()
        host_names = dict()
        rmthost_wwns = dict()
        for vdisk_id in vdisk_map.keys():
            for host_id in vdisk_map[vdisk_id].keys():
                # We only want to do anything if we haven't processed this Host
                if host_id not in host_ids and host_id not in rmthost_wwns:
                    host_ids.append(host_id)
                    # For Performance reasons group multiple Hosts in one call
                    if len(host_ids) >= 20:
                        self._get_wwpns_for_hosts(
                            host_ids, host_names, rmthost_wwns)
                        host_ids = list()
        # If there were any Hosts left, make one final call to process those
        if len(host_ids) > 0:
            self._get_wwpns_for_hosts(host_ids, host_names, rmthost_wwns)
        # Loop through the Virtual Disks a second time to build the return maps
        for vdisk_id in vdisk_map.keys():
            scsi_ids[vdisk_id] = '-1'
            rmt_wwpns[vdisk_id] = list()
            connect_info[vdisk_id] = dict()
            # Update Remote and Local WWPN's for each Virtual Disk Identifier
            for host_id in vdisk_map[vdisk_id].keys():
                host_name = host_names.get(host_id, '')
                rmt_wwpns[vdisk_id].extend(rmthost_wwns.get(host_id, []))
                # Add the Local WWPN's to the list for the given Virtual Disk
                connect_info[vdisk_id][host_id] = dict()
                connect_info[vdisk_id][host_id]['source_wwn'] =\
                    rmthost_wwns.get(host_id, [])
                connect_info[vdisk_id][host_id]['target_wwn'] =\
                    lclhost_wwns.get(host_name, {}).keys()
                # Add the SCSI Identifier for the given Virtual Disk Identifier
                connect_info[vdisk_id][host_id]['target_lun'] =\
                    vdisk_map[vdisk_id][host_id]['SCSI_id']

        return rmt_wwpns, connect_info

    def _override_zone_map(self, volume, connector, zone_map):
        LOG.debug('Enter _override_zone_map '
                  'volume volume %(volume)s '
                  'connector %(connector)s '
                  'zone_map %(zone_map)s' %
                  dict(volume=volume,
                       connector=connector,
                       zone_map=zone_map))
        target_map = connector.pop('target_map', None)
        # if there is no target_map we requested no override for
        # this zone mapping operation.
        if not target_map:
            return
        try:
            self._prune_it_node_map(volume, connector, zone_map, target_map)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error(ex)
        LOG.info('Exit _override_zone_map %s' % zone_map)

    def _prune_it_node_map(self, volume, connector, zone_map, target_map):
        """
        For each initiator map that initiator to a single back-end storage
        target on each SVC node. This is in-line with expected and stardard
        zoning policy for Storwise
        Each pair of virtual initiators will be mapped to the same target
        """
        LOG.debug('Enter _prune_it_node_map')
        new_zone_map = {}
        target_node_wwpns = target_map['target_nodes_wwpns']

        i_index_list, i_node_index = \
            pvc_fczm_utils.build_index_of_initiators(connector, zone_map)

        for fabric in zone_map:
            LOG.debug("fabric %s" % fabric)
            for i_wwpn in sorted(zone_map[fabric]):
                t_wwpns = zone_map[fabric][i_wwpn]
                LOG.debug("t_wwpns %s" % t_wwpns)

                # pull out the already generated index for this
                # initiator
                host_index = i_index_list[i_node_index[i_wwpn]]

                new_target_wwpns = []
                # For each of the nodes that we are interested in, ensure that
                # we found at least one WWPN that belongs to that node.
                for node in target_node_wwpns:
                    node_wwpns = [wwpns.lower() for wwpns in
                                  target_node_wwpns[node]]
                    LOG.debug("node_wwpns %s" % node_wwpns)
                    found_node_ports = list(set(t_wwpns) &
                                            set(node_wwpns))

                    if not found_node_ports:
                        msg = (_('Fabric %(fabric)s does not have visibility '
                                 'to any WWPNs in node %(node)s, part of '
                                 'iogrp %(iogrp)s of storage controller '
                                 '%(name)s. WWPNs visible on the fabric are '
                                 '%(visible)s. WWPNs on the controller are '
                                 '%(available)s.')
                               % {'fabric': fabric,
                                  'node': node,
                                  'iogrp': target_map['iogrp'],
                                  'name': self.configuration.host,
                                  'visible': ';'.join(t_wwpns),
                                  'available': ';'.join(node_wwpns)}
                               )
                        LOG.warning(msg)
                        continue
                    # we have a list of logged-in target ports on this fabric,
                    # for this node, and we want to select one to use.  When
                    # multiple ports are available, we don't want to select the
                    # same one each time, so we index into the list using
                    # the svc host_id
                    # We are building up a list of WWPNs in the Fabric object,
                    # one per node.
                    sorted_ports = sorted(found_node_ports)
                    hash_index = ((int(target_map['host_id']) +
                                   host_index)
                                  % len(sorted_ports))
                    new_target_wwpns.append(sorted_ports[hash_index])
                # replace the target wwpns
                LOG.info(_('SVC new_target_wwpns %s') % new_target_wwpns)
                zone_map[fabric][i_wwpn] = new_target_wwpns

    @pvc_fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        """
        Calls the parent class to mask the volume on the storage controller,
        and here performs additional processing to work out which zones should
        be created on the switches.

        For NPIV, the virtual fibre channel ports will not yet be logged in,
        but we need to know which fabrics which ports will appear on, so that
        we know which fabrics to create the zones on.  For this reason, the
        connector must also include a data structure that maps physical WWPNs
        (which will be logged in) to the virtual WWPNs that they will provide.
        To support live partition mobility, we expect each physical WWPN to be
        providing two virtual WWPNs.

        The expected structure of a fibre channel connector is:

        {
            'host': hostname,
            'wwpns': [ list of virtual WWPNs ]  <-- this is used for masking
            'phy_to_virt_initiators': {phy_wwpn_X: [virt_wwpn_a, virt_wwpn_b],
                                      phy_wwpn_Y: [virt_wwpn_c, virt_wwpn_d]
                                        }
        }

        The expected structure of an iSCSI connector is:

        {
          'host': hostname,
          'initiator': initiator  <-- IQN format
        }
        """

        # Ensure that the passed-in data structure is correct.
        expected_fc_keys = set(['host', 'wwpns'])
        expected_iscsi_keys = set(['host', 'initiator'])
        if not (expected_fc_keys.issubset(connector) |
                expected_iscsi_keys.issubset(connector)):
            err_msg = (_('The connector %(connector)s does not contain the '
                         'required keys, which are %(expected_fc_keys)s.'
                         'or %(expected_iscsi_keys)s') %
                       {'connector': connector,
                        'expected_fc_keys': expected_fc_keys,
                        'expected_iscsi_keys': expected_iscsi_keys})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        # Perform the masking on the storage controller
        """Perform the necessary work so that an iSCSI/FC connection can
        be made.

        To be able to create an iSCSI/FC connection from a given host to a
        volume, we must:
        1. Translate the given iSCSI name or WWNN to a host name
        2. Create new host on the storage system if it does not yet exist
        3. Map the volume to the host if it is not already done
        4. Return the connection information for relevant nodes (in the
           proper I/O group)

        """
        vdisk_id = self._incorporate_restricted_metadata(volume)
        LOG.info(_("Enter: initialize_connection: %s") % volume.get('name'))
        LOG.debug('volume %(vol)s with connector %(conn)s' %
                  {'vol': str(volume), 'conn': str(connector)})

        vol_opts = self._get_vdisk_params(volume['volume_type_id'])
        # host_name = connector['host']

        # Obtain the disk's attributes, raises an exception on a communications
        # error, returns None if the disk doesn't exist.
        volume_attributes = self._get_vdisk_attributes(vdisk_id)

        if not volume_attributes:
            raise powervc_exception.SVCVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id)

        # Ensure that the UID matches and we don't attach the wrong disk,
        # raise exception if the UID doesn't match.
        self._verify_uid(volume_attributes, volume)

        try:
            preferred_node = volume_attributes['preferred_node_id']
            IO_group = volume_attributes['IO_group_id']
        except KeyError as e:
            LOG.error(_('Did not find expected column name in '
                        'lsvdisk: %s') % e)
            exception_msg = (_('initialize_connection: Missing volume '
                               'attribute for vdisk %s') % vdisk_id)
            raise exception.VolumeBackendAPIException(data=exception_msg)

        # store multipath into connector for host creation
        connector['multipath'] = vol_opts['multipath']
        return_all_wwpns = connector['multipath']

        # Check if a host object is defined for this host name
        host_name = self._get_host_from_connector(connector, vdisk_id,
                                                  IO_group)
        if host_name is None:
            # Host does not exist - add a new host to Storwize/SVC
            host_name = self._create_host(connector, IO_group)
            # Verify that create_new_host succeeded
            self._driver_assert(
                host_name is not None,
                _('_create_host failed to return the host name.'))
        elif (connector.get('connection_type') and 'npiv' in
              connector.get('connection_type')):
            # We can always add the io group so don't check if it
            # is already there.
            stringify_host = "'" + host_name + "'"
            ssh_cmd = ['svctask', 'addhostiogrp', '-iogrp',
                       IO_group, stringify_host]
            self._run_ssh(ssh_cmd)

        if vol_opts['protocol'] == 'iSCSI':
            chap_secret = self._get_chap_secret_for_host(host_name)
            if chap_secret is None:
                chap_secret = self._add_chapsecret_to_host(host_name)

        #use backend volume name to map to host as the volume name from db
        #does not match backend name in case of a upgrade scenario
        backend_name = volume_attributes['name']
        multihostmap = self.configuration.storwize_svc_multihostmap_enabled

        @synchronized(host_name, 'fchost-', True)
        def map_vol_to_host():
            return self._helpers.map_vol_to_host(backend_name, host_name,
                                                 multihostmap)
        lun_id = map_vol_to_host()

        try:
            # Get preferred node and other nodes in I/O group
            preferred_node_entry = None
            io_group_nodes = []
            for k, node in self._state['storage_nodes'].iteritems():
                if vol_opts['protocol'] not in node['enabled_protocols']:
                    continue
                if node['id'] == preferred_node:
                    preferred_node_entry = node
                if node['IO_group'] == IO_group:
                    io_group_nodes.append(node)

            if not len(io_group_nodes):
                exception_msg = (_('initialize_connection: No node found in '
                                   'I/O group %(gid)s for vdisk %(vdisk_id)s')
                                 % {'gid': IO_group, 'vdisk_id': vdisk_id})
                LOG.error(exception_msg)
                raise exception.VolumeBackendAPIException(data=exception_msg)

            if not preferred_node_entry and not vol_opts['multipath']:
                # Get 1st node in I/O group
                preferred_node_entry = io_group_nodes[0]
                LOG.warn(_('initialize_connection: Did not find a preferred '
                           'node for vdisk %s') % vdisk_id)

            properties = {}
            properties['target_discovered'] = False
            properties['target_lun'] = lun_id
            properties['volume_id'] = volume['id']
            properties['vendor'] = "IBM"
            if vol_opts['protocol'] == 'iSCSI':
                type_str = 'iscsi'
                if len(preferred_node_entry['ipv4']):
                    ipaddr = preferred_node_entry['ipv4'][0]
                else:
                    ipaddr = preferred_node_entry['ipv6'][0]
                properties['target_UID'] = volume_attributes['vdisk_UID']
                properties['target_portal'] = '%s:%s' % (ipaddr, '3260')
                properties['target_iqn'] = preferred_node_entry['iscsi_name']
                properties['auth_method'] = 'CHAP'
                properties['auth_username'] = connector['initiator']
                LOG.debug('leave: initialize_connection:\n volume: %(vol)s\n '
                          'connector %(conn)s\n properties: %(prop)s'
                          % {'vol': str(volume),
                             'conn': str(connector),
                             'prop': str(properties)})
                properties['auth_password'] = chap_secret
            else:
                type_str = 'fibre_channel'
                conn_wwpns = self._get_conn_fc_wwpns(host_name,
                                                     connector,
                                                     IO_group)
                if connector['connection_type']:
                    if 'pv_vscsi' in connector['connection_type']:
                        return_all_wwpns = True
                        properties['target_UID'] =\
                            volume_attributes['vdisk_UID']
                        properties['pg83NAA'] = volume_attributes['vdisk_UID']
                # Add i_t_map just as the community driver does.
                i_t_map = self.\
                    _make_initiator_target_map(connector['wwpns'], conn_wwpns)
                properties['initiator_target_map'] = i_t_map

                    # TO-DO : May need to add information for vscsi
                if len(conn_wwpns) == 0:
                    msg = (_('Could not get FC connection information for the '
                             'host-volume connection. Is the host configured '
                             'properly for FC connections?'))
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                if not return_all_wwpns:
                    if preferred_node_entry['WWPN'] in conn_wwpns:
                        properties['target_wwn'] = preferred_node_entry['WWPN']
                    else:
                        properties['target_wwn'] = conn_wwpns[0]
                else:
                    properties['target_wwn'] = list(set(conn_wwpns))
                LOG.debug('leave: initialize_connection:\n volume: %(vol)s\n '
                          'connector %(conn)s\n properties: %(prop)s'
                          % {'vol': str(volume),
                             'conn': str(connector),
                             'prop': str(properties)})
        except Exception:
            with excutils.save_and_reraise_exception():
                self.terminate_connection(volume, connector)
                LOG.error(_('initialize_connection: Failed to collect return '
                            'properties for volume %(vol)s and connector '
                            '%(conn)s.\n') % {'vol': str(volume),
                                              'conn': str(connector)})
        info = {'driver_volume_type': type_str, 'data': properties, }

        # Removing multipath from connector, isolated use only for svc
        if 'multipath' in connector:
            del connector['multipath']
        return info

    def _get_conn_fc_wwpns(self, host_name, connector, iogrp):
        """
        We overrided this method because for NPIV the virtual ports are not yet
        logged in, so we don't expect our host to have any connectivity to the
        storage controller.
        We return the wwpns for either all the SVC ports if this in multipath
        or the ports for the requested IOgroup.
        for npiv we will use the phy_to_virt_initiators passed in to
        determine the proper zoning on the decorator
        """

        LOG.info("Enter _get_conn_fc_wwpns")
        return_all_wwpns = connector.get('multipath', False)

        # if vscsi want to start with the full set targets
        if connector.get('connection_type') and iogrp is not None:
            if 'pv_vscsi' in connector['connection_type']:
                return_all_wwpns = True
        else:
            return_all_wwpns = True

        target_wwpns = []
        if return_all_wwpns:
            for node_id, node in self._state['storage_nodes'].items():
                LOG.info(_('Incorporating WWPNS for Node %(node_id)s: '
                           '%(wwpns)s') %
                         {'node_id': node_id, 'wwpns': node['WWPN']})
                target_wwpns.extend(node['WWPN'])
        else:
            target_nodes_wwpns = {}
            for node_id, node in self._state['storage_nodes'].items():
                if node['IO_group'] == iogrp:
                    LOG.info(_('Incorporating WWPNS for Node %(node_id)s: '
                               '%(wwpns)s') %
                             {'node_id': node_id, 'wwpns': node['WWPN']})
                    LOG.info("node %s" % node['WWPN'])
                    target_wwpns.extend(node['WWPN'])
                    target_nodes_wwpns[node_id] = node['WWPN']
            target_map = {}
            target_map['target_nodes_wwpns'] = target_nodes_wwpns
            target_map['iogrp'] = iogrp
            resp = self.ssh.lshost(host=host_name)
            for id in resp.select('id'):
                target_map['host_id'] = id
                break
            connector['target_map'] = target_map

        LOG.info("Exit _get_conn_fc_wwpns %s" % target_wwpns)
        return target_wwpns

    def validate_connector(self, connector):
        """Check connector for at least one enabled protocol (iSCSI/FC)."""
        valid = False
        if 'iSCSI' in self._state['enabled_protocols'] and \
                'initiator' in connector:
            valid = True
        if 'FC' in self._state['enabled_protocols'] and 'wwpns' in connector:
            valid = True
        if not valid:
            err_msg = (_('The connector does not contain the required '
                         'information: %(connector)s, protocols %(protos)s')
                       % {'connector': connector,
                          'protos': self._state['enabled_protocols']})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

    @pvc_fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        LOG.info(_("Terminating connection with parameters %(connector)s" %
                   {'connector': connector}))

        # Perform the masking on the storage controller
        """Cleanup after an iSCSI connection has been terminated.

        When we clean up a terminated connection between a given connector
        and volume, we:
        1. Translate the given connector to a host name
        2. Remove the volume-to-host mapping if it exists
        3. Delete the host if it has no more mappings (hosts are created
           automatically by this driver when mappings are created)
        """
        LOG.debug('enter: terminate_connection: volume %(vol)s with '
                  'connector %(conn)s' % {'vol': volume['id'],
                                          'conn': str(connector)})

        vdisk_id = self._incorporate_restricted_metadata(volume)

        # Ensure that the UID matches and we don't detach the wrong disk,
        # Warn for certain exceptions.
        try:
            self._verify_uid_by_vdisk_id(vdisk_id, volume)
        except powervc_exception.SVCVdiskMismatchException as e:
            # Volume has been deleted and recreated, probably in use by
            # someone else.  Do nothing, but warn.
            LOG.warn(_("terminate_connection: vdisk mismatch: %(err)s. "
                       "Performing no operation on the storage controller")
                     % {'err': six.text_type(e)})
            return
        except powervc_exception.SVCVdiskNotFoundException as e:
            # Volume has been deleted, do nothing, but warn.
            LOG.warn(_("terminate_connection: vdisk deleted: %(err)s. "
                       "Performing no operation on the storage controller")
                     % {'err': six.text_type(e)})
            return
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.exception(e)  # unexpected

        # We are just trying to get the IO group name so we can detach
        # passing io group None here as we arn't going to create the
        host_name = self._get_host_from_connector(connector, vdisk_id,
                                                  None,
                                                  is_initconn=False)
        # Verify that _get_host_from_connector returned the host.
        # Is there any situation where this would just be a warning?
        if host_name is None:
            if self._helpers.is_vdisk_in_use(vdisk_id):
                msg = _('Detach volume: failed to lookup host name '
                        'for vdisk ID: %(vdisk)s') % dict(vdisk=vdisk_id)
                LOG.error(msg)
                # don't raise without any exception. Otherwise will receive
                # exception "exceptions must be old-style classes or derived
                # from BaseException, not NoneType"
                raise exception.VolumeDriverException(message=msg)
            else:
                #vdisk not in use and no hostname found. Could be
                #host got deleted. Nothing to terminate
                LOG.info(_("terminate_connection: No host entry found "
                           "and no mappings for vdisk ID %(vdisk)s. "
                           "Performing no op on the storage controller.") %
                         dict(vdisk=vdisk_id))
                return

        # Check if vdisk-host mapping exists, remove if it does
        mapping_data = self._get_hostvdisk_mappings(host_name)
        if vdisk_id in mapping_data:
            LOG.debug("Before rmvdiskhostmap: vdisk_id=%s, host_name"
                      "=%s, mapping_data=%s" % (vdisk_id, host_name,
                                                mapping_data[vdisk_id]))
            stringify_host = "'" + host_name + "'"
            ssh_cmd = ['svctask', 'rmvdiskhostmap', '-host', stringify_host,
                       vdisk_id]
            try:
                out, err = self._run_ssh(ssh_cmd)
                LOG.debug("After rmvdiskhostmap: stdout=%s, stderr=%s" %
                          (out, err))
            except processutils.ProcessExecutionError as e:
                LOG.exception(e)
                err = six.text_type(e)
            # check for error case. The non empty 'out' case is already
            # handled by the assert below.
            if err and len(out.strip()) == 0:
                msg = _("There was an error when attempting to remove the "
                        "host disk mappings for volume ID %(vol)s and vdsik "
                        "ID %(vdisk)s: %(err)s") %\
                    dict(vol=volume['id'], vdisk=vdisk_id, err=err)
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Verify CLI behaviour - no output is returned from rmvdiskhostmap
            self._assert_ssh_return(len(out.strip()) == 0,
                                    'terminate_connection', ssh_cmd, out, err)
            del mapping_data[vdisk_id]
        else:
            LOG.warn(_('terminate_connection: No mapping of vdisk '
                       '%(vdisk_id)s to host %(host_name)s found') %
                     {'vdisk_id': vdisk_id, 'host_name': host_name})

        # If this host has no more mappings, delete it
        LOG.debug("terminate_connection: connector in terminate_connection %s"
                  % connector)
        properties = {}
        if self.configuration.zoning_mode == 'fabric':
            type_str = 'fibre_channel'
        else:
            type_str = 'iscsi'
        if not mapping_data:
            LOG.debug("No more mappings. Delete host definition on storage "
                      "backend: %s" % host_name)
            self._delete_host(host_name)
            LOG.debug("Host definition removed.")
            if 'initiator_target_map' in connector:
                properties['initiator_target_map'] = \
                    connector['initiator_target_map']
            elif self.configuration.zoning_mode == 'fabric':
                conn_wwpns = self._get_conn_fc_wwpns(host_name,
                                                     connector,
                                                     None)
                i_t_map = self.\
                    _make_initiator_target_map(connector['wwpns'],
                                               conn_wwpns)
                properties['initiator_target_map'] = i_t_map

        LOG.debug('leave: terminate_connection: volume %(vol)s with '
                  'connector %(conn)s' % {'vol': str(volume),
                                          'conn': str(connector)})
        if self.configuration.zoning_mode == 'fabric':
            type_str = 'fibre_channel'
        else:
            type_str = 'iscsi'
        info = {'driver_volume_type': type_str, 'data': properties, }
        return info

    def _get_vdisk_attributes(self, vdisk_id):
        """Return vdisk attributes, or None if vdisk does not exist

        Exception is raised if the information from system can not be
        parsed/matched to a single vdisk.
        """

        ssh_cmd = ['svcinfo', 'lsvdisk', '-bytes', '-delim', '!', vdisk_id]
        return self._execute_command_and_parse_attributes(ssh_cmd)

    def _get_vdisk_fc_mappings(self, vdisk_id):
        """Return FlashCopy mappings that this vdisk is associated with."""

        ssh_cmd = ['svcinfo', 'lsvdiskfcmappings', '-nohdr', vdisk_id]
        out, err = self._run_ssh(ssh_cmd)

        mapping_ids = []
        if (len(out.strip())):
            lines = out.strip().split('\n')
            mapping_ids = [line.split()[0] for line in lines]
        return mapping_ids

    def _get_vdisk_params(self, type_id):
        opts = self._build_default_opts()
        if type_id:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, type_id)
        else:
            volume_type = volume_types.get_default_volume_type()
        if volume_type:
            specs = volume_type.get('extra_specs')
            for k, value in specs.iteritems():
                # Get the scope, if using scope format
                key_split = k.split(':')
                if len(key_split) == 1:
                    scope = None
                    key = key_split[0]
                else:
                    scope = key_split[0]
                    key = key_split[1]

                # We generally do not look at capabilities in the driver, but
                # protocol is a special case where the user asks for a given
                # protocol and we want both the scheduler and the driver to act
                # on the value.
                if scope == 'capabilities' and key == 'storage_protocol':
                    scope = None
                    key = 'protocol'
                    words = value.split()
                    self._driver_assert(words and
                                        len(words) == 2 and
                                        words[0] == '<in>',
                                        _('protocol must be specified as '
                                          '\'<in> iSCSI\' or \'<in> FC\''))
                    del words[0]
                    value = words[0]

                # Anything keys that the driver should look at should have the
                # 'drivers' scope.
                if scope and scope != "drivers":
                    continue

                if key in opts:
                    this_type = type(opts[key]).__name__
                    if this_type == 'int':
                        value = int(value)
                    elif this_type == 'bool':
                        value = strutils.bool_from_string(value)
                    opts[key] = value

        self._check_vdisk_opts(opts)
        return opts

    def _get_vdisk_id_from_vdisk_name(self, name):
        """Given a vdisk name, query and return the corresponding vdiskid.
           Returns None if there was no vdisk with the specified name.
        """
        ssh_cmd = ['svcinfo', 'lsvdisk', '-bytes', '-delim', '!', name]
        attrs = self._execute_command_and_parse_attributes(ssh_cmd)
        if attrs is None:
            return None

        return attrs['id']

    def _get_vdisk_name_from_vdisk_id(self, vdisk_id):
        """Given a vdisk name, query and return the corresponding vdiskid.
           Returns None if there was no vdisk with the specified name.
        """
        ssh_cmd = ['svcinfo', 'lsvdisk', '-bytes', '-delim', '!', vdisk_id]
        attrs = self._execute_command_and_parse_attributes(ssh_cmd)
        if attrs is None:
            return None

        return attrs['name']

    def _run_ssh(self, *args, **kwargs):
        """
        Wrapper for SSH command execution, captures a specific case where we
        can't establish a connection due to an overloaded SVC.
        """
        try:
            return self._run_ssh_local(*args, **kwargs)
        except paramiko.SSHException as e:
            if "Error reading SSH protocol banner" in e:
                raise powervc_exception.SVCConnectionException(
                    self.endpoint_desc)
            else:
                raise

    def extend_volume(self, volume, new_size):
        """
        IBM version of extend_volume which will not be blocked by
        the volume flashcopy. It will raise exception and return
        immediately and indicate there is an in-flight flashcopy.
        So the caller, which is nova resize function doesn't have
        to poll the volume state. The caller should call back extend
        volume again.
        """
        LOG.debug(_('enter: extend_volume: volume %s') % volume['id'])
        vdisk_id = self._incorporate_restricted_metadata(volume)
        volume_attribute = self._get_vdisk_attributes(vdisk_id)
        if not volume_attribute:
            # the vdisk has been deleted out-of-band.
            raise powervc_exception.SVCVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id)

        # Ensure that the UID matches and we don't extend the wrong disk,
        # raise exception if the UID doesn't match.
        self._verify_uid(volume_attribute, volume)

        extend_amt = int(new_size) - volume['size']
        ssh_cmd = (['svctask', 'expandvdisksize', '-size', str(extend_amt),
                    '-unit', 'gb', vdisk_id])
        out, err = self._run_ssh(ssh_cmd, check_exit_code=False)

        if 'CMMVC5837E' in err:
            # The action failed because the virtual disk (VDisk)
            # is part of a FlashCopy mapping.
            raise powervc_exception.PVCExpendvdiskFCMapException()
        else:
            # No output and stderr should be returned from expandvdisksize
            self._assert_ssh_return(len(out.strip()) + len(err.strip()) == 0,
                                    'extend_volume', ssh_cmd, out, err)

        LOG.debug(_('leave: extend_volume: volume %s') % volume['id'])

    def migrate_volume(self, ctxt, volume, host):
        """Migrate direclty if source and dest are managed by same storage.

        The method uses the migratevdisk method, which returns almost
        immediately, if the source and target pools have the same extent_size.
        Otherwise, it uses addvdiskcopy and rmvdiskcopy, which require waiting
        for the copy operation to complete.

        :param ctxt: Context
        :param volume: A dictionary describing the volume to migrate
        :param host: A dictionary describing the host to migrate to, where
                     host['host'] is its name, and host['capabilities'] is a
                     dictionary of its reported capabilities.
        """
        LOG.debug('enter: migrate_volume: id=%(id)s, host=%(host)s' %
                  {'id': volume['id'], 'host': host['host']})

        # Needs update to support vdisk IDs.
        raise NotImplementedError()

        false_ret = (False, None)
        if 'location_info' not in host['capabilities']:
            return false_ret
        info = host['capabilities']['location_info']
        try:
            (dest_type, dest_id, dest_pool) = info.split(':')
        except ValueError:
            return false_ret
        if (dest_type != 'StorwizeSVCDriver' or
                dest_id != self._state['system_id']):
            return false_ret

        if 'extent_size' not in host['capabilities']:
            return false_ret
        if host['capabilities']['extent_size'] == self._extent_size:
            # If source and dest pools have the same extent size, migratevdisk
            ssh_cmd = ['svctask', 'migratevdisk', '-mdiskgrp', dest_pool,
                       '-vdisk', volume['name']]
            out, err = self._run_ssh(ssh_cmd)
            # No output should be returned from migratevdisk
            self._assert_ssh_return(len(out.strip()) == 0, 'migrate_volume',
                                    ssh_cmd, out, err)
        else:
            # If source and dest pool extent size differ, add/delete vdisk copy
            copy_info = self._get_vdisk_copy_info(volume['name'])
            copies = list(copy_info.keys())
            self._driver_assert(len(copies) == 1,
                                _('migrate_volume started with more than one '
                                  'vdisk copy'))
            orig_copy_id = copies[0]

            opts = self._get_vdisk_params(volume['volume_type_id'])
            params = self._get_vdisk_create_params(opts)
            ssh_cmd = (['svctask', 'addvdiskcopy'] + params +
                       ['-mdiskgrp', dest_pool, volume['name']])
            out, err = self._run_ssh(ssh_cmd)
            self._assert_ssh_return(len(out.strip()), 'migrate_volume',
                                    ssh_cmd, out, err)

            # Ensure that the output is as expected
            match_obj = re.search('Vdisk \[([0-9]+)\] copy \[([0-9]+)\] '
                                  'successfully created', out)
            # Make sure we got a "successfully created" message with copy id
            self._driver_assert(
                match_obj is not None,
                _('migrate_volume %(name)s - did not find '
                  'success message in CLI output.\n '
                  'stdout: %(out)s\n stderr: %(err)s')
                % {'name': volume['name'], 'out': str(out), 'err': str(err)})

            copy_id = match_obj.group(2)
            sync = False
            while not sync:
                ssh_cmd = ['svcinfo', 'lsvdiskcopy', '-delim', '!', '-copy',
                           copy_id, volume['name']]
                attrs = self._execute_command_and_parse_attributes(ssh_cmd)
                if not attrs:
                    msg = (_('migrate_volume: Could not get vdisk copy data'))
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                if attrs['sync'] == 'yes':
                    sync = True
                else:
                    time.sleep(5)

            ssh_cmd = ['svctask', 'rmvdiskcopy', '-copy', orig_copy_id,
                       volume['name']]
            out, err = self._run_ssh(ssh_cmd)
            # No output should be returned from rmvdiskcopy
            self._assert_ssh_return(len(out.strip()) == 0, 'migrate_volume',
                                    ssh_cmd, out, err)

        LOG.debug('leave: migrate_volume: id=%(id)s, host=%(host)s' %
                  {'id': volume['id'], 'host': host['host']})
        return (True, None)

    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If we haven't gotten stats yet or 'refresh' is True,
        run update the stats first.
        """
        if not self._stats or refresh:
            self._update_volume_stats()

        return self._stats

    def check_volume_health(self, volume):
        """
        Throws exceptions if something is wrong with the passed-in volume.

        Checks for existence of the underlying vdisk
        """
        vdisk_id = self._incorporate_restricted_metadata(volume)

        if not self._check_vdisk_exists(vdisk_id):
            raise powervc_exception.SVCVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id)

    def _update_volume_stats(self):
        """ Retrieve stats on all the disk groups to get an overall
        total "total" and "free" capacity in gb and plug those
        into the volume status
        """

        total_capacity_gb = 0
        free_capacity_gb = 0
        ssh_cmd = ['svcinfo', 'lsmdiskgrp', '-bytes', '-delim', ':']
        out, err = self._run_ssh(ssh_cmd)
        # Out should return a string with the values
        # of the keys as the first item, id:name:status...
        # and the values for the keys in the following items
        # ex: 0:mdiskgrp0:online:1:175.....
        self._assert_ssh_return(out is not None and len(out) > 0,
                                '_update_volume_stats',
                                ssh_cmd, out, err)
        pool_lines = out.strip().split('\n')
        table_headers = pool_lines.pop(0).split(':')
        # pool_lines.pop(0) ex: ['id:name:status:mdisk_count:more_fields']
        # need to use .split to get ['id','name','mdisk_count','more_fields']
        for line in pool_lines:
            line = line.split(':')
            volume_pool = dict(zip(table_headers, line))
            LOG.info("volume_pool %s" % volume_pool)
            total_capacity_gb += (float(volume_pool['capacity'])/units.Gi)
            free_capacity_gb += (float(volume_pool['free_capacity'])/units.Gi)

        super(StorwizeSVCProductDriver, self)._update_volume_stats()
        self._stats['total_capacity_gb'] = total_capacity_gb
        self._stats['free_capacity_gb'] = free_capacity_gb
        self._stats['extent_size'] = self._extent_size
        self._stats['default_volume_type'] = \
            volume_types.get_default_volume_type()
        self._stats.pop('consistencygroup_support', None)

    def _port_conf_generator(self, cmd):
        ssh_cmd = cmd + ['-delim', '!']
        out, err = self._run_ssh(ssh_cmd)

        if not len(out.strip()):
            return
        port_lines = out.strip().split('\n')
        if not len(port_lines):
            return

        header = port_lines.pop(0)
        yield header
        for portip_line in port_lines:
            try:
                port_data = self._get_hdr_dic(header, portip_line, '!')
            except exception.VolumeBackendAPIException:
                with excutils.save_and_reraise_exception():
                    self._log_cli_output_error('_port_conf_generator',
                                               ssh_cmd, out, err)
            yield port_data

    def _get_vdisk_copy_info(self, vdisk):
        ssh_cmd = ['svcinfo', 'lsvdiskcopy', '-delim', '!', vdisk]
        out, err = self._run_ssh(ssh_cmd)

        self._assert_ssh_return(len(out.strip()), '_get_vdisk_copy_info',
                                ssh_cmd, out, err)
        copy_lines = out.strip().split('\n')
        self._assert_ssh_return(len(copy_lines), '_get_vdisk_copy_info',
                                ssh_cmd, out, err)

        header = copy_lines.pop(0)
        ret = {}
        for copy_line in copy_lines:
            try:
                copy_data = self._get_hdr_dic(header, copy_line, '!')
            except exception.VolumeBackendAPIException:
                with excutils.save_and_reraise_exception():
                    self._log_cli_output_error('_get_vdisk_copy_info',
                                               ssh_cmd, out, err)
            ret[copy_data['copy_id']] = copy_data
        return ret

    def _get_vdisk_create_params(self, opts):
        easytier = 'on' if opts['easytier'] else 'off'

        # Set space-efficient options
        if opts['rsize'] == -1:
            params = []
        else:
            params = ['-rsize', '%s%%' % str(opts['rsize']),
                      '-autoexpand', '-warning',
                      '%s%%' % str(opts['warning'])]
            if not opts['autoexpand']:
                params.remove('-autoexpand')

            if opts['compression']:
                params.append('-compressed')
            else:
                params.extend(['-grainsize', str(opts['grainsize'])])

        params.extend(['-easytier', easytier])
        return params

    def _check_vdisk_opts(self, opts):
        # Check that rsize is either -1 or between 0 and 100
        if not (opts['rsize'] >= -1 and opts['rsize'] <= 100):
            raise exception.InvalidInput(
                reason=_('Illegal value specified for storwize_svc_vol_rsize: '
                         'set to either a percentage (0-100) or -1'))

        # Check that warning is either -1 or between 0 and 100
        if not (opts['warning'] >= -1 and opts['warning'] <= 100):
            raise exception.InvalidInput(
                reason=_('Illegal value specified for '
                         'storwize_svc_vol_warning: '
                         'set to a percentage (0-100)'))

        # Check that grainsize is 32/64/128/256
        if opts['grainsize'] not in [32, 64, 128, 256]:
            raise exception.InvalidInput(
                reason=_('Illegal value specified for '
                         'storwize_svc_vol_grainsize: set to either '
                         '32, 64, 128, or 256'))

        # Check that compression is supported
        if opts['compression'] and not self._state['compression_enabled']:
            raise exception.InvalidInput(
                reason=_('System does not support compression'))

        # Check that rsize is set if compression is set
        if opts['compression'] and opts['rsize'] == -1:
            raise exception.InvalidInput(
                reason=_('If compression is set to True, rsize must '
                         'also be set (not equal to -1)'))

        # Check that the requested protocol is enabled
        if opts['protocol'] not in self._state['enabled_protocols']:
            raise exception.InvalidInput(
                reason=_('Illegal value %(prot)s specified for '
                         'storwize_svc_connection_protocol: '
                         'valid values are %(enabled)s')
                % {'prot': opts['protocol'],
                   'enabled': ','.join(self._state['enabled_protocols'])})

        if opts['iogrp'] not in self._state['available_iogrps']:
            raise exception.InvalidInput(
                reason=_('I/O group %(iogrp)d is not valid; available '
                         'I/O groups are %(avail)s')
                % {'iogrp': opts['iogrp'],
                   'avail': ''.join(str(e) for e in
                                    self._state['available_iogrps'])})

    def _execute_command_and_parse_attributes(self, ssh_cmd):
        """Execute command on the Storwize/SVC and parse attributes.

        Exception is raised if the information from the system
        can not be obtained.

        'None' is returned if the object cannot be found.
        """

        LOG.debug('enter: _execute_command_and_parse_attributes: '
                  ' command %s' % str(ssh_cmd))

        try:
            out, err = self._run_ssh(ssh_cmd)
        except processutils.ProcessExecutionError as e:
            if 'CMMVC5753E' in e.stderr:
                # CMMVC5753E: The specified object does not exist or is not a
                #             suitable candidate.
                return None
            else:
                # something bad happened
                LOG.error(_('CLI Exception output:\n command: %(cmd)s\n '
                            'stdout: %(out)s\n stderr: %(err)s') %
                          {'cmd': ssh_cmd,
                           'out': e.stdout,
                           'err': e.stderr})
                raise

        self._assert_ssh_return(len(out),
                                '_execute_command_and_parse_attributes',
                                ssh_cmd, out, err)
        attributes = {}
        for attrib_line in out.split('\n'):
            # If '!' not found, return the string and two empty strings
            attrib_name, foo, attrib_value = attrib_line.partition('!')
            if attrib_name is not None and len(attrib_name.strip()):
                attributes[attrib_name] = attrib_value

        LOG.debug('leave: _execute_command_and_parse_attributes:\n'
                  'command: %(cmd)s\n'
                  'attributes: %(attr)s'
                  % {'cmd': str(ssh_cmd),
                     'attr': str(attributes)})

        return attributes

    def _get_hdr_dic(self, header, row, delim):
        """Return CLI row data as a dictionary indexed by names from header.
        string. The strings are converted to columns using the delimiter in
        delim.
        """

        attributes = header.split(delim)
        values = row.split(delim)
        self._driver_assert(
            len(values) ==
            len(attributes),
            _('_get_hdr_dic: attribute headers and values do not match.\n '
              'Headers: %(header)s\n Values: %(row)s')
            % {'header': str(header),
               'row': str(row)})
        dic = dict((a, v) for a, v in map(None, attributes, values))
        return dic

    def _log_cli_output_error(self, function, cmd, out, err):
        LOG.error(_('%(fun)s: Failed with unexpected CLI output.\n '
                    'Command: %(cmd)s\nstdout: %(out)s\nstderr: %(err)s\n')
                  % {'fun': function, 'cmd': cmd,
                     'out': str(out), 'err': str(err)})

    def _driver_assert(self, assert_condition, exception_message):
        """Internal assertion mechanism for CLI output."""
        if not assert_condition:
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(data=exception_message)

    def _assert_ssh_return(self, test, fun, ssh_cmd, out, err):
        self._driver_assert(
            test,
            _('%(fun)s: Failed with unexpected CLI output.\n '
              'Command: %(cmd)s\n stdout: %(out)s\n stderr: %(err)s')
            % {'fun': fun,
               'cmd': ssh_cmd,
               'out': str(out),
               'err': str(err)})

    def _handle_keyerror(self, function, header):
        msg = (_('Did not find expected column in %(fun)s: %(hdr)s') %
               {'fun': function, 'hdr': header})
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(
            data=msg)

    def _verify_uid_by_vdisk_id(self, vdisk_id, volume):
        """
        Wrapper for _verify_uid when vdisk attributes are not readily
        available - obtains the attributes and passes through.
        """
        attributes = self._get_vdisk_attributes(vdisk_id)

        # Raise an error if the vdisk doesn't exist
        if not attributes:
            raise powervc_exception.SVCVdiskNotFoundException(
                self.endpoint_desc, volume['id'], vdisk_id=vdisk_id)

        return self._verify_uid(attributes, volume)

    def _verify_uid(self, vdisk_attributes, volume):
        """
        Raises an exception if the UID in the passed-in vdisk_attributes is
        not the same at the UID that we have embedded in the restricted
        metadata of the volume.
        """
        expected_uid = \
            volume['restricted_metadata'][RESTRICTED_METADATA_VDISK_UID_KEY]

        current_uid = vdisk_attributes['vdisk_UID']

        if current_uid != expected_uid:
            raise powervc_exception.\
                SVCVdiskMismatchException(self.endpoint_desc,
                                          volume['id'],
                                          vdisk_id=vdisk_attributes['id'],
                                          expected_uid=expected_uid,
                                          current_uid=current_uid)

    def log_driver_status(self):
        """
        Logs internal state of the driver if something looks interesting.
        """

        # If we have waiting connections, or there are in-flight commands,
        # then we display the number of free connections, the number of
        # requests for connections, and the duration and parameters for all
        # currently oustanding SSH commands.
        if self.sshpool.waiting() or self.pending_ssh:
            LOG.info(_("SSH Pool: %d free, %d waiting, in-flight commands:") %
                     (self.sshpool.free(), self.sshpool.waiting()))
            for ssh_id, info in self.pending_ssh.iteritems():
                duration = time.time() - info['time']
                LOG.info(_(" %(duration).2f %(commands)s")
                         % {'duration': duration,
                            'commands': info['command']})

    def _prepare_fcmap_for_deletion(self, map_id, map_attr):
        """ Helper function to process a fcmap to reduce the wait
        time on the dependent_mappings.
        :param map_id: The flashcopy mapping id to handle.
        :map_attr: The lsfcmap output that describes the fcmap attribute
        """
        copyclean_rate = CONF.storwize_fcmap_delete_copy_clean_rate
        copy_rate = map_attr['copy_rate']
        if 'clean_rate' in map_attr:
            clean_rate = map_attr['clean_rate']
        else:
            clean_rate = 0
        if 'autodelete' not in map_attr:
            autodelete = False
        else:
            autodelete = True if map_attr['autodelete'] == 'on' else False

        if int(copy_rate) == 0:
            # Don't do anything let svc driver take care of it.
            LOG.debug(_("Flashcopy mapping copy rate is 0. Copy is not"
                        " in progress."))
            return

        # Speed up the vdisk we want to delete and it's dependent maps
        if int(copy_rate) < int(copyclean_rate):
            # set new cleanup and copy rate.
            self._run_ssh(['svctask', 'chfcmap',
                           '-copyrate',
                           copyclean_rate,
                           '-cleanrate',
                           copyclean_rate,
                           '-autodelete', 'on', map_id])
        elif int(clean_rate) < int(copyclean_rate):
            # set new cleanup rate
            self._run_ssh(['svctask', 'chfcmap',
                           '-cleanrate',
                           copyclean_rate,
                           '-autodelete', 'on', map_id])
        elif not autodelete:
                # wait for fcmap to auto delete
            self._run_ssh(['svctask', 'chfcmap',
                           '-autodelete', 'on', map_id])

    def _dump_vdisk_flashcopy_mapping(self, vdisk_id):
        """
        Dumps the complete list of host currently used by
        the vdisk and the complete list of flashcopy mapping.
        """
        # Dump the list of hosts currently using the vdisk
        ssh_cmd = ['lsvdiskhostmap', '-delim', ',', vdisk_id]
        out, err = self._run_ssh(ssh_cmd, attempts=2)
        LOG.info(_("Hosts mapped to vdisk %(vdisk)s:\n%(out)s") %
                 {'vdisk': vdisk_id, 'out': out})

        # Dump the complete list of flashcopy mappings
        ssh_cmd = ['lsfcmap']
        out, err = self._run_ssh(ssh_cmd, attempts=2)
        LOG.info(_("Existing FlashCopy mappings:\n%(out)s") %
                 {'out': out})
        return out, err

    def _get_vdisk_dep_fc_mappings(self, vdisk):
        """Return all FlashCopy mappings that this vdisk is associated with."""

        def lsvdiskfcmappings():
            """Return the dependent mappings."""
            ssh_cmd = ['svcinfo', 'lsvdiskdependentmaps', '-nohdr',
                       '-delim', '!', vdisk]
            out, error = self._run_ssh(ssh_cmd, attempts=2)
            LOG.info(_("Existing FlashCopy dependancy mappings:\n%(out)s"
                       " %(error)s") %
                     {'out': out, 'error': error})
            mapping_ids = []
            if (len(out.strip())):
                lines = out.strip().split('\n')
                mapping_ids = [line.split('!')[0] for line in lines]
            return mapping_ids

        mapping_ids = self._helpers._get_vdisk_fc_mappings(vdisk)
        mapping_ids = mapping_ids + lsvdiskfcmappings()
        mapping_ids = sorted(set(mapping_ids))
        return mapping_ids

    def _delete_vdisk(self, vdisk_id, force, volume=None):
        """Deletes existing vdisks.

        It is very important to properly take care of mappings before deleting
        the disk:
        1. If no mappings, then it was a vdisk, and can be deleted
        2. If it is the source of a flashcopy mapping and copy_rate is 0, then
           it is a vdisk that has a snapshot.  If the force flag is set,
           delete the mapping and the vdisk, otherwise set the mapping to
           copy and wait (this will allow users to delete vdisks that have
           snapshots if/when the upper layers allow it).
        3. If it is the target of a mapping and copy_rate is 0, it is a
           snapshot, and we should properly stop the mapping and delete.
        4. If it is the source/target of a mapping and copy_rate is not 0, it
           is a clone or vdisk created from a snapshot.  We wait for the copy
           to complete (the mapping will be autodeleted) and then delete the
           vdisk.

        """

        LOG.debug('enter: _delete_vdisk: vdisk %s' % vdisk_id)

        # Try to delete volume only if found on the storage
        vdisk_defined = self._is_vdisk_defined(vdisk_id)
        if not vdisk_defined:
            LOG.info(_('warning: Tried to delete vdisk %s but it does not '
                       'exist.') % vdisk_id)
            return

        # up the copy/clean rates if one was set.
        vdisk_name = self._get_vdisk_name_from_vdisk_id(vdisk_id)
        mapping_ids = []
        try:
            mapping_ids = self._get_vdisk_dep_fc_mappings(vdisk_name)
            for map_id in mapping_ids:
                map_attr = self._helpers.\
                    _get_flashcopy_mapping_attributes(map_id)
                self._prepare_fcmap_for_deletion(map_id, map_attr)
        except Exception as ex:
            LOG.debug("Ignoring prep for deletion failure %s" %
                      six.text_type(ex))

        # raise an warning if flash copy(ies) are still in progress
        if len(mapping_ids) > 0:
            # set visk display name
            if volume:
                display_name = volume['display_name']
            else:
                display_name = vdisk_name
            info = _("A flashcopy associated with volume %(volname)s, volume "
                     "ID: %(volid)s, is in progress. Volume ID %(volid)s will "
                     "be deleted after the flash copy is complete.") \
                % dict(volname=display_name, volid=vdisk_id)
            LOG.info(info)
            # send messages to console
            anotifier = rpc.get_notifier('volume', self.host)
            anotifier.info(context.get_admin_context(),
                           'volume.info', info)

        try:
            self._helpers.delete_vdisk(vdisk_name, force)
        except Exception:
            # SVC doesn't return anything useful so slap
            # the map out there if there is one and go on
            with excutils.save_and_reraise_exception():
                # Dump the list of hosts currently using the vdisk
                self._dump_vdisk_flashcopy_mapping(vdisk_id)

        LOG.debug('leave: _delete_vdisk: vdisk %s' % vdisk_id)

    def create_volume(self, volume):
        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % volume)
                ctxt = context.get_admin_context()
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})
        try:
            opts = self._get_vdisk_params(volume_type_id)
        except Exception as e:
            # For an exception, we need to tag the Volume to show on the UI
            with excutils.save_and_reraise_exception():
                errors = {'Create Failure description': e.message}
                # We need to add the error message to the Volume Metadata
                ctxt = context.get_admin_context()
                meta = db_api.volume_metadata_get(ctxt, volume['id'])
                meta = meta if meta else dict()
                meta.update(errors)
                # Update the Volume Metadata in the DB with the Error Message
                ctxt = context.get_admin_context()
                db_api.volume_metadata_update(
                    ctxt, volume['id'], meta, False)

        candidate_names = [volume['name']]
        vdisk_id, name = self._create_vdisk(candidate_names,
                                            str(volume['size']), 'gb', opts)
        # Place the vdisk ID, name and UID into restricted metadata so that it
        # can be viewed by users and retrieved by this driver when it
        # subsequently performs operations on this volume.
        vdisk = self._get_vdisk_attributes(vdisk_id)
        self._update_restricted_metadata(volume['id'], vdisk)

        if opts['qos']:
            self._helpers.add_vdisk_qos(volume['name'], opts['qos'])

        model_update = None
        ctxt = context.get_admin_context()
        if 'replication' in opts and opts['replication']:
            try:
                model_update = self.replication.create_replica(ctxt, volume)
            except Exception as ex:
                exception_message = (_("Volume replication failed for "
                                       "%(volume_name)s %(volume_id)s "
                                       "with exception. Refer to cinder"
                                       " logs for details.") %
                                     {'volume_name': volume['name'],
                                      'volume_id': volume['id']})
                LOG.error(six.text_type(ex))
                raise exception.VolumeBackendAPIException(
                    data=exception_message)
        if model_update is None:
            model_update = {}
        meta = db_api.volume_metadata_get(ctxt, volume['id'])
        meta = meta if meta else dict()
        meta[METADATA_VOLUME_WWN] = vdisk['vdisk_UID']
        model_update['metadata'] = meta
        return model_update

    def delete_volume(self, volume):
        vdisk_id = self._incorporate_restricted_metadata(volume)

        # If there is no vdisk_id assigned then we probably failed during
        # creation.  Emit a message but do nothing.
        if vdisk_id is None:
            LOG.warn(_("Starting deletion of volume %(volume_id)s.  There is "
                       "no associated vdisk, so no operation will be "
                       "performed on the storage controller.") %
                     {'volume_id': volume['id']})
            return

        LOG.info(_("Starting deletion of volume %(volume_id)s, backed by "
                   "vdisk %(vdisk_id)s") %
                 {'volume_id': volume['id'],
                  'vdisk_id': vdisk_id})

        # Ensure that the UID matches and we don't delete the wrong disk,
        # raises exceptions if the UID doesn't match.
        try:
            self._verify_uid_by_vdisk_id(vdisk_id, volume)
        except powervc_exception.SVCVdiskMismatchException as e:
            # Volume has been deleted and recreated, probably in use by
            # someone else.  Do nothing, but warn.
            LOG.warn(_("delete_volume: vdisk mismatch: %(err)s. "
                       "Performing no operation on the storage controller")
                     % {'err': six.text_type(e)})
            return
        except powervc_exception.SVCVdiskNotFoundException as e:
            # Volume has been deleted, do nothing, but warn.
            LOG.warn(_("delete_volume: vdisk deleted: %(err)s. "
                       "Performing no operation on the storage controller")
                     % {'err': six.text_type(e)})
            return
        self._delete_vdisk(vdisk_id, False, volume)

        return None

    def create_snapshot(self, snapshot):
        ctxt = context.get_admin_context()
        source_vol = self.db.volume_get(ctxt, snapshot['volume_id'])
        src_vdisk_id = self._incorporate_restricted_metadata(source_vol)

        opts = self._get_vdisk_params(source_vol['volume_type_id'])

        tgt_vdisk_names = [snapshot['name']]
        self._create_copy(src_vdisk_id=src_vdisk_id,
                          tgt_vdisk_names=tgt_vdisk_names,
                          full_copy=False,
                          opts=opts,
                          src_id=snapshot['volume_id'],
                          from_vol=True)

        return None

    def delete_snapshot(self, snapshot):
        vdisk_id = self._get_vdisk_id_from_vdisk_name(snapshot['name'])

        # If the disk didn't exist for some reason, just return without trying
        # to initiate deletion.
        if vdisk_id is None:
            return

        self._delete_vdisk(vdisk_id, False)

    def create_volume_from_snapshot(self, volume, snapshot):
        if volume['size'] != snapshot['volume_size']:
            exception_message = (_('create_volume_from_snapshot: '
                                   'Source and destination size differ.'))
            raise exception.VolumeBackendAPIException(data=exception_message)

        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % volume)
                ctxt = context.get_admin_context()
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})
        opts = self._get_vdisk_params(volume_type_id)
        src_vdisk_id = self._get_vdisk_id_from_vdisk_name(snapshot['name'])

        # If the snapshot vdisk doesn't exist for some reason, log error and
        # raise an exception
        if src_vdisk_id is None:
            exception_msg = (
                _('create_volume_from_snapshot: '
                  ' Source vdisk %s does not exist')
                % snapshot['name'])
            LOG.error(exception_msg)
            raise exception.VolumeNotFound(exception_msg,
                                           volume_id=snapshot['id'])

        # Generate a preference-ordered list of names for the new volume
        candidate_names = [volume['name']]

        vdisk_id, name = self._create_copy(src_vdisk_id=src_vdisk_id,
                                           tgt_vdisk_names=candidate_names,
                                           full_copy=True,
                                           opts=opts,
                                           src_id=snapshot['id'],
                                           from_vol=False)

        # Place the vdisk ID, name and UID into restricted metadata so that it
        # can be viewed by users and retrieved by this driver when it
        # subsequently performs operations on this volume.
        vdisk = self._get_vdisk_attributes(vdisk_id)
        self._update_restricted_metadata(volume['id'], vdisk)

        if opts['qos']:
            self._helpers.add_vdisk_qos(volume['name'], opts['qos'])

        model_update = None
        ctxt = context.get_admin_context()
        if 'replication' in opts and opts['replication']:
            try:
                model_update = self.replication.create_replica(ctxt, volume)
            except Exception as ex:
                exception_message = (_("Volume replication failed for "
                                       "%(volume_name)s %(volume_id)s "
                                       "with exception. Refer to cinder"
                                       " logs for details.") %
                                     {'volume_name': volume['name'],
                                      'volume_id': volume['id']})
                LOG.error(six.text_type(ex))
                raise exception.VolumeBackendAPIException(
                    data=exception_message)
        if model_update is None:
            model_update = {}
        meta = db_api.volume_metadata_get(ctxt, volume['id'])
        meta = meta if meta else dict()
        meta[METADATA_VOLUME_WWN] = vdisk['vdisk_UID']
        model_update['metadata'] = meta
        return model_update

    def create_cloned_volume(self, tgt_volume, src_volume):
        if src_volume['size'] != tgt_volume['size']:
            exception_message = (_('create_cloned_volume: '
                                   'Source and destination size differ.'))
            LOG.error(exception_message)
            raise exception.VolumeBackendAPIException(data=exception_message)

        volume_type_id = tgt_volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                tgt_volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % tgt_volume)
                ctxt = context.get_admin_context()
                self.db.volume_update(ctxt, tgt_volume['id'],
                                      {'volume_type_id': volume_type_id})
        opts = self._get_vdisk_params(volume_type_id)

        src_vdisk_id = self._incorporate_restricted_metadata(src_volume)

        # Ensure that the UID matches and we don't clone the wrong disk,
        # raise exception if the UID doesn't match.
        self._verify_uid_by_vdisk_id(src_vdisk_id, src_volume)

        candidate_names = [tgt_volume['name']]

        vdisk_id, name = self._create_copy(src_vdisk_id=src_vdisk_id,
                                           tgt_vdisk_names=candidate_names,
                                           full_copy=True,
                                           opts=opts,
                                           src_id=src_volume['id'],
                                           from_vol=True)

        # Place the vdisk ID, name and UID into restricted metadata so that it
        # can be viewed by users and retrieved by this driver when it
        # subsequently performs operations on this volume.
        vdisk = self._get_vdisk_attributes(vdisk_id)
        self._update_restricted_metadata(tgt_volume['id'], vdisk)

        if opts['qos']:
            self._helpers.add_vdisk_qos(tgt_volume['name'], opts['qos'])

        model_update = None
        ctxt = context.get_admin_context()
        if 'replication' in opts and opts['replication']:
            try:
                model_update = self.replication.create_replica(
                    ctxt, tgt_volume)
            except Exception as ex:
                exception_message = (_("Volume replication failed for "
                                       "%(volume_name)s %(volume_id)s "
                                       "with exception. Refer to cinder"
                                       " logs for details.") %
                                     {'volume_name': tgt_volume['name'],
                                      'volume_id': tgt_volume['id']})
                LOG.error(six.text_type(ex))
                raise exception.VolumeBackendAPIException(
                    data=exception_message)
        if model_update is None:
            model_update = {}
        meta = db_api.volume_metadata_get(ctxt, tgt_volume['id'])
        meta = meta if meta else dict()
        meta[METADATA_VOLUME_WWN] = vdisk['vdisk_UID']
        model_update['metadata'] = meta
        return model_update

    """
    For debug purposes, we maintain a map of SSH connection objects to
    information about the command that they are currently executing.
    """
    pending_ssh = {}

    def _run_ssh_local(self, cmd_list, check_exit_code=True, attempts=1):
        """
        override _run_ssh() method from san.SanDriver.
        It is important to set connection timeout to None
        if development_ssh_san_conn is enabled to close ssh
        connection per get(). Otherwise, keepalive without
        timeout on the ssh transport will cause the service
        greenthread to be stuck on the join during close()
        when put() is invoked.
        """
        if CONF.development_ssh_san_conn:
            LOG.debug("PowerVCSSHPool: close ssh connection after _run_ssh")
            timeout = None
            # periodic task and volume driver ssh execution may
            # be launched at the same time which could cause paramiko
            # authentication error due to the attempt to create
            # multiple paramiko ssh connection at the sametime.
            attempts = attempts if attempts > 3 else 3
        else:
            timeout = self.configuration.ssh_conn_timeout

        if not self.sshpool:
            password = self.configuration.san_password
            privatekey = self.configuration.san_private_key
            min_size = self.configuration.ssh_min_pool_conn
            max_size = self.configuration.ssh_max_pool_conn
            self.sshpool = PowerVCSSHPool(self.configuration.san_ip,
                                          self.configuration.san_ssh_port,
                                          timeout,
                                          self.configuration.san_login,
                                          password=password,
                                          privatekey=privatekey,
                                          min_size=min_size,
                                          max_size=max_size)

        # cinderutils.check_ssh_injection(cmd_list)
        command = ' '.join(cmd_list)

        last_exception = None
        pauses_between_attempts = 0
        try:
            while attempts > 0:
                attempts -= 1
                try:
                    with self.sshpool.item() as ssh:
                        # Record the command that we are about to execute,
                        # along with start time (in seconds)
                        self.pending_ssh[id(ssh)] = {'command': command,
                                                     'time': time.time()}
                        (out, err) = \
                            processutils.ssh_execute(
                                ssh, command,
                                check_exit_code=check_exit_code)

                        # Remove the command from the list of commands we're
                        # executing.
                        del self.pending_ssh[id(ssh)]
                        return (out, err)
                except Exception as e:
                    LOG.info(e)
                    last_exception = e
                    if attempts > 0:
                        sleep_time = \
                            (random.randint(CONF.san_retry_min_delay,
                                            CONF.san_retry_max_delay) +
                             (pauses_between_attempts *
                              CONF.san_retry_max_delay))
                        LOG.info(_("Waiting %(seconds)d seconds before next "
                                   "attempt...")
                                 % {'seconds': sleep_time})
                        greenthread.sleep(sleep_time)
                        pauses_between_attempts += 1
            try:
                raise processutils.ProcessExecutionError(
                    exit_code=last_exception.exit_code,
                    stdout=last_exception.stdout,
                    stderr=last_exception.stderr,
                    cmd=last_exception.cmd)
            except AttributeError:
                raise processutils.ProcessExecutionError(
                    exit_code=-1,
                    stdout="",
                    stderr="Error running SSH command",
                    cmd=command)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.warn(_("Error running SSH command: %s") % command)


def add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in range(0, len(wwpn), 2)]).lower()


def strip_colons(wwpn):
    wwpn = wwpn.lower()
    if len(wwpn) > 16:
        return wwpn.replace(":", "")
    else:
        return wwpn


class FcmapWaiter(object):

    """ A flashcopy mapping event waiter. It will be blocked on the
    eventlet event created during vdisk delete operation.The fcmap waiter
    will be queued to the fcmap delete queue and will be polled by the
    fcmap polling timer every 30 seconds(configurable). Once the fcmap
    is removed for a given vdisk, the notification will be sent from the
    timer to wake up the vdisk delete thread that is blocked. deque, which
    holds the fcmap waiter, is thread safe. Once the waiter is created by
    the vdisk delete thread and enqueued, the content of the waiter will only
    be updated by the fcmap timer thread and each vdisk delete thread has
    its own waiter. This machanism makes it lock free. """

    def __init__(self, vdisk_id, allow_snaps=True):
        self.vdisk_id = vdisk_id
        self.priority = 0
        self.fcmap_ids = []
        self.exc = None
        self.event = Event()
        self.enqueue_tb = time.time()
        self.allow_snaps = allow_snaps

    def update(self, fcmap_ids, pri, exc=None):
        """ function to update the fcmap waiter properties. It is called
        by both fcmap handler timer and vdisk delete thread(not concurrently).
        """
        self.fcmap_ids = fcmap_ids
        self.priority = pri
        self.exc = exc

    def wait(self):
        """ function called by vdisk delete thread to wait during
        flashcopy mapping processing"""
        return self.event.wait()

    def is_ready(self):
        """ helper function called by fcmap handler timer thread. It makes
        sure the fcmap for a given vdisk delete operation is done."""
        return True if len(self.fcmap_ids) == 0 else False

    def done(self, ret=True):
        """function called by fcmap handler timer thread only. It wakes up
        the vdisk delete thread that is waiting for the fcmap processing."""
        if self.exc:
            self.event.send_exception(self.exc)
        else:
            self.event.send(ret)


class CLIResponse(object):

    '''Parse SVC CLI output and generate iterable'''

    def __init__(self, raw, delim='!'):
        super(CLIResponse, self).__init__()
        self.raw = raw
        self.delim = delim
        self.result = self._parse()

    def select(self, *keys):
        for a in self.result:
            vs = []
            for k in keys:
                v = a.get(k, None)
                if isinstance(v, basestring):
                    v = [v]
                if isinstance(v, list):
                    vs.append(v)
            if len(vs) > 1:
                for item in zip(*vs):
                    yield item
            else:
                for item in vs[0]:
                    yield item

    def __getitem__(self, key):
        return self.result[key]

    def __iter__(self):
        for a in self.result:
            yield a

    def __len__(self):
        return len(self.result)

    def _parse(self):
        def get_reader(content, delim):
            if content:
                for line in content.lstrip().splitlines():
                    line = line.strip()
                    if line:
                        yield line.split(delim)
                    else:
                        yield []
            else:
                yield []

        reader = get_reader(self.raw[0], self.delim)
        result = []

        cur = dict()
        for row in reader:
            if row:
                CLIResponse.append_dict(cur, row[0], ' '.join(row[1:]))
            elif cur:  # start new section
                result.append(cur)
                cur = dict()
        if cur:
            result.append(cur)
        return result

    @staticmethod
    def append_dict(dict_, key, value):
        key, value = key.strip(), value.strip()
        obj = dict_.get(key, None)
        if obj is None:
            dict_[key] = value
        elif isinstance(obj, list):
            obj.append(value)
            dict_[key] = obj
        else:
            dict_[key] = [obj, value]
        return dict_


@staticmethod
def factory(driver):
    """Use replication methods for the requested mode."""
    return PowerVCStorwizeSVCReplicationStretchedCluster(driver)

storwize_rep.StorwizeSVCReplication.factory = factory


class PowerVCStorwizeSVCReplicationStretchedCluster(
        storwize_rep.StorwizeSVCReplicationStretchedCluster):

    def create_replica(self, ctxt, volume, vol_type=None):
        # if vol_type is None, use the source volume type
        dest_pool = ''
        if vol_type is None:
            vol_type = volume['volume_type_id']
        vol_type_id = vol_type
        vol_type = volume_types.get_volume_type(ctxt, vol_type)
        # get it from vol_type/storage template
        dest_pool = volume_types.get_volume_type_extra_specs(
            vol_type_id, 'drivers:mirror_pool')
        # raise pvc exception if destination pool is not available
        if (dest_pool is None or
                (isinstance(dest_pool, six.string_types) and
                    (dest_pool == '')) or
                not isinstance(dest_pool, six.string_types)):
            exception_message = (_("Missing storwize mirror destination"
                                   "pool information (drivers:"
                                   "mirror_pool). "
                                   "Review the configuration"
                                   "storage template used."))
            raise exception.VolumeBackendAPIException(
                data=exception_message)
        self.driver.add_vdisk_copy(volume['name'], dest_pool, vol_type)
        vol_update = {'replication_status': 'copying'}

        return vol_update
