# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import base64

from cinder import context
from cinder import exception
from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import discovery_driver
from cinder.zonemanager import utils as fczm_utils
from powervc_cinder.zonemanager import powervc_utils as pvc_fczm_utils
# don't do any zoning requests
fczm_utils.AddFCZone = pvc_fczm_utils.ignore_zoning_request
fczm_utils.RemoveFCZone = pvc_fczm_utils.ignore_zoning_request
from cinder.volume.drivers.ibm import xiv_ds8k
from cinder.volume import volume_types
from powervc_cinder.volume.drivers.xiv_ds8k import common_driver
from powervc_cinder.volume.drivers.xiv_ds8k.common_driver\
    import RESTRICTED_METADATA_VDISK_ID_KEY,\
    RESTRICTED_METADATA_VDISK_UID_KEY,\
    RESTRICTED_METADATA_VDISK_NAME_KEY,\
    RESTRICTED_METADATA_BYTE_SIZE_KEY,\
    RESTRICTED_METADATA_NAA,\
    METADATA_VOLUME_WWN

from powervc_cinder import _
from oslo_log import log as logging
from oslo_utils import excutils

from oslo_utils import importutils
from powervc_cinder import exception as powervc_exception
from cinder import db as db_api

from xiv_ds8k_openstack.utils.powervc_helper import XIVPowerVCHelpers

LOG = logging.getLogger(__name__)


###############################################
#      Temporary Workaround for XIV Bug       #
###############################################
def _get_xiv_exception_class(self):
    """Temporary Workaround for XIV bug loading exception class"""
    return exception.CinderException

XIVPowerVCHelpers._get_exception = _get_xiv_exception_class


class XIVProductDriver(common_driver.XIVDS8KDriver):

    def __init__(self, *args, **kwargs):
        super(XIVProductDriver, self).__init__(*args, **kwargs)
        self.endpoint_desc = "XIVProductDriver provider"

    def _get_xiv_helpers(self):
        if not hasattr(self, '_xiv_helpers'):
            proxy = importutils.import_class(
                'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
            self._xiv_helpers = proxy()

        return self._xiv_helpers

    def _build_default_opts(self):
        default_opts = {'storage_pool': self.configuration.san_clustername}
        return default_opts

    def create_volume(self, volume):
        self._add_default_volume_type(volume)
        vol = super(XIVProductDriver, self).create_volume(volume)
        # Add restricted metadata to volume
        wwn = self._get_wwn(volume)
        return self._update_metadata(vol, volume['id'], volume['name'],
                                     volume['name'], volume['name'], wwn)

    def _get_wwn(self, volume):
        volume_type_id = volume.get('volume_type_id')
        pool = volume_types.get_volume_type_extra_specs(volume_type_id,
                                                        "drivers:storage_pool")
        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        helpers = self._get_xiv_helpers()
        context = {'user': x_userid, 'password': x_password,
                   'address': x_ip, 'pool': pool, 'volume': volume['name']}
        specific_volume = helpers.get_volume(context)
        wwn = specific_volume['wwn']
        LOG.info(_('getting wwn: %s') % wwn)
        return wwn

    def create_cloned_volume(self, tgt_volume, src_volume):
        """
        Overrides the superclass create_cloned_volume. Adds
        restricted metadata information to the cloned volume
        and sets default volume type if one is not specified
        """
        self._add_default_volume_type(tgt_volume)
        s = super(XIVProductDriver, self)
        vol = s.create_cloned_volume(tgt_volume, src_volume)
        wwn = self._get_wwn(tgt_volume)
        return self._update_metadata(vol, tgt_volume['id'],
                                     tgt_volume['name'], tgt_volume['name'],
                                     tgt_volume['name'], wwn)

    def discover_volumes(self, context, filters=None):
        """Method to Retrieve a list of volumes for a host
        Currently supported filters:

         'host_refs': [ WWPN, ... ]
           Returns only volumes that are mapped to at least one of the WWPNS in
           the specified list.  Matching is case-insensitive.
        """
        LOG.debug("discover_volumes filters %s" % filters)
        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        helpers, volumes = (self._get_xiv_helpers(), dict())
        xiv_context = {'user': x_userid, 'password': x_password,
                       'address': x_ip}
        for pool in helpers.get_pools_list(xiv_context).values():
            context2 = dict(xiv_context, pool=pool['name'])
            volumes.update(helpers.get_volume_list(context2))
        # The XIV get_volume_list helper method does not return volume ids
        # and so we use the volume name as the volume ID.
        for vol in volumes.values():
            vol['id'] = vol['name']
        connection_info = helpers.get_hosts_list(xiv_context)
        return_volumes = self._discover_volumes(context, volumes,
                                                connection_info,
                                                filters=filters)
        LOG.debug("discover_volumes return %s" % return_volumes)
        return return_volumes

    def check_for_deleted_volumes(self, context, volumes):
        """Determines which of the Volumes no longer exist on the Provider"""
        # Run the vol_list command to query the list of volumes on the XIV
        existvols = self.xiv_ds8k_proxy._call_xiv_ds8k_cli("vol_list").as_list
        existvol_ids = [v['name'] for v in existvols]
        return self._check_for_deleted_volumes(context, volumes, existvol_ids)

    def get_storage_metadata(self):
        """Returns the additional information that PowerVC uses to populate
        various UI elements.
        """
        compression = False

        pools = self.xiv_ds8k_proxy._call_xiv_ds8k_cli("pool_list")

        def transform_pool_structure(pool):
            return {'name': pool.name,
                    'capacity': str(pool.soft_size) + " GB",
                    'free_capacity': str(pool.empty_space_soft) + " GB"}
        volume_pools = map(transform_pool_structure, pools)

        return {'volume_pools': volume_pools,
                'compression_supported': compression}

    @pvc_fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        """
        Calls the parent class to mask the volume on the storage controller,
        and here performs additional processing to work out which zones should
        be created on the switches.

        For NPIV, the virtual fibre channel ports will not yet be logged in,
        but we need to know which fabrics which ports will appear on, so that
        we know which fabrics to create the zones on.  For this reason, the
        connector must also include a data structure that maps physical WWPNs
        (which will be logged in) to the virtual WWPNs that they will provide.
        To support live partition mobility, we expect each physical WWPN to be
        providing two virtual WWPNs.

        The expected structure of a fibre channel connector is:

        {
            'host': hostname,
            'wwpns': [ list of virtual WWPNs ]  <-- this is used for masking
            'phy_to_virt_initiators': {phy_wwpn_X: [virt_wwpn_a, virt_wwpn_b],
                                       phy_wwpn_Y: [virt_wwpn_c, virt_wwpn_d]
                                      }
        }

        The expected structure of an iSCSI connector is:

        {
          'host': hostname,
          'initiator': initiator  <-- IQN format
        }
        """

        # Ensure that the passed-in data structure is correct.
        expected_fc_keys = set(['host', 'wwpns'])
        expected_iscsi_keys = set(['host', 'initiator'])
        if not (expected_fc_keys.issubset(connector) |
                expected_iscsi_keys.issubset(connector)):
            err_msg = (_('The connector %(connector)s does not contain the '
                         'required keys, which are %(expected_fc_keys)s.'
                         'or %(expected_iscsi_keys)s') %
                       {'connector': connector,
                        'expected_fc_keys': expected_fc_keys,
                        'expected_iscsi_keys': expected_iscsi_keys})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        # Need to sanitize the host, since XIV doesn't handle many characters
        if connector.get('host'):
            connector['host'] = sanitize_name(connector['host'])
        # Perform the masking on the storage controller
        s = super(XIVProductDriver, self)
        info = s.initialize_connection(volume, connector)
        info['data']['vendor'] = "IBM"
        try:
            if (connector['connection_type'] and
                    'pv_vscsi' in connector['connection_type']):
                ctxt = context.get_admin_context()
                metadata = db_api.volume_metadata_get(ctxt, volume['id'])
                info['data']['pg83NAA'] = metadata['volume_wwn']
        except Exception as e:
            # don't worry if we can't add code page 83
            pass

        # A new field has been added to info data to return all
        # target wwpns.  We use that to poplulate target_wwn if it
        # returns empty from initialize_connection()
        if info['data']['all_storage_wwpns'] and \
           not info['data']['target_wwn']:
            info['data']['target_wwn'] = \
                info['data']['all_storage_wwpns']

        # LUN ID SAM format standard uses 8 bytes, if the lun id is greater
        # than 255(1 byte) then format needs to be changed for the top 2 bits
        # of lun id. Top two bits should be set to 0x4000 (16384) followed by
        # 8 byte Lun id
        # for eg if lun=256, will be passed 0x4100 = (0x4000 + 0x100) down
        # we will update the value to use the SAM format for setting the
        # So the lun id value will be set as lun id = 16384 + lun_id
        if info['data']['target_lun'] > 255:
            info['data']['target_lun'] += 16384

        return {'driver_volume_type': 'fibre_channel', 'data': info['data']}

    @pvc_fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):

        fcwwpns = []
        for fcwwpn in connector['wwpns']:
            fcwwpns.append(fcwwpn.lower())
            connector['wwpns'] = fcwwpns

        LOG.info(_("Terminating connection with parameters %(connector)s" %
                   {'connector': connector}))

        # Perform the masking on the storage controller
        s = super(XIVProductDriver, self)
        info = s.terminate_connection(volume, connector, **kwargs)

        return info

    def check_volume_health(self, volume):
        """
        Throws exceptions if something is wrong with the passed-in volume.

        Checks for existence of the underlying vdisk
        """
        # The list of volumes from the call to get_volume_list()
        # is used a couple of times.  Collect the info here to
        # avoid redundant calls.
        # Also create a new internal helper method to simplify the call.
        volumes = self._get_volume_list()

        vdisk_id = self._incorporate_restricted_metadata(volume, volumes)

        if not self._check_vdisk_exists(vdisk_id, volumes):
            raise powervc_exception.XIVVdiskNotFoundException(
                self.endpoint_desc,
                volume['id'], vdisk_id=vdisk_id)

    def _check_vdisk_exists(self, vdisk_id, volumes):
        for key, vol in volumes.items():
            if vol['existing_ref'] == vdisk_id:
                return True
        return False

    def _incorporate_restricted_metadata(self, volume_ref, volumes):
        """Query the Cinder database to retrieve any access-restricted
           metadata and attach it to the volume data.
           Returns the vdisk id, which most of the time is what we were after.
           If there is no restricted metadata, tries to find the disk and
           creates it.
        """
        ctxt = context.get_admin_context()
        metadata = powervc_db_api.\
            volume_restricted_metadata_get(ctxt, volume_ref['id'])
        # If the vdisk ID is not in the metadata, this could be because of an
        # upgrade from a version that didn't use it.  We try to find it by
        # looking for a volume on the SAN with a matching name
        if RESTRICTED_METADATA_VDISK_ID_KEY not in metadata:
            # Previous versions would have used this as the name on the SAN
            vdisk = self._find_vdisk_by_name(volume_ref['name'], volumes)
            if vdisk is not None:
                self._update_restricted_metadata(volume_ref['id'], vdisk)
                metadata = powervc_db_api.\
                    volume_restricted_metadata_get(ctxt, volume_ref['id'])
            else:
                return None

        volume_ref['restricted_metadata'] = metadata
        return metadata[RESTRICTED_METADATA_VDISK_ID_KEY]

    def _find_vdisk_by_name(self, vdisk_name, volumes):
        """ Look through the list of volumes for a match.
            For XIV, the only attribute needed is the name which
            will be used for the metadata update.
        """
        for key, vol in volumes.items():
            if vol['existing_ref'] == vdisk_name:
                attributes = {}
                attributes['name'] = vdisk_name
                return attributes
        return None

    def _update_restricted_metadata(self, volume_id, vdisk):
        """Perform a database update to associate the specified information
           with the specified volume.
        """
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: vdisk['name'],
                    RESTRICTED_METADATA_VDISK_NAME_KEY: vdisk['name'],
                    RESTRICTED_METADATA_VDISK_UID_KEY: vdisk['name']
                    }
        ctxt = context.get_admin_context()
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(ctxt, volume_id,
                                                        metadata)

    def _get_volume_list(self):
        """ Helper method to consolidate info for the
            XIV helper method call to get_volume_list()
            Returns the list of volumes.
        """
        x_userid = self.configuration.san_login
        x_password = self.configuration.san_password
        x_password = self._decode_password(x_password)
        x_ip = self.configuration.san_ip
        x_pool = self.configuration.san_clustername
        helpers = self._get_xiv_helpers()
        context = {'user': x_userid, 'password': x_password,
                   'address': x_ip, 'pool': x_pool}
        volumes = helpers.get_volume_list(context)
        return volumes

    def _b_to_volume_size(self, byte_size):
        # XIV has historically used 10^9 for this byte to volume size
        # conversion during onboarding, which is not what the community
        # has standardized on.

        # The XIV itself seems to be using GB in its backend rather than GiB.
        # The following investigation was done to determine if this code,
        # used in onboarding should move to 2^30 instead of 10^9 and it was
        # determined that it will remain 10^9:

        # In XIV GUI specify a volume be created of size 16 GiB.
        # The volume is created and the XIV GUI displays 17 GB.
        # The XIV GUI shows size in blocks to be 33554432 blocks.
        # The helper methods 'byte size' returned is 17208180736.
        # This gives a block size of 512.84375 which is definitely not right,
        # and likely points to some rounding up for the byte-size by XIV.

        # In the XIV GUI specify a volume be created with a size of 17GB.
        # The volume is created and the XIV GUI displays 17GB.
        # The XIV GUI shows size in blocks to be 33609728 blocks.
        # The helper methods byte size returned is 17208180736, which is the
        # same value as the 16GiB volume.
        # This give a block size of exactly 512.
        #
        # Given the above, the division of byte size 17208180736 by 10^9
        # to give the integer value of 17GB for the volume is the correct
        # thing to do.
        return byte_size // (10 ** 9)


def add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in range(0, len(wwpn), 2)]).lower()


def strip_colons(wwpn):
    wwpn = wwpn.lower()
    if len(wwpn) > 16:
        return wwpn.replace(":", "")
    else:
        return wwpn


###################################################
#       Override Create Volume on XIV Proxy       #
###################################################
# Since the XIV Driver doesn't currently provide a way to specify a
# different pool on create_volume than what is the default specified
# when registered, we need to override the proxy to allow that pool
# to be specified until the XIV team provides support in the driver.
__import__('powervc_oslo.xiv_xsf')
gigabytes_to_blocks = importutils.import_class(
    'xiv_ds8k_openstack.gigabytes_to_blocks')
_trace_time = importutils.import_class(
    'xiv_ds8k_openstack.nova_proxy._trace_time')
XIVNovaProxy = importutils.import_class(
    'xiv_ds8k_openstack.xiv_nova_proxy.XIVNovaProxy')
SystemOutOfSpaceError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.SystemOutOfSpaceError')
PoolOutOfSpaceError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.PoolOutOfSpaceError')
XCLIError = importutils.import_class(
    'xsf.xiv.client.xcli.errors.XCLIError')
# Load the various constant string literals that we need to use from XIV
CREATE_VOLUME_BASE_ERROR = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_BASE_ERROR')
CREATE_VOLUME_SYSTEM_OUT_OF_SPACE = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_SYSTEM_OUT_OF_SPACE')
CREATE_VOLUME_POOL_OUT_OF_SPACE = importutils.import_class(
    'xiv_ds8k_openstack.strings.CREATE_VOLUME_POOL_OUT_OF_SPACE')
# We need to import the sanitize name to use for replacing the host
sanitize_name = importutils.import_class(
    'powervc_cinder.db.sqlalchemy.models._sanitize_volume_name')


class XIVProductProxy(XIVNovaProxy):

    @_trace_time
    def create_volume(self, volume):
        """Override the create_volume to specify the pool name"""
        size = gigabytes_to_blocks(float(volume['size']))
        pool = self._get_pool_from_extra_specs(volume)
        # If a Pool override wasn't specified, use the default configured
        if not pool:
            pool = self.xiv_ds8k_info.get('xiv_ds8k_vol_pool')
        try:
            self._call_xiv_ds8k_cli('vol_create', vol=volume['name'],
                                    size_blocks=size, pool=pool)
        # If the System is Out-of-Space then log and raise that error
        except SystemOutOfSpaceError:
            self._log('info', CREATE_VOLUME_BASE_ERROR,
                      details=CREATE_VOLUME_SYSTEM_OUT_OF_SPACE)
            raise self._get_exception()(
                message=CREATE_VOLUME_BASE_ERROR %
                {'details': CREATE_VOLUME_SYSTEM_OUT_OF_SPACE})
        # If the Pool is Out-of-Space then log and raise that error
        except PoolOutOfSpaceError:
            err = CREATE_VOLUME_POOL_OUT_OF_SPACE % {'pool': pool}
            self._log('info', CREATE_VOLUME_BASE_ERROR, details=err)
            raise self._get_exception()(
                message=CREATE_VOLUME_BASE_ERROR % {'details': err})
        # If it is a Generic CLI Error, then just log and raise general
        except (XCLIError) as e:
            details = self._get_code_and_status_or_message(e)
            self._log('info', CREATE_VOLUME_BASE_ERROR, details=details)
            raise self.meta.exception.VolumeBackendAPIException(
                data=CREATE_VOLUME_BASE_ERROR % {'details': details})
        return {}

    def _get_xiv_helpers(self):
        if not hasattr(self, '_xiv_helpers'):
            proxy = importutils.import_class(
                'xiv_ds8k_openstack.utils.powervc_helper.XIVPowerVCHelpers')
            self._xiv_helpers = proxy()

        return self._xiv_helpers

    def _update_stats(self):
        """Override the Volume Stats to return all of the Pools"""
        total, free = (0, 0)
        # Call the Parent Class to retrieve most of the Statistics
        super(XIVProductProxy, self)._update_stats()
        # Retrieve the Capacity from each of the Pools and use the total
        x_userid = self.driver.configuration.safe_get('san_login')
        x_password = self.driver.configuration.safe_get('san_password')
        x_password = base64.b64decode(x_password)
        x_ip = self.driver.configuration.safe_get('san_ip')
        context = {'user': x_userid, 'password': x_password, 'address': x_ip}
        helpers = self._get_xiv_helpers()

        for pool in helpers.get_pools_list(context).values():
            total += int(pool['soft_size'])
            free += int(pool['free_size'])
        # Now we can actually override the Free/Total Capacity values
        self.meta.stat['total_capacity_gb'] = total
        self.meta.stat['free_capacity_gb'] = free

    def _get_pool_from_extra_specs(self, volume):
        """Helper method to get the Pool Name from the Extra Specs"""
        if volume and volume.get('volume_type_id') is not None:
            specs = volume_types.\
                get_volume_type_extra_specs(volume['volume_type_id'])
            return specs.get('drivers:storage_pool')
        return None
