# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
import base64
from cinder import context
from cinder import db as db_api
from cinder import rpc
from powervc_cinder.volume import discovery_driver
from cinder.volume.drivers.ibm import xiv_ds8k
from cinder.volume import volume_types
from powervc_cinder import _
from powervc_cinder.db import api as powervc_db_api
from oslo_log import log as logging

RESTRICTED_METADATA_VDISK_ID_KEY = "vdisk_id"
RESTRICTED_METADATA_VDISK_UID_KEY = "vdisk_uid"
RESTRICTED_METADATA_VDISK_NAME_KEY = "vdisk_name"
RESTRICTED_METADATA_BYTE_SIZE_KEY = "byte_size"
RESTRICTED_METADATA_NAA = "wwn"
METADATA_VOLUME_WWN = "volume_wwn"

LOG = logging.getLogger(__name__)


class XIVDS8KDriver(xiv_ds8k.XIVDS8KDriver,
                    discovery_driver.VolumeDiscoveryDriver):
    """A common parent class for PowerVC XIV and DS8K volume drivers to allow
    code sharing between the drivers
    """

    def __init__(self, *args, **kwargs):
        super(XIVDS8KDriver, self).__init__(*args, **kwargs)

    def get_default_vol_type(self):
        """Returns the default volume type from the .conf file"""
        vtn = self.configuration.default_volume_type
        return vtn.decode('utf-8') if vtn else vtn

    def _decode_password(self, password):
        """Shared Helper Method to specify the deCryption method to use"""
        return base64.b64decode(password)

    def _encode_password(self, password):
        """Shared Helper Method to specify the encryption method to use"""
        return base64.b64encode(password)

    def _discover_volumes(self, context, volumes, connection_info,
                          filters=None):
        """Helper method for the discover volumes method.
        """
        LOG.debug("_discover_volumes volumes %s" % volumes)
        LOG.debug("_discover_volumes connection_info %s" % connection_info)
        LOG.debug("_discover_volumes filters %s" % filters)
        # If we are filtering, create a set from the WWPN list to make
        # filtering easier.  Convert the WWPNs to upper-case for consistency
        filter_set = None
        upper_refs = None
        if filters is not None and 'host_refs' in filters:
            upper_refs = [ref.upper() for ref in filters['host_refs']]
            filter_set = set(upper_refs)

        # If we are filtering get the volume references to get the k2udid
        vol_refs = None
        if filters is not None and 'vol_refs' in filters:
            vol_refs = filters['vol_refs']

        return_volumes = []
        for vol in volumes.values():
            volume = {}
            volume['name'] = vol['name']
            volume['wwn'] = vol['wwn']
            volume['storage_pool'] = vol['pool']
            volume['status'] = 'available'
            volume['size'] = self._b_to_volume_size(int(vol['size_in_bytes']))
            vdisk_uid = vol['id']
            volume['restricted_metadata'] = {
                RESTRICTED_METADATA_NAA: vol['wwn'],
                RESTRICTED_METADATA_VDISK_ID_KEY: vol['id'],
                RESTRICTED_METADATA_VDISK_UID_KEY: vol['id'],
                RESTRICTED_METADATA_VDISK_NAME_KEY: vol['name'],
                RESTRICTED_METADATA_BYTE_SIZE_KEY: vol['size_in_bytes']
            }
            volume['metadata'] = {METADATA_VOLUME_WWN: vol['wwn'],
                                  'storage_pool': vol['pool']}
            volume['support'] = {"status": "supported"}
            volume['mapped_wwpns'] = vol['wwpns']
            if filter_set is not None:
                rmt_wwpns = set([wwpn.upper() for wwpn in vol['wwpns']])

            if len(vol['wwpns']) > 0:
                volume['status'] = 'in-use'
                volume['support'] = {
                    "status": "not_supported",
                    "reasons": [
                        _('This volume is not a candidate for management '
                          'because it is already attached to a virtual '
                          'machine.  To manage this volume with PowerVC, '
                          'you must bring the virtual machine under '
                          'management.  Select to manage the virtual '
                          'machine that has the volume attached.  The '
                          'attached volume will be automatically included '
                          'for management.')
                    ]
                }
                conninfo = {}
                for k, v in connection_info.items():
                    hwwpns = v['host_wwpns']
                    for w in vol['wwpns']:
                        if w in hwwpns:
                            conninfo[k] = v
                            volume['connection_info'] = conninfo

            self._check_supported_size(volume, vol)
            # Store the volume k2udid
            if vol_refs:
                for vol_idx in vol_refs:
                    wwn = vol['wwn'].upper()
                    if wwn in vol_idx['uid']:
                        volume['k2udid'] = vol_idx['k2udid']
                        break
            # If we are not filtering on WWPNs or UID's, or we ARE filtering
            # and there is a match, add this disk to the list to be returned.
            if ((filter_set is None or len(filter_set & rmt_wwpns) > 0 or
                 vdisk_uid in filter_set)):
                return_volumes.append(volume)
                continue

            # This is to support onboarding of XIV volume vSCSI attached
            # to a VM.
            # get wwn from vol and match it in host_ref
            # for vSCSI filter looks like :
            # filters={'host_refs':[u'261120017380075B012A7072810XIV03IBMfcp',
            #                       u'261120017380075B01458072810XIV03IBMfcp']}
            # vol['wwn'] looks like '0017380075B01458'
            if upper_refs is not None:
                for ref in upper_refs:
                    # upper_refs are upper case, convert wwn too
                    wwn = vol['wwn'].upper()
                    if wwn in ref:
                        if volume not in return_volumes:
                            return_volumes.append(volume)
                            break
            if volume in return_volumes:
                continue

            # This is to support onboarding of DS8K volume vSCSI attached
            # to a VM.
            # get wwn from vol, which is the pg83NAA value and match it to the
            # pg83NAA value in the vol_ref.
            # filters={'vol_refs': [{'k2udid': u'01...==',
            #                   'pg83NAA': u'6005076306FFD5F40000000000001169',
            #                    'uid': u'200B75DRR61116907210790003IBMfcp'}],
            # vol['wwn'] looks like u'6005076306ffd5f40000000000001169'
            if vol_refs:
                for vol_ref in vol_refs:
                    vol_ref_pg83NAA = vol_ref.get('pg83NAA')
                    if (vol_ref_pg83NAA and
                            vol_ref_pg83NAA == vol['wwn'].upper()):
                        k2udid = vol_ref.get('k2udid')
                        if k2udid:
                            volume['k2udid'] = k2udid
                        return_volumes.append(volume)
                        break
        LOG.debug("discover_volumes return %s" % return_volumes)
        if not return_volumes:
            msg_input = {'volumes': volumes,
                         'connection_info': connection_info,
                         'filters': filters}
            LOG.warn(_('No volumes were found. '
                       'Volumes from backend %(volumes)s. '
                       'Connection info from backend %(connection_info)s '
                       'Volume filters %(filters)s.' % msg_input))
        return return_volumes

    def _check_supported_size(self, volume, volume_from_backend):
        # Check if the volume is supported given its size in the backend
        pass

    def _check_for_deleted_volumes(self, context, volumes_to_check,
                                   volume_ids_from_backend):
        LOG.debug('Checking the existence of volumes on the backend.'
                  ' Existing backend volume_ids: %s' % volume_ids_from_backend)
        """Determines which of the Volumes no longer exist on the Provider"""
        vols = []
        for volume in volumes_to_check:
            metadata = volume.get('volume_restricted_metadata', list())
            metadata = dict([(m['key'], m['value']) for m in metadata])
            voluid = metadata.get(RESTRICTED_METADATA_VDISK_UID_KEY)
            if voluid and voluid not in volume_ids_from_backend:
                vols.append(volume)
        return vols

    def _get_vscsi_itl_info(self, server_infos, volume):
        """
        Return the vsci connection info for a given volume
        and given instance
        """
        valid_itls = []
        itl_list = volume.get('connection_info', [])
        if itl_list:
            for cinfo in itl_list.values():
                itl = {}
                itl['vios_wwpn'] = cinfo.get('host_wwpns', None)
                itl['target_wwpn'] = cinfo.get('storage_wwpns', None)
                if 'luns' in cinfo:
                    itl['target_lun'] = cinfo.get('luns')
                for server_wwpns in server_infos.values():
                    # make the wwpn to upper
                    vol_wwpns = itl.get('vios_wwpn', [])
                    if vol_wwpns:
                        vol_wwpns = [wwpn.upper() for wwpn in vol_wwpns]
                        server_wwpns = [wwpn.upper() for wwpn in server_wwpns]
                        common_wwpn = list(set(vol_wwpns) & set(server_wwpns))
                        if common_wwpn and len(common_wwpn) > 0:
                            valid_itls.append(itl)
            # END FOR-Loop
        return valid_itls

    def query_volumes(self, ctxt, volumes, server_info={}, mark_boot=True):
        LOG.debug("Enter: query_volumes %(vol)s %(srv_info)s" %
                  dict(vol=volumes, srv_info=server_info))
        uid_map, wwpn_map, pg83NAA_map = (dict(), dict(), dict())
        returned_volumes, boot_map = (list(), dict())

        for server_id, server in server_info.iteritems():
            mark_boot = server.get('boot') is not True
            for volume in server.get('volumes', []):
                uid = volume.get('unique_device_id')
                wwpns = volume.get('wwpns', [])
                pg83NAA = volume.get('pg83NAA')
                vios_wwpn = volume.get('vios_wwpns', None)
                # Build maps of the UID and WWPN depending on which is set
                if uid is not None:
                    if uid.upper() not in uid_map:
                        server_dict = {server_id: (mark_boot, vios_wwpn)}
                    else:
                        server_dict = uid_map[uid.upper()]
                        server_dict[server_id] = (mark_boot, vios_wwpn)
                    uid_map[uid.upper()] = server_dict
                for wwpn in wwpns:
                    wwpn_map[wwpn.upper()] = (server_id, mark_boot)
                # Build map of pg83NAA to server dict
                # Use by DS8K vSCSI
                if pg83NAA:
                    server_dict = pg83NAA_map.get(pg83NAA.upper())
                    if not server_dict:
                        server_dict = {}
                        pg83NAA_map[pg83NAA.upper()] = server_dict
                    server_dict[server_id] = (mark_boot, vios_wwpn)

        LOG.info("uid_map %s wwpn_map %s" % (uid_map, wwpn_map))
        server_to_volumes = {}
        vm_blocking_volumes = []
        for volume in volumes:
            volume_copy = volume.copy()
            server_id, mark_boot = (None, False)
            # Parse the UID out of the MetaData to use for matching Servers
            metadata = volume_copy.get('restricted_metadata', {})
            uid = metadata.get(RESTRICTED_METADATA_VDISK_UID_KEY)
            metadata.get(RESTRICTED_METADATA_VDISK_ID_KEY)
            udid = None
            if 'k2udid' in volume_copy.keys():
                udid = volume_copy['k2udid']
            volume_copy['attach_info'] = {}
            if volume_copy.get('block_vm_onboard'):
                vm_blocking_volumes.append(volume_copy)
            # Parse the correct Connection Info based on the WWPN's
            connect_infos = volume_copy.get('connection_info', {})
            if len(connect_infos) > 0:
                for cinfo in connect_infos.values():
                    wwpns = cinfo.get('host_wwpns', [])
                    # Loop through each of the WWPN's looking for the Server
                    for wwpn in wwpns:
                        # If we found a Server for the WWPN, use this Info
                        if wwpn_map.get(wwpn.upper()) is not None:
                            server_id = wwpn_map[wwpn.upper()][0]
                            if server_id not in volume_copy['attach_info']:
                                mark_boot = wwpn_map[wwpn.upper()][1]
                                connection_info = None
                                # sanitize connection info for npiv
                                if cinfo is not None:
                                    cinfo['target_wwn'] = cinfo.pop(
                                        'storage_wwpns', None)
                                    conn_type = \
                                        discovery_driver.CONNECTION_TYPE_NPIV
                                    connection_info =\
                                        self._get_connection_info(
                                            uid, cinfo, conn_type)
                                    self._add_attach_info(
                                        volume_copy, boot_map,
                                        connection_info, server_id,
                                        mark_boot=mark_boot)
                            if server_id not in server_to_volumes:
                                server_to_volumes[server_id] = []
                            server_to_volumes[server_id].append(volume_copy)

            wwn = volume_copy['wwn']
            server_dict = None
            # Look the WWN up in the pg83NAA map.
            # The DS8K puts the pg83NAA value in the wwn field.
            # This is the DS8K vSCSI path.
            server_dict = pg83NAA_map.get(wwn.upper())
            # If that wasn't found, look up the WWN in the uid_map
            # key to get the server dict. This is the XIV vSCSI path.
            if not server_dict:
                for uidkey in uid_map.keys():
                    if wwn in uidkey:
                        wwnFound = uidkey
                        server_dict = uid_map[wwnFound]

            if server_dict:
                for server_id in server_dict.keys():
                    if server_id not in volume_copy['attach_info']:
                        mark_boot = server_dict[server_id][0]
                        udid = volume.get('k2udid')
                        connection_info = None
                        server_wwpn = server_dict[server_id][1]
                        # get correct connection info and itl list
                        valid_itls = self._get_vscsi_itl_info(server_wwpn,
                                                              volume_copy)
                        if valid_itls:
                            LOG.info(_("Query Volume Valid ITLS=%s") %
                                     valid_itls)
                            # If given the VDisk UID from the VM, then
                            # it isn't NPIV
                            # pass the first itl to extract the target_wwn
                            cinfo = valid_itls[0]
                            cinfo['target_wwn'] = cinfo.get('target_wwpn',
                                                            None)
                            conn_type = discovery_driver.CONNECTION_TYPE_VSCSI
                            pg83value = (wwn.upper() if wwn.upper()
                                         in pg83NAA_map else uid)
                            connection_info = self._get_connection_info(
                                pg83value, cinfo, conn_type, udid=udid)
                            connection_info['vscsi_target_lun'] =\
                                valid_itls
                            connection_info['data']['vendor'] =\
                                "IBM"
                            self._add_attach_info(volume_copy, boot_map,
                                                  connection_info,
                                                  server_id,
                                                  mark_boot=mark_boot)
                            if server_id not in server_to_volumes:
                                server_to_volumes[server_id] = []
                            server_to_volumes[server_id].append(volume_copy)

            # Add the Copy of the Volume to the list to be returned
            volume_copy.pop('connection_info', None)
            returned_volumes.append(volume_copy)
        # Check return value to ensure every volume has some attach info.
        for volume in returned_volumes:
            if server_info and not volume['attach_info']:
                LOG.warn(_('The volume %(volume)s does not have attachment '
                           'information. Input volumes: %(volumes)s '
                           'Input server information: %(servers)s') %
                         {'volume': volume,
                         'volumes': volumes,
                         'servers': server_info})
        for volume in vm_blocking_volumes:
            attach_info = volume.get('attach_info', {})
            for server_id in attach_info.keys():
                error = _('Volume %(name)s is preventing the virtual machine'
                          ' from being managed because: %(reasons)s' %
                          {'name': volume['name'],
                           'reasons': volume['block_vm_onboard']['reasons']})
                LOG.error(error)
                # send messages to console
                payload = {}
                payload['msg'] = error
                anotifier = rpc.get_notifier('volume', self.host)
                anotifier.error(context.get_admin_context(),
                                'volume.log', payload)
                # For every server this volume is attached to, remove
                # attachment information to this server for all other volumes.
                # This prevents the VM from being onboarded because it will
                # have no volumes.
                server_volumes = server_to_volumes[server_id]
                for sv in server_volumes:
                    attach_info = sv.get('attach_info', {})
                    attach_info.pop(server_id, 0)
                    if not attach_info and sv in returned_volumes:
                        returned_volumes.remove(sv)
                        LOG.info(_('Volume %(name)s will not be managed '
                                   'because it is only attached to virtual '
                                   'machines which are prevented from '
                                   'management by an unsupported volume.' %
                                   {'name': sv['name']}))

        LOG.debug("exit: query_volumes %s" % returned_volumes)
        return returned_volumes

    def _add_default_volume_type(self, volume):
        """Adds the default volume type to the volume if the volume does not
        have a volume type yet."""
        ctxt = context.get_admin_context()
        volume_type_id = volume.get('volume_type_id')
        if not volume_type_id:
            volume_type = volume_types.get_default_volume_type()
            if volume_type:
                volume_type_id = volume_type['id']
                volume['volume_type_id'] = volume_type_id
                # Update db to preserve volume_type
                LOG.info(_('adding volume_type_id to volume=%s') % volume)
                self.db.volume_update(ctxt, volume['id'],
                                      {'volume_type_id': volume_type_id})

    def _update_metadata(self, db_vol, volume_uuid,
                         vdisk_id, vdisk_name, vdisk_uid, wwn):
        ctxt = context.get_admin_context()
        metadata = {RESTRICTED_METADATA_VDISK_ID_KEY: vdisk_id,
                    RESTRICTED_METADATA_VDISK_NAME_KEY: vdisk_name,
                    RESTRICTED_METADATA_VDISK_UID_KEY: vdisk_uid,
                    RESTRICTED_METADATA_NAA: wwn
                    }
        powervc_db_api.\
            volume_restricted_metadata_update_or_create(
                ctxt, volume_uuid, metadata)
        # db_vol can be none in the create from clone case
        if db_vol is None:
            LOG.debug('db_vol is none, creating an empty one')
            db_vol = {}

        # DS8K volume create returns a db_vol object that already
        # has metadata on it and we need to use that versus read from the
        # database to preserve the community drivers metadata additions
        if db_vol.get('metadata'):
            meta = db_vol.get('metadata')
            LOG.debug('using metadta from db_vol %s' % meta)
        else:
            meta = db_api.volume_metadata_get(ctxt, volume_uuid)
            if not meta:
                meta = {}
        meta[METADATA_VOLUME_WWN] = wwn
        db_vol['metadata'] = meta
        return db_vol

    def _b_to_volume_size(self, byte_size):
        # Cinder volume size is GiB, 1 GiB = 2^30 bytes
        # Previous to this bug it as ambiguous between 10^9 bytes and 2^30
        # bytes: https://bugs.launchpad.net/openstack-api-site/+bug/1456631
        return byte_size // (2 ** 30)
