#  vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2015 IBM Corp.
# All Rights Reserved.
#    (c) Copyright 2013 Brocade Communications Systems Inc.
#    All Rights Reserved.
#    Copyright 2013 OpenStack LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import random

from eventlet import greenthread
from oslo_concurrency import processutils

from cinder import ssh_utils
from cinder import utils

from oslo_log import log as logging
from oslo_utils import excutils
from cinder import exception
from powervc_cinder import _
from cinder.i18n import _LE

from cinder.zonemanager.drivers.brocade.brcd_fc_zone_client_cli \
    import BrcdFCZoneClientCLI
import cinder.zonemanager.drivers.brocade.fc_zone_constants as \
    ZoneConstant


LOG = logging.getLogger(__name__)


class PowerVCBrcdFCZoneClientCLI(BrcdFCZoneClientCLI):

    def __init__(self, *args, **kwargs):
        super(PowerVCBrcdFCZoneClientCLI, self).__init__(*args, **kwargs)

    def get_active_zone_map(self):
        """
        Returns the zoneset &
        the zones - a dictionary where keys are zone names and values are
        lists of WWPNs in that zone.  WWPNs are in the 'expanded' colon format.
        """
        zoneset = self.get_active_zone_set()
        zones = zoneset[ZoneConstant.CFG_ZONES]
        return zoneset, zones

    # override community to remove random sleep.
    # yfeng TODO: remove when community integrates in Liberty
    def _ssh_execute(self, cmd_list, check_exit_code=True, attempts=1):
        """Execute cli with status update.

        Executes CLI commands such as cfgsave where status return is expected.
        """
        utils.check_ssh_injection(cmd_list)
        command = ' '. join(cmd_list)

        if not self.sshpool:
            self.sshpool = ssh_utils.SSHPool(self.switch_ip,
                                             self.switch_port,
                                             None,
                                             self.switch_user,
                                             self.switch_pwd,
                                             min_size=1,
                                             max_size=5)
        stdin, stdout, stderr = None, None, None
        LOG.debug("Executing command via ssh: %s" % command)
        last_exception = None
        try:
            with self.sshpool.item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        stdin, stdout, stderr = ssh.exec_command(command)
                        stdin.write("%s\n" % ZoneConstant.YES)
                        channel = stdout.channel
                        exit_status = channel.recv_exit_status()
                        LOG.debug("Exit Status from ssh:%s", exit_status)
                        # exit_status == -1 if no exit code was returned
                        if exit_status != -1:
                            LOG.debug('Result was %s' % exit_status)
                            if check_exit_code and exit_status != 0:
                                raise processutils.ProcessExecutionError(
                                    exit_code=exit_status,
                                    stdout=stdout,
                                    stderr=stderr,
                                    cmd=command)
                            else:
                                return True
                        else:
                            return True
                    except Exception as e:
                        LOG.error(e)
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                LOG.debug("Handling error case after "
                          "SSH:%s", last_exception)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Error executing command via ssh: %s"), e)
        finally:
            if stdin:
                stdin.flush()
                stdin.close()
            if stdout:
                stdout.close()
            if stderr:
                stderr.close()
