#  vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2014, 2015 IBM Corp.
# All Rights Reserved.
#    (c) Copyright 2013 Brocade Communications Systems Inc.
#    All Rights Reserved.
#    Copyright 2013 OpenStack LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from cinder.zonemanager.drivers.cisco.cisco_fc_zone_client_cli \
    import CiscoFCZoneClientCLI
import cinder.zonemanager.drivers.cisco.fc_zone_constants as \
    ZoneConstant
import six
import random
from eventlet import greenthread

from cinder import utils
from cinder import ssh_utils

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
from powervc_cinder import _

LOG = logging.getLogger(__name__)


class PowerVCCiscoFCZoneClientCLI(CiscoFCZoneClientCLI):

    def __init__(self, *args, **kwargs):
        super(PowerVCCiscoFCZoneClientCLI, self).__init__(*args, **kwargs)

    def get_active_zone_map(self):
        """
        Returns the zoneset &
        the zones - a dictionary where keys are zone names and values are
        lists of WWPNs in that zone.  WWPNs are in the 'expanded' colon format.
        """
        zoneset = self.get_active_zone_set()
        zones = zoneset[ZoneConstant.CFG_ZONES]
        return zoneset, zones

    # Adding additional paramateres needed for community add_zones
    # need: vsan, active_zone_set, zone_status
    def add_zones(self, zones, activate, fabric_vsan=None,
                  active_zone_set=None, zoneStatus=None):
        """Add zone configuration.

        This method will add the zone configuration passed by user.
            input params:
            zones - zone names mapped to members and VSANs.
            zone members are colon separated but case-insensitive
            {   zonename1:[zonememeber1,zonemember2,...],
                zonename2:[zonemember1, zonemember2,...]...}
            e.g: {'openstack50060b0000c26604201900051ee8e329':
                    ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']
                }
            activate - True/False
        """
        if (fabric_vsan is None):
            fabric_vsan = self.fabric_vsan
        if (active_zone_set is None):
            active_zone_set = self.get_active_zone_set()
        if (zoneStatus is None):
            zoneStatus = self.get_zoning_status()

        s = super(PowerVCCiscoFCZoneClientCLI, self)
        try:
            s.add_zones(zones, activate, fabric_vsan,
                        active_zone_set, zoneStatus)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_("Failed adding zones= %(zones)s, error= %(error)s")
                          % {'zones': zones, 'error': six.text_type(e)})

    def delete_zones(self, zones, activate, fabric_vsan=None,
                     active_zone_set=None, zoneStatus=None):
        """Delete zones from fabric.

        Method to delete the active zone config zones

        params zone_names: zoneNames separated by semicolon
        params activate: True/False
        """
        if (fabric_vsan is None):
            fabric_vsan = self.fabric_vsan
        if (active_zone_set is None):
            active_zone_set = self.get_active_zone_set()
        if (zoneStatus is None):
            zoneStatus = self.get_zoning_status()

        s = super(PowerVCCiscoFCZoneClientCLI, self)
        try:
            s.delete_zones(zones, activate, fabric_vsan,
                           active_zone_set, zoneStatus)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_('Failed deleting zones= %(zones)s, '
                            'error= %(error)s') %
                          {'zones': zones, 'error': six.text_type(e)})

    # This is for encountered non harmful errors from Cisco,
    # because of Cisco does not support concurrent writs...
    # wrapping around a try/except to catch errors and ignoring
    def _cfg_save(self):
        try:
            s = super(PowerVCCiscoFCZoneClientCLI, self)
            s._cfg_save()
        except Exception as e:
            # concurrent writes to Cisco nvram are not supported.
            # if we get a failure during the copy, log it and
            # ignore it as these copies are not necessary for
            # switch functionality.
            LOG.info(_("Error while running copy running-config"
                       "to startup-config command."
                       "Ignoring this error."))
            LOG.error(six.text_type(e))

    # override community to remove random sleep.
    # yfeng TODO: remove when community integrates in Liberty
    def _ssh_execute(self, cmd_list, check_exit_code=True, attempts=1):
        """Execute cli with status update.

        Executes CLI commands where status return is expected.

        cmd_list is a list of commands, where each command is itself
        a list of parameters.  We use utils.check_ssh_injection to check each
        command, but then join then with " ; " to form a single command.
        """

        # Check that each command is secure
        for cmd in cmd_list:
            utils.check_ssh_injection(cmd)

        # Combine into a single command.
        command = ' ; '.join(map(lambda x: ' '.join(x), cmd_list))

        if not self.sshpool:
            self.sshpool = ssh_utils.SSHPool(self.switch_ip,
                                             self.switch_port,
                                             None,
                                             self.switch_user,
                                             self.switch_pwd,
                                             min_size=1,
                                             max_size=5)
        stdin, stdout, stderr = None, None, None
        LOG.debug("Executing command via ssh: %s" % command)
        last_exception = None
        try:
            with self.sshpool.item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        stdin, stdout, stderr = ssh.exec_command(command)
                        channel = stdout.channel
                        exit_status = channel.recv_exit_status()
                        LOG.debug("Exit Status from ssh:%s", exit_status)
                        # exit_status == -1 if no exit code was returned
                        if exit_status != -1:
                            LOG.debug('Result was %s' % exit_status)
                            if check_exit_code and exit_status != 0:
                                raise processutils.ProcessExecutionError(
                                    exit_code=exit_status,
                                    stdout=stdout,
                                    stderr=stderr,
                                    cmd=command)
                            else:
                                return True
                        else:
                            return True
                    except Exception as e:
                        msg = _("Exception: %s") % six.text_type(e)
                        LOG.error(msg)
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                LOG.debug("Handling error case after SSH:%s", last_exception)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                msg = (_("Error executing command via ssh: %s") %
                       six.text_type(e))
                LOG.error(msg)
        finally:
            if stdin:
                stdin.flush()
                stdin.close()
            if stdout:
                stdout.close()
            if stderr:
                stderr.close()
