# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
"""
ZoneManager is responsible to manage access control using FC zoning
when zoning mode is set as 'fabric'.
ZoneManager provides interfaces to add connection and remove connection
for given initiator and target list associated with a FC volume attach and
detach operation.

This PowerVC implementation does not have a pluggable driver architecture.
It speaks directly to the Brocade backend library.

It maintains persistent connections to the fabrics, and performs operations
across fabrics in parallel using greenthreads.

It also changes the API used to add connections, by requiring the caller to
be explicit about which zones to create on which fabrics.  This results in
higher performance, because we don't have to query connectivity add the time
that connections are added.
"""

from oslo_utils import encodeutils
from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_context import context as osloctxt
from oslo_log import log as logging
from cinder.volume.configuration import Configuration
from cinder.zonemanager import fc_zone_manager as cmt_zone_manager
from powervc_cinder import _
from powervc_cinder import exception
from powervc_cinder.zonemanager.drivers.brocade. \
    powervc_brcd_fc_zone_client_cli import PowerVCBrcdFCZoneClientCLI
from \
    powervc_cinder.zonemanager.drivers.cisco.powervc_cisco_fc_zone_client_cli \
    import PowerVCCiscoFCZoneClientCLI
from eventlet.greenpool import GreenPool
import re

LOG = logging.getLogger(__name__)

zone_manager_opts = [
    cfg.StrOpt('zoning_policy',
               default='initiator-target',
               help='Zoning policy configured by user'),
    cfg.BoolOpt('zone_activate',
                default=True,
                help="Indicates whether zone should be activated or not"),
    cfg.StrOpt('zone_name_prefix',
               default="powervc",
               help="A prefix to be used when naming zone"),
    cfg.StrOpt('fc_fabric_names',
               default='',
               help='FC Fabric names')
]

common_zone_fabric_opts = [
    cfg.StrOpt('fc_fabric_address',
               default='',
               help='Management IP of fabric'),
    cfg.StrOpt('fc_fabric_user',
               default='',
               help='Fabric user ID'),
    cfg.StrOpt('fabric_type',
               default='brocade',
               help='Fabric type'),
    cfg.StrOpt('fc_fabric_display_name',
               default='',
               help='Display name'),
    cfg.StrOpt('fc_fabric_password',
               default='',
               help='Password for user',
               secret=True),
    cfg.IntOpt('fc_fabric_port',
               default=22,
               min=1, max=65535,
               help='Connecting port'),
    cfg.StrOpt('fc_fabric_uniqueid',
               default=None,
               help='UniqueID of fabric')
]

CONF = cfg.CONF
CONF.register_opts(zone_manager_opts)


def get_formatted_wwn(wwn_str):
    """Utility API that formats WWN to insert ':'.
    """
    if (len(wwn_str) != 16):
        return wwn_str.lower()
    else:
        return ':'.join(
            [wwn_str[i:i + 2] for i in range(0, len(wwn_str), 2)]).lower()


class PowerVCZoneManager:
    fabrics = {}

    def __init__(self, configuration, overrides=None):
        """
        Here we register the configuration options and extract all the
        configuration data.

        We also establish connections to the fabrics for use in subsequent
        operations.

        This will raise a ZoneManagerParallel exception if anything goes wrong
        during the connection process.
        """
        self.conf = configuration
        self.conf.register_opts(zone_manager_opts)
        self.fc_fabric_conf = {}
        # the fabric names are are registered in the group opt,
        # 'fc-zone-manager', in the community fc_zone_manager
        fabric_names = self.conf.get('fc-zone-manager') \
                                .get('fc_fabric_names').split(',')

        # Check for uniqueness of names
        if len(fabric_names) != len(set(fabric_names)):
            msg = _("Fabric names are not unique: %(fabric_names)s") % \
                {'fabric_names': ','.join(fabric_names)}
            raise exception.ZoneManagerMisconfig(reason=msg)

        # For each fabric we will register the common opts first.
        # Then depending the fabric type, additional fabrics opts
        # specific to that vendor will be registered.
        for fabric_name in fabric_names:
            fabric_conf_common = \
                Configuration(common_zone_fabric_opts, fabric_name)
            # Handle case where the fabric type is passed in as an
            # override such as happens from map_initiators.
            if overrides and overrides[fabric_name]:
                over_dict = dict(overrides[fabric_name])
                if 'fabric_type' in over_dict:
                    fabric_conf_common.fabric_type = over_dict['fabric_type']
            fabric_type = fabric_conf_common.fabric_type

            if fabric_type == "brocade":
                self._register_brocade_options(fabric_conf_common)
            elif fabric_type == "cisco":
                self._register_cisco_options(fabric_conf_common)
            else:
                raise

            self.fc_fabric_conf[fabric_name] = fabric_conf_common

            if overrides:
                # set overrides for specific fabric type
                self._handle_overrides(fabric_name, overrides)

        # Now we initialise a connection to the switch for each of the fabrics

        # This function is called in a GreenThread for each registered switch.
        def _do_connect(context, fabric_name):
            # Mark this thread as running the passed-in context, so that log
            # messages can be correlated.
            context.update_store()

            def _do_locked_connect(fabric_name):
                fabric_display_name = encodeutils.safe_decode(
                    self.fc_fabric_conf[fabric_name]
                    .safe_get('fc_fabric_display_name'))
                fabric_ip = self.fc_fabric_conf[fabric_name] \
                                .safe_get('fc_fabric_address')
                fabric_user = self.fc_fabric_conf[fabric_name] \
                                  .safe_get('fc_fabric_user')
                fabric_pwd = self.fc_fabric_conf[fabric_name] \
                                 .safe_get('fc_fabric_password')
                fabric_port = self.fc_fabric_conf[fabric_name] \
                                  .safe_get('fc_fabric_port')
                fabric_type = self.fc_fabric_conf[fabric_name] \
                                  .safe_get('fabric_type')
                fabric_vsan = 'n/a'  # default for loging

                if fabric_type == "brocade":
                    conn = PowerVCBrcdFCZoneClientCLI(fabric_ip,
                                                      fabric_user,
                                                      fabric_pwd,
                                                      fabric_port)
                elif fabric_type == "cisco":
                    fabric_vsan = self.fc_fabric_conf[fabric_name] \
                                      .safe_get('fc_fabric_vsan')
                    conn = PowerVCCiscoFCZoneClientCLI(fabric_ip,
                                                       fabric_user,
                                                       fabric_pwd,
                                                       fabric_port,
                                                       fabric_vsan)
                LOG.info(_("Created zone client CLI: Fabric designator="
                           "%(fabric_name)s, fabric type=%(fabric_type)s, "
                           "display name=%(fabric_display_name)s, "
                           "IP addr=%(fabric_ip)s, user=%(fabric_user)s, "
                           "port=%(fabric_port)s, VSAN=%(fabric_vsan)s.") %
                         locals())
                return conn

            return _do_locked_connect(fabric_name)

        # Start a GreenThread for each fabric that we will connect to and
        # initiate the connection in it.
        pool = GreenPool(size=len(fabric_names))

        # Obtain our current context so that we can make sure that our child
        # threads have the same context, so that we can correlate log messages
        # that they generate.
        context = osloctxt.get_current()

        threads = {}
        for fabric_name in fabric_names:
            thread = pool.spawn(_do_connect, context, fabric_name)
            threads[fabric_name] = thread

        # Collect the resulting connection objects.
        # The wait() will raise an exception if something went wrong.
        exceptions = []
        for fabric_name, thread in threads.iteritems():
            try:
                self.fabrics[fabric_name] = thread.wait()
                LOG.info(_("Connection established to fabric %(f_name)s") %
                         dict(f_name=fabric_name))
            except Exception as e:
                exceptions.append(e)

        # If any exceptions were raised, we throw an exception that
        # encapsulates them all.
        if exceptions:
            raise exception.ZoneManagerParallel(exceptions)

    def _register_brocade_options(self, fabric_conf):
        brocade_fabric_opts = [
            cfg.FloatOpt('fc_fabric_timeout',
                         default=10.0,
                         help='Connection timeout '
                         '(seconds)'),
            cfg.FloatOpt('fc_fabric_cmd_timeout',
                         default=100.0,
                         help='Command execution '
                         'timeout (seconds)')]

        fabric_conf.append_config_values(brocade_fabric_opts)

    def _register_cisco_options(self, fabric_conf):
        cisco_fabric_opts = [
            cfg.StrOpt('fc_fabric_vsan',
                       default="1",
                       help='Cisco VSAN to provision to')]

        fabric_conf.append_config_values(cisco_fabric_opts)

    def _handle_overrides(self, fabric_name, overrides):
        """
        If running in a process without fabric CLI options passed, allow
        those options to be given and set via the overrides param
        Overrides are specific for a fabric of a particular type.  That is,
        certain brocade overrides do not apply to cisco fabrics.
        Input: fabric_name - either 'a' or 'b' currently, but could go beyond
                             that in the future
               overrides - a dictionary containing list of (key, value) tuples
                           for the specified fabric.
               e.g.  {'a': [('fc_fabric_user', 'admin')],
                      'b': [('fc_fabric_user', 'admin')]
                     }
        """
        # if the fabric_name [a,b, etc..] exists in overrides,
        # will override that specific fabric opts

        if fabric_name in overrides.keys():
            override_list = overrides[fabric_name]
            for item in override_list:
                self.conf.local_conf \
                    .set_override(item[0], item[1], fabric_name)

    def _parallel_execute(self, operation, *args):

        def _spawn(context, operation, fabric_name, conn, *args):
            # Inherit this thread's context from the parent
            context.update_store()

            @lockutils.synchronized(fabric_name, 'fcfabric-', True)
            def _locked_spawn(operation, fabric_name, conn, *args):
                return operation(fabric_name, conn, *args)

            return _locked_spawn(operation, fabric_name, conn, *args)

        """
        Perform an operation against all fabrics, consolidate the responses
        into a dictionary keyed on fabric name.
        """
        pool = GreenPool(size=len(self.fabrics))

        # Obtain our current context so that we can make sure that our child
        # threads have the same context, so that we can correlate log messages
        # that they generate.
        context = osloctxt.get_current()
        threads = {}
        for fabric_name, conn in self.fabrics.iteritems():
            thread = pool.spawn(_spawn, context, operation, fabric_name, conn,
                                *args)
            threads[fabric_name] = thread

        # Collect the responses.  This may raise exceptions when we call wait()
        # If they do, we collect them and raise a collective exception at the
        # end.
        responses = {}
        exceptions = []

        for fabric_name, thread in threads.iteritems():
            try:
                responses[fabric_name] = thread.wait()
            except Exception as e:
                """
                FabricExceptions can indicate that a backtrace is not required
                if they contain sufficient debug information themselves.
                """
                if (not isinstance(e, exception.FabricException) or
                        e.backtrace_needed):
                    LOG.exception(e)
                exceptions.append(e)

        # If any exceptions were raised, we throw an exception that
        # encapsulates them all.
        if exceptions:
            raise exception.ZoneManagerParallel(exceptions)

        return responses

    def close(self):
        """
        Initiate a parallel close operation on all of the registered fabrics.
        Silently suppress any exceptions that are thrown.
        """
        def _do_close(fabric_name, conn):
            try:
                # Replaced close_connection() -> cleanup(),
                # THIS DOES NOT REMOVE/CLOSE ssh paramiko connection.
                # Connection will remain in ssh pool
                conn.cleanup()
            except Exception:
                # Not worth reporting anything that went wrong, we're closing.
                pass

        self._parallel_execute(_do_close)

    def add_connection(self, host_name, fabric_initiator_target_map):
        """
        Add new zones to the fabrics.

        The host_name is used to build the zone names.

        The fabric_initiator_target_map is like this:
        { 'a':  { 'initiator_A': [ target_W, target_X ],
                  'initiator_B': [ target_W, target_X ]
                }
          'b':  { 'initiator_C': [ target_Y, target_Z ],
                  'initiator_D': [ target_Y, target_Z ]
                }
        }

        So there is an initiator->target map per fabric.  We process this
        request on all fabrics in parallel, passing each one the appropriate
        piece of this data structure.
        """
        # This is the function that is called for each fabric.
        def _do_add_connection(fabric_name, conn,
                               fabric_initiator_target_map):
            # We build up a list of zones and ensure that the WWPNs are in the
            # correct format.
            if fabric_name in fabric_initiator_target_map:
                it_map = fabric_initiator_target_map[fabric_name]

                zone_map = {}
                for initiator, targets in it_map.iteritems():
                    for target in targets:
                        zone_name = self.make_zone_name(host_name, initiator,
                                                        target)
                        zone_map[zone_name] = [get_formatted_wwn(initiator),
                                               get_formatted_wwn(target)]

                exception_desc = getattr(conn, 'exception_desc', fabric_name)
                # Perform name check for zones on fabric, do not create
                # if zones already exist

                active_zoneset, active_zone_dict = conn.get_active_zone_map()
                zone_map_zones = zone_map.keys()

                for zone in zone_map_zones:
                    if zone in active_zone_dict:
                        del zone_map[zone]
                if len(zone_map) == 0:
                    LOG.info(_("Zones attempted to be created already exist,"
                               "skipping creating %s") % zone_map_zones)
                    return
                LOG.info(_("Zones to be created for %(fabric)s are "
                           "%(zones)r") % {'fabric': exception_desc,
                                           'zones': zone_map})
                conn.add_zones(zone_map,
                               self.conf.zone_activate,
                               active_zone_set=active_zoneset)

        # Perform the operation in parallel across all fabrics
        return self._parallel_execute(_do_add_connection,
                                      fabric_initiator_target_map)

    def delete_connection(self, initiator_target_map):
        """
        Removes all the zones that were created with our prefix and contains
        one of our initiators.
        """
        def _do_delete_zones(fabric_name, conn, initiators, targets):
            # We look in the zone list for zones containing precisely one of
            # the specified initiators and one other thing, and remove it.
            # The zone name must also use our prefix.
            initiator_set = set(map(get_formatted_wwn, initiators))
            target_set = set(map(get_formatted_wwn, targets))
            zones_to_delete = []

            # Maps zone name to list of WWPNs.  WWPNs are in colon-expanded
            # format.
            active_zoneset, active_zone_dict = conn.get_active_zone_map()
            for zone_name, zone_wwpns in active_zone_dict.iteritems():
                zone_set = set(map(get_formatted_wwn, zone_wwpns))
                if (zone_name.startswith(self.conf.zone_name_prefix)
                        and len(zone_set & target_set) == 1
                        and len(zone_set & initiator_set) == 1):
                    zones_to_delete.append(zone_name)

            if len(zones_to_delete) > 0:
                exception_desc = getattr(conn, 'exception_desc', fabric_name)
                LOG.info(_("%(fabric_descriptor)s: deleting zones "
                           "%(zone_list)s") %
                         {'fabric_descriptor': exception_desc,
                          'zone_list': ', '.join(zones_to_delete)}
                         )
                # Community cli expects zones to be deleted to be a
                # single string separated by ;
                zone_name_string = ''
                num_zones = len(zones_to_delete)
                for i in range(0, num_zones):
                    if i == 0:
                        zone_name_string = zones_to_delete[i]
                    else:
                        zone_name_string = ('%s%s%s' % (
                                            zone_name_string, ';',
                                            zones_to_delete[i]))

                conn.delete_zones(zone_name_string,
                                  self.conf.zone_activate,
                                  active_zone_set=active_zoneset)
        # ? initiators = connector['wwpns']
        initiator_wwpns = []
        target_wwpns = []
        for initiators in initiator_target_map:
            initiator_wwpns.append(initiators)
        for targets in initiator_target_map.values():
            target_wwpns = target_wwpns + targets
        # Perform the operation in parallel across all fabrics
        return self._parallel_execute(_do_delete_zones,
                                      initiator_wwpns,
                                      target_wwpns)

    def get_san_context(self, wwn_lists):
        """
        SAN lookup for end devices.

        wwn_lists is a dictionary of strings -> lists of WWPNs

        We pass this dictionary to every fabric, which returns the same
        dictionary, but only including the WWPNs that are present on the
        switch.

        We return a dictionary of fabric_name to filtered dictionary.

        e.g.

        ->   { 'initiators': [ A, B, C, D ],
               'targets': [ W, X, Y, Z ] }

        returns

            {
              'fabric_a': { 'initiators': [ A, B ],
                            'targets': [ W, X ] },
              'fabric_b': { 'initiators': [ C, D ],
                            'targets': [ Y, Z ] }
            }
        """
        def _do_get_san_context(fabric_name, conn, wwn_sets):
            nsinfo = conn.get_nameserver_info()
            LOG.info(_("Fabric %(fab_name)s. Returned data from switch "
                       "driver's get_nameserver_info(): %(info)r") %
                     dict(fab_name=fabric_name, info=nsinfo))
            nsset = set(nsinfo)

            found_wwns = {}
            for name, wwn_set in wwn_sets.iteritems():
                found_wwns[name] = list(nsset & wwn_set)

            return found_wwns

        # Convert the lists of WWPNs into sets to make it easier to find the
        # intersection with what is on the switches.
        wwn_sets = {}
        for name, wwpns in wwn_lists.iteritems():
            wwn_sets[name] = set(list(wwpns))

        return self._parallel_execute(_do_get_san_context, wwn_sets)

    def get_zones(self, initiators=None, targets=None):
        """
        Lookup for zones in end devices.
        initiator: list of initiator wwpns to filter against.
        targets  : list of target wwpns to filter against
        returns  : a dictionary dictionary of fabric_name to zonenames
                   and the zonemembers.
                   If no filter is supplied, then all the current
                   active zonenames will be returned.

                {
                  'fabric_a': {zone1: [initiators, targets]},
                               zone2: [initiators, targets]},
                  'fabric_b': {zone3: [initiators, targets]},
                               zone4: [initiators, targets]}
                }
        """
        def _get_zones(fabric_name, conn, initiators, targets):
            active_zoneset, active_zone_dict = conn.get_active_zone_map()
            filter_list = []
            filtered_zones = {}
            if initiators:
                initiators_set = set(map(get_formatted_wwn, initiators))
                filter_list.append(initiators_set)
            if targets:
                targets_set = set(map(get_formatted_wwn, targets))
                filter_list.append(targets_set)
            if not filter_list:
                filtered_zones = active_zone_dict
            else:
                for zonename, zone_wwpns in active_zone_dict.iteritems():
                    zone_set = set(map(get_formatted_wwn, zone_wwpns))
                    append_zone = True
                    # filter against initiators (and or) targets
                    for filter_set in filter_list:
                        if len(zone_set & filter_set) == 0:
                            append_zone = False
                            break
                    if append_zone:
                        filtered_zones[zonename] = zone_wwpns
            LOG.info(_("Fabric %(fab_name)s. Returned zonenames from switch "
                       "driver's : %(info)r") %
                     dict(fab_name=fabric_name, info=filtered_zones))
            return filtered_zones
        return self._parallel_execute(_get_zones, initiators, targets)

    def make_zone_name(self, host_name, initiator, target):
        # Maximum Brocade name length is 64 characters
        # Naming scheme: "powervc_"+"host_name"(trunc)+"initiator"+"target"
        chars_remaining = 64 - len(self.conf.zone_name_prefix +
                                   initiator +
                                   target)

        trunc_host = host_name[:chars_remaining]
        trunc_host = re.sub(r'[^a-zA-Z0-9_]', "_", trunc_host)
        return self.conf.zone_name_prefix + trunc_host + initiator + target
