# vim: tabstop=4 shiftwidth=4 softtabstop=4

# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""
Monkeypatch utility functions for use with zonemanager.

"""

from powervc_cinder import _
from oslo_log import log as logging
from oslo_utils import excutils
from cinder import exception
from powervc_cinder.zonemanager.powervc_fc_zone_manager \
    import PowerVCZoneManager
LOG = logging.getLogger(__name__)
# We need to import the sanitize name to use for replacing the host
from oslo_utils import importutils
sanitize_name = importutils.import_class(
    'powervc_cinder.db.sqlalchemy.models._sanitize_volume_name')


def _sanitize_connector_host(connector):
        # sanitize the host. Some back-ends can't handle some chars
        if 'host' in connector:
            connector['host'] = sanitize_name(connector['host'])


def AddFCZone(initialize_connection):
    """Decorator to add fibre channel zones"""
    def decorator(self, *args, **kwargs):
        # Validate the connector.
        volume = args[0]
        connector = args[1]
        if ('connection_type' not in connector or
                not connector['connection_type']):
            if self.configuration.zoning_mode == 'fabric':
                exception_msg = (_('initialize_connection: '
                                   'Invalid connector missing '
                                   'connection_type. Volume %(vol)s '
                                   'Connector %(con)s') %
                                 dict(con=connector, vol=volume))
                raise exception.VolumeBackendAPIException(data=exception_msg)
            connection_type = None
            do_zoning = False
        elif 'npiv' in connector['connection_type']:
            connection_type = 'npiv'
            do_zoning = True
        elif 'pv_vscsi' in connector['connection_type']:
            # to zone for vscsi or not to zone ...
            zone_vscsi = self.configuration.safe_get('zone_vscsi')
            if zone_vscsi:
                do_zoning = True
            else:
                do_zoning = False
            connection_type = 'pv_vscsi'
        else:
            # if we got here we have a bad value in connection type.
            exception_msg = (_('initialize_connection: '
                               'Invalid connection_type in connector '
                               'Volume %(vol)s Connector %(con)s') %
                             dict(con=connector, vol=volume))
            raise exception.VolumeBackendAPIException(data=exception_msg)
        # if initialize_connection fails log the error so we can debug
        # that failure.
        try:
            # sanitize the host.
            _sanitize_connector_host(connector)
            info = initialize_connection(self, *args, **kwargs)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(ex)

        if not info:
            LOG.info(_("No connection info returned by Volume driver. "
                       "No Zoning! Volume %(vol)s Connector %(con)s") %
                     dict(con=connector, vol=volume))
            return info

        vol_type = info.get('driver_volume_type', None)
        if vol_type != 'fibre_channel':
            LOG.info(_("AddFCZone driver_volume_type %(vtype)s Volume %(vol)s "
                       "Connector %(con)s Connection info %(c_info)s") %
                     dict(vtype=vol_type, con=connector, vol=volume,
                          c_info=info))
            return info

        if 'initiator_target_map' not in info['data']:
            LOG.info(_("AddFCZone no initiator_target_map Volume %(vol)s "
                       "Connector %(con)s %(c_info)s") %
                     dict(con=connector, vol=volume, c_info=info))
            return info
        LOG.debug("AddFCZone Connection info %s" % info)
        # add npiv or vscsi into the return info
        if connection_type:
            info['data']['connector_type'] = connection_type

        try:
            itm = info['data']['initiator_target_map']
            # to zone or not to zone that is the question
            if do_zoning:
                info['data']['initiator_target_map'] = \
                    _create_zones(self, volume, connector, itm)
        except Exception as ex:
            LOG.exception(ex)
            with excutils.save_and_reraise_exception():
                LOG.info(_("Creation of zones failed. "
                           "Terminate the connection to the volume"))
                try:
                    self.terminate_connection(*args, **kwargs)
                except Exception as ex2:
                    LOG.info(_("Cleanup of volume connection failed"))
                    LOG.exception(ex2)
        LOG.info(_("AddFCZone returning connection info %(c_info)s "
                   "for volume %(vol)s and connector %(con)s") %
                 dict(c_info=info, vol=volume['id'], con=connector))
        return info

    return decorator


def ignore_zoning_request(connection_request):
    """Decorator used to ignore FC Zoning decorator."""
    def decorator(self, *args, **kwargs):
        info = connection_request(self, *args, **kwargs)
        return info
    return decorator


def _add_colons(wwpn):
    return ':'.join([wwpn[i:i + 2] for i in range(0, len(wwpn), 2)]).lower()


def _strip_colons(wwpn):
    wwpn = wwpn.lower()
    if len(wwpn) > 16:
        return wwpn.replace(":", "")
    else:
        return wwpn


def _get_fabric_mapping(zone_manager, phy_initiator_wwpns,
                        target_wwpns):
    # Given a list of physical initiator wwpns, and a list of target wwpns,
    # asks our registered SAN lookup service to tell is which of those
    # wwpns are visible on which fabrics.
    LOG.info(_('Requested mapping for initiators  %(initiators)s '
               'and targets %(targets)s')
             % {'initiators': phy_initiator_wwpns,
                'targets': target_wwpns})

    # The brocade code only works if we add colons to the incoming WWPNs.
    initiator_colons = map(_add_colons, phy_initiator_wwpns)
    target_colons = map(_add_colons, target_wwpns)

    # Produce a mapping that shows which of the specified initiator and
    # target wwpns are logged into which switches.
    #               'a': [...]}
    #               'b': [...]}
    wwpns = {'initiator_port_wwn_list': initiator_colons,
             'target_port_wwn_list': target_colons}

    mapping = zone_manager.get_san_context(wwpns)

    LOG.info(_("Returned mapping is %(mapping)s")
             % {'mapping': mapping})

    # Remove any colons that might be in the returned structure
    for fabric_name, wwpn_lists in mapping.iteritems():
        for name, wwpns in wwpn_lists.iteritems():
            wwpn_lists[name] = \
                map(_strip_colons, wwpn_lists[name])

    LOG.info(_("Returned mapping is %(mapping)s")
             % {'mapping': mapping})

    return mapping


def _build_zone_map(self, zonemanager, connector, initiator_target_map):
        """
        Construct a Zone map to interact with the fabrics
        """
        LOG.info("_build_zone_map")

        # Sanitize the initiator_targe_map to be lower case, no colons
        new_it_map = {}
        i_wwpns = []
        for initiator, targets in initiator_target_map.items():
            new_it_map[_strip_colons(initiator)] = map(_strip_colons, targets)
            i_wwpns.append(initiator)
        initiator_target_map = new_it_map

        # Extract the map of physical-to-virtual WWPNs, sanitising to all
        # lowercase with no colons, and also build a list of physical WWPNs.

        if 'phy_to_virt_initiators' in connector:
            phy_to_virt_initiators = connector['phy_to_virt_initiators']
        else:
            # phy_to_virt_initiators might None for the novalink
            # and will be None for vscsi
            # create phys to virt mapping as a virt to virt mapping
            phy_to_virt_initiators = dict([(I, [I]) for I in i_wwpns])
            LOG.info(_("Expect the connector wwpns are active without "
                     "phy to virt mapping. phy to virt simulated "
                     "%s " % phy_to_virt_initiators))

        phy_to_virt_map = {}
        for phy, virt in phy_to_virt_initiators.items():
            phy_to_virt_map[phy.lower()] = map(lambda x: x.lower(), virt)
        phy_initiators = phy_to_virt_map.keys()

        target_wwpns = set()
        for targets in initiator_target_map.values():
            for target in targets:
                target_wwpns.add(_strip_colons(target))

        mapping = _get_fabric_mapping(zonemanager, phy_initiators,
                                      list(target_wwpns))

        # We now know which physical initiators and targets are on which fabric
        # We need to filter the initiator_target_map

        zones = {}
        LOG.info(_("Mapping is %(mapping)s\n"
                   "phy_to_virt_map is %(phy_to_virt_map)s\n"
                   "initiator_target_map is %(initiator_target_map)s")
                 % locals())

        for fabric_name, wwpn_lists in mapping.iteritems():

            # Make a set of all the target ports on this fabric.
            target_fabric_set = set(wwpn_lists['target_port_wwn_list'])

            # For every physical initiator found on the fabric...
            for phy_initiator in wwpn_lists['initiator_port_wwn_list']:

                # ...for every virtual initiator that will be hosted by that
                # physical initiator...
                for virt_initiator in phy_to_virt_map[phy_initiator]:

                    # Get the list of target ports that we might want to
                    # zone to
                    if virt_initiator in initiator_target_map:
                        target_zone_set = set(initiator_target_map
                                              [virt_initiator])

                        # Filter the ports to only those on the fabric
                        filtered_target_ports = (target_zone_set &
                                                 target_fabric_set)

                        # If there is at least one target port in the
                        # same fabric as the initiator
                        if filtered_target_ports:
                            if fabric_name not in zones:
                                zones[fabric_name] = {}

                            zones[fabric_name][virt_initiator] = \
                                sorted(list(filtered_target_ports))

        return zones


def build_index_of_initiators(connector, zone_map):
    """
    Build an dict of virtual wwpns that create an ordered index that
    will allow us to associate the pair of virtual wwpns with the same
    target wwpn. This is important during LPM so the targets don't change
    as that is one of the VIOS LPM check conditions.
    """
    i_index = 0
    i_node_index = {}
    if 'phy_to_virt_initiators' in connector:
        p2v = connector['phy_to_virt_initiators']
        for key, wwpns in p2v.iteritems():
            for wwpn in wwpns:
                i_node_index[wwpn.lower()] = i_index
            i_index += 1
    else:
        for wwpn in connector['wwpns']:
            i_node_index[wwpn.lower()] = i_index
            i_index += 1
    i_index_list = [-1]*(i_index)
    # For each fabric process the initiators to create an index that will
    # eventually be used as input into the selection of a target wwpn.
    # Initiators accociated with the same physical wwpn will have the same
    # index.
    for fabric in zone_map:
        i_index = 0
        # For each initiator wwpn in the zone map of the fabric, establish
        # an index, where the index number is the same for all the virtual
        # initiators belongs to the same physical wwpn.
        for i_wwpn in sorted(zone_map[fabric]):
            # Establish an index for the initiator if one has not already
            # been established.
            if i_index_list[i_node_index[i_wwpn]] == -1:
                i_index_list[i_node_index[i_wwpn]] = i_index
                i_index += 1
    return i_index_list, i_node_index


def _map_i_wwpn_2_t_wwpn(connector, zone_map):
    """
    For each initiator wwpn in the connector, map that initiator wwpn to a
    single back-end storage target wwpn. This is in-line with expected
    zoning policy for the average PowerVM FC user.
    The pair of virtual initiator wwpns are mapped to the same target
    wwpn, so the targets wwpns on an LPM operation don't change.
    """
    i_index_list, i_node_index = build_index_of_initiators(connector, zone_map)

    # we hash the lowest wwpn in the connector
    # as we don't expect this to ever change between
    # attachments this gives us a mapping index into the list
    # of target wwpns so that we can have one initiator mappped to
    # one target.
    map_index = hash(sorted(connector['wwpns'])[0])

    # loop thru each fabric and pick one target wwpn from the list of
    # target wwpns to connect to this initiator. Both of the virtual
    # initiators in the phys 2 virt mapping will be mapped to the same
    # target.
    for fabric in zone_map:
        #LOG.debug("fabric %s" % fabric)
        i_index = 0
        for i_wwpn in sorted(zone_map[fabric]):
            t_wwpns = zone_map[fabric][i_wwpn]
            wwpn_index = i_index_list[i_node_index[i_wwpn]]
            target_index = ((map_index + wwpn_index) %
                            len(zone_map[fabric][i_wwpn]))
            new_target_wwpns = sorted(zone_map[fabric][i_wwpn])
            zone_map[fabric][i_wwpn] = [new_target_wwpns[target_index]]


def _create_zones(self, volume, connector, initiator_target_map):
        """
        Construct a Zone manager to interact with the fabrics
        """
        LOG.debug("Enter _create_zones")
        zonemanager = PowerVCZoneManager(self.configuration)

        zones = _build_zone_map(self,
                                zonemanager,
                                connector,
                                initiator_target_map)

        # Does the driver have a special routine to override the selection
        # of the initiator to target map
        if hasattr(self, '_override_zone_map'):
            LOG.debug("Driver has _override_zone_map")
            self._override_zone_map(volume,
                                    connector,
                                    zones)
            LOG.debug("Driver override_zone_map zones %s" % zones)
        else:
            # Map each initiator to only one target which is the expected
            # behavior by most PowerVM customers.
            _map_i_wwpn_2_t_wwpn(connector, zones)
        # Merge the zone list across all fabrics.
        new_initiator_target_map = {}
        for fabric_name in zones:
            for i_wwpn in zones[fabric_name]:
                if i_wwpn in new_initiator_target_map:
                    new_initiator_target_map[i_wwpn].\
                        append(zones[fabric_name][i_wwpn])
                else:
                    new_initiator_target_map[i_wwpn] = \
                        zones[fabric_name][i_wwpn]

        if not new_initiator_target_map:
            msg = (_('The virtual machine does not have '
                     'connectivity to the storage. The zone '
                     'map created from the initiator_target_map '
                     'is empty for volume %(vol)s with '
                     'connector %(con)s and '
                     'initiator_target_map %(itm)s') %
                   dict(con=connector, vol=volume['id'],
                        itm=initiator_target_map))
            raise exception.VolumeBackendAPIException(data=msg)

        # Now that we have a data structure describing the zones, zone it.
        LOG.info(_("Creating zones: %(zones)s initiator_target_map "
                   "%(itm_map)s") %
                 dict(zones=zones, itm_map=new_initiator_target_map))
        host = connector['host']
        zonemanager.add_connection(host, zones)
        zonemanager.close()

        return new_initiator_target_map


def RemoveFCZone(terminate_connection):
    """Decorator for FC drivers to remove zone."""
    def decorator(self, *args, **kwargs):
                # Validate the connector.
        volume = args[0]
        connector = args[1]
        if ('connection_type' not in connector or
                not connector['connection_type']):
            if self.configuration.zoning_mode == 'fabric':
                msg = (_('terminate_connection: '
                         'connector missing '
                         'connection_type. Volume %(vol)s '
                         'Connector %(con)s') %
                       dict(con=connector, vol=volume['id']))
                LOG.info(msg)
                do_zoning = True
            else:
                do_zoning = False
        elif 'npiv' in connector['connection_type']:
            do_zoning = True
        elif 'pv_vscsi' in connector['connection_type']:
            # to zone for vscsi or not to zone ...
            zone_vscsi = self.configuration.safe_get('zone_vscsi')
            if zone_vscsi:
                do_zoning = True
            else:
                do_zoning = False
        else:
            # We had a connection_type so we must be powerVM
            # so we must be be either vscsi or npiv
            msg = (_('terminate_connection: Invalid connector, '
                     'connection type unsupported '
                     'on volume %(vol)s connector %(con)s') %
                   dict(con=connector, vol=volume['id']))
            raise exception.VolumeBackendAPIException(data=msg)

        try:
            info = terminate_connection(self, *args, **kwargs)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.exception(ex)

        if not info:
            LOG.debug(_("No connection info returned by Volume driver. "
                        "No Zoning to remove! "
                        "Volume %(vol)s Connector %(con)s") %
                      dict(con=connector, vol=volume['id']))
            return info

        # can only remove zones if the config type is fabric
        vol_type = info.get('driver_volume_type', None)
        if vol_type != 'fibre_channel':
            LOG.info(_("RemoveCZone driver_volume_type %(vtype)s "
                       "Volume %(vol)s "
                       "Connector %(con)s Connection info %(c_info)s") %
                     dict(vtype=vol_type, con=connector, vol=volume,
                          c_info=info))
            return info

        # Have we been told to remove zones?
        if 'data' not in info or 'initiator_target_map' not in info['data']:
            LOG.info(_("RemoveFCZone no initiator_target_map Volume %(vol)s "
                       "Connector %(con)s %(c_info)s") %
                     dict(con=connector, vol=volume, c_info=info))
            return info

        LOG.debug("RemoveFCZone %s" % info)
        # to remove or not remove zones that is the question.
        if do_zoning:
            LOG.info(_("Removing zones for initiators: "
                       "%(initiator_list)s") % {'initiator_list': ',\
                       '.join(connector['wwpns'])})
            # Create a zonemanager and ask it to remove zones for these
            # initiators
            try:
                _remove_zones(self.configuration,
                              info['data']['initiator_target_map'])
            except Exception as ex:
                LOG.exception(ex)
                LOG.error(_("There was a failure when the zone manager "
                            "attempted to remove zones created for WWPNs "
                            "%(wwpns)s during detach of volume %(vol)s. "
                            "Connection info %(c_info)s connector %(conn)s") %
                          dict(wwpns=connector['wwpns'], vol=volume['id'],
                               conn=connector, c_info=info))
        LOG.info(_("RemoveFCZone returning connection info %(c_info)s "
                   "for volume %(vol)s with connector %(con)s") %
                 dict(c_info=info, vol=volume['id'], con=connector))
        return info

    return decorator


def _remove_zones(config, itm):
    zonemanager = PowerVCZoneManager(config)
    zonemanager.delete_connection(itm)
    zonemanager.close()
