# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Registers or adds a Storage Provider or Fabric using a Pluggable Driver.
   There are really 4 operations provided as part of this command:
     1) add - Registers or adds the Storage Provider or Fabric
     2) edit - Updates the registration for the Storage Provider or Fabric
     3) remove - Unregisters or removes the Storage Provider or Fabric
     3) list - Lists all of the Storage Providers or Fabrics registered
"""
import os
import sys
import argparse
import commands
import ConfigParser
from datetime import datetime

from oslo_utils import encodeutils
from cinderclient import client as cinderclient
from cinderclient import exceptions

from powervc_discovery.locale import _

_stdout, _stderr = sys.stdout, sys.stderr


#################################################
# #########  Storage Operation Methods ##########
#################################################
def _add_storage(parser, display_name, driver,
                 new_name, properties_file, force):
    """Registers or adds the Storage Provider to PowerVC"""
    client = _get_cinder_client()
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    _verify_required_parameter(parser, driver, 'driver')
    _verify_required_parameter(parser, properties_file, 'properties_file')
    # Parse the Properties File into the Configuration Properties dictionary
    config_props = _parse_config_file(properties_file)
    # Make sure the specified name doesn't already exist for anything managed
    _get_storage_provider(client, display_name, 'add')
    # Call the Rest API to add the Storage Provider to PowerVC
    reg = dict(host_display_name=display_name, volume_driver=driver,
               host_type='generic', configuration_properties=config_props)
    _restreq(client.post, '/os-hosts', body=dict(host=dict(registration=reg)))
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The Storage Provider '%s' added to PowerVC.") % display_name)


def _edit_storage(parser, display_name, driver,
                  new_name, properties_file, force):
    """Updates or edits the Storage Provider managed by PowerVC"""
    client, reg = (_get_cinder_client(), dict())
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    # Parse the Properties File into the Configuration Properties dictionary
    config_props = _parse_config_file(properties_file)
    # Query the Name for the Storage Provider if they gave us the Display Name
    host_name = _get_storage_provider(client, display_name, 'edit')
    # If a New Display Name was specified, pass them into the Edit API
    if new_name:
        reg['host_display_name'] = new_name
    # If any Configuration Properties were specified, pass them into the API
    if config_props:
        reg['configuration_properties'] = config_props
    # Call the Rest API to update the Storage Provider in PowerVC
    _restreq(client.put, '/os-hosts/' + host_name +
             '//update-registration', body=dict(registration=reg))
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The Storage Provider '%s' updated in PowerVC.") % display_name)


def _remove_storage(parser, display_name, driver,
                    new_name, properties_file, force):
    """Unregisters or removes the managed Storage Provider from PowerVC"""
    client = _get_cinder_client()
    # Tack on the Force API Suffix if they asked to force remove it
    force = '/force' if unicode(force).lower() == 'true' else ''
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    # Query the Name for the Storage Provider if they gave us the Display Name
    host_name = _get_storage_provider(client, display_name, 'remove')
    # Call the Rest API to remove the Storage Provider from PowerVC
    _restreq(client.delete, '/os-hosts/' + host_name + force)
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The Storage Provider '%s' removed from PowerVC.") % display_name)


def _list_storage(parser, display_name, driver,
                  new_name, properties_file, force):
    """Lists the Storage Providers that are managed by PowerVC"""
    client = _get_cinder_client()
    providers = _get_storage_providers(client)
    # Print out the list of Storage Providers in a user-friendly format
    _print_resources(providers, 24)


#################################################
# #########  Fabric Operation Methods  ##########
#################################################
def _add_fabric(parser, display_name, driver,
                new_name, properties_file, force):
    """Registers or adds the SAN Fabric to PowerVC"""
    client = _get_cinder_client()
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    _verify_required_parameter(parser, driver, 'driver')
    _verify_required_parameter(parser, properties_file, 'properties_file')
    # Parse the Properties File into the Configuration Properties dictionary
    config_props = _parse_config_file(properties_file)
    # Make sure the specified name doesn't already exist for anything managed
    _get_san_fabric(client, display_name, 'add')
    # Call the Rest API to add the SAN Fabric to PowerVC
    reg = dict(fabric_display_name=display_name, zone_driver=driver,
               fabric_type='generic', configuration_properties=config_props)
    _restreq(client.post, '/san-fabrics',
             body=dict(fabric=dict(registration=reg)))
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The SAN Fabric '%s' added to PowerVC.") % display_name)


def _edit_fabric(parser, display_name, driver,
                 new_name, properties_file, force):
    """Updates or edits the SAN Fabric managed by PowerVC"""
    client, reg = (_get_cinder_client(), dict())
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    # Parse the Properties File into the Configuration Properties dictionary
    config_props = _parse_config_file(properties_file)
    # Query the Name for the SAN Fabric if they gave us the Display Name
    fabric_name = _get_san_fabric(client, display_name, 'edit')
    # If a New Display Name was specified, pass them into the Edit API
    if new_name:
        reg['fabric_display_name'] = new_name
    # If any Configuration Properties were specified, pass them into the API
    if config_props:
        reg['configuration_properties'] = config_props
    # Call the Rest API to update the SAN Fabric in PowerVC
    _restreq(client.put, '/san-fabrics/' + fabric_name,
             body=dict(registration=reg))
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The SAN Fabric '%s' updated in PowerVC.") % display_name)


def _remove_fabric(parser, display_name, driver,
                   new_name, properties_file, force):
    """Unregisters or removes the managed SAN Fabric from PowerVC"""
    client = _get_cinder_client()
    # Make sure that they specified the required parameters on input
    _verify_required_parameter(parser, display_name, 'display_name')
    # Query the Name for the SAN Fabric if they gave us the Display Name
    fabric_name = _get_san_fabric(client, display_name, 'remove')
    # Call the Rest API to remove the SAN Fabric from PowerVC
    _restreq(client.delete, '/san-fabrics/' + fabric_name)
    # Print a statement to standard out, so the caller knows it worked
    stdout(_("The SAN Fabric '%s' removed from PowerVC.") % display_name)


def _list_fabric(parser, display_name, driver,
                 new_name, properties_file, force):
    """Lists the SAN Fabrics that are managed by PowerVC"""
    client = _get_cinder_client()
    fabrics = _get_san_fabrics(client)
    # Print out the list of Storage Providers in a user-friendly format
    _print_resources(fabrics, 10)


def _fill_text(self, text, width, indent):
    """Override the Help Formatter to allow new lines"""
    _HelpFormatter = argparse.HelpFormatter
    if text.find('\n') > 0:
        return ''.join([indent + line for line in text.splitlines(True)])
    return _HelpFormatter._fill_text2(self, text, width, indent)


#################################################
# ##########  Common Utility Methods  ###########
#################################################
def _parse_arguments():
    """Parses the arguments from the command line that was input"""
    argparse.HelpFormatter._fill_text2 = argparse.HelpFormatter._fill_text
    argparse.HelpFormatter._fill_text = _fill_text
    parser = argparse.ArgumentParser(
        description=_("The powervc-register command is used to register or "
                      "add a new Storage Provider or Fabric to be managed by "
                      "PowerVC.  This command provides a mechanism to have "
                      "PowerVC manage any OpenStack Volume/Zone Driver in a "
                      "generic fashion.  It takes as input the various "
                      "configuration parameters to configure a new cinder-"
                      "volume service in OpenStack using the driver for that "
                      "type of Storage Provider or Fabric."),
        epilog=_("exit status:\n  0  successful \n  2  missing either the "
                 "operation or resource type parameter\n  3  invalid resource "
                 "type parameter value\n  4  invalid operation parameter value"
                 "\n  5  no resource managed with the given name\n  6  resourc"
                 "e already managed with the given name\n  7  more than one "
                 "resource managed with the given name\n  8  only pluggable "
                 "drivers can be edited\n  9  missing a required parameter\n  "
                 "10 the properties file specified does not exist\n  11 the "
                 "user does not have permission to the properties file\n  12 "
                 "error parsing the properties file\n  13 could connect to "
                 "the openstack-cinder-api service\n  14 error calling the "
                 "calling the resource registration api"),
        usage="powervc-register [-h]\n       "
              "powervc-register -o add -r RESOURCE_TYPE -d DRIVER\n       "
              "                        -n DISPLAY_NAME -p PROPERTIES_FILE"
              "\n       "
              "powervc-register -o edit -r RESOURCE_TYPE -n DISPLAY_NAME"
              "\n       "
              "                         [-e NEW_NAME] [-p PROPERTIES_FILE]"
              "\n       "
              "powervc-register -o remove -r RESOURCE_TYPE -n DISPLAY_NAME"
              "\n       "
              "powervc-register -o list -r RESOURCE_TYPE")
    parser.add_argument('-o', '--operation', dest='operation',
                        help=_("specifies the operation to perform (the "
                               "supported operations are 'list', 'add', "
                               "'edit' and 'remove')"))
    parser.add_argument('-r', '--resource', dest='resource_type',
                        help=_("specifies the type of resource to register "
                               "(only 'storage' and 'fabric' are supported)"))
    parser.add_argument('-d', '--driver', dest='driver',
                        help=_("specifies the python class name of the "
                               "volume/zone driver for the storage provider "
                               "or fabric"))
    parser.add_argument('-n', '--name', dest='display_name',
                        help=_("specifies a user-friendly name given "
                               "to the storage provider or fabric"))
    parser.add_argument('-e', '--newname', dest='new_name',
                        help=_("specifies a new user-friendly name to give "
                               "to the storage provider or fabric on edit"))
    parser.add_argument('-p', '--properties', dest='properties_file',
                        help=_("specifies a properties file containing the "
                               "name/value pairs for the configuration "
                               "properties"))
    parser.add_argument('-f', '--force', dest='force',
                        default='false', help=argparse.SUPPRESS)
    return parser, parser.parse_args()


def _get_storage_provider(client, name, operation):
    """Helper method to get lookup a Storage Providers based on name"""
    providers = _get_storage_providers(client)
    return _find_matching_resource(
        providers, _("Storage Provider"), name, operation)


def _get_san_fabric(client, name, operation):
    """Helper method to get lookup a Storage Providers based on name"""
    frabrics = _get_san_fabrics(client)
    return _find_matching_resource(
        frabrics, _("SAN Fabric"), name, operation)


def _get_storage_providers(client):
    """Helper method to get a list of the Storage Providers managed"""
    providers = list()
    for host in _restreq(client.get, '/os-hosts')[1]['hosts']:
        if host['service'] != 'cinder-volume':
            continue
        reg = _find_registration(
            client.get('/os-hosts/' + host['host_name'])[1])
        providers.append(
            (reg['host_name'], reg['host_display_name'], reg['host_type']))
    return providers


def _get_san_fabrics(client):
    """Helper method to get a list of the Storage Providers managed"""
    fabrics = list()
    for sw in _restreq(client.get, '/san-fabrics/detail')[1]['fabrics']:
        fabrics.append(
            (sw['fabric_name'], sw['fabric_display_name'], sw['fabric_type']))
    return fabrics


def _find_matching_resource(resources, resource_type, name, operation):
    """Helper method to find the resource in the list with the matching name"""
    match_list, info = (list(), dict(type=resource_type, name=name))
    # Loop through the List of Resources trying to find matches based on name
    for resource in resources:
        if resource[0] == name or resource[1] == name:
            match_list.append(resource)
    # Unless this is an add, error out if the name isn't found
    if len(match_list) <= 0 and operation != 'add':
        stderr(_("No %(type)s is managed with name '%(name)s'.") % info)
        sys.exit(5)
    # If this is an add, error out if one already exists with that name
    if len(match_list) > 0 and operation == 'add':
        stderr(_("A %(type)s is already managed with name '%(name)s'.") % info)
        sys.exit(6)
    # We aren't able to select the right one if there is more than one match
    if len(match_list) > 1:
        stderr(_("More than 1 %(type)s exists with display "
                 "name '%(name)s'.  Use the unique name "
                 "returned from the list operation instead.") % info)
        sys.exit(7)
    # If this isn't a Generic Registrar case, we can't let them edit it
    if operation == 'edit' and match_list[0][2] != 'generic':
        stderr(_("Only a %(type)s registered for Pluggable "
                 "Drivers can be edited with this command.") % info)
        sys.exit(8)
    return match_list[0][0] if match_list else None


def _find_registration(host):
    """Helper method to parse out the registration from the response body"""
    for item in host['host']:
        if 'registration' in item:
            return item['registration']
    return None


def _print_resources(resources, padding):
    """Prints the list of Resources out in a user-friendly format"""
    _format = "%-" + str(padding) + "s %s"
    stdout(_format % (_("Name"), _("Display Name")))
    stdout(_format % ("====", "============"))
    for resource in resources:
        stdout(_format % (resource[0], resource[1]))


def _parse_config_file(properties_file):
    """Helper method to parse the CONF File into a dictionary of properties"""
    parser, props = (ConfigParser.RawConfigParser(), dict())
    # If the Properties File wasn't specified on Edit, just return
    if not properties_file:
        return dict()
    # If the Properties File didn't exist, error out since can't continue
    if os.path.exists(properties_file) is False:
        stderr(_("The properties file '%s' does not exist") % properties_file)
        sys.exit(10)
    # If the Properties File didn't exist, error out since can't continue
    if _has_permission(properties_file) is False:
        stderr(_("The user does not have permission "
                 "to properties file '%s'") % properties_file)
        sys.exit(11)
    try:
        # Generate a name for a temporary file if we need to use it
        temp_file = '/tmp/' + os.path.basename(properties_file)
        temp_file += '_' + str(datetime.now().microsecond)
        try:
            # Now we can parse the Properties File and return the output
            parser.read(properties_file)
        except ConfigParser.MissingSectionHeaderError:
            # If it isn't in INI format, then we need to write it out as INI
            with open(temp_file, 'w') as outf:
                with open(properties_file, 'r') as inf:
                    outf.write('[DEFAULT]\n')
                    # Copy all of the lines from the original file to new one
                    for line in inf.readlines():
                        outf.write(line)
            # Now we can try reading the new temporary file again
            parser.read(temp_file)
        # Cleanup the temporary file we created when we are done
        os.remove(temp_file) if os.path.exists(temp_file) else None
    except Exception as exc:
        stderr(_("Error parsing the properties file - %s") % unicode(exc))
        os.remove(temp_file) if os.path.exists(temp_file) else None
        sys.exit(12)
    # Parse all of the Default Properties from the Configuration File
    for key in parser.defaults():
        props[key] = parser.get('DEFAULT', key)
    # Next we need to add all the Key/Values from the other Sections
    for section in parser.sections():
        # We only want the options that explicitly in the given Section
        for key in parser._sections[section]:
            if key != '__name__':
                props[section + ':' + key] = parser.get(section, key)
    return props


def _verify_required_parameter(parser, attr, attr_name):
    """Helper method to verify the required parameter was provided"""
    if not attr:
        stderr(_("The required parameter '%s' was not specified.") % attr_name)
        parser.print_usage()
        sys.exit(9)


def _restreq(method, url, body=None):
    """Helper method to call a REST API while handling exceptions that occur"""
    try:
        kwargs = dict(body=body) if body is not None else dict()
        return method(url, **kwargs)
    except exceptions.ConnectionError as connexp:
        stderr(_("Could not connect to the openstack-"
                 "cinder-api service - %s") % unicode(connexp))
        sys.exit(13)
    except exceptions.ClientException as cliexp:
        stderr(unicode(cliexp.message))
        sys.exit(14)
    except Exception as exp:
        stderr(_("Unexpected error calling the "
                 "registration api - %s") % unicode(exp))
        sys.exit(14)


def _get_cinder_client():
    """Helper method to get a Cinder Client to use for Rest API calls"""
    auth_token, project_id = _get_auth_token()
    clientv2 = cinderclient.Client(
        2, '', '', project_id=project_id, tenant_id=project_id, insecure=True)
    clientv2.client.management_url = "http://localhost:9000/v2/" + project_id
    clientv2.client.auth_token = auth_token
    return clientv2.client


def _has_permission(afile):
    """Helper method to make sure the user has permission to the given file"""
    permission = False
    try:
        with open(afile, 'r'):
            permission = True
    except:
        pass
    return permission


def _get_auth_token():
    """Helper method to get the Authentication Token and Project ID to use"""
    output = commands.getoutput('/usr/bin/powervc-get-token --project')
    # In case there is any logging done, make sure to get the last line only
    output = output.strip().split('\n')[-1].split()
    return output[0], output[1]


def stderr(astr):
    """Helper method to print a string to standard error"""
    _stderr.write(encodeutils.safe_encode(astr) + '\n')


def stdout(astr):
    """Helper method to print a string to standard error"""
    _stdout.write(encodeutils.safe_encode(astr) + '\n')


#################################################
# #############  Main Entry Point  ##############
#################################################
def main():
    """Main invocation for the powervc-manage command"""
    # Parse the arguments that were provided on the command line
    parser, args = _parse_arguments()
    # Make sure they passed in at least the Resource Type and Operation
    if not args.resource_type or not args.operation:
        parser.print_usage()
        sys.exit(2)
    # Make sure they specified one of the supported Resource Types
    if args.resource_type not in ['storage', 'fabric']:
        stderr(_("The resource type specified must be "
                 "either 'storage' or 'fabric'."))
        parser.print_usage()
        sys.exit(3)
    # Make sure they specified one of the supported Operations
    if args.operation not in ['add', 'edit', 'remove', 'list']:
        stderr(_("The operation specified must be "
                 "'add', 'edit', 'remove' or 'list'."))
        parser.print_usage()
        sys.exit(4)
    # Dynamically invoke the correct method based on the type/operation
    mthd = getattr(sys.modules[__name__], "_%s_%s" %
                   (args.operation, args.resource_type))
    mthd(parser, args.display_name, args.driver,
         args.new_name, args.properties_file, args.force)
