# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import tarfile
import logging
from powervc_oslo.config import data_mgr as mgr
from powervc_oslo.config import data_utils as utils
from powervc_discovery.locale import _

IMG_PATH = '/opt/ibm/powervc/images/%s/'
IMG_PREFIX = 'powervc-%s-compute-'
MANAGE_CMD = '/opt/ibm/powervc/bin/powervc-manage'

LOG = logging.getLogger(__name__)


######################################################
# #######  Discovery Data Manager Definition  ########
######################################################
class DiscoveryDataManager(mgr.BaseDataManager):

    def get_component_name(self):
        """Provides the name of the Component providing the Data Manager"""
        return 'discovery'

    def is_supported_environment(self):
        """Specifies whether the Component is supported in this Environment"""
        return utils.is_management_system()

    def get_dependent_managers(self):
        """Specifies the Component Name of the Managers it is dependent on"""
        return ['db2', 'rabbitmq', 'httpd', 'keystone', 'cinder', 'nova']

    ###############################################
    # ## Data Manager - System Changed Interface ##
    ###############################################
    def management_ip_changed(self, upgrading):
        """Updates the Management IP in the CONF files when they change"""
        mgmt_ip = utils.get_config_data()['mgmt_ip']
        # We need to update the old mgmt_ip property since TTV relies on it
        for host in self.__get_remote_hosts():
            utils.set_openstack_config_value(
                '/etc/nova/nova-%s.conf' % host, 'mgmt_ip', mgmt_ip)

    def management_system_changed(self, upgrading):
        """Performs Component-specific deltas when Management System changes"""
        # For now we just do what we did in the passed, and set a flag in the
        # conf file so the periodic task in nova-api does the Remote Host work
        for host in self.__get_remote_hosts():
            utils.set_openstack_config_value(
                '/etc/nova/nova-%s.conf' % host, 'restored_new_system', 'True')
            # If we don't need to upgrade first, then re-manage the Remote Host
            if not upgrading:
                self._remanage_remote_host(host)

    ###############################################
    # ####  Data Manager - Upgrade Interface  #####
    ###############################################
    def upgrade_component(self, version):
        """Performs Upgrade specific to the given Component"""
        # Loop through each of the Remote Hosts doing a re-manage of the Host
        for host in self.__get_remote_hosts():
            new_system = utils.get_openstack_config_value(
                '/etc/nova/nova-%s.conf' % host, 'restored_new_system')
            new_system = new_system is not None
            # Run an un-manage first, so that we un-manage on the old version
            # We only want to the un-manage/re=manage if the system changed
            if new_system:
                self._unmanage_remote_host(host)
            # Then we can upgrade the Remote Host and re-manage if needed
            if self._update_remote_host(host, not new_system) and new_system:
                self._remanage_remote_host(host)

    def _update_remote_host(self, host, restart=False):
        """Updates the Remote Host to the latest version available"""
        creds = self.__get_remote_host_credentials(host)
        ivers = self.__get_latest_remote_image_version(creds)
        image = self.__get_matching_remote_image(creds, ivers)
        # Try to get the PowerVC version from the Remote Host
        output = utils.run_ssh_command(
            creds, '/bin/cat /opt/ibm/powervc/version.properties')[1]
        hvers = utils.get_property_value(output, 'version')
        # If the version was retrieved and older, then update it
        if hvers and utils.compare_version(ivers, hvers) > 0:
            info = dict(host=creds['name'], vers=hvers)
            LOG.log(utils.CONSOLE, _("Upgrading the Remote Host %(host)s "
                                     "from version %(vers)s....") % info)
            # If we got an error updating the Remote Host, log to the console
            if self.__update_remote_host_image(creds, image, ivers) != 0:
                LOG.log(utils.CONSOLE, _("Failed to upgrade the Remote "
                                         "Host %s.") % creds['name'])
                return False
            # If we were asked to restart the services, do that now
            if restart:
                svccmd = '/opt/ibm/powervc/bin/powervc-services restart'
                utils.run_ssh_command(creds, svccmd, requires_root=True)
        return True

    ###############################################
    # #####  Data Manager - Backup Interface  #####
    ###############################################
    def prepare_component_restoration(self):
        """Performs any Component-specific preparation needed for restore"""
        config, hosts = utils.get_config_data(), self.__get_remote_hosts()
        # Nothing to do if there aren't any remote hosts registered
        if len(hosts) <= 0:
            return
        # Read the list of file names that are inside of the restore file
        rstdir, rstfile = config.get('restore_dir'), config.get('restore_file')
        with tarfile.open(rstdir + '/' + rstfile) as f:
            filenames = f.getnames()
        # Loop through the hosts and if any don't exist in the restore
        # archive, then we want to un-manage and delete those host files
        for host in self.__get_remote_hosts():
            conf_file = '/etc/nova/nova-%s.conf' % host
            # If the file isn't in the archive, then remove it
            if conf_file[1:] not in filenames:
                ipcreds = utils.get_remote_host_credentials(conf_file)
                self._unmanage_remote_host(host, ipcreds)
                utils.remove_file(conf_file)

    ###############################################
    # ### Internal Data Manager Helper Methods ####
    ###############################################
    @staticmethod
    def get_remotehost_certificates(csr_dir):
        """Helper method to get the Certificates to transfer to Remote Host"""
        certs = list()
        for manager in mgr.get_registered_managers():
            ca_info = manager._get_certificate_authority_info()
            cert_info = manager._get_remotehost_certificate_info()
            # If there are any CA Certificates to transfer, add them now
            if cert_info and cert_info.get('cacert'):
                certs.append((cert_info['cacert'][0], cert_info['cacert'][1],
                              ca_info.get('component')))
            # Loop through the CSR's, adding any of those to the transfer list
            if cert_info and cert_info.get('csrs'):
                cert_dir = cert_info['certdir']
                for csr in cert_info['csrs']:
                    cert_file = ca_info['component'] + '-' + csr['certfile']
                    certs.append((csr_dir + '/' + cert_file,
                                  cert_dir + '/' + csr['certfile'],
                                  ca_info['component']))
        return certs

    def __get_remote_hosts(self):
        """Helper method to provide a list of Remote Hosts managed"""
        # Get a list of each of the Remote Host Nova Proxy Configuration Files
        files = utils.get_matching_files('/etc/nova', 'nova-', '.conf')
        return [f[5:-5] for f in files if self.__is_remote_host_conf(f)]

    def __is_remote_host_conf(self, conf_file):
        """Helper method to determine if the CONF file is for a remote host"""
        full_path = conf_file.startswith('/etc/nova')
        conf_file = conf_file if full_path else ('/etc/nova/' + conf_file)
        # Query the Remote Credentials to use to determine if Proxy CONF
        config = utils.get_remote_host_credentials(conf_file)
        return config.get('ip') is not None and config.get('user') is not None

    def __get_remote_host_credentials(self, host):
        """Helper method to provide a list of Remote Hosts managed"""
        conf_file = '/etc/nova/nova-%s.conf' % host
        # Use the Data Manager Utility to get the IP/Credential Information
        creds = utils.get_remote_host_credentials(conf_file)
        # Add in the extra properties such as the Display Name and Storage Type
        extras = dict(name='host_display_name', domain='dhcp_domain')
        extras['ephemeral_path'] = 'kvm_host_ephemeral_path'
        for key, val in extras.iteritems():
            creds[key] = utils.get_openstack_config_value(conf_file, val)
        return creds

    def __get_image_file_prefix(self, creds):
        """Helper method to determine get the Image Path and Prefix"""
        if utils.get_offering() == 'powervm':
            return IMG_PATH % 'powervm', IMG_PREFIX % 'powervm'
        # If this is KVM then we need to determine 2.1 versus 3.1
        output = utils.run_ssh_command(creds, '/bin/cat /etc/issue')[1].strip()
        version = 'powerkvm31' if output.find('3.1') > 0 else 'powerkvm'
        return IMG_PATH % 'kvm', IMG_PREFIX % version

    def __get_matching_remote_image(self, creds, version):
        """Helper method to find the Host Image that matches the version"""
        img_path, img_prefix = self.__get_image_file_prefix(creds)
        image_file = img_path + img_prefix + str(version) + '.tgz'
        # If the image file exists, return it else return None for nonexistent
        return image_file if utils.file_exists(image_file) else None

    def __get_latest_remote_image_version(self, creds):
        """Helper method to find the latest available Host Image"""
        latest_version = None
        img_path, img_prefix = self.__get_image_file_prefix(creds)
        # Loop through each of the matching file names, finding the latest
        file_names = utils.get_matching_files(img_path, img_prefix, '.tgz')
        for file_name in file_names:
            file_version = file_name[len(img_prefix):len(file_name) - 4]
            # If the version is the greatest yet, then use that version
            if utils.compare_version(file_version, latest_version) > 0:
                latest_version = file_version
        return latest_version

    def __update_remote_host_image(self, creds, image, version):
        """Helper method to update the image on the given Remote Host"""
        imgfile, rc = image[image.rfind('/') + 1:], 1
        # Make the directory on the remote system to transfer the image to
        todir = utils.create_remote_unique_temp_directory(creds)
        tofile = todir + '/' + imgfile
        # If the transfer succeeded then we want to do the install
        if utils.sftp_transfer_file(creds, image, tofile) == 0:
            untar_cmd = '/bin/tar -xf ' + todir + '/%s -C ' + todir
            inst_dir = (todir + '/powervc-%s/') % version
            # First untar the image zip, then run the installer command
            utils.run_ssh_command(creds, untar_cmd % imgfile)
            rc = utils.run_ssh_command(
                creds, inst_dir + 'install', requires_root=True)[0]
        utils.run_ssh_command(creds, '/bin/rm -rf ' + todir)
        return rc

    def _unmanage_remote_host(self, host, creds=None):
        """Helper method to un-manage the Remote Host when MgmtSys changes"""
        uncmd = '%s -o unmanage' % MANAGE_CMD
        mgmtip = utils.get_config_data()['mgmt_ip']
        # Query the Remote Host to figure out the Management IP used
        creds = creds if creds else self.__get_remote_host_credentials(host)
        old_mgmtip = self.__get_remote_mgmt_ip(creds)
        # Tack on the old IP to notify if it was a different Management Server
        if old_mgmtip and old_mgmtip != mgmtip:
            uncmd += ' -n %s' % mgmtip
        # Try to un-manage the existing Mgmt System from the Remote Host
        return utils.run_ssh_command(creds, uncmd, requires_root=True)[0]

    def _remanage_remote_host(self, host):
        """Helper method to re-manage the Remote Host when MgmtSys changes"""
        creds = self.__get_remote_host_credentials(host)
        # Create the temporary directory on the local and remote system
        lcldir = utils.create_unique_temp_directory()
        rmtdir = utils.create_remote_unique_temp_directory(creds)
        # Try to un-manage the existing Mgmt System from the Remote Host
        if self._unmanage_remote_host(host) != 0:
            LOG.warn(_("The un-manage failed for %s") % host)
            return
        # Next try to generate the CSR's on the Remote Host
        csrcmd = '%s -o gencsr -f %s' % (MANAGE_CMD, rmtdir)
        utils.run_ssh_command(creds, csrcmd, requires_root=True)
        # Next try to transfer the signing request to the Mgmt System
        rc = utils.sftp_transfer_directory(creds, lcldir, rmtdir, True)
        if rc != 0:
            LOG.warn(_("The CSR transfer failed for %s") % host)
            utils.run_ssh_command(creds, '/bin/rm -rf ' + rmtdir)
            utils.remove_directory(lcldir)
        # Next try to generate the signed certificates for the Remote Host
        crtcmd = '%s -o gencert -f %s' % (MANAGE_CMD, lcldir)
        if utils.run_command(crtcmd) != 0:
            LOG.warn(_("The Certificate Signing failed for %s") % host)
            utils.run_ssh_command(creds, '/bin/rm -rf ' + rmtdir)
            utils.remove_directory(lcldir)
        # Next try to transfer the signed certificates back to the Remote Host
        for fromcert, tocert, comp in self.get_remotehost_certificates(lcldir):
            tocert = fromcert[fromcert.rfind('/')+1:]
            if comp and tocert.find(comp) < 0:
                tocert = comp + '-' + tocert
            tocert = rmtdir + '/' + tocert
            utils.sftp_transfer_file(creds, fromcert, tocert)
        # Create the Management Information File to transfer to the Remote Host
        dmpini = '/mgmtinfo.ini'
        utils.run_command('%s -o dump -f %s' % (MANAGE_CMD, lcldir + dmpini))
        self.__add_extra_mgmt_info(creds, host, lcldir + dmpini)
        utils.sftp_transfer_file(creds, lcldir + dmpini, rmtdir + dmpini)
        utils.remove_directory(lcldir)
        # Finally we will try to manage the Host with the new Mgmt System
        mancmd = '%s -o manage -f %s' % (MANAGE_CMD, rmtdir + dmpini)
        rc = utils.run_ssh_command(creds, mancmd, requires_root=True)[0]
        utils.run_ssh_command(creds, '/bin/rm -rf ' + rmtdir)
        # If the Manage succeeded than we can say we restored successfully
        if rc == 0:
            LOG.info(_("Successfully re-managed the Remote Host %s") % host)
            utils.del_openstack_config_value(
                '/etc/nova/nova-%s.conf' % host, 'restored_new_system')

    def __get_remote_mgmt_ip(self, creds):
        """Helper method get the Management IP used by the remote host"""
        try:
            output = utils.run_ssh_command(
                creds, '/bin/cat /etc/nova/nova.conf', False, True, True)[1]
            return utils.get_property_value(output, 'management_ip')
        except Exception as exc:
            LOG.info(_("Error retrieving Remote Management IP: %s") % exc)

    def __add_extra_mgmt_info(self, creds, host, mgmt_ini):
        """Helper method to add Extra Management Info to the File"""
        set_config = utils.set_openstack_config_value
        dispname = creds.get('name') or host
        storage = creds.get('storage') or 'local'
        path = creds.get('ephemeral_path') or '/var/lib/libvirt/images'
        set_config(mgmt_ini, 'host', host)
        set_config(mgmt_ini, 'dhcp_domain', creds.get('domain'))
        set_config(mgmt_ini, 'host_display_name', dispname)
        set_config(mgmt_ini, 'host_storage_type', storage)
        set_config(mgmt_ini, 'instances_path', path + '/nova/instances')
