# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to On-board VM's for a Compute Host"""

import webob.exc
from nova import policy
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from nova.objects import block_device
from oslo_log import log as logging

from powervc_nova.db import api as powervc_db
from powervc_nova.objects.compute import dom as compute_dom

from powervc_discovery.onboarding import importer_messages as msg
from powervc_discovery.onboarding.compute import vm_importer
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
ALIAS = 'host-onboard'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'host_onboard')


class HostOnboardControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for On-boarding VM's"""

    @wsgi.action("index")
    def index(self, req, host_name_id):
        """Add Rest GET Method to list On-boarding Tasks"""
        tasks = []
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'index')
        # Parameter for showing only running tasks
        running = False
        if 'running' in req.GET and req.GET['running'] == 'true':
            running = True
        # Retrieve the list of On-boarding Tasks from the Database
        result = powervc_db.onboard_task_find_all(context, host_name_id)
        for task in result:
            task_uri = self._get_task_uri(req, task['id'])
            if(not running or (running and task['status'] != 'completed' and
                               task['status'] != 'failed')):
                tasks.append({'id': task['id'], 'links': [{'href': task_uri}]})
        return {'tasks': tasks}

    @wsgi.action("show")
    def show(self, req, host_name_id, id):
        """Add Rest GET Method to show a specified On-boarding Task"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'show')
        # Retrieve the list of On-boarding Tasks from the Database
        result = powervc_db.onboard_task_find(context, int(id))
        # If the Task wasn't found in the Database, then throw an exception
        if result is None or result['host'] != host_name_id:
            text = msg.BaseImporterMessageBundle().\
                get_message_text(context, msg.NO_IMPORT_TASK, taskid=id)
            raise webob.exc.HTTPNotFound(explanation=text)
        # Construct the Task dictionary to return in the REST response
        task = {'id': result['id']}
        task['links'] = [{'href': req.url}]
        task['servers'] = []
        # Loop through all the attributes on the task we want to include
        task_attrs = ['status', 'progress', 'started', 'ended']
        for attr in task_attrs:
            task[attr] = result[attr]
        # Add in the list of VM Results into the overall Task entry
        result_vms = result.get('servers', [])
        for result_vm in result_vms:
            vm = {'id': result_vm['server_uuid']}
            vm['name'] = result_vm['server_name']
            vm['status'] = result_vm['status']
            if result_vm.get('fault_message'):
                vm['fault'] = result_vm['fault_message']
            task['servers'].append(vm)
        return {'task': task}

    @wsgi.action("create")
    def onboard(self, req, host_name_id, body):
        """Add Rest POST Method to On-board VM's for a given Host"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'import')
        # Parse out the key attributes from the request body
        servers = body.pop('servers', None)
        preparation = body.pop('preparation', {})
        synch = body.pop('synchronous', False) is True
        unsupported = body.pop('allow_unsupported', False) is True
        # Delegate to the VM Importer to actually perform the on-boarding
        importer = vm_importer.VmImporter(context, host_name_id)
        response = importer.import_resources(servers, preparation, body, synch,
                                             unsupported)
        # If this is asynchronous, then return the Task Identifier Info
        if not synch:
            # Build the Task URI (to use for progress) to return the response
            task_uri = self._get_task_uri(req, response)
            return {'task': {'id': response, 'links': [{'href': task_uri}]}}
        # Otherwise we will get the Values back on the Synchronous call
        return {'servers': response}

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def _get_task_uri(req, task_id):
        """Internal Helper Method to format the URI for the On-boarding Task"""
        return ('%s/%s' % (req.url, str(task_id))).replace('?running=true', '')


class HostUnmanageControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for Unmanaging VM's"""

    @wsgi.action("create")
    @wsgi.response(204)
    def unmanage(self, req, host_name_id, body):
        """Add Rest POST Method to Unmanage VM's for a given Host"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'unmanage')
        force = body.get('force', False)
        # Delegate to the VM Importer to actually perform the unmanage
        importer = vm_importer.VmImporter(context, host_name_id)
        importer.unmanage_resources(body.get('servers'), bool(force))

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class HostVmListControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for listing VM's"""

    @wsgi.action("index")
    def index(self, req, host_name_id):
        """Add Rest GET Method to list existing VM's for a given Host"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'list_servers')
        # Delegate to the VM Importer to actually perform the listing of vm's
        importer = vm_importer.VmImporter(context, host_name_id)
        # Allow querying the details for internal usage
        if 'details' in req.params and bool(req.params['details']) is True:
            vms = importer._query_resource_details(['*all'], {}, True)
            return {'servers': vms}
        vms = importer.discover_resources()
        return {'servers': vms}

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class ServerFinishOnboardControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for listing VM's"""

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Rest GET Method to add the Pending On-board Actions attribute to"""
        ctxt = req.environ['nova.context']
        # Add the Pending On-board Actions attribute to the given Server
        self._add_pending_onboard_attribute(ctxt, req, resp_obj.obj['server'])
        self._add_finish_onboard_attributes(ctxt, req, resp_obj.obj['server'])

    @wsgi.extends
    def detail(self, req, resp_obj):
        """Rest GET Method to add the Pending On-board Actions attributes to"""
        context = req.environ['nova.context']
        # Add the Pending On-board Actions attribute to each of the Servers
        for server in resp_obj.obj['servers']:
            self._add_pending_onboard_attribute(context, req, server)
            self._add_finish_onboard_attributes(context, req, server)

    @wsgi.extends
    def delete(self, req, id):
        """REST DELETE Method to verify that the Server can be deleted"""
        context = req.environ['nova.context']
        self._verify_valid_action(context, id, 'delete')
        yield

    @wsgi.extends(action='createImage')
    def capture(self, req, id, body):
        """REST POST Method to verify that the Server can be captured"""
        context = req.environ['nova.context']
        self._verify_valid_action(context, id, 'capture')
        yield

    @wsgi.extends(action='resize')
    def resize(self, req, id, body):
        """REST POST Method to verify that the Server can be resized"""
        context = req.environ['nova.context']
        self._verify_valid_action(context, id, 'resize')
        yield

    @wsgi.action('finishOnboard')
    @wsgi.response(204)
    def finish_onboard(self, req, id, body):
        """REST POST Method to finish the on-boarding process for the Server"""
        # Parse the on-boarding info out of the request body
        onboard_info = body.get('finishOnboard')
        self._post_boot_volumes(req, id, body, onboard_info)

    @wsgi.action('editBootVolumes')
    @wsgi.response(204)
    def edit_boot_volumes(self, req, id, body):
        """REST POST Method to edit the boot volume."""
        onboard_info = body.get('editBootVolumes')
        self._post_boot_volumes(req, id, body, onboard_info)

    def _post_boot_volumes(self, req, id, body, onboard_info):
        context = req.environ['nova.context']
        image_os_distro = onboard_info.get('os_distro')
        image_endianness = onboard_info.get('endianness')
        boot_volume_id = onboard_info.get('boot_volume_id')
        boot_volume_ids = onboard_info.get('boot_volume_ids')
        attach_order = onboard_info.get('attach_order')
        use_ae = onboard_info.get('use_activation_engine', True)
        # If they gave us the old-style boot_volume_id, use it
        if boot_volume_ids is None and boot_volume_id is not None:
            boot_volume_ids = [boot_volume_id]
        # If attach_order is not specified, use boot_volume_ids instead
        if not attach_order:
            attach_order = boot_volume_ids
        # Look up the host name for the VM that was specified
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        db_inst = instfact.find_instance_by_uuid(context, id)
        # Verify that the user has Authorization to Update the VM
        policy.enforce(context, 'compute:update', db_inst)
        context = context.elevated()
        # Delegate to the VM Importer to actually finish the on-boarding
        importer = vm_importer.VmImporter(context, db_inst['host'])
        importer.finish_import(db_inst, image_os_distro, image_endianness,
                               boot_volume_ids, attach_order, use_ae)

    @staticmethod
    def _verify_valid_action(context, inst_uuid, action):
        """Internal Helper Method to verify that the Action is Valid"""
        power_specs = dict()
        try:
            # Retrieve the Instance that was specified from the Database
            specs = powervc_db.instance_power_specs_find_all(
                context, dict(instance_uuid=inst_uuid))
            power_specs = specs[0] if specs else dict()
        # If we couldn't get the Instance, let the main class throw exception
        except Exception:
            return
        # If the action is still pending finishing on-boarding, throw exception
        actions = power_specs.get('pending_onboard_actions')
        pending_actions = actions.split(',') if actions is not None else []
        if action in pending_actions:
            msgb = msg.BaseImporterMessageBundle()
            text = msgb.get_message_text(context, msg.INVALID_PENDING_ACTIONS,
                                         instuuid=inst_uuid, action=action)
            raise webob.exc.HTTPBadRequest(explanation=text)

    @staticmethod
    def _add_pending_onboard_attribute(context, req, server):
        """Internal Helper Method to add the Pending On-board Attribute"""
        try:
            # Retrieve the cached Server DB Instance for the given ID
            db_inst = req.get_db_instance(server['id'])
            power_specs = db_inst.get('power_specs')
            power_specs = power_specs if power_specs is not None else dict()
            pending_onboard = power_specs.get('pending_onboard_actions')
            # If the Pending On-board Actions is set on the Server Instance
            if pending_onboard is not None and len(pending_onboard) > 0:
                pending_onboard = pending_onboard.split(',')
                server['pending_onboard_actions'] = pending_onboard
        # If we got an exception, just let the Server API still succeed
        except Exception as exc:
            LOG.warn(_("Error retrieving Server "
                       "DB Instance: %s") % unicode(exc))

    @staticmethod
    def _add_finish_onboard_attributes(context, req, server):
        """Internal Helper Method to add attributes for finish onboarding"""
        try:
            # Retrieve the cached Server DB Instance for the given ID
            db_inst = req.get_db_instance(server['id'])
            metadata = db_inst.get('system_metadata')
            metadata = metadata if metadata is not None else dict()
            image_endianness = metadata.get('image_endianness')
            image_os_distro = metadata.get('image_os_distro')
            image_use_ae = metadata.get('image_configuration_strategy')
            # If the image_endianness Attribute is set on the Server Instance
            if image_endianness:
                server['endianness'] = image_endianness
            # If the image_os_distro Attribute is set on the Server Instance
            if image_os_distro:
                server['os_distro'] = image_os_distro
            # Only want to return the use_activiation_engine attribute if set
            if image_use_ae == 'NEEDED':
                server['use_activation_engine'] = True
        # If we got an exception, just let the Server API still succeed
        except Exception as exc:
            LOG.warn(_("Error retrieving Server "
                       "DB Instance: %s") % unicode(exc))

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class ServerAttachmentControllerExtension(wsgi.Controller):
    """Controller Class to extend the volume-attachments API for VM's"""

    @wsgi.extends
    def index(self, req, resp_obj, server_id):
        """Rest GET Method to add the Boot Index Attribute to"""
        context = req.environ['nova.context'].elevated()
        bdms = block_device.BlockDeviceMappingList.\
            get_by_instance_uuid(context, server_id)
        bdms = dict([(bdm['volume_id'], bdm) for bdm in bdms])
        # Add in the Boot Index for the Volume Attachments queried
        for attachment in resp_obj.obj['volumeAttachments']:
            self.__add_boot_index(bdms, attachment)

    @wsgi.extends
    def show(self, req, resp_obj, server_id, id):
        """Rest GET Method to add the Boot Index Attribute to"""
        context = req.environ['nova.context'].elevated()
        bdms = block_device.BlockDeviceMappingList.\
            get_by_instance_uuid(context, server_id)
        bdms = dict([(bdm['volume_id'], bdm) for bdm in bdms])
        # Add in the Boot Index for the Volume Attachment queried
        self.__add_boot_index(bdms, resp_obj.obj['volumeAttachment'])

    def __add_boot_index(self, bdms, attachment):
        """Helper method to add the Boot Index in for the attachment"""
        if attachment['id'] in bdms:
            attachment['bootIndex'] = bdms[attachment['id']]['boot_index']


class HostOnboarding(extensions.ExtensionDescriptor):
    """Capability to On-board VM's for a Compute Host."""
    name = "Compute Host On-boarding"
    alias = "ibm-host-onboarding"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_host_onboarding/api/v1.2"
    updated = "2013-07-21T21:00:00-06:00"

    def get_controller_extensions(self):
        controller1 = ServerFinishOnboardControllerExtension()
        controller2 = ServerAttachmentControllerExtension()
        extension1 = extensions.ControllerExtension(
            self, 'servers', controller1)
        extension2 = extensions.ControllerExtension(
            self, 'os-volume_attachments', controller2)
        return [extension1, extension2]

    def get_resources(self):
        """Provide a Sub-Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'onboard', controller=HostOnboardControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource2 = extensions.ResourceExtension(
            'unmanage', controller=HostUnmanageControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource3 = extensions.ResourceExtension(
            'all-servers', controller=HostVmListControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        return [resource1, resource2, resource3]


class HostOnboardingV21(extensions.V21APIExtensionBase):
    """Capability to On-board VM's for a Compute Host."""
    name = "Compute Host On-boarding"
    alias = "ibm-host-onboarding"
    version = 1

    def get_controller_extensions(self):
        controller1 = ServerFinishOnboardControllerExtension()
        controller2 = ServerAttachmentControllerExtension()
        extension1 = extensions.ControllerExtension(
            self, 'servers', controller1)
        extension2 = extensions.ControllerExtension(
            self, 'os-volume_attachments', controller2)
        return [extension1, extension2]

    def get_resources(self):
        """Provide a Sub-Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'onboard', controller=HostOnboardControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource2 = extensions.ResourceExtension(
            'unmanage', controller=HostUnmanageControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource3 = extensions.ResourceExtension(
            'all-servers', controller=HostVmListControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        return [resource1, resource2, resource3]
