# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""On-boarding Classes for importing and unmanaging Virtual Machines"""

import uuid
import hashlib

from neutron.common.constants import DEVICE_OWNER_DHCP
from nova import db
from nova import quota
from nova import exception
from nova import network as network_api
from nova import notifications as notifyutil
from nova import rpc as rpc_api
from nova import volume as volume_api
from nova import utils as base_utils
from nova.compute import flavors
from nova.compute import utils
from nova.network import base_api
from nova.objects import flavor as flavor_obj
from nova.objects import network_request
from oslo_concurrency import lockutils
from oslo_serialization import jsonutils
from oslo_log import log as logging
from oslo_utils import timeutils
from nova.network import neutronv2
from nova.volume import cinder

from powervc_nova import compute
from powervc_nova.db import api as powervc_db
from powervc_nova.network.ibmpowervm import adapter_mapping as mapping_task
from powervc_nova.network.neutronv2 import ibmpowervm_api as neutron_api
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.network import dom as network_dom
from powervc_nova.objects.storage import dom as storage_dom

from powervc_discovery.onboarding import importer
from powervc_discovery.onboarding import importer_messages as msg
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)


########################################################
# ##########  VM On-boarding Implementation  ###########
########################################################
class VmImporter(importer.BaseImporter):
    """On-boarding Classes for importing and unmanaging Virtual Machines"""

    def __init__(self, context, hostname, **kwargs):
        """Constructor for Importer Class for VM On-boarding"""
        super(VmImporter, self).__init__(context, hostname, **kwargs)
        self.compute_api = compute.HostAPI()
        self.network_api = neutron_api.PowerVMAPI()
        self.network_client = neutronv2.api.get_client(context)
        self.secgroup = self._get_default_security_group()
        # Retrieve the Host from the DB to use Attributes later
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        self.compute_host = hostfact.find_host_by_name(context, hostname)
        self._provider_list = None

    ###############################################
    # ###  Implementing BaseImporter Interface  ###
    ###############################################
    def _get_resource_id_attribute(self):
        """Provides the name of the attribute for the Resource Identifier"""
        return 'uuid'

    def _get_resource_type_string(self):
        """Provides the translated Resource Type for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.RSCTYPE_VM)

    def _get_resource_types_string(self):
        """Provides the translated Resource Types for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.RSCTYPE_VMS)

    def _get_system_type_string(self):
        """Provides the translated System Type for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTYPE_HOST)

    def _get_host_config_file(self):
        """Provides the name of the Config File created for this Host"""
        return '/etc/nova/nova-' + self.host + '.conf'

    def _get_host_resource(self):
        """Retrieves the Host Resource from the DB to make sure it exists"""
        if not self.compute_host or not self.compute_host.hypervisor_hostname:
            return None
        return self.compute_host

    def _query_existing_resources(self, filters):
        """Discovers/Queries the Existing Resources on the Managed System"""
        # First verify the Service is up so we don't want for the long timeout
        self._verify_service_running()
        # If the Service is up, then we can discover the VM's from it
        vm_data = self.compute_api.discover_instances(self.context, self.host)
        return self._handle_instance_chunking(vm_data)

    def _query_managed_resources(self):
        """Queries the existing Managed Resources from the Database"""
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        return instfact.find_all_instances_by_host(self.context, self.host)

    def _query_resource_details(self, resource_ids,
                                extra_parms, allow_unsupported):
        """Queries Details about the Resources from the Managed System"""
        # First verify the Service is up so we don't wait for the long timeout
        self._verify_service_running()
        # If the Service is up, then we can query the VM's from it
        vm_data = self.compute_api.query_instances(
            self.context, self.host, resource_ids)
        return self._handle_instance_chunking(vm_data)

    def _verify_service_running(self):
        """Verifies that the Service is running to Query Resources from"""
        # Verify that the Nova-Compute Service is running to Query VM's
        if not self.compute_api.verify_host_running(self.context, self.host):
            service = 'nova-compute-' + self.host
            self._throw_server_error(msg.SERVICE_DOWN, service=service)

    def _prepare_resource(self, resource, extra_parms):
        """Prepares the specified resource so it can be onboarded"""
        return self.compute_api.prepare_instance(self.context, self.host,
                                                 resource, extra_parms)

    @lockutils.synchronized('import_resource', 'powervc-', True)
    def _import_resource(self, resource, resource_prep, preparation,
                         extra_parms):
        """On-boards the specified resource into the Management System"""
        LOG.info(_("Importing Instance %s") % str(resource))
        LOG.info("Importing instance resource prep %s " % resource_prep)
        anotifier = rpc_api.get_notifier('compute', self.host)
        # Parse out the Ports/PowerSpecs to be used later in the Import
        resource.pop('volumes', [])
        ports = resource.pop('ports', [])
        flavor = resource.pop('flavor', {})
        volume_connector = resource.pop('volume_connector', None)
        # If the resource was prepared, modify the ports list to inject vswitch
        if preparation:
            resource_id = self._get_resource_id(resource)
            if resource_id in preparation.get('servers', []):
                vswitch = preparation.get('network_vswitch')
                for port in ports:
                    # If the user gave a vswitch to use, we will use that one
                    if vswitch:
                        port['vswitch'] = vswitch
                        port['physical_network'] = vswitch
        # First verify the Attached Volumes for the VM were imported in Cinder
        volumes = self._verify_attached_volumes(resource, resource_prep)
        # If no volumes means the Storage Provider query failed, so abort
        # import
        if volumes is None:
            return None
        # Handle the concurrent on-board case where one may have just imported
        filters = dict(vm_uuid=resource.get('power_specs', {}).get('vm_uuid'))
        if powervc_db.instance_power_specs_find_all(self.context, filters):
            LOG.info(_("Skipping instance %s because already exists, likely "
                       "due to concurrent requests.") % resource['new_uuid'])
            instance = dict(uuid=resource['new_uuid'])
            bdms = db.block_device_mapping_get_all_by_instance(
                self.context, instance['uuid'])
            volids = [b['volume_id'] for b in bdms if b.get('volume_id')]
            # Only cleanup the volumes that weren't part of the other VM
            bdms = [dict(volume_id=v['id']) for
                    v in volumes if v['id'] not in volids]
            self._cleanup_attached_volumes(instance, bdms)
            return dict()
        # Setup all the default values on the Instance
        vm = self._prep_instance_for_import(resource)
        # Import the Flavor into the DB first, so we have the Flavor ID
        self._import_instance_flavor(vm, flavor, resource)
        # Need to set image_os_distro so that health/migration could
        # distinguish between ibmi and other platforms.
        power_specs = resource.get('power_specs', {})
        if (power_specs.get('lpar_type') == 'OS400'):
            vm['system_metadata']['image_os_distro'] = 'ibmi'
        # Build the Notification Info manually, to add the System MetaData
        system_metadata = vm['system_metadata']
        info = notifyutil.info_from_instance(self.context, vm, None, None)
        info.update(system_metadata=system_metadata)
        # Send a notification to listeners that we are importing this VM
        anotifier.info(self.context, 'compute.instance.import.start', info)
        # Reserve the Quota Reservation saying we are going to add Usage
        reserv = quota.QUOTAS.reserve(self.context,
                                      project_id=vm['project_id'],
                                      instances=1, cores=vm['vcpus'],
                                      ram=vm['memory_mb'])
        try:
            # Create the VM in the DB and increment the Usage Count
            vm.save()
            quota.QUOTAS.commit(self.context, reserv)
        except Exception:
            # If the VM wasn't created, roll-back the increment
            quota.QUOTAS.rollback(self.context, reserv)
            raise
        uuid = vm.get('uuid')
        # Now that the VM is inserted, can insert the Volume Attachment Records
        self._import_volume_attachments(vm, volumes, volume_connector)
        # Now we can Import the Allocated Networks for the VM into Neutron
        self._import_allocated_networks(vm, ports)
        # Call the finish_onboard method on the host
        self.compute_api.finish_onboard(
            self.context, self.host, vm, preparation)
        # Build the Notification Info manually, to add the System MetaData
        info = notifyutil.info_from_instance(self.context, vm, None, None)
        info.update(uuid=uuid, system_metadata=system_metadata)
        # Send a notification to listeners that we imported this VM
        anotifier.info(self.context, 'compute.instance.import.end', info)
        # Associate the Instance with the Default Security Group
        db.instance_add_security_group(self.context, uuid, self.secgroup)
        # We will just run the Instance UUID for now, since nobody is using
        return dict(id=vm['uuid'])

    def _unmanage_resource(self, resource, extra_params=None):
        """Stops managing the specified resource by removing it from the DB"""
        LOG.info(_("Unmanaging Instance %s") % str(resource.get('uuid')))
        anotifier = rpc_api.get_notifier('compute', self.host)
        # Retrieve the Meta-data about the Instance from the DB to use
        system_meta = resource.get('system_metadata', {}).copy()
        # Send a notification to listeners that we are deleting this vm
        utils.notify_about_instance_usage(anotifier, self.context,
                                          resource, "delete.start")
        # Reserve the Quota Reservation saying we are going to remove Usage
        reserv = quota.QUOTAS.reserve(self.context,
                                      project_id=resource['project_id'],
                                      instances=-1,
                                      cores=-resource['vcpus'],
                                      ram=-resource['memory_mb'])
        # Try to Remove from the Database any Networks that are Allocated
        self._cleanup_allocated_networks(resource)
        # Try to Remove from the Database any Volumes that are Attached
        self._cleanup_attached_volumes(resource)
        try:
            # Update the VM's state in the DB to say it has been Deleted
            updates = dict(vm_state='soft-delete', task_state=None,
                           cleaned=True, terminated_at=timeutils.utcnow())
            resource.update(updates)
            resource.save()
            # Remove the VM instance from the Database
            resource.delete()
            # Commit the decrement of the Usage Count for this Instance
            quota.QUOTAS.commit(self.context, reserv, resource['project_id'])
        except exception.InstanceNotFound:
            # If the Instance was deleted in the middle, roll-back the
            # decrement
            quota.QUOTAS.rollback(self.context, reserv, resource['project_id'])
        # Send a notification to listeners that we deleted this vm
        utils.notify_about_instance_usage(anotifier, self.context,
                                          resource, "delete.end",
                                          system_metadata=system_meta)

    def _import_resources_preparation(self, resources):
        """Perform any preparation needed prior to importing the resources"""
        return self._import_attached_volumes(resources)

    def _create_onboard_task(self):
        """Creates an On-boarding Task in the DB to use for this Action"""
        task = powervc_db.onboard_task_create(self.context, self.host)
        anotifier = rpc_api.get_notifier('compute', self.host)
        anotifier.info(self.context, 'compute.task.import.create',
                       dict(host=self.host, task_id=task['id']))
        return task

    def _update_onboard_task(self, task_id, status, progress=None):
        """Updates an On-boarding Task in the DB to use for this Action"""
        if task_id is not None:
            info = dict(status=status, progress=progress)
            task = powervc_db.onboard_task_update(self.context, task_id, info)
            info.update(**dict(host=self.host, task_id=task_id))
            anotifier = rpc_api.get_notifier('compute', self.host)
            anotifier.info(self.context, 'compute.task.import.update', info)
            return task

    def _create_onboard_taskrsc(self, task_id, resource_id,
                                resource_name, status, fault_message=None):
        """Creates an On-boarding Task Record in the DB for this Resource"""
        if task_id is not None:
            values = dict(server_name=resource_name,
                          status=status, fault_message=fault_message)
            return powervc_db.onboard_task_server_create(
                self.context, task_id, resource_id, values)

    def _update_onboard_taskrsc(self, task_id, resource_id,
                                status, fault_message=None):
        """Updates an On-boarding Task Record in the DB for this Resource"""
        if task_id is not None:
            values = dict(status=status, fault_message=fault_message)
            return powervc_db.onboard_task_server_update(
                self.context, task_id, resource_id, values)

    def _send_console_message(self, priority, message):
        """Shared Helper method to send a Message to the Console Log"""
        anotifier = rpc_api.get_notifier('compute', self.host)
        txt = discovery_locale.translate_message(self.context, message)
        if priority == 'INFO':
            anotifier.info(self.context, 'compute.instance.log', {'msg': txt})
        elif priority == 'WARN':
            anotifier.warn(self.context, 'compute.instance.log', {'msg': txt})

    ###############################################
    # ###  Implementing VmImporter Interface  #####
    ###############################################
    def finish_import(self, resource, image_os_distro, image_endianness,
                      boot_volume_ids, attach_order, use_ae):
        """Allows the caller to specify the Boot Volume to finish import"""
        anotifier = rpc_api.get_notifier('compute', self.host)
        # Retrieve values from the resource
        updates, uuid = (dict(), resource['uuid'])
        power_specs = resource.get('power_specs', {}).copy()
        system_metadata = resource.get('system_metadata', {}).copy()
        actions = power_specs.get('pending_onboard_actions')
        pend_actions = actions.split(',') if actions is not None else []
        # Retrieve all of the Volume Attachments that are in the Nova DB
        bdms = db.block_device_mapping_get_all_by_instance(self.context, uuid)
        # If they gave us Boot Volumes, update it to reflect those Boot Volumes
        if boot_volume_ids is not None and len(boot_volume_ids) > 0:
            # Selecting Boot Volume for Local or NFS Storage is not allowed
            if self._is_local_storage():
                self._throw_bad_request(msg.BOOT_FOR_LOCAL)
            # Update the Boot Volume to be whichever one was specified
            self._update_boot_volumes(uuid, boot_volume_ids, updates,
                                      bdms, attach_order)
            # Remove the Delete and Resize operations from the Pending Actions
            for action in ['delete', 'resize']:
                if action in pend_actions:
                    pend_actions.remove(action)
        # If they gave us an OS Distro Type, update it on the VM instance
        if image_os_distro is not None:
            system_metadata['image_os_distro'] = image_os_distro
            updates['system_metadata'] = system_metadata
            # If we had the Distro Type remove Capture from the pending actions
            # We are ignoring if the Boot Volume is set to handle Local Storage
            if 'capture' in pend_actions:
                pend_actions.remove('capture')
        if image_endianness is not None:
            system_metadata['image_endianness'] = image_endianness
            updates['system_metadata'] = system_metadata
        # If they asked to use the Activation Engine, set the CfgStrategy
        if self.compute_host.hypervisor_type in ['powervm', 'phyp']:
            if str(use_ae).lower() == 'true':
                system_metadata['image_configuration_strategy'] = 'NEEDED'
            else:
                system_metadata.pop('image_configuration_strategy', None)
            updates['system_metadata'] = system_metadata
        # If there were some updates, then update the Instance in the DB
        if len(updates) > 0:
            actions = ','.join(pend_actions) if len(pend_actions) > 0 else None
            # We also want to update the Pending Actions in the Database
            updates['power_specs'] = dict(pending_onboard_actions=actions)
            resource.update(updates)
            resource.save()
            # We also need to send a notification to the UI for the change
            info = notifyutil.info_from_instance(
                self.context, resource, None, None)
            info.update(system_metadata=resource['system_metadata'])
            anotifier.info(self.context, 'compute.instance.update', info)

    ###############################################
    #     Internal On-boarding Helper Methods     #
    ###############################################
    def _handle_instance_chunking(self, vm_data):
        """Internal Helper Method to Handle Chunking for RPC Calls"""
        instance_list = []
        # If this is a partial list due to chunking, get the next chunk
        while vm_data['chunking'] is True:
            instance_list.extend(vm_data['servers'])
            # Make an RPC call to the Compute Manager to get the next chunk
            vm_data = self.compute_api.get_next_instances_chunk(
                self.context, self.host, vm_data['identifier'])
        # Add the final list of VM's to the overall list to be returned
        instance_list.extend(vm_data['servers'])
        return instance_list

    def _get_default_security_group(self):
        """Internal Helper Method to get the Default Security Group from DB"""
        db.security_group_ensure_default(self.context)
        return db.security_group_get_by_name(
            self.context, self.context.project_id, 'default')['id']

    def _is_local_storage(self):
        """Internal Helper method to determine if this is host local-storage"""
        return self.compute_host.host_storage_type in ['local', 'nfs']

    def _only_generic_storage(self):
        """Internal Helper method to determine if all Storage is local"""
        try:
            providers = self._find_storage_providers()
            # If there weren't any Providers, then they aren't all Generic
            if len(providers) <= 0:
                return False
            # Check to see if any of the Providers are not Generic Providers
            for prov in providers:
                if prov.host_type != 'generic':
                    return False
            return True
        # If we got an exception, we will assume something other than Generic
        except:
            return False

    def _prep_instance_for_import(self, resource):
        """Internal Helper Method to Prepare the VM for Import"""
        instance = dict(resource, id=None)
        instance['uuid'] = resource.pop('new_uuid', str(uuid.uuid4()))
        # Set all the necessary properties on the Volume that is being imported
        instance['host'] = self.host
        instance['node'] = self.compute_host.hypervisor_hostname
        instance['hostname'] = base_utils.sanitize_hostname(instance['name'])
        instance['user_id'] = self.context.user_id
        instance['project_id'] = self.context.project_id
        instance['display_name'] = instance.pop('name', None)
        instance['availability_zone'] = None
        instance['created_at'] = timeutils.utcnow()
        instance['launched_at'] = timeutils.utcnow()
        instance['image_ref'] = 'ONBOARDED_NO_IMAGE'
        instance['metadata'] = dict()
        instance['locked'] = False
        instance['progress'] = 0
        # Return the Instance object as a DOM Model object to use later
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        return instfact.construct_resource(
            compute_dom.ManagedInstance, self.context, **instance)

    def _import_instance_flavor(self, instance, flavor_dict, resource):
        """Internal Helper Method to Import the Instance Type into the DB"""
        instance['instance_type_id'] = None
        instance['system_metadata'] = {}
        # If there wasn't any flavors specified, then nothing to do here
        if not flavor_dict:
            return
        # Parse the main Flavor Attributes out the Dictionary provided
        flavor_dict.pop('vcpus', 0)
        vcpus = resource.get('vcpus', 0)
        if resource.get('system_metadata'):
            meta_dict = resource['system_metadata']
        else:
            meta_dict = {}
        root_gb = flavor_dict.pop('root_gb', 0)
        ephem_gb = flavor_dict.pop('ephemeral_gb', 0)
        memory_mb = flavor_dict.pop('memory_mb', 0)
        # We want to create a unique/consistent Flavor Name based on the Values
        text_name = ("auto-flavor-%(vcpus)s-%(memory_mb)s-"
                     "%(root_gb)s-%(ephem_gb)s" % locals())
        # Loop through the Extra Specs provided, adding them to the Name
        sorted_keys = sorted(flavor_dict.keys())
        for key in sorted_keys:
            text_name = '-'.join([text_name, key, str(flavor_dict[key])])
        # Create an MD5 Hash to use for the Name of the Flavor
        flavor_name = hashlib.md5(text_name).hexdigest()
        try:
            # Query the DB to see if there is an existing Flavor we can reuse
            inst_type = flavors.get_flavor_by_name(flavor_name, self.context)
        except exception.FlavorNotFoundByName:
            # If there isn't a matching Flavor, will have to create a new one
            inst_type = flavor_obj.Flavor(
                context=self.context,
                name=flavor_name, memory_mb=memory_mb, vcpus=vcpus,
                root_gb=root_gb, ephemeral_gb=ephem_gb,
                flavorid=flavor_name, swap=0, is_public=False,
                projects=[self.context.project_id],
                extra_specs=flavor_dict.get('extra_specs', {}))
            inst_type.create()
        # Update the Instance when reference to the Flavor information
        instance['flavor'] = inst_type
        instance['instance_type_id'] = inst_type['id']
        # Temporarily set the VCPU's/Memory in the Meta-data for an EGO issue
        instance['system_metadata']['vcpus'] = vcpus
        instance['system_metadata']['memory_mb'] = memory_mb
        # Need to set Image Properties so Scheduler knows it is PowerVM or KVM
        instance['system_metadata']['image_hypervisor_type'] =\
            self.compute_host.hypervisor_type.lower()
        instance['system_metadata']['image_architecture'] = 'ppc64'
        # Copy over the metadata dict
        for prop in meta_dict:
            instance['system_metadata'][prop] = meta_dict[prop]

    def _import_allocated_networks(self, instance, ports):
        """Internal Helper Method to Import Allocated Networks into the DB"""
        inst_uuid = instance['uuid']
        netfact = network_dom.NetworkAssociationFactory.get_factory()
        # Build a map of Networks up-front to use to see if ones already exist
        network_map = self._retrieve_existing_networks()
        # Maintain a network list, so that we allocate at once (thus
        # maintaining the instance cache)
        port_req_list = network_request.NetworkRequestList()
        # Loop through each Port adding them to the Database if they don't
        # exist, and also creating a corresponding Network if one doesn't exist
        for port in ports:
            vswitch = port.get('vswitch')
            vlanid = port.get('segmentation_id', -1)
            macaddress = port.get('mac_address')
            vlanid = str(vlanid)
            # Update the mac address format
            if len(macaddress) == 12:
                macaddress = ':'.join(a + b for a, b in zip(macaddress[::2],
                                                            macaddress[1::2]))
            # The network key is the segmentation id plus the physical network
            phys_net = port.get('physical_network', 'default')
            segment = vlanid
            if segment == "-1":
                segment = 'Flat'
            net_key = segment + ":" + phys_net
            # If the VLAN ID or MAC Address is missing we can't import the Port
            if vlanid is None or macaddress is None or vswitch is None:
                LOG.error(_("Missing required port attributes for %(uuid)s:"
                            "%(port)s") % dict(uuid=inst_uuid, port=str(port)))
                continue
            # At this point, we may need to create a network.  Save off the
            # information we'd need for that...
            name = 'Imported Network for VLAN ' + vlanid + ' - ' + vswitch
            if segment is 'Flat':
                name = 'Imported Flat Network (' + vswitch + ")"
            anetwork = {'name': name,
                        'provider:physical_network': phys_net}
            if segment is not 'Flat':
                anetwork['provider:segmentation_id'] = vlanid
                anetwork['provider:network_type'] = 'vlan'
            else:
                anetwork['provider:network_type'] = 'flat'
            # See if a suitable Network exists already for the given VLAN ID
            port_networks = network_map.get(net_key)
            port_network = None
            secondary_network = None
            # If the Network doesn't already exist then we need to create one
            if port_networks is None or len(port_networks) == 0:
                # Call the Neutron Client to create the Network for the VLAN
                port_network = self.network_client.\
                    create_network({'network': anetwork})['network']
                # Update our cached Networks to include the one we created
                network_map[net_key] = [port_network]
            elif self.compute_host.hypervisor_type in ['powervm', 'phyp']:
                # There were networks existing...try to find one on the
                # appropriate vSwitch.
                for p_net in port_networks:
                    netassn = netfact.find_network_assoc(
                        self.context, self.host, p_net['id'])
                    # If there is a network without a network association, its
                    # a candidate...but if we can find one with a proper
                    # association on this vswitch, we want to use that instead
                    if netassn is None:
                        secondary_network = p_net
                        continue
                    # Check to see if this network association has an SEA,
                    # and if it does, is that SEA no the same vSwitch
                    if netassn.sea is None:
                        continue
                    if netassn.sea.primary_vea.vswitch_name == vswitch:
                        port_network = p_net
                        break
            # If not PowerVM just want to use the first network that was found
            else:
                port_network = port_networks[0]
            # If the port_network is already set, then we're fine.  If not,
            # utilize the secondary network...but even that may not be set.
            # If that's the case, then we need to create a new network
            if port_network is None:
                if secondary_network is not None:
                    port_network = secondary_network
                else:
                    # Call the Neutron Client to create the Network
                    port_network = self.network_client.\
                        create_network({'network': anetwork})['network']
                    # Update our cached Networks to include the one we created
                    network_map[net_key] = [port_network]
            # Create a Default Host Network Association Mapping
            if self.compute_host.hypervisor_type in ['powervm', 'phyp']:
                self._create_dflt_network_assn(
                    instance['display_name'], port_network, vlanid, vswitch)
            # Need to first find an existing port (perhaps this instance is
            # being imported after being unmanaged) or create the port
            search_opts = {'mac_address': macaddress,
                           'network_id': port_network['id']}
            ports = self.network_client.list_ports(**search_opts).get('ports')
            # If we didn't find the mac above, try again in lowercase since
            # custom macs (P7 and above) are generated in lowercase (as opposed
            # to P6 macs from K2 stored to the neutron db are in uppercase)
            if ports is None or len(ports) == 0:
                macaddress = macaddress.lower()
                search_opts = {'mac_address': macaddress,
                               'network_id': port_network['id']}
                ports = self.network_client.list_ports(**search_opts)\
                    .get('ports')
            if ports is not None and len(ports) > 0:
                # Found a port, be sure it's one we can use
                orig_port = ports[0]
                # If this is already bound to an instance
                if orig_port.get('device_owner') == 'compute:None':
                    # This is not likely to happen, but to maintain parity
                    # with the original code, just skip this port.
                    continue
                # Update the port to wipe out the device_owner so
                # allocate_for_instance can set it.
                port_req_body = {'port': {}}
                port_req_body['port']['device_owner'] = ''
                port = self.network_client.update_port(orig_port.get('id'),
                                                       port_req_body)
            else:
                # No existing port found, create one.
                port_req_body = {'port': {}}
                port_req_body['port']['network_id'] = port_network['id']
                port_req_body['port']['admin_state_up'] = True
                port_req_body['port']['tenant_id'] = instance['project_id']
                port_req_body['port']['mac_address'] = macaddress
                port_req_body['port']['fixed_ips'] = []
                port = self.network_client.create_port(port_req_body)
            port_id = port['port']['id']
            # Extend the requested networks with this tuple.  port_id is all
            # to use here.
            port_req = network_request.NetworkRequest()
            port_req.port_id = port_id
            port_req.network_id = None
            port_req.address = None
            port_req_list.objects.append(port_req)
        # use the Nova Neutron API to allocate in one swoop
        allocate_args = dict(requested_networks=port_req_list)
        nw_info = self.network_api.allocate_for_instance(self.context,
                                                         instance,
                                                         **allocate_args)
        # Since the neutron port existed prior to calling allocate_for_instance
        # above, it turned on 'preserve_on_delete' for the VIFs.  Spin through
        # and turn off 'preserve_on_delete'
        if nw_info:
            for vif in nw_info:
                LOG.debug('Removing preserve_on_delete for %s' % vif)
                vif['preserve_on_delete'] = False
            # Update instance cache
            base_api.update_instance_cache_with_nw_info(self.network_api,
                                                        self.context,
                                                        instance,
                                                        nw_info)

    def _import_attached_volumes(self, instances):
        """Internal Helper Method to Import Attached Volumes into Cinder"""
        import_prep, server_info = (dict(), dict())
        inserted_volumes, missing_volumes = list(), list()
        # Loop through each of the Instances
        for instance in instances:
            # Need to know the UUID going in the DB for the Volume Attachment.
            instance['new_uuid'] = self._get_resource_id(instance)
            volumes = instance.get('volumes', [])
            # If given Volumes, build the Server Info to use to Import Volumes
            if len(volumes) > 0:
                server_info[instance['new_uuid']] =\
                    dict(name=instance['name'], volumes=volumes, boot=False)
            # If the driver gave no Volumes, then this is Local Storage
            else:
                import_prep[instance['new_uuid']] = list()
        # Log a Message specifying the Server Info provided for Volume Import
        LOG.info(_("Importing Volumes for Instances %s") % str(server_info))
        # No point trying to Import Volumes if no Volumes to Import
        if len(server_info) <= 0:
            return import_prep
        # Retrieve the list of Storage Providers registered to import against
        cinder_client = cinder.cinderclient(self.context).client
        # Loop through the Storage Provider trying to Import Applicable Volumes
        for provider in self._find_storage_provider_names():
            try:
                uri = '/os-hosts/' + provider + '/onboard'
                body = dict(server_info=server_info, volumes=['*all'],
                            synchronous=True, allow_unsupported=True,
                            local_storage=self._is_local_storage())
                # Call the on-board API to Import Volumes matching the Filter
                body = cinder_client.post(uri, body=body)[1]
                LOG.info("cinder_client onboard response %s" % body)
                # Loop through Volumes returned creating the Volume Attachment
                for volume in body['volumes']:
                    # If we found a boot volume for this VM, tag it as found
                    if volume.get('bootable'):
                        for attach in volume.get('attachments'):
                            server_dict = server_info.get(
                                attach.get('instance_uuid'))
                            if server_dict:
                                server_dict['boot'] = True
                    inserted_volumes.append(volume)
                LOG.info("Imported volumes are %(volumes)s"
                         % dict(volumes=body['volumes']))
            # For now we will just log a warning if a Volume Import fails
            except Exception as exc:
                LOG.warn(_("Volume Import Failure for %s") % provider)
                LOG.exception(exc)
        # Loop though each of the Inserted Volumes, organizing by Server
        for volume in inserted_volumes:
            # if volume.get('instance_uuid') is not None:
            for attach in volume.get('attachments'):
                instance_uuid = attach['instance_uuid']
                volumes = import_prep.get(instance_uuid)
                # If this is the first Volume for the Server, create the list
                if volumes is None:
                    volumes = list()
                    import_prep[instance_uuid] = volumes
                # sanitize volume attachments. Add only those for this instance
                sanitized_volume = self.__add_instance_attach_info(
                    instance_uuid, volume)
                volumes.append(sanitized_volume)
        # See which VM's we need to say that there was no Storage Provider for
        for instance in instances:
            if import_prep.get(instance['new_uuid']) is None:
                missing_volumes.append(instance['name'])
        # If VM's were missing Volume (and not Local), send a Console Message
        if missing_volumes and not self._is_local_storage():
            msgb = self._get_message_bundle()
            missing_volumes = self._format_string_list(missing_volumes)
            text = msgb.get_message_text(
                self.context, msg.IMPORT_NO_STORAGE, vmnames=missing_volumes)
            # If all Storage Providers are Generic, then do a different message
            if self._only_generic_storage() is True:
                text = msgb.get_message_text(
                    self.context, msg.IMPORT_ONLY_GENERIC)
            self._send_console_message('WARN', text)
        return import_prep

    def __add_instance_attach_info(self, instance_uuid, volume):
        """
        Add attachment info for that instance into the return
        value to be used by import_prep
        """
        ret_vol = dict(id=volume['id'], bootable=volume.get('bootable', False),
                       size=volume.get('size', None))

        for attach in volume.get('attachments', None):
            if attach.get('instance_uuid') == instance_uuid:
                for key in attach.keys():
                    if key != 'id':
                        ret_vol[key] = attach[key]
                break
        return ret_vol

    def _verify_attached_volumes(self, instance, resources_prep):
        """Internal Helper Method to Verify Attached Volumes were Imported"""
        volumes = resources_prep.get(instance['new_uuid'])
        # If there weren't any Volumes inserted, this means that the backing
        # Storage Provider wasn't registered with PowerVC so don't import
        # We also will bypass the check if it is local storage with volumes
        # attached since we know they must be data volumes if driver allowed it
        if volumes is None and not self._is_local_storage():
            return None
        return volumes if volumes is not None else []

    def _import_volume_attachments(self, instance, volumes, connector):
        """Internal Helper Method to Import Volume Attachments into the DB"""
        # Loop through Volumes returned creating the Volume Attachment
        for volume in volumes:
            self._create_block_device_mapping(instance, volume, connector)
            # If this is the Boot Volume, update the Instance Disk Size
            if volume.get('bootable', False):
                updates = dict(root_gb=volume['size'])
                updates['root_device_name'] = volume['mountpoint']
                instance.update(updates)
        power_specs = {'pending_onboard_actions': ['capture']}
        # If there is more than one Boot Volume, then they need to select
        # If this is local storage we know the volumes are all data volumes
        if len(volumes) > 1 and not self._is_local_storage():
            power_specs['pending_onboard_actions'].append('delete')
            power_specs['pending_onboard_actions'].append('resize')
        # Update the Pending On-board Actions flag in the DB for the VM
        power_specs['pending_onboard_actions'] =\
            ','.join(power_specs['pending_onboard_actions'])
        instance.update(dict(power_specs=power_specs))
        instance.save()

    def _update_boot_volumes(self, uuid, boot_volume_ids, inst_updates,
                             bdms, attach_order):
        """Helper method to update the Boot Volume for the VM"""
        found_match = False
        volapi = volume_api.API()
        cinder_client = cinder.cinderclient(self.context).client
        # First make sure the specified Volumes are attached to this VM
        for boot_volume_id in boot_volume_ids:
            for bdm in bdms:
                if boot_volume_id == bdm['volume_id']:
                    found_match = True
            # If the Volume specified is not attached, throw an exception
            if not found_match:
                self._throw_bad_request(
                    msg.NO_BOOT_VOLUME, volid=boot_volume_id)
        # Now loop through all the attachments, resetting the Boot records
        for bdm in (bdm for bdm in bdms if bdm['volume_id'] is not None):
            delbdm = dict()
            new_meta = dict()
            meta_changed = False
            volume_id = bdm['volume_id']
            volume = volapi.get(self.context, volume_id)
            new_meta = volume.get('volume_metadata', {}).copy()
            was_boot = new_meta.get('is_boot_volume')
            # If this is a Boot Volume, update the info on the Instance for it
            if volume_id in boot_volume_ids:
                if was_boot != 'True':
                    new_meta['is_boot_volume'] = 'True'
                    meta_changed = True
                delbdm['delete_on_termination'] = True
            else:
                # If this isn't the Boot Volume anymore, remove the Boot Flag
                if was_boot == 'True':
                    new_meta.pop('is_boot_volume', None)
                    meta_changed = True
                # Update the Attachment to say not to delete the Volume
                if bdm.get('delete_on_termination', False):
                    delbdm['delete_on_termination'] = False
            if volume_id in attach_order:
                # Only set the root gb/device based on the first attach order
                if volume_id == attach_order[0]:
                    inst_updates['root_gb'] = volume['size']
                    inst_updates['root_device_name'] = bdm['device_name']
                delbdm['boot_index'] = attach_order.index(volume_id)
            # If the Boot Index was set previously, then we want to unset it
            elif bdm.get('boot_index') is not None:
                delbdm['boot_index'] = None
            # If the Boot Volume changed, update the Delete on Termination Flag
            if len(delbdm) > 0:
                db.block_device_mapping_update(self.context, bdm['id'], delbdm)
            # If the Boot Volume flag changed, call the Cinder API to update it
            if meta_changed:
                body = {'volume': {'metadata': new_meta}}
                # Update the Volume Meta-data with the Boot change
                cinder_client.put('/volumes/' + volume_id, body=body)

    def _retrieve_existing_networks(self):
        """Internal Helper Method to build a Map of Networks that exist"""
        network_map = dict()
        network_list = self.network_client.list_networks()['networks']
        for anetwork in network_list:
            # The 'VLAN' is 'Flat' if one does not exist
            vlanid = anetwork.get('provider:segmentation_id', 'Flat')
            phys_net = anetwork.get('provider:physical_network', 'default')
            vlanid = vlanid if vlanid else 'Flat'
            phys_net = phys_net if phys_net else 'default'
            # A dictionary key for the network
            net_pass = str(vlanid) + ":" + phys_net
            if network_map.get(net_pass, None) is None:
                network_map[net_pass] = [anetwork]
            else:
                new_map = network_map.get(net_pass)
                new_map.append(anetwork)
                network_map[net_pass] = new_map
        return network_map

    def _create_dflt_network_assn(self, vm_name, anetwork, vlanid, vswitch):
        """Internal Helper Method to Create the Default Host Network Mapping"""
        hostfact = network_dom.ManagedHostFactory.get_factory()
        netfact = network_dom.NetworkAssociationFactory.get_factory()
        # Since the call requires a Host DOM, get the VIOS's to construct one
        host_dom = hostfact.find_host_by_name(self.context, self.host)
        netw_assn = netfact.find_network_assoc(
            self.context, self.host, anetwork['id'])
        if netw_assn is not None:
            self._check_for_sea_conflict(
                vm_name, anetwork, vlanid, vswitch, netw_assn, host_dom)
            return
        # Call the Nova Neutron API to build the default network association
        self.network_api.build_default_network_assn(
            self.context, host_dom, vlanid, anetwork['id'], vswitch)
        netw_assn = netfact.find_network_assoc(
            self.context, self.host, anetwork['id'])
        if netw_assn is not None:
            self._check_for_sea_conflict(
                vm_name, anetwork, vlanid, vswitch, netw_assn, host_dom)

    def _check_for_sea_conflict(self, vm_name, anetwork, vlanid,
                                vswitch, netw_assn, host_dom):
        """
            Internal Helper Method to check and update sea conflict
            This method checks if any existing network association has
            different sea than the one which actually used by the VM
            then this method will try update the network association
            with the correct sea.Also updates peer networks (using same
            vlan).If this update is success full then user will get a
            message about the network sea update.
            If it is not then user will get a message about error.
        """
        seaInUse = None
        sea_list = host_dom.find_all_primary_seas()
        vlan_id = int(vlanid)
        for sea in sea_list:
            # See if vlan is on the sea's primary vea
            if sea.primary_vea.is_vlan_configured(vlan_id, vswitch,
                                                  False):
                seaInUse = sea
                break
            # Next, loop through all the additional VEAs to see if it exists
            for vea in sea.additional_veas:
                if vea.is_vlan_configured(vlan_id, vswitch, False):
                    seaInUse = sea
                    break
        # if the Network Association for the Host already exists
        # check if seas are same, if not update sea of the network association
        if netw_assn is not None and netw_assn.sea is not None and\
           seaInUse is not None and netw_assn.sea.name != seaInUse.name:
            # update the network association and inform the user
            netw_assn.sea = seaInUse
            # calling network association put with force
            exception_list = mapping_task.AdapterMappingTaskClass().\
                network_association_put_obj(
                    self.context, [netw_assn], True, None, True)
            if exception_list and len(exception_list) > 0:
                for e in exception_list:
                    LOG.error(_("Error while updating network "
                                "association: %(msg)s") %
                              {'msg': unicode(e.message)})
                msgb = self._get_message_bundle()
                text = msgb.get_message_text(
                    self.context, msg.IMPORT_NETWORK_FAILED, vmname=vm_name)
                raise Exception(text)
            else:
                msgb = self._get_message_bundle()
                text = msgb.get_message_text(
                    self.context, msg.IMPORT_NETWORK_CHANGED, vmname=vm_name,
                    networkname=anetwork['name'], seaname=seaInUse.name)
                LOG.warn(text)
                self._send_console_message('WARN', text)

    def _find_storage_providers(self):
        """Internal Helper Method to build a List of Storage Providers"""
        if self._provider_list is None:
            prvfact = storage_dom.StorageProviderFactory.get_factory()
            self._provider_list = prvfact.find_all_providers(self.context)
        return self._provider_list

    def _find_storage_provider_names(self):
        """Internal Helper Method to build a List of Storage Providers"""
        providers = self._find_storage_providers()
        return [prov.host_name for prov in providers]

    def _create_block_device_mapping(self, instance, volume, connector):
        """Internal Helper Method to Create a Block Device Mapping"""
        # Insert the Block Device Mapping into the Database for the Volume
        values = {'instance_uuid': instance['uuid'], 'volume_id': volume['id'],
                  'device_name': volume['mountpoint']}
        values['source_type'] = values['destination_type'] = 'volume'
        values['delete_on_termination'] = volume.get('bootable') is True
        values['boot_index'] = 0 if volume.get('bootable') is True else None
        # Add the Connection Info to the Block Device Mapping if it is provided
        if volume.get('connection_info') is not None:
            connect_info = volume['connection_info']
            # Add in the Volume Connector Info retrieved from the Driver
            if connect_info.get('volume_connector') is not None:
                # Don't want to use the connector if this is SSP and is mixed
                if connect_info.get('driver_volume_type') == 'ibm_ssp':
                    connect_info['volume_connector']['connector'] = None
                elif connect_info.get('driver_volume_type') == 'fibre_channel':
                    gen_connector = connect_info['volume_connector']
                    if (gen_connector.get('connection-type') == 'npiv'):
                        connect_info['volume_connector']['connector'] = \
                            connector['npiv_connector_data']
                        connect_info['volume_connector']['connector']['host'] \
                            = connector['host']
                    elif (gen_connector.get('connection-type') == 'pv_vscsi'):
                        connect_info['volume_connector']['connector'] = \
                            connector['vscsi_connector_data']
                        connect_info['volume_connector']['connector']['host'] \
                            = connector['host']
            values['connection_info'] = jsonutils.dumps(connect_info)
        db.block_device_mapping_create(self.context, values, False)

    def _cleanup_allocated_networks(self, instance):
        """Internal Helper Method to Cleanup Allocated Networks from the DB"""
        net_port_map = dict()
        try:
            # Get all the Ports up front so we have those for this Instance
            all_ports = network_api.API().list_ports(self.context)['ports']
            # List the ports for instance and maintain a list of ports that are
            # going to be deleted.
            search_opts = {'device_id': instance['uuid']}
            data = self.network_client.list_ports(**search_opts)
            ports = [port for port in data.get('ports', [])]
            # Reserve each port in case they want to remanage the same vm.
            for port in ports:
                # Set up a request to update the port and reset some fields
                port_req = {'port': {}}
                # Unbind from the host in case they want to remanage
                port_req['port']['binding:host_id'] = ''
                port_req['port']['device_id'] = ''
                # See if there are fixed IPs (device_owner is different)
                fixed_ips = port.get('fixed_ips', [])
                # Should only be one fixed_ip (if any) per port
                fixed_ip = fixed_ips[0] if fixed_ips else None
                if fixed_ip:
                    # Add description 'Unmanaged VM <vm name>'.
                    port_req['port']['device_owner'] = 'PowerVC: Unmanaged ' +\
                        'VM ' + instance['display_name']
                else:
                    # For dhcp network, set owner to network:dhcp so these
                    # ports can be auto-cleaned up on network delete.
                    port_req['port']['device_owner'] = DEVICE_OWNER_DHCP
                # Now update the port.
                self.network_client.update_port(port.get('id'), port_req)
            # We need to figure out which Networks will still have Ports left
            for port in all_ports:
                if port.get('device_id') != instance['uuid']:
                    network_id = str(port.get('network_id'))
                    net_ports = net_port_map.get(network_id, [])
                    net_ports.append(port)
                    net_port_map[network_id] = net_ports
            # We want to cleanup any Imported Networks that have no Ports left
            for port in all_ports:
                if port.get('device_id') == instance['uuid']:
                    network_id = str(port.get('network_id'))
                    if len(net_port_map.get(network_id, [])) <= 0:
                        self._cleanup_imported_network(network_id)
        except Exception as ex:
            LOG.warn(_("Error Removing Network from the DB: %s") % unicode(ex))

    def _cleanup_imported_network(self, network_id):
        """Internal Helper Method to Cleanup Unused Imported Networks"""
        if network_id is None or str(network_id) == 'None':
            return
        try:
            network = self.network_client.show_network(network_id)['network']
            # If this is an Imported Network, then we want to clean it up
            if network['name'].startswith('Imported Network for VLAN') or\
                    network['name'].startswith('Imported Flat Network'):
                self.network_client.delete_network(network_id)
        except Exception as exc:
            LOG.warn(_("Error Removing the "
                       "Imported Network: %s") % unicode(exc))

    def _cleanup_attached_volumes(self, instance, bdms=None):
        """Internal Helper Method to Cleanup Attached Volumes from the DB"""
        try:
            cinder_client = cinder.cinderclient(self.context).client
            # Retrieve all of the Volume Attachments that are in the Nova DB
            if bdms is None:
                bdms = db.block_device_mapping_get_all_by_instance(
                    self.context, instance['uuid'])
            for bdm in bdms:
                try:
                    # If it is a valid ID in the Attachment, remove the Volume
                    if bdm['volume_id']:
                        instance_id_dict = dict(instance_uuid=instance['uuid'])
                        uri = '/volumes/' + bdm['volume_id']
                        volume = cinder_client.get(uri)[1]['volume']
                        # Get the Host for the Volume in order to UnManage it
                        host = volume['os-vol-host-attr:host']
                        uri = '/os-hosts/' + host + '/unmanage'
                        body = dict(volumes=[bdm['volume_id']], force=True,
                                    extra_params=instance_id_dict)
                        cinder_client.post(uri, body=body)
                    # Remove the Volume Attachment from the Database
                    if 'id' in bdm:
                        db.block_device_mapping_destroy(self.context,
                                                        bdm['id'])
                    elif 'volume_id' in bdm:
                        db.block_device_mapping_destroy_by_instance_and_volume(
                            self.context, instance['uuid'], bdm['volume_id'])
                    else:
                        LOG.info(_("cleanup_attached_volumes:No bdm found"))
                except Exception as exc1:
                    info = dict(uuid=str(bdm['volume_id']), exc=unicode(exc1))
                    LOG.warn(_("Error Removing Volume %(uuid)s "
                               "from the database: %(exc)s") % info)
        except Exception as ex:
            LOG.warn(_("Error Removing Volumes from the DB: %s") % unicode(ex))
