# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base Classes for all On-boarding related Activity"""

import uuid
import webob.exc
import ConfigParser
from eventlet import greenthread
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils

from powervc_discovery.onboarding import importer_messages as msg
from powervc_discovery.locale import _

cfg.CONF.register_opt(
    cfg.BoolOpt('allow_unsupported_onboard', default=False,
                help='Override to allow Importing of Unsupported Resources')
)

LOG = logging.getLogger(__name__)


########################################################
# ##########  Base On-boarding Implementation  #########
########################################################
class BaseImporter(object):
    """Abstract Base Class to handle On-boarding Existing Resources"""

    def __init__(self, context, hostname, **kwargs):
        """Constructor for Abstract Class for Host Registration"""
        self.host = hostname
        self.context = context
        self.dispname = self._get_display_name()

    ###############################################
    # #####  Defining BaseImporter Interface  #####
    ###############################################
    def discover_resources(self, filters=None):
        """Discovers/Queries the Existing Resources on the Managed System"""
        self._verify_host_resource()
        return self._query_existing_resources(filters=filters)

    def prepare_resources(self, resources, extra_parms={},
                          synchronous=False, onboard=False):
        """Prepare Resources to be imported into the Management System"""
        # Verify that the Host provided exists on this Management System
        self._verify_host_resource()
        resource_ids = [self._get_resource_id(rsc) for rsc in resources]
        # Verify that the actually did provide a valid syntax input
        self._verify_resource_ids(msg.IMPORT_ACTION, resource_ids)
        # Verify that the appropriate Services are running to Query Resources
        self._verify_service_running()
        # If they want us to run synchronously, just call the prepare directly
        if synchronous:
            self._prepare_resources(resources, extra_parms, onboard)
        # If this is an asynchronous request, spawn a thread to do what we need
        else:
            greenthread.spawn(self._prepare_resources, resources, extra_parms,
                              onboard)

    def import_resources(self, resources, preparation, extra_parms={},
                         synchronous=False, allow_unsupported=False):
        """Imports Existing Resources into the Management System"""
        # Verify that the Host provided exists on this Management System
        self._verify_host_resource()
        # The input list can be mixed dicts/strings, pull out just the id's
        resource_ids = [self._get_resource_id(rsc) for rsc in resources]
        # Verify that the actually did provide a valid syntax input
        self._verify_resource_ids(msg.IMPORT_ACTION, resource_ids)
        # Verify that the appropriate Services are running to Query Resources
        self._verify_service_running()
        # Insert a On-board Task record into the Database to use
        task_id = None if synchronous else self._create_onboard_task()['id']
        # If they want us to run synchronously, just call the import directly
        if synchronous:
            return self._import_resources(
                None, resources, resource_ids, preparation, extra_parms,
                allow_unsupported)
        # If this is an asynchronous request, spawn a thread to do what we need
        else:
            greenthread.spawn(self._import_resources, task_id, resources,
                              resource_ids, preparation, extra_parms,
                              allow_unsupported)
            return task_id

    def unmanage_resources(self, resource_ids, force=False, extra_params=None):
        """Removes Resources from the DB so they are no longer Managed"""
        failed_resources = []
        invalid_resources = []
        # Verify that the Host provided exists on this Management System
        self._verify_host_resource(force)
        # Verify that the actually did provide a valid syntax input
        self._verify_resource_ids(msg.UNMANAGE_ACTION, resource_ids)
        # Query the list of currently managed resources from the DB
        valid_resources = self._query_managed_resources()
        # Filter the input list down to only those that are valid to remove
        resources = self._filter_unmanage_resources(
            resource_ids, valid_resources, invalid_resources)
        # Give the sub-class a chance to verify all resources are remove-able
        nonremove_error = self._filter_unremoveable_resources(resources, force)
        # Call unmanage on each resource individually to handle failures
        for resource in resources:
            self.__unmanage_resource_wrapper(resource, failed_resources,
                                             extra_params)
        # Verify that there were no failures during the unmanage operation
        self._check_for_failures(msg.UNMANAGE_ACTION, invalid_resources,
                                 failed_resources, not force)
        # If any resources weren't remove-able, then throw the exception given
        if nonremove_error is not None:
            raise webob.exc.HTTPBadRequest(explanation=nonremove_error)

    ###############################################
    # ### Defining Abstract On-boarding Methods ###
    ###############################################
    def _get_resource_type_string(self):
        """Provides the translated Resource Type for this Importer"""
        raise NotImplementedError()

    def _get_resource_types_string(self):
        """Provides the translated Resource Types for this Importer"""
        raise NotImplementedError()

    def _get_system_type_string(self):
        """Provides the translated System Type for this Importer"""
        raise NotImplementedError()

    def _get_display_name(self):
        """Retrieves the Display Name for the given Managed System"""
        dispname = None
        parser = ConfigParser.RawConfigParser()
        if self._get_host_config_file():
            parser.read(self._get_host_config_file())
        # If the Display Name is in the Configuration File use that
        if parser.has_option('DEFAULT', 'host_display_name'):
            dispname = parser.get('DEFAULT', 'host_display_name')
        dispname = dispname if dispname else self.host
        return encodeutils.safe_decode(dispname)

    def _get_host_config_file(self):
        """Provides the name of the Config File created for this Host"""
        return None

    def _get_host_resource(self):
        """Retrieves the Host Resource from the DB to make sure it exists"""
        raise NotImplementedError()

    def _query_existing_resources(self, filters):
        """Discovers/Queries the Existing Resources on the Managed System"""
        raise NotImplementedError()

    def _query_managed_resources(self):
        """Queries the existing Managed Resources from the Database"""
        raise NotImplementedError()

    def _query_resource_details(self, resource_ids,
                                extra_parms, allow_unsupported):
        """Queries Details about the Resources from the Managed System"""
        raise NotImplementedError()

    def _verify_service_running(self):
        """Verifies that the Service is running to Query Resources from"""
        raise NotImplementedError()

    def _prepare_resource(self, resource, resource_prep, extra_parms):
        """Prepares the specified resource so it can be onboarded"""
        raise NotImplementedError()

    def _import_resource(self, resource, resource_prep, preparation,
                         extra_parms):
        """On-boards the specified resource into the Management System"""
        raise NotImplementedError()

    def _unmanage_resource(self, resource, extra_params=None):
        """Stops managing the specified resource by removing it from the DB"""
        raise NotImplementedError()

    def _import_resources_preparation(self, resources):
        """Perform any preparation needed prior to importing the resources"""
        return None

    def _filter_unremoveable_resources(self, resources, force):
        """Filters the resources that can't be removed and returns exception"""
        return None

    def _create_onboard_task(self):
        """Creates an On-boarding Task in the DB to use for this Action"""
        raise NotImplementedError()

    def _update_onboard_task(self, task_id, status, progress=None):
        """Updates an On-boarding Task in the DB to use for this Action"""
        raise NotImplementedError()

    def _create_onboard_taskrsc(self, task_id, resource_id,
                                status, fault_message=None):
        """Creates an On-boarding Task Record in the DB for this Resource"""
        raise NotImplementedError()

    def _update_onboard_taskrsc(self, task_id, resource_id,
                                status, fault_message=None):
        """Updates an On-boarding Task Record in the DB for this Resource"""
        raise NotImplementedError()

    ###############################################
    # ####  Shared On-boarding Helper Methods  ####
    ###############################################
    def _get_message_bundle(self):
        """Provides the Message Bundle class to use for this Importer"""
        return msg.BaseImporterMessageBundle()

    def _get_resource_id(self, resource):
        """Provides the name of the attribute for the Resource Identifier"""
        if isinstance(resource, basestring):
            return resource
        else:
            return resource.get('uuid', resource.get('id', None))

    def _prepare_resources(self, task_id, preparation):
        """Main Prepare Method that can be spawned in a separate thread"""
        prepared_resources, failed_resources = (list(), list())
        try:
            # Find resources that require preparation
            resource_ids = preparation.get('servers', [])
            # If there weren't any resource found, just return an empty list
            if len(resource_ids) <= 0:
                # Send a message to the Console saying that the prep completed
                # Commenting out until the prepare is fully in-place
                # TODO FIXME
                # self._send_completed_console_message(
                #    None, [], [], failed_resources, True)
                return prepared_resources
            self._update_onboard_task(task_id, 'preparing', 45)
            # Call Discover Resources first so we can make sure the resources
            # that we are passing to the driver really do exist or filter out
            exist_rscs = self.discover_resources()
            exist_rscs = dict([(rsc['id'], rsc) for rsc in exist_rscs])
            # Call prepare on each resource individually to handle failures
            for resource_id in resource_ids:
                resource = {'id': resource_id}
                new_resource = None
                # First make sure the Resource exists, otherwise fail it
                if exist_rscs.get(self._get_resource_id(resource)) is not None:
                    # Add the name to the Resource for the Driver to use
                    resource['name'] = \
                        exist_rscs.get(self._get_resource_id(resource)).\
                        get('name')
                    new_resource = self.__prepare_resource_wrapper(
                        resource, preparation)
                # If a Resource is returned, it means it was successful
                if new_resource is not None:
                    prepared_resources.append(new_resource)
                else:
                    failed_resources.append(resource)
            # Send a message to the Console saying that the prepare completed
            self._send_completed_console_message(
                None, prepared_resources, [], failed_resources, True)
            self._update_onboard_task(task_id, 'prepared', 46)
        # If there was an error running the task, update the status to failed
        except Exception as exc:
            LOG.warn(_("Error running prepare on-board task"))
            LOG.exception(exc)
            # Send a Failure Message to the Console Log with the Fault
            self._send_escape_console_message(unicode(exc))
            self._update_onboard_task(task_id, 'prepare_failed')
        return prepared_resources

    def _import_resources(self, task_id, resources, resource_ids, preparation,
                          extra_parms, allow_unsupported):
        """Main Import Method spawned in a separate thread"""
        progress = 50
        imported_resources, failed_resources = (list(), list())
        try:
            # First make sure the resource_ids list is formatted properly
            if '*all' in resource_ids or '*ALL' in resource_ids:
                resource_ids = ['*all']
            # Next check to see if we should override allowing unsupported
            if not allow_unsupported:
                allow_unsupported = cfg.CONF.allow_unsupported_onboard
            # Perform preparation on any resources that require it
            preped = []
            if preparation:
                preped = self._prepare_resources(task_id, preparation)
            # Remove prepared resource IDs from the list of IDs to query for
            resource_ids = [r for r in resource_ids if not
                            any(self._get_resource_id(p) == r for p in preped)]
            # Now we can call the Manager to retrieve the Resource Details
            resources = self._query_resource_details(
                resource_ids, extra_parms, allow_unsupported)
            # Combine the list of queried and prepared resources
            resources.extend(preped)
            resource_ids = [self._get_resource_id(r) for r in resources]
            # Filter out the resources that aren't applicable for import
            resources, unsupported_ids = self._filter_import_resources(
                task_id, resource_ids, resources, preparation,
                allow_unsupported)
            # If there weren't any resource found, just return an empty list
            if len(resources) <= 0:
                self._update_onboard_task(task_id, 'completed', 100)
                # Send a message to the Console saying the import completed
                self._send_completed_console_message(
                    task_id, list(), unsupported_ids, failed_resources)
                return imported_resources
            increment = 50 / len(resources)
            # Allow the sub-class to perform any necessary preparation
            resource_prep = self._import_resources_preparation(resources)
            # Call import on each resource individually to handle failures
            for resource in resources:
                # If previously managed, we need to remove the DB instance
                if self._prev_managed(self.context, resource):
                    self._hard_delete_from_db(resource)
                self._update_onboard_task(task_id, 'importing', progress)
                new_resource = self.__import_resource_wrapper(
                    task_id, resource, resource_prep, preparation, extra_parms)
                # If a Resource is returned, it means it was successful
                if new_resource:
                    imported_resources.append(new_resource)
                elif new_resource is None:
                    failed_resources.append(resource)
                progress += increment
            # If everything completed, then mark the overall task as completed
            self._update_onboard_task(task_id, 'completed', 100)
            # Send a message to the Console saying that the import completed
            self._send_completed_console_message(
                task_id, imported_resources, unsupported_ids, failed_resources)
        # If there was an error running the task, update the status to failed
        except Exception as exc:
            LOG.warn(_("Error running on-board task %s") % str(task_id))
            LOG.exception(exc)
            # Update the Import Task in the Database for the Failure
            self._update_onboard_task(task_id, 'failed')
            # Send a Failure Message to the Console Log with the Fault
            self._send_escape_console_message(unicode(exc))
        return imported_resources

    def _filter_import_resources(self, task_id, resource_ids, resources,
                                 preparation, allow_unsupported):
        """Helper Method to return the subset of Valid Resources specified"""
        invalid_resource_ids = list()
        importable_resources, unsupported_ids = (list(), list())
        allow_unsupported = allow_unsupported is True
        resource_map = dict([(self._get_resource_id(resource),
                              resource) for resource in resources])
        preped = []
        if preparation:
            preped = preparation.get('servers', [])
        # Loop through each of the Resources, seeing which ones are valid
        for resource_id in resource_ids:
            if resource_id != '*all' and resource_map.get(resource_id) is None:
                invalid_resource_ids.append(resource_id)
                # Create an On-board Task record for the specified Resource
                self._create_onboard_taskrsc(task_id, resource_id,
                                             '', 'nonexistent', None)
        # Loop through the resources returned seeing which ones to import
        for resource in resources:
            fault = None
            status = 'waiting'
            name = resource.get('name', '')
            managed = resource.get('managed', False)
            support = resource.get('support', {})
            supported = support.get('status', 'supported')
            resource_id = str(self._get_resource_id(resource))
            # If it is already managed, then we will skip the import
            if managed:
                status = 'already_managed'
            # If it was prepared, run the import  TODO - verify import passed
            elif supported == 'preparation' and resource_id in preped:
                importable_resources.append(resource)
            # If it isn't supported, then skip and give reason why
            elif supported != 'supported':
                status = supported
                unsupported_ids.append(resource_id)
                fault = support.get('reasons', [''])[0].encode('utf-8')[:255]
            # If it is valid, then add to the list of resources to import
            else:
                importable_resources.append(resource)
            # Create an On-board Task record for the specified Resource
            self._create_onboard_taskrsc(task_id, resource_id,
                                         name, status, fault)
        # If there were all Invalid Resource ID's specified, throw an exception
        if len(invalid_resource_ids) > 0:
            parm_str = self._format_string_list(invalid_resource_ids)
            if len(invalid_resource_ids) == len(resource_ids):
                self._throw_bad_request(msg.INVALID_RESOURCES, rscids=parm_str)
        return (importable_resources, unsupported_ids)

    def _filter_unmanage_resources(self, resource_ids,
                                   valid_resources, invalid_resources):
        """Helper Method to return the subset of Valid Resources specified"""
        selected_resources = []
        resource_map = dict([(self._get_resource_id(resource),
                              resource) for resource in valid_resources])
        # Loop through all the Resource ID's finding the Resource for it
        for rscid in resource_ids:
            # Skip this Resource ID if they asked for All Resources
            if rscid != '*all' and rscid != '*ALL':
                resource = resource_map.get(rscid)
                # If the resource exists, add it to the map of those selected
                if resource is not None:
                    selected_resources.append(resource)
                # If the resource does not exist, then flag it as invalid
                else:
                    invalid_resources.append(rscid)
        # If they specified All Resources, return all Valid Resources
        if '*all' in resource_ids or '*ALL' in resource_ids:
            return valid_resources
        return selected_resources

    def _send_completed_console_message(self, task_id, imported_resources,
                                        nosupport_resources, failed_resources,
                                        prepare=False):
        """Shared Helper method to send a Completion Message to the Console"""
        dispname = self._get_display_name()
        # Only send the message to the Console if it was asynchronous
        if task_id is not None:
            msgid = msg.PREPARE_FINISHED2 if prepare else msg.IMPORT_COMPLETED4
            msgb = self._get_message_bundle()
            failed = str(len(failed_resources))
            success = str(len(imported_resources))
            nosupport = str(len(nosupport_resources))
            rsctypes = self._get_resource_types_string()
            systype = self._get_system_type_string()
            # We want to leave out failed/unsupported from the message if 0
            if nosupport == '0' and failed == '0':
                msgid = msg.IMPORT_COMPLETED1
                msgid = msg.PREPARE_FINISHED1 if prepare else msgid
            elif nosupport == '0':
                msgid = msg.IMPORT_COMPLETED2
            elif failed == '0':
                msgid = msg.IMPORT_COMPLETED3
            # Lookup the translated message for the import completion
            text = msgb.get_message_text(self.context, msgid,
                                         rsctypes=rsctypes, systype=systype,
                                         hostname=dispname, success=success,
                                         nosupport=nosupport, failed=failed)
            # Send a Message to the Console Log for the completion
            self._send_console_message('INFO', text)
            LOG.info(text)

    def _send_escape_console_message(self, fault):
        """Shared Helper method to send an Escape Message to the Console"""
        msgb = self._get_message_bundle()
        dispname = self._get_display_name()
        rsctypes = self._get_resource_types_string()
        systype = self._get_system_type_string()
        # We need to trim out the call-stack from the Escape Message
        if fault and fault.find('Traceback (') > 0:
            fault = fault[:fault.find('Traceback (')]
        # Lookup the translated message for the import completion
        text = msgb.get_message_text(
            self.context, msg.IMPORT_ESCAPE, rsctypes=rsctypes,
            systype=systype, hostname=dispname, fault=fault)
        # Send a Message to the Console Log for the completion
        self._send_console_message('WARN', text)

    def _send_failure_console_message(self, resource, fault):
        """Shared Helper method to send a Failure Message to the Console"""
        msgb = self._get_message_bundle()
        dispname = self._get_display_name()
        rscname = str(resource.get('name'))
        rsctype = self._get_resource_type_string()
        systype = self._get_system_type_string()
        # Lookup the translated message for the import completion
        text = msgb.get_message_text(
            self.context, msg.IMPORT_FAILED, rsctype=rsctype,
            rscname=rscname, systype=systype, hostname=dispname, fault=fault)
        # Send a Message to the Console Log for the completion
        self._send_console_message('WARN', text)

    def _send_console_message(self, priority, message):
        """Shared Helper method to send a Message to the Console Log"""
        raise NotImplementedError()

    def _verify_host_resource(self, force=False):
        """Shared Helper method to verify the host specified exists"""
        # For now we will just try to retrieve it that throws an exception
        if not force and self._get_host_resource() is None:
            self._throw_not_found(msg.NO_HOST, hostname=self.host)

    def _verify_resource_ids(self, action_message, resource_ids):
        """Shared Helper method to throw an exception if no resources"""
        msgb = self._get_message_bundle()
        action_text = msgb.get_message_text(self.context, action_message)
        # If we weren't given any Resource Identifiers, throw an exception
        if resource_ids is None or len(resource_ids) <= 0:
            self._throw_bad_request(msg.NO_RESOURCES, action=action_text)
        # Loop through each of the Resource Identifiers, validating them
        for resource_id in resource_ids:
            # If the Resource Identifier isn't a UUID, then throw an exception
            if self._is_valid_uuid(resource_id) is False:
                self._throw_bad_request(
                    msg.INVALID_RESOURCE_IDS, rscid=str(resource_id))

    def _check_for_failures(self, action_message, invalid_resources,
                            failed_resources, throw_exception):
        """Shared Helper method to throw an exception if any resource failed"""
        msgb = self._get_message_bundle()
        action_text = msgb.get_message_text(self.context, action_message)
        # First check to see if any resources failed and throw an exception
        if throw_exception and len(failed_resources) > 0:
            parm_str = self._format_string_list(failed_resources)
            self._throw_server_error(msg.ACTION_FAILURE,
                                     action=action_text, rscids=parm_str)
        # Check to see if any resources were invalid and throw an exception
        if throw_exception and len(invalid_resources) > 0:
            parm_str = self._format_string_list(invalid_resources)
            self._throw_bad_request(msg.INVALID_RESOURCES,
                                    action=action_text, rscids=parm_str)

    def _is_valid_uuid(self, resource_id):
        """Shared Helper method to see if it is a valid UUID format"""
        if not resource_id:
            return False
        # Special case if they passed in *ALL for the Resource Identifier
        if resource_id.lower() == '*all':
            return True
        try:
            return str(uuid.UUID(resource_id)) == resource_id
        except Exception:
            return False

    def _throw_bad_request(self, msgid, **msgparms):
        """Shared Helper method to Throw a Bad Request exception"""
        self._throw_exception(webob.exc.HTTPBadRequest, msgid, **msgparms)

    def _throw_not_found(self, msgid, **msgparms):
        """Shared Helper method to Throw a Not Found exception"""
        self._throw_exception(webob.exc.HTTPNotFound, msgid, **msgparms)

    def _throw_server_error(self, msgid, **msgparms):
        """Shared Helper method to Throw a Internal Server Error exception"""
        self._throw_exception(webob.exc.HTTPInternalServerError,
                              msgid, **msgparms)

    def _throw_exception(self, exctype, msgid, **msgparms):
        """Shared Helper method to Throw a translated exception"""
        # For now we are just going to return the English-specific message text
        msgb = self._get_message_bundle()
        msgparms['rsctype'] = self._get_resource_type_string()
        msgparms['rsctypes'] = self._get_resource_types_string()
        raise exctype(explanation=msgb.get_message_text(self.context,
                                                        msgid, **msgparms))

    @staticmethod
    def _format_string_list(alist):
        """Shared Helper method to format a String List for display"""
        astr = ''
        # If isn't a List, then just convert it to a String (shouldn't happen)
        if not isinstance(alist, list):
            return str(alist)
        # If it is a List, then convert it to a comma-separate list of strings
        for anitem in alist:
            astr += str(anitem) + ', '
        return astr[:len(astr) - 2] if len(astr) > 0 else astr

    def _prev_managed(self, context, instance_uuid):
        # Check to see if the VM was previously managed by PowerVC
        # TODO in nova/cinder agnostic fashion, or move to subclasses
        return False

    def _hard_delete_from_db(self, resource):
        # Hard deletes a resource from the DB
        pass

    ###############################################
    # ### Internal On-boarding Helper Methods  ####
    ###############################################
    def __prepare_resource_wrapper(self, resource, extra_parms):
        """Wrapper method to handle exceptions from prepare_resource"""
        resource_id = str(self._get_resource_id(resource))
        try:
            return self._prepare_resource(resource, extra_parms)
        except Exception as exc:
            # If we got an exception on prepare, log it and add it to the list
            LOG.warn(_("Error Preparing Resource %s") % resource_id)
            LOG.exception(exc)
            # Send a failure message to the Consoles with the fault
            self._send_failure_console_message(resource, unicode(exc))
            return None

    def __import_resource_wrapper(self, task_id, resource, resource_prep,
                                  preparation, extra_parms):
        """Wrapper method to handle exceptions from import_resource"""
        resource_id = str(self._get_resource_id(resource))
        try:
            self._update_onboard_taskrsc(task_id, resource_id, 'importing')
            resource = self._import_resource(
                resource, resource_prep, preparation, extra_parms)
            status = 'imported' if resource else 'failed'
            self._update_onboard_taskrsc(task_id, resource_id, status)
            return resource
        except Exception as exc:
            # If we got an exception on unmanage, log it and add it to the list
            LOG.warn(_("Error Importing Resource %s") % resource_id)
            LOG.exception(exc)
            # Update the Task in the Database saying that the Import failed
            fault = unicode(exc).encode('utf-8')
            self._update_onboard_taskrsc(task_id, resource_id,
                                         'failed', fault[:255])
            # Send a failure message to the Consoles with the fault
            self._send_failure_console_message(resource, unicode(exc))
            return None

    def __unmanage_resource_wrapper(self, resource, failed_resources,
                                    extra_params):
        """Wrapper method to handle exceptions from unmange_resource"""
        resource_id = str(self._get_resource_id(resource))
        try:
            self._unmanage_resource(resource, extra_params)
        except Exception as exc:
            # If we got an exception on unmanage, log it and add it to the list
            LOG.warn(_("Error Unmanaging Resource %s") % resource_id)
            LOG.exception(exc)
            failed_resources.append(resource_id)
