# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to On-board Volumes for a Storage Host"""

import webob.exc

from cinder.api import extensions
from cinder.api.openstack import wsgi
from oslo_log import log as logging

from powervc_cinder.db import api as powervc_db
from powervc_discovery.onboarding import importer_messages as msg
from powervc_discovery.onboarding.storage import volume_importer
from powervc_discovery import locale as discovery_locale

LOG = logging.getLogger(__name__)
AUTHORIZE = extensions.extension_authorizer('volume', 'host_onboard')


class StorageOnboardControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for On-boarding Volumes"""

    @wsgi.action("index")
    def index(self, req, host_name_id):
        """Add Rest GET Method to list On-boarding Tasks"""
        tasks = []
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'index')
        # Parameter for showing only running tasks
        running = False
        if 'running' in req.GET and req.GET['running'] == 'true':
            running = True
        # Retrieve the list of On-boarding Tasks from the Database
        result = powervc_db.onboard_task_find_all(context, host_name_id)
        for task in result:
            task_uri = self._get_task_uri(req, task['id'])
            if(not running or (running and task['status'] != 'completed' and
                               task['status'] != 'failed')):
                tasks.append({'id': task['id'], 'links': [{'href': task_uri}]})
        return {'tasks': tasks}

    @wsgi.action("show")
    def show(self, req, host_name_id, id):
        """Add Rest GET Method to show a specified On-boarding Task"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'show')
        # Retrieve the list of On-boarding Tasks from the Database
        result = powervc_db.onboard_task_find(context, int(id))
        # If the Task wasn't found in the Database, then throw an exception
        if result is None or result['host'] != host_name_id:
            text = msg.BaseImporterMessageBundle().\
                get_message_text(context, msg.NO_IMPORT_TASK, taskid=id)
            raise webob.exc.HTTPNotFound(explanation=text)
        # Construct the Task dictionary to return in the REST response
        task = {'id': result['id']}
        task['links'] = [{'href': req.url}]
        task['volumes'] = []
        # Loop through all the attributes on the task we want to include
        task_attrs = ['status', 'progress', 'started', 'ended']
        for attr in task_attrs:
            task[attr] = result[attr]
        # Add in the list of Volume Results into the overall Task entry
        result_vols = result.get('volumes', [])
        for result_vol in result_vols:
            vol = {'id': result_vol['volume_uuid']}
            vol['name'] = result_vol['volume_name']
            vol['status'] = result_vol['status']
            if result_vol.get('fault_message'):
                vol['fault'] = result_vol['fault_message']
            task['volumes'].append(vol)
        return {'task': task}

    @wsgi.action("create")
    def onboard(self, req, host_name_id, body):
        """Add Rest POST Method to On-board Volumes for a given Host"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'import')
        # Parse out the key attributes from the request body
        volumes = body.pop('volumes', None)
        synch = body.pop('synchronous', False) is True
        unsupported = body.pop('allow_unsupported', False) is True
        # Delegate to the Volume Importer to actually perform the on-boarding
        importer = volume_importer.VolumeImporter(context, host_name_id)
        response = importer.import_resources(volumes, None, body, synch,
                                             unsupported)
        # If this is asynchronous, then return the Task Identifier Info
        if not synch:
            # Build the Task URI (to use for progress) to return the response
            task_uri = self._get_task_uri(req, response)
            return {'task': {'id': response, 'links': [{'href': task_uri}]}}
        # Otherwise we will get the Values back on the Synchronous call
        return {'volumes': response}

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def _get_task_uri(req, task_id):
        """Internal Helper Method to format the URI for the On-boarding Task"""
        return ('%s/%s' % (req.url, str(task_id))).replace('?running=true', '')


class StorageUnmanageControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for Unmanaging Volumes"""

    @wsgi.action("create")
    @wsgi.response(204)
    def unmanage(self, req, host_name_id, body):
        """Add Rest POST Method to Unmanage Volumes for a given Host"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'unmanage')
        force = body.get('force', False)
        # get extra parameters from body
        extra_params = body.get('extra_params', None)
        # Delegate to the Volume Importer to actually perform the unmanage
        importer = volume_importer.VolumeImporter(context, host_name_id)
        importer.unmanage_resources(body.get('volumes'), bool(force),
                                    extra_params)

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class StorageVolumeListControllerExtension(wsgi.Controller):
    """Controller Class to extend the os-hosts API for listing Volumes"""

    @wsgi.action("index")
    def index(self, req, host_name_id):
        """Add Rest GET Method to list existing Volumes for a given Host"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'list_volumes')
        # Delegate to the Volume Importer to actually perform query volumes
        importer = volume_importer.VolumeImporter(context, host_name_id)
        # Perform any conversion on passed-in filter arguments.
        filters = req.params.copy()
        # host_refs filter is a list of strings separated by ',' characters.
        # The meaning of strings themselves are specific to connection type;
        # for FC that would be Host WWPNs, for SSP the UUID of the LU
        if 'host_refs' in filters:
            filters['host_refs'] = filters['host_refs'].split(',')
        # Now run the Discover Resources on the Importer to get the Volumes
        volumes = importer.discover_resources(filters=filters)
        return {'volumes': volumes}

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class StorageOnboarding(extensions.ExtensionDescriptor):
    """Capability to On-board Volumes for a Storage Host."""
    name = "Storage Host On-boarding"
    alias = "ibm-storage-onboarding"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_storage_onboarding/api/v1.2"
    updated = "2013-07-21T21:00:00-06:00"

    def get_resources(self):
        """Provide a Sub-Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'onboard', controller=StorageOnboardControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource2 = extensions.ResourceExtension(
            'unmanage', controller=StorageUnmanageControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        resource3 = extensions.ResourceExtension(
            'all-volumes', controller=StorageVolumeListControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'))
        return [resource1, resource2, resource3]
