# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""On-boarding Classes for importing and unmanaging Volumes"""

import re
import string
import uuid

from oslo_config import cfg
from cinder import db
from cinder import exception
from cinder import quota
from cinder import rpc as rpc_api
from cinder.image import glance
from oslo_log import log as logging
from oslo_utils import timeutils
from cinder.volume import utils

from powervc_cinder.db import api as powervc_db
from powervc_cinder.volume import rpcapi_product
from powervc_discovery.onboarding import importer
from powervc_discovery.onboarding import importer_messages as msg
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _


LOG = logging.getLogger(__name__)


########################################################
# ########  Volume On-boarding Implementation  #########
########################################################
class VolumeImporter(importer.BaseImporter):
    """On-boarding Classes for importing and unmanaging Volumes"""

    def __init__(self, context, hostname, **kwargs):
        """Constructor for Importer Class for Volume On-boarding"""
        super(VolumeImporter, self).__init__(context, hostname, **kwargs)
        self.volume_api = rpcapi_product.VolumeAPIProduct()

    ###############################################
    # ### Implementing BaseImporter Interface  ####
    ###############################################
    def _get_resource_type_string(self):
        """Provides the translated Resource Type for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.RSCTYPE_VOLUME)

    def _get_resource_types_string(self):
        """Provides the translated Resource Types for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.RSCTYPE_VOLUMES)

    def _get_system_type_string(self):
        """Provides the translated System Type for this Importer"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTYPE_STORAGE)

    def _get_host_config_file(self):
        """Provides the name of the Config File created for this Host"""
        return '/etc/cinder/cinder-' + self.host + '.conf'

    def _get_host_resource(self):
        """Retrieves the Host Resource from the DB to make sure it exists"""
        services = powervc_db.service_get_all_by_host(self.context, self.host)
        return services[0] if services and len(services) > 0 else None

    def _query_existing_resources(self, filters):
        """Discovers/Queries the Existing Resources on the Managed System"""
        # First verify the Service is up so we don't want for the long timeout
        self._verify_service_running()
        # If the Service is up, then we can discover the Volumes from it
        vol_data = self.volume_api.discover_volumes(
            self.context, self.host, filters)
        return self._handle_volume_chunking(vol_data)

    def _query_managed_resources(self):
        """Queries the existing Managed Resources from the Database"""
        return db.volume_get_all_by_host(self.context, self.host)

    def _query_resource_details(self, resource_ids,
                                extra_parms, allow_unsupported):
        """Queries Details about the Resources from the Managed System"""
        parms = extra_parms if extra_parms is not None else dict()
        # First verify the Service is up so we don't want for the long timeout
        self._verify_service_running()
        # If the Service is up, then we can query the VM's from it
        vol_data = self.volume_api.query_volumes(
            self.context, self.host, resource_ids, parms, allow_unsupported)
        return self._handle_volume_chunking(vol_data)

    def _verify_service_running(self):
        """Verifies that the Service is running to Query Resources from"""
        # Verify that the Cinder-Volume Service is running to Query VM's
        if not self.volume_api.verify_host_running(self.context, self.host):
            service = 'cinder-volume-' + self.host
            self._throw_server_error(msg.SERVICE_DOWN, service=service)

    def _import_resource(self, resource, resource_prep, preparation,
                         extra_parms):
        """On-boards the specified resource into the Management System"""
        LOG.info(_("Importing Volume %s") % str(resource))
        server_info = extra_parms.get('server_info', {})
        local_storage = extra_parms.get('local_storage', False) is True
        volume = self._prep_volume_for_import(
            resource, server_info, local_storage)
        restricted_metadata = volume.pop('restricted_metadata', None)
        metadata = volume.get('metadata', {})
        bootable = metadata.get('is_boot_volume') == 'True'
        attachments = volume.pop('attachments', [])
        # First make sure the Volume doesn't already exist we are importing
        if self._volume_exists_in_db(resource['id']):
            # The valid case for it already existing if for attaching or shared
            volume_copy = volume.copy()
            for attachment in attachments:
                attachment['volume_id'] = resource['id']
                if self._volume_attachment_exists_in_db(attachment):
                    LOG.warn("The volume attachment already exists for %(vol)s"
                             " with %(inst)s" %
                             dict(vol=resource['id'],
                                  inst=attachment.get('instance_uuid', None)))
                    continue
                atrs = ['metadata', 'attach_status', 'status', 'multiattach']
                # Set the multiattach property since its a shared volume
                volume_copy['multiattach'] = True
                values = dict([(key, volume_copy[key]) for key in atrs])
                # Update the Status on the Volume, and insert the Attachment
                volume = db.volume_update(self.context, resource['id'],
                                          values)
                self._add_attachment_info(self.context, attachment)
            return dict(id=resource['id'], bootable=bootable,
                        attachments=attachments, size=volume['size'])
        # Send a notification to listeners that we are importing this volume
        utils.notify_about_volume_usage(self.context, volume, "import.start")
        # Reserve the Quota Reservation saying we are going to add Usage
        reserv = quota.QUOTAS.reserve(self.context, volumes=1,
                                      gigabytes=volume['size'])
        try:
            # Check for multi-attach
            if attachments and len(attachments) > 1:
                volume['multiattach'] = True
            # Create the Volume instance in the Database
            volume = db.volume_create(self.context, volume)
            for attachment in attachments:
                attachment['volume_id'] = volume['id']
                self._add_attachment_info(self.context, attachment)
            # Commit the increment of the Usage Count for this Volume
            quota.QUOTAS.commit(self.context, reserv)
        except Exception as exc:
            # If the Volume wasn't created, roll-back the increment
            quota.QUOTAS.rollback(self.context, reserv)
            raise exc
        # Send a notification to listeners that we imported this volume
        utils.notify_about_volume_usage(self.context, volume, "import.end")
        # If there was any restricted metadata provided, insert that in the DB
        if restricted_metadata:
            powervc_db.volume_restricted_metadata_update_or_create(
                self.context, volume['id'], restricted_metadata)
        # Return very basic information about the Volume back to the exploiter
        return dict(id=volume['id'], bootable=bootable,
                    attachments=attachments, size=volume['size'])

    def _add_attachment_info(self, context, attachment, new_vol_id=None):
        """Helper method to add the Attachment Info to the Volume"""
        # Need to insert an attachment entry for the Volume
        if attachment:
            connection_info = attachment.get('connection_info', None)
            LOG.debug("Adding attachment_info %(attach)s %(cinfo)s" %
                      dict(attach=attachment, cinfo=connection_info))
            db.volume_attach(self.context, attachment)
            # If the Connection Info was returned, add the Volume ID to it
            if connection_info is not None:
                connection_info['serial'] = attachment['volume_id']
                connection_info['data']['volume_id'] = attachment['volume_id']

    def _unmanage_resource(self, resource, extra_params=None):
        """Stops managing the specified resource by removing it from the DB"""
        LOG.info(_("Unmanaging Volume %s") % str(resource.get('id')))
        # Send a notification to listeners that we are deleting this volume
        utils.notify_about_volume_usage(self.context, resource, "delete.start")
        # Reserve the Quota Reservation saying we are going to remove Usage
        reserv = quota.QUOTAS.reserve(self.context, volumes=-1,
                                      gigabytes=-resource['size'])
        try:
            instance_to_detach = None
            volume_id = resource.get('id')
            if extra_params and 'instance_uuid' in extra_params:
                instance_to_detach = extra_params['instance_uuid']

            volume_attach = db.volume_attachment_get_used_by_volume_id(
                self.context, volume_id)
            # If there are more than one attachment do not remove the volume
            if volume_attach and len(volume_attach) > 1:
                attach = None
                for idx in volume_attach:
                    if idx['instance_uuid'] == instance_to_detach:
                        attach = idx
                        break
                if instance_to_detach and attach:
                    db.volume_detached(self.context, volume_id, attach['id'])
                else:
                    msg = _("Volume %(vol)s has more than one attachment"
                            " and instance id not given") % dict(vol=volume_id)
                    raise Exception(msg)
            else:
                # Remove the Volume instance from the Database
                db.volume_destroy(self.context, resource['id'])
                # Commit the decrement of the Usage Count for this Volume
                quota.QUOTAS.commit(self.context, reserv)
                # Send a notification to listeners that we deleted this volume
                utils.notify_about_volume_usage(self.context, resource,
                                                "delete.end")
        except exception.VolumeNotFound as exc:
            # If the Volume was deleted, roll-back the decrement
            quota.QUOTAS.rollback(self.context, reserv)
            raise exc

    def _filter_unremoveable_resources(self, resources, force):
        """Filters the resources that can't be removed and returns exception"""
        msgb = self._get_message_bundle()
        new_resources, resource_map = (list(), dict())
        # Only do the verification if they didn't ask to force the remove
        if force or cfg.CONF.force_deregister:
            return None
        # Get a list of volumes that are attached to vm's that can't be removed
        attach_vols = self._get_attached_volumes(resources)
        # Get a list of volumes that are used by images in glance
        image_vols = self._get_image_volumes(resources)
        # If there were no attached volumes and no images, we can remove all
        if len(attach_vols) <= 0 and len(image_vols) <= 0:
            return None
        # Loop through the Resources, filtering out those that aren't removable
        for resource in resources:
            # Build a Map of the Resources to use for getting the Display Name
            if resource['id'] in attach_vols or resource['id'] in image_vols:
                resource_map[resource['id']] = resource
            # Build a new list of the Resources that are valid to remove
            else:
                new_resources.append(resource)
        # Replace the original list with only those that are removable
        resources[:] = new_resources
        # Get the list of Names of the Volumes that can't be removed
        attach_names = self._get_volume_names(resource_map, attach_vols)
        image_names = self._get_volume_names(resource_map, image_vols)
        # Create the proper message to throw for what can't be removed
        rmv_msg = msg.ATTACHED_VOLUMES if len(attach_vols) > 0 else None
        if len(image_vols) > 0:
            rmv_msg = msg.ATTACHED_BOTH if rmv_msg else msg.ATTACHED_IMAGES
        return msgb.get_message_text(self.context, rmv_msg,
                                     image_volumes=image_names,
                                     attached_volumes=attach_names)

    def _create_onboard_task(self):
        """Creates an On-boarding Task in the DB to use for this Action"""
        task = powervc_db.onboard_task_create(self.context, self.host)
        anotifier = rpc_api.get_notifier('compute', self.host)
        anotifier.info(self.context, 'volume.task.import.create',
                       dict(host=self.host, task_id=task['id']))
        return task

    def _update_onboard_task(self, task_id, status, progress=None):
        """Updates an On-boarding Task in the DB to use for this Action"""
        if task_id is not None:
            info = dict(status=status, progress=progress)
            task = powervc_db.onboard_task_update(self.context, task_id, info)
            info.update(**dict(host=self.host, task_id=task_id))
            anotifier = rpc_api.get_notifier('compute', self.host)
            anotifier.info(self.context, 'volume.task.import.update', info)
            return task

    def _create_onboard_taskrsc(self, task_id, resource_id,
                                resource_name, status, fault_message=None):
        """Creates an On-boarding Task Record in the DB for this Resource"""
        if task_id is not None:
            values = dict(volume_name=resource_name,
                          status=status, fault_message=fault_message)
            return powervc_db.onboard_task_volume_create(
                self.context, task_id, resource_id, values)

    def _update_onboard_taskrsc(self, task_id, resource_id,
                                status, fault_message=None):
        """Updates an On-boarding Task Record in the DB for this Resource"""
        if task_id is not None:
            values = dict(status=status, fault_message=fault_message)
            return powervc_db.onboard_task_volume_update(
                self.context, task_id, resource_id, values)

    def _send_console_message(self, priority, message):
        """Shared Helper method to send a Message to the Console Log"""
        anotifier = rpc_api.get_notifier('compute', self.host)
        txt = discovery_locale.translate_message(self.context, message)
        if priority == 'INFO':
            anotifier.info(self.context, 'compute.instance.log', {'msg': txt})
        elif priority == 'WARN':
            anotifier.warn(self.context, 'compute.instance.log', {'msg': txt})

    ###############################################
    # ###  Internal On-boarding Helper Methods  ###
    ###############################################
    def _handle_volume_chunking(self, volume_data):
        """Internal Helper Method to Handle Chunking for RPC Calls"""
        volume_list = []
        # If this is a partial list due to chunking, get the next chunk
        while volume_data['chunking'] is True:
            volume_list.extend(volume_data['volumes'])
            # Make an RPC call to the Volume Manager to get the next chunk
            volume_data = self.volume_api.get_next_volumes_chunk(
                self.context, self.host, volume_data['identifier'])
        # Add the final list of Volumes to the overall list to be returned
        volume_list.extend(volume_data['volumes'])
        return volume_list

    def _volume_exists_in_db(self, volume_id):
        """Internal Helper Method to Retrieve the Volume from the DB"""
        try:
            volume = db.volume_get(self.context, volume_id)
            return volume is not None
        # Since the Database API currently throws an exception when not found
        except Exception:
            return False

    def _volume_attachment_exists_in_db(self, attachment):
        """
        Check if attachment entry for given volume and instance
        already exists.
        """
        try:
            volume_id = attachment['volume_id']
            db_attachments = db.volume_attachment_get_used_by_volume_id(
                self.context, volume_id)
            for db_attach in db_attachments:
                db_inst_id = db_attach['instance_uuid']
                if attachment['instance_uuid'] == db_inst_id:
                    #same attachment already exists
                    return True
            return False
        # Since the Database API currently throws an exception when not found
        except Exception:
            return False

    def _find_avail_device_name(self, inst_uuid):
        """Internal Helper Method to find the next available Device Name"""
        used_letters = set([])
        dev_expr = re.compile('^/dev/')
        pref_expr = re.compile('^((x?v|s)d)')
        # Retrieve the Volumes from the Database that are for this Instance
        LOG.debug("get volumes for instance %s" % inst_uuid)
        volumes = powervc_db.ibm_volume_get_all_by_instance(self.context,
                                                            inst_uuid)
        LOG.debug("Returned volumes %s" % volumes)
        # Loop through each Volume, finding the Letters used in the Device Name
        MOUNTPOINT = 3
        for volumeTuple in volumes:
            if volumeTuple[MOUNTPOINT] is not None:
                dev_name = volumeTuple[MOUNTPOINT]
                used_letters.add(pref_expr.sub('', dev_expr.sub('', dev_name)))
        # Figure out which is the first Unused Letter in the Alphabet
        doubles = [first + second for second in string.ascii_lowercase
                   for first in string.ascii_lowercase]
        all_letters = set(list(string.ascii_lowercase) + doubles)
        letters = list(all_letters - used_letters)
        letters.sort(key=lambda x: x.rjust(2, '`'))
        return '/dev/sd' + letters[0]

    def _prep_volume_for_import(self, resource, server_info, local_storage):
        """Internal Helper Method to Prepare the Volume for Import"""
        volume = dict(resource)
        volume['id'] = str(uuid.uuid4())
        # Set all the necessary properties on the Volume that is being imported
        volume['host'] = self.host
        volume['user_id'] = self.context.user_id
        volume['project_id'] = self.context.project_id
        volume['display_name'] = volume.pop('name', None)
        volume['attach_status'] = 'detached'
        volume['availability_zone'] = 'nova'
        # Initialize the additional attributes to null to insert in the DB
        default_attrs = ['id', 'volume_type_id', 'snapshot_id',
                         'created_at', 'launched_at', 'replication_status',
                         'replication_extended_status',
                         'replication_driver_data']
        for attr in default_attrs:
            if attr not in volume:
                volume[attr] = None
        # If this is supposed to be attached to a Server, then set it up
        attach_infos = volume.get('attach_info', {})
        volume['attachments'] = []
        for server_id in attach_infos:
            # if server_id is not None:
            attachment = dict(instance_uuid=server_id)
            # Don't set the host until OpenStack starts setting on attach
            attachment['attach_time'] = timeutils.utcnow()
            attachment['attach_status'] = 'attached'
            attachment['volume_id'] = volume['id']
            attachment['attach_mode'] = 'rw'
            attachment['connection_info'] =\
                attach_infos[server_id].get('connection_info', None)

            # If the driver didn't set the Mount Point, then generate one
            mount_point = attach_infos[server_id].get('mountpoint', None)
            if not mount_point:
                mount_point = self._find_avail_device_name(server_id)
            attachment['mountpoint'] = mount_point
            volume['attachments'].append(attachment)
        if attach_infos and len(attach_infos) > 0:
            volume['attach_status'] = 'attached'
            volume['status'] = 'in-use'
            metadata = volume.get('metadata', {})
            metadata['readonly'] = 'False'
            metadata['attached_mode'] = 'rw'
            volume['metadata'] = metadata
            bootable = volume.pop('bootable', False)
            # Set it as boot-able if the driver tagged it that way and it isn't
            # local storage, since we know it is a data volume if local storage
            if bootable and not local_storage:
                volume['metadata']['is_boot_volume'] = 'True'
        return volume

    def _get_volume_names(self, volume_map, volume_ids):
        """Internal Helper Method to get a list of names of the Volumes"""
        volume_names = list()
        # Loop through each of the Volume ID's building a list of Volume Names
        for volume_id in volume_ids:
            volume_name = volume_map[volume_id].get('display_name')
            # If the Display Name isn't set, fall back to building off the ID
            if volume_name is None:
                volume_name = 'volume-' + volume_map[volume_id]['id']
            volume_names.append(volume_name)
        return self._format_string_list(volume_names)

    def _get_attached_volumes(self, volumes):
        """Internal Helper Method to find which Volumes are Attached to VM's"""
        attached_volumes = []
        # Loop through each of the Volumes, adding the ones that are attached
        for volume in volumes:
            if volume.get('attach_status') == 'attached':
                attached_volumes.append(volume['id'])
        return attached_volumes

    def _get_image_volumes(self, volumes):
        """Internal Helper Method to find which Volumes are used by Images"""
        image_volumes = []
        for volume in volumes:
            image, metadata = (None, None)
            try:
                metadata = db.volume_metadata_get(self.context, volume['id'])
                if metadata and ('is_image_volume' in metadata and
                                 bool(metadata['is_image_volume'])):
                    image_service, image_id = \
                        glance.get_remote_image_service(self.context,
                                                        metadata['image_id'])
                    try:
                        image = image_service.show(self.context.elevated(),
                                                   image_id)
                    except exception.ImageNotFound:
                        pass  # this is actually what we want here
            except Exception:
                LOG.exception(_("Failed checking whether volume %(uuid)s backs"
                                " an image - metadata=%(metadata)s")
                              % dict(uuid=volume['id'], metadata=metadata))
            if image is not None and image.get('status') != 'deleted':
                image_volumes.append(volume['id'])
        return image_volumes
