# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""HMC-managed PowerVM Registrar Classes for performing Host Registration"""

from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_log import log as logging

from powervc_oslo.k2 import k2operator as k2o
from powervc_nova.db import api as db
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.storage import dom as storage_dom
from powervc_nova.virt.ibmpowervm.common import constants as comm_const
from powervc_nova.virt.ibmpowervm.hmc import cod_pool

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration.compute import host_registrar
from powervc_discovery.registration.compute import powervm_registrar
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
HOST_UUID_MAP = dict()
NONCLASSIC = cfg.CONF.nonclassic_supported

cfg.CONF.import_opt('ibmpowervm_ssp_hmc_version_check',
                    'powervc_nova.virt.ibmpowervm.hmc')
cfg.CONF.import_opt('ibmpowervm_register_ssps',
                    'powervc_nova.virt.ibmpowervm.hmc')


########################################################
# #######  HMC PowerVM Registrar Implementation  #######
########################################################
class HMCPowerVMHostRegistrar(powervm_registrar.PowerVMHostRegistrar):
    """HMC-managed PowerVM Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the HMC PowerVM Class for Host Registration"""
        super(HMCPowerVMHostRegistrar, self).__init__(context, **kwargs)
        self._powervm_switch_type = None

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['host_name', 'host_type', 'hmc_uuids']

    def get_sensitive_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        attrs = super(HMCPowerVMHostRegistrar,
                      self).get_sensitive_attributes()
        attrs.append('host_proc_level_type')
        attrs.append('is_master')
        return attrs

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # We only currently support managing a Host from a Single HMC
        if len(config['hmc_uuids']) != 1:
            self._throw_bad_request(msg.SINGLE_HMC)
        hmc_uuid = config['hmc_uuids'][0]
        # Retrieve the IP/Credentials from the DB and load the info to use
        hmc = hmcfact.find_hmc_by_uuid(self.context, hmc_uuid)
        # Make sure that the HMC provided actually has been registered
        if not hmc:
            self._throw_bad_request(msg.HMC_NOTFOUND, hmcname=hmc_uuid)
        # Update the Configuration Properties with the DB Info we Retrieved
        ipcred = {'access_ip': hmc['access_ip'],
                  'user_id': hmc['user_id'], 'password': hmc['password'], }
        new_config = dict(config, **ipcred)
        # Get the K2 Entry for the given Host from the HMC
        k2entry = self.__get_host_from_k2(new_config)
        # Parse the Host Information about the K2 Response from the HMC
        host_info = self.__parse_host_from_k2(k2entry)
        # Copy the necessary attributes parsed from K2 into the Connection Info
        for ck in ['host_name', 'host_mtms', 'host_uuid', 'host_display_name']:
            connect_info[ck] = host_info[ck]
        # Figure out what we are going to set for the Unique Identifier to
        connect_info['unique_id'] = self._get_unique_identifier(connect_info)
        # Retrieve the DHCP Domain to use from what the System has set
        connect_info['dhcp_domain'] = self._get_dhcp_domain(config)
        # Retrieve the Host Processor Level Type to use (POWER6/POWER7)
        connect_info['host_proc_level_type'] = \
            self._get_host_proc_level_type(k2entry)
        # Add the is_master flag to the Connection Info for future reference
        connect_info['is_master'] = \
            HMCPowerVMHostRegistrar.__k2_is_master(k2entry)

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_HMC)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['unauthorized']),
            (msg.CONNECTION_NOTKNOWN, [
                'name or service not known',
                'no address associated with hostname']),
            (msg.NO_HOST_MODES_RETURNED_K2, ['no_host_modes_returned_k2']),
            (msg.NO_HOST_MODES_P6_OR_P7, ['no_host_modes_p6_or_p7']),
            (msg.NO_HOST_FOUND_ON_HMC_FOR_MODES,
             ['no_host_found_on_HMC_for_nodes']),
            (msg.INVALID_CERTIFICATE, ['certificate verify']),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired']),
            (msg.INVALID_VIOS_DB, [
                '- unable to connect to database',
                'rmcvioscmdexception: unable to connect to database']),
            (msg.INVALID_RMC_DOWN, ['ibmpowervmallrmcdown']),
            (msg.INVALID_RMC_NO_SEA, ['ibmpowervmrmcdownnosea']),
            (msg.NO_VIOS_ON_HOST, [
                'no vios was found on host',
                'vios was not found on host']),
            (msg.INVALID_MISSING_SEA, [
                'ibmpowervmvalidseanotfound',
                'sea was not found',
                'vios has no sea configured',
                'vios has no shared ethernet adapter',
                'a shared ethernet adapter (sea) is not configured']),
            (msg.CONNECTION_TIMEOUT_HMC, ['timed out', 'etimedout'])
        ]

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 12443

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def register(self):
        """Registrar Method to update the System Registration"""
        super(HMCPowerVMHostRegistrar, self).register()

    def deregister(self, force=False, uninstall=False, soft=False):
        """Registrar Method to delete a System Registration"""
        # Cleanup any Clusters for this Host here before the HMC is deleted
        if cfg.CONF.ibmpowervm_register_ssps:
            self.__deregister_clusters()
        self.__deregister_cod_pools()
        super(HMCPowerVMHostRegistrar, self).deregister(force, uninstall, soft)

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(HMCPowerVMHostRegistrar,
                        self)._get_service_param_mapping()
        mapping['host_display_name'] = 'host_display_name'
        mapping['host_proc_level_type'] = 'host_proc_level_type'
        mapping['powervm_img_remote_path'] = 'image_remote_path'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(HMCPowerVMHostRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['compute_driver'] = \
            'powervc_nova.virt.ibmpowervm.hmc.driver.PowerVMDriver'
        parm_mapping['compute_manager'] = \
            'powervc_nova.compute.hmc_manager.PowerVCHmcComputeManager'
        parm_mapping['powervm_vif_driver'] = 'powervc_nova.' \
            'virt.ibmpowervm.vif.hmc.driver_hmc.IBMPowerVMVlanVIFDriverHMC'
        parm_mapping['powervm_vif_topo_driver'] = 'powervc_nova.' \
            'virt.ibmpowervm.vif.hmc.topo_hmc.IBMPowerVMNetworkTopoHMC'
        parm_mapping['powervm_mgr_type'] = 'hmc'
        # If the Display Name isn't set, set it to the value on the HMC
        if not parm_mapping.get('host_display_name'):
            parm_mapping['host_display_name'] = \
                connect_info['host_display_name']
        # We need to set the MTMS in the Config File for the Driver to use
        parm_mapping['powervm_mgr_mtms'] = \
            connect_info['host_mtms'].replace(' ', '-')
        # We will also set the Host UUID in the Config File for quicker lookup
        parm_mapping['powervm_mgr_uuid'] = connect_info['host_uuid']
        # If switching from PVM, flag so the driver can do data conversion
        if self._powervm_switch_type is not None:
            parm_mapping['powervm_switch_type'] = self._powervm_switch_type

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        super(HMCPowerVMHostRegistrar, self)._parse_service_config(config)
        hostname = self.config['host_name']
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # Populate the hmc_uuids attribute based on the query from the db
        hmcs = hmcfact.find_all_hmcs_by_host(self.context, hostname)
        self.config['hmc_uuids'] = [hmc['hmc_uuid'] for hmc in hmcs]

    def _post_service_registration_hook(self, action,
                                        connect_info=None, config={}):
        """Hook to be able to take actions after doing the registration"""
        context = self.context
        host_name = self.config['host_name']
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        super(HMCPowerVMHostRegistrar, self).\
            _post_service_registration_hook(action, connect_info, config)
        # First we need to query the HMC to get the list of Hosts
        hmc = hmcfact.find_hmc_by_uuid(context, self.config['hmc_uuids'][0])
        # Adds the New Host to the Member list for the registered HMC
        if action == 'register':
            if host_name not in hmc.managed_hosts:
                hmc.managed_hosts.append(host_name)
                hmc.managed_hosts = hmc.managed_hosts
                hmc.save()
        # As a safety-net cleanup, we are going to cleanup the HMC Hosts table
        if action == 'deregister':
            db.hmc_host_delete(context, host_name)
        # Add the Host to the new HMC's Member List from the old one
        if action == 'update-registration':
            hmc_uuid2 = config.get('hmc_uuids', self.config['hmc_uuids'])[0]
            # If the associated HMC has changed, update it in the DB
            if hmc_uuid2 != self.config['hmc_uuids'][0]:
                hmc2 = hmcfact.find_hmc_by_uuid(context, hmc_uuid2)
                clusters = db.host_cluster_find_all(
                    self.context, dict(host_name=host_name))
                if host_name not in hmc2.managed_hosts:
                    hmc2.managed_hosts.append(host_name)
                    hmc2.managed_hosts = hmc2.managed_hosts
                    hmc2.save()
                # Remove the Host from the old HMC since it now changed
                if host_name in hmc.managed_hosts:
                    hmc.managed_hosts.remove(host_name)
                    hmc.managed_hosts = hmc.managed_hosts
                    hmc.save()
                # Need to update the HMC on the Storage Provider if changed
                provfact = storage_dom.StorageProviderFactory.get_factory()
                for clust in clusters:
                    prov = provfact.find_provider_by_name(
                        self.context, clust['cluster_id'])
                    if prov is not None and prov.hmc_uuids[0] != hmc_uuid2:
                        prov.hmc_uuids = [hmc_uuid2]
                        prov.save()

    def _registration_type_switch_hook(self, action,
                                       connect_info=None, config={}):
        """Hook to be able to switch registration types if needed"""
        hostname = self.config['host_name']
        config, dflt = config or self.config, action != 'register'
        # If they add/edit the host and it isn't the master, don't allow
        if connect_info and connect_info.get('is_master') is False:
            msgid = msg.HOST_HMC_NOT_MASTER \
                if NONCLASSIC else msg.HOST_HMC_NOT_MASTER2
            self._throw_bad_request(msgid, hostname=hostname)
        # Check next if it is currently PVM-managed and need to switch
        if self.__get_manager_type(self.context, hostname) == 'pvm':
            # If they didn't say to force the switch, error out
            if not config.get('force_switch', dflt):
                self._throw_bad_request(
                    msg.HOST_PVM_FORCE_SWITCH, hostname=hostname)
            # Now we can un-configure the old host (and leave data)
            registrar = host_registrar.get_registrar(
                self.context, host_name=hostname, throw_exc=False)
            registrar.deregister(soft=True)
            self._powervm_switch_type = 'pvm'
            return True
        return False

    #############################################
    # ## Implementing HostRegistrar Interface ###
    #############################################
    def switching_registration_type(self, config):
        """Specifies whether the Values will cause the type to switch"""
        is_pvm = 'access_ip' in config and 'user_id' in config
        return is_pvm and 'hmc_uuids' not in config

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        return config['host_uuid']

    def _get_host_proc_level_type(self, k2entry):
        """Retrieves the host processor level (P6/P7) given host System"""
        modeParms = []
        p6ModeFound = False
        p7ModeFound, p8ModeFound = (False, False)
        # Use the SystemName Attribute returned for the Display Name
        dispName = k2entry.element.findtext('SystemName')
        # If the HMC doesn't have a connection to the Host, can't add it
        if k2entry.element.findtext('State') == 'no connection':
            self._throw_bad_request(
                msg.HMC_NO_HOST_CONNECTION, hostname=dispName)
        # If the host is in recovery state on the HMC, can't add it
        if k2entry.element.findtext('State') == 'recovery':
            self._throw_bad_request(
                msg.HMC_HOST_IN_RECOVERY, hostname=dispName)
        # Parse out the Compatibility Modes from the Host's K2 Entry
        elements = k2entry.element.findall(
            './AssociatedSystemProcessorConfiguration/'
            'SupportedPartitionProcessorCompatibilityModes')
        # Build a list of the Compatibility Mode String that are set
        for element in elements:
            if element.gettext():
                modeParms.append(element.gettext())
        # If there weren't any, throw an error that we couldn't fine any
        if len(modeParms) == 0:
            self._throw_bad_request(msg.NO_HOST_MODES_RETURNED_K2,
                                    hostname=dispName)
        # Loop through each of the Mode Strings, seeing if this is P6 or P7
        for modeIndex in range(len(modeParms)):
            if modeParms[modeIndex] == comm_const.POWER8_MODE:
                p8ModeFound = True
            if modeParms[modeIndex] == comm_const.POWER7_MODE:
                p7ModeFound = True
            if modeParms[modeIndex] == comm_const.POWER6_PLUS_MODE:
                p6ModeFound = True
            if modeParms[modeIndex] == comm_const.POWER6_MODE:
                p6ModeFound = True
        # Return the Mode that is the greatest supported Compatibility Mode
        if p8ModeFound:
            return comm_const.PowerSystemType.POWER8
        if p7ModeFound:
            return comm_const.PowerSystemType.POWER7
        if p6ModeFound:
            return comm_const.PowerSystemType.POWER6
        # otherwise if p6 or p7 mode NOT found could be power 5
        self._throw_bad_request(msg.NO_HOST_MODES_P6_OR_P7, hostname=dispName)

    ###############################################
    # #### Shared Registration Helper Methods #####
    ###############################################
    @staticmethod
    def get_hmc_hosts(context, config):
        """Retrieves the Information about the Hosts on the HMC"""
        hosts = list()
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # Retrieve the list of Entries for the Hosts from K2 on the HMC
        k2entries = HMCPowerVMHostRegistrar.__get_hosts_from_k2(config)
        # Loop through each of the K2 Entries parsing out the Host Info
        for entry in k2entries.values():
            host = HMCPowerVMHostRegistrar.__parse_host_from_k2(entry)
            # Provide the Exploiter with the HMC's currently managing the Host
            hmcs = hmcfact.find_all_hmcs_by_host(context, host['host_name'])
            host['hmc_uuids'] = [hmc['hmc_uuid'] for hmc in hmcs]
            # Get the Manager Type from the Host Registrar to fill-in
            host['manager_type'] = HMCPowerVMHostRegistrar.\
                __get_manager_type(context, host['host_name'])
            # Fill-in whether this is a Master HMC or is a Slave
            host['is_master'] = HMCPowerVMHostRegistrar.__k2_is_master(entry)
            # Add an attribute saying whether an HMC or PVM is the master
            host['master_manager_type'] = \
                HMCPowerVMHostRegistrar.__k2_master_manager_type(entry)
            hosts.append(host)
        return hosts

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def __deregister_clusters(self):
        """Helper method to auto-delete Clusters when DeRegistering a Host"""
        host_name = self.config['host_name']
        disp_name = self._get_display_name(self.config)
        db_clusters, host_clusters = (dict(), list())
        prvfact = storage_dom.StorageProviderFactory.get_factory()
        try:
            # Build a mapping of what hosts are registered for each cluster
            for clust in db.host_cluster_find_all(self.context):
                if clust['cluster_id'] not in db_clusters:
                    db_clusters[clust['cluster_id']] = list()
                # If the Cluster Mapping is to this Host, add it to host list
                if clust['host_name'] == host_name:
                    if clust['cluster_id'] not in host_clusters:
                        host_clusters.append(clust['cluster_id'])
                # Add to the db host list if it is for a different host
                else:
                    db_clusters[clust['cluster_id']].append(clust['host_name'])
            # Delete any clusters for this host that isn't part of another Host
            for cluster_id in host_clusters:
                try:
                    # Delete the Provider if no other Hosts for this cluster
                    if len(db_clusters[cluster_id]) <= 0:
                        prvfact.delete_provider(self.context, cluster_id)
                    db.host_cluster_delete(self.context,
                                           cluster_id, host_name)
                except Exception as exc:
                    LOG.warn(_("Error removing Cluster %s") % cluster_id)
                    raise
        # We don't want the thread crashing the process, so catch the exception
        except Exception as exc:
            if unicode(exc).find("PD-A000017") >= 0:
                self._throw_bad_request(msg.SSP_REMOVE_IMGVOLS,
                                        hostname=disp_name,
                                        providername=cluster_id)
            LOG.warn(_("Error Removing Clusters for %s") % host_name)
            LOG.exception(exc)

    def __deregister_cod_pools(self):
        """Helper method to auto-delete CoD pool when DeRegistering a Host"""
        host_name = self.config['host_name']
        host_cod_pool = cod_pool.get_cod_pool_for_host(self.context, host_name)
        if host_cod_pool is None:
            return
        try:
            new_hosts = host_cod_pool.host_names
            del new_hosts[host_name]
            host_cod_pool.host_names = new_hosts
            if host_cod_pool.host_names:
                host_cod_pool.save()
            else:
                # Pool members empty - delete it
                cod_pool_fact = compute_dom.CodPoolFactory.get_factory()
                cod_pool_fact.delete_cod_pool(self.context, host_cod_pool.uuid)
        # We don't want the thread crashing the process, so catch the exception
        except Exception as exc:
            LOG.error(_("Error updating or removing CoD pool for host %s.") %
                      host_name)
            LOG.exception(exc)

    @staticmethod
    @lockutils.synchronized('registrar_host_uuid_cache', 'nova-')
    def __get_uuid_from_cache(hmc_uuid, host_name):
        """Helper method to retrieve the UUID for a Host from the Cache"""
        host_uuid = None
        hmc_hosts = HOST_UUID_MAP.get(hmc_uuid)
        if hmc_hosts is not None:
            host_uuid = hmc_hosts.get(host_name)
        return host_uuid

    @staticmethod
    @lockutils.synchronized('registrar_host_uuid_cache', 'nova-')
    def __put_uuids_in_cache(hmc_uuid, host_uuids):
        """Helper method to update the UUIDs for an HMC in the Cache"""
        hmc_hosts = HOST_UUID_MAP.get(hmc_uuid)
        if hmc_hosts is None:
            HOST_UUID_MAP[hmc_uuid] = hmc_hosts = dict()
        hmc_hosts.update(host_uuids)

    @staticmethod
    def __get_k2_operator(hmc):
        """Helper method to retrieve a list of registered Clusters"""
        session = k2o.K2Session(
            hmc['access_ip'], hmc['user_id'], hmc['password'])
        return k2o.K2Operator(session)

    @staticmethod
    def __k2_operator_read(operator, rootType, rootId=None, timeout=-1):
        """Helper method to perform a K2 read and log the exception"""
        try:
            return operator.read(rootType, rootId, timeout=timeout)
        except Exception as exc:
            # If the exception is a K2 Error, print out the K2 Response
            if isinstance(exc, k2o.K2Error):
                resp, uri = (exc.k2response, ('/' + rootType))
                uri = (uri + '/' + rootId) if rootId else uri
                body = resp.body if resp is not None else None
                LOG.warn(_("Error reading K2 %(uri)s: %(exc)s") %
                         dict(uri=uri, exc=str(body)))
            raise exc

    def __get_host_from_k2(self, config):
        """Helper method to get the K2 Entry for the given Host"""
        host_name = config['host_name']
        hmc_uuid = config['hmc_uuids'][0]
        # Get the map of the Host Name's to UUID's that we have in our Cache
        host_uuid = self.__get_uuid_from_cache(hmc_uuid, host_name)
        # If the given Host Name isn't in the Cache, we need to query all Hosts
        if host_uuid is None:
            k2entries = self.__get_hosts_from_k2(config)
            # The previous line updated the Cache, so get the UUID from it
            # again
            host_uuid = self.__get_uuid_from_cache(hmc_uuid, host_name)
            # If the UUID is still not found, it means the Host doesn't exist
            if host_uuid is None or k2entries.get(host_uuid) is None:
                self._throw_bad_request(msg.NO_HOST_ON_HMC, hostname=host_name)
            return k2entries[host_uuid]
        # If the UUID was in the Cache, then get the K2 Entry based on the UUID
        else:
            operator = self.__get_k2_operator(config)
            timeout = cfg.CONF.service_registration_timeout
            k2resp = self.__k2_operator_read(
                operator, 'ManagedSystem', host_uuid, timeout=timeout)
            return k2resp.entry

    @staticmethod
    def __get_hosts_from_k2(config):
        """Helper method to get the K2 Entries for all of the Hosts"""
        hmc_uuid = config['hmc_uuids'][0]
        k2entries, hostuuids = (dict(), dict())
        # Get a K2 Session to Log-on/Connect to the HMC's K2 API's
        operator = HMCPowerVMHostRegistrar.__get_k2_operator(config)
        # Retrieve the list of /ManagedSystem resources from the K2 API
        timeout = cfg.CONF.service_registration_timeout
        k2resp = HMCPowerVMHostRegistrar.__k2_operator_read(
            operator, 'ManagedSystem', timeout=timeout)
        # Loop through K2 Entries updating the Cache and returning the Entries
        for entry in k2resp.feed.entries:
            host_info = HMCPowerVMHostRegistrar.__parse_host_from_k2(entry)
            hostuuids[host_info['host_name']] = host_info['host_uuid']
            k2entries[host_info['host_uuid']] = entry
        # Update the cached map of Host Name's to UUID's for the given HMC
        HMCPowerVMHostRegistrar.__put_uuids_in_cache(hmc_uuid, hostuuids)
        return k2entries

    @staticmethod
    def __parse_host_from_k2(k2entry):
        """Helper method to Parse Host Info from the K2 Response"""
        uid = k2entry.properties['id']
        # Use the SystemName Attribute returned for the Display Name
        dispname = k2entry.element.findtext('SystemName')
        # Parse each part of the MTMS out of the MTMS Element
        mtms = k2entry.element.find('MachineTypeModelAndSerialNumber')
        model = mtms.findtext('Model')
        mtype = mtms.findtext('MachineType')
        serial = mtms.findtext('SerialNumber')
        # Retrieve the List of Logical Partitions to figure out the Count
        lpars = k2entry.element.find('AssociatedLogicalPartitions')
        num_lpars = len(lpars.findall('link')) if lpars else 0
        hostname = mtype + model + '_' + serial
        hostmtms = mtype + '-' + model + ' ' + serial
        # Add the Name/DisplayName/MTMS/VmCount to the Attributes returned
        return dict(host_name=hostname, host_uuid=uid, host_mtms=hostmtms,
                    host_display_name=str(dispname), host_vms=num_lpars)

    @staticmethod
    def __k2_is_master(k2entry):
        """Helper method to determine if this HMC is the Master or not"""
        reg = HMCPowerVMHostRegistrar
        # Get the properties to see if this is Classic Mode or a Slave
        is_classic = reg.__find_k2_boolean(k2entry, 'IsClassicHMCManagement')
        is_slave = reg.__find_k2_boolean(
            k2entry, 'IsNotPowerVMManagementMaster')
        # If this is Classic Mode (or pre-Classic Mode), then is the Master
        if is_classic is None or is_classic:
            return True
        # Since Partition Profiles aren't support unless it is in Classic
        # Mode we are only able to manage in Classic Mode currently, but
        # we will have a CONF property to override this behavior for testing
        return not is_slave and NONCLASSIC

    @staticmethod
    def __k2_master_manager_type(k2entry):
        """Helper method to determine if this HMC is the Master or not"""
        reg = HMCPowerVMHostRegistrar
        # Get the properties to see if this PVM is the Master or not
        pvm_master = reg.__find_k2_boolean(
            k2entry, 'IsManagementPartitionPowerVMManagementMaster')
        return 'pvm' if pvm_master else 'hmc'

    @staticmethod
    def __get_manager_type(context, host_name):
        """Helper method to determine the Manager Type for the given host"""
        # Get the Manager Type from the Host Registrar to fill-in
        registrar = host_registrar.get_registrar(
            context, host_name=host_name, throw_exc=False)
        # Only populate the Manager Type if it is actually available
        if registrar is not None and registrar.config.get('manager_type'):
            return registrar.config['manager_type']

    @staticmethod
    def __find_k2_boolean(k2entry, key):
        """Helper method to return the boolean value for the given key"""
        val = k2entry.element.findtext(key)
        return utils.bool_from_string(val) if val is not None else None


########################################################
# ########  HMC Periodic Task Implementation  ##########
########################################################
def periodic_task(context):
    """Periodic Task invocation point for fixing up HMC Host Mappings"""
    if cfg.CONF.powervm_mgr_type != 'hmc':
        return False
    try:
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        hosts = hostfact.find_all_hosts(context)
        # Loop through each of the Hosts seeing if any need fixing up
        for host in hosts:
            # Try to fix-up the HMC Host mapping if deleted by accident.
            # If we fixed up either the DB or CONF, then restart the service.
            if _fixup_hmc_host_mapping(context, host):
                svc = 'openstack-nova-compute-' + host.host_name
                utils.run_command('service ' + svc + ' restart', True)
                LOG.info(_("After fix up, restarted service %s") % svc)
    # If we got an error trying to update, just log it and continue on
    except Exception as exc:
        LOG.warn(_("Error trying to fix-up HMC Host Mappings"))
        LOG.exception(exc)
    return False


def _fixup_hmc_host_mapping(context, host):
    """Helper method to fix-up a HMC Host Mapping deleted by bug 24717"""
    try:
        # Nothing to do if the HMC Mapping is already set on the Host
        if not host.host_type or host.hmc_uuids:
            return False
        # Import the SQL ALchemy Modules necessary for the DB queries
        import nova.db.sqlalchemy.api as base_dbapi
        from powervc_nova.db.sqlalchemy import models as dbmodels
        # Create a Session to query on the HMC Host Mapping Table
        hmchost, session = (None, base_dbapi.get_session())
        query = session.query(dbmodels.HmcHostsDTO)
        # If for some reason got here even though there is a mapping, return
        if len(query.filter_by(host_name=host.host_name, deleted=0).all()) > 0:
            return False
        # Loop through each HMC Host Mapping finding the newest one
        for cand_hmchost in query.filter_by(host_name=host.host_name).all():
            if hmchost is None or _is_newer_date_time(cand_hmchost, hmchost):
                hmchost = cand_hmchost
        # If we found a match, mark that one as no longer deleted
        if hmchost is not None:
            hmchost.update(dict(deleted_at=None, deleted=0))
            hmchost.save(session=session)
            LOG.info(_("Fixed up the HMC Host Mapping for %s to "
                       "no longer be deleted.") % host.host_name)
            return True
    # If we got an error trying to update, just log it and continue on
    except Exception as exc:
        LOG.warn(_("Error fixing up HMC Host Mapping for %s") % host.host_name)
        LOG.exception(exc)
    return False


def _is_newer_date_time(new_hmchost, old_hmchost):
    """Helper method to compare if the new date is older than the old"""
    if new_hmchost.deleted_at is None:
        return False
    if old_hmchost.deleted_at is None:
        return True
    return new_hmchost.deleted_at > old_hmchost.deleted_at


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return HMCPowerVMHostRegistrar(context, **kwargs)
