# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""HMC Registrar Classes for performing HMC Registration"""

import eventlet
import six
from oslo_config import cfg
from oslo_log import log as logging
from oslo_concurrency import lockutils
from nova.volume import cinder

from powervc_oslo.k2 import k2operator as k2o
from powervc_nova.db import api as db
from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.objects.storage import dom as storage_dom

from powervc_discovery.registration import registrar
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.compute import powervm_registrar
from powervc_discovery.registration.compute import hmc_powervm_registrar
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)

cfg.CONF.import_opt('powervm_mgr_type', 'powervc_nova.virt.ibmpowervm.ivm')


########################################################
# #######  HMC PowerVM Registrar Implementation  #######
########################################################
class HMCRegistrar(registrar.BaseRegistrar):
    """HMC Registrar Class to handle HMC Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the HMC Class for HMC Registration"""
        inchosts = kwargs.get('include_hosts')
        self.include_hosts = False if inchosts is None else inchosts
        super(HMCRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['access_ip', 'user_id', 'password']

    def get_sensitive_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['password', 'unique_id', 'include_hosts']

    def register(self):
        """Sub-classes needs to override to do the Host Registration"""
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # Make sure that this is an HMC-managed environment before registering
        if cfg.CONF.powervm_mgr_type != 'hmc':
            self._throw_bad_request(msg.NONHMC_ENVIRONMENT)
        # Verify that we can establish a connection to the specified HMC
        connect_info = self.establish_connection(self.config.copy())
        # Use the UUID from the K2 API's as the Unique Identifier
        hmc_uuid = connect_info['unique_id']
        self.config['hmc_uuid'] = hmc_uuid
        self.config['unique_id'] = hmc_uuid
        self.config['mtmsn'] = connect_info['mtmsn']
        # Retrieve the HMC Information from the Database for the Name given
        hmc = hmcfact.find_hmc_by_uuid(self.context, hmc_uuid)
        # If the HMC already exists in the Database, then throw an exception
        if hmc:
            hmcname = hmc['hmc_display_name']
            self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=hmcname)
        # If the Display Name wasn't provided, retrieve it from the HMC
        if not self.config.get('hmc_display_name'):
            self.config['hmc_display_name'] = connect_info['hmc_display_name']
        # Adding the HMC Entry to the Database for our registration
        dbcfg = self.__get_config_for_db(self.config)
        hmc = hmcfact.construct_hmc(self.context, auto_create=True, **dbcfg)
        # We need to get the Registered At Attribute from the DB
        self.__update_config_from_db(self.config, hmc)
        # Synchronize the HMC IP/Credential Info in Cinder for SSP usage
        self.__sync_hmc_for_storage(hmc_uuid, self.config, 'register')

    def deregister(self, force=False, uninstall=False):
        """Sub-classes needs to override to do the Host Deletion"""
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        regtype = self._get_registration_type_string()
        hmc_uuid = self.config['hmc_uuid']
        LOG.info(_("Removing %(regtype)s %(host)s...") %
                 dict(regtype=regtype, host=hmc_uuid))
        # First we need to remove the hosts that are registered for the HMC
        self.__remove_host_registrars()
        # Retrieve the HMC Information from the Database for the Name given
        hmc = hmcfact.find_hmc_by_uuid(self.context, hmc_uuid)
        # If the HMC wasn't found in the Database, then throw an exception
        if not hmc:
            self._throw_not_found(msg.SYSTEM_NOTFOUND, hostname=hmc_uuid)
        # Update the Config Attributes with the Info retrieved from the DB
        self.__update_config_from_db(self.config, hmc)
        # Delete the HMC from the Database that we were asked to deregister
        hmcfact.delete_hmc(self.context, self.config['hmc_uuid'])
        # Synchronize the HMC IP/Credential Info in Cinder for SSP usage
        self.__sync_hmc_for_storage(self.config['hmc_uuid'], {}, 'deregister')
        LOG.info(_("Removed %(regtype)s %(host)s.") %
                 dict(regtype=regtype, host=hmc_uuid))

    def update_registration(self, updated_values):
        """Sub-classes needs to override to update the Host Registration"""
        hmc_uuid = self.config['hmc_uuid']
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # First make sure we can establish a connection with the new HMC config
        new_config = dict(self.config, **updated_values)
        connect_info = self.establish_connection(new_config)
        change = new_config.get('certificate_changed') is True
        change = change or self.config['access_ip'] != new_config['access_ip']
        change = change or self.config['user_id'] != new_config['user_id']
        change = change or self.config['password'] != new_config['password']
        # Verify that the IP specified is not changing it to a different HMC
        self._verify_updating_same_system(connect_info, updated_values)
        # Update the new HMC Config Attributes provided in the Database
        hmc = hmcfact.find_hmc_by_uuid(self.context, hmc_uuid)
        hmc.update(self.__get_config_for_db(new_config))
        hmc.save()
        # Synchronize the HMC IP/Credential Info in Cinder for SSP usage
        self.__sync_hmc_for_storage(hmc_uuid, new_config, 'update-register')
        # Need to Notify the Host Registrars so they can restart nova-compute
        self.__update_host_registrars(force_restart=change)

    def _retrieve_registration(self):
        """Sub-classes needs to override to retrieve the Host Registration"""
        hmc_uuid = self.config['hmc_uuid']
        hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
        # Retrieve the HMC Information from the Database for the Name given
        hmc = hmcfact.find_hmc_by_uuid(self.context, hmc_uuid)
        # If the HMC wasn't found in the Database, then throw an exception
        if not hmc:
            self._throw_not_found(msg.SYSTEM_NOTFOUND, hostname=hmc_uuid)
        # Update the Config Attributes with the Info retrieved from the DB
        self.__update_config_from_db(self.config, hmc)
        # If they asked to retrieve the hosts managed, connect to query them
        if self.include_hosts:
            try:
                cfg = dict(self.config, query_hosts=True)
                cfg['auto_add_certificate'] = False
                cfg['hmc_uuids'] = [hmc_uuid]
                with eventlet.timeout.Timeout(60):
                    connect_info = self.establish_connection(cfg)
                    self.config['available_hosts'] =\
                        [h['host_name'] for h in connect_info['hosts']]
            # If we got a timeout exception, just log it and continue on
            except eventlet.timeout.Timeout:
                LOG.warn(_("Timed out querying hosts for HMC %s") % hmc_uuid)
            # If we got any other exception, just log it and continue on
            except Exception as exc:
                info = (hmc_uuid, exc)
                LOG.warn(_("Exception querying hosts for HMC %s: %s") % info)

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        resp = None
        # Initialize protocol necessary for connection
        self._initialize_connection(config)
        # Retrieve the list of Hosts that are managed by the HMC
        if config.get('query_hosts'):
            connect_info['hosts'] = hmc_powervm_registrar.\
                HMCPowerVMHostRegistrar.get_hmc_hosts(self.context, config)
        else:
            # Get a K2 Session to Log-on/Connect to the HMC's K2 API's
            session = k2o.K2Session(
                config['access_ip'], config['user_id'], config['password'])
            operator = k2o.K2Operator(session)
            try:
                # Retrieve the list of /ManagementConsole resources from K2
                resp = operator.read('ManagementConsole')
            except Exception as exc:
                # If the exception is a K2 Error, print out the K2 Response
                if isinstance(exc, k2o.K2Error):
                    resp, uri = (exc.k2response, '/ManagementConsole')
                    body = resp.body if resp is not None else None
                    LOG.warn(_("Error reading K2 %(uri)s: %(exc)s") %
                             dict(uri=uri, exc=six.text_type(body)))
                raise exc
            # Retrieve the Unique Identifier to use for the HMC
            connect_info['unique_id'] = resp.feed.entries[0].properties['id']
            # Retrieve the Display Name the user provided on the HMC
            connect_info['hmc_display_name'] = \
                resp.feed.entries[0].element.findtext('ManagementConsoleName')
            # Verify that this isn't really a NovaLink partition instead
            if resp.headers and resp.headers.get('x-mc-type') == 'PVM':
                hmcname = config.get(
                    'hmc_display_name', connect_info['hmc_display_name'])
                self._throw_bad_request(msg.PVM_NOT_HMC, hmcname=hmcname)
            # Retrieve the MTMSN (machine type, model, and serial number)
            mtmsn_elem = resp.feed.entries[0].element.find(
                'MachineTypeModelAndSerialNumber')
            machine_type = mtmsn_elem.findtext('MachineType')
            model = mtmsn_elem.findtext('Model')
            serial_number = mtmsn_elem.findtext('SerialNumber')
            connect_info['mtmsn'] = "{0}-{1} {2}".format(machine_type,
                                                         model, serial_number)

    def _get_registration_type_string(self):
        """Sub-classes needs to override to do the Host Registration"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.REG_TYPE_HMC)

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_HMC)

    def _get_registration_log_file(self):
        """Describes the name of the Log File running the Registration"""
        return '/var/log/nova/nova-api.log'

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['unauthorized']),
            (msg.CONNECTION_TIMEOUT_HMC, ['timed out', 'etimedout']),
            (msg.CONNECTION_NOTKNOWN, [
                'label empty or too long',
                'name or service not known',
                'no address associated with hostname']),
            (msg.CONNECTION_NOROUTE, ['enetunreach', 'ehostunreach']),
            (msg.CONNECTION_FAILED_HMC, ['connection', 'econnrefused']),
            (msg.INVALID_CERTIFICATE, ['certificate verify', "doesn't match"]),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired'])
        ]

    def _get_name_attribute(self):
        """Provides the Name of the Attribute to use for the system name"""
        return 'hmc_uuid'

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 12443

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        autoadd, old_cert = (True, None)
        if config.get('auto_add_certificate') is not None:
            autoadd = config['auto_add_certificate']
            # If we might update the certificate, get the existing to compare
            if autoadd:
                old_cert = utils.get_existing_certificate(config['access_ip'])
        new_cert = utils.verify_certificate(
            config, self._get_certificate_port())
        # If we have an existing certificate, see if it has changed
        if old_cert is not None:
            # If the certificate changed, then we need to flag for processing
            if new_cert is not None and old_cert != new_cert:
                config['certificate_changed'] = True
        return new_cert

    def __remove_host_registrars(self):
        """Tells the Host Registrars that the HMC is being removed"""
        # Loop through the Hosts, telling the Registrar the HMC is removed
        for host in self.config.get('managed_hosts', []):
            reg = powervm_registrar.get_registrar(self.context, host_name=host)
            reg.deregister()

    def __update_host_registrars(self, force_restart=False):
        """Notifies the Host Registrars that the HMC IP/Creds was updated"""
        # Loop through the Hosts, notifying the Registrar the HMC was updated
        for host in self.config.get('managed_hosts', []):
            reg = powervm_registrar.get_registrar(self.context, host_name=host)
            reg.update_registration({}, force_restart=force_restart,
                                    wait_for_restart=False)

    @lockutils.synchronized('sync_hmc', 'nova-')
    def __sync_hmc_for_storage(self, hmc_uuid, config, action):
        """Helper method to sync the HMC info in the Cinder DB for SSP's"""
        try:
            registration = dict()
            # Set up the Configuration Information to call the REST API with
            registration['hmc_uuid'] = config.get('hmc_uuid')
            registration['hmc_display_name'] = config.get('hmc_display_name')
            registration['access_ip'] = config.get('access_ip')
            registration['user_id'] = config.get('user_id')
            registration['password'] = config.get('password')
            hmcinfo = dict(registration=registration)
            # Get the Cinder REST Client and make the REST call
            client = cinder.cinderclient(self.context).client
            # If this is a Register, do the HTTP POST to /ibm-hmcs
            if action == 'register':
                client.post('/ibm-hmcs', body=dict(hmc=hmcinfo), log=False)
            # If this is a ReRegister, do the HTTP PUT to /ibm-hmcs
            elif action == 'update-register':
                client.put('/ibm-hmcs/' + hmc_uuid, body=hmcinfo, log=False)
            # If this is a DeRegister, do the HTTP DELETE to /ibm-hmcs
            elif action == 'deregister':
                self.__switch_hmc_for_storage(hmc_uuid)
                client.delete('/ibm-hmcs/' + hmc_uuid)
        # Log the exception and continue on if the Synchronization failed
        except Exception as exc:
            actstr = action + ' - ' + str(hmc_uuid)
            LOG.warn(_("Error synchronizing HMC in Cinder for %s") % actstr)
            LOG.exception(exc)

    def __switch_hmc_for_storage(self, hmc_uuid):
        """Internal Helper Method to Switch HMC for Storage when deleted"""
        cluster_hmcs = dict()
        # Build a list of the HMC's available for each VIOS Cluster
        for clust in db.host_cluster_find_all(self.context):
            if clust['cluster_id'] not in cluster_hmcs:
                cluster_hmcs[clust['cluster_id']] = list()
            if clust['hmc_uuid'] not in cluster_hmcs[clust['cluster_id']]:
                cluster_hmcs[clust['cluster_id']].append(clust['hmc_uuid'])
        # Get a list of Storage Providers to determine which use this HMC
        prvfact = storage_dom.StorageProviderFactory.get_factory()
        for prov in prvfact.find_all_providers_by_type(self.context, 'ssp'):
            # If it is using this HMC currently, see if we can switch it
            if len(prov.hmc_uuids) == 1 and hmc_uuid in prov.hmc_uuids:
                hmc_list = list(cluster_hmcs.get(prov.host_name, list()))
                # If there is another HMC, update the Storage Provider to use
                if hmc_uuid in hmc_list:
                    hmc_list.remove(hmc_uuid)
                if len(hmc_list) > 0:
                    prov.hmc_uuids = [hmc_list[0]]
                    prov.save()

    @staticmethod
    def __get_config_for_db(config):
        """Internal Helper Method to parse the DB Config Values"""
        attrs = ['hmc_uuid', 'mtmsn', 'access_ip', 'user_id',
                 'password', 'hmc_display_name']
        return dict([(attr, config[attr]) for attr in attrs if attr in config])

    @staticmethod
    def __update_config_from_db(config, hmcdb):
        """Internal Helper Method to Update the Config from the DB Values"""
        attrs = ['hmc_uuid', 'mtmsn', 'access_ip', 'user_id', 'password',
                 'hmc_display_name', 'registered_at', 'managed_hosts']
        # Copy over the attributes from the DB to the Config to return
        for attr in attrs:
            config[attr] = hmcdb.get(attr)


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return HMCRegistrar(context, **kwargs)


def list_hmc_registrars(context, include_hosts=False):
    """Factory Method to retrieve the list of all of the Registrars"""
    registrars = []
    hmcfact = compute_dom.ManagementConsoleFactory.get_factory()
    # Retrieve all of the HMC's from the DB to be able to list Registrars
    hmcs = hmcfact.find_all_hmcs(context)
    for hmc in hmcs:
        registrar = HMCRegistrar(
            context, hmc_uuid=hmc['hmc_uuid'], include_hosts=include_hosts)
        registrars.append(registrar)
    return registrars
