# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Register a new HMC"""

from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging

from powervc_discovery.registration.compute import hmc_registrar
from powervc_discovery import locale as discovery_locale

LOG = logging.getLogger(__name__)
ALIAS = 'hmc-registration'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'hmc_registration')


class HMCRegistrationControllerExtension(wsgi.Controller):
    """Main Controller Class to provide the ibm-hmcs API for Registration"""

    def index(self, req):
        """Implement Rest GET Method to return a list of HMC's"""
        hmcs = []
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'index')
        # Get the list of registered HMC's to return basic info about them
        registrars = hmc_registrar.list_hmc_registrars(context)
        for registrar in registrars:
            hmcs.append(
                {'hmc_uuid': registrar.config['hmc_uuid'],
                 'hmc_display_name': registrar.config['hmc_display_name']})
        return {'hmcs': hmcs}

    def show(self, req, id):
        """Implement Rest GET Method to return a details of a specific HMC"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'show')
        # Check for the query parameter for including all available Hosts
        inchosts = req.GET.get('include_available_hosts')
        inchosts = inchosts is not None and str(inchosts).lower() == 'true'
        # Load the appropriate Registrar to Retrieve the Registration Info
        registrar = hmc_registrar.get_registrar(
            context, hmc_uuid=id, include_hosts=inchosts)
        registration = self.filter_config(registrar)
        return {'hmc': {'registration': registration}}

    def detail(self, req):
        """Implement Rest GET Method to return a list of HMC details"""
        hmcs = []
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'detail')
        # Check for the query parameter for including all available Hosts
        inchosts = req.GET.get('include_available_hosts')
        inchosts = inchosts is not None and str(inchosts).lower() == 'true'
        # Get the list of registered HMC's to return basic info about them
        registrars = hmc_registrar.list_hmc_registrars(
            context, include_hosts=inchosts)
        for registrar in registrars:
            hmcs.append(self.filter_config(registrar))
        return {'hmcs': hmcs}

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Create an HMC Registration"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'create')
        registration = body['hmc']['registration']
        # Load the appropriate Registrar to do the HMC Registration
        registrar = hmc_registrar.get_registrar(context, create=True,
                                                **registration)
        # Delegate to the Registrar to do the HMC Registration
        registrar.register()
        # Return on the Response the Registration Info
        return {'hmc': {'registration': self.filter_config(registrar)}}

    @wsgi.action("update")
    def update(self, req, id, body):
        """Add Rest PUT Method to Update an HMC Registration"""
        hmc = {'hmc': id}
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'update')
        registration = body.pop('registration', None)
        # If no Registration Info is provided, then updating a base attribute
        if registration is not None:
            # Load the appropriate Registrar to do the HMC Re-Registration
            registrar = hmc_registrar.get_registrar(context, hmc_uuid=id)
            # Delegate to the Registrar to do the HMC Re-Registration
            registrar.update_registration(registration)
            hmc['registration'] = self.filter_config(registrar,
                                                     config=registration)
        return hmc

    @wsgi.action("delete")
    @wsgi.response(204)
    def delete(self, req, id):
        """Add Rest DELETE Method to Remove an HMC Registration"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'delete')
        # Load the appropriate Registrar to do the HMC De-Registration
        registrar = hmc_registrar.get_registrar(context, hmc_uuid=id)
        # Delegate to the Registrar to do the HMC De-Registration
        registrar.deregister()

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def filter_config(registrar, config=None):
        """Internal Helper Method to Remove any Sensitive Attributes"""
        new_config = registrar.config.copy()
        if config is not None:
            new_config = config.copy()
        # Loop through each of the Sensitive Attributes, removing them
        for attr in registrar.get_sensitive_attributes():
            new_config.pop(attr, None)
        return new_config


class HMCHostsControllerExtension(wsgi.Controller):
    """Main Controller Class to provide the ibm-hmcs API for Registration"""

    def index(self, req, hmc_id):
        """Implement Rest GET Method to return a list of HMC's"""
        context = req.environ['nova.context']
        mainext = HMCRegistrationControllerExtension
        context = mainext._validate_authorization(context, req, 'list_hosts')
        # Load the appropriate Registrar to retrieve the HMC Info
        registrar = hmc_registrar.get_registrar(context, hmc_uuid=hmc_id)
        # Retrieve the list of Host that the HMC is currently managing
        registrar.config['auto_add_certificate'] = False
        registrar.config['query_hosts'] = True
        registrar.config['hmc_uuids'] = [hmc_id]
        connection_info = registrar.establish_connection(registrar.config)
        return {'hosts': connection_info['hosts']}


class HMCRegistration(extensions.ExtensionDescriptor):
    """Capability to Register a new Hardware Management Console."""
    name = "IBM Hardware Management Console Registration"
    alias = "ibm-hmc-registration"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_hmc_registration/api/v1.1"
    updated = "2013-03-21T21:00:00-06:00"

    def get_resources(self):
        """Provide a Sub-Resource Extension to the ibm-hmcs Resource"""
        resource1 = extensions.ResourceExtension(
            'ibm-hmcs', controller=HMCRegistrationControllerExtension(),
            collection_actions={'detail': 'GET', 'update': 'PUT'})
        resource2 = extensions.ResourceExtension(
            'hosts', controller=HMCHostsControllerExtension(),
            parent=dict(collection_name='ibm-hmcs', member_name='hmc'))
        return [resource1, resource2]


class HMCRegistrationV21(extensions.V21APIExtensionBase):
    """Capability to Register a new Hardware Management Console."""
    name = "IBM Hardware Management Console Registration"
    alias = "ibm-hmc-registration"
    version = 1

    def get_resources(self):
        """Provide a Sub-Resource Extension to the ibm-hmcs Resource"""
        resource1 = extensions.ResourceExtension(
            'ibm-hmcs', controller=HMCRegistrationControllerExtension(),
            collection_actions={'detail': 'GET', 'update': 'PUT'})
        resource2 = extensions.ResourceExtension(
            'hosts', controller=HMCHostsControllerExtension(),
            parent=dict(collection_name='ibm-hmcs', member_name='hmc'))
        return [resource1, resource2]

    def get_controller_extensions(self):
        return []
