# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base Registrar Classes for performing Host Registration"""

import re
import sys
import cgi
import commands
from oslo_config import cfg

from nova import db
from nova import exception as nova_exception
from nova import rpc as rpc_api
from nova.objects import compute_node as compute_obj
from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_utils import timeutils

from powervc_nova import compute
from powervc_nova.common import aggregate_utils as agg_utils
from powervc_nova.objects.compute import dom as compute_dom

from powervc_discovery.onboarding.compute import vm_importer
from powervc_discovery.registration import registrar as reg
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.ListOpt('supported_host_types', default=['powervm'],
                help='The list of Host Types that are supported'),
    cfg.BoolOpt('nonclassic_supported', default=False,
                help='Whether or not HMC Non-Classic Mode is supported')
])

LOG = logging.getLogger(__name__)


########################################################
# ##########  Host Registrar Implementation  ###########
########################################################
class HostRegistrar(reg.ServiceRegistrar):
    """Abstract Base Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Host Registration"""
        super(HostRegistrar, self).__init__(context, **kwargs)
        self.compute_api = compute.HostAPI()

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        return ['password', 'private_key', 'private_key_data',
                'unique_id', 'dhcp_domain']

    def get_immutable_attributes(self):
        """Provides the list of Attributes that can't be updated"""
        return ['host_name', 'host_type']

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        # Initialize protocol necessary for connection
        self._initialize_connection(config)
        # Retrieve the Unique Identifier to use for the System
        connect_info['unique_id'] = self._get_unique_identifier(config)
        # Retrieve the DHCP Domain to use from what the System has set
        connect_info['dhcp_domain'] = self._get_dhcp_domain(config)
        # Retrieve the Display Name to use from what the System has set
        connect_info['host_display_name'] = self._get_display_name(config)
        # Retrieve the host_proc_level_type (power 7)
        connect_info['host_proc_level_type'] =\
            self._get_host_proc_level_type(config)

    def _get_registration_type_string(self):
        """Provides the translated Registration Type for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.REG_TYPE_HOST)

    def _get_registration_log_file(self):
        """Describes the name of the Log File running the Registration"""
        return '/var/log/nova/nova-api.log'

    def _get_name_attribute(self):
        """Provides the Name of the Attribute to use for the system name"""
        return 'host_name'

    def _get_private_key_prepend(self):
        """Provides the Directory and Prefix for the Private Key Location"""
        return '/etc/nova/rsakeys/nova_'

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def register(self):
        """Registrar Method to perform the System Registration"""
        super(HostRegistrar, self).register()
        self.__update_aggregates_for_register()

    def _get_service_name(self):
        """Provides the name of the Linux Service created by this Registrar"""
        return 'openstack-nova-compute-' + self.config['host_name']

    def _get_service_config_file(self):
        """Provides the name of the Config File created by this Registrar"""
        return '/etc/nova/nova-' + self.config['host_name'] + '.conf'

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        return {'host_type': 'host_type',
                'host_uniqueid': 'unique_id', 'dhcp_domain': 'dhcp_domain',
                'host_display_name': 'host_display_name'}

    def _pre_service_registration_hook(self, action, connect_info=None,
                                       config={}, force=False):
        """Hook to be able to take actions before doing the registration"""
        super(HostRegistrar, self).\
            _pre_service_registration_hook(action, connect_info, config, force)
        # If the Display Name wasn't set by the User, add it to the config
        if action == 'register':
            self.config['unique_id'] = connect_info.get('unique_id')
            self.config['dhcp_domain'] = connect_info.get('dhcp_domain')
            self.config['host_display_name'] = \
                connect_info.get('host_display_name')
            if connect_info.get('host_proc_level_type'):
                self.config['host_proc_level_type'] = \
                    connect_info['host_proc_level_type']
        # Confirm that this Compute Host is able to be ReRegistered
        if action == 'update-registration':
            self.__verify_allowed_to_edit(force)
        # Confirm that this Compute Host is able to be DeRegistered
        if action == 'deregister':
            self.__verify_allowed_to_delete(force)
            self.__driver_host_cleanup(self.config['host_name'])

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        super(HostRegistrar, self)._parse_service_config(config)
        # TODO: Can remove once we have the Host Type always in the Config File
        if self.config.get('host_type') is None:
            self.config['host_type'] = 'powervm'

    def _get_service_instance(self, verify_existence=False):
        """Retrieves the Service Instance from the Database for this System"""
        hostname = self.config['host_name']
        services = db.service_get_all(self.context)
        # Since the topic/host filter excludes disabled, filter ourselves
        services = [svc for svc in services if svc['topic'] == 'compute']
        services = [svc for svc in services if svc['host'] == hostname]
        service = services[0] if services else None
        # We will wait until the Compute Node exists if this is verification
        if verify_existence and service is not None:
            try:
                compute_obj.ComputeNodeList.\
                    get_all_by_host(self.context, hostname)
            # If the Compute Node wasn't found, it will throw an exception
            except nova_exception.NotFound:
                return None
        return service

    def _destroy_service_instance(self, service):
        """Cleans up the Service Instance from the Database on Deregister"""
        context = self.context
        hostname = self.config['host_name']
        hostfact = compute_dom.ManagedHostFactory.get_factory()
        anotifier = rpc_api.get_notifier('compute', hostname)
        # Cleanup each of the Instances/VM's returned from the database
        importer = vm_importer.VmImporter(context, hostname)
        importer.unmanage_resources(['*ALL'], force=True)
        # Retrieve the Compute Node to have info to send in the notification
        compute_node = self.__get_compute_node_instance(service)
        # If the Compute Node was found, send a notification we are deleting
        if compute_node:
            anotifier.info(context, 'compute.node.delete.start', compute_node)

        @lockutils.synchronized('reconcile_aggregates',
                                'nova-reconcile-aggregates-',
                                external=True)
        def _remove_host_from_aggregates_and_delete_host():
            self.__update_aggregates_for_deregister()
            # Clean up the Service and Compute Host entries from the database
            hostfact.delete_host(self.context, hostname)

        _remove_host_from_aggregates_and_delete_host()
        # If the Compute Node was found, send a notification we deleted
        if compute_node:
            anotifier.info(context, 'compute.node.delete.end', compute_node)

    def _verify_no_existing_service(self, connect_info):
        """Helper Method to see if the System already has been Registered"""
        new_id = connect_info.get('unique_id')
        # Loop through the configuration files to make sure there aren't any
        # existing systems registered that are for the same system as this one
        for filenam in utils.get_matching_files('/etc/nova', 'nova-', '.conf'):
            hostname = filenam[5: len(filenam) - 5]
            registrar = HostRegistrar(self.context, host_name=hostname)
            old_id = registrar.config.get('unique_id')
            # If the Unique Identifier matches between 2, block adding the new
            if old_id and new_id and old_id == new_id:
                dispname = registrar.config.get('host_display_name', hostname)
                self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=dispname)

    def _wait_for_service_initialization(self):
        """Allows sub-class to wait longer for service initialization"""
        host = self.config['host_name']
        # Wait up to an additional 60 seconds for the service to start
        self.compute_api.verify_host_running(self.context, host, 60)

    def _describe_host_service_error(self, force_check=False):
        """Determine what the Reason was that the Host Service didn't start"""
        status = self._host_service_status()
        # Check first to see if the Service is Started, Stopped or Crashed
        if status == 'started' and not force_check:
            return ('ok', 'Nova Compute Service is Started')
        if status == 'stopped':
            return ('service_stopped', 'Nova Compute Service is Stopped')
        # If the Service crashed, let's try to figure out why it crashed
        hostlog, output = self._gather_host_service_log()
        # Call the Helper Method to parse the Error Output provided
        return self._describe_connection_error(output, hostlog)

    def _gather_host_service_log(self):
        """Run the command to gather the Host's Service Log for errors"""
        hostlog = '/var/log/nova/nova-compute-'
        hostlog += self.config['host_name'] + '.log'
        return (hostlog, commands.getoutput('tail -n 75 ' + hostlog))

    ###############################################
    #      Defining HostRegistrar Interface       #
    ###############################################
    def update_software(self, software_versions):
        """Registrar Method to update the PowerVC Software"""
        dispname = self._get_display_name(self.config)
        self._throw_bad_request(msg.NO_SOFTWARE_UPDATE, hostname=dispname)

    def switching_registration_type(self, config):
        """Specifies whether the Values will cause the type to switch"""
        return False

    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        raise NotImplementedError()

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        raise NotImplementedError()

    def _get_host_proc_level_type(self, config):
        """Retrieves the host processor level (P6/P7) given host System"""
        return None

    def _get_host_dhcp_domain(self, config):
        """Retrieves the DHCP Domain Name for the specified System"""
        return None

    ###############################################
    # ######  Common Shared Helper Methods  #######
    ###############################################
    def _get_display_name(self, config):
        """Retrieves the Display Name for the given System"""
        name = self.config.get('host_name')
        access_ip = self.config.get('access_ip')
        display_name = self.config.get('host_display_name')
        # Don't even bother to generate a Display Name if one is already set
        if display_name:
            return cgi.escape(display_name)
        # If the IP Address is set, we will generate it based on the IP Address
        if access_ip:
            return utils.formulate_hostname(access_ip)
        return name if name else None

    def _get_dhcp_domain(self, config):
        """Retrieves the DHCP Domain Name for the specified System"""
        domain, dflt_domain = (None, cfg.CONF.get('dhcp_domain', 'novalocal'))
        # If the Customer overrode the Default Config value, use that value
        if dflt_domain != 'novalocal':
            return dflt_domain
        # Next we will ask the sub-class to retrieve the Domain from the Host
        host_domain = self._get_host_dhcp_domain(config)
        if host_domain:
            return host_domain
        try:
            # Run a command on the Mgmt System to get the DHCP Domain to use
            output = utils.run_command('cat /etc/resolv.conf', True, False)
            # Parse the Domain out of the resolv.conf output
            domain = self._parse_domain_entry(output)
        except Exception as exc:
            LOG.warn(_("Couldn't retrieve domain name, using localdomain"))
            LOG.exception(exc)
        # If the domain wasn't found then we will use localdomain as default
        return domain if domain else 'localdomain'

    @staticmethod
    def _parse_domain_entry(output):
        """Shared Helper Method to Parse the Domain Entry from resolve.conf"""
        domain = None
        # Parse the domain or entry of the resolv.conf file ouptut
        # The C resolver uses the last entry if multiple exist, so follow that
        for line in output[0].splitlines() if output else []:
            columns = re.split('[ \t]', line)
            # If found a domain or search entry is in the file, use it
            if len(columns) > 1 and columns[0] in ['domain', 'search']:
                domain = columns[1]
        return domain

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def __get_compute_node_instance(self, service):
        """Internal Helper Method to Retieve the Compute Node from the DB"""
        try:
            # Try to retrieve the Compute Node Instance from the Database
            node = compute_obj.ComputeNodeList.\
                get_all_by_host(self.context, service['host'])[0]
            # Only provide the essentials attributes for the notification
            return {'compute_node_id': node['id'], 'host': service['host'],
                    'service_id': service['id'],
                    'hypervisor_hostname': node['hypervisor_hostname']}
        except Exception as exc:
            # Log the Error to help in debug, but we still want to do cleanup
            LOG.warn(_("Error Retrieving Compute Node %(host)s on "
                       "Cleanup, continuing on: %(exc)s") %
                     dict(host=service['host'], exc=unicode(exc)))
            return None

    def __driver_host_cleanup(self, host):
        """Helper Method to allow the Driver to do cleanup on Host Removal"""
        try:
            self.compute_api.unmanage_host(self.context, host)
        except Exception as ex:
            LOG.warn(_("Exception on Driver cleanup, "
                       "moving on: %s") % unicode(ex))
            LOG.exception(ex)

    def __verify_allowed_to_delete(self, force):
        """Internal Helper method to Verify the Compute Host can be deleted"""
        host = self.config['host_name']
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        # Only do the verification if they didn't ask to force the delete
        if not force and not cfg.CONF.force_deregister:
            instances = instfact.find_all_instances_by_host(self.context, host)
            # First make sure that none of the Instances have Tasks in-progress
            self.__verify_no_tasks_inprogress(
                instances, msg.REMOVED_ACTION, msg.REMOVING_ACTION)

    def __update_aggregates_for_register(self):
        """Adds the host to the default host aggregate"""
        try:
            agg_utils.add_host_to_aggregate(
                self.context, self.config['host_name'],
                cfg.CONF.default_host_aggregate_name)
        except nova_exception.AggregateHostExists:
            LOG.debug(("Host %(host)s was already added to aggregate %(agg)s "
                       "as part of a periodic task.") %
                      {'host': self.config['host_name'],
                       'agg': cfg.CONF.default_host_aggregate_name})

    def __update_aggregates_for_deregister(self):
        """Removes the host from any aggregates in which it resides"""
        agg_ids = agg_utils.get_aggregate_ids_for_host(
            self.context, self.config['host_name'])
        for agg_id in agg_ids:
            agg_utils.remove_host_from_aggregate(self.context,
                                                 self.config['host_name'],
                                                 agg_id)

    def __verify_allowed_to_edit(self, force):
        """Internal Helper method to Verify the Compute Host can be edited"""
        host = self.config['host_name']
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        # Only do the verification if they didn't ask to force the edit
        if not force and not cfg.CONF.force_deregister:
            instances = instfact.find_all_instances_by_host(self.context, host)
            # First make sure that none of the Instances have Tasks in-progress
            self.__verify_no_tasks_inprogress(
                instances, msg.EDITED_ACTION, msg.EDITING_ACTION)

    def __verify_no_tasks_inprogress(self, instances, action, actioning):
        """Internal Helper method to Verify that No Tasks are In-progress"""
        host_name = self.config['host_name']
        # If the Compute Service isn't running, then no tasks are really
        # running
        if self._host_service_status() != 'started':
            return
        # If we can't contact the Compute Service, assume it has hung
        if not self.compute_api.verify_host_running(self.context, host_name):
            return
        vms_inprogress = list()
        # Loop through each of the VM's seeing if any have tasks in-progress
        for instance in instances:
            task_state = instance.get('task_state')
            updated_at = instance.get('updated_at')
            if updated_at is None:
                updated_at = instance.get('created_at')
            # Check the task state of the VM to see if any tasks are running
            if task_state is None or updated_at is None:
                continue
            # If the VM wasn't updated within the last 1/2 hour we will assume
            # the task got stuck and skip it, otherwise since the task state
            # showed a task in-progress, don't allow the Compute Host removal
            if not timeutils.is_older_than(updated_at, 1800):
                vms_inprogress.append(unicode(instance.get('display_name')))
        # If any tasks were in-progress throw an exception blocking removal
        if len(vms_inprogress) > 0:
            b = self._get_message_bundle()
            action = b.get_message_text(self.context, action)
            actioning = b.get_message_text(self.context, actioning)
            rsctype = b.get_message_text(self.context, msg.RSC_TYPE_SERVER)
            rsctypes = b.get_message_text(self.context, msg.RSC_TYPE_SERVERS)
            # Include only the first 15 VM names in the list to avoid clutter
            vm_names = utils.format_string_list(vms_inprogress[:15])
            self._throw_bad_request(msg.TASKS_INPROGRESS, action=action,
                                    rsctypes=rsctypes, rscnames=vm_names,
                                    rsctype=rsctype, actioning=actioning)


########################################################
# ########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    throw_exc = kwargs.pop('throw_exc', True)
    try:
        # If we are not attempting to create the host, we need to get the type
        tempreg = HostRegistrar(context, **kwargs)
        host_type = tempreg.config.get('host_type')
        # If the host_type isn't set then they didn't supply the attribute
        if not host_type:
            tempreg._throw_bad_request(msg.MISSING_ATTRIBUTE, attr='host_type')
        # If the Host Type isn't in the Supported List, throw an exception
        if host_type not in cfg.CONF.supported_host_types:
            tempreg._throw_bad_request(msg.INVALID_HOST_TYPE,
                                       hosttype=host_type)
        try:
            # Attempt to dynamically load the corresponding module for the type
            regmodule = __package__ + '.' + str(host_type) + '_registrar'
            __import__(regmodule)
            return getattr(sys.modules[regmodule],
                           'get_registrar')(context, **kwargs)
        except Exception:
            tempreg._throw_bad_request(msg.INVALID_HOST_TYPE,
                                       hosttype=host_type)
    except:
        # Only throw the exception if we were asked to throw it on error
        if throw_exc:
            raise
    return None
