# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Register a new Compute Host"""

import webob
from nova.api.openstack import extensions
from nova.api.openstack import wsgi
from oslo_log import log as logging

from powervc_nova.objects.compute import dom as compute_dom
from powervc_nova.prs import constants as pvc_prs_const
from powervc_nova.prs import utils as prs_utils
from powervc_discovery.registration import registrar_utilities
from powervc_discovery.registration.compute import host_registrar
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
ALIAS = 'host-registration'
AUTHORIZE_V21 = extensions.os_compute_authorizer(ALIAS)
AUTHORIZE_V2 = extensions.extension_authorizer('compute', 'host_registration')


class HostRegistrationControllerExtension(wsgi.Controller):
    """Controller Extension Class to extend os-hosts API for Registration"""

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Extend Rest GET Method to Add Attributes for a specific Host"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'show')
        # Load the appropriate Registrar to Retrieve the Registration Info
        registrar = host_registrar.get_registrar(context, host_name=id,
                                                 throw_exc=True)
        # Only add the Registration Info if it was already registered
        if registrar:
            registration = self.filter_config(registrar)
            resp_obj.obj['host'].append({'registration': registration})

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Create a Host Registration"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'create')
        registration = body['host']['registration']
        # Load the appropriate Registrar sub-class to do the Host Registration
        registrar = host_registrar.get_registrar(context, create=True,
                                                 **registration)
        # Delegate to the Registrar to do the Host Registration
        registrar.register()
        # Return on the Response the Registration Info
        return {'host': {'registration': self.filter_config(registrar)}}

    @wsgi.action("update")
    def update(self, req, host_name_id, body):
        """Extend Rest PUT Method to Update a Host Registration"""
        host = {'host': host_name_id}
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'update')
        registration = body.pop('registration', None)
        # If no Registration Info is provided, then updating a base attribute
        if registration is not None:
            # Load the appropriate Registrar to do the Host Re-Registration
            registrar = host_registrar.get_registrar(context,
                                                     host_name=host_name_id)
            # If they are trying to switch types, then do a register instead
            if registrar.switching_registration_type(registration):
                new_registration = dict(registration, host_name=host_name_id)
                new_registration.setdefault('force_switch', True)
                new_registration['host_type'] = registrar.config['host_type']
                registrar = host_registrar.get_registrar(
                    context, create=True, **new_registration)
                registrar.register()
            # Delegate to the Registrar to do the Host Re-Registration
            else:
                registrar.update_registration(registration)
            host['registration'] = self.filter_config(registrar,
                                                      config=registration)
        dro_enabled = body.pop(pvc_prs_const.DRO_ENABLED_KEY, None)
        if dro_enabled is not None:
            try:
                dro_enabled = dro_enabled.lower()
            except Exception:
                msg = (_("Invalid value for dro_enabled (%s). It "
                         "must either be 'true' or 'false'.") % dro_enabled)
                raise webob.exc.HTTPBadRequest(explanation=msg)
            if dro_enabled != 'true' and dro_enabled != 'false':
                msg = (_("Invalid value for dro_enabled (%s). It must "
                         "either be 'true' or 'false'.") % dro_enabled)
                raise webob.exc.HTTPBadRequest(explanation=msg)
            node = prs_utils.get_compute_node_by_service_host(
                context, host_name_id)
            if not node:
                msg = (_("Compute host '%s' is not found.") %
                       host_name_id)
                raise webob.exc.HTTPNotFound(explanation=msg)
            prs_utils.update_dro_enabled_in_compute_node_metadata(
                context, node.id, dro_enabled)
            host[pvc_prs_const.DRO_ENABLED_KEY] = dro_enabled
            return host

        return host

    @wsgi.action("update_software")
    @wsgi.response(202)
    def update_software(self, req, host_name_id, body):
        """Extend Rest PUT Method to Update Software for Host Registration"""
        host = {'host': host_name_id}
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'update')
        # Load the appropriate Registrar to do the Host Software update
        registrar = host_registrar.\
            get_registrar(context, host_name=host_name_id)
        # Delegate to the Registrar to do the Host Software update
        versions = registrar.update_software(body)
        host.update(versions)
        return host

    @wsgi.action("delete")
    @wsgi.response(204)
    def delete(self, req, id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        self._delete(req, id, force=False, uninstall=False)

    @wsgi.action("force_delete")
    @wsgi.response(204)
    def force_delete(self, req, host_name_id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        self._delete(req, host_name_id, force=True, uninstall=False)

    @wsgi.action("force_uninstall")
    @wsgi.response(204)
    def force_uninstall(self, req, host_name_id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        self._delete(req, host_name_id, force=False, uninstall=True)

    @registrar_utilities.host_inprogress_decorator('remove')
    def _delete(self, req, host_name_id, force=False, uninstall=False):
        """Internal Helper Method to Remove a Host Registration"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'delete')
        try:
            # Load the appropriate Registrar to do the Host DeRegistration
            registrar = host_registrar.\
                get_registrar(context, host_name=host_name_id)
            # Delegate to the Registrar to do the Host DeRegistration
            registrar.deregister(force=force, uninstall=uninstall)
        # If the CONF file wasn't found, see if we can clean up the DB entry
        except webob.exc.HTTPNotFound:
            hostfact = compute_dom.ManagedHostFactory.get_factory()
            host = hostfact.find_host_by_name(context, host_name_id)
            # If the Service/ComputeNode existed, then try to clean that up
            if host is not None:
                host.delete()
                return
            # Otherwise we want to raise the Host Not Found exception
            raise

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def filter_config(registrar, config=None):
        """Internal Helper Method to Remove any Sensitive Attributes"""
        new_config = registrar.config.copy()
        if config is not None:
            new_config = config.copy()
        # Loop through each of the Sensitive Attributes, removing them
        for attr in registrar.get_sensitive_attributes():
            new_config.pop(attr, None)
        return new_config


class HostConnectControllerExtension(wsgi.Controller):
    """Main Controller Extension Class to provide the os-hosts/connect API"""

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Verify the Connection to the Host"""
        context = req.environ['nova.context']
        registration = body['host']['registration']
        mainext = HostRegistrationControllerExtension
        verify_only = registration.pop('verify_only', None)
        context = mainext._validate_authorization(context, req, 'connect')
        # Load the appropriate Registrar sub-class to do the Host Registration
        registrar = host_registrar.get_registrar(context, create=True,
                                                 **registration)
        # Delegate to the Registrar to connect to the Storage Host
        connection_info = registrar.establish_connection(registrar.config)
        connection_info.pop('host_proc_level_type', None)
        connection_info.pop('host_display_name', None)
        # If they only wanted to verify, then we wanted to return no content
        if str(verify_only).lower() == 'true':
            return webob.Response(status_int=204)
        return {'host': mainext.filter_config(registrar, connection_info)}


class HypervisorHostControllerExtension(wsgi.Controller):
    """Controller Extension Class to add additional data to os-hypervisors"""

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Extend Rest GET Method to Add Attributes for a Hypervisor"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'show')
        # Get the Hypervisor that the Base Extension retrieved
        hypervisor = resp_obj.obj.get('hypervisor', {})
        # Add the additional attributes from the Host Registrar
        self._add_host_attributes(context, hypervisor)

    @wsgi.extends
    def detail(self, req, resp_obj):
        """Extend Rest GET Method to Add Attributes for all Hypervisors"""
        context = req.environ['nova.context']
        context = self._validate_authorization(context, req, 'detail')
        # Loop through each of the Hypervisors the Base Extension retrieved
        for hypervisor in resp_obj.obj.get('hypervisors', []):
            # Add the additional attributes from the Host Registrar
            self._add_host_attributes(context, hypervisor)

    def _add_host_attributes(self, context, hypervisor):
        """Helper method to add additional Host Attributes to a Hypervisor"""
        service = hypervisor.get('service', {})
        host = service.get('host')
        # We only have anything to add if we could find the given Host
        if host:
            # Delegate to the Host Registrar to retrieve Host registration
            registrar = host_registrar.get_registrar(
                context, host_name=host,
                skip_access_state=True, throw_exc=False)
            # If the Host was registered, add the additional attributes
            if registrar:
                additional_attrs = ['host_display_name', 'hmc_uuids',
                                    'host_storage_type', 'access_ip',
                                    'host_type', 'ephemeral_disk_path',
                                    'host_network_server', 'manager_type']
                for attr in additional_attrs:
                    if registrar.config.get(attr) is not None:
                        service[attr] = registrar.config[attr]
                # We need to map the storage type to user display-able values
                if service.get('host_storage_type') is not None:
                    if service['host_storage_type'] == 'local':
                        service['host_storage_type'] = 'Local'
                    if service['host_storage_type'] == 'san':
                        service['host_storage_type'] = 'SAN'
                # Default the Storage Type to SAN if it isn't set
                else:
                    service['host_storage_type'] = 'SAN'

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        if 'v2.1' in req.url:
            AUTHORIZE_V21(context, action=action)
        else:
            AUTHORIZE_V2(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class HostRegistration(extensions.ExtensionDescriptor):
    """Capability to Register a new Compute Host."""
    name = "Compute Host Registration"
    alias = "ibm-host-registration"
    namespace = "http://docs.openstack.org/compute/ext/" + \
                "ibm_host_registration/api/v1.1"
    updated = "2013-03-21T21:00:00-06:00"

    def __init__(self, ext_mgr=None):
        """Register this extension with the extension manager"""
        super(HostRegistration, self).__init__(ext_mgr)
        # We need to cleanup any left over tagged hosts as in-progress
        registrar_utilities.remove_all_hosts_inprogress()

    def get_resources(self):
        """Provide a Sub-Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'update-registration',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'update': 'PUT'})
        resource2 = extensions.ResourceExtension(
            'force',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'force_delete': 'DELETE'})
        resource3 = extensions.ResourceExtension(
            'uninstall',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'force_uninstall': 'DELETE'})
        resource4 = extensions.ResourceExtension(
            'os-hosts/connect', HostConnectControllerExtension(),
            collection_actions={'create': 'POST'})
        resource5 = extensions.ResourceExtension(
            'update-software',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'update_software': 'PUT'})
        return [resource1, resource2, resource3, resource4, resource5]

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hosts Resource"""
        extension1 = extensions.ControllerExtension(
            self, 'os-hosts', HostRegistrationControllerExtension())
        extension2 = extensions.ControllerExtension(
            self, 'os-hypervisors', HypervisorHostControllerExtension())
        return [extension1, extension2]


class HostRegistrationV21(extensions.V21APIExtensionBase):
    """Capability to Register a new Compute Host."""
    name = "Compute Host Registration"
    alias = "ibm-host-registration"
    version = 1

    def __init__(self, extension_info=None):
        """Register this extension with the extension manager"""
        # super(HostRegistration, self).__init__(ext_mgr)
        # We need to cleanup any left over tagged hosts as in-progress
        registrar_utilities.remove_all_hosts_inprogress()

    def get_resources(self):
        """Provide a Sub-Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'update-registration',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'update': 'PUT'})
        resource2 = extensions.ResourceExtension(
            'force',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'force_delete': 'DELETE'})
        resource3 = extensions.ResourceExtension(
            'uninstall',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'force_uninstall': 'DELETE'})
        resource4 = extensions.ResourceExtension(
            'os-hosts/connect', HostConnectControllerExtension(),
            collection_actions={'create': 'POST'})
        resource5 = extensions.ResourceExtension(
            'update-software',
            controller=HostRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'update_software': 'PUT'})
        return [resource1, resource2, resource3, resource4, resource5]

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hosts Resource"""
        extension1 = extensions.ControllerExtension(
            self, 'os-hosts', HostRegistrationControllerExtension())
        extension2 = extensions.ControllerExtension(
            self, 'os-hypervisors', HypervisorHostControllerExtension())
        return [extension1, extension2]
