# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""KVM Registrar Classes for performing Host Registration"""

import sys
import ConfigParser

from oslo_config import cfg
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.compute import remote_host_registrar
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.StrOpt('kvm_platform_type', default='ppc',
               help='Specifies the Platform (x86 or ppc) for KVM')
])

LOG = logging.getLogger(__name__)


########################################################
# ##########  KVM Registrar Implementation  ############
########################################################
class KVMHostRegistrar(remote_host_registrar.RemoteHostRegistrar):
    """KVM Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the KVM Class for Host Registration"""
        super(KVMHostRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_KVM)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.INVALID_SYSTEM, ['/sbin/lsmod: ']),
            (msg.KVM_LIBVIRT_UNAVAILABLE, ['hypervisorunavailable: ']),
        ] + super(KVMHostRegistrar, self)._get_registration_error_mappings()

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        if self.skip_remote_commands is False:
            super(KVMHostRegistrar, self).\
                _retrieve_connection_info(connect_info, config)
            # Retrieve the File Systems from the KVM Host for the Ephemeral
            # Path
            connect_info['fileSystems'] = self._get_file_systems(config)
            # If they gave a Ephemeral Disk Path, verify that it is valid
            path = config.get('ephemeral_disk_path')
            if path and path != '/var/lib/libvirt/images':
                self._verify_ephemeral_disk_path(connect_info, config, path)

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(KVMHostRegistrar,
                        self)._get_service_param_mapping()
        mapping['kvm_host_ip'] = 'access_ip'
        mapping['kvm_host_user'] = 'user_id'
        mapping['kvm_host_passwd'] = 'password'
        mapping['kvm_host_private_key'] = 'private_key'
        mapping['kvm_host_ephemeral_path'] = 'ephemeral_disk_path'
        mapping['host_storage_type'] = 'host_storage_type'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(KVMHostRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['kvm_platform_type'] = cfg.CONF.kvm_platform_type
        parm_mapping['host_storage_type'] = 'local'
        if parm_mapping.get('kvm_host_ephemeral_path') is None:
            parm_mapping['kvm_host_ephemeral_path'] = '/var/lib/libvirt/images'
        # We need to set some extra parameters if they are using NFS
        self._set_nfs_config_values(parm_mapping, connect_info)

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        nfs_ip = config.get('kvm_host_network_server')
        nfs_share = config.get('kvm_host_network_share')
        # Call the Super Class to parse most of the configuration properties
        super(KVMHostRegistrar, self)._parse_service_config(config)
        # If both the IP and Share were set, then combine for a single property
        if nfs_ip and nfs_share:
            self.config['host_network_server'] = nfs_ip + ':' + nfs_share

    #############################################
    # ## Implementing HostRegistrar Interface ###
    #############################################
    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        # Run the lsmod command on the KVM Host to see if it is KVM or not
        output = utils.run_ssh_command(config, "/sbin/lsmod")[0]
        kvmmods = [line for line in output.splitlines() if 'kvm' in line]
        # If no modules returned had KVM in it, it isn't a valid system
        if len(kvmmods) <= 0:
            raise Exception('/sbin/lsmod: No KVM Modules loaded')
        # Now let the sub-class perform any additional validation they need
        self._validate_kvm_host(config)
        # Lastly, lets make sure the user has root privileges on the KVM Host
        self._validate_root_privileges(config)
        # For now we will just use the SSH Finger-print as the Unique
        # Identifier
        return utils.verify_host_key(config)['host_key']['fingerprint']

    ###############################################
    # ### Defining KVMHostRegistrar Interface  ####
    ###############################################
    def _get_host_image_file(self):
        """Provides the path on the Mgmt System with the image payload"""
        raise NotImplementedError()

    def _validate_kvm_host(self, config):
        """Allows the sub-class to further validate the KVM Host"""
        pass

    ###############################################
    # ###  Shared Registration Helper Methods  ####
    ###############################################
    def _get_file_systems(self, config):
        """Shared Helper method to retrieve File Systems from the KVM Host"""
        nfs_mounts, mount_dirs = (dict(), list())
        file_systems, inaccessible_ips = (list(), list())
        # Get the output of /etc/mtab
        output = utils.run_ssh_command(config, '/bin/cat /etc/mtab')[0]
        # Parse the mount directories from /etc/mtab if of type nfs or ext
        for line in output.splitlines():
            ln_splt = line.split()
            if ln_splt[2] in ['ext', 'ext2', 'ext3', 'ext4']:
                mount_dirs.append(ln_splt[1])
            # If the file system type is NFS, gather its information as well
            if ln_splt[2] in ['nfs', 'nfs2', 'nfs3', 'nfs4']:
                nfssvr = line.split(':')
                mount_dirs.append(ln_splt[1])
                if nfssvr[0] in nfs_mounts:
                    nfs_mounts[nfssvr[0]]['fs_name'].append(ln_splt[0])
                    nfs_mounts[nfssvr[0]]['fs_mount'].append(ln_splt[1])
                    nfs_mounts[nfssvr[0]]['fs_type'].append(ln_splt[2])
                else:
                    nfs_mounts[nfssvr[0]] = dict(fs_name=[ln_splt[0]])
                    nfs_mounts[nfssvr[0]]['fs_mount'] = [ln_splt[1]]
                    nfs_mounts[nfssvr[0]]['fs_type'] = [ln_splt[2]]
        # Check the NFS services. If they are not running, add the ip to the
        # inaccessible list and remove the mount directory from mount_dirs
        for ip in nfs_mounts:
            mounts = nfs_mounts[ip]['fs_mount']
            if self.__nfs_service_check(config, ip, mounts) is False:
                inaccessible_ips.append(ip)
                for mount in mounts:
                    mount_dirs.remove(mount)
        # Build the df command with the mount directories list
        cmd = '/bin/df -hT ' + ' '.join(mount_dirs) + ' 2> /dev/null'
        # Run the df command on the KVM to get the Mounted File Systems
        output = utils.run_ssh_command(config, cmd, True, True)[0]
        # Parse the File Systems from the output of the df command
        if output and len(output) > 0:
            for line in output.splitlines():
                columns = line.split()
                # We want to skip the Header line if it isn't a valid format
                if len(columns) >= 7 and columns[0] != 'Filesystem':
                    # We want to exclude some internal file system types
                    file_system = dict(fs_name=columns[0])
                    # Parse each of the columns from the File System output
                    file_system['fs_type'] = columns[1]
                    file_system['fs_mount'] = columns[6]
                    file_system['fs_free_capacity_gb'] = \
                        self.__parse_file_system_size(columns[4])
                    file_system['fs_total_capacity_gb'] = \
                        self.__parse_file_system_size(columns[2])
                    file_system['fs_accessible'] = True
                    file_systems.append(file_system)
        # Add the inaccessible NFS file systems to the list
        for nfs in inaccessible_ips:
            for index in range(len(nfs_mounts[nfs]['fs_name'])):
                file_system = dict(fs_name=nfs_mounts[nfs]['fs_name'][index])
                file_system['fs_type'] = nfs_mounts[nfs]['fs_type'][index]
                file_system['fs_mount'] = nfs_mounts[nfs]['fs_mount'][index]
                file_system['fs_free_capacity_gb'] = None
                file_system['fs_total_capacity_gb'] = None
                file_system['fs_accessible'] = False
                file_systems.append(file_system)
        return file_systems

    def _verify_ephemeral_disk_path(self, connect_info, config, disk_path):
        """Shared Helper method to verify the Ephemeral Disk Path is valid"""
        crtdir, crtfile = (False, False)
        dir_path, file_path = (disk_path, None)
        fsystems = connect_info.get('fileSystems', [])
        # Make sure that the Ephemeral Disk Path is an absolute path
        if not disk_path.strip().startswith('/'):
            self._throw_bad_request(msg.KVM_ABSOLUTE_PATH, path=disk_path)
        # Find the file system that is the best match for the given disk path
        fsys = self.__get_matching_file_system(fsystems, disk_path)
        # If the file system is marked not accessible then throw an exception
        if fsys is None or fsys['fs_accessible'] is False:
            self._throw_bad_request(msg.KVM_INVALID_EPHEMERAL, path=disk_path)
        try:
            # First run an ls command to see if the directory already exists
            lserr = utils.run_ssh_command(config, '/bin/ls ' + dir_path,
                                          True)[1]
            # If the directory doesn't exist, then trying creating it
            if lserr is not None and len(lserr.strip()) > 0:
                dir_path += utils.get_unique_file_postfix()
                utils.run_ssh_command(config, '/bin/mkdir -p ' + dir_path)
                crtdir = True
            # Now we can try creating a file in the directory to verify write
            file_path = dir_path + '/pvc' + utils.get_unique_file_postfix()
            utils.run_ssh_command(config, '/bin/touch ' + file_path)
            crtfile = True
        except Exception as exc:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error Validating Ephemeral "
                       "Disk Path: %s") % unicode(exc))
            LOG.exception(exc)
            self._throw_bad_request(msg.KVM_INVALID_EPHEMERAL, path=disk_path)
        finally:
            # Now we can do cleanup to delete the temporary files we created
            if file_path and crtfile is True:
                utils.run_ssh_command(self.config, '/bin/rm -f ' + file_path,
                                      True)
            if dir_path and crtdir is True:
                utils.run_ssh_command(self.config, '/bin/rm -rf ' + dir_path,
                                      True)

    def _set_nfs_config_values(self, parm_mapping, connect_info):
        """Helper method to update the Config File with the NFS properties"""
        fsystems = connect_info.get('fileSystems', [])
        disk_path = parm_mapping.get('kvm_host_ephemeral_path', '/')
        # Find the file system that is the best match for the given disk path
        fsys = self.__get_matching_file_system(fsystems, disk_path)
        # If an NFS-backed File System was selected, set its properties
        if fsys is not None and fsys['fs_type'].startswith('nfs'):
            fsname = fsys['fs_name'].split(':')
            parm_mapping['host_storage_type'] = 'nfs'
            parm_mapping['kvm_host_network_server'] = fsname[0]
            parm_mapping['kvm_host_network_share'] = fsname[1]

    ###############################################
    #    Internal Registration Helper Methods     #
    ###############################################
    def _add_extra_management_info(self, mgmt_file, parm_mapping):
        """Helper method add the extra attribute to the Management Info File"""
        super(KVMHostRegistrar,
              self)._add_extra_management_info(mgmt_file, parm_mapping)
        config = dict(host_storage_type=parm_mapping['host_storage_type'])
        # We want to override the Ephemeral Disk Path if they specified one
        config['instances_path'] = parm_mapping['kvm_host_ephemeral_path']
        config['instances_path'] += '/nova/instances'
        # Loop through each Attribute adding it to the Management Info file
        for key, val in config.iteritems():
            utils.set_openstack_config_value(mgmt_file, key, val)

    def __parse_file_system_size(self, sizestr):
        """Helper method to parse the float size from string provided"""
        size = None
        if sizestr.endswith('K'):
            size = float(sizestr[:len(sizestr) - 1]) / (1024 ** 2)
        elif sizestr.endswith('M'):
            size = float(sizestr[:len(sizestr) - 1]) / (1024 ** 1)
        elif sizestr.endswith('G'):
            size = float(sizestr[:len(sizestr) - 1]) / (1024 ** 0)
        elif sizestr.endswith('T'):
            size = float(sizestr[:len(sizestr) - 1]) / (1024 ** -1)
        else:
            size = float(sizestr) / (1024 ** 3)
        return "%.2f" % size if size is not None else None

    def __get_matching_file_system(self, file_systems, disk_path):
        """Helper method to find the file system closes matching the path"""
        fsys = None
        # Loop through each fo the file systems, seeing which matches
        for fs in file_systems:
            # If the fs mount is a parent of the path, it is a possible match
            if utils.is_sub_path(disk_path, fs['fs_mount']):
                # If this is a better match than the previous, use this one
                if not fsys or len(fs['fs_mount']) > len(fsys['fs_mount']):
                    fsys = fs
        return fsys

    def __nfs_service_check(self, config, nfs_ip, mounts):
        """Helper method to check if NFS servces are running on the server"""
        rpc_info, serv_up = (list(), dict())
        # Get a table of RPC info
        cmd = '/usr/sbin/rpcinfo -s ' + nfs_ip
        output = utils.run_ssh_command(config, cmd, True)[0]
        for line in output.splitlines():
            rpc_info.append(line.split())
        # If the service is running, add it to the service up list
        for i in range(1, len(rpc_info)):
            if rpc_info[i][3] == 'nfs':
                serv_up['nfs'] = True
            elif rpc_info[i][3] == 'nfs_acl':
                serv_up['nfs_acl'] = True
            elif rpc_info[i][3] == 'nlockmgr':
                serv_up['nlockmgr'] = True
            elif rpc_info[i][3] == 'mountd':
                serv_up['mountd'] = True
        # Check to see if all NFS services are running
        if len(serv_up) >= 4:
            output = utils.run_ssh_command(
                config, '/bin/ls %s' % mounts[0], True, False, 30)
            # Also make sure that we can access the mount for the share
            return output[1] != 'eventlet.timeout.Timeout'
        return False


########################################################
#           KVM Periodic Task Implementation           #
########################################################
def periodic_task(context):
    """Periodic Task invocation point for updating the Mgmt IP on Hosts"""
    pending_registrars = list()
    # We need to cleanup any left over tagged hosts as in-progress
    utils.remove_all_hosts_inprogress()
    # Loop through each of the Hosts, updating its IP Address
    matching_registrars = _get_kvm_registrars(context)
    for registrar in matching_registrars:
        # If the registrar failed to update, then add it for next interval
        if registrar.update_ip() is False:
            pending_registrars.append(registrar)
    return len(pending_registrars) > 0


def _get_kvm_registrars(context):
    """Helper method to get the list of KVM Hosts needing updating"""
    kvm_registrars = list()
    # Loop through each host, seeing if it the kvm host needs ip's updated
    for filename in utils.get_matching_files('/etc/nova', 'nova-', '.conf'):
        host_type, changed_systems = (None, None)
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/nova/' + filename)
        # Check to see if the Host Type is set and set to KVM
        if parser.has_option('DEFAULT', 'host_type'):
            host_type = parser.get('DEFAULT', 'host_type')
        # Check if the Management System was retored from a different system
        if parser.has_option('DEFAULT', 'restored_new_system'):
            changed_systems = parser.get('DEFAULT', 'restored_new_system')
        # If it is KVM and restored from a different system, then return i
        if host_type == 'kvm' and changed_systems == 'True':
            hostname = filename[5:len(filename) - 5]
            kvm_registrars.append(get_registrar(context, host_name=hostname))
    return kvm_registrars


########################################################
# ########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    # Attempt to dynamically load the corresponding module for the host type
    regmodule = (__package__ + '.' +
                 cfg.CONF.kvm_platform_type + '_kvm_registrar')
    __import__(regmodule)
    return getattr(sys.modules[regmodule], 'get_registrar')(context, **kwargs)
