# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import six
import sys
import json
import codecs
import locale
import requests
import ConfigParser

from powervc_keystone.encrypthandler import EncryptHandler
from powervc_discovery.locale import _
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# set encoding for 'print'
_stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)
_stderr = codecs.getwriter(locale.getpreferredencoding())(sys.stderr)
_sysexit = sys.exit

NOVA_URL = 'http://localhost:8774/v2/'
KEYSTONE_URL = 'https://localhost:5000/v3/'


def update_host_software():
    """Updates all of the KVM Hosts to to the latest version available"""
    # Construct the HTTP Client to the Nova API Server
    identity = _get_identity_info()
    # Query the Hypervisors to determine the Hosts needing updating
    hosts = _get_updateable_hosts(identity)
    # Update any of the Hosts that need to be updated
    for host in hosts:
        _update_host_software(identity, host)


def _get_updateable_hosts(identity):
    """Helper Method to retrieve a list of hosts needing to be updated"""
    hosts = list()
    try:
        response = _nova_rest_request(
            identity, 'os-hypervisors/detail', 'GET')
    # If there was an exception query the hypervisors, log it and exit
    except Exception as exc:
        _log_error(_("Error querying hypervisors, failed to update the "
                     "PowerKVM hosts: %(exc)s") % dict(exc=_('%s') % exc))
        _sysexit(2)
    # If there was an error status code on the query, log it and exit
    if response['status_code'] != 200:
        info = dict(code=response['status_code'], reason=response['reason'])
        _log_error(_("Error querying hypervisors, failed to update the "
                     "PowerKVM hosts: (%(code)s) %(reason)s") % info)
        _sysexit(3)
    # Loop through each of the Hypervisors, seeing which need to be updated
    for hypervisor in response['body'].get('hypervisors', []):
        # We will only update the given Host if the Host Type is KVM
        if hypervisor['service'].get('host_type') == 'kvm':
            host = hypervisor['service'].get('host')
            cur_version = hypervisor.get('current_mgmt_version')
            new_version = hypervisor.get('latest_mgmt_version')
            # See if there is a later version than the current version
            if _compare_version(new_version, cur_version) > 0:
                hosts.append(dict(host=host, curvers=cur_version,
                                  newvers=new_version))
    # If there weren't any hosts that need to be updated, log it here
    if len(hosts) <= 0:
        _log_info(_("All PowerKVM hosts are at the latest "
                    "version.  No updates are needed."))
    return hosts


def _update_host_software(identity, host):
    """Helper Method to update the PowerVC software on a single host"""
    url = 'os-hosts/' + host['host'] + '/update-software'
    try:
        # Log a message saying that we are starting to update the KVM Host
        _log_info(_("Updating PowerKVM host %(host)s from "
                    "%(curvers)s to %(newvers)s...") % host)
        # Do the HTTP PUT to update the PowerVC Software on the KVM Host
        response = _nova_rest_request(
            identity, url, 'PUT', dict(mgmt_version='*latest'))
        # If there was an error status code on the query, log it and exit
        if response['status_code'] != 202:
            info = dict(host=host['host'], code=response['status_code'],
                        reason=response['reason'])
            _log_error(_("Error updating PowerKVM host %(host)s: "
                         "(%(code)s) %(reason)s") % info)
            return
        # Log a message saying that we have finished updating the KVM Host
        _log_info(_("Updated PowerKVM host %(host)s to %(newvers)s.") % host)
    # Log the error if one failed, but continue updating the next host
    except Exception as exc:
        info = dict(host=host['host'], exc=_('%s') % exc)
        _log_error(_("Error updating PowerKVM host %(host)s, "
                     "continuing on: %(exc)s") % info)


def _compare_version(new_version, old_version):
    """Helper method to compare 2 versions and see if the first is newer"""
    # If the old version isn't specified, the new version is greater if set
    if not old_version:
        return 1 if new_version else 0
    # If no new version is provided, the old is newer since it must be set
    if not new_version:
        return -1
    # We need to break apart the version to single digits for comparison
    old_split = old_version.split('.')
    new_split = new_version.split('.')
    max_digits = max(len(new_split), len(old_split))
    # Loop through each of the digits and do the comparison
    for index in range(max_digits):
        old_num = int(old_split[index]) if len(old_split) > index else 0
        new_num = int(new_split[index]) if len(new_split) > index else 0
        # If a digit is different, then return since 1 since greater
        if old_num != new_num:
            return 1 if new_num > old_num else -1
    # If all digits were the same, the 2 versions must be equal
    return 0


def _log_error(msg):
    """Helper method to log an error message to standard error"""
    _stderr.write(six.text_type(msg) + '\n')


def _log_info(msg):
    """Helper method to log an error message to standard error"""
    _stdout.write(six.text_type(msg) + '\n')


def _get_identity_info():
    """Helper method to construct an HTTP Client to the Nova API process"""
    try:
        pwd = _get_nova_password()
        token = _get_authentication_token('nova', pwd)
        tenant = _get_tenant_identifier(token)
        return (token, tenant)
    # If we got an exception authenticating, log it and exit
    except Exception as exc:
        info = dict(exc=_('%s') % exc)
        _log_error(_("Error connecting to the Keystone Server, failed "
                     "to update the PowerKVM hosts: %(exc)s") % info)
        _sysexit(1)


def _get_nova_password():
    """Helper method to parse the Nova Password fromt the CONF file"""
    parser = ConfigParser.RawConfigParser()
    parser.read('/etc/nova/nova.conf')
    admin_pwd = parser.get('keystone_authtoken', 'password')
    # DeCrypt the Password String, so we can retrieve the token
    return EncryptHandler().decode(admin_pwd)


def _nova_rest_request(identity, url_postfix, method, body=None):
    """Helper method to issue a REST Request for the given Nova API"""
    body = json.dumps(body) if body else body
    output, url = (None, NOVA_URL + identity[1] + '/' + url_postfix)
    header = {'Content-Type': 'application/json', 'X-Auth-Token': identity[0]}
    # Use the Requests Library to Retrieve the List of Tenants available
    resp = requests.request(method, url, headers=header, data=body)
    # If the response body is a JSON string, then parse it
    if resp.text and resp.text[0] == '{':
        output = json.loads(resp.text)
    return dict(status_code=resp.status_code, reason=resp.reason, body=output)


def _get_authentication_token(user, pwd):
    """Helper method to get the Authentication Token from Keystone"""
    domain, header = ({'id': 'default'}, {'Content-Type': 'application/json'})
    pwddict = dict(user=dict(domain=domain, name=user, password=pwd))
    # Construct the input Identity info into the Request Body
    identity = dict(methods=['password'], password=pwddict)
    scope = dict(project=dict(domain=dict(id='default'), name='service'))
    data = json.dumps(dict(auth=dict(identity=identity, scope=scope)))
    # Use Requests Library to execute a POST to get the Token
    resp = requests.request('POST', KEYSTONE_URL + 'auth/tokens',
                            headers=header, data=data, verify=False)
    # If we didn't get an OK return code, we can't continue without a token
    if resp.status_code != 201:
        respdict = dict(status=resp.status_code, text=resp.text)
        raise Exception(_("Error Code %(status)s retrieving "
                          "Authentication Token: %(text)s") % respdict)
    # Parse the Token out of the Response Header to use on this call
    return resp.headers['X-Subject-Token']


def _get_tenant_identifier(token):
    """Helper method to Retrieve the Tenant Identifier to use"""
    header = {'Content-Type': 'application/json', 'X-Auth-Token': token}
    # Use the Requests Library to Retrieve the List of Tenants available
    resp = requests.request(
        'GET', KEYSTONE_URL + 'projects', headers=header, verify=False)
    # If we didn't get an OK return code, we can't continue without a tenant
    if resp.status_code != 200:
        respdict = dict(status=resp.status_code, text=resp.text)
        raise Exception(_("Error Code %(status)s retrieving "
                          "Tenant Identifier: %(text)s") % respdict)
    # Parse out the Identifier from the List for the IBM Default Tenant Name
    for project in json.loads(resp.text)['projects']:
        if project['name'] == 'service':
            return project['id']
    # If didn't find our matching Tenant, then give up and throw an exception
    raise Exception(_("Unable to find a matching Tenant for service"))
