# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PowerKVM Registrar Classes for performing Host Registration"""

from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.compute import kvm_registrar

LOG = logging.getLogger(__name__)


########################################################
# ########  PowerKVM Registrar Implementation  #########
########################################################
class PowerKVMHostRegistrar(kvm_registrar.KVMHostRegistrar):
    """PowerKVM Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the PowerKVM Class for Host Registration"""
        super(PowerKVMHostRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_POWERKVM)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return super(PowerKVMHostRegistrar, self).\
            _get_registration_error_mappings() + [
                (msg.INVALID_SYSTEM, ['the /etc/issue file'])]

    ###############################################
    # ## Implementing KVMHostRegistrar Interface ##
    ###############################################
    def _get_host_image_file(self):
        """Provides the path on the Mgmt System with the image payload"""
        output = utils.run_ssh_command(
            self.config, '/bin/cat /etc/issue', True)[0].strip()
        # Check the PowerKVM version and if it is 3.1 tack on 31 to the image
        vers = '31' if output.find('3.1') > 0 else ''
        return '/opt/ibm/powervc/images/kvm/powervc-powerkvm%s-compute-' % vers

    def _validate_kvm_host(self, config):
        """Allows the sub-class to further validate the KVM Host"""
        # Check the /etc/issue file to see if this is PowerKVM or not
        output = utils.run_ssh_command(config, "/bin/cat /etc/issue",
                                       suppress=True)
        output = output[0].strip().lower() if output and output[0] else ''
        if output.find('ibm powerkvm') < 0 and output.find('ibm_powerkvm') < 0:
            raise Exception("The /etc/issue file doesn't contain IBM PowerKVM")


########################################################
# ########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return PowerKVMHostRegistrar(context, **kwargs)
