# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""PVM-managed PowerVM Registrar Classes for performing Host Registration"""

from oslo_log import log as logging
from nova import rpc as rpc_api
from nova.objects import block_device

from powervc_nova.objects.compute import dom as compute_dom
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.compute import host_registrar
from powervc_discovery.registration.compute import remote_host_registrar

LOG = logging.getLogger(__name__)
NOVALINK_MIN_VERSION = '1.0.0.2'


########################################################
# ##########  PVM Registrar Implementation  ############
########################################################
class PVMPowerVMHostRegistrar(remote_host_registrar.RemoteHostRegistrar):
    """PVM-managed PowerVM Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the NovaLink Class for Host Registration"""
        super(PVMPowerVMHostRegistrar, self).__init__(context, **kwargs)
        self._powervm_switch_type = None

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def establish_connection(self, config, throw_exception=True):
        """Retrieval Method to get the Compute Host's Connection Info"""
        # Call the parent class to establish the actual connection
        connect_info = super(PVMPowerVMHostRegistrar, self).\
            establish_connection(config, throw_exception)
        # Need to update the host_name now that we connected to the system
        if self.config.get('host_name') == 'unknown':
            self.config['host_name'] = connect_info['unique_id']
        return connect_info

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_NOVALINK)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        parent_map = super(PVMPowerVMHostRegistrar,
                           self)._get_registration_error_mappings()
        return [
            (msg.INVALID_SYSTEM, ['pvmctl: ']),
            (msg.PVM_NEED_PVMADMIN,
             ['permission denied: \'/opt/pmc',
              'unable to read file /opt/pvm',
              'unable to read file /opt/pmc'])
        ] + parent_map

    def _generate_name_value(self):
        """Generates the Default Value to use for the Name Attribute"""
        try:
            return self._get_unique_identifier(self.config)
        # If couldn't connect to get the MTMS, will throw Exception later
        except Exception:
            return 'unknown'

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(PVMPowerVMHostRegistrar,
                        self)._get_service_param_mapping()
        mapping['powervm_host_ip'] = 'access_ip'
        mapping['powervm_host_user'] = 'user_id'
        mapping['powervm_host_passwd'] = 'password'
        mapping['powervm_host_private_key'] = 'private_key'
        mapping['powervm_mgr_type'] = 'manager_type'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(PVMPowerVMHostRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['powervm_mgr_type'] = 'pvm'

    def _registration_type_switch_hook(self, action,
                                       connect_info=None, config={}):
        """Hook to be able to switch registration types if needed"""
        hostname = self.config['host_name']
        config, dflt = config or self.config, action != 'register'
        # Determine whether PHYP thinks it is HMC managed
        hyp_hmc_managed = self.__is_hypervisor_hmc_managed(config)
        # Determine if the Host is already HMC managed to switch
        registrar = host_registrar.get_registrar(
            self.context, host_name=hostname, throw_exc=False)
        hmc_managed = registrar is not None and \
            registrar.config.get('manager_type') == 'hmc'
        # If HMC-managed, make sure now SSP/SPP VM's since PVM doesn't support
        if hmc_managed:
            self.__verify_vms_supported(hostname)
        # If they didn't say to force the switch, may need to error out
        if not config.get('force_switch', dflt):
            # If PHYP has it managed by HMC, then need to throw an exception
            if hyp_hmc_managed:
                self._throw_bad_request(
                    msg.HOST_PVM_NOT_MASTER, hostname=hostname)
            # If it is PowerVC HMC-managed, throw an exception
            if hmc_managed:
                self._throw_bad_request(
                    msg.HOST_HMC_FORCE_SWITCH, hostname=hostname)
        # If we were told to force the switch, change PHYP to PVM-managed
        if hyp_hmc_managed:
            self.__switch_hypervisor_to_pvm_managed(config)
        # If we are doing an install, verify that the package is install-able
        if hmc_managed or action == 'register':
            self._verify_package_installable(config)
        # Now we can un-configure the old host (and leave data)
        if hmc_managed:
            self.__synchronize_uuids()
            registrar.deregister(soft=True)
            self._powervm_switch_type = 'hmc'
            return True
        return False

    #############################################
    #   Implementing HostRegistrar Interface    #
    #############################################
    def switching_registration_type(self, config):
        """Specifies whether the Values will cause the type to switch"""
        is_pvm = 'access_ip' in config or 'user_id' in config
        return not is_pvm and 'hmc_uuids' in config

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        # Run the pvmctl command on the Host to see if it is NovaLink or not
        cmd = '/usr/bin/pvmctl ManagedSystem list -d ' \
              'MachineTypeModelAndSerialNumber.mtms_str'
        output = utils.run_ssh_command(config, cmd)[0].strip()
        # The format has changed, so split on either : or =
        splitchar = ':' if output.find(':') > 0 else '='
        mtms = output.split(splitchar)[1].strip()
        # Next, lets make sure the user has root privileges on the Host
        self._validate_root_privileges(config)
        # Parse the MTMS (8247-22L*2125D0A) to use of the Unique Identifier
        return mtms[0:4] + mtms[5:8] + '_' + mtms[9:]

    def _verify_package_installable(self, config):
        """Allows sub-class a hook to verify the package is installable"""
        # We want to verify that NovaLink is at at least the minimum version
        cmd = '/usr/bin/dpkg-query -W pvm-novalink'
        output = utils.run_ssh_command(config, cmd, True)[0].strip().split()
        # First we will check the package installed current on the system
        if ((output[0] != 'pvm-novalink' or
             utils.compare_version(output[1], NOVALINK_MIN_VERSION) >= 0)):
            return
        # Next we will refresh the apt repository to make sure it is current
        cmd = '/usr/bin/apt-get update'
        utils.run_ssh_command(config, cmd, True, timeout=30)
        # We also want to make sure it won't update from the apt repository
        cmd = '/usr/bin/apt-cache show pvm-novalink'
        output = utils.run_ssh_command(config, cmd, True)[0].strip()
        # Loop through each line seeing if any of the version is acceptable
        for line in output.split('\n'):
            parts = [p.strip() for p in line.split(':')]
            if ((parts[0] == 'Version' and
                 utils.compare_version(parts[1], NOVALINK_MIN_VERSION) >= 0)):
                return
        # If we couldn't get an acceptable version, then throw an exception
        self._throw_bad_request(
            msg.HOST_PVM_MIN_VERISON, version=NOVALINK_MIN_VERSION)

    ###############################################
    # ## Defining RemoteHostRegistrar Interface ###
    ###############################################
    def _get_host_image_file(self):
        """Provides the path on the Mgmt System with the image payload"""
        return '/opt/ibm/powervc/images/powervm/powervc-powervm-compute-'

    def _add_extra_management_info(self, mgmt_file, parm_mapping):
        """Helper method add the extra attribute to the Management Info File"""
        super(PVMPowerVMHostRegistrar, self).\
            _add_extra_management_info(mgmt_file, parm_mapping)
        # If switching from HMC, flag so the driver can do data conversion
        if self._powervm_switch_type is not None:
            utils.set_openstack_config_value(
                mgmt_file, 'powervm_switch_type', self._powervm_switch_type)

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def __is_hypervisor_hmc_managed(self, config):
        """Helper method to determine if PHYP is HMC-managed"""
        try:
            # Use the pvmctl command to check if PVM is the master or not
            cmd = '/usr/bin/pvmctl ManagedSystem list -d ' \
                  'ManagedSystem.session_is_master'
            output = utils.run_ssh_command(config, cmd)[0].strip()
            # The format has changed, so split on either : or =
            splitchar = ':' if output.find(':') > 0 else '='
            return output.split(splitchar)[1].strip().lower() == 'false'
        # This isn't supported on the Proxy, so ignore the exceptions for now
        except:
            return False

    def __switch_hypervisor_to_pvm_managed(self, config):
        """Helper method to update PHYP to make the host PVM-managed"""
        try:
            # Use the pvmctl command to update make PVM as the master
            cmd = '/usr/sbin/pvmutil -o setmastermode -i force'
            utils.run_ssh_command(config, cmd)
        except Exception as exc:
            LOG.warn(_("Error switching to PVM-managed: %s") % exc)
            LOG.exception(exc)
            # If there is an error, throw a more reasonable error
            self._throw_server_error(
                msg.HOST_PVM_SWITCH_ERROR, hostname=self.config['host_name'])

    def __synchronize_uuids(self):
        """Helper method to synchronize the VM UUID's with the Hypervisor"""
        hostname = self.config['host_name']
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        instances = instfact.find_all_instances_by_host(self.context, hostname)
        # Only need to synchronize anything if there are VM's for the Host
        if len(instances) > 0:
            utils.run_command('powervc-manage -o syncuuids -s %s' % hostname)
            # Notify the UI that it needs to refresh the Instance Cache
            anotifier = rpc_api.get_notifier('compute', hostname)
            anotifier.info(
                self.context, 'compute.instance.refresh', dict(host=hostname))

    def __verify_vms_supported(self, hostname):
        """Helper method verify the VM's aren't SSP/SPP and are supported"""
        instfact = compute_dom.ManagedInstanceFactory.get_factory()
        instances = instfact.find_all_instances_by_host(self.context, hostname)
        # Loop through each of the VM's, seeing if there is an SSP/SPP info
        for inst in instances:
            bdms = block_device.BlockDeviceMappingList.\
                get_by_instance_uuid(self.context, inst.uuid)
            pool_id = inst.power_specs.get('shared_proc_pool_id')
            # Loop through the BDM's checking for SSP-attached volumes
            for bdm in bdms:
                if bdm.connection_info and 'ibm_ssp' in bdm.connection_info:
                    self._throw_bad_request(msg.HOST_PVM_NO_SSP)
            # Check to see if the VM is in a different shared processor pool
            if pool_id is not None and pool_id != 0:
                self._throw_bad_request(msg.HOST_PVM_NO_SPP)


########################################################
# ########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return PVMPowerVMHostRegistrar(context, **kwargs)
