# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015, 2016 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Registrar Classes for performing Remote Host Registration"""

import ConfigParser
from datetime import datetime
from eventlet import greenthread

from oslo_config import cfg
from nova import rpc as rpc_api
from nova.objects import service as service_obj
from oslo_log import log as logging

from powervc_nova.objects.compute import dom as compute_dom
from powervc_discovery.config import data_mgr
from powervc_discovery.registration import registrar as reg
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.compute import host_registrar
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)


########################################################
# #######  Remote Host Registrar Implementation  #######
########################################################
class RemoteHostRegistrar(reg.RemoteServiceRegistrar,
                          host_registrar.HostRegistrar):
    """Remote Registrar Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the Remote Class for Host Registration"""
        kwargs['supports_sudo'] = True
        super(RemoteHostRegistrar, self).__init__(context, **kwargs)
        self.skip_remote_commands = False
        self.added_as_inprogress = False
        self.wait_for_service = True

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attrs = super(RemoteHostRegistrar, self).get_sensitive_attributes()
        attrs.extend(['mgmt_ip', 'supports_sudo'])
        return attrs

    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['host_type', 'access_ip', 'user_id',
                'private_key|private_key_data|password']

    def establish_connection(self, config, throw_exception=True):
        """Retrieval Method to get the Compute Host's Connection Info"""
        config['supports_sudo'] = True
        # Call the parent class to establish the actual connection
        return super(RemoteHostRegistrar, self).\
            establish_connection(config, throw_exception)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.HOST_NEED_SUDOER, ['not in the sudoers file']),
            (msg.HOST_SUDOER_NOPASS, ['sudo: a password is required']),
            (msg.PRIVKEY_ERROR, ['unable to compute factors p and q']),
            (msg.HOST_CANNOT_CONNECT_MGMT, ['unable to connect to amqp serv']),
            (msg.HOST_CANNOT_CONNECT_MGMT, ['ailed to connect to the managem'])
        ] + utils.ssh_reg_messages()

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def register(self):
        """Registrar Method to perform the System Registration"""
        # Before we do the registration, cleanup any leftover shadow CONF files
        self._cleanup_orphaned_conf_files()
        try:
            # Now we can call the parent class to do the registration
            super(RemoteHostRegistrar, self).register()
        finally:
            # Cleanup the Host from the In-progress list if it was added
            if self.added_as_inprogress and self.config.get('host_name'):
                utils.remove_host_inprogress(self.config['host_name'])

    def deregister(self, force=False, uninstall=False, soft=False):
        """Registrar Method to delete a System Registration"""
        uninstall = cfg.CONF.force_deregister_uninstall or uninstall is True
        private_key = self.config.get('private_key')
        # We need to copy the Private Key so it is there for the uninstall
        if uninstall is True and private_key is not None:
            command = '/bin/cp -p -f %s %s.tmp' % (private_key, private_key)
            private_key = private_key + '.tmp'
            utils.run_command(command, suppress=True, run_as_root=False)
        try:
            super(RemoteHostRegistrar, self).deregister(force, uninstall, soft)
            # If they asked to UnInstall as Part of Removal, do that now
            if uninstall is True:
                config = dict(self.config, private_key=private_key)
                self._uninstall_openstack_rpms(config)
        finally:
            # Clean up the temporary Private Key that we just created
            if uninstall is True and private_key is not None:
                utils.run_command('/bin/rm -f ' + private_key, suppress=True,
                                  run_as_root=False)

    def _get_service_name(self):
        """Provides the name of the Linux Service created by this Registrar"""
        return 'openstack-nova-compute'

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(RemoteHostRegistrar,
                        self)._get_service_param_mapping()
        mapping['mgmt_ip'] = 'mgmt_ip'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(RemoteHostRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['mgmt_ip'] = utils.get_local_system_ip(self.context)

    def _pre_service_registration_hook(self, action, connect_info=None,
                                       config={}, force=False):
        """Hook to be able to take actions before doing the registration"""
        super(RemoteHostRegistrar, self).\
            _pre_service_registration_hook(action, connect_info, config, force)
        # Make sure the OpenStack/PowerVC RPM's are installed on the Host
        if action == 'register':
            config = self.config.copy()
            mgmt_ip = self._get_management_ip(config)
            lcl_version = self._get_installed_version(config, remote=False)
            rmt_version = self._get_installed_version(config, remote=True)
            # We need to fail Registration if non-PowerVC Mgmt System managing
            if rmt_version is None and mgmt_ip:
                self._throw_bad_request(msg.HOST_OTHER_MANAGED, ip=mgmt_ip)
            # Add this Host to the In-Progress Hosts, to avoid timing issues
            if utils.add_host_inprogress(self.config['host_name']):
                disp = self._get_display_name(self.config)
                # We don't want to let them try to add the same host twice
                self._throw_bad_request(
                    msg.HOST_ADD_INPROGRESS, hostname=disp)
            self.added_as_inprogress = True
            # See if this version is installed or if need to install/update it
            if utils.compare_version(lcl_version, rmt_version) > 0:
                self._deploy_openstack_image(config, rmt_version)

    def _gather_host_service_log(self):
        """Run the command to gather the Host's Service Log for errors"""
        hostlog = '/var/log/nova/nova-compute.log'
        command = '/usr/bin/tail -n 75 ' + hostlog
        output = utils.run_ssh_command(
            self.config, command, suppress=True, requires_root=True)
        return (hostlog, output[0])

    def _start_linux_service(self, restart=False):
        """Provides the implementation for starting/restarting the Service"""
        # Since the Manage/UnManage/Update restart the service, NoOp this one
        pass

    def _verify_service_started(self, updated_at=None):
        """Internal Helper Method to Block until the Service has Started"""
        # If we didn't change anything on the Host, don't want to restart
        if updated_at is None or self.wait_for_service is True:
            super(RemoteHostRegistrar,
                  self)._verify_service_started(updated_at)

    ###############################################
    # ## Implementing RmtSvcRegistrar Interface ###
    ###############################################
    def _manage_remote_service(self, parm_mapping):
        """Performs any necessary updates to the remote service to manage"""
        disp = self._get_display_name(self.config)
        mgmt_ip = self._get_management_ip(self.config)
        # If it was already managed the auth_host should have been set
        if mgmt_ip:
            # If they didn't say to force an un-manage throw an exception
            if self.config.get('force_unmanage', False) is not True:
                self._throw_bad_request(msg.HOST_ALREADY_MANAGED, ip=mgmt_ip)
            # Otherwise we will force it to be un-managed and notify the other
            self._unmanage_remote_service(remanage=True)
        # Now that we know it isn't managed currently, we can manage it
        tmp_dir = '/tmp/powervc_mgmtinfo' + utils.get_unique_file_postfix()
        dump_file = tmp_dir + '/mgmtinfo.ini'
        try:
            # First we need to generate the certificates to use on the Host
            self._generate_certificate_files(self.config, tmp_dir)
            # Make sure that nova owns the file so we can transfer it
            utils.run_command('/bin/touch ' + dump_file, run_as_root=False)
            utils.run_command('/bin/chmod 600 ' + dump_file, run_as_root=False)
            # First we need to dump out the IP/Creds for the Management System
            utils.run_command('powervc-manage -o dump -f %s' % dump_file)
            # Next we need to add the extra attributes to the Mgmt Info file
            self._add_extra_management_info(dump_file, parm_mapping)
            # Next we need to transfer the Management Info file to the Host
            utils.sftp_transfer_file(self.config, dump_file, dump_file)
            # Before we can manage the Host, the Certificates need to exist
            self._transfer_certificate_files(self.config, tmp_dir)
            # Next we need to run the PowerVC Manage command on the Host
            cmd = '/opt/ibm/powervc/bin/powervc-manage -o manage -f '
            utils.run_ssh_command(self.config, cmd + dump_file,
                                  english=True, requires_root=True)
        except Exception as exc:
            # If the time is off on the Host, throw a specific exception
            if unicode(exc).find("The hostname is set to localhost") >= 0:
                self._throw_bad_request(msg.HOST_INVALID_HOSTNAME)
            # If the time is off on the Host, throw a specific exception
            if unicode(exc).find("Time difference from Management Sys") >= 0:
                self._throw_bad_request(msg.HOST_INVALID_TIME_SKEW)
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error configuring OpenStack: %s") % unicode(exc))
            LOG.exception(exc)
            self._throw_server_error(msg.HOST_CONFIGURE_FAILED, hostname=disp)
        finally:
            # Now we can do cleanup to delete the temporary files we created
            utils.run_ssh_command(self.config, '/bin/rm -rf ' + tmp_dir, True)
            utils.run_command('/bin/rm -rf ' + tmp_dir, suppress=True,
                              run_as_root=False)

    def _update_remote_service(self, parm_mapping, updated_values):
        """Performs any necessary updates to the remote service on edit"""
        self.wait_for_service = False
        disp = self._get_display_name(self.config)
        try:
            # We need to shadow the Host's Display Name if that changed
            if ((utils.has_value_changed(
                 self.config, updated_values, 'host_display_name'))):
                val = updated_values['host_display_name']
                val = utils.escape_config_value(val)
                # Run a Command to update the Display Name on the Remote Host
                cmd = '/usr/bin/openstack-config --set /etc/nova/nova.conf '
                cmd += 'DEFAULT host_display_name "' + unicode(val) + '"'
                utils.run_ssh_command(self.config, cmd, requires_root=True)
                self.wait_for_service = True
            # Run a Command to restart the Nova Services on the Remote Host
            if self.wait_for_service is True:
                cmd = '/opt/ibm/powervc/bin/powervc-services nova restart'
                utils.run_ssh_command(self.config, cmd, requires_root=True)
        except Exception as exc:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error reconfiguring OpenStack: %s") % unicode(exc))
            LOG.exception(exc)
            self._throw_server_error(
                msg.HOST_RECONFIGURE_FAILED, hostname=disp)

    def _unmanage_remote_service(self, remanage=False):
        """Performs any necessary updates to the remote service to un-manage"""
        mgmt_ip = self._get_management_ip(self.config)
        local_ip = utils.get_local_system_ip(self.context)
        # If this is not the management server for the Remote Host, do nothing
        if not remanage and mgmt_ip != local_ip:
            return
        disp = self._get_display_name(self.config)
        command = '/opt/ibm/powervc/bin/powervc-manage -o unmanage'
        # If we want to bypass remote commands, then skip the unmanage
        if self.skip_remote_commands is True:
            return
        # There is nothing to do if PowerVC RPM's aren't installed on the Host
        if self._get_installed_version(self.config, True) is None:
            return
        try:
            # We need to tell the Command to notify other Management Systems
            if remanage and mgmt_ip != local_ip:
                command += ' -n ' + local_ip
            # Run the powervc-manage -unmanage command on the Remote Host
            utils.run_ssh_command(self.config, command,
                                  english=True, requires_root=True)
        except Exception as exc:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error unconfiguring OpenStack : %s") % unicode(exc))
            LOG.exception(exc)
            self._throw_server_error(
                msg.HOST_UNCONFIGURE_FAILED, hostname=disp)

    #############################################
    # ## Implementing HostRegistrar Interface ###
    #############################################
    def update_software(self, software_versions):
        """Registrar Method to update the PowerVC Software"""
        disp = self._get_display_name(self.config)
        image_prefix = self._get_host_image_file()
        new_ver = software_versions.get('mgmt_version')
        # If they didn't give us the Management Version, then not valid
        if not new_ver:
            self._throw_bad_request(msg.MISSING_ATTRIBUTE, attr='mgmt_version')
        # If they requested the latest version, figure out what that is
        if new_ver.lower() == '*latest':
            new_ver = self._get_latest_host_image_file()[1]
        # Check to see if the version specified exists on the Management System
        if not utils.file_exists(image_prefix + new_ver + '.tgz'):
            self._throw_bad_request(msg.NO_SOFTWARE_VERSION, version=new_ver)
        # Try to establish a connection to the Host to get connection errors
        self.establish_connection(dict(self.config))
        # Figure out the Current Version and the Version they specified
        cur_ver = self._get_installed_version(self.config, True)
        # If the Remote Host is already at at least that given version, end now
        if utils.compare_version(new_ver, cur_ver) <= 0:
            self._throw_bad_request(
                msg.SOFTWARE_UPTODATE, hostname=disp, version=cur_ver)
        # Need to let the sub-class verify the package is really install-able
        self._verify_package_installable(self.config)
        # Now we can actually kick off a thread to update the PowerVC Software
        greenthread.spawn(self._update_openstack_image, self.config, new_ver)
        return dict(mgmt_version=new_ver)

    def update_ip(self):
        """Registrar Method to update the Mgmt IP on the Host on change"""
        success, conf_file = (False, self._get_service_config_file())
        new_ip = utils.get_local_system_ip(self.context)
        host = self.config['host_name']
        # Log a message saying we are updating the Mgmt IP on the Remote Host
        LOG.info(_("Updating remote host %(hostname)s to new Management "
                   "IP %(mgmtip)s...") % dict(hostname=host, mgmtip=new_ip))
        try:
            # Try to connect to the Remote Host to get connection errors
            connect_info = self.establish_connection(dict(self.config))
            # We are going to force re-manage no matter if is already managed
            # in order to handle the case where they reinstalled same system
            LOG.info(_("Remanaging the Remote Host %s...") % host)
            # Ask the sub-class to give us a mapping of config values
            parm_mapping = self._get_service_param_mapping().copy()
            # Loop through each of the Mappings, adding a Parameter
            for key, val in parm_mapping.iteritems():
                parm_mapping[key] = self.config.get(val)
            self._update_service_param_mapping(parm_mapping, connect_info)
            self.config['force_unmanage'] = True
            # Now we can call the manage method to re-manage from the new IP
            self._manage_remote_service(parm_mapping)
            # We need to reflect in the configuration file are updates
            utils.set_openstack_config_value(conf_file, 'mgmt_ip', new_ip)
            utils.del_openstack_config_value(conf_file, 'restored_new_system')
            success = True
        except Exception as exc:
            # Lot the error saying we will try again later if it failed
            info = dict(hostname=host, exc=unicode(exc))
            LOG.info(_("Error updating Management IP on the remote host "
                       "%(hostname)s, trying again later: %(exc)s") % info)
        return success

    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_host_key(config)

    def _get_host_dhcp_domain(self, config):
        """Retrieves the DHCP Domain Name for the specified System"""
        try:
            # Run a command on the VIOS to get the DHCP Domain to use
            output = utils.run_ssh_command(config, '/bin/cat /etc/resolv.conf')
            return self._parse_domain_entry(output)
        # The Host may not have a resolv.conf file, so ignore the exception
        except Exception:
            return None

    ###############################################
    # ## Defining RemoteHostRegistrar Interface ###
    ###############################################
    def _get_host_image_file(self):
        """Provides the path on the Mgmt System with the image payload"""
        raise NotImplementedError()

    def _verify_package_installable(self, config):
        """Allows sub-class a hook to verify the package is installable"""
        pass

    ###############################################
    # ###  Shared Registration Helper Methods  ####
    ###############################################
    def _send_console_message(self, message):
        """Shared Helper method to send a Message to the Console Log"""
        txt = discovery_locale.translate_message(self.context, message)
        anotifier = rpc_api.get_notifier('compute', self.config['host_name'])
        anotifier.info(self.context, 'compute.instance.log', {'msg': txt})

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _get_management_ip(self, config):
        """Helper method to determine the IP that is Managing the Host"""
        mgmtip_cmd = '/bin/cat /etc/nova/nova.conf | /bin/grep management_ip'
        try:
            # Grep to see if this host is managed by another Management System
            output = utils.run_ssh_command(
                self.config, mgmtip_cmd, requires_root=True)[0]
            return utils.get_property_value(output, 'management_ip')
        # If an exception we will assume it isn't actually Installed/Managed
        except Exception:
            return None

    def _validate_root_privileges(self, config):
        """Helper method to verify that the user has root privileges on Host"""
        output = utils.run_ssh_command(
            config, '/usr/bin/id -u', requires_root=True)[0]
        # If this is the root user or we couldn't determine, consider it valid
        if len(output.strip()) <= 0 or output.strip() == '0':
            return
        output = utils.run_ssh_command(
            config, '/usr/bin/id -g', requires_root=True)[0]
        # If this is the root group or we couldn't determine, consider it valid
        if len(output.strip()) <= 0 or output.strip() == '0':
            return
        # Since it wasn't the root user or root group, throw an exception
        self._throw_bad_request(msg.HOST_NON_ROOT_USER)

    def _get_latest_host_image_file(self):
        """Helper method to find the latest version image file to use"""
        latest_vers, latest_path = (None, None)
        path_prefix = self._get_host_image_file()
        # We need to parse out the Path and the Name Prefix from the File
        last_index = path_prefix.rfind('/')
        image_path = path_prefix[:last_index]
        file_prefix = path_prefix[last_index + 1:]
        # Loop through each of the matching file names, finding the latest
        file_names = utils.get_matching_files(image_path, file_prefix, '.tgz')
        for file_name in file_names:
            file_version = file_name[len(file_prefix):len(file_name) - 4]
            # If the version is the greatest yet, then use that version
            if utils.compare_version(file_version, latest_vers) > 0:
                latest_path = image_path + '/' + file_name
                latest_vers = file_version
        return latest_path, latest_vers

    def _get_installed_version(self, config, remote=True):
        """Helper method to get the PowerVC Version installed on a System"""
        output, vers_file = ('', 'version.properties')
        try:
            # If they asked for the Host's Version, run cat through SSH
            if remote is True:
                cmd = '/bin/cat /opt/ibm/powervc/' + vers_file
                output = utils.run_ssh_command(config, cmd)[0]
                # Parse the version out of the version.properties file
                return utils.get_property_value(output, 'version')
            # If they asked for the Image Version, just parse the file name
            else:
                return self._get_latest_host_image_file()[1]
        # If the file doesn't exist, we will just return None for the version
        except Exception:
            return None

    def _add_extra_management_info(self, mgmt_file, parm_mapping):
        """Helper method add the extra attribute to the Management Info File"""
        config = dict(host=self.config['host_name'])
        config['host_display_name'] = parm_mapping['host_display_name']
        # We need to add the current time so the manage command can determine
        # if the end-point's time is skewed too much from the Management System
        date_secs = datetime.utcnow() - datetime(1970, 1, 1)
        config['time_now'] = str((date_secs.days * 86400L) + date_secs.seconds)
        # Set the DHCP Domain Name to use for the Remote Host
        config['dhcp_domain'] = parm_mapping['dhcp_domain']
        # Loop through each Attribute adding it to the Management Info file
        for key, val in config.iteritems():
            utils.set_openstack_config_value(mgmt_file, key, val)

    def _deploy_openstack_image(self, config, rmt_version):
        """Helper method to Deploy OpenStack/PowerVC RPM's to the Host"""
        img_path, img_ver = self._get_latest_host_image_file()
        lcl_dir = img_path[:img_path.rfind('/')]
        img_files = [img_path[img_path.rfind('/') + 1:]]
        rmt_dir = '/tmp/powervc_img' + utils.get_unique_file_postfix()
        try:
            # First Transfer the OpenStack Image to the Remote Host
            self._transfer_openstack_rpms(config, rmt_dir, lcl_dir, img_files)
            # Next Install the OpenStack Image onto the Remote Host
            self._install_openstack_rpms(config, rmt_dir, [img_ver], True)
        finally:
            # We need to cleanup the Temporary Directory when we are done
            utils.run_ssh_command(config, '/bin/rm -rf ' + rmt_dir, True)

    def _update_openstack_image(self, config, version):
        """Helper method to Deploy OpenStack/PowerVC RPM's to the Host"""
        disp = self._get_display_name(self.config)
        img_path = self._get_host_image_file() + version + '.tgz'
        lcl_dir = img_path[:img_path.rfind('/')]
        img_files = [img_path[img_path.rfind('/') + 1:]]
        svc_cmd = '/opt/ibm/powervc/bin/powervc-services'
        rmt_dir = '/tmp/powervc_img' + utils.get_unique_file_postfix()
        tmp_dir = '/tmp/powervc_mgmtinfo' + utils.get_unique_file_postfix()
        # Send a notification to the Console saying the version will be updated
        text = self._get_message_text(
            msg.HOST_SOFTWARE_UPDATING, hostname=disp, version=version)
        self._send_console_message(text)
        # We will first stop the services, so they don't update the state
        utils.run_ssh_command(config, svc_cmd + ' stop',
                              suppress=True, requires_root=True)
        # We want to set the hypervisor state to updating for the ui to show
        old_state = self._update_compute_node_state('updating')
        try:
            # First Transfer the OpenStack Image to the Remote Host
            self._transfer_openstack_rpms(config, rmt_dir, lcl_dir, img_files)
            # Next Install the OpenStack Image onto the Remote Host
            self._install_openstack_rpms(config, rmt_dir, [version], False)
            # Next we need to generate the certificates to use on the Host
            self._generate_certificate_files(config, tmp_dir)
            # Next need to transfer any certificates that don't already exist
            self._transfer_certificate_files(config, tmp_dir, False)
        finally:
            # We need to set the hypervisor state back to what it was before
            self._update_compute_node_state(old_state)
            # We need to cleanup the Temporary Directory when we are done
            utils.run_ssh_command(config, '/bin/rm -rf ' + rmt_dir, True)
            # Finally we want to make sure that the services are started up
            utils.run_ssh_command(config, svc_cmd + ' start',
                                  suppress=True, requires_root=True)

    def _transfer_openstack_rpms(self, config, rmt_dir, lcl_dir, img_files):
        """Helper method to Transfer the OpenStack/PowerVC RPM's to Host"""
        disp = self._get_display_name(self.config)
        untar_cmd = '/bin/tar -xf ' + rmt_dir + '/%s -C ' + rmt_dir
        # First verify that we have enough free space to do the transfer
        self._verify_enough_free_space(config)
        # Send a notification to the Console saying the install started
        text = self._get_message_text(msg.HOST_TRANSFER_STARTED, hostname=disp)
        self._send_console_message(text)
        try:
            # First we need to create the remote directory to put the files in
            utils.run_ssh_command(config, '/bin/mkdir -p ' + rmt_dir)
            # Then transfer and unzip each of the image files specified
            for img_file in img_files:
                utils.sftp_transfer_large_file(
                    config, lcl_dir + '/' + img_file, rmt_dir + '/' + img_file)
                utils.run_ssh_command(config, untar_cmd % img_file)
        except Exception as ex:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error transferring OpenStack Image: %s") % unicode(ex))
            LOG.exception(ex)
            self._throw_server_error(msg.HOST_TRANSFER_FAILED, hostname=disp)

    def _install_openstack_rpms(self, config, rmt_dir, versions, is_add):
        """Helper method to Install the OpenStack/PowerVC RPM's on Host"""
        disp = self._get_display_name(self.config)
        # Send a notification to the Console saying the install started
        md = msg.HOST_INSTALL_STARTED1 if is_add else msg.HOST_INSTALL_STARTED2
        self._send_console_message(self._get_message_text(md, hostname=disp))
        try:
            # Run the Install/Update Script on the Remote Host for each version
            for version in versions:
                inst_dir = (rmt_dir + '/powervc-%s/') % version
                # Run the Install Script through SSH to Install the RPM's
                utils.run_ssh_command(
                    config, inst_dir + 'install', requires_root=True)
        except Exception as exc:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error installing OpenStack Image: %s") % unicode(exc))
            LOG.exception(exc)
            # If the error is not enough Disk Space, throw that one instead
            if unicode(exc).lower().find('enough disk space') > 0:
                self._throw_server_error(msg.HOST_NO_SPACE, freespace='100')
            self._throw_server_error(msg.HOST_INSTALL_FAILED, hostname=disp)
        # Send a notification to the Console saying the install started
        md = msg.HOST_INSTALL_DONE1 if is_add else msg.HOST_INSTALL_DONE2
        self._send_console_message(self._get_message_text(md, hostname=disp))

    def _uninstall_openstack_rpms(self, config):
        """Helper method to UnInstall OpenStack/PowerVC RPM's from Host"""
        disp = self._get_display_name(self.config)
        # There is nothing to do if PowerVC RPM's aren't installed on the Host
        if self._get_installed_version(config, True) is None:
            return
        try:
            # Run a Command to UnInstall PowerVC from the Remote Host
            cmd = '/opt/ibm/powervc/bin/powervc-uninstall -y'
            utils.run_ssh_command(config, cmd, requires_root=True)
        except Exception as ex:
            # If we got an exception, log it and throw a user friendly one
            LOG.warn(_("Error uninstalling OpenStack Image: %s") % unicode(ex))
            LOG.exception(ex)
            self._throw_server_error(msg.HOST_UNINSTALL_FAILED, hostname=disp)

    def _generate_certificate_files(self, config, tmp_dir):
        """Helper method to transfer the certificate files to the Host"""
        crtcmd = 'powervc-manage -o gencert -f ' + tmp_dir
        csrcmd = '/opt/ibm/powervc/bin/powervc-manage -o gencsr -f ' + tmp_dir
        # Create the Temporary Directory on the Mgmt System and Host to use
        utils.run_command('/bin/mkdir -m 700 ' + tmp_dir, run_as_root=False)
        # Create the Remote Directory first, so that it is as the user
        utils.run_ssh_command(self.config, '/bin/mkdir -m 700 ' + tmp_dir)
        # Run the Command to Generate the Signing Requests for the Certificates
        utils.run_ssh_command(self.config, csrcmd, requires_root=True)
        # Transfer the Signing Requests to the Mgmt System to use in Generation
        utils.sftp_get_directory(self.config, tmp_dir, tmp_dir)
        # Run the Command to Generate the Certificates for the Remote Host
        utils.run_command(crtcmd)

    def _transfer_certificate_files(self, config, tmp_dir, overwrite=True):
        """Helper method to transfer the certificate files to the Host"""
        dmr = data_mgr.DiscoveryDataManager
        # Loop through each of the certificates, transferring the file to Host
        for fromfile, tofile, comp in dmr.get_remotehost_certificates(tmp_dir):
            tofile = fromfile[fromfile.rfind('/') + 1:]
            if comp and tofile.find(comp) < 0:
                tofile = comp + '-' + tofile
            tofile = tmp_dir + '/' + tofile
            utils.sftp_transfer_file(config, fromfile, tofile, overwrite)

    def _verify_enough_free_space(self, config):
        """Helper method to make sure we have enough free space available"""
        output = utils.run_ssh_command(config, '/bin/df -m /tmp')[0]
        availstr = output.strip().split('\n')[1].split()[3].strip()
        # Make sure that there are at least 100 MB's free to transfer/install
        if long(availstr) < 100:
            self._throw_server_error(msg.HOST_NO_SPACE, freespace='100')

    def _update_compute_node_state(self, new_state):
        """Helper method to update the hypervisor state to a new value"""
        old_state = None
        anotifier = rpc_api.get_notifier('compute', self.config['host_name'])
        try:
            # Query the database to get the Stats in order to update the State
            compute_node = service_obj.Service.get_by_host_and_topic(
                self.context, self.config['host_name'], 'compute').compute_node
            old_state = compute_node.stats.get('hypervisor_state')
            # Update the Hypervisor State to the new value that was passed in
            compute_node.stats['hypervisor_state'] = new_state
            compute_node.stats = compute_node.stats
            # Construct the info to send in the notification to the console
            info = dict(host=self.config['host_name'])
            service = service_obj.Service.get_by_compute_host(
                self.context, compute_node.host)
            info['service_id'] = service.id
            info['compute_node_id'] = compute_node.id
            info['hypervisor_hostname'] = compute_node.hypervisor_hostname
            info['stats'] = compute_node.stats
            # Send the notification that were are starting to update the state
            anotifier.info(self.context, 'compute.node.update.start', info)
            # Call Save on the Compute Node object to update it in the database
            compute_node.save()
            # Send a notification that the Hypervisor State has changed
            anotifier.info(self.context, 'compute.node.update.end', info)
        # Just log the exception if we couldn't update the hypervisor state
        except Exception as exc:
            LOG.warn(_('Error Updating Compute Node Hypervisor State'))
            LOG.exception(exc)
        return old_state

    def _cleanup_orphaned_conf_files(self):
        """Helper method to cleanup any left over Shadow CONF Files"""
        try:
            # Query the database for the the hosts that exist still
            hostfact = compute_dom.ManagedHostFactory.get_factory()
            hosts = hostfact.find_all_hosts(self.context)
            hosts = [host.host_name for host in hosts]
            # Now query all the shadown CONF files seeing if any orphaned
            for fn in utils.get_matching_files('/etc/nova', 'nova-', '.conf'):
                # If this isn't in the DB anymore, remove the the file
                if fn[5: len(fn) - 5] not in hosts:
                    parser = ConfigParser.RawConfigParser()
                    parser.read('/etc/nova/' + fn)
                    # Make sure is a Remote Host (so we don't orphan a service)
                    if parser.has_option('DEFAULT', 'host_type'):
                        if parser.get('DEFAULT', 'host_type') == 'kvm':
                            # Skip cleanup if the Host is currently being added
                            if utils.is_host_inprogress(fn[5: len(fn) - 5]):
                                continue
                            utils.run_command('rm -f /etc/nova/' + fn, True)
        # Just log exception if can't do the cleanup, since just preventative
        except Exception as exc:
            LOG.warn(_("Error cleaning up orphaned files: %s") % unicode(exc))
            LOG.exception(exc)
