# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base Classes for all Registration-related Activity"""

import getpass
from time import sleep
import webob.exc

from oslo_config import cfg
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.IntOpt('service_registration_timeout', default=360,
               help='The amount of time to wait for the service to start'),
    cfg.BoolOpt('force_deregister', default=False,
                help='Ignore any safety checks and force the de-register'),
    cfg.BoolOpt('force_deregister_uninstall', default=False,
                help='Performs an uninstall as part of de-registration')
])

LOG = utils.LOGGING.getLogger(__name__)


########################################################
# ##########  Base Registrar Implementation  ###########
########################################################
class BaseRegistrar(object):
    """Abstract Base Class to handle Host Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Host Registration"""
        is_create = kwargs.pop('create', False)
        self.context = context
        self.config = kwargs.copy()
        # If this is a New Registration, they will provide us all attributes
        if is_create:
            hostname = self.config.get(self._get_name_attribute())
            # If they didn't set the host-name, then try to evaluate one
            if hostname is None or len(hostname) == 0:
                attr_name = self._get_name_attribute()
                self.config[attr_name] = self._generate_name_value()
        # If the Registration already exists, they will provide us a hostname
        else:
            self._retrieve_registration()

    ###############################################
    # ####  Defining BaseRegistrar Interface  #####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        raise NotImplementedError()

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        raise NotImplementedError()

    def get_immutable_attributes(self):
        """Provides the list of Attributes that can't be updated"""
        return []

    def register(self):
        """Registrar Method to update the System Registration"""
        raise NotImplementedError()

    def deregister(self, force=False, uninstall=False):
        """Registrar Method to delete a System Registration"""
        raise NotImplementedError()

    def update_registration(self, updated_values):
        """Registrar Method to update the System Registration"""
        raise NotImplementedError()

    def establish_connection(self, config, throw_exception=True):
        """Retrieval Method to get the Compute Host's Connection Info"""
        connect_info = {}
        host = config.get(self._get_name_attribute())
        # Verify that all of the Required Attributes were Specified as Input
        self._verify_required_attributes()
        # Verify that attribute values have correct types
        self._verify_attribute_values()
        # Trim any input that needs it, like access_ip
        utils.trim_config_values(config)
        # If they specified a Private Key, then we make a Temporary copy
        postfix = utils.get_unique_file_postfix()
        self._update_key_files(config, '_rsa_' + postfix)
        try:
            # Retrieve the Key Information off the System being connected to
            self._retrieve_connection_info(connect_info, config)
        except webob.exc.HTTPError:
            # If this was a HTTP Error being thrown, just re-raise it
            raise
        except Exception as exc:
            authentic = None
            exctype = webob.exc.HTTPBadRequest
            hostlog = self._get_registration_log_file()
            excstr = unicode(exc.__class__.__name__ + ': ') + unicode(exc)
            host_error = self._describe_connection_error(excstr, hostlog)
            host_msgid = host_error[0]
            # If this is an unknown error, throw a 500 else a 400 error code
            if host_error[0] == msg.UNKNOWN_ERROR:
                LOG.error(_("Error retrieving Connection info: %s") % exc)
                LOG.exception(exc)
                exctype = webob.exc.HTTPInternalServerError
            elif (host_error[0] == msg.UNKNOWN_HOST_KEY or
                  host_error[0] == msg.MISMATCH_HOST_KEY):
                exctype = webob.exc.HTTPForbidden
                fingerprint = utils.format_key_fingerprint(exc.key)
                authentic = {'host_key': {'name': exc.key.get_name(),
                                          'bits': exc.key.get_bits(),
                                          'fingerprint': fingerprint}}
                host_error[1]['known_hosts'] = utils.SSH_KNOWN_HOST_FILE
                if host_error[0] == msg.MISMATCH_HOST_KEY:
                    host_msgid = msg.MISMATCH_HOST_KEY_I
            elif host_error[0] == msg.INVALID_CERTIFICATE:
                exctype = webob.exc.HTTPForbidden
                port = self._get_certificate_port()
                authentic = utils.get_certificate(config['access_ip'], port)
                if authentic:
                    authentic = utils.certificate_to_dict(authentic)
            if throw_exception:
                self._throw_exception(exctype, host_msgid,
                                      authenticity=authentic, **host_error[1])
        finally:
            # Cleanup the Private Key that we Temporarily copied over to /etc
            key_prepend = self._get_private_key_prepend()
            if key_prepend:
                utils.run_command(['/usr/bin/rm', '-f',
                                   key_prepend + host + '_rsa_' + postfix],
                                  suppress=True, run_as_root=False)
        return connect_info

    def _retrieve_registration(self):
        """Registrar method to retrieve the System Registration"""
        raise NotImplementedError()

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Compute Host's Connection Info"""
        raise NotImplementedError()

    def _get_message_bundle(self):
        """Provides the Message Bundle class to use for this Registrar"""
        return msg.BaseRegistrarMessageBundle()

    def _get_registration_type_string(self):
        """Provides the translated Registration Type for this Registrar"""
        raise NotImplementedError()

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_SYSTEM)

    def _get_registration_log_file(self):
        """Describes the name of the Log File running the Registration"""
        raise NotImplementedError()

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return []

    def _get_name_attribute(self):
        """Provides the Name of the Attribute to use for the system name"""
        raise NotImplementedError()

    def _generate_name_value(self):
        """Generates the Default Value to use for the Name Attribute"""
        access_ip = self.config.get('access_ip')
        if access_ip:
            return utils.formulate_hostname(access_ip)
        return None

    def _get_private_key_prepend(self):
        """Provides the Directory and Prefix for the Private Key Location"""
        return None

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        raise NotImplementedError()

    ###############################################
    # ###  Shared Registration Helper Methods  ####
    ###############################################
    def _verify_updating_same_system(self, connect_info, updated_values):
        """Shared Helper Method to make sure the IP is for the same host"""
        old_ip = self.config.get('access_ip')
        new_ip = updated_values.get('access_ip')
        new_id = connect_info.get('unique_id')
        # First see if we are even changing the IP, if not then nothing to do
        if new_ip and new_ip != old_ip:
            # Attempt to retrieve the unique identifier using the old IP
            old_info = self.establish_connection(self.config.copy(), False)
            # Use the identifier retrieved for the old IP, config as fall-back
            old_id = old_info.get('unique_id', self.config.get('unique_id'))
            if old_id and new_id and old_id.lower() != new_id.lower():
                self._throw_bad_request(msg.SWITCHING_SYSTEMS)

    def _update_key_files(self, config, postfix='_rsa'):
        """Shared Helper Method to make our own copy of the Private Key"""
        # Determine the Key Location given and the new Location to Copy to
        key_prepend = self._get_private_key_prepend()
        # First make sure that the Registrar supports Private Keys
        if key_prepend and config.get('private_key_data'):
            key_data = config.get('private_key_data')
            key_file = key_prepend + \
                self.config[self._get_name_attribute()] + postfix
            try:
                # If the Private Key is specified, copy it to the Key Directory
                if key_data and key_data != key_file:
                    utils.store_key_file(key_file, key_data, key_prepend)
                    config['private_key'] = key_file
            except Exception:
                self._throw_bad_request(msg.PRIVKEY_ERROR)

    def _obfuscate_password(self, config, decrypt=False):
        """Shared Helper Method to Encrypt the Passwords Provided"""
        if config.get('password'):
            try:
                password = config['password']
                if decrypt:
                    config['password'] = self._decode_password(password)
                else:
                    config['password'] = self._encode_password(password)
            except Exception, exc:
                LOG.warn(_("Error encrypting password: %s") % unicode(exc))
                self._throw_server_error(msg.PASSWORD_ENCRYPT_ERROR)

    def _decode_password(self, password):
        """Shared Helper Method to specify the deCryption method to use"""
        return utils.decode_password(password)

    def _encode_password(self, password):
        """Shared Helper Method to specify the encryption method to use"""
        return utils.encode_password(password)

    def _describe_connection_error(self, output, log):
        """Determine what the Reason was that the communication failed"""
        error_msg_id = msg.UNKNOWN_ERROR
        error_msg_parms = {'log': log}
        system_type = self._get_system_type_string()
        # If we found the log file, search through it for specific error codes
        if output and len(output) > 0:
            output = output.lower()
            # Loop through the log error strings to make the user error strings
            for msg_id, match_list in self._get_registration_error_mappings():
                for log_match in match_list:
                    # If found a match, use the message id/parms we mapped from
                    if output.find(log_match) >= 0:
                        error_msg_id = msg_id
                        error_msg_parms = {'type': system_type}
                        break
                # We need this line to break out of the second loop on match
                if error_msg_id == msg_id:
                    break
        return (error_msg_id, error_msg_parms)

    def _handle_password_key_switch(self, config):
        """Helper method to handle when they switch from password to a key"""
        cleanup_privkey = False
        # If the updated CONF contains passwords, wipe out private keys
        if config.get('password') and self.config.get('private_key'):
            config['private_key'] = None
            cleanup_privkey = True
        elif config.get('private_key_data') and self.config.get('password'):
            config['password'] = None
        return cleanup_privkey

    def _verify_required_attributes(self):
        """Shared Helper Method to Verify all Required Attributes"""
        reqlist = self.get_required_attributes()
        for reqattr in reqlist:
            match = False
            attrset = reqattr.split('|')
            for attr in attrset:
                if attr in self.config:
                    match = True
            if not match:
                reqattr = reqattr.replace('|', ' or ')
                self._throw_bad_request(msg.MISSING_ATTRIBUTE, attr=reqattr)

    def _verify_attribute_values(self):
        """Shared Helper Method to Verify correct Attribute Values"""
        keys = ['auto_add_host_key', 'auto_add_certificate',
                'force_unmanage', 'force_switch']
        for attr in self.config:
            if attr not in keys:
                continue
            try:
                self.config[attr] = utils.bool_from_string(self.config[attr])
            except ValueError:
                self._throw_bad_request(msg.INVALID_ATTRIBUTE_VALUE, attr=attr)

    def _verify_immutable_attributes(self, updated_values):
        """Shared Helper Method to Verify no Immutable Attributes specified"""
        attrlist = self.get_immutable_attributes()
        for attr in attrlist:
            if attr in updated_values:
                self._throw_bad_request(msg.IMMUTABLE_ATTRIBUTE, attr=attr)

    def _get_message_text(self, msgid, authenticity=None, **msgparms):
        """Shared Helper method to get the Translated Message Text"""
        msgb = self._get_message_bundle()
        # For now we are just going to return the english-specific message text
        try:
            msgparms['regtype'] = self._get_registration_type_string()
            msgparms['systype'] = self._get_system_type_string()
            msgparms['myip'] = utils.get_local_system_ip(self.context)
        except:
            pass
        # If this is an authenticity error, we need to format special
        if authenticity:
            text = msgb.get_message_text(self.context, msgid, **msgparms)
            text = discovery_locale.translate_message(self.context, text)
            return {'text': text, 'authenticity': authenticity}
        else:
            return msgb.get_message_text(self.context, msgid, **msgparms)

    def _throw_bad_request(self, msgid, **msgparms):
        """Shared Helper method to Throw a Bad Request exception"""
        self._throw_exception(webob.exc.HTTPBadRequest, msgid, **msgparms)

    def _throw_not_found(self, msgid, **msgparms):
        """Shared Helper method to Throw a Not Found exception"""
        self._throw_exception(webob.exc.HTTPNotFound,  msgid, **msgparms)

    def _throw_server_error(self, msgid, **msgparms):
        """Shared Helper method to Throw a Internal Server Error exception"""
        self._throw_exception(
            webob.exc.HTTPInternalServerError, msgid, **msgparms)

    def _throw_exception(self, exctype, msgid, authenticity=None, **msgparms):
        """Shared Helper method to Throw a translated exception"""
        if 'ignore_lookup' in msgparms and msgparms['ignore_lookup'] is True:
            raise exctype(explanation=msgid)
        text = self._get_message_text(
            msgid, authenticity=authenticity, **msgparms)
        raise exctype(explanation=text)


########################################################
# ########  Service Registrar Implementation  ##########
########################################################
class ServiceRegistrar(BaseRegistrar):
    """Abstract Base Class to handle Service Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Service Registration"""
        # Determine if they want to skip querying the access state
        self._skip_access_state = kwargs.pop('skip_access_state', None)
        # Now call the super class to do the rest of the construction
        super(ServiceRegistrar, self).__init__(context, **kwargs)
        hostname = self.config.get(self._get_name_attribute())
        # Make sure they gave us valid characters for the hostname to use
        if hostname and utils.is_valid_hostname(hostname) is False:
            self._throw_bad_request(msg.INVALID_CHARACTERS)

    #############################################
    # ## Implementing BaseRegistrar Interface ###
    #############################################
    def register(self):
        """Registrar Method to update the System Registration"""
        hostname = self.config.get(self._get_name_attribute())
        # Verify that all of the Required Attributes were Specified as Input
        self._verify_required_attributes()
        # Verify that attribute values have correct types
        self._verify_attribute_values()
        # Trim any input that needs it, like access_ip
        utils.trim_config_values(self.config)
        # Verify that we can establish a connection to the System specified
        connect_info = self.establish_connection(self.config.copy())
        # Allow the sub-class to switch registration types if needed
        switch = self._registration_type_switch_hook('register', connect_info)
        # Make sure that the Host specified doesn't already exist on the system
        if utils.file_exists(self._get_service_config_file()):
            config = utils.read_config_file(self._get_service_config_file())
            dispname = config.get('host_display_name', hostname)
            self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=dispname)
        # Verify that this Host doesn't already exist with a different name
        if not switch:
            self._verify_no_existing_service(connect_info)
        # If they specified a Private Key, then we will copy it over to /etc
        self._update_key_files(self.config)
        # If they specified a Password, we want to encrypt the password
        self._obfuscate_password(self.config)
        # Provide a Hook for the sub-class to do any pre-registration items
        self._pre_service_registration_hook('register', connect_info)
        # Ask the sub-class to give us a mapping of config values to parameters
        parm_mapping = self._get_service_param_mapping().copy()
        # Loop through each of the Mappings, adding a Parameter for the Mapping
        for key, val in parm_mapping.iteritems():
            parm_mapping[key] = self.config.get(val)
        self._update_service_param_mapping(parm_mapping, connect_info)
        # Create the Linux Service used to manage this given resource
        self._create_linux_service(parm_mapping)
        # Provide a Hook for the sub-class to do any post-registration items
        self._post_service_registration_hook('register', connect_info)
        # Now that we let the sub-class's hook run, we can start the service
        self._start_linux_service(restart=switch)
        # Wait for the Service to finish starting after we created it
        self._verify_service_started()
        # Add the Registration Time to the set of Attributes returned
        self.__add_registration_time()

    def deregister(self, force=False, uninstall=False, soft=False):
        """Registrar Method to delete a System Registration"""
        host = self.config.get(self._get_name_attribute())
        regtype = self._get_registration_type_string()
        service = self._get_service_instance()
        LOG.info(_("Removing %(regtype)s %(host)s...") %
                 dict(regtype=regtype, host=host))
        # If the System wasn't found in the DB, throw an exception
        if not service and not force:
            self._throw_not_found(msg.SYSTEM_NOTFOUND, hostname=host)
        # Provide a Hook for the sub-class to do any pre-registration items
        self._pre_service_registration_hook('deregister', force=force)
        # Delete the Linux Service used to manage this given resource
        self._delete_linux_service()
        # Cleanup any RSA Keys that were copied to the /etc directory
        key_prepend = self._get_private_key_prepend()
        if key_prepend:
            utils.run_command(['/usr/bin/rm', '-f',
                               key_prepend + host + '_rsa'],
                              suppress=True, run_as_root=False)
        # Provide a Hook for the sub-class to do any post-registration items
        self._post_service_registration_hook('deregister')
        # Cleanup the Service from the Database after we are done
        if service and not soft:
            self._destroy_service_instance(service)
        LOG.info(_("Removed %(regtype)s %(host)s.") %
                 dict(regtype=regtype, host=host))

    def update_registration(self, updated_values,
                            force_restart=False, wait_for_restart=True):
        """Registrar Method to update the System Registration"""
        value_changed = False
        hostname = self.config.get(self._get_name_attribute())
        service = self._get_service_instance()
        if not service:
            self._throw_not_found(msg.SYSTEM_NOTFOUND, hostname=hostname)
        # Make sure they didn't specify any attribute that cannot be changed
        self._verify_immutable_attributes(updated_values)
        # Trim any input that needs it, like access_ip
        utils.trim_config_values(updated_values)
        # Loop through each of the Updated Values, seeing if any changed
        for key in updated_values:
            if utils.has_value_changed(self.config, updated_values, key):
                value_changed = True
        # No-op if they didn't provide us with any parameters we can change
        if value_changed is True:
            # Since they may have switched between Passwords and Private Keys
            cleanup_privkey = self._handle_password_key_switch(updated_values)
            # First make sure we can establish a connection with the new config
            new_config = dict(self.config, **updated_values)
            connect_info = self.establish_connection(new_config)
            # Allow the sub-class to switch registration types if needed
            self._registration_type_switch_hook('update-registration',
                                                connect_info, updated_values)
            # Verify that the IP specified is not for a different system
            self._verify_updating_same_system(connect_info, updated_values)
            # If they specified a Password, we want to encrypt the password
            self._obfuscate_password(updated_values)
            # If they specified a Private Key, then we will copy it to /etc
            self._update_key_files(updated_values)
            # Provide a Hook for the sub-class to do any pre-registration items
            self._pre_service_registration_hook('update-registration',
                                                connect_info, updated_values)
            # Ask the sub-class to give us a mapping of config values to parms
            parm_mapping = self._get_service_param_mapping().copy()
            # Loop through each of the Mappings, adding a Parm for the Mapping
            self._update_service_param_mapping(parm_mapping, connect_info)
            # Update the Linux Service used to manage this given resource
            self._update_linux_service(parm_mapping, updated_values)
            # Cleanup the Private Key if they switched from Key to Password
            if cleanup_privkey is True and self.config.get('private_key'):
                utils.run_command('rm -f ' + self.config['private_key'], True)
            # Provide a Hook for the sub-class to do post-registration items
            self._post_service_registration_hook('update-registration',
                                                 connect_info, updated_values)
            # We will refresh the values from the CONF file for use later
            self._retrieve_registration()
        # Restart the service if the values changed or asked to force restart
        if value_changed is True or force_restart is True:
            self._start_linux_service(restart=True)
            # Wait for the Service to finish starting after we updated it
            if wait_for_restart is True:
                self._verify_service_started(service['updated_at'])

    def _retrieve_registration(self):
        """Registrar method to retrieve the System Registration"""
        host_name = self.config.get(self._get_name_attribute())
        config_file = self._get_service_config_file()
        # We first want to see if the File exists, otherwise it isn't added
        if not utils.file_exists(config_file):
            self._throw_not_found(msg.SYSTEM_NOTFOUND, hostname=host_name)
        # Parse the Configuration File for the given Service
        config = utils.read_config_file(config_file)
        # Allow the subclass to parse any attributes out of the config file
        self._parse_service_config(config)
        # We need to decrypt the password retrieved from the file
        self._obfuscate_password(self.config, decrypt=True)
        # Skip the access-state/register-time if they asked for performance
        if not self._skip_access_state:
            # Figure out the access_state based on if the Service is running
            self.config['access_state'] = self._determine_access_state()
            # Add the Registration Time to the set of Attributes returned
            self.__add_registration_time()

    ###############################################
    # ### Defining ServiceRegistrar Interface  ####
    ###############################################
    def _get_service_name(self):
        """Provides the name of the Linux Service created by this Registrar"""
        raise NotImplementedError()

    def _get_service_config_file(self):
        """Provides the name of the Config File created by this Registrar"""
        raise NotImplementedError()

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        raise NotImplementedError()

    def _get_service_list_type_params(self):
        """Provides the list of Properties for the Registrar that are lists"""
        return []

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        pass

    def _pre_service_registration_hook(self, action, connect_info=None,
                                       config={}, force=False):
        """Hook to be able to take actions prior to doing the registration"""
        pass

    def _post_service_registration_hook(self, action,
                                        connect_info=None, config={}):
        """Hook to be able to take actions after doing the registration"""
        pass

    def _registration_type_switch_hook(self, action,
                                       connect_info=None, config={}):
        """Hook to be able to switch registration types if needed"""
        return False

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        parm_mappings = self._get_service_param_mapping().copy()
        for key, val in parm_mappings.iteritems():
            if config.get(key) is not None:
                newval = self._parse_service_config_value(key, config.get(key))
                self.config[val] = newval

    def _get_service_instance(self, verify_existence=False):
        """Retrieves the Service Instance from the Database for this System"""
        raise NotImplementedError()

    def _destroy_service_instance(self, service):
        """Cleans up the Service Instance from the Database on Deregister"""
        raise NotImplementedError()

    def _verify_no_existing_service(self, connect_info):
        """Helper Method to see if the System already has been Registered"""
        raise NotImplementedError()

    def _wait_for_service_initialization(self):
        """Allows sub-class to wait longer for service initialization"""
        pass

    def _describe_host_service_error(self, force_check=False):
        """Determine what the Reason was that the Host Service didn't start"""
        raise NotImplementedError()

    def _create_linux_service(self, parm_mapping):
        """Provides the implementation for starting/restarting the Service"""
        svcname, config = self._get_service_name(), dict()
        # Loop through each of the Mappings, adding a Parameter for the Mapping
        for key, val in parm_mapping.iteritems():
            val = self._format_service_config_value(val)
            key, sect = utils.parse_section_from_key(key, 'DEFAULT')
            config.setdefault(sect, dict())[key] = val
        # Write out the configuration properties to a temporary file
        tempfile = utils.write_temp_config_file(config)
        # Run the powervc-manage command to Create the Service and Start it
        # Use relative path on command for rootwrap's sake
        utils.run_command(
            'powervc-manage -o addhost -s ' + svcname + ' -f ' + tempfile)
        utils.run_command('/bin/rm -f ' + tempfile, True, False)

    def _update_linux_service(self, parm_mapping, updated_values):
        """Provides the implementation for updating the Service"""
        svcname, config = self._get_service_name(), dict()
        # Loop through each of the Mappings, adding a Parm for the Mapping
        for key, val in parm_mapping.iteritems():
            if utils.has_value_changed(self.config, updated_values, val):
                val = updated_values.get(val)
                val = self._format_service_config_value(val)
                key, sect = utils.parse_section_from_key(key, 'DEFAULT')
                config.setdefault(sect, dict())[key] = val
        # Write out the configuration properties to a temporary file
        tempfile = utils.write_temp_config_file(config)
        # Run the powervc-manage command to Update the Service and Restart it
        utils.run_command(
            'powervc-manage -o edithost -s ' + svcname + ' -f ' + tempfile)
        utils.run_command('/bin/rm -f ' + tempfile, True, False)

    def _delete_linux_service(self):
        """Provides the implementation for deleting/removing the Service"""
        svcname = self._get_service_name()
        # Run the powervc-manage command to cleanup the System Registration
        utils.run_command('powervc-manage -o removehost -s ' + svcname)

    def _start_linux_service(self, restart=False):
        """Provides the implementation for starting/restarting the Service"""
        srvname = self._get_service_name()
        action = 'restart' if restart is True else 'start'
        utils.run_command('service ' + srvname + '  ' + action)

    def _service_status_command(self, srvname):
        """Runs the command to get the status of the primary Service"""
        return utils.run_command(
            'powervc-services single ' + srvname + ' status', True)

    ###############################################
    # ###  Shared Registration Helper Methods  ####
    ###############################################
    def _host_service_status(self, retry_on_timeout=True):
        """Shared Helper Method to override to say if the Service is running"""
        srvname = self._get_service_name()
        out, err = self._service_status_command(srvname)
        # Check if the Status is Started or Stopped for the Service
        if out.find('is running') > 0 or out.find('Active: active (run') > 0:
            return 'started'
        if out.find('is stopped') > 0 or out.find('Active: inactive (s') > 0:
            return 'stopped'
        # If we timed out, just try again...but do not do indefinitely
        if err is not None and err.find('timed out') >= 0 and retry_on_timeout:
            return self._host_service_status(False)
        # Else return 'error' (Note: The if/else is to use the err variable)
        return 'error' if err else 'error'

    def _format_service_config_value(self, val):
        """Shared Helper Method to Format the Config File Value properly"""
        val = val if val is not None else ''
        # If this is a list, we need to properly format it for the config file
        if isinstance(val, list):
            val = ','.join(val)
        return unicode(val)

    def _parse_service_config_value(self, key, val):
        """Internal Helper Method to Parse the Config File Value properly"""
        val = val if val != '' else None
        # If this is a comma-separated list, we need to turn it a list type
        if val and key in self._get_service_list_type_params():
            val = val.split(',')
        return val

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _determine_access_state(self):
        """Internal Helper Method to Determine the Service's Access State"""
        # Invoke the Method to Describe the Error, and get the Access State
        return self._describe_host_service_error()[0]

    def __add_registration_time(self):
        """Internal Helper Method to Set the Registration Date on the Config"""
        try:
            # Retrieve the Created Time from the Service for Registered Time
            service = self._get_service_instance()
            if service is not None:
                self.config['registered_at'] =\
                    utils.format_isotime(service['created_at'])
        except Exception:
            pass

    def _throw_service_start_failed(self, host_error, is_running):
        """Shared Helper Method to throw exception for service not starting"""
        # If the Service is still trying to Start, just say it timed out
        if is_running:
            svc_name = self._get_service_name()
            self._throw_server_error(msg.SERVICE_TIMEOUT, service=svc_name)
        self._throw_server_error(host_error[0], **host_error[1])

    def _verify_service_started(self, updated_at=None):
        """Internal Helper Method to Block until the Service has Started"""
        started = False
        # Loop for Up to the timeout to wait to See if the Service Started
        for index in range(cfg.CONF.service_registration_timeout / 3):
            # If the Process is Dead for the Service, no need to keep checking
            # We do give it some buffer however.  The network sometimes gets
            # restarted as part of the registration, but only a short outage
            # We run the service status command first because the first time
            # may time out (re-saturation of the network)
            if self._host_service_status() != 'started' and index >= 30:
                break
            # Check to See if the Service Entry is in the DB yet or not
            service = self._get_service_instance(True)
            if service:
                # It is started if this is a Create or UpdatedAt has Changed
                if updated_at is None or service['updated_at'] != updated_at:
                    started = True
                    break
            # Sleep for 3 seconds in between checks if the Service has started
            sleep(3)
        # If the Service didn't end up starting, then figure out the reason
        if started is False:
            is_running = self._host_service_status() == 'started'
            # Ask the sub-class to give a reason why the Service didn't start
            host_error = self._describe_host_service_error(True)
            # If the Service didn't register, we need to clean it up
            if updated_at is None:
                self.deregister(force=True)
            # Throw an exception saying the service failed to start in time
            self._throw_service_start_failed(host_error, is_running)
        # Allow the sub-class to wait additional time for initialization
        self._wait_for_service_initialization()


########################################################
# #####  Remote Service Registrar Implementation  ######
########################################################
class RemoteServiceRegistrar(ServiceRegistrar):
    """Abstract Base Class to handle Remote Service Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Remote Service Registration"""
        super(RemoteServiceRegistrar, self).__init__(context, **kwargs)

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _create_linux_service(self, parm_mapping):
        """Provides the implementation for starting/restarting the Service"""
        config = dict()
        cfg_file = self._get_service_config_file()
        f_owner = getpass.getuser()
        # Create a shadow/proxy config file for the service
        # Use relative path on commands for rootwrap's sake
        utils.run_command('touch ' + cfg_file)
        utils.run_command('chown ' + f_owner + ':' + f_owner + ' ' + cfg_file)
        utils.run_command('chmod 640 ' + cfg_file)
        # Loop through each of the Mappings, adding a Parm for the Mapping
        for key, val in parm_mapping.iteritems():
            config[key] = self._format_service_config_value(val)
        # Now we can write the value to the Nova Configuration File
        utils.write_config_file(cfg_file, config, False)
        # First let the sub-class perform the Management on the Remote System
        try:
            self._manage_remote_service(parm_mapping)
        # If the configuration failed, then clean up the Proxy File
        except:
            utils.run_command('rm -f ' + cfg_file)
            raise

    def _update_linux_service(self, parm_mapping, updated_values):
        """Provides the implementation for updating the Service"""
        cfg_file = self._get_service_config_file()
        # First let the sub-class update the Management on the Remote System
        self._update_remote_service(parm_mapping, updated_values)
        # Loop through each of the Mappings, adding a Parm for the Mapping
        for key, val in parm_mapping.iteritems():
            if utils.has_value_changed(self.config, updated_values, val):
                val = updated_values.get(val)
                val = self._format_service_config_value(val)
                # Run the OpenStack Config Command to update the CONF property
                utils.set_openstack_config_value(cfg_file, key, val)

    def _delete_linux_service(self):
        """Provides the implementation for deleting/removing the Service"""
        cfg_file = self._get_service_config_file()
        # First lest the sub-class remove the Management from the Remote System
        self._unmanage_remote_service()
        # Run a command to remove the proxy config file for the service
        utils.run_command('rm -f ' + cfg_file)

    def _start_linux_service(self, restart=False):
        """Provides the implementation for starting/restarting the Service"""
        srvcmd = '/opt/ibm/powervc/bin/powervc-services'
        action = 'restart' if restart is True else 'start'
        utils.run_ssh_command(
            self.config, srvcmd + ' ' + action, requires_root=True)

    def _service_status_command(self, srvname):
        """Runs the command to get the status of the primary Service"""
        cmd = '/opt/ibm/powervc/bin/powervc-services single '
        return utils.run_ssh_command(self.config, cmd + srvname + ' status',
                                     suppress=True, requires_root=True)

    def _determine_access_state(self):
        """Internal Helper Method to Determine the Service's Access State"""
        return 'started'

    def _throw_service_start_failed(self, host_error, is_running):
        """Shared Helper Method to throw exception for service not starting"""
        # If the Service is still starting and don't know why, say it timed out
        if is_running and host_error[0] == msg.UNKNOWN_ERROR:
            svc_name = self._get_service_name()
            self._throw_server_error(msg.SERVICE_TIMEOUT, service=svc_name)
        self._throw_server_error(host_error[0], **host_error[1])

    ###############################################
    #  Defining RemoteServiceRegistrar Interface  #
    ###############################################
    def _manage_remote_service(self, parm_mapping):
        """Performs any necessary updates to the remote service to manage"""
        pass

    def _update_remote_service(self, parm_mapping, updated_values):
        """Performs any necessary updates to the remote service on edit"""
        pass

    def _unmanage_remote_service(self, notify=False):
        """Performs any necessary updates to the remote service to un-manage"""
        pass
