# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Security Utilities for all Registration-related Activity"""

import ConfigParser
import commands
import eventlet
import getpass
import grp
import os.path
import paramiko
import pexpect
import re
import shlex
import six
import socket
import ssl
import tempfile
import webob.exc

from backports.ssl_match_hostname import match_hostname, CertificateError
from binascii import hexlify
from datetime import datetime
from hashlib import sha1 as sha1
from paramiko import SSHException as SSHException
from paramiko.config import SSH_PORT as PMK_SSH_PORT
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.ber import decoder as ber_decoder
from pyasn1_modules import rfc2459
from os import chmod, chown, listdir, getpid, close as os_close

from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_concurrency import processutils as procutils
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import timeutils
from oslo_utils import strutils

from powervc_keystone.encrypthandler import EncryptHandler
from powervc_discovery.registration import registrar_messages as msg

if cfg.CONF.config_file[0].find('nova') >= 0:
    from nova import utils
    ROOT_HELPER = 'sudo nova-rootwrap %s' % cfg.CONF.rootwrap_config
else:
    from cinder import utils
    ROOT_HELPER = 'sudo cinder-rootwrap %s' % cfg.CONF.rootwrap_config

LOG = logging.getLogger(__name__)
LOGGING = logging
UNIQUE_COUNTER = 1
SSH_TIMEOUT = 30
SSH_PORT = PMK_SSH_PORT
SSH_KNOWN_HOST_FILE = '/opt/ibm/powervc/data/known_hosts'
CERT_PATH = '/opt/ibm/powervc/data/certs/'
CERT_EXT = '.crt'
LOCAL_IP = None
NAME_TYPES_MAP = {
    'dNSName': 'DNS'  # IP not included because match_hostname lacks support
}


########################################################
# ############  External Utility Methods  ##############
########################################################
@lockutils.synchronized('registrar_command', 'powervc-')
def run_command(cmd, suppress=False, run_as_root=True):
    """Shared Helper Method to Run a Command on the local system"""
    if isinstance(cmd, list):
        for index in range(len(cmd)):
            if isinstance(cmd[index], unicode):
                cmd[index] = cmd[index].encode('utf8')
    else:
        cmd = shlex.split(cmd.encode('utf8'))
    if suppress is False:
        return procutils.execute(*cmd, run_as_root=run_as_root,
                                 root_helper=ROOT_HELPER)
    else:
        return procutils.trycmd(*cmd, check_exit_code=False,
                                run_as_root=run_as_root,
                                root_helper=ROOT_HELPER)


@lockutils.synchronized('registrar_command', 'powervc-')
def run_ssh_command(config, cmd, suppress=False, english=False,
                    timeout=1800, requires_root=False):
    """Shared Helper Method to Run an Command thru SSH to the System"""
    try:
        # Get the Connection to the SSH Server on the Host provided
        ssh, proc_input = _get_ssh_connection(config), None
        # If this command requires SUDO, then tack it onto the front
        # We will pass in the password to STDIN on SUDO if they gave us one
        # If using a Private Key, do non-interactive so no password prompt
        if requires_root and config.get('supports_sudo'):
            # Only do this if the user id isn't root, since root doesn't need
            if config.get('user_id') != 'root':
                option = '-S ' if config.get('password') else '-n '
                passwd = config['password'] if config.get('password') else None
                proc_input = decode_password(passwd) if passwd else None
                cmd = '/usr/bin/sudo ' + option + cmd
        # If they asked to use English for the language, add to command
        cmd = ('LANG=en_US; ' + cmd) if english else cmd
        # Put an overall timeout over the command so that it doesn't hang
        with eventlet.timeout.Timeout(timeout):
            # Execute the given SSH command that was provided on input
            out, err = _ssh_execute(ssh, cmd, proc_input, suppress is False)
        # Make sure to close the SSH connection when we are done with it
        ssh.close()
        return (out, err)
    # Raise an exception so that we can parse appropriately
    except eventlet.timeout.Timeout:
        if suppress is False:
            raise Exception('Timeout for cmd %s' % cmd)
        return ('', 'eventlet.timeout.Timeout')
    except Exception as exc:
        if suppress is False:
            raise
        return ('', unicode(exc))


def run_command_pwd_prompt(cmd, password, prompt):
    """Shared Helper Method to Run a Command prompting for Password"""
    p = pexpect.spawn(cmd)
    p.expect(prompt)
    p.sendline(password)
    p.expect(pexpect.EOF, timeout=30)
    p.close()
    # If there was a failure, throw the STDOUT/STDERR in an Exception
    if p.exitstatus != 0:
        raise pexpect.ExceptionPexpect(p.before)
    return p.before


@lockutils.synchronized('registrar_command', 'powervc-')
def sftp_transfer_file(config, local_file, remote_file, overwrite=True):
    """Shared Helper Method to Transfer a File thru SFTP to the System"""
    ssh = _get_ssh_connection(config)
    # If they don't want to overwrite the file, check if it exists first
    if overwrite is False:
        try:
            # Execute the given SSH command that was provided on input
            procutils.ssh_execute(
                ssh, '/bin/cat ' + remote_file, check_exit_code=True)
            # If the Cat Command is successful, then it exists so return
            ssh.close()
            return
        except:
            pass
    # Open an SFTP Session and transfer the file to the remote system
    sftp = ssh.open_sftp()
    sftp.put(local_file, remote_file)
    # Make sure to close the SSH connection when we are done with it
    (sftp.close(), ssh.close())


def sftp_transfer_large_file(config, local_file, remote_file):
    """Shared Helper Method to Transfer a Large File thru SFTP to the System"""
    parser = ConfigParser.RawConfigParser()
    parser.optionxform = str
    # Since there is still an issue with Paramiko when Monkey-patching Eventlet
    # where it doesn't always transfer the entire packet size at once and ends
    # up causing packet corruption on the receiving side at times transferring
    # large files, so we need to do the transfer in a separate process that
    # OpenStack hasn't done the patching to, so will use the Data Manager
    # Utilities, since it does the same thing as we are doing here for transfer
    transfer_cmd = '/usr/bin/python -c "import sys; from powervc_oslo.config' \
        ' import data_utils; data_utils.setup_logging(use_console=True); sys' \
        '.exit(data_utils.sftp_transfer_file(\'%s\', \'%s\', \'%s\'))"'
    # Set up the properties in the format the Data Manager Utilities expects
    props = dict(ip='access_ip', user='user_id',
                 passwd='password', private_key='private_key')
    for key, val in props.iteritems():
        if config.get(val):
            parser.set('DEFAULT', 'remote_host_' + key, config[val])
    # Create a temporary file to write the IP and Credentials to read from
    with tempfile.NamedTemporaryFile(prefix='powervc_temp_', dir='/tmp') as f:
        parser.write(f.file)
        f.file.flush()
        transfer_cmd = transfer_cmd % (f.name, local_file, remote_file)
        run_command(transfer_cmd, run_as_root=False)


@lockutils.synchronized('registrar_command', 'powervc-')
def sftp_transfer_directory(config, local_dir, remote_dir):
    """Shared Helper Method to Transfer a Director thru SFTP to the System"""
    ssh = _get_ssh_connection(config)
    # First we need to make sure that the directory exists on the target
    procutils.ssh_execute(ssh, '/bin/mkdir -p ' + remote_dir,
                          check_exit_code=True)
    # Open an SFTP Session and transfer the directory to the remote system
    sftp = ssh.open_sftp()
    # Loop through each file in the local directory and transfer each file
    for filename in listdir(local_dir):
        sftp.put(local_dir + '/' + filename, remote_dir + '/' + filename)
    # Make sure to close the SSH connection when we are done with it
    (sftp.close(), ssh.close())


@lockutils.synchronized('registrar_command', 'powervc-')
def sftp_get_directory(config, remote_dir, local_dir):
    """Shared Helper Method to Get a Directory thru SFTP from the System"""
    ssh = _get_ssh_connection(config)
    # Open an SFTP Session and transfer the directory from the remote system
    sftp = ssh.open_sftp()
    # Loop through each file in the remote directory and transfer each file
    for filename in sftp.listdir(remote_dir):
        sftp.get(remote_dir + '/' + filename, local_dir + '/' + filename)
    # Make sure to close the SSH connection when we are done with it
    (sftp.close(), ssh.close())


def file_exists(afile, run_as_root=False):
    """Shared Helper Method to Determine as Root if a file exists"""
    # If they asked us to run as root, we need to do something special
    if run_as_root:
        try:
            # Attempt to Read the File as Root to see if the File exist
            utils.read_file_as_root(afile)
            return True
        except Exception:
            return False
    return os.path.exists(afile)


def get_matching_files(adir, prefix, suffix):
    """Shared Helper Method to return the list of files matching input"""
    filenames = []
    for filename in listdir(adir):
        # We need to exclude the config files for the health services
        if filename == (prefix + 'health' + suffix):
            continue
        if filename.startswith(prefix) and filename.endswith(suffix):
            filenames.append(filename)
    return filenames


def read_config_file(filename):
    """Shared Helper Method to read from a Configuration file"""
    config = dict()
    parser = ConfigParser.RawConfigParser()
    parser.read(filename)
    for key in parser.defaults():
        val = encodeutils.safe_decode(parser.get('DEFAULT', key))
        config[key] = unescape_config_value(val)
    return config


def write_config_file(filename, config_dict, set_permissions=True):
    """Shared Helper Method to write to a Configuration file"""
    parser = ConfigParser.RawConfigParser()
    parser.optionxform = str
    # If they didn't provide the sections, we need to convert
    if ((config_dict and
         not isinstance(config_dict.itervalues().next(), dict))):
        config_dict = dict(DEFAULT=config_dict)
    # Loop through the Dictionary adding the Configuration Values
    for sect in config_dict.keys():
        if sect != 'DEFAULT' and not parser.has_section(sect):
            parser.add_section(sect)
        for key, val in config_dict[sect].iteritems():
            val = escape_config_value(val)
            if isinstance(val, unicode):
                val = val.encode('utf-8')
            parser.set(sect, key, val)
    # Make sure we set the permissions so only the owner has access
    if set_permissions:
        run_command('/bin/touch ' + filename, True, False)
        run_command('/bin/chmod 600 ' + filename, True, False)
    # Now we can write the values to the Configuration File
    with open(filename, 'w') as f:
        parser.write(f)


def write_temp_config_file(config_dict):
    """Shared Helper Method to write to a Temporary Configuration file"""
    # Let the temporary file library create the temporary file for us
    fh, filename = tempfile.mkstemp(
        suffix='.conf', prefix='powervc_temp_', dir='/tmp')
    os_close(fh)
    # Now that we have a temporary file created, we can write to it
    write_config_file(filename, config_dict, False)
    return filename


def read_from_file(filename):
    """Shared Helper Method to read the data from a given file"""
    with open(filename, 'r') as f:
        return f.read()


def write_to_file(filename, data):
    """Shared Helper Method to write the data to the given file"""
    with open(filename, 'w') as f:
        f.write(data)


def get_unique_file_postfix():
    """Shared Helper Method to Get a Unique File Postfix"""
    global UNIQUE_COUNTER
    UNIQUE_COUNTER += 1
    time_delta = datetime.now() - datetime(1970, 1, 1)
    time_secs = (time_delta.days * 86400) + time_delta.seconds
    return '_temp_' + str(int(time_secs)) + \
        '_' + str(getpid()) + str(UNIQUE_COUNTER)


def encode_password(password):
    """Utility for encrypting a password with the AES Encryption"""
    encrypt_handler = EncryptHandler()
    return encrypt_handler.encode(password)


def decode_password(password):
    """Utility for decrypting a password with the AES Encryption"""
    encrypt_handler = EncryptHandler()
    return encrypt_handler.decode(password)


def verify_certificate(config, port, try_ca=True):
    """Verifies that the certificate is trusted"""
    autoadd = config.get('auto_add_certificate')
    autoadd = True if (autoadd is None or autoadd) else False
    if autoadd or not try_ca:
        # Either caller has told us not to try CA validation, or they've told
        # us to auto-add this certificate as trusted, in which case CA
        # validation is also unnecessary. So mark that failed, though it's
        # really unattempted, so that we'll move on to standard certificate
        # retrieval, etc. without attempting CA validation.
        ca_validation_failed = True
    else:
        # Retrieve and validate the new certificate
        ca_validation_failed = False  # init
        try:
            # Attempt to validate using the local CA bundle
            rawcert = get_certificate(config['access_ip'], port,
                                      auto_add=False, validate=True)
        except ssl.SSLError:
            ca_validation_failed = True
    if ca_validation_failed:
        # still need the cert
        rawcert = get_certificate(config['access_ip'], port, auto_add=False)
    newcert = certificate_to_dict(rawcert) if rawcert else rawcert
    if not ca_validation_failed:
        # Attempt additional hostname validation not done by ssl.wrap_socket
        # This ensures that valid certificates match up with the host to which
        # they belong. We shouldn't see the cert for A when we connect to B.
        names = {'subject': ((('commonName',
                               newcert['certificate']['subject']),),)}
        alt_names = _get_subject_alt_names(rawcert)
        if alt_names:
            names['subjectAltName'] = alt_names
        try:
            match_hostname(names, config['access_ip'])
            # CA validation passed, so remove any previously-stored cert
            remove_certificate(config['access_ip'])
        except CertificateError:
            ca_validation_failed = True
    if ca_validation_failed:
        if autoadd:
            # Caller has told us to auto-add this certificate as trusted, so
            # do that without attempting validation
            _store_certificate(rawcert, config['access_ip'])
        else:
            # Attempt to validate against old stored certificate
            oldcert = get_existing_certificate(config['access_ip'])
            oldcert = certificate_to_dict(oldcert) if oldcert else oldcert
            if not oldcert:
                raise Exception("Certificate Verify Needed")
            elif oldcert != newcert:
                raise Exception("Certificate doesn't Match")
    return newcert


def get_certificate_file(address):
    """Returns the path of the file the certificate will be stored in"""
    return CERT_PATH + re.sub('[:.]', '_', address) + CERT_EXT


def get_certificate(address, port, auto_add=True, store=False, validate=False):
    """Retrieves a certificate issuer, subject, validity as a dict"""
    try:
        # Use the EventLet Timeout to make sure we don't hang
        with eventlet.timeout.Timeout(30):
            conn = socket.create_connection((address, port))
            if validate:
                sock = ssl.wrap_socket(
                    conn, cert_reqs=ssl.CERT_REQUIRED,
                    ca_certs='/etc/pki/tls/certs/ca-bundle.crt')
            else:
                sock = ssl.wrap_socket(conn)
            rawcert = ssl.DER_cert_to_PEM_cert(sock.getpeercert(True))
    # Raise an exception so that we can parse appropriately
    except eventlet.timeout.Timeout:
        raise Exception('Timed out getting SSL Certificate')
    # Check to see if the Certificate retrieved has expired
    _check_certificate_expiration(rawcert)
    # If user has allowed an untrusted certificate, persist for future
    if store is True and auto_add is True:
        _store_certificate(rawcert, address)
    return rawcert


def get_existing_certificate(address):
    """Retrieves the existing certificate stored on the file system"""
    try:
        # If the Certificate has been persisted, read its contents
        if file_exists(get_certificate_file(address)):
            with open(get_certificate_file(address), 'r') as f:
                return f.read()
    # If got an exception just ignore, since we currently don't care
    except Exception:
        pass
    return None


def certificate_to_dict(rawcert):
    """Internal Helper Method to convert a Certificate to a Dictionary"""
    dercert = ssl.PEM_cert_to_DER_cert(rawcert)
    cert = der_decoder.decode(dercert, asn1Spec=rfc2459.Certificate())[0]
    tbs = cert.getComponentByName('tbsCertificate')
    subject = _get_cert_common_name(tbs, 'subject')
    issuer = _get_cert_common_name(tbs, 'issuer')
    validity = tbs.getComponentByName('validity')
    not_before = validity.getComponentByName('notBefore').getComponent()
    not_after = validity.getComponentByName('notAfter').getComponent()
    not_before = datetime.strptime(str(not_before), '%y%m%d%H%M%SZ')
    not_after = datetime.strptime(str(not_after), '%y%m%d%H%M%SZ')
    fingerprint = ':'.join(x.encode('hex') for x in sha1(dercert).digest())
    return {'certificate': {'subject': subject, 'issuer': issuer,
                            'not_before': not_before.isoformat(),
                            'not_after': not_after.isoformat(),
                            'fingerprint': fingerprint}}


def get_certificate_identifiers(rawcert):
    """Internal Helper Method to return the Certificate Identifiers"""
    dercert = ssl.PEM_cert_to_DER_cert(rawcert)
    cert = der_decoder.decode(dercert, asn1Spec=rfc2459.Certificate())[0]
    tbs = cert.getComponentByName('tbsCertificate')
    issuer = _get_cert_component_string(tbs, 'issuer')
    serial = tbs.getComponentByName('serialNumber')
    serial = str(hex(long(str(serial)))).lower()
    serial = serial[2:len(serial) - 1]
    return serial, issuer


def remove_certificate(address):
    """Removes the certificate file for the specified address if exists"""
    try:
        os.remove(get_certificate_file(address))
    except OSError:
        pass


def verify_host_key(config):
    """Initialize a connection over SSH forcing a check of host key"""
    ssh = paramiko.SSHClient()
    if not os.path.exists(SSH_KNOWN_HOST_FILE):
        _create_known_hosts()
    ssh.load_host_keys(SSH_KNOWN_HOST_FILE)
    # Specify whether to throw an Exception if the Key is Missing
    if (('auto_add_host_key' not in config or
         str(config['auto_add_host_key']).lower() == 'true')):
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    else:
        ssh.set_missing_host_key_policy(HostKeyPolicy())
    # Try to Connect to SSH on the System using the IP and Port specified
    port = int(config.get('ssh_port', SSH_PORT))
    try:
        ssh.connect(config['access_ip'], port=port, timeout=SSH_TIMEOUT)
    except UnknownHostKeyException:
        raise
    # If we got a Bad Key Error, see if we should Auto-Add the Key or not
    except paramiko.BadHostKeyException as e:
        if (('auto_add_host_key' not in config or
             str(config['auto_add_host_key']).lower() == 'true')):
            hostkeys = ssh.get_host_keys()
            hostkeys.add(e.hostname, e.key.get_name(), e.key)
            hostkeys.save(SSH_KNOWN_HOST_FILE)
        else:
            raise
    except SSHException:
        pass
    # Get the Keys from the Connection before we close it
    hostkeys = ssh.get_host_keys()
    key = hostkeys.get(config['access_ip'], {}).get('ssh-rsa', None)
    ssh.close()
    # Return a dictionary representation of the Authenticity
    if key is not None:
        fingerprint = format_key_fingerprint(key)
        return {'host_key': {'name': key.get_name(), 'bits': key.get_bits(),
                             'fingerprint': fingerprint}}
    return dict()


def format_key_fingerprint(key):
    """Shared Helper method to format the SSH Fingerprint"""
    hex_fp = str(hexlify(key.get_fingerprint()))
    fp = ':'.join(hex_fp[i:i + 2] for i in range(0, len(hex_fp), 2))
    return fp


@lockutils.synchronized('registrar_privkeys', 'powervc-')
def store_key_file(key_file, key_data, key_prepend):
    """Shared Helper Method to make our own copy of the Private Key"""
    owner = getpass.getuser()
    key_dir = key_file[:key_file.rfind('/')]
    # If the RSA Keys Directory doesn't exist, we will need to create it
    if file_exists(key_dir) is False:
        # use relative path on cmds for rootwrap's sake
        # TODO: have install set the parent directory's permissions to allow
        # this service to modify itself
        run_command(['mkdir', '-m', '770', key_dir])
        run_command(['chown', owner + ':' + owner, key_dir])
        # now that we own the dir, we can further restrict the permissions
        run_command(['/usr/bin/chmod', '750', key_dir], run_as_root=False)
    # Write out the Key Data to the file in the Private Key Directory
    with open(key_file, 'w') as f:
        f.write(key_data)
    try:
        # Use Paramiko to confirm it is a valid Private Key format
        rsaKey = paramiko.RSAKey.from_private_key_file(key_file)
        rsaKey.write_private_key_file(key_file)
        # group read permission will allow validators to access
        run_command(['/usr/bin/chmod', '640', key_file], run_as_root=False)
    except SSHException:
        # If we got an exception, remove the copied file and throw it
        try:
            os.remove(key_file)
        except OSError:
            pass
        raise


def set_openstack_config_value(conf_file, key, val, sect='DEFAULT',
                               run_as_root=False):
    """Shared Helper Method to update a Value in an OpenStack Config File"""
    key, sect = parse_section_from_key(key, sect)
    # If they didn't give us a value then we just want to remove the key
    if not val:
        del_openstack_config_value(
            conf_file, key, sect=sect, run_as_root=run_as_root)
    # If it was an actual value, then we can run the command to update
    else:
        cmd = '/usr/bin/openstack-config'
        if run_as_root:
            # use relative path on cmd for rootwrap's sake
            cmd = 'openstack-config'
        # Escape the special interpolate configuration characters
        val = escape_config_value(val)
        run_command([cmd, "--set", conf_file, sect, key, val],
                    run_as_root=run_as_root)


def del_openstack_config_value(conf_file, key, sect='DEFAULT',
                               run_as_root=False):
    """Shared Helper Method to delete a Key from an OpenStack Config File"""
    key, sect = parse_section_from_key(key, sect)
    cmd = '/usr/bin/openstack-config'
    if run_as_root:
        # use relative path on cmd for rootwrap's sake
        cmd = 'openstack-config'
    run_command([cmd, "--del", conf_file, sect, key],
                run_as_root=run_as_root)


def parse_section_from_key(key, sect):
    """Internal Helper Method to parse the Section from the Key name"""
    if key and key.find(':') > 0:
        key, sect = (key[key.find(':') + 1:], key[0:key.find(':')])
    return key, sect


def escape_config_value(val):
    """Helper method to escape the special interpolate config characters"""
    return val.replace('$', '\$') if isinstance(val, basestring) else val


def unescape_config_value(val):
    """Helper method to un-escape the special interpolate config characters"""
    return val.replace('\$', '$') if isinstance(val, basestring) else val


@lockutils.synchronized('host_inprogress', 'powervc-')
def add_host_inprogress(hostname, operation='add'):
    """Shared Helper Method to tag a host as inprogress for being added"""
    filename = '/tmp/powervc_host_inprogress_%s_%s' % (operation, hostname)
    if not file_exists(filename):
        run_command('/bin/touch ' + filename, True, False)
        return False
    return True


@lockutils.synchronized('host_inprogress', 'powervc-')
def remove_host_inprogress(hostname, operation='add'):
    """Shared Helper Method to tag a host as inprogress for being added"""
    filename = '/tmp/powervc_host_inprogress_%s_%s' % (operation, hostname)
    run_command('/bin/rm -f ' + filename, True, False)


@lockutils.synchronized('host_inprogress', 'powervc-')
def remove_all_hosts_inprogress():
    """Shared Helper Method to tag a host as inprogress for being added"""
    commands.getoutput('/bin/rm -f /tmp/powervc_host_inprogress_*')


@lockutils.synchronized('host_inprogress', 'powervc-')
def is_host_inprogress(hostname, operation='add'):
    """Shared Helper Method to tag a host as inprogress for being added"""
    filename = '/tmp/powervc_host_inprogress_%s_%s' % (operation, hostname)
    return file_exists(filename)


def host_inprogress_decorator(operation='add'):
    """Decorator to check if the operation on the host is in-progress"""
    def wrap(f):
        def inner(*args, **kwargs):
            hostname = args[2]
            # First check if the host is already in-progress or not
            if is_host_inprogress(hostname, operation):
                msgb = msg.BaseRegistrarMessageBundle()
                msgid = msg.HOST_ADD_INPROGRESS if \
                    operation == 'add' else msg.HOST_DEL_INPROGRESS
                text = msgb.get_message_text(None, msgid, hostname=hostname)
                raise webob.exc.HTTPBadRequest(explanation=text)
            # Now tag the Host as in-progress to not allow other operations
            try:
                add_host_inprogress(hostname, operation)
                return f(*args, **kwargs)
            # Cleanup the file to allow other operations to progress now
            finally:
                remove_host_inprogress(hostname, operation)
        return inner
    return wrap


def _get_local_system_ip():
    """Shared Helper Method to retrieve the IP Address for the Local System"""
    global LOCAL_IP
    # We will cache the IP once we get it so we don't take the hit every time
    if LOCAL_IP is None:
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/nova/nova.conf')
        # Let the customer override by setting the my_ip in the nova.conf file
        if parser.has_option('DEFAULT', 'my_ip'):
            LOCAL_IP = parser.get('DEFAULT', 'my_ip')
        else:
            # Fall-back to querying the IP Address through the IfConfig Command
            cmd = "/sbin/ifconfig | grep -Eo 'inet (addr:)?([0-9]*\.){3}[0-" \
                  "9]*' | grep -Eo '([0-9]*\.){3}[0-9]*' | grep -v '127.0.'"
            # Loop through the IP Addresses, picking the first non-local one
            for ip in commands.getoutput(cmd).split('\n'):
                if not ip.startswith("127."):
                    LOCAL_IP = ip
                    break
    return LOCAL_IP


def get_local_system_ip(context):
    """Shared Helper Method to retrieve the IP Address for the Local System"""
    local_ip = None
    if context.service_catalog:
        # If the Service Catalog is in the Context, used that Hostname/Address
        catalog = context.service_catalog[0]
        # Make sure the Service Catalog is formatting correctly to parse
        if 'endpoints' in catalog and catalog['endpoints']:
            if catalog['endpoints'][0].get('adminURL'):
                url = catalog['endpoints'][0]['adminURL']
                url = url[url.find('//') + 2:]
                # If has the brackets for IPv6, need to look for closing
                if url.startswith('['):
                    local_ip = url[:url.find(']') + 1]
                # Otherwise we can look for the colon to strip the port
                else:
                    local_ip = url[:url.find(':')]
    if not local_ip:
        local_ip = _get_local_system_ip()
    # If this is an IPv6 Address we need to add brackets around it
    if local_ip.find(':') >= 0 and not local_ip.startswith('['):
        local_ip = '[' + local_ip + ']'
    return local_ip


def formulate_hostname(access_ip):
    """Shared Helper Method to formulate a hostname based on the IP"""
    hostname = access_ip
    default_timeout = socket.getdefaulttimeout()
    try:
        # Do a getnameinfo to see if we can resolve the IP's hostname
        socket.setdefaulttimeout(5)
        host_info = socket.getnameinfo((access_ip, 80), 0)
        hostname = host_info[0]
    except (socket.herror, socket.timeout, socket.gaierror):
        pass
    # If they weren't ascii characters, throw an exception
    except UnicodeEncodeError:
        msgb = msg.BaseRegistrarMessageBundle()
        text = msgb.get_message_text(
            None, msg.INVALID_ATTRIBUTE_VALUE, attr='access_ip')
        raise webob.exc.HTTPBadRequest(explanation=text)
    # Reset the Default Timeout back to its original value
    socket.setdefaulttimeout(default_timeout)
    try:
        # If this is an IP Address, then we want to replace . and : with _
        socket.inet_aton(hostname)
        hostname = re.sub('[:.]', '_', hostname)
    except socket.error:
        # If a hostname was already specified, try to return the short name
        if hostname.find('.') > 0:
            hostname = hostname[:hostname.find('.')]
    # Make sure the hostname isn't too long and doesn't have invalid characters
    hostname = hostname.replace('-', '_')[:128]
    return re.sub('[^A-Za-z0-9_]+', 'x', hostname)


def is_valid_hostname(hostname):
    """Shared Helper Method to verify a valid Hostname or IP Address"""
    return re.compile('^[A-Za-z0-9_-]+$').match(hostname) is not None


def ssh_reg_messages():
    """Provides a list of Mappings between Log Errors and User Messages"""
    return [
        (msg.AUTHENTICATION_ERROR_SSH, ['authorization has failed',
                                        'authentication failed',
                                        'bad authentication']),
        (msg.MISMATCH_HOST_KEY, ['does not match!',
                                 'ssh: invalid key', 'badhostkey']),
        (msg.UNKNOWN_HOST_KEY, ['not found in known_hosts']),
        (msg.CONNECTION_NOROUTE, ['ehostunreach', 'enetunreach']),
        (msg.CONNECTION_TIMEOUT_SSH, ['timedout', 'timed out']),
        (msg.CONNECTION_FIREWALL, ['error reading ssh protocol banner']),
        (msg.CONNECTION_NOTKNOWN, [
            'label empty or too long',
            'name or service not known',
            'no address associated with hostname']),
        (msg.CONNECTION_FAILED_SSH, [
            'error connecting', 'connectionfailed', 'connrefused'])
    ]


def format_isotime(timestamp):
    """Utility for formating a time-stamp to ISO 8601 format"""
    return timeutils.isotime(timestamp)


def compare_version(new_version, old_version):
    """Utility to compare 2 versions and see if the first is newer"""
    # If the old system isn't installed, the new system is greater if installed
    if not old_version:
        return 1 if new_version else 0
    # If the new system isn't installed, the old is newer since it must be set
    if not new_version:
        return -1
    # We don't deal with the dash at the end of the version, so strip it off
    dasho, dashn = old_version.find('-'), new_version.find('-')
    old_version = old_version[:dasho] if dasho > 0 else old_version
    new_version = new_version[:dashn] if dashn > 0 else new_version
    # We need to break apart the version to individual digits for comparison
    old_split = old_version.split('.')
    new_split = new_version.split('.')
    max_digits = max(len(new_split), len(old_split))
    # Loop through each of the digits and do the comparison
    for index in range(max_digits):
        old_num = int(old_split[index]) if len(old_split) > index else 0
        new_num = int(new_split[index]) if len(new_split) > index else 0
        # If a digit is different, then return since 1 is greater than the
        # other
        if old_num != new_num:
            return 1 if new_num > old_num else -1
    # If all digits were the same, the 2 versions must be equal
    return 0


def is_sub_path(full_path, sub_path):
    """Utility to see if a Partial Path is part of another Full Path"""
    if sub_path is None or full_path is None:
        return False
    # If the Paths don't start with a /, make them an absolute path
    if not full_path.startswith('/'):
        full_path = '/' + full_path
    if not sub_path.startswith('/'):
        sub_path = '/' + sub_path
    # If the Paths end with a /, remove it since it isn't needed to compare
    if full_path.endswith('/'):
        full_path = full_path[:len(full_path) - 1]
    if sub_path.endswith('/'):
        sub_path = sub_path[:len(sub_path) - 1]
    # Break the Paths into their directory components
    full_path = full_path.split('/')
    sub_path = sub_path.split('/')
    # If the full path isn't part of the sub-path, then it is not true
    if len(full_path) < len(sub_path):
        return False
    # Loop through each of the parts, making sure each one is the same
    for index in range(len(sub_path)):
        if sub_path[index] != full_path[index]:
            return False
    return True


def get_property_value(output, key, default=None):
    """Utility to get the property value for a key from a command output"""
    val = default
    # Since the output might have multiple lines, look at each line separate
    for line in output.strip().split('\n'):
        line = line.strip()
        # If this line is commented out than we want to skip it
        if len(line) > 0 and not line.startswith('#'):
            parts = line.split('=')
            # If this is a property line and the value is set, use it
            if len(parts) > 1 and len(parts[1].strip()) > 0:
                if parts[0].strip() == key:
                    val = parts[1].strip()
    return val


def has_value_changed(config, new_vals, key):
    """Shared Helper Method to determine one of the properties changed"""
    # We can only compare key values if they are strings
    if isinstance(key, bool) or isinstance(key, list):
        return False
    old_val = config.get(key)
    new_val = new_vals.get(key)
    # It is considered changed if one side is set but the other side isn't
    if key not in new_vals:
        return False
    # Since we copy the Private Key, so we consider it changed if it is set
    if key == 'private_key' or key == 'private_key_data':
        return True
    return old_val != new_val


def trim_config_values(config):
    """Shared Helper Method to strip the spaces from the access ip"""
    if config.get('access_ip') is not None:
        config['access_ip'] = str(config['access_ip']).strip()


def format_string_list(alist):
    """Shared Helper method to format a String List for display"""
    astr = ''
    # If isn't a List, then just convert it to a String (shouldn't happen)
    if not isinstance(alist, list):
        return str(alist)
    # If it is a List, then convert it to a comma-separate list of strings
    for anitem in alist:
        astr += str(anitem) + ', '
    return astr[:len(astr) - 2] if len(astr) > 0 else astr


def bool_from_string(subject, strict=True):
    """Utility for converting a boolean to a string"""
    return strutils.bool_from_string(subject, strict)


########################################################
# ############  Internal Utility Methods  ##############
########################################################
def _get_ssh_connection(config):
    """Internal Helper Method to Establish a connection to SSH on a System"""
    ssh = paramiko.SSHClient()
    ssh.load_host_keys(SSH_KNOWN_HOST_FILE)
    ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
    port = int(config.get('ssh_port', SSH_PORT))
    # If they specified a Password, use that to try to connect to the SVC
    if config.get('password'):
        pwd = decode_password(config['password'])
        ssh.connect(config['access_ip'], port=port,
                    username=config['user_id'],
                    password=pwd, timeout=SSH_TIMEOUT)
    # If they specified a Private Key, use that to try to connect to the SVC
    elif config.get('private_key'):
        pkfile = config['private_key']
        privatekey = paramiko.RSAKey.from_private_key_file(pkfile)
        ssh.connect(config['access_ip'], port=port,
                    username=config['user_id'], pkey=privatekey,
                    timeout=SSH_TIMEOUT)
    else:
        raise Exception('Expect Password/PrivateKey to be set on Connect')
    # Now we can return the SSH Session to the caller
    return ssh


def _ssh_execute(ssh, cmd, process_input=None, check_exit_code=True):
    """Internal Helper method to execute an SSH command with standard input"""
    nopwd_cmd = strutils.mask_password(cmd)
    LOG.debug('Running an SSH command: %s', nopwd_cmd)
    # Execute the SSH Command getting the standard out/in/err back
    stdin_stream, stdout_stream, stderr_stream = ssh.exec_command(cmd)
    # If they gave us something for standard input, then write it out
    if process_input:
        stdin_stream.write(process_input + '\n')
        stdin_stream.flush()
    stdin_stream.close()
    # Now we can read the standard output/error streams and check exit code
    stdout, stderr = stdout_stream.read(), stderr_stream.read()
    exit_status = stdout_stream.channel.recv_exit_status()
    # Strip passwords from the output so they don't end up in the logs
    stdout = strutils.mask_password(stdout)
    stderr = strutils.mask_password(stderr)
    # If they asked to check the exit code and it isn't 0, throw an exception
    if check_exit_code and exit_status != 0:
        raise procutils.ProcessExecutionError(
            exit_code=exit_status, stdout=stdout, stderr=stderr, cmd=nopwd_cmd)
    return (stdout, stderr)


def _store_certificate(certificate, address):
    """Retrieves and stores an untrusted but user accepted certificate"""
    with open(get_certificate_file(address), 'w') as f:
        f.write(certificate)


def _create_known_hosts():
    """Internal Helper Method to create the known_hosts file"""
    with open(SSH_KNOWN_HOST_FILE, 'w') as f:
        f.write('')
    chmod(SSH_KNOWN_HOST_FILE, 0660)
    group = grp.getgrnam('pvcservices')
    chown(SSH_KNOWN_HOST_FILE, -1, group.gr_gid)


def _check_certificate_expiration(rawcert):
    """Internal Helper Method to see if the Certificate has expired"""
    cert_dict = certificate_to_dict(rawcert)
    not_after = cert_dict['certificate']['not_after']
    expire_date = datetime.strptime(not_after, '%Y-%m-%dT%H:%M:%S')
    # If the Expiration Date is before the current date, say it has expired
    if datetime.utcnow() >= expire_date:
        raise Exception('The SSL Certificate has expired')


def _get_cert_common_name(tbs, name):
    """Internal Helper Method to get the Common Name for a Certificate"""
    subject = tbs.getComponentByName(name)
    for rdnss in subject:
        for rdns in rdnss:
            for name in rdns:
                oid = name.getComponentByName('type')
                value = name.getComponentByName('value')
                if oid != rfc2459.id_at_commonName:
                    continue
                value = ber_decoder.decode(str(value))[0]
                return "" if value is None else str(value)
    return ""


def _get_cert_component_string(tbs, name):
    """Internal Helper Method to get the Component String for a Certificate"""
    pieces, compstr = (list(), '')
    oid_map = {'2.5.4.3': 'CN', '2.5.4.6': 'C', '2.5.4.7': 'L',
               '2.5.4.8': 'ST', '2.5.4.10': 'O', '2.5.4.11': 'OU',
               '2.5.4.4': 'SN', '2.5.4.42': 'GN', '2.5.4.43': 'initials',
               '2.5.4.12': 'title', '2.5.4.46': 'dnQualifier'}
    subject = tbs.getComponentByName(name)
    # Loop through the pieces of the component parsing out the oid/value
    for rdnss in subject:
        for rdns in rdnss:
            for name in rdns:
                oid = name.getComponentByName('type')
                value = name.getComponentByName('value')
                value = ber_decoder.decode(str(value))[0]
                oid = '.'.join(map(str, oid._value))
                pieces.append([str(oid), str(value)])
    # Return the string in reverse order of how it was placed in the list
    for piece in reversed(pieces):
        if oid_map.get(piece[0]):
            compstr += ',' + oid_map[piece[0]] + '=' + piece[1]
    return compstr[1:] if compstr else compstr


def _get_subject_alt_names(rawcert):
    names = []
    dercert = ssl.PEM_cert_to_DER_cert(rawcert)
    cert = der_decoder.decode(dercert, asn1Spec=rfc2459.Certificate())[0]
    tbs = cert.getComponentByName('tbsCertificate')
    extensions = tbs.getComponentByName('extensions')
    for ext in extensions:
        if ext.getComponentByName('extnID') == (2, 5, 29, 17):
            octet_string = der_decoder.decode(
                ext.getComponentByName('extnValue').asOctets())[0]
            gen_names = der_decoder.decode(
                octet_string, rfc2459.GeneralNames())[0]
            for name in gen_names:
                t = NAME_TYPES_MAP[name.getName()]
                v = six.text_type(name.getComponent())
                names.append((t, v))
    return tuple(iter(names)) if names else None


########################################################
# ############  Utility Exception Classes  #############
########################################################
class HostKeyPolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        raise UnknownHostKeyException(hostname, key)


class UnknownHostKeyException(SSHException):
    def __init__(self, hostname, key):
        SSHException.__init__(
            self, 'Server %r not found in known_hosts' % hostname)
        self.hostname = hostname
        self.key = key
