# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

""" DS8K Registrar Class for performing Storage Registration"""

import copy
import base64
import requests
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar

__import__('powervc_oslo.storage.ds8k')
LOG = logging.getLogger(__name__)


# #######################################################
# ##########  DSK8K Registrar Implementation  ###########
# #######################################################
class DS8KStorageRegistrar(storage_registrar.StorageRegistrar):
    """DS8K Registrar class to handle storage registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the DS8K class for storage registration"""
        super(DS8KStorageRegistrar, self).__init__(context, **kwargs)
        self._token = None

    ###############################################
    ###  Implementing BaseRegistrar Interface   ###
    ###############################################
    def get_required_attributes(self):
        """ Provides the list of required attributes for this Registrar"""
        attr = ['host_type', 'access_ip', 'user_id', 'password',
                'volume_pool_name']
        return attr

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attr = super(DS8KStorageRegistrar, self).get_sensitive_attributes()
        attr.append('storage_unit_name')
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_DS8K)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['unauthorized']),
            (msg.CONNECTION_TIMEOUT, ['timed out', 'timeout exceeded']),
            (msg.CONNECTION_NOTKNOWN, ['name or service not known']),
            (msg.CONNECTION_NOROUTE, ['enetunreach', 'ehostunreach']),
            (msg.CONNECTION_FAILED, ['econnrefused']),
            (msg.INVALID_CERTIFICATE, ['certificate verify', "doesn't match"]),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired'])]

    def _get_service_param_mapping(self):
        """Provides a list of mappings between log errors and user messages"""
        mapping = super(DS8KStorageRegistrar,
                        self)._get_service_param_mapping()
        mapping['san_ip'] = 'access_ip'
        mapping['san_login'] = 'user_id'
        mapping['san_password'] = 'password'
        mapping['san_clustername'] = 'volume_pool_name'
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        return mapping

    def _update_service_param_mapping(self, param_mapping, connect_info):
        """Provides any additional properties to set by this Registrar"""
        super(DS8KStorageRegistrar, self).\
            _update_service_param_mapping(param_mapping, connect_info)
        param_mapping['volume_driver'] = \
            'powervc_cinder.volume.drivers.ds8k.ds8k_product.DS8KProductDriver'
        param_mapping['xiv_ds8k_proxy'] = \
            'powervc_cinder.volume.drivers.ds8k.ds8k_product.DS8KProxy'
        param_mapping['xiv_ds8k_connection_type'] = 'fibre_channel'
        param_mapping['xiv_ds8k_storage_array_type'] = 'ds8k'
        # They only want the ID portion of the Pool Name in the property
        index = param_mapping['san_clustername'].rfind('_')
        if index > 0:
            param_mapping['san_clustername'] = \
                'P' + param_mapping['san_clustername'][index+1:]

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        super(DS8KStorageRegistrar, self).\
            _retrieve_connection_info(connect_info, config)

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 8452

    def _decode_password(self, password):
        """Shared Helper Method to specify the deCryption method to use"""
        enc_pwd = super(DS8KStorageRegistrar, self)._decode_password(password)
        return base64.b64decode(enc_pwd)

    def _encode_password(self, password):
        """Shared Helper Method to specify the encryption method to use"""
        enc_pwd = base64.b64encode(password)
        return super(DS8KStorageRegistrar, self)._encode_password(enc_pwd)

    #############################################
    ## Implementing StorageRegistrar Interface ##
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_certificate(config, self._get_certificate_port())

    def _get_unique_identifier(self, config):
        """Retrieves the unique identifier for the given system."""
        resp = self.__rest_request(config, 'GET', 'systems')['data']['systems']
        return resp[0]['MTM'] + '-' + resp[0]['sn']

    def _get_storage_unit_name(self, config):
        """Retrieves the unique identifier for the given system."""
        resp = self.__rest_request(config, 'GET', 'systems')['data']['systems']
        return resp[0]['name']

    def _get_volume_pools(self, config):
        """Retrieves the volume pools for the specified system"""
        pools = list()
        data = self.__rest_request(config, 'GET', 'pools')['data']['pools']
        # Loop through the Pools returns getting the key info out
        for pool in data:
            volume_pool = dict(pool_name=(pool['name'] + '_' + pool['id'][1:]))
            volume_pool['pool_total_capacity_gb'] = \
                float(pool['cap']) / (1024 ** 3)
            volume_pool['pool_free_capacity_gb'] = \
                float(pool['capavail']) / (1024 ** 3)
            pools.append(volume_pool)
        return pools

    def _verify_volume_pool_exists(self, pool_name, volume_pools):
        """Internal Helper Method to verify that the specified Pool exists"""
        try:
            super(DS8KStorageRegistrar, self).\
                _verify_volume_pool_exists(pool_name, volume_pools)
        # If the pool isn't valid, see if it the P<id> format instead
        except Exception:
            super(DS8KStorageRegistrar, self)._verify_volume_pool_exists(
                pool_name, self.__convert_pool_name_to_id(volume_pools))

    ###############################################
    ###  Internal Registration Helper Methods   ###
    ###############################################
    def __rest_request(self, config, method, rsctype, data=None):
        """Helper method to convert the IP/User/Password for XIV"""
        session = requests.Session()
        # If we haven't yet gotten the token to use, retrieve it now
        if self._token is None and rsctype != 'tokens':
            self._token = self.__get_rest_token(config)
        # Construct the URI, Header, and Certificate Path to use
        header = {'X-Auth-Token': self._token}
        uri = 'httpsds8k://' + config['access_ip'] + ':8452/api/v1/'
        # Setup the adapter to use our added URLLIB3 Connection Pool
        session.mount("httpsds8k://", requests.adapters.HTTPAdapter())
        session.verify = self.__get_rest_certificate(config['access_ip'])
        try:
            # Now we can actually make the REST GET Request to the System
            resp = session.request(
                method, uri + rsctype, data=data, headers=header)
            # If we got an error making the request, parse the reason
            if resp.status_code != 200:
                raise Exception('%s - %s' % (resp.status_code, resp.reason))
            # They may also return error information in the server-status
            if resp.json()['server']['status'] != 'ok':
                svr = resp.json()['server']
                raise Exception('%s - %s' % (svr['code'], svr['message']))
            return resp.json()
        finally:
            session.close()

    def __get_rest_certificate(self, access_ip):
        """Helper method to get the certificate file to use for rest"""
        cert_file = utils.get_certificate_file(access_ip)
        #Use the certificate if it exists, otherwise use the System CA Bundle
        return cert_file if utils.file_exists(cert_file) else True

    def __get_rest_token(self, config):
        """Helper method to get the Token to use on Rest Requests"""
        data = '{"request": {"params": {"username": "%s", '\
               '"password": "%s"}}}' % (config['user_id'], config['password'])
        # Make the REST POST Request to get an Authentication Token
        resp = self.__rest_request(config, 'POST', 'tokens', data)
        return resp['token']['token']

    def __convert_pool_name_to_id(self, volume_pools):
        """Helper method to parse the Pool ID out of the Pool Info"""
        volume_pools = copy.deepcopy(volume_pools)
        # Loop through the Pools converting the Name to use the ID
        for pool in volume_pools:
            index = pool['pool_name'].rfind('_')
            pool['pool_name'] = 'P' + pool['pool_name'][index+1:]
        return volume_pools


########################################################
##########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return DS8KStorageRegistrar(context, **kwargs)
