# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""EMC Registrar class for performing Storage Registration"""

from xml.dom.minidom import parseString
from xml.dom.minidom import getDOMImplementation
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
try:
    from powervc_oslo.wbem import pywbem
except ImportError as ie:
    LOG.info(_('Module PyWBEM not installed.  '
               'Install PyWBEM using the python-pywbem package.'))
    raise ie


########################################################
# ########  EMC VMAX Registrar Implementation  #########
########################################################
class EMC_VMAXStorageRegistrar(storage_registrar.StorageRegistrar):
    """EMC Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the EMC Class for Storage Registration"""
        super(EMC_VMAXStorageRegistrar, self).__init__(context, **kwargs)
        self._wbem_connection = None
        # If there wasn't a WBEM Port specified we will default to 5989
        if self.config.get('wbem_port') is None:
            self.config['wbem_port'] = '5989'

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['access_ip', 'user_id', 'password', 'volume_pool_name']

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attr = super(EMC_VMAXStorageRegistrar, self).get_sensitive_attributes()
        attr.extend(['port_groups', 'portGroups'])
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_EMC_VMAX)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['authorization']),
            (msg.CONNECTION_TIMEOUT, ['timedout', 'timed out']),
            (msg.CONNECTION_NOTKNOWN, ['name or service not known']),
            (msg.CONNECTION_FAILED, ['connection failed']),
            (msg.CONNECTION_VMAXPORT, ['hello:unknown protocol']),
            (msg.INVALID_SYSTEM, ['not a valid classname']),
            (msg.INVALID_SYSTEM_VMAX, ['not a vmax system']),
            (msg.VMAX_NOT_MANAGED, ['storage array not managed']),
            (msg.INVALID_CERTIFICATE, ['certificate verify', "doesn't match"]),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired'])
        ]

    def _generate_name_value(self):
        """Generates the Default Value to use for the Name Attribute"""
        arrayid = self.config.get('storage_array_id')
        # We will tolerate if they put the SYMMETRIX+ on the front
        if arrayid and arrayid.find('+') > 0:
            arrayid = arrayid[arrayid.find('+') + 1:]
        return arrayid.lower() if arrayid else arrayid

    def _get_private_key_prepend(self):
        """Provides the Directory and Prefix for the Private Key Location"""
        return None

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return int(self.config['wbem_port'])

    def _obfuscate_password(self, config, decrypt=False):
        """Shared Helper Method to Encrypt the Passwords Provided"""
        pass

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        super(EMC_VMAXStorageRegistrar,
              self)._retrieve_connection_info(connect_info, config)
        # Verify that the Storage Array is really a VMAX Array
        self.__verify_vmax_storage(config)
        # Query the Port Groups so we can stick them in the CONF file
        connect_info['portGroups'] = self.__get_port_groups(config)
        # Query the Storage Arrays to return the list to the user
        connect_info['storageArrays'] = self.__get_storage_arrays(config)
        # If they gave the Array ID, they don't need the Arrays returned
        if config.get('storage_array_id'):
            connect_info.pop('storageArrays', None)
        LOG.debug("VMAX connection info: %s " % connect_info)

    def register(self):
        """Registrar Method to update the System Registration"""
        # If the Storage Array ID wasn't sent on the Create, throw exception
        if not self.config.get('storage_array_id'):
            self._throw_bad_request(
                msg.MISSING_ATTRIBUTE, attr='storage_array_id')
        # Call the super class to do the actual registration
        super(EMC_VMAXStorageRegistrar, self).register()

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(
            EMC_VMAXStorageRegistrar, self)._get_service_param_mapping()
        mapping['default_volume_type'] = 'default_volume_type'
        mapping['volume_backend_name'] = 'storage_array_id'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(EMC_VMAXStorageRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['cinder_emc_config_file'] = \
            '/etc/cinder/cinder_emc_config_%s.xml' % self.config['host_name']
        parm_mapping['volume_driver'] = 'powervc_cinder.volume.drivers.emc.'\
                                        'emc_vmax_product.EMCVMAXProductDriver'

    def _post_service_registration_hook(self, action,
                                        connect_info=None, config={}):
        """Hook to be able to take actions after doing the registration"""
        super(EMC_VMAXStorageRegistrar, self).\
            _post_service_registration_hook(action, connect_info, config)
        LOG.debug("_post_service_registration_hook(): action=%s, config=%s."
                  % (action, config))
        # Write out the EMC SMI-S Provider XML Configuration File on Register
        if action == 'register':
            if not self.config.get('port_groups'):
                port_groups = ','.join(connect_info.get('portGroups'))
                self.config['port_groups'] = port_groups
            self.__write_emc_xml_config(self.config)
        # Update the EMC SMI-S Provider XML Configuration File on Re-Register
        if action == 'update-registration':
            new_config = dict(self.config, **config)
            self.__write_emc_xml_config(new_config)
        # Remove the EMC SMI-S Provider XML Configuration File on De-Register
        if action == 'deregister':
            host = self.config['host_name']
            emc_xml_file = '/etc/cinder/cinder_emc_config_' + host + '.xml'
            utils.run_command('rm -f ' + emc_xml_file, True)

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        super(EMC_VMAXStorageRegistrar, self)._parse_service_config(config)
        # Parse the Registration Info out of the EMC Provider XML Config File
        self.__parse_emc_xml_config(config.get('cinder_emc_config_file'))

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_certificate(config, self._get_certificate_port())

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        array_id = config.get('storage_array_id')
        if array_id:
            if '-+-' in array_id:
                sep = '-+-'  # SMI-S v8.0 separator
            elif '+' in array_id:
                sep = '+'  # old SMI-S separator
            else:
                return array_id
            array_id = array_id[array_id.find(sep)+len(sep):]
        return array_id

    def _get_display_name(self, config):
        """Retrieves the Display Name for the given System"""
        array_id = unicode(config.get('storage_array_id'))
        nametxt = _("EMC VMAX storage %s") % array_id
        return discovery_locale.translate_message(self.context, nametxt)

    def _get_volume_pools(self, config):
        """Retrieves the Volume Pools for the specified System"""
        volume_pools = list()
        # Get a WBEM Connection to the SMIS Proxy Provider
        conn = self.__get_wbem_connection(config)
        # Get the Unique Identifier to match against the Controller
        unique_id = self._get_unique_identifier(config)
        # First check V3 pools
        pools = None
        try:
            pools = conn.EnumerateInstances('Symm_SRPStoragePool')
        except Exception:
            LOG.debug('Symm_SRPStoragePool class likely does not exist so '
                      'assuming V2 generation.')
        if not pools:
            # Use the Symm_VirtualProvisioningPool CIM Class to get the Pools
            pools = conn.EnumerateInstances('Symm_VirtualProvisioningPool')
        for pool in pools:
            array_id = self.__parse_array_id(pool['InstanceID'])
            # See if the Storage Pool is for this Array (or wasn't specified)
            if not unique_id or unique_id == array_id:
                volume_pool = dict(pool_name=pool['PoolID'])
                # free space is computed using subscribed capacity per
                # Kilo support limitation of cinder driver. In Liberty
                # we anticipate being able to switch to raw free capacity
                # here to match expected new driver support.
                volume_pool['pool_total_capacity_gb'] = \
                    self.__format_capacity(pool['TotalManagedSpace'])
                unsubscribed = (int(pool['TotalManagedSpace']) -
                                int(pool['EMCSubscribedCapacity']))
                # SMI-S v8.0 may not return these pool properties as expected.
                # In that case, we need to match the logic that the PowerVC
                # driver extension uses to report the free capacity based on
                # RemainingManagedSpace instead.
                if int(pool['RemainingManagedSpace']) < unsubscribed:
                    unsubscribed = int(pool['RemainingManagedSpace'])
                volume_pool['pool_free_capacity_gb'] = self.__format_capacity(
                    unsubscribed)
                if not unique_id:
                    volume_pool['pool_storage_array_id'] = array_id
                volume_pools.append(volume_pool)
        LOG.debug("VMAX pool registration info: %s. " % volume_pools)
        return volume_pools

    ###############################################
    # ### Internal Registration Helper Methods  ###
    ###############################################
    def __get_emc_config_mapping(self):
        """Internal Helper Method to Write out the EMC XML Config File"""
        return dict(access_ip='EcomServerIp', user_id='EcomUserName',
                    password='EcomPassword', wbem_port='EcomServerPort',
                    volume_pool_name='Pool', storage_array_id='Array')

    def __write_emc_xml_config(self, config):
        """Internal Helper Method to Write out the EMC XML Config File"""
        # Create a XML DOM Object with the given EMC XML Attributes
        dom_impl = getDOMImplementation()
        xml_doc = dom_impl.createDocument(None, "EMC", None)
        for key, val in self.__get_emc_config_mapping().iteritems():
            node = xml_doc.createElement(val)
            node.appendChild(xml_doc.createTextNode(str(config.get(key, ''))))
            xml_doc.documentElement.appendChild(node)
        # Handle EMC port groups special since they are nested
        if config['port_groups']:
            node = xml_doc.createElement('PortGroups')
            for pg in config['port_groups'].split(','):
                child = xml_doc.createElement('PortGroup')
                child.appendChild(xml_doc.createTextNode(pg))
                node.appendChild(child)
            xml_doc.documentElement.appendChild(node)
        # Add in the Properties for the Certificate and to use SSL
        cert_file = utils.get_certificate_file(config['access_ip'])
        # 1.2.3 changes: Cert --> EComCACert, UseSSL --> EcomUseSSL
        # The PowerVC driver extension supports both so no upgrade needed
        cert_node = xml_doc.createElement('EcomCACert')
        ssl_node = xml_doc.createElement('EcomUseSSL')
        noverify_node = xml_doc.createElement('EcomNoVerification')
        cert_node.appendChild(xml_doc.createTextNode(cert_file))
        ssl_node.appendChild(xml_doc.createTextNode('True'))
        noverify_node.appendChild(xml_doc.createTextNode('False'))
        xml_doc.documentElement.appendChild(cert_node)
        xml_doc.documentElement.appendChild(ssl_node)
        xml_doc.documentElement.appendChild(noverify_node)
        # Add in retry config properties
        interval = xml_doc.createElement('Interval')
        retries = xml_doc.createElement('Retries')
        interval.appendChild(xml_doc.createTextNode('5'))
        retries.appendChild(xml_doc.createTextNode('720'))
        xml_doc.documentElement.appendChild(interval)
        xml_doc.documentElement.appendChild(retries)
        # Write the XML out to the EMC XML Configuration File
        postfix = utils.get_unique_file_postfix()
        emc_xml_file = 'cinder_emc_config_' + self.config['host_name'] + '.xml'
        temp_xml_file = '/tmp/' + emc_xml_file + postfix
        LOG.debug("emc_xml_file=%s, temp_xml_file=%s." %
                  (emc_xml_file, temp_xml_file))
        xml_data = xml_doc.toprettyxml(indent='', newl='', encoding='UTF-8')
        utils.write_to_file(temp_xml_file, xml_data)
        # Copy the Temporary File to the Final Location and Cleanup the Temp
        utils.run_command('cp -f ' + temp_xml_file +
                          ' /etc/cinder/' + emc_xml_file)
        utils.run_command('chown cinder:cinder /etc/cinder/' + emc_xml_file)
        utils.run_command('chmod 640 /etc/cinder/' + emc_xml_file)
        utils.run_command('rm -f ' + temp_xml_file, True)

    def __parse_emc_xml_config(self, emc_xml_file):
        """Internal Helper Method to Write out the EMC XML Config File"""
        # Parse the XML out of the EMC XML Configuration File
        if utils.file_exists(emc_xml_file):
            xml_data = utils.read_from_file(emc_xml_file)
            xml_dom = parseString(xml_data)
            # Get the Key Attributes out of the EMC XML Configuration
            for key, val in self.__get_emc_config_mapping().iteritems():
                nodes = xml_dom.getElementsByTagName(val)
                if nodes and nodes.length > 0:
                    self.config[key] = \
                        nodes.item(0).firstChild.nodeValue.strip()
            # Handle EMC port groups special since they are nested
            nodes = xml_dom.getElementsByTagName('PortGroup')
            groups = [n.firstChild.nodeValue.strip() for n in nodes]
            self.config['port_groups'] = ','.join(groups)

    def __verify_vmax_storage(self, config):
        """Verify the ECOM managed array is a VMAX"""
        # Get a WBEM Connection to the SMIS Proxy Provider
        conn = self.__get_wbem_connection(config)
        # Use the Symm_VirtualProvisioningPool CIM Class to get the Pools
        products = conn.EnumerateInstances('EMC_ArrayProduct')
        for product in products:
            # This Product is VMAX, so we are good and can continue
            if product['Version'].lower().startswith('vmax'):
                return
        # It wasn't a VMAX System (or weren't any), so throw an exception
        raise Exception('Not a VMAX System')

    def __get_storage_arrays(self, config):
        """Retrieves the storage arrays for the specified ECOM service"""
        array_ids, storage_arrays = (list(), list())
        # Get a WBEM Connection to the SMIS Proxy Provider
        conn = self.__get_wbem_connection(config)
        # Get the Unique Identifier to match against the Controller
        unique_id = self._get_unique_identifier(config)
        # Use the Symm_VirtualProvisioningPool CIM Class to get the Pools
        arrays = conn.EnumerateInstances('Symm_StorageSystem')
        for array in arrays:
            if '-+-' in array['Name']:
                sep = '-+-'
            else:
                sep = '+'
            array_id = array['Name'][array['Name'].find(sep)+len(sep):]
            #Build a Dictionary with the Array ID and Description
            array_dict = dict(array_id=array_id,
                              array_name=array['Description'])
            storage_arrays.append(array_dict)
            array_ids.append(array_id)
        # If we were given an array, make sure it is one managed
        if unique_id and unique_id not in array_ids:
            raise Exception('Storage Array not managed')
        return storage_arrays

    def __get_port_groups(self, config):
        """Retrieves the Volume Pools for the specified System"""
        port_groups = list()
        # Get a WBEM Connection to the SMIS Proxy Provider
        conn = self.__get_wbem_connection(config)
        # Get the Unique Identifier to match against the Controller
        unique_id = self._get_unique_identifier(config)
        # Use the Symm_VirtualProvisioningPool CIM Class to get the Pools
        mask_groups = conn.EnumerateInstances('CIM_TargetMaskingGroup')
        for group in mask_groups:
            array_id = self.__parse_array_id(group['InstanceID'])
            # See if the Storage Pool is for this Array (or wasn't specified)
            if not unique_id or unique_id == array_id:
                port_groups.append(group['ElementName'])
        return port_groups

    def __get_wbem_connection(self, config):
        """Internal Helper Method to Get a WBEM Connection"""
        # We will cache the Connection, so we only get it once per REST call
        if self._wbem_connection is None:
            user, pwd = (config['user_id'], config['password'])
            cert = utils.get_certificate_file(config['access_ip'])
            url = 'https://%s:%s' % (config['access_ip'], config['wbem_port'])
            self._wbem_connection = pywbem.WBEMConnection(
                url, (user, pwd), default_namespace='root/emc', ca_certs=cert)
        return self._wbem_connection

    def __smis_proxy_registered(self, access_ip):
        """Internal Helper Method to see if the SMI-S Proxy is registered"""
        try:
            emc_xml_files = utils.get_matching_files(
                '/etc/cinder', 'cinder_emc_config_', '.xml')
            # Loop through each of the XML Files seeing if any are a match
            for emc_xml_file in emc_xml_files:
                xml_data = utils.read_from_file('/etc/cinder/' + emc_xml_file)
                xml_dom = parseString(xml_data)
                nodes = xml_dom.getElementsByTagName('EcomServerIp')
                # Parse out the EcomServerIp tag to get the compare IP
                if nodes and nodes.length > 0:
                    comp_ip = nodes.item(0).firstChild.nodeValue.strip()
                    # Only return True if there is a match of the IP Addresses
                    if comp_ip == access_ip:
                        return True
        # If there is an exception just log it and say that there isn't a match
        except Exception as exc:
            LOG.exception(exc)
        return False

    @staticmethod
    def __parse_array_id(inst_id):
        """Internal Helper Method to parse the Array Id from Instance Path"""
        # Try newer SMI-S separator first, then older.
        if '-+-' in inst_id:
            sep = '-+-'
        elif '+' in inst_id:
            sep = '+'
        else:
            return inst_id
        array_id = inst_id[inst_id.find(sep)+len(sep):]
        return array_id[:array_id.find(sep)]

    @staticmethod
    def __format_capacity(capstr):
        """Internal Helper Method to Round the Capacity to the nearest 0.5"""
        cap_gb = float(capstr) / (1024 ** 3)
        return float(round(cap_gb * 2) / 2)


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return EMC_VMAXStorageRegistrar(context, **kwargs)
