# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

""" EMC VNX Registrar Class for performing Storage Registration"""

import platform
from oslo_config import cfg
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.BoolOpt('support_ppc_naviseccli', default=False,
                help='Specifies whether it is supported on PowerPC')
])

LOG = logging.getLogger(__name__)
NAVCLI_CMD = '/opt/Navisphere/bin/naviseccli'


########################################################
# #########  EMC VNX Registrar Implementation  #########
########################################################
class EMC_VNXStorageRegistrar(storage_registrar.StorageRegistrar):
    """ VNX Registrar class to handle storage registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the VNX class for storage registration"""
        super(EMC_VNXStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """ Provides the list of required attributes for this Registrar"""
        return ['host_type', 'access_ip', 'user_id',
                'password|security_file', 'volume_pool_name']

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attr = super(EMC_VNXStorageRegistrar, self).get_sensitive_attributes()
        attr.append('security_file')
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_EMC_VNX)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['authentication failed']),
            (msg.CONNECTION_TIMEOUT, ['http request/response',
                                      'timed out', 'timeout exceeded']),
            (msg.INVALID_SYSTEM, ['terminate called after']),
            (msg.INVALID_CERTIFICATE, ['certificate verify', "doesn't match"]),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired']),
            (msg.NO_NAVISECCLI, ['naviseccli is not installed'])]

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 443

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides a list of mappings between log errors and user messages"""
        mapping = super(EMC_VNXStorageRegistrar, self).\
            _get_service_param_mapping()
        mapping['san_ip'] = 'access_ip'
        mapping['san_login'] = 'user_id'
        mapping['storage_vnx_pool_names'] = 'volume_pool_name'
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        mapping['storage_vnx_security_file_dir'] = 'security_file'
        return mapping

    def _update_service_param_mapping(self, param_mapping, connect_info):
        """Provides any additional properties to set by this Registrar"""
        super(EMC_VNXStorageRegistrar, self).\
            _update_service_param_mapping(param_mapping, connect_info)
        param_mapping['volume_driver'] = 'powervc_cinder.volume.drivers.emc.'\
                                         'emc_vnx_product.EMCVNXProductDriver'
        param_mapping['naviseccli_path'] = NAVCLI_CMD
        param_mapping['storage_vnx_security_file_dir'] = \
            self.__get_security_file_directory(connect_info)
        param_mapping['xiv_ds8k_connection_type'] = 'fibre_channel'
        param_mapping['destroy_empty_storage_group'] = 'False'
        param_mapping['initiator_auto_registration'] = 'True'
        param_mapping['default_timeout'] = 525600

    def _post_service_registration_hook(self, action,
                                        connect_info=None, config={}):
        """Hook to be able to take actions after doing the registration"""
        super(EMC_VNXStorageRegistrar, self).\
            _post_service_registration_hook(action, connect_info, config)
        # Write out the EMC VNX Security Files to the directory to use
        if action == 'register':
            secfiledir = self.__get_security_file_directory(connect_info)
            utils.run_command('mkdir -p -m 750 ' + secfiledir, False, False)
            # Run the NavCLI command to add the Security File to the system
            subcmd = '-AddUserSecurity -secfilepath %s' % secfiledir
            self.__run_navcli_cmd(self.config, subcmd, 1)
        # Update EMC VNX Security Files in the directory on Re-Register
        if action == 'update-registration':
            secfiledir = self.__get_security_file_directory(connect_info)
            olduser, newuser = (self.config['user_id'], config.get('user_id'))
            # If the user id was changed, make sure they gave us the password
            if newuser and olduser != newuser and not config.get('password'):
                self._throw_bad_request(msg.USER_NEEDS_PWD)
            # If they provide a new password or the user id changed, update it
            if config.get('password'):
                newcfg = dict(self.config, **config)
                # Run the NavCLI command to remove the old Security File
                subcmd = '-RemoveUserSecurity -secfilepath %s' % secfiledir
                self.__run_navcli_cmd(newcfg, subcmd, 0)
                # Run the NavCLI command to add the Security File to the system
                subcmd = '-AddUserSecurity -secfilepath %s' % secfiledir
                self.__run_navcli_cmd(newcfg, subcmd, 1)
        # Remove the EMC VNX Security Files from the directory on DeRegister
        if action == 'deregister':
            secfiledir = self.__get_security_file_directory(connect_info)
            # Run the NavCLI command to remove the Security File
            subcmd = '-RemoveUserSecurity -secfilepath %s' % secfiledir
            self.__run_navcli_cmd(self.config, subcmd, 0)
            utils.run_command('rm -rf ' + secfiledir, run_as_root=False)

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        # Verify that the NavCLI is installed on the system to run the command
        self.__verify_navcli_installed()
        # Next try to retrieve and verify the SSL Certificate on SMI-S
        oldcert = utils.get_existing_certificate(config['access_ip'])
        # try_ca=False because VNX does not support local CA bundle validation
        cert = utils.verify_certificate(config, self._get_certificate_port(),
                                        try_ca=False)
        newcert = utils.get_existing_certificate(config['access_ip'])
        # Also add the certificate from the EMC CLI store
        if oldcert != newcert:
            self.__add_navcli_certificate()
        return cert

    def _get_unique_identifier(self, config):
        """Retrieves the unique identifier for the given system"""
        # Command for getting node Unique Identifier from agent properties
        output = self.__run_navcli_cmd(config, 'getagent -node', 1)
        # Loop through each of the lines finding the "Node" line
        for line in output.splitlines():
            if line.strip().startswith('Node:'):
                return line.strip()[5:].strip()

    def _get_volume_pools(self, config):
        """Retrieves the volume pools for the specified system"""
        volume_pool, volume_pools = (None, list())
        # Command for getting the list of Storage Pools from the VNX
        pool_cmd = 'storagepool -list -availableCap -userCap -state'
        output = self.__run_navcli_cmd(config, pool_cmd, 1)
        # Loop through each of the lines finding the Storage Pools
        for line in output.splitlines():
            line = line.strip()
            if line.startswith('Pool Name:'):
                volume_pool = dict(pool_name=line[10:].strip())
                volume_pools.append(volume_pool)
            elif line.startswith('User Capacity (GBs):'):
                volume_pool['pool_total_capacity_gb'] = line[20:].strip()
            elif line.startswith('Available Capacity (GBs):'):
                volume_pool['pool_free_capacity_gb'] = line[25:].strip()
        return volume_pools

    ###############################################
    # ### Internal Registration Helper Methods  ###
    ###############################################
    def __get_security_file_directory(self, connect_info=None):
        """Internal Helper Method to get the Security File Directory"""
        unique_id = self.config.get('unique_id')
        unique_id = connect_info['unique_id'] if connect_info else unique_id
        return '/opt/ibm/powervc/data/emc_vnx/' + unique_id

    def __add_navcli_certificate(self):
        """Internal Helper Method to add the Certificate to the EMC store"""
        navseccmd = 'naviseccli security -certificate'
        # Verify that the NavCLI is installed on the system to run the command
        self.__verify_navcli_installed()
        # Since the EMC CLI doesn't work for certificate on Power skip for now
        if platform.machine().lower().find('ppc') >= 0:
            return
        # If the level wasn't set to medium, then change it to medium
        if utils.run_command(navseccmd + ' -getLevel')[0].strip() != 'medium':
            utils.run_command(navseccmd + ' -setLevel medium')
        # We need to cleanup the old certificate if it previously existed
        self.__cleanup_navcli_certificate(False)
        # Now we can add the certificate to the EMC Certificate Store
        certfile = utils.get_certificate_file(self.config['access_ip'])
        certcmd = ' security -certificate -add -file %s' % certfile
        utils.run_command(NAVCLI_CMD + certcmd, False, False)

    def __verify_navcli_installed(self):
        """Internal Helper Method to verify that the NavCLI is installed"""
        # Since the EMC CLI isn't supported on PPC Linux anymore, don't allow
        if platform.machine().lower().find('ppc') >= 0:
            # See if they want to override the lack of support for PPC Linux
            if not cfg.CONF.support_ppc_naviseccli:
                self._throw_bad_request(msg.NO_PPC_NAVISECCLI)
        # First verify that the NaviSphere CLI is actually installed
        if not utils.file_exists(NAVCLI_CMD):
            raise Exception('The naviseccli is not installed')

    def __cleanup_navcli_certificate(self, log_error=True):
        """Internal Helper Method to add the Certificate to the EMC store"""
        # Since the EMC CLI doesn't work for certificate on Power skip for now
        if platform.machine().lower().find('ppc') >= 0:
            return
        try:
            cert = utils.get_existing_certificate(self.config['access_ip'])
            serial, issuer = utils.get_certificate_identifiers(cert)
            # Run the command to remove the certificate from the EMC store
            cmd = ['security', '-certificate', '-remove',
                   '-serialNumber', serial, '-issuer', issuer]
            utils.run_command([NAVCLI_CMD] + cmd, False, False)
        except Exception as exc:
            if log_error:
                LOG.warn(_("Error removing certificate from the EMC store"))
                LOG.exception(exc)

    def __run_navcli_cmd(self, config, subcmd, pwd_prompt):
        """Internal Helper Method to run a Navisphere CLI Command"""
        # Verify that the NavCLI is installed on the system to run the command
        self.__verify_navcli_installed()
        # If they provided a password (add or edit) use that to connect
        if pwd_prompt and config.get('password'):
            pwd = self._decode_password(config['password'])
            cmd = '%s -Address %s -User %s -Scope 0 -Timeout 30 %s'
            cmd = cmd % (NAVCLI_CMD, config['access_ip'],
                         config['user_id'], subcmd)
            return utils.run_command_pwd_prompt(cmd, pwd, 'Enter password:')
        # If this is an edit without providing the password, use security file
        elif pwd_prompt:
            cmd = '%s -Address %s -secfilepath %s -Timeout 30 %s'
            cmd = cmd % (NAVCLI_CMD, config['access_ip'],
                         config['security_file'], subcmd)
            return utils.run_command(cmd, run_as_root=False)[0]
        # If we don't need to provide the password, just run a normal command
        return utils.run_command(
            '%s -Scope 0 %s' % (NAVCLI_CMD, subcmd), run_as_root=False)[0]


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return EMC_VNXStorageRegistrar(context, **kwargs)
