# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Generic Registrar Classes for performing Storage Registration"""

from oslo_concurrency import lockutils
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar

LOG = logging.getLogger(__name__)


########################################################
# #########  Generic Registrar Implementation  #########
########################################################
class GenericStorageRegistrar(storage_registrar.StorageRegistrar):
    """Generic Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the Generic Class for Storage Registration"""
        if kwargs.get('create', False) is True and not kwargs.get('host_name'):
            kwargs['host_name'] = self._find_next_available_name()
        super(GenericStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['host_type', 'host_display_name',
                'volume_driver', 'configuration_properties']

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attrs = super(GenericStorageRegistrar, self).get_sensitive_attributes()
        attrs.append('configuration_properties')
        return attrs

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_GENERIC_STG)

    def _retrieve_registration(self):
        """Registrar method to retrieve the System Registration"""
        super(GenericStorageRegistrar, self)._retrieve_registration()
        self.config['configuration_properties'] = dict()

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(GenericStorageRegistrar,
                        self)._get_service_param_mapping()
        mapping['volume_driver'] = 'volume_driver'
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        mapping['configuration_properties'] = 'configuration_properties'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(GenericStorageRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        # We temporarily put the Configuration Properties in the Mappings
        cfg_props = parm_mapping.pop('configuration_properties', None)
        cfg_props = cfg_props if isinstance(cfg_props, dict) else dict()
        # Copy each Configuration Property as a Mapping to the CONF File
        for key, val in cfg_props.iteritems():
            parm_mapping[key] = val
        # Since don't know the communication protocol, we can't verify keys
        parm_mapping['strict_ssh_host_key_policy'] = 'False'

    def _update_linux_service(self, parm_mapping, updated_values):
        """Provides the implementation for updating the Service"""
        # We need to break out the Configuration Properties each separately
        if updated_values.get('configuration_properties'):
            cfg_props = updated_values.pop('configuration_properties')
            for key, val in cfg_props.iteritems():
                parm_mapping[key], updated_values[key] = (key, val)
        # Call the Parent Class to actually update the Configuration File
        super(GenericStorageRegistrar, self).\
            _update_linux_service(parm_mapping, updated_values)

    ###############################################
    # ###  Defining StorageRegistrar Interface  ###
    ###############################################
    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        return None

    def _get_volume_pools(self, config):
        """Retrieves the Volume Pools for the specified System"""
        return list()

    @lockutils.synchronized('next_storage_name', 'powervc-')
    def _find_next_available_name(self):
        """Helper method to find the next generated hostname available"""
        for index in range(1000):
            afile = '/etc/cinder/cinder-generic%d.conf' % index
            if not utils.file_exists(afile):
                return 'generic' + str(index)
        return None


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return GenericStorageRegistrar(context, **kwargs)
