# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Hitachi Registrar Classes for performing Storage Registration"""

from oslo_config import cfg
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar
from powervc_discovery.locale import _

cfg.CONF.register_opts([])

LOG = logging.getLogger(__name__)
HORCM_CMD = '/root/hitachi/HORCM/usr/bin'


########################################################
# #########  Hitachi Registrar Implementation  #########
########################################################
class HitachiStorageRegistrar(storage_registrar.StorageRegistrar):
    """Hitachi Storage Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the Hitachi Class for Storage Registration"""
        super(HitachiStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        attr = ['host_type', 'volume_pool_name', 'access_ip', 'user_id',
                'password']
        return attr

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attr = super(HitachiStorageRegistrar, self).get_sensitive_attributes()
        attr.extend(['ports'])
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_HITACHI)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ["unauthorized"]),
            (msg.HORCM_STOPPED, ["Can't be attached"]),
            (msg.NO_HORCM, ['horcmstart.sh: command not found'])
        ]

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(HitachiStorageRegistrar, self).\
            _get_service_param_mapping()
        mapping['hitachi_storage_ip'] = 'access_ip'
        mapping['volume_backend_name'] = 'host_name'
        mapping['hitachi_horcm_user'] = 'user_id'
        mapping['hitachi_horcm_password'] = 'password'
        mapping['volume_pool_name'] = 'volume_pool_name'
        mapping['default_volume_type'] = 'default_volume_type'
        mapping['hitachi_horcm_numbers'] = 'ports'
        return mapping

    def _update_service_param_mapping(self, param_mapping, connect_info):
        """Provides any additional properties to be set by this Registrar"""
        super(HitachiStorageRegistrar, self).\
            _update_service_param_mapping(param_mapping, connect_info)
        param_mapping['volume_driver'] = \
            'powervc_cinder.volume.drivers.' \
            'hitachi.hitachi_product.HitachiProductDriver'
        param_mapping['horcm_path'] = HORCM_CMD
        param_mapping['hitachi_serial_number'] = connect_info['unique_id']
        param_mapping['hitachi_pool_id'] = \
            self._get_volume_pool_id(connect_info,
                                     param_mapping['volume_pool_name'])
        param_mapping['hitachi_group_request'] = True
        param_mapping['hitachi_horcm_add_conf'] = False

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _create_required_files(self, ports):
        """Method for creating the horcm conf files"""
        conf_files = dict()
        # Use relative path on command for rootwrap's sake
        cmd = 'touch'
        # First check to see if the files exist and create them if not
        for port in ports:
            file_path = '/etc/horcm%s.conf' % port
            cmd = cmd + ' ' + file_path
            conf_files[file_path] = port
        utils.run_command(cmd)
        # Use relative path on command for rootwrap's sake
        cmd = 'mkdir -m 777 /var/lock/hbsd/'
        utils.run_command(cmd, True)
        return conf_files

    def _setup_horcm_conf_files(self, config):
        """Method for writing the horcm conf files"""
        # First we need to create the conf files to be configured
        conf_files = self._create_required_files(self.config['ports'])
        # For each file we want to specify the ip address and port to listen on
        for filename in conf_files:
            try:
                # Need to change the permissions in order to to write to file
                cmd = 'chmod 606 ' + filename
                utils.run_command(cmd)
                content = """HORCM_MON
#ip_address       service    poll(10ms)    timeout(10ms)
127.0.0.1         31%(port)s      6000          3000

HORCM_CMD
#dev_name                dev_name                dev_name
\\\\.\IPCMD-%(ip_address)s-31001""" % dict(port=conf_files[filename],
                                           ip_address=config['access_ip'])
                utils.write_to_file(filename, content)
            except Exception:
                self._throw_bad_request(msg.WRITING_ERROR)

    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        # We will check to see if the ports numbers are in the conf file
        ports = self.config.get('ports')
        # If no ports, this is registration so we create and write horcm files
        if not ports:
            ports = ['200', '201']
            # Loop through the ports to determine a suitable pair
            while (utils.file_exists('/etc/horcm%s.conf' % ports[0]) or
                   utils.file_exists('/etc/horcm%s.conf' % ports[1])):
                ports[0] = str(int(ports[0]) + 100)
                ports[1] = str(int(ports[1]) + 100)
        self.config['ports'] = ports
        self._setup_horcm_conf_files(config)
        # Make sure that Raid manager is installed and HORCM is running
        for port in ports:
            start_horcm = 'horcmstart.sh ' + port
            utils.run_command(start_horcm, True)
            try:
                # Login to the RAID Manager using the credentials supplied
                login = 'raidcom -login %(username)s %(password)s ' \
                        '-ITC%(port)s' % dict(username=config['user_id'],
                                              password=config['password'],
                                              port=port)
                utils.run_command(login)
            except Exception as exc:
                LOG.exception(_(exc))
                raise exc

    def _get_property_index(self, prop, content):
        """Method for retrieving the index for a given property"""
        for p in enumerate(content[0].split('\n')[0].split()):
            if p[1] == prop:
                return p[0]

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        # We will run the raidqry command to retrieve the Serial Number
        cmd = 'raidqry -l -ITC%s' % self.config['ports'][0]
        output = utils.run_command(cmd)
        # Get the index for the serial number
        idx = self._get_property_index('Serial#', output)
        # Parse the Serial Number from the output
        return output[0].split('\n')[1].split()[idx]

    def _get_volume_pools(self, config):
        """Retrieves the Volume Pools and ID for the specified System"""
        # We will run the raidcom command to get the volume pools
        volume_pools = list()
        cmd = 'raidcom get pool -key opt -ITC%s' % self.config['ports'][0]
        output = utils.run_command(cmd)
        # Separate each line in the output
        pool_data = output[0].split('\n')
        # Get the index for the volume pool id and volume pool name
        pid = self._get_property_index('PID', output)
        pname = self._get_property_index('POOL_NAME', output)
        # Loop through all the pools getting the necessary info
        for i in range(1, len(pool_data) - 1):
            # Get the volume pool name and id for each volume pool listed
            # We also need a way to pass the ports being used so we do it here
            pool_info = {'pool_name': pool_data[i].split()[pname],
                         'pool_id': int(pool_data[i].split()[pid])}
            pool_info.update(self._get_pool_capacity(pool_info['pool_id']))
            volume_pools.append(pool_info)
        return volume_pools

    def _get_pool_capacity(self, pool_id):
        """Retrieve the free and total capacity of the volume pool"""
        cmd = 'raidcom get pool -pool_id %(pool_id)s -ITC%(port)s' % \
            dict(pool_id=str(pool_id), port=self.config['ports'][0])
        output = utils.run_command(cmd)
        pool_info = output[0].split('\n')[1].split()
        free = self._get_property_index('Available(MB)', output)
        cap = self._get_property_index('Capacity(MB)', output)
        return {'pool_total_capacity_gb': float(pool_info[cap]) / 1024,
                'pool_free_capacity_gb': float(pool_info[free]) / 1024}

    def _get_volume_pool_id(self, connect_info, pool_name):
        """Method for retrieving the volume pool id being registered"""
        for pool in connect_info['volumePools']:
            if pool_name == pool['pool_name']:
                return pool['pool_id']


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return HitachiStorageRegistrar(context, **kwargs)
