# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""SSP Registrar Classes for performing Storage Registration"""

from oslo_config import cfg
from oslo_log import log as logging

from cinder import db

from powervc_oslo.k2 import k2operator as k2o
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration.storage import storage_registrar
from powervc_cinder.db import api as db_api

LOG = logging.getLogger(__name__)


########################################################
# ###########  SSP Registrar Implementation  ###########
########################################################
class SSPStorageRegistrar(storage_registrar.StorageRegistrar):
    """SSP Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the SSP Class for Storage Registration"""
        super(SSPStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['host_name', 'host_type', 'hmc_uuids']

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_HMC)

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        hostname = self.config['host_name']
        # We only currently support managing a Host from a Single HMC
        if len(config['hmc_uuids']) != 1:
            self._throw_bad_request(msg.SINGLE_HMC)
        hmc_uuid = config['hmc_uuids'][0]
        # Retrieve the IP/Credentials from the DB and load the info to use
        hmc = db_api.ibm_hmc_get_by_uuid(self.context, hmc_uuid)
        # Make sure that the HMC provided actually has been registered
        if not hmc:
            self._throw_bad_request(msg.HMC_NOTFOUND, hmcname=hmc_uuid)
        # Update the Configuration Properties with the DB Info we Retrieved
        ipcred = {'access_ip': hmc['access_ip'],
                  'user_id': hmc['user_id'], 'password': hmc['password'], }
        # Call K2 to retrieve the entry for the requesting Cluster
        k2oper, k2entry = self._get_k2_cluster(dict(config, **ipcred))
        # If the Cluster doesn't exist on the specified HMC, throw an exception
        if k2entry is None:
            self._throw_bad_request(msg.NO_CLUSTER_ON_HMC, hostname=hostname)
        # Set the Unique ID and Display Name based on the info from K2
        connect_info['unique_id'] = k2entry.properties['id']
        connect_info['host_display_name'] = \
            k2entry.element.findtext('ClusterName')
        # We only need to retrieve the Volume Pools if this is a connect calls
        if config.get('volume_pool_name') == '[temp]':
            connect_info['volumePools'] = \
                self._get_volume_pools(k2oper, k2entry)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['unauthorized']),
            (msg.CONNECTION_TIMEOUT_HMC, ['timed out']),
            (msg.CONNECTION_NOTKNOWN, ['name or service not known']),
            (msg.CONNECTION_FAILED_HMC, ['connection']),
            (msg.INVALID_CERTIFICATE, ['certificate verify']),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired'])
        ]

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 12443

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def deregister(self, force=False, uninstall=False, internal=False):
        """Registrar Method to delete a System Registration"""
        # We only allow SSP's to be removed via removing a Host or by Force
        if not internal and not force:
            self._throw_bad_request(msg.SSP_REMOVE_INTERNAL)
        if internal:
            host_name = self.config['host_name']
            volumes = db.volume_get_all_by_host(self.context, host_name)
            super(SSPStorageRegistrar, self).verify_no_image_volumes(volumes)
            force = True
            super(SSPStorageRegistrar, self).deregister(force, uninstall,
                                                        internal)
        else:
            super(SSPStorageRegistrar, self).deregister(force, uninstall,
                                                        internal)

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(SSPStorageRegistrar, self)._get_service_param_mapping()
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        return mapping

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        super(SSPStorageRegistrar, self)._parse_service_config(config)
        hmc_uuids = []
        host_name = self.config['host_name']
        # Query database for the related HMC's
        hmcs = db_api.ibm_hmc_get_all_by_cluster(self.context, host_name)
        for hmc in hmcs:
            hmc_uuids.append(hmc['hmc_uuid'])
        self.config['hmc_uuids'] = hmc_uuids

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(SSPStorageRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['volume_driver'] = \
            'powervc_cinder.volume.drivers.ssp.ssp_product.SSPProductDriver'
        # If the Display Name isn't set, set it to the value on the HMC
        if not parm_mapping.get('host_display_name'):
            parm_mapping['host_display_name'] = \
                connect_info['host_display_name']
        parm_mapping['powervm_mgr_cluster'] = connect_info['unique_id']
        # To accommodate slow SSP GET change periodic_interval, default is 60s
        parm_mapping['periodic_interval'] = '60'

    def _post_service_registration_hook(self, action,
                                        connect_info=None, config={}):
        """Hook to be able to take actions after doing the registration"""
        context, hmc_uuids = self.context, self.config['hmc_uuids']
        hmc_uuid = hmc_uuids[0] if hmc_uuids else None
        host_name = self.config['host_name']
        # Adds the New Cluster to the Member list for the registered HMC
        if action == 'register':
            db_api.ibm_hmc_clusters_create(context, hmc_uuid, host_name)
        # Removes the Cluster from the Member list for the registered HMC
        if action == 'deregister' and hmc_uuid:
            db_api.ibm_hmc_clusters_delete(context, hmc_uuid, host_name)
        # Updates the HMC for the Cluster if a new HMC was specified
        if action == 'update-registration':
            new_hmc = config.get('hmc_uuids', list())
            # If the HMC has changed, we want change it over to the new one
            if new_hmc and new_hmc[0] != hmc_uuid:
                db_api.ibm_hmc_clusters_delete(context, hmc_uuid, host_name)
                db_api.ibm_hmc_clusters_create(context, new_hmc[0], host_name)

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        return config['unique_id']

    def _get_volume_pools(self, k2oper, k2entry):
        """Retrieves the Volume Pools for the specified System"""
        volume_pools = []
        # Retrieve a list SSP's on cluster and get their properties
        for link in k2entry.element.findall('ClusterSharedStoragePool'):
            resp2 = k2oper.readbyhref(link.get('href'))
            name = resp2.entry.element.findtext('StoragePoolName')
            total = resp2.entry.element.findtext('Capacity')
            free = resp2.entry.element.findtext('FreeSpace')
            volume_pools.append({'pool_name': name,
                                 'pool_total_capacity_gb': total,
                                 'pool_free_capacity_gb': free})
        return volume_pools

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _get_k2_cluster(self, config):
        """Retrieves a dictionary of Clusters and ClusterID's on the HMC"""
        timeout = cfg.CONF.service_registration_timeout
        # Get a K2 session for connecting to HMC's K2 API's
        session = k2o.K2Session(config['access_ip'],
                                config['user_id'], config['password'])
        operator = k2o.K2Operator(session)
        # We need to search thru the list of Clusters for a match
        resp = operator.read('Cluster', timeout=timeout)
        # Loop through the clusters and get their properties
        for entry in resp.feed.entries:
            cluster_id = entry.element.findtext('ClusterID')
            # Check if the Host Name matches the Cluster ID given
            if cluster_id == config['host_name']:
                return operator, entry
        return None, None


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return SSPStorageRegistrar(context, **kwargs)
