# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base Registrar Classes for performing Storage Registration"""

import sys
import cgi
import commands
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import timeutils

from cinder import db
from cinder import exception
from cinder import rpc as rpc_api
from cinder.image import glance
from cinder.volume import volume_types

from powervc_cinder.db import api as powervc_db_api
from powervc_cinder.volume import rpcapi_product

from powervc_discovery.onboarding.storage import volume_importer
from powervc_discovery.registration import registrar as reg
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.ListOpt('supported_host_types', default=['svc'],
                help='The list of Storage Types that are supported'),
    cfg.ListOpt('volume_type_name_template',
                help='The template used for formatting the Volume Type')
])

LOG = logging.getLogger(__name__)


########################################################
# ########  Storage Registrar Implementation  ##########
########################################################
class StorageRegistrar(reg.ServiceRegistrar):
    """Abstract Base Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Storage Registration"""
        super(StorageRegistrar, self).__init__(context, **kwargs)
        self.volume_api = rpcapi_product.VolumeAPIProduct()

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        return ['password', 'private_key', 'private_key_data', 'unique_id']

    def get_immutable_attributes(self):
        """Provides the list of Attributes that can't be updated"""
        return ['host_name', 'host_type']

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        # Initialize protocol necessary for connection
        self._initialize_connection(config)
        # Retrieve the Unique Identifier to use for the System
        connect_info['unique_id'] = self._get_unique_identifier(config)
        connect_info['volumePools'] = self._get_volume_pools(config)
        # Retrieve the Display Name to use from what the System has set
        connect_info['host_display_name'] = self._get_display_name(config)
        # Storage Drivers still need to have a default  Volume Pool
        self._verify_volume_pool_exists(config.get('volume_pool_name'),
                                        connect_info['volumePools'])

    def _get_registration_type_string(self):
        """Provides the translated Registration Type String for this
        Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.REG_TYPE_STORAGE)

    def _get_registration_log_file(self):
        """Describes the name of the Log File running the Registration"""
        return '/var/log/cinder/api.log'

    def _get_name_attribute(self):
        """Provides the Name of the Attribute to use for the system name"""
        return 'host_name'

    def _get_private_key_prepend(self):
        """Provides the Directory and Prefix for the Private Key Location"""
        return '/etc/cinder/rsakeys/cinder_'

    ###############################################
    #   Implementing ServiceRegistrar Interface   #
    ###############################################
    def deregister(self, force=False, uninstall=False, internal=False):
        """Registrar Method to delete a System Registration"""
        super(StorageRegistrar, self).deregister(force, uninstall)
        # After the Storage is removed, then delete the default volume type
        self.__delete_volume_types()

    def verify_no_image_volumes(self, volumes):
        """Internal Helper method to Verify that No Images are using Volumes"""
        image_volumes = []
        # Loop through each Volume, keeping track of the ones that back images
        for volume in volumes:
            image = None
            metadata = None
            try:
                metadata = db.volume_metadata_get(self.context, volume['id'])
                if metadata and ('is_image_volume' in metadata and
                                 bool(metadata['is_image_volume'])):
                    image_service, image_id = \
                        glance.get_remote_image_service(self.context,
                                                        metadata['image_id'])
                    try:
                        image = image_service.show(self.context.elevated(),
                                                   image_id)
                    except exception.ImageNotFound:
                        pass  # this is actually what we want here
            except Exception:
                LOG.exception(_("Failed checking whether volume %(uuid)s backs"
                                " an image - metadata=%(metadata)s")
                              % dict(uuid=volume['id'], metadata=metadata))
            if image is not None and image.get('status') != 'deleted':
                image_volumes.append(volume['id'])
        # If any Volumes were still backing images, throw an exception as such
        if len(image_volumes) > 0:
            volumes_parm = utils.format_string_list(image_volumes)
            self._throw_bad_request(
                msg.REMOVE_IMAGE_VOLUMES, volumes=volumes_parm)

    def _get_service_name(self):
        """Provides the name of the Linux Service created by this Registrar"""
        return 'openstack-cinder-volume-' + self.config['host_name']

    def _get_service_config_file(self):
        """Provides the name of the Config File created by this Registrar"""
        return '/etc/cinder/cinder-' + self.config['host_name'] + '.conf'

    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        return {'host_type': 'host_type', 'host_uniqueid': 'unique_id',
                'host_display_name': 'host_display_name'}

    def _pre_service_registration_hook(self, action, connect_info=None,
                                       config={}, force=False):
        """Hook to be able to take actions after doing the registration"""
        super(StorageRegistrar, self).\
            _pre_service_registration_hook(action, connect_info, config, force)
        # If the Display Name wasn't set by the User, add it to the config
        if action == 'register':
            self.config['unique_id'] = connect_info.get('unique_id')
            if not self.config.get('host_display_name'):
                self.config['host_display_name'] = \
                    connect_info.get('host_display_name')
            if not self.config.get('default_volume_type'):
                self.config['default_volume_type'] = \
                    self._create_default_volume_type_name()
            # Confirm that the Display Name provided isn't too long
            if len(unicode(self.config.get('host_display_name'))) > 247:
                self._throw_bad_request(msg.STORAGE_DISPLAY_LEN)
        # Confirm that this Storage Host is able to be ReRegistered
        if action == 'update-registration':
            self.__verify_allowed_to_edit(force)
        # Confirm that this Storage Host is able to be DeRegistered
        if action == 'deregister':
            self.__verify_allowed_to_delete(force)

    def _parse_service_config(self, config):
        """Parses the Config Properties into those returned by the Registrar"""
        super(StorageRegistrar, self)._parse_service_config(config)
        # TODO: Can remove once we have the Host Type always in the Config File
        if self.config.get('host_type') is None:
            self.config['host_type'] = 'svc'

    def _get_service_instance(self, verify_existence=False):
        """Retrieves the Service Instance from the Database for this System"""
        hostname = self.config['host_name']
        services = powervc_db_api.\
            service_get_all_by_host(self.context, hostname)
        # We will wait until the Storage Node exists if this is verification
        if verify_existence and services is not None and len(services) > 0:
            storage_node = powervc_db_api.\
                storage_node_get_by_host(self.context, hostname)
            return services[0] if storage_node is not None else None
        return services[0] if services and len(services) > 0 else None

    def _destroy_service_instance(self, service):
        """Cleans up the Service Instance from the Database on Deregister"""
        context = self.context
        hostname = self.config['host_name']
        anotifier = rpc_api.get_notifier('volume', hostname)
        # Cleanup each of the Volumes returned from the database
        importer = volume_importer.VolumeImporter(context, hostname)
        importer.unmanage_resources(['*ALL'], force=True)
        # Retrieve the Storage Node to have info to send in the notification
        storage_node = self.__get_storage_node_instance(service)
        if storage_node:
            anotifier.info(context, 'storage.node.delete.start', storage_node)
        # Clean up the Service and Storage Host entries from the database
        powervc_db_api.service_delete(context, service['id'])
        if storage_node:
            anotifier.info(context, 'storage.node.delete.end', storage_node)

    def _verify_no_existing_service(self, connect_info):
        """Helper Method to see if the System already has been Registered"""
        new_id = connect_info.get('unique_id')
        # Loop through the config files to make sure there aren't any existing
        # systems registered for the same system as the one being registered
        for fil in utils.get_matching_files('/etc/cinder', 'cinder-', '.conf'):
            hostname = fil[7:len(fil) - 5]
            registrar = StorageRegistrar(self.context, host_name=hostname)
            old_id = registrar.config.get('unique_id')
            # If the Unique Identifier matches between 2, block adding
            if old_id and new_id and old_id == new_id:
                dispname = registrar.config.get('host_display_name', hostname)
                self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=dispname)

    def _wait_for_service_initialization(self):
        """Allows sub-class to wait longer for service initialization"""
        host = self.config['host_name']
        # Wait up to an additional 60 seconds for the service to start
        self.volume_api.verify_host_running(self.context, host, 60)

    def _delete_linux_service(self):
        """Provides the implementation for deleting/removing the Service"""
        super(StorageRegistrar, self)._delete_linux_service()
        # Since cinder-volume has a watch-dog, may need to kill the processes
        processes = list()
        procstr = '/usr/bin/python /usr/bin/cinder-volume '
        procstr += '--config-file /etc/cinder/cinder-%s.conf'
        procstr = procstr % self.config['host_name']
        # Run the PS command to get a list of running cinder-volume processes
        output = utils.run_command('/bin/ps -ef', True, False)[0]
        output = encodeutils.safe_decode(output)
        # Loop through each line adding the process id to the list
        for line in output.strip().splitlines():
            # Make sure this matches our exact storage provider's name
            if line.find(procstr) > 0:
                processes.append(line.strip().split()[1].strip())
        # If there were any processes running, then run kill on the full list
        if len(processes) > 0:
            # Use relative path on command for rootwrap's sake
            utils.run_command('kill -9 ' + ' '.join(processes), True)

    def _describe_host_service_error(self, force_check=False):
        """Determine what the Reason was that the Host Service didn't start"""
        status = self._host_service_status()
        # Check to see if the Service is Started/Stopped rather than Crashed
        if status == 'started' and not force_check:
            return ('ok', 'Cinder Volume Service is Started')
        if status == 'stopped':
            return ('service_stopped', 'Cinder Volume Service is Stopped')
        # If the Service crashed, lets try to figure out why it crashed
        hostlog = '/var/log/cinder/volume-' + self.config['host_name'] + '.log'
        output = commands.getoutput('tail -n 30 ' + hostlog)
        # Call the Helper Method to parse the Error Output provided
        return self._describe_connection_error(output, hostlog)

    ###############################################
    # #### Defining StorageRegistrar Interface ####
    ###############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        pass

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        raise NotImplementedError()

    def _get_volume_pools(self, config):
        """Retrieves the Volume Pools for the specified System"""
        raise NotImplementedError()

    def _get_display_name(self, config):
        """Retrieves the Display Name for the given System"""
        name = self.config.get('host_name')
        access_ip = self.config.get('access_ip')
        display_name = self.config.get('host_display_name')
        # Don't even bother to generate a Display Name if one is already set
        if display_name:
            return cgi.escape(display_name)
        # If the IP Address is set, we will generate it based on the IP Address
        if access_ip:
            return utils.formulate_hostname(access_ip)
        return name if name else None

    ###############################################
    #     Shared Registration Helper Methods      #
    ###############################################
    def _create_default_volume_type_name(self):
        """Internal Helper Method to generate the default volume type name"""
        # Since the default Volume Type will generally use the Display Name,
        # we want to make sure there aren't going to be 2 with the same Name
        self.__verify_not_duplicate_names(self.config.get('host_display_name'))
        # If they gave a template, generate the name based on that template
        # They need to specify for substitution parameters the attributes
        # provide on the REST API, so like host_name, host_display_name, etc
        if cfg.CONF.volume_type_name_template:
            return cfg.CONF.volume_type_name_template % self.config
        # Otherwise we will use the display name with a postfix appended
        nametxt = _("%(host_display_name)s base template") % self.config
        return discovery_locale.translate_message(self.context, nametxt)

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def __get_storage_node_instance(self, service):
        """Internal Helper Method to Retrieve the Storage Node from the DB"""
        try:
            # Try to retrieve the Storage Node Instance from the Database
            node = powervc_db_api.\
                storage_node_get_by_host(self.context, service['host'])
            # Only provide the essentials attributes for the notification
            if node is not None:
                return {'storage_node_id': node['id'], 'host': service['host'],
                        'service_id': node['service_id'],
                        'storage_hostname': node['storage_hostname']}
        except Exception as exc:
            # Log the Error to help in debug, but we still want to do cleanup
            LOG.warn(_("Error Retrieving Storage Node %(host)s on "
                       "Cleanup, continuing on: %(exc)s") %
                     dict(host=service['host'], exc=unicode(exc)))
        return None

    def __verify_not_duplicate_names(self, dispname):
        """Internal Helper method to verify not a duplicate Display Name"""
        for fn in utils.get_matching_files('/etc/cinder', 'cinder-', '.conf'):
            config = utils.read_config_file('/etc/cinder/' + fn)
            # If the Display Name is the same as another, throw the exception
            if dispname and config.get('host_display_name') == dispname:
                self._throw_bad_request(msg.DUP_DISP_NAMES, dispname=dispname)

    def _verify_volume_pool_exists(self, pool_name, volume_pools):
        """Internal Helper Method to verify that the specified Pool exists"""
        # The Storage Drivers still need to have a default Volume Pool
        if(pool_name and pool_name != '[temp]' and volume_pools and
           len(volume_pools) > 0):
            for volume_pool in volume_pools:
                if volume_pool['pool_name'] == pool_name:
                    return
            self._throw_bad_request(msg.INVALID_POOL_ERROR)

    def __delete_volume_types(self):
        """Internal Helper Method to Delete the Volume Types for Host"""
        host_name = self.config['host_name']
        vol_types = volume_types.get_all_types(self.context).values()
        for vol_type in vol_types:
            extra_specs = vol_type['extra_specs']
            vol_type_id = vol_type['id']
            if "capabilities:volume_backend_name" in extra_specs:
                name = extra_specs["capabilities:volume_backend_name"]
                if name == host_name:
                    volume_types.destroy(self.context, vol_type_id)

    def __verify_allowed_to_delete(self, force):
        """Internal Helper method to Verify the Storage Host can be deleted"""
        host_name = self.config['host_name']
        # Only do the verification if they didn't ask to force the delete
        if not force and not cfg.CONF.force_deregister:
            volumes = db.volume_get_all_by_host(self.context, host_name)
            # First make sure that none of the Volumes are attached to VM's
            self.__verify_no_attached_volumes(volumes)
            # Next make sure that non Images are using any of the Volumes
            self.verify_no_image_volumes(volumes)
            # Next make sure that none of the Volumes have Tasks in-progress
            self.__verify_no_tasks_inprogress(
                volumes, msg.REMOVED_ACTION, msg.REMOVING_ACTION)

    def __verify_allowed_to_edit(self, force):
        """Internal Helper method to Verify the Compute Host can be deleted"""
        host_name = self.config['host_name']
        # Only do the verification if they didn't ask to force the edit
        if not force and not cfg.CONF.force_deregister:
            volumes = db.volume_get_all_by_host(self.context, host_name)
            # First make sure that none of the Volumes have Tasks in-progress
            self.__verify_no_tasks_inprogress(
                volumes, msg.EDITED_ACTION, msg.EDITING_ACTION)

    def __verify_no_attached_volumes(self, volumes):
        """Internal Helper method to Verify that No Volumes are Attached"""
        attached_volumes = []
        # Loop through each Volume, keeping track of the ones that are attached
        for volume in volumes:
            if volume.get('attach_status') == 'attached':
                volume_name = volume.get('display_name')
                # If this is an internal one, use the Volume UUID for the Name
                if volume_name is None or len(volume_name) <= 0:
                    volume_name = 'volume-' + volume['id']
                attached_volumes.append(volume_name)
        # If any Volumes were still attached, throw an exception as such
        if len(attached_volumes) > 0:
            volumes_parm = utils.format_string_list(attached_volumes)
            self._throw_bad_request(
                msg.REMOVE_ATTACHED_VOLUMES, volumes=volumes_parm)

    def __verify_no_tasks_inprogress(self, volumes, action, actioning):
        """Internal Helper method to Verify that No Tasks are In-progress"""
        host_name = self.config['host_name']
        # If the Volume Service isn't running, then no tasks are really running
        if self._host_service_status() != 'started':
            return
        # If we can't contact the Volume Service, assume it has hung
        if not self.volume_api.verify_host_running(self.context, host_name):
            return
        volumes_inprogress = list()
        # Loop through each of the Volumes seeing if any have tasks in-progress
        for volume in volumes:
            status = volume.get('status')
            updated_at = volume.get('updated_at')
            if updated_at is None:
                updated_at = volume.get('created_at')
            # If the Status of Volume is Available or In-Use, no task running
            if status is None or status in ['available', 'in-use']:
                continue
            # If the Status of the Volume is in error, means the task failed
            if status.startswith('error') or updated_at is None:
                continue
            # If the Volume wasn't updated within the last 1/2 hour we will
            # assume the task got stuck and skip it, otherwise since the
            # status showed a task in-progress, don't allow the Storage removal
            if not timeutils.is_older_than(updated_at, 1800):
                volume_name = volume.get('display_name')
                # If this is an internal one, use the Volume UUID for the Name
                if volume_name is None or len(volume_name) <= 0:
                    volume_name = 'volume-' + volume['id']
                volumes_inprogress.append(volume_name)
        # If any tasks were in-progress throw an exception blocking removal
        if len(volumes_inprogress) > 0:
            b = self._get_message_bundle()
            action = b.get_message_text(self.context, action)
            actioning = b.get_message_text(self.context, actioning)
            rsctype = b.get_message_text(self.context, msg.RSC_TYPE_VOLUME)
            rsctypes = b.get_message_text(self.context, msg.RSC_TYPE_VOLUMES)
            # Include only first 15 volume names in the list to avoid clutter
            volume_names = utils.format_string_list(volumes_inprogress[:15])
            self._throw_bad_request(msg.TASKS_INPROGRESS, action=action,
                                    rsctypes=rsctypes, rscnames=volume_names,
                                    rsctype=rsctype, actioning=actioning)


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    throw_exc = kwargs.pop('throw_exc', True)
    try:
        # If this is not a register, we need to figure out the type
        tempreg = StorageRegistrar(context, **kwargs)
        host_type = tempreg.config.get('host_type')
        # If the Host Type isn't in the Supported List, throw an exception
        if host_type not in (cfg.CONF.supported_host_types + ['generic']):
            tempreg._throw_bad_request(msg.INVALID_HOST_TYPE,
                                       hosttype=host_type)
        try:
            # Attempt to dynamically load the corresponding module for the type
            regmodule = __package__ + '.' + str(host_type) + '_registrar'
            __import__(regmodule)
            return getattr(sys.modules[regmodule],
                           'get_registrar')(context, **kwargs)
        except Exception as exc:
            LOG.exception(exc)
            tempreg._throw_bad_request(msg.INVALID_HOST_TYPE,
                                       hosttype=host_type)
    except:
        # Only throw the exception if we were asked to throw it on error
        if throw_exc:
            raise
    return None
