# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Register a new Storage Host"""

import webob.exc
from oslo_log import log as logging
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.volume import volume_types

from powervc_cinder.db import api as pwrvcdb_api
from powervc_discovery.registration.storage import storage_registrar
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
AUTHORIZE = extensions.extension_authorizer('volume', 'host_registration')
AUTHORIZEH = extensions.extension_authorizer('volume', 'hmc_registration')


class StorageRegistrationControllerExtension(wsgi.Controller):
    """Main Controller Class to extend the os-hosts API for Registration"""

    def __init__(self):
        """Constructor for the Host Registration Controller Extension"""
        super(StorageRegistrationControllerExtension, self).__init__()

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Extend Rest GET Method to Add Attributes for a specific Host"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'show')
        # Add the Storage Capacity of the Host to what is returned for details
        self._add_host_capacity(context, resp_obj, id)
        # Load the appropriate Registrar to Retrieve the Registration #Info
        registrar = storage_registrar.get_registrar(context, host_name=id,
                                                    throw_exc=True)
        # Only add the Registration Info if it was already registered
        if registrar:
            registration = self.filter_config(registrar)
            resp_obj.obj['host'].append({'registration': registration})

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Create a Host Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'create')
        registration = body['host']['registration']
        # Load the appropriate Registrar to do the Storage Registration
        registrar = storage_registrar.get_registrar(context, create=True,
                                                    **registration)
        # Delegate to the Registrar to do the Storage Registration
        registrar.register()
        # Return on the Response the Registration Info
        return {'host': {'registration': self.filter_config(registrar)}}

    @wsgi.action("update_old")
    def update_old(self, req, host_name_id, body):
        """Extend Rest PUT Method to Update a Host Registration"""
        host = {'host': host_name_id}
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'update')
        registration = body.pop('registration', None)
        # If no Registration Info is provided, then updating a base attribute
        if registration is not None:
            # Load the appropriate Registrar to do the Storage Re-Registration
            registrar = storage_registrar.get_registrar(context,
                                                        host_name=host_name_id)
            # Delegate to the Registrar to do the Storage Re-Registration
            registrar.update_registration(registration)
            host['registration'] = self.filter_config(registrar,
                                                      config=registration)
        return host

    @wsgi.extends
    def update(self, req, id, body):
        """Extend Rest PUT Method to Update a Host Registration"""
        if body.get('registration') is not None:
            our_resp = self.update_old(req, id, body)
            resp_obj = (yield)
            self._add_our_response(req, resp_obj, our_resp)

    @wsgi.action("delete")
    @wsgi.response(204)
    def delete(self, req, id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'delete')
        # Load the appropriate Registrar to do the Storage De-Registration
        registrar = storage_registrar.get_registrar(context, host_name=id)
        # Delegate to the Registrar to do the Storage De-Registration
        registrar.deregister()

    @wsgi.action("force_delete")
    @wsgi.response(204)
    def force_delete(self, req, host_name_id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'delete')
        # Load the appropriate Registrar to do the Storage De-Registration
        registrar = storage_registrar.\
            get_registrar(context, host_name=host_name_id)
        # Delegate to the Registrar to do the Storage De-Registration
        registrar.deregister(force=True)

    @wsgi.action("delete_internal")
    @wsgi.response(204)
    def delete_internal(self, req, host_name_id):
        """Add Rest DELETE Method to Remove a Host Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'delete')
        # Load the appropriate Registrar to do the Storage De-Registration
        registrar = storage_registrar.\
            get_registrar(context, host_name=host_name_id)
        # Delegate to the Registrar to do the Storage De-Registration
        registrar.deregister(internal=True)

    def _add_host_capacity(self, context, resp_obj, host):
        """Internal Helper Method to populate the Capacity of the Storage"""
        storage_node = None
        try:
            # Try to retrieve the Storage Node from the Cinder Database
            storage_node = pwrvcdb_api.storage_node_get_by_host(context, host)
        except Exception as exc:
            # Log the exception that came trying to query the database
            LOG.warn(_("Error querying the Storage Node from the Database"))
            LOG.exception(exc)
        # Make sure that the Storage Node was retrieved from the Database
        if storage_node is not None:
            for items in resp_obj.obj['host']:
                # Only want to use the Resource elements in the list returned
                rsc = items.get('resource')
                # Add the Total/Free Capacity to the Resource for all projects
                if rsc and rsc['host'] == host and rsc['project'] == '(total)':
                    rsc['free_capacity_gb'] = \
                        str(storage_node['free_capacity_gb'])
                    rsc['total_capacity_gb'] = \
                        str(storage_node['total_capacity_gb'])

    def _add_our_response(self, req, resp_obj, our_resp):
        """Internal Helper Method to add our response for the Update"""
        resp_obj.obj.update(our_resp)

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        # We want to use the more granular version, but can't until it exists
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def filter_config(registrar, config=None):
        """Internal Helper Method to Remove any Sensitive Attributes """
        new_config = registrar.config.copy()
        if config is not None:
            new_config = config.copy()
        # Loop through each of the Sensitive Attributes, removing them
        for attr in registrar.get_sensitive_attributes():
            new_config.pop(attr, None)
        return new_config


class StorageConnectControllerExtension(wsgi.Controller):
    """Main Controller Extension Class to provide the os-hosts/connect API"""

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Get the Storage Host's Volume Pools"""
        context = req.environ['cinder.context']
        registration = body['host']['registration']
        mainext = StorageRegistrationControllerExtension
        context = mainext._validate_authorization(context, req, 'connect')
        # Load the appropriate Registrar sub-class to do the Host Registration
        registrar = storage_registrar.get_registrar(context, create=True,
                                                    **registration)
        # Set the Volume Pool to a dummy value, so it passes verification
        registrar.config['volume_pool_name'] = '[temp]'
        # Delegate to the Registrar to connect to the Storage Host
        connection_info = registrar.establish_connection(registrar.config)
        connection_info.pop('host_display_name', None)
        return {'host': mainext.filter_config(registrar, connection_info)}


class StorageConnectControllerExtensionOld(StorageConnectControllerExtension):
    """Old Controller Extension Class to provide the os-hosts/connect API"""

    @wsgi.action("create")
    def create(self, req, host_name_id, body):
        """Add Rest POST Method to Get the Storage Host's Volume Pools"""
        # We only want to allow this API if it is v1 for Legacy purposes
        if req.path.startswith('/v1') is False:
            text = _("Use the /os-hosts/connect API instead")
            raise webob.exc.HTTPNotFound(explanation=text)
        return super(StorageConnectControllerExtensionOld,
                     self).create(req, body)


class StorageProviderHostControllerExtension(wsgi.Controller):
    """Controller Extension Class to add data to storage-providers"""

    @wsgi.extends
    def show(self, req, resp_obj, id):
        """Extend Rest GET Method to Add Attributes for a Storage Provider"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'show')
        # Get the Storage Provider that the Base Extension retrieved
        provider = resp_obj.obj.get('storage_provider', {})
        # Add the additional attributes from the Storage Registrar
        self._add_host_attributes(context, provider)

    @wsgi.extends
    def detail(self, req, resp_obj):
        """Extend Rest GET Method to Add Attributes for all Storage Nodes"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'detail')
        # Loop through the Storage Providers the Base Extension retrieved
        for provider in resp_obj.obj.get('storage_providers', []):
            # Add the additional attributes from the Storage Registrar
            self._add_host_attributes(context, provider)

    def _add_host_attributes(self, context, provider):
        """Helper method to add additional Host Attributes to a Provider"""
        service = provider.get('service', {})
        host = service.get('host')
        # We only have anything to add if we could find the given Host
        if host:
            # Delegate to the Storage Registrar to retrieve the registration
            registrar = storage_registrar.get_registrar(
                context, host_name=host, throw_exc=False)
            # If the Host was registered, add the additional attributes
            if registrar:
                additional_attrs = [
                    'host_display_name', 'hmc_uuids', 'access_ip']
                for attr in additional_attrs:
                    if registrar.config.get(attr) is not None:
                        service[attr] = registrar.config[attr]
                # Now we can lookup and add the Default Volume Type
                if registrar.config.get('default_volume_type'):
                    try:
                        # Look up the Volume Type ID for the given Name
                        vtn = registrar.config['default_volume_type']
                        vt = volume_types.get_volume_type_by_name(context, vtn)
                        service['default_volume_type'] = vt['id']
                    # If the Volume Type wasn't found, we just won't set it
                    except:
                        pass

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class HMCRegistrationControllerExtension(wsgi.Controller):
    """Main Controller Class to provide the ibm-hmcs API for Cinder Sync"""

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Create an HMC Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'create')
        # Call the DB API to create the HMC in the Database
        pwrvcdb_api.ibm_hmc_create(context, body['hmc']['registration'])
        return body

    @wsgi.action("update")
    def update(self, req, id, body):
        """Add Rest PUT Method to Update an HMC Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'update')
        # Call the DB API to update the HMC in the Database
        pwrvcdb_api.ibm_hmc_update(context, id, body['registration'])
        return {'hmc': id, 'registration': body['registration']}

    @wsgi.action("delete")
    @wsgi.response(204)
    def delete(self, req, id):
        """Add Rest DELETE Method to Remove an HMC Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'delete')
        # Call the DB API to delete the HMC from the Database
        pwrvcdb_api.ibm_hmc_delete(context, id)

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZEH(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)


class StorageRegistration(extensions.ExtensionDescriptor):
    """Capability to Register a new Storage Host."""
    name = "Storage Host Registration"
    alias = "ibm-storage-registration"
    namespace = "http://docs.openstack.org/volume/ext/" + \
                "ibm_host_registration/api/v1.1"
    updated = "2013-03-21T21:00:00-06:00"

    def get_resources(self):
        """Provide a Resource Extension to the os-hosts Resource"""
        resource1 = extensions.ResourceExtension(
            'connect', StorageConnectControllerExtensionOld(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'create': 'POST'})
        resource2 = extensions.ResourceExtension(
            'update-registration',
            controller=StorageRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'update_old': 'PUT'})
        resource3 = extensions.ResourceExtension(
            'os-hosts/connect', StorageConnectControllerExtension(),
            collection_actions={'create': 'POST'})
        resource4 = extensions.ResourceExtension(
            'force',
            controller=StorageRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'force_delete': 'DELETE'})
        resource5 = extensions.ResourceExtension(
            'internal',
            controller=StorageRegistrationControllerExtension(),
            parent=dict(collection_name='os-hosts', member_name='host_name'),
            collection_actions={'delete_internal': 'DELETE'})
        # Internal REST API for synchronizing the HMC info in Cinder
        resource6 = extensions.ResourceExtension(
            'ibm-hmcs', controller=HMCRegistrationControllerExtension())
        return [resource1, resource2, resource3, resource4, resource5,
                resource6]

    def get_controller_extensions(self):
        """Provide a Controller Extension to the os-hosts Resource"""
        extension1 = extensions.ControllerExtension(
            self, 'os-hosts', StorageRegistrationControllerExtension())
        extension2 = extensions.\
            ControllerExtension(self, 'storage-providers',
                                StorageProviderHostControllerExtension())
        return [extension1, extension2]
