# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""SVC Registrar Classes for performing Storage Registration"""

from oslo_config import cfg
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar
from powervc_discovery.locale import _

cfg.CONF.register_opts([
    cfg.StrOpt('storwize_svc_connection_protocol', default='fc',
               help='Connection protocol (iSCSI/FC) used to the SVC')
])

LOG = logging.getLogger(__name__)


########################################################
# ##########  SVC Registrar Implementation  ############
########################################################
class SVCStorageRegistrar(storage_registrar.StorageRegistrar):
    """SVC Registrar Class to handle Storage Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the SVC Class for Storage Registration"""
        super(SVCStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        attr = ['host_type', 'access_ip', 'user_id',
                'private_key|private_key_data|password']
        attr.append('volume_pool_name')
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_SVC)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.INVALID_SYSTEM, ['lssystem: command not',
                                  'lsmdiskgrp: command not']),
            (msg.INVALID_SVC_ISCSI, ['illegal value iscsi specified'])
        ] + utils.ssh_reg_messages()

    #############################################
    #  Implementing ServiceRegistrar Interface  #
    #############################################
    def _get_service_param_mapping(self):
        """Provides the mapping of Properties to be used by the Registrar"""
        mapping = super(SVCStorageRegistrar, self)._get_service_param_mapping()
        mapping['san_ip'] = 'access_ip'
        mapping['san_login'] = 'user_id'
        mapping['san_password'] = 'password'
        mapping['san_private_key'] = 'private_key'
        mapping['san_ssh_port'] = 'ssh_port'
        mapping['storwize_svc_volpool_name'] = 'volume_pool_name'
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        mapping['storwize_svc_connection_protocol'] = 'driver_volume_type'
        mapping['storwize_svc_iscsi_chap_enabled'] = 'chap_enabled'
        return mapping

    def _update_service_param_mapping(self, parm_mapping, connect_info):
        """Provides any additional Properties to set by this Registrar"""
        super(SVCStorageRegistrar, self).\
            _update_service_param_mapping(parm_mapping, connect_info)
        parm_mapping['volume_driver'] = \
            'powervc_cinder.volume.drivers.svc.'\
            'storwize_svc_product.StorwizeSVCProductDriver'
        if parm_mapping.get('san_ssh_port') is None:
            parm_mapping['san_ssh_port'] = '22'
        # If the Connection Protocol wasn't specified as Input, use Default
        if parm_mapping.get('storwize_svc_connection_protocol') is None:
            parm_mapping['storwize_svc_connection_protocol'] = \
                cfg.CONF.storwize_svc_connection_protocol
        if parm_mapping['storwize_svc_iscsi_chap_enabled'] is None:
            parm_mapping['storwize_svc_iscsi_chap_enabled'] = False

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_host_key(config)

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        # Run the lssystem command on the SVC to get the Unique Identifier
        output = utils.run_ssh_command(config, 'lssystem -delim :')
        # Confirm that the user specified has Admin privileges on the SVC
        self._verify_admin_privileges(config)
        # Parse the Unique Identifier out of the output returned from lssystem
        if output[0] and len(output[0]) > 0:
            for line in output[0].splitlines():
                columns = line.split(':')
                if len(columns) == 2 and columns[0] == 'id':
                    return columns[1]
        return None

    def _get_volume_pools(self, config):
        """Retrieves the Volume Pools for the specified System"""
        volume_pools = []
        # Run the lsmdiskgrp command on the SVC to get the Volume Pools
        output = utils.run_ssh_command(config, 'lsmdiskgrp -bytes -delim :')
        # Parse the Volume Pools out of the output returned from lsmdiskgrp
        if output[0] and len(output[0]) > 0:
            for line in output[0].splitlines():
                columns = line.split(':')
                if len(columns) >= 8 and columns[0] != 'id':
                    volume_pool = {'pool_name': columns[1]}
                    volume_pool['pool_total_capacity_gb'] = \
                        (float(columns[5]) / (1024 ** 3))
                    volume_pool['pool_free_capacity_gb'] = \
                        (float(columns[7]) / (1024 ** 3))
                    volume_pools.append(volume_pool)
        return volume_pools

    def _verify_admin_privileges(self, config):
        """Confirm that the user specified has Admin privileges on the SVC"""
        has_privileges = True
        try:
            # Run the lsuser and lsusergrp commands to determine the Role
            usr_output = utils.run_ssh_command(config, 'lsuser -delim :')
            grp_output = utils.run_ssh_command(config, 'lsusergrp -delim :')
            # Parse the output to get the User/Groups and Group/Roles
            usr_map = self._parse_command_output(usr_output, 1, 6, ':')
            grp_map = self._parse_command_output(grp_output, 1, 2, ':')
            role = grp_map.get(usr_map.get(config['user_id'], ''), '')
            # Verify the Role is Administrator or SecurityAdmin
            has_privileges = len(role) <= 0 or role.find('Admin') >= 0
        # If we got an exception, continue on since this is a usability check
        except Exception as exc:
            LOG.warn(_("Error trying to verify Admin privileges on SVC"))
            LOG.exception(exc)
        # If the user didn't have sufficient privileges, throw an exception
        if not has_privileges:
            self._throw_bad_request(msg.INVALID_SVC_PRIVILEGES)

    def _parse_command_output(self, cmd_output, key_idx, val_idx, delim):
        """Parses the Command Output for an SVC Command into a Dictionary"""
        cmd_dict = dict()
        first_line = True
        # Check to see if there was an command output returned to parse
        if cmd_output[0] and len(cmd_output[0]) > 0:
            for line in cmd_output[0].splitlines():
                columns = line.split(':')
                # If this is the Header Line, don't add it to the output
                if first_line:
                    first_line = False
                # Make sure that they gave us enough columns to parse
                elif len(columns) > val_idx:
                    cmd_dict[columns[key_idx]] = columns[val_idx]
        return cmd_dict


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return SVCStorageRegistrar(context, **kwargs)
