# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

""" XIV Registrar Class for performing Storage Registration"""

import base64

from xiv_ds8k_openstack.utils.powervc_helper import XIVPowerVCHelpers

from oslo_log import log as logging
from cinder import exception

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.storage import storage_registrar

__import__('powervc_oslo.xiv_xsf')
LOG = logging.getLogger(__name__)


########################################################
# ###########  XIV Registrar Implementation  ############
########################################################
class XIVStorageRegistrar(storage_registrar.StorageRegistrar):
    """XIV Registrar class to handle storage registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the XIV class for storage registration"""
        super(XIVStorageRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """ Provides the list of required attributes for this Registrar"""
        attr = ['host_type', 'access_ip', 'user_id', 'password',
                'volume_pool_name']
        return attr

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_XIV)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.AUTHENTICATION_ERROR, ['user credentials error']),
            (msg.CONNECTION_TIMEOUT, ['timed out', 'timeout exceeded']),
            (msg.CONNECTION_FAILED, ['unexpected network er', 'econnrefused']),
            (msg.INVALID_CERTIFICATE, ['certificate verify', "doesn't match"]),
            (msg.CERTIFICATE_EXPIRED, ['certificate has expired'])]

    def _get_service_param_mapping(self):
        """Provides a list of mappings between log errors and user messages"""
        mapping = super(XIVStorageRegistrar, self)._get_service_param_mapping()
        mapping['san_ip'] = 'access_ip'
        mapping['san_login'] = 'user_id'
        mapping['san_password'] = 'password'
        mapping['san_clustername'] = 'volume_pool_name'
        mapping['volume_backend_name'] = 'host_name'
        mapping['default_volume_type'] = 'default_volume_type'
        return mapping

    def _update_service_param_mapping(self, param_mapping, connect_info):
        """Provides any additional properties to set by this Registrar"""
        super(XIVStorageRegistrar, self).\
            _update_service_param_mapping(param_mapping, connect_info)
        param_mapping['volume_driver'] = \
            'powervc_cinder.volume.drivers.xiv.xiv_product.XIVProductDriver'
        param_mapping['xiv_ds8k_proxy'] = \
            'powervc_cinder.volume.drivers.xiv.xiv_product.XIVProductProxy'
        param_mapping['xiv_ds8k_connection_type'] = 'fibre_channel'

    def _get_certificate_port(self):
        """Provides the Port Number to use to retrieve the Certificate"""
        return 7778

    def _decode_password(self, password):
        """Shared Helper Method to specify the deCryption method to use"""
        return base64.b64decode(password)

    def _encode_password(self, password):
        """Shared Helper Method to specify the encryption method to use"""
        return base64.b64encode(password)

    #############################################
    #  Implementing StorageRegistrar Interface  #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_certificate(config, self._get_certificate_port())

    def _get_unique_identifier(self, config):
        """Retrieves the unique identifier for the given system."""
        self._verify_admin_priveleges(config)
        context = self.__get_context_for_xiv(config)
        return XIVPowerVCHelpers().get_unique_identifier(context)

    def _verify_admin_priveleges(self, config):
        """Confirm that the user specified has admin priveleges on the SVC"""
        context = self.__get_context_for_xiv(config)
        XIVPowerVCHelpers().verify_admin_privileges(context)

    def _get_volume_pools(self, config):
        """Retrieves the volume pools for the specified system"""
        pools = list()
        context = self.__get_context_for_xiv(config)
        helper_pools = XIVPowerVCHelpers().get_pools_list(context)
        # Loop through the Pools returns getting the key info out
        for pool in helper_pools.keys():
            volume_pool = dict(pool_name=helper_pools[pool]['name'])
            volume_pool['pool_total_capacity_gb'] = \
                helper_pools[pool]['soft_size']
            volume_pool['pool_free_capacity_gb'] = \
                helper_pools[pool]['free_size']
            pools.append(volume_pool)
        return pools

    ###############################################
    # ### Internal Registration Helper Methods  ###
    ###############################################
    def __get_context_for_xiv(self, config):
        """Helper method to convert the IP/User/Password for XIV"""
        return dict(address=config['access_ip'],
                    user=config['user_id'], password=config['password'])


###############################################
# ####  Temporary Workaround for XIV Bug  #####
###############################################
def _get_xiv_exception_class(self):
    """Temporary Workaround for XIV bug loading exception class"""
    return exception.CinderException

XIVPowerVCHelpers._get_exception = _get_xiv_exception_class


########################################################
# ########  Registrar Factory Implementation  ##########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return XIVStorageRegistrar(context, **kwargs)
