# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Brocade Registrar Classes for performing Fabric Registration"""

import re
from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.switch import fabric_registrar

LOG = logging.getLogger(__name__)


########################################################
# ########  Brocade Registrar Implementation  ##########
########################################################
class BrocadeFabricRegistrar(fabric_registrar.FabricRegistrar):
    """Brocade Registrar Class to handle Fabric Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the Brocade Class for Fabric Registration"""
        super(BrocadeFabricRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_BROCADE)

    def _get_registration_error_mappings(self):
        """Provides a list of Mappings between Log Errors and User Messages"""
        return [
            (msg.INVALID_SYSTEM, ['wwn: command not', 'wwn:  not found',
                                  "syntax error while parsing 'wwn'"])
        ] + utils.ssh_reg_messages()

    #############################################
    #  Implementing FabricRegistrar Interface   #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_host_key(config)

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        # Run the WWN command on the Switch to get the Unique Identifier
        return utils.run_ssh_command(config, 'wwn')[0].strip()

    def _verify_switch_config(self, config):
        """Verifies that the given Switch is suitable for our use"""
        required, found = (dict(), set())
        # Check the properties of the switch, making sure that we are
        # the principal switch and that zoning is enabled for the switch
        required['switchRole'] = {
            'regex': 'Principal', 'err': msg.SWITCH_IS_NOT_PRINCIPAL}
        required['zoning'] = {
            'regex': 'ON\s.*', 'err': msg.SWITCH_IS_NOT_ZONING}
        # Run the switchshow command and get the key properties to check
        res = utils.run_ssh_command(config, 'switchshow')
        for line in res[0].split('\n'):
            # Identify key/value pair lines, i.e. switchName: FC5022
            m = re.match("(\w+):\s+(.*)$", line)
            # If this is one of the keys that we are looking for, add it
            if m and m.groups()[0] in required:
                key, val = (m.groups()[0], m.groups()[1])
                # Throw an exception if the value is different than expected
                if not re.match(required[key]['regex'], val):
                    self._throw_bad_request(required[key]['err'], current=val)
                found.add(key)
        missing_keys = set(required.keys()) - found
        # If any keys were missing, throw an exception saying we can't validate
        if missing_keys:
            self._throw_bad_request(msg.SWITCH_PROPERTY_MISSING,
                                    missing=', '.join(missing_keys))

    def _get_static_config_params(self):
        """Provides any static CONF Properties to add on Register"""
        params = dict()
        params['zone_driver'] = 'cinder.zonemanager.'\
            'drivers.brocade.brcd_fc_zone_driver.BrcdFCZoneDriver'
        params['fc_san_lookup_service'] = 'powervc_cinder.zonemanager.'\
            'powervc_fc_san_lookup_service.PvcFCSanLookupService'
        params['brcd_sb_connector'] = 'cinder.zonemanager.drivers.'\
            'brocade.brcd_fc_zone_client_cli.BrcdFCZoneClientCLI'
        params['fc_fabric_type'] = 'brocade'
        return params


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return BrocadeFabricRegistrar(context, **kwargs)
