# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""CISCO Registrar Class for performing fabric registrations"""

from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.switch import fabric_registrar

LOG = logging.getLogger(__name__)


########################################################
# #########  CISCO Registrar Implementation  ###########
########################################################
class CiscoFabricRegistrar(fabric_registrar.FabricRegistrar):
    """Cisco Registrar Class to handle Fabric Registration"""

    def __init__(self, context, **kwargs):
        super(CiscoFabricRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def _get_system_type_string(self):
        """Provides the translated system type string for this registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_CISCO)

    def _get_registration_error_mappings(self):
        """Provides a list of mappings between log errors and user messages"""
        return [
            (msg.INVALID_SYSTEM, ['command not found', 'invalid command'])
        ] + utils.ssh_reg_messages()

    #############################################
    #  Implementing FabricRegistrar Interface   #
    #############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        return utils.verify_host_key(config)

    def _get_unique_identifier(self, config):
        """Retrieves the unique identifier for the given System"""
        return utils.run_ssh_command(config, 'show wwn switch')[0].strip()

    def _verify_switch_config(self, config):
        """Verifies that the given switch is suitable for our use"""
        pass

    def _get_static_config_params(self):
        """Provides any static CONF properties to add on register"""
        params = dict()
        params['zone_driver'] = 'cinder.zonemanager.'\
            'drivers.cisco.cisco_fc_zone_driver.CiscoFCZoneDriver'
        params['fc_san_lookup_service'] = 'powervc_cinder.zonemanager.'\
            'powervc_fc_san_lookup_service.PvcFCSanLookupService'
        params['cisco_sb_connector'] = 'cinder.zonemanager.drivers.'\
            'cisco.cisco_fc_zone_client_cli.CiscoFCZoneClientCLI'
        params['fc_fabric_type'] = 'cisco'
        return params


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper registrar"""
    return CiscoFabricRegistrar(context, **kwargs)
