# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Base Registrar Classes for performing Fabric Registration"""

import sys
import cgi
import ConfigParser
from oslo_config import cfg
from oslo_log import log as logging

from cinder import rpc as rpc_api

from powervc_discovery.registration import registrar as reg
from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
MAX_FABRICS = 26


########################################################
# #########  Fabric Registrar Implementation  ##########
########################################################
class FabricRegistrar(reg.ServiceRegistrar):
    """Abstract Base Registrar Class to handle Fabric Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for Abstract Class for Fabric Registration"""
        # First we need to read the Fabric Configuration File
        self.fabric_conf = self._read_fabric_config_file()
        # Then we can call the parent class to initialize the constructor
        super(FabricRegistrar, self).__init__(context, **kwargs)
        # Since didn't require the fabric_type set before, default to Brocade
        if not self.config.get('fabric_type'):
            self.config['fabric_type'] = 'brocade'

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['access_ip', 'user_id', 'password']

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        return ['password', 'unique_id']

    def get_immutable_attributes(self):
        """Provides the list of Attributes that can't be updated"""
        return ['fabric_name', 'fabric_type']

    def register(self):
        """Registrar Method to update the System Registration"""
        fabrics = self.__get_current_fabric_list()
        fabric_conf = self._get_config_file()
        fabric_name = self.config.get(self._get_name_attribute())
        fabric_name_l = fabric_name.lower()
        # Make sure that we aren't at the maximum number of registered Fabrics
        if len(fabrics) >= MAX_FABRICS:
            self._throw_bad_request(msg.FABRIC_LIMIT)
        # Make sure that the Fabric specified isn't already registered
        if fabric_name_l in fabrics:
            self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=fabric_name)
        fabrics.append(fabric_name_l)
        # Verify that another type of Fabric isn't already registered
        self.__verify_homogeneous_fabrics()
        # Verify that we can establish a connection to the specified Switch
        connect_info = self.establish_connection(self.config.copy())
        # If the Switch already exists in the CONF file, throw an exception
        unique_ids = self.__get_current_unique_ids()
        if connect_info['unique_id'] in unique_ids:
            display_name = unique_ids[connect_info['unique_id']]
            self._throw_bad_request(msg.SYSTEM_EXISTS, hostname=display_name)
        self.config['unique_id'] = connect_info['unique_id']
        # Verify that the switch is suitable.
        self._verify_switch_config(self.config.copy())
        # If they specified a Password, we want to encrypt the password
        self._obfuscate_password(self.config)
        # Send a notification that we are starting to add the Fabric
        self.__send_fabric_notification(self.config, 'create', 'start')
        # Add any of the Static Driver-specific properties to the CONF file
        switch_param = self._get_static_config_params()
        for key, val in switch_param.iteritems():
            utils.set_openstack_config_value(fabric_conf, key, val,
                                             run_as_root=True)
        # Add PowerVC Lookupservice to fc-zone-manager
        utils.set_openstack_config_value(fabric_conf, 'fc_san_lookup_service',
                                         switch_param['fc_san_lookup_service'],
                                         'fc-zone-manager',
                                         run_as_root=True)
        # Add the Fabric to the list of Fabrics in the Fabric CONF File
        for sect in ['DEFAULT', 'fc-zone-manager']:
            utils.set_openstack_config_value(fabric_conf, 'fc_fabric_names',
                                             ','.join(fabrics), sect,
                                             run_as_root=True)
        # Ensure that zoning is enabled
        utils.set_openstack_config_value(fabric_conf, 'zoning_mode', 'fabric',
                                         run_as_root=True)
        # Update each of the Configuration Properties for the given Fabric
        self._update_fabric_config_file(self.config)
        # Send a notification that we are done adding the Fabric
        self.__send_fabric_notification(self.config, 'create', 'end')
        # Restart all of the Cinder Volume Services when we are done
        self.__restart_cinder_volume_services()

    def deregister(self, force=False, uninstall=False):
        """Registrar Method to delete a System Registration"""
        fabrics = self.__get_current_fabric_list()
        regtype = self._get_registration_type_string()
        fabric_conf = self._get_config_file()
        fabric_name = self.config.get(self._get_name_attribute())
        fabrics.remove(fabric_name.lower())
        LOG.info(_("Removing %(regtype)s %(host)s...") %
                 dict(regtype=regtype, host=fabric_name))
        # Send a notification that we are starting to remove the Fabric
        self.__send_fabric_notification(self.config, 'delete', 'start')
        # Remove the Fabric from the list of Fabrics in the Fabric CONF File
        fabrics = ','.join(fabrics)
        for sect in ['DEFAULT', 'fc-zone-manager']:
            utils.set_openstack_config_value(fabric_conf, 'fc_fabric_names',
                                             fabrics, sect, run_as_root=True)
        # If we now have no fabrics, disable zoning
        if not fabrics:
            utils.set_openstack_config_value(fabric_conf, 'zoning_mode',
                                             'none', run_as_root=True)
        # Remove each of the Configuration Properties for the given Fabric
        for param in self._get_config_param_mappings().values():
            utils.del_openstack_config_value(
                fabric_conf, param + '_' + fabric_name.lower(),
                run_as_root=True)
        # We need to clean up the section that is specific for this Fabric
        self._delete_section_from_config_file(self.config['fabric_name'])
        # Send a notification that we are done removing the Fabric
        self.__send_fabric_notification(self.config, 'delete', 'end')
        # Restart all of the Cinder Volume Services when we are done
        self.__restart_cinder_volume_services()
        LOG.info(_("Removed %(regtype)s %(host)s.") %
                 dict(regtype=regtype, host=fabric_name))

    def update_registration(self, updated_values):
        """Registrar Method to update the System Registration"""
        # First make sure we can establish a connection with the new Config
        new_config = dict(self.config, **updated_values)
        connect_info = self.establish_connection(new_config)
        # Verify that the IP specified is not changing it to a different Switch
        self._verify_updating_same_system(connect_info, updated_values)
        # If they specified a Password, we want to encrypt the password
        self._obfuscate_password(updated_values)
        # Send a notification that we are starting to update the Fabric
        self.__send_fabric_notification(new_config, 'update', 'start')
        # Update each of the Configuration Properties for the given Fabric
        if self._update_fabric_config_file(updated_values):
            # Restart all of the Cinder Volume Services when we are done
            self.__restart_cinder_volume_services()
        # Send a notification that we are done updating the Fabric
        self.__send_fabric_notification(new_config, 'update', 'end')

    def _retrieve_registration(self):
        """Registrar method to retrieve the System Registration"""
        fabric_list = self.__get_current_fabric_list()
        fabric_name = self.config.get(self._get_name_attribute())
        fabric_name_l = fabric_name.lower()
        # If the Fabric wasn't found in the CONF File, then throw an exception
        if fabric_name_l not in fabric_list:
            self._throw_bad_request(msg.SYSTEM_NOTFOUND, hostname=fabric_name)
        # Loop through each of the Parameter Mappings, parsing from CONF file
        for key, val in self._get_config_param_mappings().iteritems():
            self.config[key] = self.fabric_conf.get(val + '_' + fabric_name_l)
        # We need to decrypt the password retrieved from the file
        self._obfuscate_password(self.config, decrypt=True)

    def _retrieve_connection_info(self, connect_info, config):
        """Retrieval Method to get the Systems's Connection Info"""
        # Initialize protocol necessary for connection
        self._initialize_connection(config)
        # Retrieve the Unique Identifier to use for the System
        connect_info['unique_id'] = self._get_unique_identifier(config)
        # If the Display Name wasn't provided, use the resolved DNS name
        if not self.config.get('fabric_display_name'):
            fabric_display_name = utils.formulate_hostname(config['access_ip'])
        else:
            fabric_display_name = self.config.get('fabric_display_name')
            fabric_display_name = cgi.escape(fabric_display_name)
        self.config['fabric_display_name'] = fabric_display_name

    def _get_registration_type_string(self):
        """Provides the translated Registration Type String for Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.REG_TYPE_FABRIC)

    def _get_registration_log_file(self):
        """Describes the name of the Log File running the Registration"""
        return '/var/log/cinder/api.log'

    def _get_name_attribute(self):
        """Provides the Name of the Attribute to use for the system name"""
        return 'fabric_name'

    def _generate_name_value(self):
        """Generates the Default Value to use for the Name Attribute"""
        fabric_names = map(chr, range(ord('a'), ord('z')+1))
        fabric_list = self.__get_current_fabric_list()
        for fabric_name in fabric_names:
            if fabric_name not in fabric_list:
                return fabric_name.upper()
        return ''

    ###############################################
    # ###  Defining FabricRegistrar Interface  ####
    ###############################################
    def _initialize_connection(self, config):
        """Initialize protocol necessary for connection"""
        pass

    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        raise NotImplementedError()

    def _verify_switch_config(self, config):
        """Verifies that the given Switch is suitable for our use"""
        raise NotImplementedError()

    def _get_static_config_params(self):
        """Provides any static CONF Properties to add on Register"""
        return dict()

    def _get_config_param_mappings(self):
        """Provides any static CONF Properties to add on Register"""
        param_mappings = {'fabric_display_name': 'fc_fabric_display_name'}
        param_mappings['access_ip'] = 'fc_fabric_address'
        param_mappings['user_id'] = 'fc_fabric_user'
        param_mappings['password'] = 'fc_fabric_password'
        param_mappings['unique_id'] = 'fc_fabric_uniqueid'
        param_mappings['fabric_type'] = 'fabric_type'
        param_mappings['port'] = 'fc_fabric_port'
        param_mappings['vsan'] = 'fc_fabric_vsan'
        return param_mappings

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _get_config_file(self):
        """Provides the Name of the Fabric Configuration File"""
        return '/etc/cinder/fabrics.conf'

    def __get_current_fabric_list(self):
        """Helper method to get the list of existing registered Fabrics"""
        fabric_list = self.fabric_conf.get('fc_fabric_names', '').strip()
        return [] if len(fabric_list) <= 0 else fabric_list.split(',')

    def __get_current_unique_ids(self):
        """Helper method to get the list of existing registered Fabrics"""
        unique_ids = dict()
        fabric_list = self.__get_current_fabric_list()
        # Loop through each of the registered Fabrics, finding its unique id
        for fabric in fabric_list:
            unique_id = self.fabric_conf.get('fc_fabric_uniqueid_' + fabric)
            dspname = self.fabric_conf.get('fc_fabric_display_name_' + fabric)
            # If for some reason it isn't in the CONF file, just skip it
            if unique_id is not None:
                unique_ids[unique_id] = dspname if dspname is not None else ''
        return unique_ids

    def __verify_homogeneous_fabrics(self):
        """Helper method to disallow mixed types of fabrics"""
        fabtypN = self.config.get('fabric_type', 'brocade')
        # Loop through the existing fabrics seeing if are of a different type
        # We will allow mixed as long as one of them isn't the generic one
        for fabname in self.__get_current_fabric_list():
            fabtypO = self.fabric_conf.get('fabric_type_' + fabname, 'brocade')
            # If one of the Fabrics is Generic, then throw exception
            if 'generic' in [fabtypN, fabtypO] and fabtypN != fabtypO:
                registrar = get_registrar(self.context, fabric_name=fabname)
                systype = registrar._get_system_type_string()
                self._throw_bad_request(msg.TWO_FABRIC_TYPES, systype2=systype)

    def __send_fabric_notification(self, config, event_type, step):
        """Helper method to send a notification when the Fabric changed"""
        anotifier = rpc_api.get_notifier(publisher_id=cfg.CONF.volume_topic)
        attrs = ['fabric_name', 'fabric_display_name',
                 'fabric_type', 'access_ip', 'user_id']
        payload = dict([(k, config.get(k)) for k in attrs])
        anotifier.info(self.context,
                       'san.fabric.' + event_type + '.' + step, payload)

    def _update_fabric_config_file(self, config):
        """Helper method to update the entries in the Fabric CONF file"""
        changes = False
        fabric_conf = self._get_config_file()
        fabric_name = self.config.get(self._get_name_attribute())
        # Loop through each of the Parameter Mappings, updating the CONF file
        for key, val in self._get_config_param_mappings().iteritems():
            param = val + '_' + fabric_name.lower()
            value = config.get(key)
            if value is not None:
                utils.set_openstack_config_value(fabric_conf, param, value,
                                                 run_as_root=True)
                # We need to duplicate the property for the Community drivers
                utils.set_openstack_config_value(fabric_conf, val, value,
                                                 fabric_name.lower(),
                                                 run_as_root=True)
                changes = True
        return changes

    @staticmethod
    def _read_fabric_config_file():
        """Helper method to read the entries from the Fabric CONF file"""
        return utils.read_config_file('/etc/cinder/fabrics.conf')

    @staticmethod
    def _delete_section_from_config_file(section):
        """Helper method to delete this fabrics section from the CONF file"""
        section = section.lower()
        parser = ConfigParser.RawConfigParser()
        parser.read('/etc/cinder/fabrics.conf')
        if parser.has_section(section):
            for key in parser._sections[section]:
                if key != '__name__':
                    utils.del_openstack_config_value(
                        '/etc/cinder/fabrics.conf', key, section,
                        run_as_root=True)

    @staticmethod
    def __restart_cinder_volume_services():
        """Helper method to restart an Cinder Volume Services registered"""
        filenames = utils.get_matching_files('/etc/cinder', 'cinder-', '.conf')
        # Loop through each of the Cinder Volume Services, restarting each one
        for filename in filenames:
            svcname = 'openstack-cinder-volume-' + \
                filename[7:len(filename) - 5]
            # use relative path on cmd for rootwrap's sake
            utils.run_command('service ' + svcname + ' restart')


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    throw_exc = kwargs.pop('throw_exc', True)
    try:
        # If this is not a register, we need to figure out the type
        tempreg = FabricRegistrar(context, **kwargs)
        # Default the Fabric Type to Brocade since it is the only one
        fabric_type = tempreg.config.get('fabric_type', 'brocade')
        try:
            # Attempt to dynamically load the corresponding module for the type
            regmodule = __package__ + '.' + str(fabric_type) + '_registrar'
            __import__(regmodule)
            return getattr(sys.modules[regmodule],
                           'get_registrar')(context, **kwargs)
        except Exception as exc:
            LOG.exception(exc)
            tempreg._throw_bad_request(msg.INVALID_HOST_TYPE,
                                       hosttype=fabric_type)
    except:
        # Only throw the exception if we were asked to throw it on error
        if throw_exc:
            raise
    return None


def list_fabric_registrars(context):
    """Factory Method to retrieve the list of all of the Registrars"""
    registrars = []
    config = FabricRegistrar._read_fabric_config_file()
    fabric_list = config.get('fc_fabric_names', '').strip()
    fabric_list = [] if len(fabric_list) <= 0 else fabric_list.split(',')
    for fabric in fabric_list:
        registrar = get_registrar(context, fabric_name=fabric.upper())
        registrars.append(registrar)
    return registrars
