# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2015 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""An API Extension for being able to Register a new Storage Host"""

from oslo_log import log as logging
from cinder.api import extensions
from cinder.api.openstack import wsgi

from powervc_cinder.fabric.map_initiators import MapInitiators
from powervc_discovery.registration.switch import fabric_registrar
from powervc_discovery import locale as discovery_locale
from powervc_discovery.locale import _

LOG = logging.getLogger(__name__)
AUTHORIZE = extensions.extension_authorizer('volume', 'fabric_registration')


class FabricRegistrationControllerExtension(wsgi.Controller):
    """Main Controller Class to extend the os-hosts API for Registration"""

    def index(self, req):
        """Implement Rest GET Method to return a list of Fabrics"""
        fabrics = []
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'index')
        # Get the list of registered Fabrics to return basic info about them
        registrars = fabric_registrar.list_fabric_registrars(context)
        for registrar in registrars:
            name = registrar.config['fabric_name']
            dspname = registrar.config['fabric_display_name']
            fabrics.append(dict(fabric_name=name, fabric_display_name=dspname))
        return {'fabrics': fabrics}

    def show(self, req, id):
        """Implement Rest GET Method to return details of a specific Fabric"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'show')
        # Load the appropriate Registrar to Retrieve the Registration Info
        registrar = fabric_registrar.get_registrar(context, fabric_name=id)
        registration = self.filter_config(registrar)
        return {'fabric': {'registration': registration}}

    def detail(self, req):
        """Implement Rest GET Method to return a list of Fabric details"""
        fabrics = []
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'detail')
        # Get the list of registered Fabrics to return basic info about them
        registrars = fabric_registrar.list_fabric_registrars(context)
        for registrar in registrars:
            fabrics.append(self.filter_config(registrar))
        return {'fabrics': fabrics}

    @wsgi.action("create")
    def create(self, req, body):
        """Add Rest POST Method to Create a Fabric Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'create')
        registration = body['fabric']['registration']
        # Load the appropriate Registrar to do the Fabric Registration
        registrar = fabric_registrar.get_registrar(context, create=True,
                                                   **registration)
        # Delegate to the Registrar to do the Fabric Registration
        registrar.register()
        # Return on the Response the Registration Info
        return {'fabric': {'registration': self.filter_config(registrar)}}

    @wsgi.action("update")
    def update(self, req, id, body):
        """Add Rest PUT Method to Update a Fabric Registration"""
        fabric = {'fabric': id}
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'update')
        registration = body.pop('registration', None)
        # If no Registration Info is provided, then updating a base attribute
        if registration is not None:
            # Load the appropriate Registrar to do the Fabric Re-Registration
            registrar = fabric_registrar.get_registrar(context, fabric_name=id)
            # Delegate to the Registrar to do the Fabric Re-Registration
            registrar.update_registration(registration)
            fabric['registration'] = self.filter_config(registrar,
                                                        config=registration)
        return fabric

    @wsgi.action("delete")
    @wsgi.response(204)
    def delete(self, req, id):
        """Add Rest DELETE Method to Remove a Fabric Registration"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'delete')
        # Load the appropriate Registrar to do the Fabric De-Registration
        registrar = fabric_registrar.get_registrar(context, fabric_name=id)
        # Delegate to the Registrar to do the Fabric De-Registration
        registrar.deregister()

    @staticmethod
    def _validate_authorization(context, req, action):
        """Internal Helper Method to Confirm the Requester is Authorized"""
        AUTHORIZE(context, action=action)
        return discovery_locale.populate_locale(context.elevated(), req)

    @staticmethod
    def filter_config(registrar, config=None):
        """Internal Helper Method to Remove any Sensitive Attributes"""
        new_config = registrar.config.copy()
        if config is not None:
            new_config = config.copy()
        # Loop through each of the Sensitive Attributes, removing them
        for attr in registrar.get_sensitive_attributes():
            new_config.pop(attr, None)
        return new_config

    @wsgi.action("map_ports")
    def map_ports(self, req, body):
        """Attempt to match FC port WWPNs with registered fabrics"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'map_ports')
        LOG.info(_("map_ports request with body: %s") % body)
        mapper = MapInitiators()
        mapping = mapper.get_wwpn_fabric_map(body['wwpn_list'])
        LOG.debug("RETURNS %s" % mapping)
        return {'wwpn_mapping': mapping}

    @wsgi.action("map_zones")
    def map_zones(self, req, body):
        """Attempt to match zones' wwpns with registered fabrics"""
        context = req.environ['cinder.context']
        context = self._validate_authorization(context, req, 'map_zones')
        LOG.info(_("map_zones request with body: %s") % body)
        mapper = MapInitiators()
        initiators = body['initiators'] if body['initiators'] else None
        targets = body['targets'] if body['targets'] else None
        zones = mapper.get_zones_fabric_map(initiators,
                                            targets)
        LOG.debug("RETURNS Zones: %s" % zones)
        return zones


class FabricRegistration(extensions.ExtensionDescriptor):
    """Capability to Register a new SAN Fabric."""
    name = "SAN Fabric Registration"
    alias = "ibm-san-fabric-registration"
    namespace = "http://docs.openstack.org/volume/ext/" + \
                "ibm_san_fabric_registration/api/v1.1"
    updated = "2013-08-25T21:00:00-06:00"

    def get_resources(self):
        """Provide a Sub-Resource Extension to the san-fabrics Resource"""
        resource = extensions.ResourceExtension(
            'san-fabrics', controller=FabricRegistrationControllerExtension(),
            collection_actions={'detail': 'GET', 'update': 'PUT',
                                'map_ports': 'POST', 'map_zones': 'POST'})
        return [resource]
