# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2014 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

"""Generic Registrar Classes for performing Fabric Registration"""

from oslo_log import log as logging

from powervc_discovery.registration import registrar_messages as msg
from powervc_discovery.registration import registrar_utilities as utils
from powervc_discovery.registration.switch import fabric_registrar

LOG = logging.getLogger(__name__)


########################################################
# #########  Generic Registrar Implementation  #########
########################################################
class GenericFabricRegistrar(fabric_registrar.FabricRegistrar):
    """Generic Registrar Class to handle Fabric Registration"""

    def __init__(self, context, **kwargs):
        """Constructor for the Generic Class for Fabric Registration"""
        super(GenericFabricRegistrar, self).__init__(context, **kwargs)

    ###############################################
    # ### Implementing BaseRegistrar Interface ####
    ###############################################
    def get_required_attributes(self):
        """Provides the list of Required Attributes for this Registrar"""
        return ['fabric_type', 'fabric_display_name',
                'zone_driver', 'configuration_properties']

    def get_sensitive_attributes(self):
        """Provides the list of Sensitive Attributes for this Registrar"""
        attrs = super(GenericFabricRegistrar, self).get_sensitive_attributes()
        attrs.append('configuration_properties')
        attrs.append('access_ip')
        attrs.append('user_id')
        return attrs

    def _get_system_type_string(self):
        """Provides the translated System Type String for this Registrar"""
        msgb = self._get_message_bundle()
        return msgb.get_message_text(self.context, msg.SYSTEM_TYPE_GENERIC_FAB)

    def _retrieve_registration(self):
        """Registrar method to retrieve the System Registration"""
        super(GenericFabricRegistrar, self)._retrieve_registration()
        self.config['configuration_properties'] = dict()
        # Retrieve the Zone Driver property from the Configuration File
        self.config['zone_driver'] = self.fabric_conf.get('zone_driver')

    #############################################
    #  Implementing FabricRegistrar Interface   #
    #############################################
    def _get_unique_identifier(self, config):
        """Retrieves the Unique Identifier for the given System"""
        return None

    def _verify_switch_config(self, config):
        """Verifies that the given Switch is suitable for our use"""
        pass

    def _get_static_config_params(self):
        """Provides any static CONF Properties to add on Register"""
        params = dict()
        # If they gave us a Driver, specify that in the Configuration File
        if self.config.get('zone_driver'):
            params['zone_driver'] = self.config['zone_driver']
        # If they gave additional properties, put those in the Configuration
        if self.config.get('configuration_properties'):
            cfg_props = self.config['configuration_properties']
            for key, val in cfg_props.iteritems():
                # We will map the instance-specific values in a different
                # method
                if key not in self._get_config_param_mappings().values():
                    params[key] = val
        return params

    ###############################################
    # ### Internal Registration Helper Methods ####
    ###############################################
    def _update_fabric_config_file(self, config):
        """Helper method to update the entries in the Fabric CONF file"""
        fabric_conf = self._get_config_file()
        config_props = config.pop('configuration_properties', None)
        config_props = config_props if config_props else dict()
        # Next call the super-class to actually update the fabrics.conf
        changes = super(GenericFabricRegistrar,
                        self)._update_fabric_config_file(config)
        # We need to explicitly update the Configuration Properties ourselves
        for key, val in config_props.iteritems():
            key = self.__format_section_in_key(key)
            utils.set_openstack_config_value(fabric_conf, key, val)
        return changes or config_props

    def __format_section_in_key(self, key):
        """Helper method to put the key in the right section in the file"""
        keyl, keyt = (key.lower(), key[key.find(':') + 1:])
        sect_specific = ['fc_fabric_address', 'fc_fabric_user',
                         'fc_fabric_password', 'fc_fabric_port']
        # First check if they tried to pass in the section on the key itself
        if keyl[0].islower() and keyl[1] == ':':
            return self.config['fabric_name'].lower() + ':' + keyt
        # Now lets see if it is one of the predefined section-specific keys
        for speckey in sect_specific:
            if keyl == speckey:
                return self.config['fabric_name'].lower() + ':' + key
            if keyl.endswith(':' + speckey):
                return self.config['fabric_name'].lower() + ':' + keyt
        return key


########################################################
# #########  Registrar Factory Implementation  #########
########################################################
def get_registrar(context, **kwargs):
    """Factory Method to retrieve an instance of the proper Registrar"""
    return GenericFabricRegistrar(context, **kwargs)
